#!/bin/bash

# This script looks for a GMP installation in a standard location.
# If a single GMP installation is found, it prints out the full path of
# the (static) library and returns with exit code 0.
# If none is found or several are found, it prints out an error message
# and returns with a non-zero exit code.

##################################################################
# Use find to search through various standard directories.
# NB look through all directories, even if a GMP has already been found.

# List of directories under which libgmp.a and/or libgmp.so is normally found.
STD_GMP_LIBDIRS="/usr/lib64:/usr/local:/usr/lib:/opt/local/lib:/sw/lib:/usr/sfw/lib"
IFS=":"
LIBGMPPATHS=libgmp-paths
/bin/rm -rf $LIBGMPPATHS
for directory in $STD_GMP_LIBDIRS
do
  if [ -d "$directory" ]
  then
    find "$directory" -name  libgmp.a   -print >> $LIBGMPPATHS  2> /dev/null
    find "$directory" -name  libgmp.so  -print >> $LIBGMPPATHS  2> /dev/null
  fi
done

if [ \! -s $LIBGMPPATHS ]
then
  # Did not find any plausible GMP installation, so return empty handed.
  echo "No GMP installation found; looked inside $STD_GMP_LIBDIRS"
  /bin/rm $LIBGMPPATHS
  exit 1
fi

# Slightly odd call to wc is to avoid it printing out the file name.
if [ `wc -l < $LIBGMPPATHS` -ne 1 ]
then
  echo "*** Found multiple GMP libraries ***"
  cat $LIBGMPPATHS
  /bin/rm $LIBGMPPATHS
  exit 2
fi

# We have found a single file called libgmp.a; do a couple of quick
# sanity checks before declaring our search successful...
GMP_LIB=`cat $LIBGMPPATHS`
/bin/rm $LIBGMPPATHS
GMP_DIR=`dirname "$GMP_LIB"`
# Special handling if libgmp.a is not fully installed...
if [ `basename "$GMP_DIR"` = ".libs" ]
then
  GMP_HDR="$GMP_DIR/../gmp.h"
else
  GMP_HDR="$GMP_DIR/../include/gmp.h"
fi
if [ -f "$GMP_LIB" -a -r "$GMP_LIB" -a -f "$GMP_HDR" -a -r "$GMP_HDR" ]
then
  # We've found a plausible libgmp.a and (probably) its corresponding gmp.h.
  echo "$GMP_LIB"
  exit 0
fi

# Get here probably only if there's a damaged GMP installation.
echo "Trouble reading file $GMP_LIB or its header file $GMP_HDR."
exit 3
