// Copyright (c) 2005 Anna Bigatti
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Program showing how to compute the Alexander Dual of a mon'l sqfr ideal.\n";

const string LongDescription =
  "It offers the function \"PrimaryDecomposition\" for a mon'l sqfr ideal\n"
  "which will be included in CoCoALib (sooner or later).\n";
//----------------------------------------------------------------------


bool IsMonomialGend(const ideal& I)
{
  const vector<RingElem>& G = gens(I);
  for (vector<RingElem>::const_iterator it=G.begin(); it!=G.end() ; ++it)
    if ( !IsMonomial(*it) ) return false;
  return true;
}

// ??? copied from GReductor.C
bool BoolCmpLPP(ConstRefRingElem f, ConstRefRingElem g)
{
  CoCoA_ASSERT(owner(f)==owner(g));//BUG HUNTING  ???
  CoCoA_ASSERT(!IsZero(f));//BUG HUNTING  ???
  CoCoA_ASSERT(!IsZero(g));//BUG HUNTING  ???
  
  return AsSparsePolyRing(owner(f))->myCmpLPP(raw(f), raw(g)) > 0;
}


// ??? copied from ex-QuotientBasis.C
bool IsDivisible(ConstRefPPMonoidElem pp, const vector<PPMonoidElem>& ByL)
{
  for (size_t i=0; i < ByL.size(); ++i)
    if ( IsDivisible(pp, ByL[i])) return true;
  return false;
}


// ??? it would be handy to make an ideal from a vector of PPMonoidElem
// ??? what's the name for const_reverse_iterator???
ideal interreduce(const ideal& I)
{
  SparsePolyRing P = AsSparsePolyRing(AmbientRing(I));
  vector<RingElem> g = gens(I);
  vector<PPMonoidElem> p;
  sort(g.begin(), g.end(), BoolCmpLPP);
  for (vector<RingElem>::reverse_iterator it=g.rbegin(); it!=g.rend() ; ++it)
    if ( !IsDivisible(LPP(*it), p) )
      back_inserter(p) = LPP(*it);
  g.clear();
  for (vector<PPMonoidElem>::const_iterator it=p.begin(); it!=p.end() ; ++it)
    back_inserter(g) = monomial(P, 1, *it);
  return ideal(P, g);
}


// ??? should become intersect(MonIdeal J1, MonIdeal J2)
ideal MonIdIntersect(const ideal& J1, const ideal& J2)
{
  if ( AmbientRing(J1) != AmbientRing(J2) )
    CoCoA_ERROR(ERR::MixedRings, "MonIdIntersect");
  if (!(IsMonomialGend(J1) && IsMonomialGend(J2)))
    CoCoA_ERROR("only for monomial ideals", "intersect");
  SparsePolyRing P = AsSparsePolyRing(AmbientRing(J1));
  const vector<RingElem>& g1 = gens(J1), g2 = gens(J2);
  vector<RingElem> g;
  for (vector<RingElem>::const_iterator i1=g1.begin(); i1!=g1.end() ; ++i1)
    for (vector<RingElem>::const_iterator i2=g2.begin(); i2!=g2.end() ; ++i2)
      back_inserter(g) = monomial(P, 1, lcm(LPP(*i1), LPP(*i2)));
  return interreduce(ideal(P, g));
}


// ??? should become product(MonIdeal J1, MonIdeal J2)
ideal MonIdProduct(const ideal& J1, const ideal& J2)
{
  if ( AmbientRing(J1) != AmbientRing(J2) )
    CoCoA_ERROR(ERR::MixedRings, "MonIdProduct");
  if (!(IsMonomialGend(J1) && IsMonomialGend(J2)))
    CoCoA_ERROR("only for monomial ideals", "product");
  SparsePolyRing P = AsSparsePolyRing(AmbientRing(J1));
  const vector<RingElem>& g1 = gens(J1), g2 = gens(J2);
  vector<RingElem> g;
  for (vector<RingElem>::const_iterator i1=g1.begin(); i1!=g1.end() ; ++i1)
    for (vector<RingElem>::const_iterator i2=g2.begin(); i2!=g2.end() ; ++i2)
      back_inserter(g) = monomial(P, 1, LPP(*i1) * LPP(*i2));
  //  return interreduce(ideal(P, g));
  return ideal(P, g);
}


// ??? sum of monomial ideals should not be special

ideal AlexanderDual(const PolyRing& P, ConstRefPPMonoidElem pp)
{
  /*
    L := Log(X);
    Return Ideal([ Indet(I) | I In 1..NumIndets() And L[I]<>0 ]);
  */
  vector<RingElem> g;
  for (size_t i=0 ; i<NumIndets(owner(pp)) ; ++i )
    if ( exponent(pp, i) != 0 )
    {
      if ( exponent(pp, i) != 1 ) CoCoA_ERROR("input must be square free", "AlexanderDual");
      back_inserter(g) = indet(P, i);
    }
  return ideal(P, g);
}


ideal AlexanderDual(const ideal& I)
{
  //--  Return IntersectionList([$.AlexanderDual(G) | G In Gens(X)]); 
  SparsePolyRing P = AsSparsePolyRing(AmbientRing(I));
  ideal AD(P, 1);
  const vector<RingElem>& g = gens(I);
  for (vector<RingElem>::const_iterator it=g.begin(); it!=g.end() ; ++it)
    AD = MonIdIntersect(AD, AlexanderDual(P, LPP(*it)));
  
  return AD;
}


//vector<ideal> PrimaryDecomposition(SqFrMonIdeal I)
vector<ideal> PrimaryDecomposition(const ideal& I)
{
  //--  Return [$.AlexanderDual(G) | G In Gens(AD)];
  SparsePolyRing P = AsSparsePolyRing(AmbientRing(I));
  vector<RingElem> g = gens(AlexanderDual(I));
  vector<ideal> PD;
  for (vector<RingElem>::const_iterator it=g.begin(); it!=g.end() ; ++it)
    PD.push_back(AlexanderDual(P, LPP(*it)));
  return PD;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  ring Fp = NewZmod(32003);          // coefficient ring
  SparsePolyRing Fpx = NewPolyRing(Fp, 8); // Fp[x[0..7]]
  SparsePolyRing P = Fpx;
  double t0;  // for CpuTime
  
  const vector<RingElem>& x = indets(P);
  vector<RingElem> g;
  back_inserter(g) = power(x[2],2) * power(x[5],4);
  back_inserter(g) = power(x[1],3) * power(x[4],4);
  back_inserter(g) = power(x[1],3) * power(x[5],4);
  back_inserter(g) = power(x[3],3) * power(x[6],4);
  back_inserter(g) = power(x[3],4) * power(x[6],3);
  vector<RingElem> f;
  back_inserter(f) = x[1]*x[2];
  back_inserter(f) = x[2]*x[3];

  ideal J1(P, g), J2(P, f);
  GlobalOutput() << "J1  = " << J1 << endl; 
  GlobalOutput() << "J2  = " << J2 << endl << endl;

  t0 = CpuTime();
  ideal I = MonIdIntersect(J1, J2);
  GlobalOutput() << "Cpu Time = " << CpuTime()-t0 << endl;
  GlobalOutput() << "MonIdIntersect(J1, J2) = " << I << endl;
  GlobalOutput() << endl;

  t0 = CpuTime();
  vector<ideal> PrimDec = PrimaryDecomposition(J2);
  GlobalOutput() << "Cpu Time = " << CpuTime()-t0 << endl;
  GlobalOutput() << "PrimaryDecomposition(J2) = " << PrimDec << endl;
  //  copy(PrimDec.begin(), PrimDec.end(), ostream_iterator<ideal>(GlobalOutput(), ", "));
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-AlexanderDual.C,v 1.2 2007/12/04 14:27:07 bigatti Exp $
// $Log: ex-AlexanderDual.C,v $
// Revision 1.2  2007/12/04 14:27:07  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.8  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/02/28 13:51:59  bigatti
// -- added function IsMonomial
//
// Revision 1.6  2007/02/26 15:49:08  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.5  2007/02/12 15:29:07  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.4  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.3  2006/11/21 10:57:35  cocoa
// -- cleaned up ex-AlexanderDual.C and added to Makefile
//
// Revision 1.2  2006/11/17 18:14:55  cocoa
// -- added const & to some arguments (embarrassing!)
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1  2006/04/28 11:26:06  cocoa
// -- first import
//
