// Copyright (c) 2006  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "A short example showing how to use the approximate point preprocessing algorithms.\n"
  "See the file ex-ApproxPt1.in for a sample input.\n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
//#include <iostream> // using std::ostream; using std::endl;


void ConvertToRingElem(RefRingElem x, double z)
{
  QQ Q;
  convert(Q, z);
  x = Q;
}

void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;
  
  GlobalOutput() << "Insert the space dimension: " << endl;
  size_t dim;
  GlobalInput() >> dim;
  if ( !GlobalInput() )
    CoCoA_ERROR("Input must be a positive integer", "program");
  if (dim < 1 || dim > 1000000)
    CoCoA_ERROR("Ridiculous input dimension", "main program in ex-ApproxPts1");
  
  ring Q = RingQ();
  vector<RingElem> tolerance(dim, zero(Q));
  GlobalOutput() << "Insert the tolerance in each dimension: " << endl;
  for (size_t i=0; i < dim; ++i)
  {
    double compt;
    GlobalInput() >> compt;
    if (!GlobalInput() || compt < 0) { CoCoA_ERROR("bad input", "main program in ex-ApproxPts1"); }
    ConvertToRingElem(tolerance[i], compt);
  }

  GlobalOutput() << "Insert the number of points to be preprocessed: " << endl;
  size_t NumPts;
  GlobalInput() >> NumPts;
  if (NumPts < 1 || NumPts > 1000000)
    CoCoA_ERROR("Ridiculous number of points", "main program in ex-ApproxPts1");

  vector<ApproxPts::PointR> OrigPts(NumPts, ApproxPts::PointR(dim, zero(Q)));
  GlobalOutput() << "Insert the coordinates of the points " << endl;
  for (size_t i=0; i < NumPts; ++i)
    {
      for (size_t j=0; j < dim; ++j)
	{
          double compt;
	  GlobalInput() >> compt;
	  if (!GlobalInput()) { CoCoA_ERROR("bad input", "main program in ex-ApproxPts1"); }
          ConvertToRingElem(OrigPts[i][j], compt);
	}
    }

  GlobalOutput() << endl
                 << "Read " << OrigPts.size() << " original points." << endl
                 << endl;

  double t0;
  GlobalOutput() << "-------------------------------------------------------" << endl;
  vector<ApproxPts::PointR> PreprocessedPts;
  vector<size_t> weights;
  t0 = CpuTime();
  ApproxPts::PreprocessGridAlgm(PreprocessedPts, weights, OrigPts, tolerance);
  t0 = CpuTime()-t0;
  GlobalOutput() << "Grid algm produces " << PreprocessedPts.size() << " preprocessed points." << endl
                 << endl
                 << "The preprocessed points are: " << PreprocessedPts << endl
                 << "and their weights are: " << weights << endl
                 << endl
                 << "CPU time for grid algm: " << t0 << endl;

  GlobalOutput() << "-------------------------------------------------------" << endl;
  t0 = CpuTime();
  ApproxPts::PreprocessAggrAlgm(PreprocessedPts, weights, OrigPts, tolerance);
  t0 = CpuTime()-t0;
  GlobalOutput() << "Aggr algm produces " << PreprocessedPts.size() << " preprocessed points." << endl
                 << endl
                 << "The preprocessed points are: " << PreprocessedPts << endl
                 << "and their weights are: " << weights << endl
                 << endl
                 << "CPU time for aggr algm: " << t0 << endl;

  GlobalOutput() << "-------------------------------------------------------" << endl;
  t0 = CpuTime();
  ApproxPts::PreprocessSubdivAlgm(PreprocessedPts, weights, OrigPts, tolerance);
  t0 = CpuTime()-t0;
  GlobalOutput() << "Subdiv algm produces " << PreprocessedPts.size() << " preprocessed points." << endl
                 << endl
                 << "The preprocessed points are: " << PreprocessedPts << endl
                 << "and their weights are: " << weights << endl
                 << endl
                 << "CPU time for subdiv algm: " << t0 << endl;
}

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-ApproxPts1.C,v 1.4 2009/12/11 11:46:32 abbott Exp $
// $Log: ex-ApproxPts1.C,v $
// Revision 1.4  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.3  2008/11/23 18:23:22  abbott
// Now uses points with coords in RingQ rather than represedented as C++ doubles.
//
// Revision 1.2  2008/11/21 21:12:19  abbott
// Update inline with the changes to the preprocessing code.
// Changed semiwidth into tolerance.
// Program now prints out the weights too.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.4  2007/02/12 15:29:07  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.3  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.2  2006/06/21 17:05:47  cocoa
// Major overhaul of approx point preprocessing algms.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.3  2006/05/29 16:17:10  cocoa
// Changed a variable name.
//
// Revision 1.2  2006/05/22 15:52:16  cocoa
// Added preprocess-disg algorithm to ApproxPts.
// Sundry minor improvements.
//
// Revision 1.1  2006/05/12 13:16:30  cocoa
// Added functions for preprocessing approximate points.
//
