// Copyright (c) 2005  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Program to find lengths of 3n+1 sequences. (uses CoCoA::GMPAllocator) \n";

const string LongDescription =
  "This example shows how a CoCoA::GMPAllocator can be used.  If you do     \n"
  "computations involving many small big-integers (up to about 40 decimal   \n"
  "digits) then the custom allocator may give slightly better run-time      \n"
  "performance -- you must compile with debugging turned off!               \n"
  "The crucial command is the very first line of \"program\" just comment it \n"
  "out and recompile to use the default system allocator for GMP values.    \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


// This function counts the number of iterations of the 3n+1 sequence
// are needed to reach 1 (the first time) starting from N
size_t NumIters(ZZ N)
{
  N = abs(N); // ignore sign of N
  size_t iters = 0;
  while (N > 1)
  {
    if (IsEven(N)) N /= 2;
    else N = 3*N+1;
    ++iters;
  }
  return iters;
}


void program()
{
  CoCoA::GMPAllocator sentinel; // use CoCoALib's custom GMP allocator
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  ZZ Nmax;
  GlobalOutput() << "Enter highest starting value to try (e.g. 1000000): ";
  GlobalInput() >> Nmax;
  size_t MaxIters = 0;

  for (ZZ N(3); N <= Nmax; N += 2)
  {
    size_t iters = NumIters(N);
    if (iters > MaxIters)
    {
      MaxIters = iters;
      GlobalOutput() << "The sequence starting from " << N << " has length " << MaxIters << endl;
    }
  }
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-GMPAllocator.C,v 1.1.1.1 2007/03/09 15:16:11 abbott Exp $
// $Log: ex-GMPAllocator.C,v $
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2007/03/07 15:45:42  cocoa
// GMPAllocator must be created before GlobalManager, otherwise bus error.
//
// Revision 1.4  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.3  2007/02/12 15:29:07  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.1  2005/04/29 15:42:02  cocoa
// Improved documentation for GMPAllocator.
// Added example program for GMPAllocator.
// Added example program for simple ops on polynomials.
// Added two new ctors for (principal) ideals (from long, and from ZZ).
// Added (crude) printing for PPMonoids.
// Updated library.H (#included GMPAllocator.H).
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.3  2004/12/09 15:08:42  cocoa
// -- added log info
//
