// Copyright (c) 2006  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "This program illustrates the use of some basic number theoretic functions.  \n";

const string LongDescription =
  "This programs show how to use some of the basic number theoretic functions.  \n"
  "Many of the examples use machine integers for convenience, but the functions \n"
  "work equally well with ZZ values (except NextPrime and PrevPrime).           \n";

//----------------------------------------------------------------------

void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  const int n = 123456789;
  const int m = 987654321;
  GlobalOutput() << "Some example compuations with m=" << m << "  and n=" << n << endl;
  GlobalOutput() << "gcd(m,n) = " << gcd(m,n) << endl;
  GlobalOutput() << "gcd(-m,-n) = " << gcd(-m,-n) << "    **result is always non-negative**" << endl;
  // Compute the cofactors for m and n: note that a & b must be of type long!
  long a,b;
  ExtGcd(a,b,m,n);
  GlobalOutput() << "Cofactors for the gcd are: a = " << a << "    b = " << b << endl << endl;
  GlobalOutput() << "To compute lcm in this case we must use big integers because" << endl
                 << "the result is too big to fit into a machine integer (on a 32-bit CPU)" << endl
                 << "lcm(m,n) = " << lcm(ZZ(m), n) << endl;

  // Partial factorization:
  GlobalOutput() << "SmoothFactor(m,999) = " << SmoothFactor(m,999) << endl;
  GlobalOutput() << "SmoothFactor(n,999) = " << SmoothFactor(n,999) << endl;

  // Complete factorization:
  const factorization<long> nfactors = factor(n);
  GlobalOutput() << "factor(n) = " << nfactors << endl;
  GlobalOutput() << "First prime dividing n is " << nfactors.myFactors[0] << endl;

  // Testing primality & generating primes
  GlobalOutput() << endl << "Testing primality & generating primes." << endl;

  // Example showing that IsPrime can be much slower than IsProbPrime.
  const ZZ N = power(2,64);
  const ZZ P = NextProbPrime(N); // NB cannot use NextPrime as it is only for machine integers!
  GlobalOutput() << "The next (probable) prime after 2^64 is P = 2^64 + " << P-N << endl;
  double StartTime = CpuTime();
  IsProbPrime(P);
  GlobalOutput() << "Time for IsProbPrime(P): " << CpuTime() - StartTime << "     [relatively quick]" << endl;
  StartTime = CpuTime();
  IsPrime(P);
  GlobalOutput() << "Time for IsPrime(P): " << CpuTime() - StartTime << "       [relatively slow] " << endl;
}

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-NumTheory1.C,v 1.1 2010/03/03 10:41:50 abbott Exp $
// $Log: ex-NumTheory1.C,v $
// Revision 1.1  2010/03/03 10:41:50  abbott
// Added example for basic number theory functions.
//
// Revision 1.4  2008/10/07 12:12:54  abbott
// Removed useless commented out #include.
//
// Revision 1.3  2007/05/31 16:06:16  bigatti
// -- removed previous unwanted checked-in version
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/07 11:51:40  bigatti
// -- improved test alignment
//
// Revision 1.8  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/03/02 17:46:40  bigatti
// -- unique RingZ and RingQ
// -- requires foundations.H ;  foundations blah;  (thik of a better name)
//
// Revision 1.6  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.5  2007/03/01 13:52:59  bigatti
// -- minor: fixed typo
//
// Revision 1.4  2007/02/28 15:15:56  bigatti
// -- minor: removed quotes in description
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
