// Copyright (c) 2005  Anna Bigatti
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"

#include <cstdlib>
// using exit

using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "This program shows a way to read a polynomial from the input.\n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include<iostream> // using std::endl; using std::flush;
#include<fstream>  // using std::ifstream; using std::ofstream;


RingElem ReadNewPoly(const SparsePolyRing& P)
{
  RingElem f(P);
  size_t InputNumTerms, NumInds = NumIndets(P);
  ZZ IntCoeff;
  vector<long> expv(NumInds);

  // look in program() for changing Global*put from default values
  GlobalLogput() << "how many terms? " << flush;
  GlobalInput() >> InputNumTerms;
  if ( !GlobalInput() )
  {
    GlobalErrput() << "*** ERROR *** Input must be a positive integer" << endl;
    exit(1);
  }
  for (size_t t=1 ; t <= InputNumTerms ; ++t) 
  {
    GlobalLogput() << "[" << t << "]\tcoeff (in ZZ)? " << flush;
    GlobalInput()  >> IntCoeff;
    GlobalLogput() << "   \t" << NumInds << " exponents? (separated by spaces) " << flush;
    for (size_t i=0; i<NumInds; ++i)  GlobalInput() >> expv[i];
    f += monomial(P, IntCoeff, expv);
  }
  return f;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  SparsePolyRing Qxy = NewPolyRing(RingQ(), symbols("x","y"));
  SparsePolyRing Fpabc = NewPolyRing(NewZmod(32003), symbols("a","b","c"));
  SparsePolyRing Pxy = NewPolyRing(Fpabc, symbols("x","y"));

  GlobalOutput() << "We start by reading a polynomial from stdin: " << endl;
  GlobalOutput() << "-- reading h ..." << endl;
  RingElem h = ReadNewPoly(Qxy);
  GlobalOutput() << "-- h in " << owner(h) << endl;
  GlobalOutput() << "-- h     = " << h << endl;

  GlobalOutput() << endl;
  GlobalOutput() << "... and now we read f and g from file \"ex-PolyInput1.in\": " << endl;
  // we can, at any time, change the default streams in, out, log, err like this:
  const char* FileName = "ex-PolyInput1.in";
  ifstream in(FileName);
  if (in == 0)
  {
    GlobalErrput() << "Cannot find input file `" << FileName << "'.  Aborting." << endl;
    abort();
  }
  SetGlobalInput(in);
  ofstream trash("/dev/null");
  SetGlobalLogput(trash);


  GlobalOutput() << "-- reading f ..." << endl;
  RingElem f = ReadNewPoly(Fpabc);
  GlobalOutput() << "-- reading g ..." << endl;
  RingElem g = ReadNewPoly(Pxy);
  GlobalOutput() << endl;

  GlobalOutput() << "-- f in " << owner(f) << endl;
  GlobalOutput() << "-- f     = " << f << endl;
  GlobalOutput() << endl;

  GlobalOutput() << "The printed form of g will have parentheses around each" << endl;
  GlobalOutput() << "coefficient to remind the user that coefficients" << endl;
  GlobalOutput() << "are not \"simple numbers\"" << endl;
  GlobalOutput() << "-- g in " << owner(g) << endl;
  GlobalOutput() << "-- g     = " << g << endl;
  GlobalOutput() << endl;

  RingHom phi = CoeffEmbeddingHom(Pxy);
  GlobalOutput() << "-- phi(f) * g = " << phi(f) * g << endl;
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-PolyInput1.C,v 1.2 2009/10/29 18:24:49 abbott Exp $
// $Log: ex-PolyInput1.C,v $
// Revision 1.2  2009/10/29 18:24:49  abbott
// Added missing include of cstdlib for exit.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.7  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.6  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.5  2007/02/26 15:49:08  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.4  2007/02/12 15:31:57  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.3  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.2  2006/08/17 10:05:27  cocoa
// -- changed: SmallExponent_t --> long
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.4  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.3  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.2  2005/05/04 16:32:03  cocoa
// -- simplified syntax for NewPolyRing
// -- input read both from stdin and file
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.4  2005/04/29 15:42:02  cocoa
// Improved documentation for GMPAllocator.
// Added example program for GMPAllocator.
// Added example program for simple ops on polynomials.
// Added two new ctors for (principal) ideals (from long, and from ZZ).
// Added (crude) printing for PPMonoids.
// Updated library.H (#included GMPAllocator.H).
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.4  2004/12/09 15:07:33  cocoa
// -- simplified code using monomial instead of PushBack
//
// Revision 1.3  2004/12/09 13:31:30  cocoa
// -- choice of I/O moved into program()
//
