// Copyright (c) 2005 John Abbott.
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Program showing how to homogenize a sparse polynomial using iterators.\n";

const string LongDescription =
  "This is just an example!  If you want to homogenize polynomials  \n"
  "you should use the library function \"homog\".                   \n";
// ----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


void HomogCheck(const SparsePolyRing& P, const vector<size_t>& h)
{
  // Verify that wdeg(indet(h[i])) is (0,..,0,1,0,..,0) with 1 in i-th position
  const size_t D = GradingDim(P);
  CoCoA_ASSERT(h.size() == D);
  for (size_t i=0; i < D; ++i)
  {
    CoCoA_ASSERT(h[i] < NumIndets(P));
    degree d = wdeg(indet(P,h[i]));
    for (size_t j=0; j < D; ++j)
      if (i == j)
        CoCoA_ASSERT(IsOne(d[j]));
      else
        CoCoA_ASSERT(IsZero(d[j]));
  }
}


// This procedure assumes that args have been sanity checked.
void MultiHomog(RefRingElem hf, ConstRefRingElem f, const vector<size_t>& h)
{
  CoCoA_ASSERT(owner(hf) == owner(f));
  SparsePolyRing P = AsSparsePolyRing(owner(f));
  const size_t D = GradingDim(P);

  if (IsZero(f)) { hf = f; return; }// trivial case

  // Compute in H the "top" of the degrees of the PPs in f
  //??? if (D == 1) ... SPECIAL CASE
  degree H(D);
  for (SparsePolyIter i=BeginIter(f); !IsEnded(i); ++i)
  {
    H = top(H, wdeg(PP(i)));
  }

  // Now homogenize f.  Accumulate result into a geobucket for speed.
  geobucket ans(P);
  vector<long> expv(NumIndets(P));
  for (SparsePolyIter i=BeginIter(f); !IsEnded(i); ++i)
  {
    degree diff = H - wdeg(PP(i));
    for (size_t j=0; j < D; ++j)
      if (!IsConvertible(expv[h[j]], diff[j]))
        CoCoA_ERROR("Exponent too big", "MultiHomog(ans,f,h)");

    ans.myAddClear(monomial(P, coeff(i), PP(i)*PPMonoidElem(PPM(P),expv)), 1);
  }
  hf = 0; // NB cannot do this earlier in case hf aliases f.
  AddClear(hf, ans);
}


RingElem MultiHomog(ConstRefRingElem f, const vector<size_t>& h)
{
  CoCoA_ASSERT(IsSparsePolyRing(owner(f)));
  SparsePolyRing P = AsSparsePolyRing(owner(f));
  HomogCheck(P, h);
  RingElem ans(P);
  MultiHomog(ans, f, h);
  return ans;
}


RingElem MultiHomog(ConstRefRingElem f, const vector<RingElem>& h)
{
  if (!IsSparsePolyRing(owner(f)))
    CoCoA_ERROR(ERR::NotSparsePolyRing, "homog(f,h)");
  SparsePolyRing P = AsSparsePolyRing(owner(f));

  size_t k = GradingDim(P);
  vector<size_t> indices(k);
  if (h.size() != k)
    CoCoA_ERROR(ERR::BadArraySize, "homog(f,h)");
  for (size_t i=0; i < k; ++i)
  {
    if (owner(h[i]) != P)
      CoCoA_ERROR(ERR::MixedRings, "homog(f,h)");
    size_t index;
    if (!IsIndet(index, h[i]))
      CoCoA_ERROR(ERR::NotIndet, "homog(f,h)");

    indices[k] = index;
  }
  HomogCheck(P, indices);
  RingElem ans(P);
  MultiHomog(ans, f, indices);
  return ans;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  ring Fp = NewZmod(32003);                 // coefficient ring
  SparsePolyRing Fpx = NewPolyRing(Fp, 4);  // Fpx is Z/(32003)[x[0..3]]
  const vector<RingElem>& x = indets(Fpx);
  vector<size_t> h;
  h.push_back(0);
  RingElem f = x[0] + 3*power(x[1],2) + 5*power(x[2],4) + 7*power(x[3],8);
  GlobalOutput() << "Original f  = " << f << endl;
  GlobalOutput() << "MultiHomog(f, h) =  " << MultiHomog(f, h) << endl;
  GlobalOutput() << "homog(f, x[0]) = " << homog(f, x[0]) << endl;
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-PolyIterator2.C,v 1.2 2009/12/11 11:46:32 abbott Exp $
// $Log: ex-PolyIterator2.C,v $
// Revision 1.2  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.8  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.7  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.6  2007/02/26 17:41:16  bigatti
// -- changed "homog" --> "MultiHomog"
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.5  2007/02/22 17:31:20  bigatti
// -- added call to CoCoA::homog
//
// Revision 1.4  2007/02/12 15:45:15  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.3  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.2  2006/08/17 10:05:27  cocoa
// -- changed: SmallExponent_t --> long
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.5  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.4  2005/10/12 15:52:09  cocoa
// Completed test-RingFp1 and corrected/cleaned the SmallFp*
// and RingFp* files.
//
// Some minor tidying elsewhere.
//
// Revision 1.3  2005/10/07 17:08:09  cocoa
// Fixed an embarrassing buglet -- I ought to try compiling before
// I check (and there are plenty of other things I ought to do too...)
//
// Revision 1.2  2005/10/06 16:36:42  cocoa
// Added the capability find out build information at run-time.
// The Makefiles should be a little tidier too.
//
// Revision 1.1  2005/08/08 16:36:33  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.2  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.1  2005/05/04 16:35:06  cocoa
// -- new examples
//
