// Copyright (c) 2009  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Program illustrating basic use of QQ values (i.e. rational numbers) \n"
  "showing that they can be used with a natural syntax.                \n";

const string LongDescription =
  "Program giving example of basic arithmetic with exact rational numbers \n"
  "represented as values of type QQ.  The syntax recalls that used for    \n"
  "the built-in C++ numerical types.  Emphasis is on convenience rather   \n"
  "utmost execution speed.  To understand better the difference between   \n"
  "a QQ value and an element of the ring RingQ, contrast this example     \n"
  "with ex-RingQ1.C.                                                      \n";
//----------------------------------------------------------------------

void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  // Rational numbers can be constructed from a pair of integers
  // (being numerator and denominator).  If the denominator is absent
  // it is taken to be 1.
  GlobalOutput() << "The number 3 as a rational is  " << QQ(3) << endl
                 << "Its reciprocal is  " << QQ(1,3) << "  which is the same as  " << 1/QQ(3) << endl
                 << "You can get the **WRONG ANSWER** by writing  QQ(1/3) = " << QQ(1/3) << " clearly **WRONG**" << endl
                 << "You can also make a rational from a string: " << QQ("22/7") << endl
                 << endl;

  // Rational numbers are always exact; they are not approximated.
  QQ OneThird(1,3); // OneThird = 1/3
  GlobalOutput() << "1/3 + 1/3 + 1/3 - 1 = " << OneThird + OneThird + OneThird - 1 << endl;
  GlobalOutput() << "3*(1/3) - 1 = " << 3*OneThird - 1 << endl;

  // Here is an important caveat: be very careful about rational constants.
  // One might reasonably expect  OneThird + 2/3  to produce 1, but it does not!
  // The C++ compiler interprets the expression  2/3  as an integer division.
  GlobalOutput() << "This value is NOT equal to one: " << OneThird + 2/3 << endl;

  // Here is one way to obtain the desired behaviour:
  GlobalOutput() << "This value IS equal to one: " << OneThird + QQ(2,3) << endl;

  // The functions  num  and  den  give the numerator and denominator (as a ZZ).
  // den(Q) is always positive, and  num and den are always coprime.
  QQ q(123,456);
  GlobalOutput() << "num(" << q << ") = " << num(q) << endl;
  GlobalOutput() << "den(" << q << ") = " << den(q) << endl;

  // The usual arithmetic operators work as you would expect, but note that
  // operator% is NOT DEFINED as it does not make sense.
  QQ q1 = (2*q+1)/(4*q-3);
  QQ q2 = power(q,2);

  // The usual comparison operators work as you would expect.
  // There is also the function  cmp(a,b)  which returns a machine integer
  // which is <0, =0, >0 according as a<b, a=b, a>b.
  if (q1 < q2)   GlobalOutput() << q1 << " is smaller than " << q2 << endl;
  if (q1 == q2)  GlobalOutput() << q1 << " is equal to " << q2 << endl;
  if (q1 > q2)   GlobalOutput() << q1 << " is larger than " << q2 << endl;

  GlobalOutput() << "cmp(" << q1 << ", " << q2 << ") = " << cmp(q1, q2) << endl;

  // There are a few specific tests:
  if (IsZero(q))      GlobalOutput() << q << " is zero" << endl;
  if (IsOne(q))       GlobalOutput() << q << " is one" << endl;
  if (IsMinusOne(q))  GlobalOutput() << q << " is minus one" << endl;

  // Conversion of a rational into an integer.
  // If you want to know whether a rational number is an integer...
  if (IsOneDen(q)) GlobalOutput() << q << " is the integer " << num(q) << endl;
  // Three functions to convert to a nearby integer:
  GlobalOutput() << "floor(" << q << ") = " << floor(q) << endl;
  GlobalOutput() << "ceil(" << q << ") = " << ceil(q) << endl;
  GlobalOutput() << "round(" << q << ") = " << round(q) << endl;
}

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-QQ1.C,v 1.2 2010/03/22 11:47:33 abbott Exp $
// $Log: ex-QQ1.C,v $
// Revision 1.2  2010/03/22 11:47:33  abbott
// Added example using QQ ctor from a string; plus a warning about QQ(1/3).
//
// Revision 1.1  2009/07/08 12:26:53  abbott
// Added floor and ceil functions for QQs.
// Added example program for QQs.
// Minor correction to convert.C; minor cleaning to ex-ZZ1.C
//
// Revision 1.4  2008/10/07 12:12:54  abbott
// Removed useless commented out #include.
//
// Revision 1.3  2007/05/31 16:06:16  bigatti
// -- removed previous unwanted checked-in version
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/07 11:51:40  bigatti
// -- improved test alignment
//
// Revision 1.8  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/03/02 17:46:40  bigatti
// -- unique RingZ and RingQ
// -- requires foundations.H ;  foundations blah;  (thik of a better name)
//
// Revision 1.6  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.5  2007/03/01 13:52:59  bigatti
// -- minor: fixed typo
//
// Revision 1.4  2007/02/28 15:15:56  bigatti
// -- minor: removed quotes in description
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
