// Copyright (c) 2005 Anna Bigatti
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Program showing how to compute a quotient basis of a 0-dimensional ideal.\n";

const string LongDescription =
  "It offers the function \"QuotientBasisRec\"          \n"
  "which will be included in CoCoALib (sooner or later).\n";
// ----------------------------------------------------------------------

// Includes from the standard C++ library
#include <functional> // using std::bind2nd; using std::ptr_fun;
// #include <iostream> // using std::endl;
#include <iterator> // using std::back_insert_iterator;
#include <list> // using std::list;


bool IsSimplePower(ConstRefPPMonoidElem pp)
{
  size_t i = 0;
  for (; i<NumIndets(owner(pp)) ; ++i)
    if (exponent(pp, i)!= 0) break;
  for (++i; i<NumIndets(owner(pp)) ; ++i)
    if (exponent(pp, i)!= 0) return false;
  return true;
}


bool IsDivisible(ConstRefPPMonoidElem pp, const list<PPMonoidElem>& ByL)
{
  for (list<PPMonoidElem>::const_iterator i=ByL.begin(); i != ByL.end(); ++i)
    if (IsDivisible(pp, *i)) return true;
  return false;
}

// apparently bind2nd cannot have (const?) reference arguments
// this is a bad hack
inline bool IsDivisibleHack(RefPPMonoidElem pp, list<PPMonoidElem>* ByL)
{
  return IsDivisible(pp, *ByL);
}

typedef bool (*bool_ppL)(ConstRefPPMonoidElem, const list<PPMonoidElem>&);

void QuotientBasisRec(back_insert_iterator<list<PPMonoidElem> > ans, 
                      const list<PPMonoidElem>& L, 
                      ConstRefPPMonoidElem prefix, 
                      size_t idx)
{
  PPMonoid PPM = owner(L.front());
  const PPMonoidElem& X = indets(PPM)[idx];
  PPMonoidElem ppXd(prefix);  // prefix * x[idx]^d
  int MaxDeg = 0;
  if (idx==NumIndets(PPM)-1)
  {
    MaxDeg = exponent(*(L.begin()),idx);
    for (int d=0; d < MaxDeg; ++d, ppXd *= X, ++ans)
      *ans = ppXd;
    return;
  }
  for (list<PPMonoidElem>::const_iterator it=L.begin(); it != L.end() ; ++it)
    if (exponent(*it,idx)>MaxDeg)  MaxDeg = exponent(*it,idx);
  list<PPMonoidElem> CutOff, tmp;
  PPMonoidElem Xd(PPM);  // x[idx]^d
  for (int d=0; d < MaxDeg; ++d, ppXd *= X, Xd *= X)
  {
    tmp.clear();
    for (list<PPMonoidElem>::const_iterator it=L.begin(); it != L.end(); ++it)
      if (exponent(*it,idx)==d)  tmp.push_back(*it/Xd);
    CutOff.remove_if(bind2nd(ptr_fun(IsDivisibleHack), &tmp));
    //    CutOff.remove_if(bind2nd(ptr_fun(static_cast<bool_ppL>(IsDivisible)), tmp));
    CutOff.insert(CutOff.end(), tmp.begin(), tmp.end());
    QuotientBasisRec(ans, CutOff, ppXd, idx+1);
  }
}


list<PPMonoidElem> QuotientBasis(const ideal& I)
{
  SparsePolyRing P = AsSparsePolyRing(AmbientRing(I));
  vector<RingElem> GB = TidyGens(I);
  list<PPMonoidElem> LeadingPPs;
  for (size_t i=0; i < GB.size(); ++i) back_inserter(LeadingPPs)=LPP(GB[i]);

  vector<PPMonoidElem> SimplePowers;
  for (list<PPMonoidElem>::iterator it=LeadingPPs.begin() ; it!=LeadingPPs.end() ; ++it )
    if (IsSimplePower(*it)) SimplePowers.push_back(*it);
  if (SimplePowers.size()==1 && IsOne(SimplePowers[0]))
    CoCoA_ERROR("ideal must be 0-dimensional", "QuotientBasis");
  if (SimplePowers.size() != NumIndets(P))
    CoCoA_ERROR("ideal must be 0-dimensional", "QuotientBasis");
  
  list<PPMonoidElem> ans;
  QuotientBasisRec(back_inserter(ans), LeadingPPs, PPMonoidElem(PPM(P)), 0);
  return ans;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  ring Fp = NewZmod(32003);                // coefficient ring
  SparsePolyRing Fpx = NewPolyRing(Fp, 8); // Z/(32003)[x[0..7]]

  const vector<RingElem>& x = indets(Fpx);
  vector<RingElem> g;
  for (size_t i=0 ; i<NumIndets(Fpx) ; ++i )  back_inserter(g)=power(x[i],7);
  back_inserter(g) = power(x[2],2) * power(x[5],4);
  back_inserter(g) = power(x[1],3) * power(x[4],4);
  ideal I(Fpx, g);
  GlobalOutput() << "Original I  = " << I << endl;  
  double t0 = CpuTime();
  list<PPMonoidElem> QB=(QuotientBasis(I));
  GlobalOutput() << "CpuTime = " << CpuTime()-t0 << endl;
  GlobalOutput() << "len = " <<  QB.size() << endl;
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-QuotientBasis.C,v 1.2 2007/12/04 14:27:07 bigatti Exp $
// $Log: ex-QuotientBasis.C,v $
// Revision 1.2  2007/12/04 14:27:07  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.11  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.10  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.9  2007/02/26 15:50:25  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.8  2007/02/12 15:45:15  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.7  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.6  2006/11/27 13:22:23  cocoa
// -- deteted unused #include file
//
// Revision 1.5  2006/11/21 14:35:13  cocoa
// -- fixed/hacked a problem with bind2nd: what's the correct way to fix it?
//
// Revision 1.4  2006/11/20 15:52:54  cocoa
// Fixed some buglets.
//
// Revision 1.3  2006/11/17 18:15:28  cocoa
// -- added const & to QuotientBasis argument
//
// Revision 1.2  2006/08/17 10:06:47  cocoa
// -- changed: list as argument is now reference (very embarrassing..)
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.4  2006/03/09 17:11:21  cocoa
// -- changed: "insert" instead of "copy"
//
// Revision 1.3  2006/03/09 17:03:19  cocoa
// -- modified using "list" instead of "vector" to use "remove_if" member
//    function on lists
//
// Revision 1.2  2005/11/29 13:01:36  cocoa
// -- added example for passing overloaded function to bind2nd
//
// Revision 1.1  2005/11/16 14:07:25  cocoa
// -- first import
//
