// Copyright (c) 2007  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "This program illustrates use of the pseudo-random bit generator of CoCoALib.  \n"
  "The bits are independent, identically distributed; each with equal probability\n"
  "of being true or false.  It is also possible to generate biased bits.         \n"
  "See RandomLongStream if you want random (machine) integers.                   \n";

const string LongDescription =
  "CoCoALib offers a pseudorandom bit generator.  The generator can be   \n"
  "seeded when it is created; this allows different pseudorandom         \n"
  "sequences to be produced, though the sequence is completely determined\n"
  "by the initial seed value.  The `prob' function filters a random bit  \n"
  "sequence to produce `true' with a specified probability.              \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  RandomBitStream RBS;
  GlobalOutput() << "This is a brand new random bit generator: " << RBS << endl << endl;

  const int NumBits = 20;
  GlobalOutput() << "The first " << NumBits << " random bits are:";
  for (int i=0; i < NumBits; ++i)
    GlobalOutput() << " " << sample(RBS);
  GlobalOutput() << endl;

  GlobalOutput() << endl
                 << "The ""prob"" function simulates a biased coin toss." << endl
                 << "Here we do 1000000 tosses of a 0.001 probability coin." << endl;
  const double LowProb = 0.001;
  int count = 0;
  const size_t StartIndex = RBS.myIndex();
  for (int i=0; i < 1000000; ++i)
    if (prob(LowProb, RBS)) ++count;
  GlobalOutput() << "The coin came up heads " << count << " times -- this should be about 1000." << endl
                 << "The ""prob"" function uses on average about two random bits per call." << endl
                 << "The actual number of random bits used for ths trial is " << RBS.myIndex() - StartIndex << endl
                 << "We see that this value is indeed not far from 2000000" << endl
                 << endl;

  GlobalOutput() << "The final state of the generator is " << RBS << endl;
}

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-RandomBitStream1.C,v 1.4 2010/02/16 10:20:58 abbott Exp $
// $Log: ex-RandomBitStream1.C,v $
// Revision 1.4  2010/02/16 10:20:58  abbott
// Added new fn sample; SampleBool & SampleLong are obsolescent.
//
// Revision 1.3  2010/02/01 22:35:40  abbott
// Very minor changes to some printed strings.
//
// Revision 1.2  2009/01/23 16:08:08  abbott
// Added missing newline inside the LongDescription string.
//
// Revision 1.1  2007/06/06 15:16:48  abbott
// Added new RandomBitStream class (now based upon GMP's random generator).
// Consequential changes to Makefiles, etc.  There's even doc and an example!
//
// Revision 1.3  2007/05/31 16:06:16  bigatti
// -- removed previous unwanted checked-in version
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/07 11:51:40  bigatti
// -- improved test alignment
//
// Revision 1.8  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/03/02 17:46:40  bigatti
// -- unique RingZ and RingQ
// -- requires foundations.H ;  foundations blah;  (thik of a better name)
//
// Revision 1.6  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.5  2007/03/01 13:52:59  bigatti
// -- minor: fixed typo
//
// Revision 1.4  2007/02/28 15:15:56  bigatti
// -- minor: removed quotes in description
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
