// Copyright (c) 2005  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Example showing operations on RingElem for a ring or a PolyRing.\n";

const string LongDescription =
  "This is a long list of function calls from different rings.  \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


//-- TestRing ----------------------------------------------------------
// testing some ring operations over elements of a given ring

void TestRing(ring R, RingElem a, RingElem b)
{
  RingElem one(R, 1);
  GlobalOutput() << "  initial ring elements are:" << endl;
  GlobalOutput() << "    a = " << a << endl;
  GlobalOutput() << "    b = " << b << endl;
  GlobalOutput() << endl;

  GlobalOutput() << "a == b gives " << (a == b) << endl;
  GlobalOutput() << "a != b gives " << (a != b) << endl;
  GlobalOutput() << "a * b  gives " << a*b << endl;
  GlobalOutput() << "-a     gives " << -a << endl;
  GlobalOutput() << "a += b gives a = " << (a += b) << endl;
  GlobalOutput() << endl;

  GlobalOutput() << "IsZero(one)     gives " << IsZero(one) << endl;
  GlobalOutput() << "IsOne(one)      gives " << IsOne(one) << endl;
  GlobalOutput() << "IsMinusOne(one) gives " << IsMinusOne(one) << endl;
  GlobalOutput() << endl;
  GlobalOutput() << "power(a, 10)  gives "  << power(a, 10) << endl;
  GlobalOutput() << endl;

  GlobalOutput() << "IsDivisible(a, b) gives " << IsDivisible(a, b) << endl;
  if (!IsDivisible(a, b)) 
    GlobalOutput() << "  so we CANNOT compute b / a" << endl;
  else 
    GlobalOutput() << "  b / a gives " << b/a << endl;
  GlobalOutput() << endl;

  GlobalOutput() << "IsFractionField(R) gives " << IsFractionField(R) << endl;
  if (!IsFractionField(R)) 
    GlobalOutput() << "  so we CANNOT compute num(a), den(a)" << endl;
  else 
  {
    GlobalOutput() << "  num(a)   gives " << num(a) << endl;
    GlobalOutput() << "  den(a)   gives " << den(a) << endl;
  }
  GlobalOutput() << endl;

  GlobalOutput() << "IsGCDDomain(R) gives " << IsGCDDomain(R) << endl;
  if (!IsGCDDomain(R)) 
    GlobalOutput() << "  so we CANNOT compute gcd(a,b)" << endl;
  else 
    GlobalOutput() << "  gcd(a, b)   gives " << gcd(a, b) << endl;
  GlobalOutput() << endl;

  GlobalOutput() << "IsPolyRing(R) gives " << IsPolyRing(R) << endl;
  if (!IsPolyRing(R)) 
    GlobalOutput() << "  so we CANNOT compute deg(a) or StdDeg(a)" << endl;
  else 
  {
    GlobalOutput() << "  NB deg(a) and StdDeg(a) are synonyms;" << endl
                   << "  StdDeg is a more precise name but is also more cumbersome" << endl;
    GlobalOutput() << "  deg(a)    gives " << deg(a) << endl;
    GlobalOutput() << "  StdDeg(a) gives " << StdDeg(a) << endl;
  }
  GlobalOutput() << endl;

  GlobalOutput() << "IsSparsePolyRing(R) gives " << IsSparsePolyRing(R) << endl;
  if (!IsSparsePolyRing(R)) 
    GlobalOutput() << "  so we CANNOT compute LPP(a), wdeg(a)" << endl;
  else 
  {
    GlobalOutput() << "  LPP(a)  gives " << LPP(a) << endl;
    GlobalOutput() << "  wdeg(a) gives " << wdeg(a) << endl;
    GlobalOutput() << "  NB wdeg(a) agrees with deg(a) only if the ring is standard graded." << endl;
  }
  GlobalOutput() << endl;

}


//-- main --------------------------------------------------------------
// we run TestRing on some rings

void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  //------------------------------------------------------------
  // Z
  ring Z = RingZ();
  {
    RingElem a(Z), b(Z);
    
    a = 5;
    b = 7;
    GlobalOutput() << "----------------------------------------" << endl;
    GlobalOutput() << "  ring is Z" << endl;
    TestRing(Z, a, b);
  }
  
  //------------------------------------------------------------
  // Q
  ring Q = NewFractionField(Z);  
  {
    RingElem a(Q), b(Q);
    
    a = 5; a /= 2;
    b = 7; b /= 3;
    GlobalOutput() << "----------------------------------------" << endl;
    GlobalOutput() << "  ring is Q" << endl;
    TestRing(Q, a, b);
  }

  //------------------------------------------------------------
  // Q[y]

  // Indet name is y[0]
  PolyRing P = NewPolyRing(Q, symbols("y"));

  RingElem y = indet(P, 0); // set the C++ variable y to the value of the indeterminate y in ring P
  RingElem f = 15 * y + power(y,3);
  RingElem g = 4  * y - 3 * power(y,7);
  GlobalOutput() << "----------------------------------------" << endl;
  GlobalOutput() << "  ring is Q[y]" << endl;
  TestRing(P, f, g);
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-RingElem1.C,v 1.1.1.1 2007/03/09 15:16:11 abbott Exp $
// $Log: ex-RingElem1.C,v $
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/08 16:55:06  cocoa
// Changed name of "range" function to "SymbolRange".
//
// Revision 1.5  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.4  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.3  2007/02/12 15:45:15  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.3  2006/03/01 14:26:41  cocoa
// -- removed some useless "include"
//
// Revision 1.2  2006/02/13 12:08:04  cocoa
// Fixed a problem with some missing assignment ops for certain PPMonoidElems.
// Fixed a bug in RingDistrMPoly::myIndetPower.
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.3  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.2  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.2  2004/12/09 15:08:42  cocoa
// -- added log info
//
