// Copyright (c) 2005  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Example program showing how to create small prime finite fields    \n"
  "implemented in a specific manner -- normally it is best to let     \n"
  "the function NewZmod choose automatically the best implementation. \n";

const string LongDescription =
  "The program asks the user for the characteristic N, and then tries \n"
  "different ways of making Z mod N.\n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  // Ask user for the characteristic -- read it into p.
  ZZ p;
  GlobalOutput() << "Enter the field characteristic: ";
  GlobalInput() >> p;

  // Get rid of the trivial cases...
  if (p == 0 || p == 1 || p == -1)
  {
    GlobalOutput() << "There is no finite field of characteristic " << p << endl;
    return;
  }

  // NewZmod automatically chooses the "best" implementation; it will
  // happily create rings which are not fields.  NewZmod will give an error
  // only if p is 0, 1, or -1.
  ring Zmodp = NewZmod(p);
  GlobalOutput() << "NewZmod has created the ring " << Zmodp << endl;
  if (IsField(Zmodp))
    GlobalOutput() << "It is a field." << endl;
  else
    GlobalOutput() << "It is not a field." << endl;


  // Now try creating a RingFp: fast but characteristic must be small and prime.
  try
  {
    ring SmallFp1 = NewRingFp(p);
    GlobalOutput() << "NewRingFp has created the field " << SmallFp1 << endl
                   << "Since NewRingFp succeeded it is necessarily a field." << endl;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    if (err != ERR::BadSmallFpChar) throw; // rethrow unexpected exceptions
    GlobalOutput() << "NewRingFp failed: the characteristic was composite or too large." << endl;
  }


  // Now try creating a RingFpLog: mul/div using log tables; characteristic must be small and prime.
  try
  {
    ring SmallFp2 = NewRingFpLog(p);
    GlobalOutput() << "NewRingFpLog has created the field " << SmallFp2 << endl
                   << "Since NewRingFpLog succeeded it is necessarily a field." << endl;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    if (err != ERR::BadSmallFpChar) throw; // rethrow unexpected exceptions
    GlobalOutput() << "NewRingFpLog failed: the characteristic was composite or too large." << endl;
  }


  // Now try creating a RingFpDouble: failry fast and characteristic may be as big as 67 million.
  try
  {
    ring SmallFp3 = NewRingFpDouble(p);
    GlobalOutput() << "NewRingFpDouble has created the field " << SmallFp3 << endl
                   << "Since NewRingFpDouble succeeded it is necessarily a field." << endl;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    if (err != ERR::BadSmallFpChar) throw; // rethrow unexpected exceptions
    GlobalOutput() << "NewRingFpDouble failed: the characteristic was composite or too large." << endl;
  }
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-RingFp2.C,v 1.2 2008/10/07 15:45:22 abbott Exp $
// $Log: ex-RingFp2.C,v $
// Revision 1.2  2008/10/07 15:45:22  abbott
// Changed ErrorInfo objects so they include the name of their own error ID.
// Changed catch statements to catch const objects.
// Removed calls to the member fn which accessed the error ID member of an
// ErrorInfo; now you simply compare directly with the error ID (makes the
// code easier to read).
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.4  2007/02/26 17:41:50  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.3  2007/02/12 16:11:12  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.1  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.2  2004/12/09 15:08:42  cocoa
// -- added log info
//
