// Copyright (c) 2005  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "The example in this file shows how to create and use some \n"
  "homomorphisms between rings.                              \n";

const string LongDescription =
  "We compute these polynomials (with parameters) in some rings: \n"
  "f = (2*a/3-1)*x[0] + 1/a;  g = x[0]-a;                        \n";

//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


// A simple function to test embedding ring homs in CoCoALib.
// We know that Kx is of the form R(a)[x[0..N]]
void SimpleTest(const PolyRing& Kx) 
{
  GlobalOutput() << "Kx is " << Kx << endl;
  // Give a handy name to the indeterminate x[0]
  RingElem x0 = indet(Kx, 0);

  ring K = CoeffRing(Kx); // this is R(a)
  ring Ra = BaseRing(AsFractionField(K));  // this is R[a]

  RingElem a_R = indet(AsPolyRing(Ra), 0); // a as element of R[a]

  RingHom EmbedIntoFrF = CanonicalHom(Ra, K);  // K = FractionField(R[a])
  RingHom EmbedCoeff   = CanonicalHom(K, Kx);
//   RingElem a = EmbedCoeff(EmbedIntoFrF(a_R)); // a as element of Kx
// or equivalently:
  RingHom phi = EmbedCoeff(EmbedIntoFrF);  
  RingElem a = phi(a_R);       // a as element of Kx

  RingElem f = (2*a/3-1)*x0 + 1/a;               // This is an element of Kx.
  RingElem g = x0-a;                             // This is another one.

  GlobalOutput() << "f = " << f << endl;
  GlobalOutput() << "f*g = " << f*g << endl;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;
  GlobalOutput() << LongDescription << endl; 

  ring Q = RingQ();
  ring Fp = NewZmod(32003);

  //---------------------------------------------------------------------------
  {
    GlobalOutput() << "  --- Coeffs in " << Fp << endl;
    PolyRing Fpa = NewPolyRing(Fp, symbols("a"));              // Fpa is Fp[a]  where Fp = Z/(32003)
    ring K = NewFractionField(Fpa);                            // K   is Fp(a)
    PolyRing Kxyzt = NewPolyRing(K, symbols("x","y","z","t")); // Kxyzt is Fp(a)[x,y,z,t]
    SimpleTest(Kxyzt);
    GlobalOutput() << endl;
  }

  //---------------------------------------------------------------------------
  // we don't know yet how to deal with this:
//   {
//   GlobalOutput() << "Coeffs in " << Z << endl;
//   PolyRing Za = NewPolyRing(Z, symbols("a"));           // Z[a]
//   PolyRing Qax4 = NewPolyRing(NewFractionField(Za), 4); // Q(a)[x[0..3]]
//   SimpleTest(Qax4);
//   GlobalOutput() << endl << endl;
//   }

  //---------------------------------------------------------------------------
  {
    GlobalOutput() << "  --- Coeffs in " << Q << endl;
    PolyRing Qa = NewPolyRing(Q, symbols("a"));           // Q[a]
    PolyRing Qax4 = NewPolyRing(NewFractionField(Qa), 4); // Q(a)[x[0..3]]
    SimpleTest(Qax4);
    GlobalOutput() << endl;
  }

  GlobalOutput() << "  --- Now we supply an unsuitable input" << endl;
  GlobalOutput() << "      (it is unsuitable because IsFractionField(Fp) gives "
                 << IsFractionField(Fp) << ")" << endl;
  try
  {
    SimpleTest(NewPolyRing(Fp, symbols("a")));
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    if (err != ERR::NotFracField) throw;
    GlobalOutput() << "\nOK!  Our unsuitable input produced the expected exception." << endl;
  }
  GlobalOutput() << endl;
 }

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-RingHom5.C,v 1.2 2008/10/07 15:45:22 abbott Exp $
// $Log: ex-RingHom5.C,v $
// Revision 1.2  2008/10/07 15:45:22  abbott
// Changed ErrorInfo objects so they include the name of their own error ID.
// Changed catch statements to catch const objects.
// Removed calls to the member fn which accessed the error ID member of an
// ErrorInfo; now you simply compare directly with the error ID (makes the
// code easier to read).
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.2  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.1  2007/03/07 14:31:01  bigatti
// -- was ex-RingHom1.C
//
// Revision 1.6  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.5  2007/02/26 17:39:46  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.4  2007/02/12 16:11:12  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.3  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.2  2007/01/18 14:34:18  cocoa
// -- example of alternative syntax
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.3  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.2  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.3  2004/12/09 15:08:42  cocoa
// -- added log info
//
// Revision 1.2  2004/11/22 17:10:04  cocoa
// -- changed  "PolyRing" --> "AsPolyRing",  "FractionField" --> "AsFractionField"
//
