// Copyright (C) 2004 Daniele Venzano; modified by John Abbott.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//-----------------------------------------------------------------------------
const string ShortDescription =
 "Program to find a 'big' prime number, suitable for use in \n"
 "cryptographic applications (with some more primality testing).\n";

const string LongDescription =
  "This program illustrates that ZZ values can be used much like normal C++ ints \n"
  "except that there is a almost no limit on the magnitude of the values. \n"
  "It shows use of some more advanced functions for ZZ values.            \n"
  "NB If you need extreme efficiency then use the GMP library directly.   \n";
//-----------------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  ZZ bits;

  GlobalOutput() << "Enter the number of bits the prime should have: ";
  GlobalInput() >> bits;
  if (bits < 2)
  {
    GlobalOutput() << "There are no primes with fewer than 2 bits." << endl;
    return;
  }
  if (bits > 9999) GlobalOutput() << "WARNING: the computation will take a very long time!" << endl;
  else if (bits > 999) GlobalOutput() << "NOTE: this computation may take some time." << endl;
  ZZ min = power(2, bits-1);
  ZZ max = power(2, bits)-1;

  ZZ prime_candidate = random(min, max);
  while(!IsProbPrime(prime_candidate))
  {
    prime_candidate = random(min, max);
  }

  GlobalOutput() << prime_candidate << " seems to be a prime! (" << bits << " bits)"
                 << endl;

  GlobalOutput() << "And another test with more iterations ";
  if (IsProbPrime(prime_candidate, 50))
    GlobalOutput() << "confirms";
  else
    GlobalOutput() << "rejects";
  GlobalOutput() << " this result." << endl;
}

// We write main() like this so we can handle uncaught CoCoA errors in
// a sensible way (i.e. by announcing them).
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error" << endl;
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-ZZ-prime1.C,v 1.3 2009/06/05 12:14:56 abbott Exp $
// $Log: ex-ZZ-prime1.C,v $
// Revision 1.3  2009/06/05 12:14:56  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
// Revision 1.2  2007/03/22 22:45:31  abbott
// Removed spaces at ends of lines.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.4  2007/02/22 17:25:04  bigatti
// -- added printout of number of bits used
//
// Revision 1.3  2007/02/12 16:20:57  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.1  2004/11/29 16:50:28  cocoa
// -- first import
//
