// Copyright (c) 2006  Anna Bigatti
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "This example is just for testing the Hilbert code.                   \n"
  "It might disappear as soon as HilbertSeries is included in CoCoALib. \n";

const string LongDescription =
  "This code also shows how to create the \"chess examples\".           \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
#include <algorithm>  // using std::min;
// #include <iostream>   // using std::endl;


// ---- chess tests -----------------------------------------------------

ConstRefRingElem CsbSquareIndet(SparsePolyRing P, size_t l, size_t sq1, size_t sq2)
{
  CoCoA_ASSERT( l*l <= NumIndets(P) );
  CoCoA_ASSERT( sq1 <= l && sq2 <= l );
  return indet(P, (sq1-1)*l + (sq2-1));
}


ideal NewQueenMovesFrom(SparsePolyRing P, size_t Csb, size_t sq1, size_t sq2)
{
  ConstRefRingElem x = CsbSquareIndet(P, Csb, sq1, sq2);
  vector<RingElem> g;
  for ( size_t i=sq2+1 ; i<=Csb ; ++i )
    g.push_back(x * CsbSquareIndet(P, Csb, sq1, i));
  for ( size_t i=sq1+1 ; i<=Csb ; ++i )
    g.push_back(x * CsbSquareIndet(P, Csb, i, sq2));
  for ( size_t i=min(Csb-sq1,Csb-sq2) ; i>0 ; --i )
    g.push_back(x * CsbSquareIndet(P, Csb, sq1+i, sq2+i));
  for ( size_t i=min(Csb-sq1, sq2-1) ; i>0 ; --i )
    g.push_back(x * CsbSquareIndet(P, Csb, sq1+i, sq2-i));
  return ideal(P, g);
}


ideal NewQueenIdeal(SparsePolyRing P, size_t Csb)
{
  ideal I = ideal(P);
  for ( size_t sq1=1 ; sq1<=Csb ; ++sq1 )
    for ( size_t sq2=1 ; sq2<=Csb ; ++sq2 )
      I += NewQueenMovesFrom(P, Csb, sq1, sq2);
  return I;
}


// ----------------------------------------------------------------------

void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  ring Q = RingQ();
  DenseUPolyRing HPRing = NewPolyRing_DUP(RingZ(), symbol("lambda"));

  SparsePolyRing P = NewPolyRing(Q, 4);
  const vector<RingElem>& x = indets(P);
  vector<RingElem> g;
  g.push_back(x[1]);
  g.push_back(x[2]);
  g.push_back(x[3]);
  ideal I = ideal(P, g);
  GlobalOutput() << "gens(I) = " << gens(I) << endl;
  GlobalOutput() << "TidyGens(I) = " << TidyGens(I) << endl;
  GlobalOutput() << "HilbertNumeratorMod(HPRing, I) = "
                 << HilbertNumeratorMod(HPRing, I) << endl;

  SparsePolyRing CsbRing7 = NewPolyRing(Q,49);
  ideal Queen7 = NewQueenIdeal(CsbRing7, 7);

  double T;
  T=CpuTime();
  TidyGens(Queen7);
  GlobalOutput() << "TidyGens time = " << CpuTime()-T << endl;
  T=CpuTime();  
  GlobalOutput() << "HilbertNumeratorMod(HPRing, Queen7) = "
                 << HilbertNumeratorMod(HPRing, Queen7) << endl;
  GlobalOutput() << "Hilbert time = " << CpuTime()-T << endl;

  SparsePolyRing CsbRing = NewPolyRing(Q,64);
  ideal Queen8 = NewQueenIdeal(CsbRing, 8);
  T=CpuTime();
  TidyGens(Queen8);
  GlobalOutput() << "TidyGens time = " << CpuTime()-T << endl;
  T=CpuTime();  
  RingElem HPNumerator = HilbertNumeratorMod(HPRing, Queen8);
  GlobalOutput() << "Hilbert time = " << CpuTime()-T << endl;
  GlobalOutput() << "HilbertNumeratorMod(HPRing, Queen8) = "
                 << HPNumerator << endl;

  FractionField HPSRing = NewFractionField(HPRing);
  RingElem t = indet(HPRing, 0);
  RingHom phi = EmbeddingHom(HPSRing);
//NYI  GlobalOutput() << "HilbertSeries = "
//NYI                 << phi(HPNumerator)/phi(power(1-t,64)) << endl;
  //  GlobalOutput() << "HilbertSeriesMod(I) = " << HilbertSeriesMod(I) << end;
}

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-hilbert1.C,v 1.3 2007/10/30 17:14:12 abbott Exp $
// $Log: ex-hilbert1.C,v $
// Revision 1.3  2007/10/30 17:14:12  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.2  2007/10/18 11:34:00  bigatti
// -- fixed: using DenseUPolyRing
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/08 17:27:20  bigatti
// -- improved NewPolyRing
//
// Revision 1.8  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/02/28 15:14:14  bigatti
// -- minor: Q7 --> Queen7
//
// Revision 1.6  2007/02/26 17:40:34  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.5  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.4  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.3  2006/11/17 18:16:02  cocoa
// -- added timings
//
// Revision 1.2  2006/11/16 17:39:51  cocoa
// -- fixed some calls to rum: lots of code could be deleted, but this is
//    all much more fragile than I would like
//
// Revision 1.1  2006/10/09 16:48:58  cocoa
// -- first import
//
