// Copyright (c) 2005  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Example program illustrating the creation of matrices, and some \n"
  "some operations on them. \n";

const string LongDescription =
  "Please note that the matrix code is still rather young. \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;


// convention: a function containing a "new" should be named "New.."
matrix NewMatrixFromC(ring Z, int cmat[4][4])
{
  matrix M(NewDenseMat(Z,4,4));
  
  for (int i=0; i < 4; ++i)
    for (int j=0; j < 4; ++j)
      SetEntry(M, i, j, cmat[i][j]);
  return M;
}


void ExBasicOps(matrix M)
{
  GlobalOutput() << "M = " << M << endl;
  GlobalOutput() << "rank(M) = " << rank(M) << endl;
  GlobalOutput() << "det(M) = " << det(M) << endl;

  matrix InvM = inverse(M);  
  GlobalOutput() << "InvM = " << InvM << endl;
  mul(InvM, InvM, M);  
  GlobalOutput() << "InvM * M = " << InvM << endl;

  matrix AdjM = adjoint(M);  
  GlobalOutput() << "AdjM = " << AdjM << endl;  
  mul(AdjM, AdjM, M);
  GlobalOutput() << "AdjM * M = " << AdjM << endl;
}


void ExMatrixView(matrix M)
{
  GlobalOutput() <<"-- some examples with ConstMatrixView/MatrixView --"<< endl;
  GlobalOutput() <<"see the documentation of matrix for the full list"<< endl;
  
  // It is easy to compute the determinant of an identity matrix, even if large.
  ConstMatrixView Id40000 = IdentityMat(BaseRing(M), 40000);
  GlobalOutput() << "det(Id40000) = " << det(Id40000) << endl;

  ConstMatrixView Id4 = IdentityMat(BaseRing(M), 4);
  ConstMatrixView  A = ConcatHor(M, Id4);
  GlobalOutput() << "ConcatHor(M, Id4) = " << A << endl;
  
  ConstMatrixView Z28 = ZeroMat(BaseRing(M), 2, 8);
  ConstMatrixView  B = ConcatVer(A, Z28);
  GlobalOutput() << "ConcatVer(A, Z28) = " << B << endl;

  // with MatrixView you can change the original matrix M
  //   using a "different point of view"
  // (remember: in C/C++ indices start from 0)
  GlobalOutput() << M << endl;
  // transpose
  MatrixView TrM = transpose(M);  SetEntry(TrM, 1,0, 11);
  GlobalOutput() << 
    "   TrM = transpose(M);  SetEntry(TrM, 1,0, 11)" << endl <<
    "M = " << M << endl;

  // if you do not want to modify M make a copy into a proper matrix
  //  matrix TrM1 = NewDenseMat(transpose(M));

  vector<size_t> L12;
  L12.push_back(1); L12.push_back(2);
  // submat
  MatrixView SM = submat(M, L12, L12);  SetEntry(SM, 1,0, 22);
  GlobalOutput() << 
    "    SM = submat(M, L12, L12);  SetEntry(SM, 0,1, 22);" << endl <<
    "M = " << M << endl;

  // BlockMat
  MatrixView BM = BlockMat(M, M, M, M);  SetEntry(BM, 7,0, 33);
  GlobalOutput() << 
    "    BM = BlockMatrix(M, M, M, M);  SetEntry(BM, 7,0, 33);" << endl <<
    "BM = " << BM << endl;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

   // the first 2 rows represent the degree matrix
  int C_matrix[4][4] = {{1, 1, 1, 1},
                        {1, 2, 3, 4},
                        {5, 6, 0, 0},
                        {0, 0, 7, 7}};

  // Create two CoCoALib matrices containing the values of M.
  // M_Z5 contains the image of M in Z/(5), and M_Q the image in Q.
  matrix M_Z5(NewMatrixFromC(NewZmod(5), C_matrix));
  matrix M_Q(NewMatrixFromC(RingQ(), C_matrix));

  ExBasicOps(M_Z5);
  ExBasicOps(M_Q);

  ExMatrixView(M_Q);
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-matrix1.C,v 1.4 2009/02/13 14:55:26 bigatti Exp $
// $Log: ex-matrix1.C,v $
// Revision 1.4  2009/02/13 14:55:26  bigatti
// -- changed a lot: rearranged basic operations and added (Const)MatrixView
//
// Revision 1.3  2008/04/18 15:35:57  abbott
// (long overdue) Major revision to matrices
//
// Revision 1.2  2008/04/08 15:26:42  abbott
// Major revision to matrix implementation: added matrix views.
// Lots of changes.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.5  2007/02/26 17:40:34  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.4  2007/02/22 17:25:40  bigatti
// -- added printing of separators: "-------"
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.2  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.6  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.5  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.4  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.3  2005/03/30 15:30:43  cocoa
// -- added: computation of rank and determinant
// -- changed: declaration of I4000 as ConstMatrix
//
// Revision 1.2  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.1  2004/11/29 16:50:28  cocoa
// -- first import
//
