// Copyright (c) 2005,2008  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "This example program shows how to create various types of ring, and \n"
  "several operations one can perform on rings (e.g. querying their    \n"
  "properties).                                                        \n";

const string LongDescription =
  "This example creates several different sorts of ring,            \n"
  "and then calls PrintRingInfo on each one.                        \n"
  "PrintRingInfo calls various functions to obtain information      \n"
  "about each ring passed to it.  Naturally, some query functions   \n"
  "make sense only for certain types of ring (e.g. NumIndets(R)).   \n";

//----------------------------------------------------------------------

// Function to print out some information about a ring.
// Here you see how various query functions of a ring can be used.
// If the ring is actually PolyRing, we create a "view" of it as
// a PolyRing, and this permits us to make further queries which
// are valid only for PolyRings.

void PrintRingInfo(const ring& R)
{
  GlobalOutput() << "R                   = " << R << endl; 
  GlobalOutput() << "its zero and one elements are " << zero(R) << " and " << one(R) << endl;

  // NOTE result is a ZZ not an int:
  GlobalOutput() << "char(R)             = " << characteristic(R) << endl; 

  // NB: "Is.." functions ask about implementation (not isomorphism classes)
  // "IsBla" should be read as "Is internally implemented as Bla"
  GlobalOutput() << "IsGCDDomain(R)      = " << IsGCDDomain(R) << endl;
  GlobalOutput() << "IsIntegralDomain(R) = " << IsIntegralDomain(R) << endl;
  GlobalOutput() << "IsField(R)          = " << IsField(R) << endl;

  // The next four queries tell how the ring is implemented:
  GlobalOutput() << "IsZ(R)              = " << IsZ(R) << endl;
  GlobalOutput() << "IsFractionField(R)  = " << IsFractionField(R) << endl;
  GlobalOutput() << "IsPolyRing(R)       = " << IsPolyRing(R) << endl;
  GlobalOutput() << "IsQuotientRing(R)   = " << IsQuotientRing(R) << endl;

  // If R is a PolyRing, we print out some more details...
  if (IsPolyRing(R))
  {
    GlobalOutput() << "--------========--------" << endl;
    GlobalOutput() << "Additional information specific to a polynomial ring:" << endl;

    // P and R are two views of the same underlying ring; we can "see"
    // more detail by querying P since it "knows" that the ring is actually
    // a polynomial ring.
    PolyRing P = AsPolyRing(R);
    GlobalOutput() << "NumIndets(R)        = " << NumIndets(P) << endl; // NB CANNOT do NumIndets(R)
    GlobalOutput() << "The indeterminates are: ";
    for (size_t i=0; i < NumIndets(P); ++i)
      GlobalOutput() << indet(P, i) << "  ";
    GlobalOutput() << endl;
    GlobalOutput() << "The information for the coefficient ring is:" << endl;
    PrintRingInfo(CoeffRing(P));
    GlobalOutput() << "-- end coefficient ring --" << endl;
  }

  GlobalOutput() << endl;
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  const int SmallPrime = 13;
  const ZZ LargePrime = NextProbPrime(power(10,20) + 1);
  const ZZ composite = power(7,510);  // a large composite

  ring Z = RingZ();                    // the integers
  ring Q = RingQ();                    // the rationals
  ring Fsmallp = NewZmod(SmallPrime);  // small finite field
  ring Fbigp = NewZmod(LargePrime);    // large finite field

  // We may also create quotient rings:
  ring ZmodN = NewQuotientRing(Z, ideal(RingElem(Z, composite)));
  ring Zmod0 = NewQuotientRing(Z, ideal(Z)); // NB isomorphic to Z but not implemented as Z!

  // This is a ring of floating point values.
  ring R = NewRingTwinFloat(128);  // guarantees 128 bit of precision (otherwise throws).
  
  // Here we create two polynomial rings: see ex-PolyRing1 for more details.
  ring P = NewPolyRing(Q, symbols("x","y","z"), lex);  // Q[x,y,z]
  ring P30 = NewPolyRing(NewZmod(30), 3);              // Z/(30)[x[1..3]]

  GlobalOutput() << "-- Info for ring Z" << endl; 
  PrintRingInfo(Z);

  GlobalOutput() << "-- Info for ring Q" << endl; 
  PrintRingInfo(Q);

  GlobalOutput() << "-- Info for ring F_13" << endl; 
  PrintRingInfo(Fsmallp);

  GlobalOutput() << "-- Info for ring F_p (for some large prime p)" << endl; 
  PrintRingInfo(Fbigp);

  GlobalOutput() << "-- Info for ring Z/nZ (for some large composite n)" <<endl;
  PrintRingInfo(ZmodN);

  GlobalOutput() << "-- Info for ring Z/0Z (NB: different from Z)" << endl; 
  PrintRingInfo(Zmod0);

  GlobalOutput() << "-- Info for ring R (twin floats)" << endl; 
  PrintRingInfo(R);

  GlobalOutput() << "-- Info for ring Q[x[0],y[0],z[0]]" << endl; 
  PrintRingInfo(P);

  GlobalOutput() << "-- Info for ring (Z/30Z)[x[0],x[1],x[2]]" << endl; 
  PrintRingInfo(P30);
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-ring2.C,v 1.6 2010/03/22 13:57:16 bigatti Exp $
// $Log: ex-ring2.C,v $
// Revision 1.6  2010/03/22 13:57:16  bigatti
// -- removed ring Z from ideal ctor (no longer necessary)
//
// Revision 1.5  2010/03/05 21:35:00  abbott
// Added some consts.  Changed an example to use PPOrderingCtor when creating a PolyRing.
//
// Revision 1.4  2009/06/05 12:14:55  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
// Revision 1.3  2008/07/21 07:58:51  abbott
// Several cosmetic changes to examples.
//
// Revision 1.2  2007/06/21 21:29:47  abbott
// Changed name of RingFloat into RingTwinFloat.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.2  2007/03/08 17:28:47  bigatti
// -- improved NewPolyRing
//
// Revision 1.1  2007/03/07 14:30:38  bigatti
// -- was ex-ring1
//
// Revision 1.6  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.5  2007/02/26 17:40:34  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.4  2007/02/22 17:26:23  bigatti
// -- added printing of separator after coeff ring
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:53  cocoa
// Imported files
//
// Revision 1.3  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.2  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.3  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.2  2005/04/21 15:12:19  cocoa
// Revised NewPolyRing as Dag Arneson suggested (perhaps just an interim
// measure).
// Brought example programs up to date (new name for CoCoA error
// information objects).
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.4  2004/11/22 17:10:04  cocoa
// -- changed  "PolyRing" --> "AsPolyRing",  "FractionField" --> "AsFractionField"
//
