#ifndef CoCoA_QQ_H
#define CoCoA_QQ_H

//   Copyright (c)  2009-2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.
//   -------------------------------------------------------------------------


#include "CoCoA/MachineInteger.H"
#include "CoCoA/ZZ.H"

#include "gmp.h"
// #include <iosfwd> --- already included in ZZ.H
//using std::ostream;
//using std::istream;

namespace CoCoA
{

  class OpenMathOutput;  // fwd decl -- defined in OpenMath.H
  class OpenMathInput;   // fwd decl -- defined in OpenMath.H

  class QQ
  {
  private:
    friend mpq_t& mpqref(QQ& N);
    friend const mpq_t& mpqref(const QQ& N);
    mpq_t myRep;
    QQ(MP_RAT to_be_owned); // for efficient return values in operator+ etc.

  public:
    QQ();
    explicit QQ(const MachineInteger& n);
    explicit QQ(const ZZ& N);
    enum NormalizeFlag { NotNormalized=0, AlreadyNormalized=1 };
    QQ(const MachineInteger& n1, const MachineInteger& n2, NormalizeFlag status=NotNormalized);  ///< creates the fraction n1/n2
    QQ(const MachineInteger& n1, const ZZ& N2, NormalizeFlag status=NotNormalized);  ///< creates the fraction n1/N2
    QQ(const ZZ& N1, const MachineInteger& n2, NormalizeFlag status=NotNormalized);  ///< creates the fraction N1/n2
    QQ(const ZZ& N1, const ZZ& N2, NormalizeFlag status=NotNormalized);  ///< creates the fraction N1/N2
    explicit QQ(const std::string& str);
    explicit QQ(const char* str);
    QQ(const QQ& from);
// QQ& clone() const;

    ~QQ();

    QQ& operator=(const QQ& rhs);
//     QQ& operator=(const QQ::rtn& rhs);

    // -------- functions that modify at least one argument or `*this' ----------

    QQ& operator+=(const QQ& rhs);
    QQ& operator-=(const QQ& rhs);
    QQ& operator*=(const QQ& rhs);
    QQ& operator/=(const QQ& rhs);
  			
    // Same but with RHS a ZZ...
    QQ& operator= (const ZZ& rhs);
    QQ& operator+=(const ZZ& rhs);
    QQ& operator-=(const ZZ& rhs);
    QQ& operator*=(const ZZ& rhs);
    QQ& operator/=(const ZZ& rhs);
  			
    // Same but with RHS a MachinInteger...
    QQ& operator= (const MachineInteger& rhs);
    QQ& operator+=(const MachineInteger& rhs);
    QQ& operator-=(const MachineInteger& rhs);
    QQ& operator*=(const MachineInteger& rhs);
    QQ& operator/=(const MachineInteger& rhs);


    const QQ& operator++( );
    const QQ& operator--( );
    const QQ operator++(int); // INEFFICIENT
    const QQ operator--(int); // INEFFICIENT

  };


  // I/O FUNCTIONS

  std::ostream& operator<<(std::ostream& out, const QQ& Q);
  std::istream& operator>>(std::istream& in, QQ& Q);

  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const QQ& Q);
  OpenMathInput& operator>>(OpenMathInput& OMIn, QQ& Q);


  // STANDARD ARITHMETIC OPERATIONS

  void swap(QQ& a, QQ& b);

  const ZZ num(const QQ& Q);
  const ZZ den(const QQ& Q);
  const QQ abs(const QQ& Q);
  const QQ operator-(const QQ& Q);

  const QQ operator+(const QQ& Q1, const QQ& Q2);
  const QQ operator-(const QQ& Q1, const QQ& Q2);
  const QQ operator*(const QQ& Q1, const QQ& Q2);
  const QQ operator/(const QQ& Q1, const QQ& Q2);

  const QQ operator+(const QQ& Q, const ZZ& N);
  const QQ operator-(const QQ& Q, const ZZ& N);
  const QQ operator*(const QQ& Q, const ZZ& N);
  const QQ operator/(const QQ& Q, const ZZ& N);

  const QQ operator+(const ZZ& N, const QQ& Q);
  const QQ operator-(const ZZ& N, const QQ& Q);
  const QQ operator*(const ZZ& N, const QQ& Q);
  const QQ operator/(const ZZ& N, const QQ& Q);

  const QQ operator+(const QQ& Q, const MachineInteger& n);
  const QQ operator-(const QQ& Q, const MachineInteger& n);
  const QQ operator*(const QQ& Q, const MachineInteger& n);
  const QQ operator/(const QQ& Q, const MachineInteger& n);

  const QQ operator+(const MachineInteger& n, const QQ& Q);
  const QQ operator-(const MachineInteger& n, const QQ& Q);
  const QQ operator*(const MachineInteger& n, const QQ& Q);
  const QQ operator/(const MachineInteger& n, const QQ& Q);

  const QQ power(const QQ& base, const ZZ& exponent);
  const QQ power(const QQ& base, const MachineInteger& exponent);


  bool IsZero(const QQ& Q);
  bool IsOne(const QQ& Q);
  bool IsMinusOne(const QQ& Q);
  bool IsOneDen(const QQ& Q);
  int sign(const QQ& Q);

  // COMPARISON FUNCTIONS

  int cmp(const QQ& Q1, const QQ& Q2);
  int cmp(const QQ& Q, const ZZ& N);
  int cmp(const ZZ& N, const QQ& Q);
  int cmp(const QQ& Q, const MachineInteger& n);
  int cmp(const MachineInteger& n, const QQ& Q);

  bool operator==(const QQ& Q1, const QQ& Q2);
  bool operator!=(const QQ& Q1, const QQ& Q2);
  bool operator> (const QQ& Q1, const QQ& Q2);
  bool operator>=(const QQ& Q1, const QQ& Q2);
  bool operator< (const QQ& Q1, const QQ& Q2);
  bool operator<=(const QQ& Q1, const QQ& Q2);
			
  bool operator==(const QQ& Q, const ZZ& N);
  bool operator!=(const QQ& Q, const ZZ& N);
  bool operator> (const QQ& Q, const ZZ& N);
  bool operator>=(const QQ& Q, const ZZ& N);
  bool operator< (const QQ& Q, const ZZ& N);
  bool operator<=(const QQ& Q, const ZZ& N);
			
  bool operator==(const ZZ& N, const QQ& Q);
  bool operator!=(const ZZ& N, const QQ& Q);
  bool operator> (const ZZ& N, const QQ& Q);
  bool operator>=(const ZZ& N, const QQ& Q);
  bool operator< (const ZZ& N, const QQ& Q);
  bool operator<=(const ZZ& N, const QQ& Q);
			
  bool operator==(const QQ& Q, const MachineInteger& n);
  bool operator!=(const QQ& Q, const MachineInteger& n);
  bool operator> (const QQ& Q, const MachineInteger& n);
  bool operator>=(const QQ& Q, const MachineInteger& n);
  bool operator< (const QQ& Q, const MachineInteger& n);
  bool operator<=(const QQ& Q, const MachineInteger& n);
      		
  bool operator==(const MachineInteger& n, const QQ& Q);
  bool operator!=(const MachineInteger& n, const QQ& Q);
  bool operator> (const MachineInteger& n, const QQ& Q);
  bool operator>=(const MachineInteger& n, const QQ& Q);
  bool operator< (const MachineInteger& n, const QQ& Q);
  bool operator<=(const MachineInteger& n, const QQ& Q);
			

  // MISCELLANEOUS FUNCTIONS

  double mantissa(const QQ& Q);
  long exponent(const QQ& Q);
  double log(const QQ& Q);

  ZZ floor(const QQ& Q);   ///< greatest integer <= Q
  ZZ ceil(const QQ& Q);    ///< least integer >= Q
  ZZ round(const QQ& Q);   ///< round rational Q to nearest integer (halves round toward +infinity)


  // Inline fns

  inline mpq_t& mpqref(QQ& Q)
  {
    return Q.myRep;
  }

  inline const mpq_t& mpqref(const QQ& Q)
  {
    return Q.myRep;
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/QQ.H,v 1.5 2010/03/22 11:49:28 abbott Exp $
// $Log: QQ.H,v $
// Revision 1.5  2010/03/22 11:49:28  abbott
// Added ctor from a string.
//
// Revision 1.4  2010/03/18 16:34:10  abbott
// Added new pseudo-ctors for QQ with optional flag to indicate that value is already normalized.
// Added OpenMath I/O operators.
//
// Revision 1.3  2009/11/26 16:12:54  bigatti
// -- commented out some includes
//
// Revision 1.2  2009/07/08 12:26:53  abbott
// Added floor and ceil functions for QQs.
// Added example program for QQs.
// Minor correction to convert.C; minor cleaning to ex-ZZ1.C
//
// Revision 1.1  2009/07/02 16:29:42  abbott
// Added new class QQ to represent rational numbers.
// Consequent change to the Makefile.
//
//

#endif
