#ifndef CoCoA_SparsePolyRing_H
#define CoCoA_SparsePolyRing_H

//   Copyright (c)  2005-2007,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Header file for the abstract class SparsePolyRingBase etc.

#include "CoCoA/PPMonoid.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/error.H"
#include "CoCoA/ideal.H"

#include <cstddef>
//using std::size_t;
#include <iosfwd>
//using std::ostream;
#include <vector>
//using std::vector;
#include <memory>
// using std::auto_ptr;


namespace CoCoA
{

  class symbol;  // fwd declaration -- defined in symbol.H

  class SparsePolyIter; // fwd declaration
  class SparsePolyIterBase
  {
  public:
    virtual ~SparsePolyIterBase() {};
    virtual SparsePolyIterBase* myClone() const = 0;

    virtual bool IamEnded() const = 0;
    virtual void myForward() = 0;
//    virtual SparsePolyIter& operator++() = 0;
//    virtual SparsePolyIter operator++(int) = 0;
    virtual ConstRefRingElem myCoeff() const = 0;
    virtual ConstRefPPMonoidElem myPP() const = 0;
//    virtual bool operator==(const iter& i2) const = 0;
//    virtual bool operator!=(const iter& i2) const = 0;
  protected:
    virtual const void* myPolyPtr() const = 0;
    virtual const void* myTermPtr() const = 0;
    friend class SparsePolyIter;
    friend bool operator==(const SparsePolyIter& i1, const SparsePolyIter& i2);
  };

  // This is a "smart pointer class" -- it owns the object pointed to.
  class SparsePolyIter
  {
  public:
    SparsePolyIter(SparsePolyIterBase* ptr);
    SparsePolyIter(const SparsePolyIter& copy);
    // Default dtor works OK
    SparsePolyIter& operator=(const SparsePolyIter& rhs);
    SparsePolyIterBase* operator->();
    const SparsePolyIterBase* operator->() const;
    SparsePolyIter& operator++();
    SparsePolyIter operator++(int);
  private:
    std::auto_ptr<SparsePolyIterBase> myIterPtr;
    friend bool IsEnded(const SparsePolyIter& i);
    friend ConstRefRingElem coeff(const SparsePolyIter& i);
    friend ConstRefPPMonoidElem PP(const SparsePolyIter& i);
    friend bool operator==(const SparsePolyIter& i1, const SparsePolyIter& i2);
  };


  class SparsePolyRingBase;

  /*-----------------------------------------------------------------*/
  /** \include SparsePolyRing.txt  */
  /*-----------------------------------------------------------------*/
  class SparsePolyRing: public PolyRing
  {
  public:
    explicit SparsePolyRing(const SparsePolyRingBase* RingPtr);
    // Default copy ctor works fine.
    // Default dtor works fine.
  private: // disable assignment
    SparsePolyRing& operator=(const SparsePolyRing& rhs); ///< NEVER DEFINED -- assignment disabled
  public:
    const SparsePolyRingBase* operator->() const; ///< allow member fns to be called
  };


  class SparsePolyRingBase: public PolyRingBase
  {
  public:
    ///@name Member functions every concrete SparsePolyRing implementation must have in addition to those of PolyRingBase.
    //@{
    virtual const PPMonoid& myPPM() const = 0;
    virtual std::size_t myGradingDim() const { return GradingDim(myPPM()); }
//    virtual void myWDeg(degree& d, ConstRawPtr rawf) const = 0;
//    virtual int myCmpWDeg(ConstRawPtr rawf, ConstRawPtr rawg) const = 0; //pure???
    virtual ConstRefPPMonoidElem myLPP(ConstRawPtr rawf) const = 0;
    virtual void myMulByPP(RawPtr rawf, PPMonoidElemConstRawPtr rawpp) const = 0;
    virtual bool myIsHomog(ConstRawPtr rawf) const; ///< true iff f is homogeneous
    virtual bool myIsHomogPartial(ConstRawPtr rawf, size_t n) const; ///< true iff f is homogeneous wrt the first n rows of weights (n <= GradingDim)
    virtual void myHomog(RawPtr rawfHom, ConstRawPtr rawf, ConstRawPtr rawh) const;  ///< fHom = f homogenized with h (indet of wdeg 1, GrDim=1)
    //@}

    ///@name   Functions for creating/building polynomials
    //@{
    virtual RingElem myMonomial(ConstRawPtr rawc, PPMonoidElemConstRawPtr rawpp) const = 0; // c is element of CoeffRing!!
    virtual SparsePolyIter myBeginIter(ConstRawPtr rawf) const = 0;
    virtual SparsePolyIter myEndIter(ConstRawPtr rawf) const = 0;
    virtual void myPushFront(RawPtr rawf, ConstRawPtr rawc, const std::vector<long>& expv) const = 0; /// SHOULD BE vector<ZZ> ????
    virtual void myPushBack(RawPtr rawf, ConstRawPtr rawc, const std::vector<long>& expv) const = 0; /// SHOULD BE vector<ZZ> ????
    virtual void myPushFront(RawPtr rawf, ConstRawPtr rawc, PPMonoidElemConstRawPtr rawpp) const = 0;
    virtual void myPushBack(RawPtr rawf, ConstRawPtr rawc, PPMonoidElemConstRawPtr rawpp) const = 0;
    //@}

    ///@name functions inherited from PolyRing with a single implementation for all SparsePolyRing implementations
    //@{
    virtual long myStdDeg(ConstRawPtr rawf) const;              ///< returns standard degree of f
    virtual long myMaxExponent(ConstRawPtr rawf, std::size_t index) const;  ///< returns the maximum exponent of x[index] in f
    virtual void myContent(RawPtr rawcontent, ConstRawPtr rawf) const; ///< exception clean
    virtual void myCommonDenom(RawPtr rawcontent, ConstRawPtr rawf) const; ///< exception clean
    virtual void myRemoveBigContent(RawPtr rawf) const;                ///< exception guarantee: same as myDivByCoeff
    virtual bool myIsValid(ConstRawPtr rawf) const;                    ///< true iff f is a valid representation, for debugging
    //@}

    ///@name functions inherited from ring with a single implementation for all SparsePolyRing implementations
    //@{
    virtual void myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; ///< lhs = x*y
    //    virtual void myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; ///< lhs = x/y
    virtual bool myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; ///< lhs = x/y, if divisible
    virtual void myDeriv(RawPtr rawlhs, ConstRawPtr rawf, ConstRawPtr rawx) const; ///< lhs = deriv(f, x)
    virtual void mySymbols(std::vector<symbol>& SymList) const;            ///< append symbols in ring to SymList
    virtual void myOutput(std::ostream& out, ConstRawPtr rawx) const;      ///< out << x
    virtual bool myIsPrintAtom(ConstRawPtr rawx) const;                    ///< x^n may be printed without parentheses
    virtual void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const;  ///< OMOut << x
    virtual bool myIsOne(ConstRawPtr rawf) const;                          ///< true iff rawf == 1
    virtual bool myIsMinusOne(ConstRawPtr rawf) const;                     ///< true iff rawf == -1
    virtual bool myIsConstant(ConstRawPtr rawf) const;                     ///< true iff rawf == c (possibly 0)
    virtual bool myIsIndet(std::size_t& IndetIndex, ConstRawPtr rawf) const;///< true iff rawf is x[IndetIndex]
    virtual bool myIsInvertible(ConstRawPtr rawx) const;                   ///< true iff x is invertible
    virtual void myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; ///< lhs = gcd(x,y) if GCDDomain;
    virtual bool myIsInteger(ZZ& N, ConstRawPtr rawx) const;               ///< true iff x is integer
////    virtual bool myIsRational(ZZ& N, ZZ& D, ConstRawPtr rawx) const;       ///< true iff x is rational
    virtual bool myIsRational(QQ& Q, ConstRawPtr rawx) const;              ///< true iff x is rational
    virtual ideal myIdealCtor(const std::vector<RingElem>& gens) const;
    virtual void mySquare(RawPtr rawlhs, ConstRawPtr rawx) const;          ///< lhs = x^2
    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const; ///< lhs = x^n, n>1, x not -1,0,1
    //@}

    ///@name  Functions for homomorphisms
    //@{
    RingHom myHomCtor(const ring& codomain, const RingHom& CoeffHom, const std::vector<RingElem>& IndetImages) const;
    RingHom myCompose(const RingHom& phi, const RingHom& theta) const;
    //@}

    ///@name Special functions on polynomials needed for implementing Buchberger's Algorithm
    //@{
    virtual bool myIsZeroAddLCs(RawPtr rawf, RawPtr rawg) const = 0; ///< f+=LM(g); g-=LM(g); assumes LPP(f)==LPP(g); returns LC(f)+LC(g)==0
    virtual void myMoveLM(RawPtr rawf, RawPtr rawg) const = 0;       ///< f+=LM(g); g-=LM(g); assumes LM(f)<LM(g)
    virtual void myDeleteLM(RawPtr rawf) const = 0;            // ????? right interface
    virtual void myDivLM(RawPtr rawlhs, ConstRawPtr rawf, ConstRawPtr rawg) const = 0; ///< lhs=div(LM(f),LM(g)); assumes f!=0,g!=0
    virtual int  myCmpLPP(ConstRawPtr rawf, ConstRawPtr rawg) const = 0; ///< cmp(LPP(f),LPP(g)); assumes f!=0,g!=0
    virtual void myAddClear(RawPtr rawf, RawPtr rawg) const = 0;     ///< f+=g; g=0;
    virtual void myAppendClear(RawPtr rawf, RawPtr rawg) const = 0;  ///< f+=g; g=0; appends g to f with no checks

    enum SkipLMFlag { SkipLMg, DontSkipLMg }; // used only for AddMul
    virtual void myAddMul(RawPtr rawf, ConstRawPtr rawh, ConstRawPtr rawg) const = 0; ///<  f += LM(h)*g
    virtual void myAddMul(RawPtr rawf, ConstRawPtr rawh, ConstRawPtr rawg, SkipLMFlag) const = 0; ///<  f += LM(h)*g
    virtual void myReductionStep(RawPtr rawf, ConstRawPtr rawg) const = 0;
    // ??? aggiungere coefficiente
    virtual void myReductionStepGCD(RawPtr rawf, ConstRawPtr rawg, RefRingElem FScale) const = 0;
    // should it all be in ReductionStep ??? ANNA
    //@}

  protected: // Ideal class for this type of ring.

    class IdealImpl: public IdealBase
    {
    public:
      IdealImpl(const SparsePolyRing& P, const std::vector<RingElem>& gens);
      // default copy ctor is OK
      virtual IdealBase* clone() const;
//???    virtual void swap(ideal& other);

      virtual const SparsePolyRing& myAmbientRing() const;
      virtual bool IamZero() const;
      // virtual bool IamOne() const;  // default impl
      virtual bool IhaveMonomialGens() const;
      virtual void myReduceMod(RingElemRawPtr rawx) const; // r elem of R, where I is ideal in R
      virtual bool contains(RingElemConstRawPtr rawx) const;
      virtual void add(const ideal&);
      virtual void mul(const ideal&);
      virtual void intersect(const ideal&);
      virtual void colon(const ideal&);
      virtual void myDivMod(RingElemRawPtr rawlhs, RingElemConstRawPtr rawnum, RingElemConstRawPtr rawden) const; // lhs = num/den modulo the ideal  (lhs = 0 if quotient does not exist)

      virtual const std::vector<RingElem>& myGens() const; // gens as specified by user
      virtual const std::vector<RingElem>& myTidyGens() const; // tidier set of gens
    protected:
      virtual void myMaximalTest() const;
      virtual void myPrimeTest() const;

    protected: // Data members of SparsePolyRingBase::IdealImpl
      SparsePolyRing myP;
      std::vector<RingElem> myGensValue;
      mutable bool3 IhaveMonomialGensValue;
      mutable bool myGBasisIsValid;
      mutable std::vector<RingElem> myGBasisValue;
    protected: // Auxiliary functions.
      virtual const std::vector<RingElem>& myGBasis() const;
      static const SparsePolyRingBase::IdealImpl* GetPtr(const ideal& J);
    };  // class IdealImpl

  protected:  // General homomorphism class for this type of ring.

    class HomImpl: public RingHomBase
    {
    public:
      HomImpl(const SparsePolyRing& domain, const ring& codomain, const RingHom& CoeffHom, const std::vector<RingElem>& IndetImages);
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    private:
      virtual void myOutputSelfDetails(std::ostream& out) const;

    private: // Data members of SparsePolyRingBase::HomImpl
      const RingHom myCoeffHom;
      const std::vector<RingElem> myIndetImages;
    };  // class HomImpl

  protected:  // Special homomorphism class for this type of ring.

    class CoeffEmbeddingHomImpl: public RingHomEmbeddingBase
    {
    public:
      CoeffEmbeddingHomImpl(const SparsePolyRing& P);
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    };  // class CoeffEmbeddingHomImpl

  };


  SparsePolyRing NewPolyRing(const ring& CoeffRing, const PPMonoid& PPM);
  SparsePolyRing NewPolyRing(const ring& CoeffRing, const std::vector<symbol>& IndetNames, const PPOrdering& ord);
  SparsePolyRing NewPolyRing(const ring& CoeffRing, const std::vector<symbol>& IndetNames, const PPOrderingCtor& ord);
  SparsePolyRing NewPolyRing(const ring& CoeffRing, const std::vector<symbol>& IndetNames); ///< Default ordering is StdDegRevLex
  SparsePolyRing NewPolyRing(const ring& CoeffRing, std::size_t NumIndets); ///< Default ordering is StdDegRevLex
  SparsePolyRing NewPolyRing(const ring& CoeffRing, std::size_t NumIndets, const PPOrderingCtor& ord);
  bool IsSparsePolyRing(const ring& R);                         // inline
  SparsePolyRing AsSparsePolyRing(const ring& R);               // inline
  const PPMonoid& PPM(const SparsePolyRing& Rx);                // inline
  std::size_t GradingDim(const SparsePolyRing& Rx);             // inline
  bool AreGoodIndetNames(const ring& CoeffRing, const std::vector<symbol>& IndetNames);

  ///@name Pseudo constructors for RingElems owned by SparsePolyRing
  //@{
  RingElem monomial(const SparsePolyRing& P, ConstRefRingElem c, ConstRefPPMonoidElem pp);
  RingElem monomial(const SparsePolyRing& P, MachineInteger n, ConstRefPPMonoidElem pp);
  RingElem monomial(const SparsePolyRing& P, const ZZ& N, ConstRefPPMonoidElem pp);
  RingElem monomial(const SparsePolyRing& P, const QQ& N, ConstRefPPMonoidElem pp);
  RingElem monomial(const SparsePolyRing& P, ConstRefRingElem c, const std::vector<long>& expv);
  RingElem monomial(const SparsePolyRing& P, MachineInteger n, const std::vector<long>& expv);
  RingElem monomial(const SparsePolyRing& P, const ZZ& N, const std::vector<long>& expv);
  RingElem monomial(const SparsePolyRing& P, const QQ& N, const std::vector<long>& expv);
  RefRingElem PushFront(RefRingElem f, ConstRefRingElem c, const std::vector<long>& expv); /// SHOULD BE vector<ZZ> ????
  RefRingElem PushBack(RefRingElem f, ConstRefRingElem c, const std::vector<long>& expv); /// SHOULD BE vector<ZZ> ????
  RefRingElem PushFront(RefRingElem f, ConstRefRingElem c, ConstRefPPMonoidElem pp); ///< with checks
  RefRingElem PushBack(RefRingElem f, ConstRefRingElem c, ConstRefPPMonoidElem pp); ///< with checks
  //@}


  ///@name Special functions on RingElems owned by SparsePolyRing
  //@{
  ConstRefPPMonoidElem LPP(ConstRefRingElem f);                 // inline
  degree wdeg(ConstRefRingElem f);
  int CmpWDeg(ConstRefRingElem f1, ConstRefRingElem f2);
  int CmpWDegPartial(ConstRefRingElem f1, ConstRefRingElem f2, std::size_t GrDim);
  bool IsHomog(ConstRefRingElem f);
  bool IsHomogPartial(ConstRefRingElem f, std::size_t GrDim);
  RingElem homog(ConstRefRingElem f, ConstRefRingElem h);       ///< homogenize f with indet h (GrDim=1, wdeg(h)=1)
  RingElem NR(ConstRefRingElem f, const std::vector<RingElem>& v); ///< Normal Remainder (remainder of the Division Algorithm).  If v is a GBasis it gives the Normal Form.
  SparsePolyIter BeginIter(ConstRefRingElem f);
  SparsePolyIter EndIter(ConstRefRingElem f);
  //@}

  ///@name Special functions on ideals owned by SparsePolyRing
  //@{
  bool IsZeroDim(const ideal& I);
  //@}

  //----------------------------------------------------------------------
  // Inline functions

  inline bool IsSparsePolyRing(const ring& R)
  {
    return dynamic_cast<const SparsePolyRingBase*>(R.myRawPtr()) != 0;
  }


  inline SparsePolyRing AsSparsePolyRing(const ring& R)
  {
    if (!IsSparsePolyRing(R))
      CoCoA_ERROR(ERR::NotSparsePolyRing, "AsSparsePolyRing(ring)");
    return SparsePolyRing(static_cast<const SparsePolyRingBase*>(R.myRawPtr()));
  }


  inline SparsePolyRing::SparsePolyRing(const SparsePolyRingBase* RingPtr):
      PolyRing(RingPtr)
  {}


  inline const SparsePolyRingBase* SparsePolyRing::operator->() const
  {
    return static_cast<const SparsePolyRingBase*>(ring::operator->());
  }


  inline const PPMonoid& PPM(const SparsePolyRing& Rx)
  {
    return Rx->myPPM();
  }


  inline std::size_t GradingDim(const SparsePolyRing& Rx)
  {
    return Rx->myGradingDim();
  }


  inline ConstRefPPMonoidElem LPP(ConstRefRingElem f)
  {
    if (!IsSparsePolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemSparsePolyRing, "LPP(f)");
    if (IsZero(f)) CoCoA_ERROR(ERR::ZeroRingElem, "LPP(f)");
    return AsSparsePolyRing(owner(f))->myLPP(raw(f));
  }


  inline SparsePolyIter::SparsePolyIter(SparsePolyIterBase* ptr):
      myIterPtr(ptr)
  {}


  inline SparsePolyIter::SparsePolyIter(const SparsePolyIter& copy):
      myIterPtr(copy.myIterPtr->myClone())
  {}


  inline SparsePolyIter& SparsePolyIter::operator=(const SparsePolyIter& rhs)
  {
//    if (&copy == this.get()) return *this;
    myIterPtr.reset(rhs.myIterPtr->myClone());
    return *this;
  }


  inline SparsePolyIterBase* SparsePolyIter::operator->()
  {
    return myIterPtr.get();
  }


  inline const SparsePolyIterBase* SparsePolyIter::operator->() const
  {
    return myIterPtr.get();
  }


  inline bool IsEnded(const SparsePolyIter& i)
  {
    return i.myIterPtr->IamEnded();
  }


  inline SparsePolyIter& SparsePolyIter::operator++()
  {
    if (myIterPtr->IamEnded()) CoCoA_ERROR(ERR::PolyIterEnded, "++PolyIter");
    myIterPtr->myForward();
    return *this;
  }

  inline SparsePolyIter SparsePolyIter::operator++(int)
  {
    if (myIterPtr->IamEnded()) CoCoA_ERROR(ERR::PolyIterEnded, "PolyIter++");
    SparsePolyIter ans(*this);
    myIterPtr->myForward();
    return ans;
  }


  inline ConstRefRingElem coeff(const SparsePolyIter& i)
  {
    if (IsEnded(i)) CoCoA_ERROR(ERR::PolyIterEnded, "coeff(PolyIter)");
    return i.myIterPtr->myCoeff();
  }


  inline ConstRefPPMonoidElem PP(const SparsePolyIter& i)
  {
    if (IsEnded(i)) CoCoA_ERROR(ERR::PolyIterEnded, "PP(PolyIter)");
    return i.myIterPtr->myPP();
  }


  inline bool operator==(const SparsePolyIter& i1, const SparsePolyIter& i2)
  {
    if (i1.myIterPtr->myPolyPtr() != i2.myIterPtr->myPolyPtr())
      CoCoA_ERROR(ERR::MixedPolyIters, "PolyIter==PolyIter");
    return i1.myIterPtr->myTermPtr() == i2.myIterPtr->myTermPtr();
  }


  inline bool operator!=(const SparsePolyIter& i1, const SparsePolyIter& i2)
  {
    return !(i1==i2);
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/SparsePolyRing.H,v 1.28 2010/03/18 13:55:56 abbott Exp $
// $Log: SparsePolyRing.H,v $
// Revision 1.28  2010/03/18 13:55:56  abbott
// Added pseudo-ctors for monomials with QQ coeffs.
//
// Revision 1.27  2010/03/05 18:43:48  abbott
// Added pseudo-ctors allowing polynomial rings to be created specifying
// the ordering using a PPOrderingCtor object.
//
// Revision 1.26  2010/02/04 09:57:11  bigatti
// -- added "mul" for ideals.  Implemented only for SparsePolyRing
//
// Revision 1.25  2010/01/21 14:12:28  bigatti
// -- swapped position of IhaveMonomialGens and myGBasisIsValid
//
// Revision 1.24  2009/11/26 17:21:34  bigatti
// -- added PushFront/PushBack(f, c, pp)
// -- added in .C inline functions:
// --   CheckCompatible, CheckElemSparsePolyRing, CheckCoeffExpv, CheckCoeffPP
//
// Revision 1.23  2009/10/02 13:27:26  bigatti
// -- unique implementation of myDiv in PolyRing.C
//
// Revision 1.22  2009/09/28 16:19:43  bigatti
// -- unique implementation for myDeriv
//
// Revision 1.21  2009/09/25 13:02:09  bigatti
// -- myDiv with one implementation in SparsePolyRing
//
// Revision 1.20  2009/09/22 14:07:33  bigatti
// -- added CmpWDegPartial and IsHomogPartial
//
// Revision 1.19  2009/07/24 12:26:43  abbott
// Added CommonDenom function for polynomials.
//
// Revision 1.18  2009/07/02 16:32:11  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.17  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.16  2008/11/18 15:20:10  bigatti
// -- added const to myGBasis return value
// -- added myIdealCtor to RingWeyl for proper inheritance
//
// Revision 1.15  2008/10/09 15:15:05  bigatti
// -- added "bool3 IhaveMonomialGensValue" member field for ideals
//
// Revision 1.14  2008/04/21 12:32:54  abbott
// Corrected size_t into std::size_t in several header files; in some cases,
// replaced size_t with MachineInteger (with consequent changes to impl files).
//
// Revision 1.13  2008/04/15 14:54:28  bigatti
// -- added mySquare
//
// Revision 1.12  2008/04/10 15:00:24  bigatti
// -- added  myPushBack/Front(RawPtr, ConstRawPtr, PPMonoidElemConstRawPtr)
//
// Revision 1.11  2008/03/12 16:35:18  bigatti
// -- changed: IsHomogeneous --> IsHomog
// -- changed: ERR:ZeroPoly --> ERR::ZeroRingElem
//
// Revision 1.10  2007/12/07 15:27:01  bigatti
// -- default implementation of "IamOne" in ideal.C
//
// Revision 1.9  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.8  2007/10/30 17:14:11  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.7  2007/10/05 15:17:13  bigatti
// -- just a proper sorting on functions
//
// Revision 1.6  2007/05/31 16:31:53  bigatti
// -- adjusted comments
//
// Revision 1.5  2007/05/31 15:43:56  bigatti
// -- added mySymbols and AreGoodIndetNames
//
// Revision 1.3  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.2  2007/03/12 16:00:29  bigatti
// -- moved myLog(F, index) into unique implementation in SparsePolyRing
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.23  2007/03/08 18:42:05  cocoa
// Cleaned up whitespace.
//
// Revision 1.22  2007/03/08 11:07:12  cocoa
// Made pseudo ctors for polynomial rings more uniform.  This allowed me to
// remove an include of CoCoA/symbol.H  from the RingDistrM*.H files, but then
// I had to put the include in several .C files.
//
// Revision 1.21  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.20  2007/01/20 14:07:25  bigatti
// -- moved code for homomorphism into common implementation in SparsePolyRing
//
// Revision 1.19  2007/01/15 13:33:34  cocoa
// -- added prefix "raw" to RawPtr arguments names
//
// Revision 1.18  2006/12/07 17:36:19  cocoa
// -- migrated  myRemoveBigContent myContent myPowerSmallExp  into
//    single implementation in SparsePolyRing
// -- removed  content  from DistrMPoly(..)
//
// Revision 1.17  2006/12/06 17:25:43  cocoa
// -- removed #include "config.H"
//
// Revision 1.16  2006/11/23 17:43:53  cocoa
// -- minor change
//
// Revision 1.15  2006/11/22 17:51:31  cocoa
// -- moved printing functions into unified implementation in SparsePolyRing
//
// Revision 1.14  2006/11/21 18:09:24  cocoa
// -- added myIsMonomial
// -- implemented myIsOne, myIsMinusOne, myIsConstant, myIsIndet in SparsePolyRing
// -- removed the 4 functions from DistrMPoly(..) and RingDistrMPoly(..)
// -- changed all names of RawPtr arguments into "raw(..)"
//
// Revision 1.13  2006/11/20 15:55:02  cocoa
// ring is now a class again.  Improved definitions of operator-> in derived classes.
//
// Revision 1.12  2006/11/14 17:36:49  cocoa
// -- fixed implementation for ideal in RingWeyl
//
// Revision 1.11  2006/11/09 17:46:58  cocoa
// -- version 0.9712:
// --   IdealImpl moved to SparsePolyRing from concrete rings
// -- PolyList in GTypes is now vector<RingElem> (was list)
// -- "my" coding convention applied to DistrMPoly
//
// Revision 1.10  2006/11/03 14:01:46  cocoa
// -- changed: reference counting in ring, PPMonoids and OrdvArith now
//    uses SmartPtrIRC
//
// Revision 1.9  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.8  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.7  2006/08/17 09:45:07  cocoa
// -- added: homogenization
//
// Revision 1.6  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.5  2006/07/20 17:06:08  cocoa
// -- moved myStdDeg into SparsePolyRing
//
// Revision 1.4  2006/07/20 16:51:38  cocoa
// -- added common implementation of myStdDeg
//
// Revision 1.3  2006/07/17 11:05:53  cocoa
// -- added: myIsValid, myIsHomogeneous, IsHomogeneous
//
// Revision 1.2  2006/06/08 16:45:28  cocoa
// -- RingDistrMPoly*.H  have been "moved" into RingDistrMPoly*.C
// -- some coding conventions fixed in DistrMPoly*
// -- functions wdeg and CmpWDeg have a common implementation in SparsePolyRing
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.10  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.9  2006/05/12 17:01:28  cocoa
// -- added myIsUnit, myGcd whose implementation in RingDistr*** were identical
//
// Revision 1.8  2006/04/27 13:06:19  cocoa
// -- just spaces
//
// Revision 1.7  2006/04/26 16:44:53  cocoa
// -- myMul has now a single implementation in SparsePolyRing
// -- myMul and mul in RingDistrMPoly* and DistrMPoly* have been disabled
//
// Revision 1.6  2006/04/21 14:58:04  cocoa
// Removed myWDeg member function: it is no longer needed.
//
// Revision 1.5  2006/03/17 18:13:25  cocoa
// -- changed: myMul --> myMulByPP
//
// Revision 1.4  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
// Revision 1.3  2006/03/07 10:11:06  cocoa
// -- fixed:  PPMonoidElem myLPP/LPP now return ConstRefPPMonoidElem
//
// Revision 1.2  2006/01/19 16:34:42  cocoa
// -- added NF, myReduceMod functions (not yet tested)
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.5  2005/09/30 15:03:39  cocoa
// Minor cleaning and tidying.
// DistrMPolyInlPP: use of summands now rather cleaner.
//
// Revision 1.4  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.3  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.2  2005/07/15 16:34:33  cocoa
// Added iterators for sparse polynomials.
// The code compiles (and the old tests still run).
// It'd Friday evening -- I'm going home before
// getting any ideas about making the iterator code run.
//
// Revision 1.1  2005/07/01 16:08:16  cocoa
// Friday check-in.  Major change to structure under PolyRing:
// now SparsePolyRing and DUPolyRing are separated (in preparation
// for implementing iterators).
//
// A number of other relatively minor changes had to be chased through
// (e.g. IndetPower).
//

#endif
