#ifndef CoCoA_ZZ_H
#define CoCoA_ZZ_H

//   Copyright (c)  2003-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/MachineInteger.H"

#include <cstddef>
//using std::size_t
// #include <iosfwd> -- already included in MachineInteger.H
//using std::ostream;
//using std::istream;
#include <string>

#include "gmp.h"

namespace CoCoA
{

  /*-----------------------------------------------------------------*/
  /** \include ZZ.txt  */
  /*-----------------------------------------------------------------*/

  class OpenMathOutput;  // fwd decl -- defined in OpenMath.H
  class OpenMathInput;   // fwd decl -- defined in OpenMath.H

  enum CopyFromMPZMarker { CopyFromMPZ }; // just used for a ZZ ctor

  class ZZ
  {
  private:  // data member
    friend mpz_t& mpzref(ZZ& N);
    friend const mpz_t& mpzref(const ZZ& N);
    mpz_t myRep;

  public:
    ZZ();
    ZZ(const MachineInteger& n);
    explicit ZZ(const std::string& str);
    explicit ZZ(const char* str);
    ZZ(MP_INT to_be_owned); // for efficient return values in operator+ etc.
    ZZ(CopyFromMPZMarker, const mpz_t to_be_copied);
    ZZ(const ZZ& from);  // std copy ctor
// ZZ& clone() const;

    ~ZZ();

    ZZ& operator=(const ZZ& rhs);

    // -------- functions that modify at least one argument or `*this' ----------

    ZZ& operator+=(const ZZ& rhs);
    ZZ& operator-=(const ZZ& rhs);
    ZZ& operator*=(const ZZ& rhs);
    ZZ& operator/=(const ZZ& rhs);
    ZZ& operator%=(const ZZ& rhs);
  			
    // Same but with RHS a MachineInteger...
    ZZ& operator= (const MachineInteger& rhs);
    ZZ& operator+=(const MachineInteger& rhs);
    ZZ& operator-=(const MachineInteger& rhs);
    ZZ& operator*=(const MachineInteger& rhs);
    ZZ& operator/=(const MachineInteger& rhs);
    ZZ& operator%=(const MachineInteger& rhs);


    const ZZ& operator++( );
    const ZZ& operator--( );
    const ZZ operator++(int); // INEFFICIENT
    const ZZ operator--(int); // INEFFICIENT

  };


  // I/O FUNCTIONS

  std::ostream& operator<<(std::ostream& out, const ZZ& N);
  std::istream& operator>>(std::istream& in, ZZ& N);

  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const ZZ& N);
  OpenMathInput& operator>>(OpenMathInput& OMIn, ZZ& N);


  // STANDARD ARITHMETIC OPERATIONS

  void swap(ZZ& a, ZZ& b);

  const ZZ abs(const ZZ& N);
  const ZZ operator-(const ZZ& N);

  const ZZ operator+(const ZZ& N1, const ZZ& N2);
  const ZZ operator-(const ZZ& N1, const ZZ& N2);
  const ZZ operator*(const ZZ& N1, const ZZ& N2);
  const ZZ operator/(const ZZ& N1, const ZZ& N2);
  const ZZ operator%(const ZZ& N1, const ZZ& N2);
  const ZZ operator+(const ZZ& N1, const MachineInteger& n2);
  const ZZ operator-(const ZZ& N1, const MachineInteger& n2);
  const ZZ operator*(const ZZ& N1, const MachineInteger& n2);
  const ZZ operator/(const ZZ& N1, const MachineInteger& n2);
  unsigned long operator%(const ZZ& N1, const MachineInteger& n2);
  const ZZ operator+(const MachineInteger& n1, const ZZ& N2);
  const ZZ operator-(const MachineInteger& n1, const ZZ& N2);
  const ZZ operator*(const MachineInteger& n1, const ZZ& N2);
  const ZZ operator/(const MachineInteger& n1, const ZZ& N2);
  const ZZ operator%(const MachineInteger& n1, const ZZ& N2);
  const ZZ power(const ZZ& base, const ZZ& exponent);
  const ZZ power(const ZZ& base, const MachineInteger& exponent);
  const ZZ power(const MachineInteger& base, const ZZ& exponent);
  const ZZ power(const MachineInteger& base, const MachineInteger& exponent);


  // Efficient arithmetic functions...

  void add(ZZ& lhs, const ZZ& N1, const ZZ& N2);
  void sub(ZZ& lhs, const ZZ& N1, const ZZ& N2);
  void mul(ZZ& lhs, const ZZ& N1, const ZZ& N2);
  void div(ZZ& lhs, const ZZ& N1, const ZZ& N2);
  void mod(ZZ& lhs, const ZZ& N1, const ZZ& N2);
  void quorem(ZZ& quo, ZZ& rem, const ZZ& num, const ZZ& den);
  void divexact(ZZ& lhs, const ZZ& N1, const ZZ& N2);

  void add(ZZ& lhs, const ZZ& N1, const MachineInteger& n2);
  void sub(ZZ& lhs, const ZZ& N1, const MachineInteger& n2);
  void mul(ZZ& lhs, const ZZ& N1, const MachineInteger& n2);
  void div(ZZ& lhs, const ZZ& N1, const MachineInteger& n2);
  void mod(ZZ& lhs, const ZZ& N1, const MachineInteger& n2);
  void quorem(ZZ& quo, ZZ& rem, const ZZ& num, const MachineInteger& den); // why is rem not of type MachineInteger???
  //???divexact???

  void add(ZZ& lhs, const MachineInteger& n1, const ZZ& N2);
  void sub(ZZ& lhs, const MachineInteger& n1, const ZZ& N2);
  void mul(ZZ& lhs, const MachineInteger& n1, const ZZ& N2);
  void div(ZZ& lhs, const MachineInteger& n1, const ZZ& N2);
  void mod(ZZ& lhs, const MachineInteger& n1, const ZZ& N2);
  void quorem(ZZ& quo, ZZ& rem, const MachineInteger& num, const ZZ& den);
  //???divexact???

  void neg(ZZ& lhs, const ZZ& rhs);
  void abs(ZZ& lhs, const ZZ& rhs);


  // SUNDRY PREDICATES.

  bool IsZero(const ZZ& N);
  bool IsOne(const ZZ& N);
  bool IsMinusOne(const ZZ& N);
  int sign(const ZZ& N);
  bool IsOdd(const ZZ& N);
  bool IsEven(const ZZ& N);
//bool IsPerfectSquare(const ZZ& N);

  // Some of these predicates for MachineIntegers are in utils.H, the rest are here:
// bool IsPerfectSquare(const MachineInteger& n);


  // COMPARISON FUNCTIONS

  int cmp(const ZZ& N1, const ZZ& N2);
  int cmp(const ZZ& N1, const MachineInteger& n2);
  int cmp(const MachineInteger& n1, const ZZ& N2);

  bool operator==(const ZZ& N1, const ZZ& N2);
  bool operator!=(const ZZ& N1, const ZZ& N2);
  bool operator> (const ZZ& N1, const ZZ& N2);
  bool operator>=(const ZZ& N1, const ZZ& N2);
  bool operator< (const ZZ& N1, const ZZ& N2);
  bool operator<=(const ZZ& N1, const ZZ& N2);
			
  bool operator==(const ZZ& N1, const MachineInteger& n2);
  bool operator!=(const ZZ& N1, const MachineInteger& n2);
  bool operator> (const ZZ& N1, const MachineInteger& n2);
  bool operator>=(const ZZ& N1, const MachineInteger& n2);
  bool operator< (const ZZ& N1, const MachineInteger& n2);
  bool operator<=(const ZZ& N1, const MachineInteger& n2);
      		
  bool operator==(const MachineInteger& n1, const ZZ& N2);
  bool operator!=(const MachineInteger& n1, const ZZ& N2);
  bool operator> (const MachineInteger& n1, const ZZ& N2);
  bool operator>=(const MachineInteger& n1, const ZZ& N2);
  bool operator< (const MachineInteger& n1, const ZZ& N2);
  bool operator<=(const MachineInteger& n1, const ZZ& N2);
			

  // MISCELLANEOUS FUNCTIONS

  double mantissa(const ZZ& n);
  std::size_t exponent(const ZZ& n);

  std::size_t NumDigits(const ZZ& n, int base=2);
  double log(const ZZ& n);

  ZZ RangeFactorial(const MachineInteger& lo, const MachineInteger& hi); // computes lo*(lo+1)*(lo+2)*...*hi
  const ZZ factorial(const ZZ& N);
  const ZZ factorial(const MachineInteger& n);
  const ZZ binomial(const ZZ& N, const ZZ& R);
  const ZZ binomial(const MachineInteger& n, const ZZ& R);
  const ZZ binomial(const ZZ& N, const MachineInteger& r);
  const ZZ binomial(const MachineInteger& n, const MachineInteger& r);
  const ZZ fibonacci(const ZZ& N);
  const ZZ fibonacci(const MachineInteger& n);

  const ZZ random(const ZZ& lo, const ZZ& hi);                         ///< rnd number in range lo to hi (incl).
  const ZZ random(const MachineInteger& lo, const MachineInteger& hi); ///< rnd number in range lo to hi (incl).
  const ZZ random(const MachineInteger& lo, const ZZ& hi);             ///< rnd number in range lo to hi (incl).
  const ZZ random(const ZZ& lo, const MachineInteger& hi);             ///< rnd number in range lo to hi (incl).

  const ZZ RoundDiv(const ZZ& num, const MachineInteger& den); ///< round num/den to an integer (halves round toward +infinity)
  const ZZ RoundDiv(const MachineInteger& num, const ZZ& den); ///< round num/den to an integer (halves round toward +infinity)
  const ZZ RoundDiv(const ZZ& num, const ZZ& den);             ///< round num/den to an integer (halves round toward +infinity)

/*
ZZ root(const ZZ&, long n = 2);
void root(ZZ&, const ZZ&, long n = 2);
ZZ sqrt(const ZZ&);
void sqrt(ZZ&, const ZZ&);
void sqrt(ZZ&, ZZ&, const ZZ&); // square root and remainder
bool IsSquare(const ZZ&);
bool IsPower(const ZZ&);
int cmpabs(const ZZ&, const ZZ&);
int cmpabs(const ZZ&, const MachineInteger&);
int cmpabs(const MachineInteger&, const ZZ&);


  friend void           SqrtRem(ZZ& sqrt, ZZ& rem, const ZZ& b);
ZZ         Div2Exp(const ZZ& b, unsigned long exponent_of_2);
ZZ         Mod2Exp(const ZZ& b, unsigned long exponent_of_2);
ZZ         Mul2Exp(const ZZ& b, unsigned long exponent_of_2);
ZZ         sqrt   (const ZZ& b);
*/


  /////////////////////////////////////////////////////////////////////////////
  // Inline functions

  inline mpz_t& mpzref(ZZ& N)
  {
    return N.myRep;
  }

  inline const mpz_t& mpzref(const ZZ& N)
  {
    return N.myRep;
  }


// I don't like this function name...
//   inline void neg_(ZZ& x)
//   {
//     x.neg();
//   }


  inline ZZ::~ZZ()
  {
    mpz_clear(myRep);
  }


// inline bool IsPerfectSquare(const MachineInteger& n);

} // end of namespace CoCoA



// The next few lines contain RCS header/log information.
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/ZZ.H,v 1.18 2010/03/22 11:50:31 abbott Exp $
// $Log: ZZ.H,v $
// Revision 1.18  2010/03/22 11:50:31  abbott
// Added ctor from a string.
// Fixed stupid bug in operator-.
//
// Revision 1.17  2010/03/18 13:54:20  abbott
// Added openmath output fns (moved from OpenMath files).
//
// Revision 1.16  2010/03/05 21:35:48  abbott
// Removed unnecessary include of assert.H
//
// Revision 1.15  2010/02/16 10:22:21  abbott
// Changed include directive for gmp.h -- uses quotes instead of <...>
//
// Revision 1.14  2009/12/29 22:44:32  abbott
// Removed buggy proxy class ZZ::rtn.
// Consequent changes for function prototypes also in NumTheory.
// Corrected some minor buglets in NumTheory.
//
// Revision 1.13  2009/11/26 16:18:00  bigatti
// -- including string ZZ.C instead of ZZ.H
//
// Revision 1.12  2009/10/08 13:41:19  abbott
// Renamed "round" into "RoundDiv".
// Added two new signatures for "RoundDiv".
//
// Revision 1.11  2009/06/11 14:05:29  abbott
// CLEANING: Removed several functions which are now gathered in NumTheory.H/C
//           (for example: gcd, lcm, PowerMod, InvMod).
//
// Revision 1.10  2009/06/05 12:08:28  abbott
// Changed return type of operator%(ZZ,MachineInteger); it is now unsigned long
// instead of ZZ.
//
// Revision 1.9  2008/12/16 21:14:47  abbott
// In functions taking a machine integer changed arg type from MachineInteger
// to const-ref-MachineInteger.
//
// Revision 1.8  2008/11/18 10:25:48  abbott
// Added function round.
//
// Revision 1.7  2008/04/22 13:09:16  abbott
// Removed IsPositive and IsNegative functions for ZZs.
// Comparison between RingElem and 0 is now handled specially (specially fast?).
//
// Revision 1.6  2007/10/30 17:14:11  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.5  2007/05/22 22:51:39  abbott
// Changed name of fn ndigits to NumDigits.
// Changed return type of exponent and NumDigits.
// Changed some exceptions from nonstandard to the appropriate standard one.
//
// Revision 1.4  2007/05/21 12:57:28  abbott
// New class for passing machine integers as args; includes some simple
// operations on machine integers (cmp, gcd, IsNegative,...).  Operations
// between ZZ and machine integers modified to use the new class.  Inexact
// integer division (of a ZZ) by a negative value now triggers an error;
// new error for reporting inexact integer division by a negative value.
//
// Revision 1.3  2007/03/23 18:38:43  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.2  2007/03/16 17:43:05  abbott
// Added new convert function (from ZZ to double).
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.10  2007/03/08 18:42:05  cocoa
// Cleaned up whitespace.
//
// Revision 1.9  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.8  2006/12/06 17:17:44  cocoa
// -- removed #include "config.H"
//
// Revision 1.7  2006/11/29 11:59:35  cocoa
// -- fixed: convert(double& z, const ZZ& num, const ZZ& den) now returns
//    bool (was void) and does not throw errors
//
// Revision 1.6  2006/11/27 13:38:18  cocoa
// -- added #include <cstddef> //using std::size_t
//
// Revision 1.5  2006/11/27 13:06:23  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.4  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.3  2006/09/29 11:46:54  cocoa
// Corrected bug in convert(ZZ, ZZ, double) -- now correct and simpler.
// Previously went into infinite loop on negative doubles.
//
// Revision 1.2  2006/06/21 17:08:04  cocoa
// Removed a pointless const.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.4  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.3  2006/03/27 12:21:26  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.2  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.2  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.2  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.7  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.6  2004/11/11 13:20:47  cocoa
// -- added \include *.txt for doxygen
//
// Revision 1.5  2004/11/04 18:47:42  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.4  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.3  2004/02/03 16:16:20  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.2  2003/10/17 10:51:06  cocoa
// Major cleaning, and new naming convention.
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
// Revision 1.2  2003/06/23 16:12:44  abbott
// Minor cleaning prior to public release.
//
// Revision 1.1  2003/05/14 17:12:35  abbott
// Initial revision
//

#endif
