#ifndef CoCoA_matrix_H
#define CoCoA_matrix_H

//   Copyright (c)  2005,2008  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ring.H"
#include "CoCoA/MachineInteger.H"
#include "CoCoA/ZZ.H"

#include <iosfwd>
//using std::ostream;
#include <cstddef>
//using std::size_t;
#include <vector>
//using std::vector;


namespace CoCoA
{

  class OpenMathOutput; // forward declaration -- defined in OpenMath.H


  /*-----------------------------------------------------------------*/
  /** \include matrix.txt  */
  /*-----------------------------------------------------------------*/


  // Operations on vectors -- should be ModuleElems?  (of a FreeModule?)
  bool IsZero(const std::vector<RingElem>& v);

  class matrix;              // forward declaration
  class MatrixBase;          // forward declaration
  class MatrixViewBase;      // forward declaration
  class ConstMatrixViewBase; // forward declaration

  class ConstMatrixView
  {
  public:
    explicit ConstMatrixView(const ConstMatrixViewBase* MatPtr);
    // Default copy ctor is OK.
    // Assignment disabled (because SmartPtrIRC has no assignment).
    // Default dtor is OK.
    const ConstMatrixViewBase* operator->() const;
    ConstRefRingElem operator()(const MachineInteger& i, const MachineInteger& j) const;

  protected: // data member
    SmartPtrIRC<ConstMatrixViewBase> mySmartPtr;
    friend void swap(matrix& M1, matrix& M2);
  };


  class MatrixView: public ConstMatrixView
  {
    // Data member inherited from ConstMatrixView
  public:
    explicit MatrixView(MatrixViewBase* MatPtr);
    // Default copy ctor is OK.
    // Assignment disabled.
    // Default dtor is OK.
    MatrixViewBase* operator->() const;
  };


  class matrix: public MatrixView
  {
    // Data member inherited from ConstMatrixView via MatrixView
  public:
    explicit matrix(MatrixBase* MatPtr);
    // Default copy ctor is OK.
    // Assignment disabled.
    // Default dtor is OK.
    MatrixBase* operator->() const;

//   protected:
//     MatrixBase* myMatrixPtr(); // Needed by concrete matrix impl classes.
  };





  class ConstMatrixViewBase: protected IntrusiveReferenceCount
  {
    friend class SmartPtrIRC<ConstMatrixViewBase>;

  protected: // Hide default ctor & dtor from public.
    ConstMatrixViewBase() {}
    virtual ~ConstMatrixViewBase() {}

  public:
    typedef std::vector<RingElem> vec; ///???? is this the right thing to do???
    virtual const ring& myBaseRing() const = 0;
    virtual std::size_t myNumRows() const = 0;
    virtual std::size_t myNumCols() const = 0;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const = 0;
    void myCheckRowIndex(const MachineInteger& i, const char* where="") const;  // throws if index is too large
    void myCheckColIndex(const MachineInteger& j, const char* where="") const;  // throws if index is too large
    virtual bool myIsZeroRow(std::size_t i) const;         ///< tests whether row i is zero -- default "dense" definition
    virtual bool myIsZeroCol(std::size_t j) const;         ///< tests whether col j is zero -- default "dense" definition
    virtual void myMulByRow(vec& lhs, const vec& v) const; ///< lhs = v*M -- vector-by-matrix product, default "dense" definition
    virtual void myMulByCol(vec& lhs, const vec& v) const; ///< lhs = M*v -- matrix-by-vector product, default "dense" definition
    virtual void myDet(RefRingElem d) const;               ///< default method uses Gaussian reduction
    virtual std::size_t myRank() const;                    ///< default method uses Gaussian reduction
    virtual void myOutputSelf(std::ostream& out) const;    ///< has a default "dense" definition
    virtual void myOutputSelf(OpenMathOutput& OMOut) const;///< has a default "dense" definition
  };


  class MatrixViewBase: public ConstMatrixViewBase
  {
  protected: // Hide default ctor & dtor from public.
    MatrixViewBase() {}
    virtual ~MatrixViewBase() {}

  public:
    virtual void myAssignZero() = 0; ///< sets all entries to zero
    virtual bool myIsWritable(std::size_t i, std::size_t j) = 0; ///< true iff location (i,j) is writable
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j) = 0; ///< may be used only if (i,j) is writable
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r) = 0;
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n) = 0;
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N) = 0;
  };


  class MatrixBase: public MatrixViewBase
  {
  protected: // Hide default ctor & dtor from public.
    MatrixBase() {}
    virtual ~MatrixBase() {}

  public:
    virtual MatrixBase* myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const = 0;
    typedef std::vector<RingElem> vec;
    virtual void myResize(std::size_t NumRows, std::size_t NumCols) = 0;
    virtual void myRowMul(std::size_t i, ConstRefRingElem c) = 0; ///< row(i) = c*row(i)
    virtual void myColMul(std::size_t j, ConstRefRingElem c) = 0; ///< col(j) = c*col(j)
    virtual void myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c) = 0; ///< row(i1) += c*row(i2)
    virtual void myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c) = 0; ///< col(j1) += c*col(j2)
    virtual void mySwapRows(std::size_t i1, std::size_t i2) = 0;
    virtual void mySwapCols(std::size_t j1, std::size_t j2) = 0;
  };




  std::ostream& operator<<(std::ostream& out, const ConstMatrixView& M);
  bool IsZeroRow(const ConstMatrixView& M, const MachineInteger& i);
  bool IsZeroCol(const ConstMatrixView& M, const MachineInteger& j);
  void AssignZero(MatrixView& M);
  void SetEntry(MatrixView& M, const MachineInteger& i, const MachineInteger& j, ConstRefRingElem r);
  void SetEntry(MatrixView& M, const MachineInteger& i, const MachineInteger& j, const MachineInteger& n);
  void SetEntry(MatrixView& M, const MachineInteger& i, const MachineInteger& j, const ZZ& N);
  void swap(matrix& M1, matrix& M2);


  /////////////////////////////////////////////////////////////////////////////
  // Inline functions

  inline ConstMatrixView::ConstMatrixView(const ConstMatrixViewBase* MatPtr):
      mySmartPtr(const_cast<ConstMatrixViewBase*>(MatPtr))
  {}


  inline const ConstMatrixViewBase* ConstMatrixView::operator->() const
  { return mySmartPtr.operator->(); }



  inline MatrixView::MatrixView(MatrixViewBase* MatPtr):
      ConstMatrixView(MatPtr)
  {}


  inline MatrixViewBase* MatrixView::operator->() const
  { return dynamic_cast<MatrixViewBase*>(const_cast<ConstMatrixViewBase*>(ConstMatrixView::operator->())); }



  inline matrix::matrix(MatrixBase* MatPtr):
      MatrixView(MatPtr)
  {}


  inline MatrixBase* matrix::operator->() const
  { return dynamic_cast<MatrixBase*>(MatrixView::operator->()); }


//   inline MatrixBase* matrix::myMatrixPtr()
//   {
//     // dynamic_cast would perform needless checks
//     // reinterpret_cast might not be portable?
//     return static_cast<MatrixBase*>(const_cast<MatrixViewBase*>(myPtr));
//   }


  inline const ring& BaseRing(const ConstMatrixView& M)
  {
    return M->myBaseRing();
  }


  inline std::size_t NumRows(const ConstMatrixView& M)
  {
    return M->myNumRows();
  }


  inline std::size_t NumCols(const ConstMatrixView& M)
  {
    return M->myNumCols();
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/matrix.H,v 1.5 2008/12/17 12:11:52 abbott Exp $
// $Log: matrix.H,v $
// Revision 1.5  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.4  2008/04/16 17:24:17  abbott
// Further cleaning of the new matrix code.  Updated documentation too.
//
// Revision 1.3  2008/04/08 15:26:42  abbott
// Major revision to matrix implementation: added matrix views.
// Lots of changes.
//
// Revision 1.2  2007/10/30 17:14:11  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2006/12/06 17:18:40  cocoa
// -- removed #include "config.H"
//
// Revision 1.4  2006/11/24 17:41:18  cocoa
// -- reorganized includes of header files
//
// Revision 1.3  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.2  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.4  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.3  2006/04/27 13:45:30  cocoa
// Changed name of NewIdentityRingHom to NewIdentityHom.
// Changed name of member functions which print out their own object
// into myOutputSelf (to distinguish from "transitive" myOutput fns).
//
// Revision 1.2  2006/03/27 12:21:26  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.11  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.10  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.9  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.8  2005/03/31 16:59:42  cocoa
// Made special matrix ctors private, so a user has to pass via the
// pseudo-ctors (which do all the arg sanity checking).
//
// Revision 1.7  2005/03/30 17:15:14  cocoa
// Cleaned the SpecialMatrix code; a simple test compiles and
// runs fine.
//
// Revision 1.6  2005/03/29 17:36:47  cocoa
// Just checking in before going home -- test-matrix1 does not yet link!
//
// Revision 1.5  2005/03/29 17:19:30  cocoa
// -- added: rank and myRank
//
// Revision 1.4  2005/03/11 16:44:18  cocoa
// New abstract class structure for matrices.
// New types of special matrix.
//
// Revision 1.3  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.9  2004/11/29 16:19:02  cocoa
// -- changed syntax for function det
//
// Revision 1.8  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//

#endif
