//   Copyright (c)  2006,2008-2009  John Abbott
//   Main authors: Laura Torrente (assisted by John Abbott)

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ApproxPts2.H"
#include "CoCoA/CanonicalHom.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/MatrixArith.H"
#include "CoCoA/MatrixViews.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/QBGenerator.H"
#include "CoCoA/QQ.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/RingTwinFloat.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/apply.H"
#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/io.H"
#include "CoCoA/matrix.H"
#include "CoCoA/ring.H"
#include "CoCoA/symbol.H"

#include <algorithm>
// using std::swap;   in  SOI, NBM
#include <iostream>
using std::ostream;
using std::endl;
//#include <vector>  --- already included by ApproxPts2.H
using std::vector;
#include <list>
using std::list;


namespace CoCoA
{

  namespace ApproxPts
  {

    namespace // anonymous namespace for file local auxiliary functions and definitions.
    {

      //------------------------------------------------------------------
      // Auxiliary functions for SOI

      /// Convert the input data to the RingTwinFloat with current BitPrec
      // !!! PROTECT AGAINST ALIASING!!!
      void ConversionToRR(const vector<RingElem>& OrigTolerance,
                          const vector<PointR>& OrigPts,
                          ConstRefRingElem OrigGamma,
                          vector<RingElem>& tolerance,
                          vector<PointR>& pts,
                          RingElem& gamma,
                          const ring& RR)
      {
        ring Q = owner(OrigTolerance.front());
        RingHom QToRR = CanonicalHom(Q,RR);
        const size_t dim = OrigTolerance.size();
        tolerance.clear();
        for (size_t i=0; i < dim; ++i)
          tolerance.push_back(QToRR(OrigTolerance[i]));
        pts.clear();
        for (size_t i=0; i < OrigPts.size(); ++i)
        {
          PointR NewPt;
          for (size_t j=0; j < dim; ++j)
            NewPt.push_back(QToRR(OrigPts[i][j]));
          pts.push_back(NewPt);
        }
        gamma = QToRR(OrigGamma);
      }

      void ConversionToRR(const vector<double>& OrigTolerance,
                          const vector<PointDbl>& OrigPts,
                          double OrigGamma,
                          vector<RingElem>& tolerance,
                          vector<PointR>& pts,
                          RingElem& gamma,
                          const ring& RR)
      {
        QQ Q; // used for converting doubles into exact rationals
        tolerance.clear();
        for (size_t i=0; i < OrigTolerance.size(); ++i)
        {
          convert(Q, OrigTolerance[i]);
          tolerance.push_back(RingElem(RR,Q));
        }
        pts.clear();
        for (size_t j=0; j < OrigPts.size(); ++j)
        {
          vector<RingElem> tmp;
          for (size_t i=0; i < OrigPts[j].size(); ++i)
          {
            convert(Q, OrigPts[j][i]);
            tmp.push_back(RingElem(RR,Q));
          }
          pts.push_back(tmp);
        }
        convert(Q, OrigGamma);
        gamma = RingElem(RR,Q);
      }


//     matrix MatrixEmbedding(ConstMatrixView M, const PolyRing& P)
//     {
//       //ring R = BaseRing(M);
//       //ERROR : If (CoeffRing(P) != R)   then error!!!
//       //ERROR : If they have different size   then error!!
//       matrix NewM(NewDenseMat(P, NumRows(M), NumCols(M)));
//       RingHom phi = CoeffEmbeddingHom(P);
//       for (size_t i=0; i < NumRows(M); ++i)
//         for (size_t j=0; j < NumCols(M); ++j)
//           SetEntry(NewM, i, j, phi(M(i,j)));
//       return NewM;
//     }


//     vector<RingElem> VectorEmbedding(const vector<RingElem>& V,
//                                      const PolyRing& P)
//     {
//       //ring R = owner(V.front());
//       //ERROR : If (CoeffRing(P) != R)   then error!!!
//       RingHom phi = CoeffEmbeddingHom(P);
//       vector<RingElem> NewV;
//       for (size_t i=0; i < V.size(); ++i)
//         NewV.push_back(phi(V[i]));
//       return NewV;
//     }


      // square 2-norm of a vector
      RingElem SquareNorm(const vector<RingElem>& v)
      {
        CoCoA_ASSERT(!v.empty());
        RingElem tmpSqNorm(zero(owner(v.front())));
        for (size_t i=0; i < v.size(); ++i)
          tmpSqNorm += power(v[i], 2);
        return tmpSqNorm;
      }


      // Evaluation of a poly f on a set Pts of points
      void Evaluation(vector<RingElem>& b,
                      ConstRefRingElem f,
                      const list<RingHom>& EvalAtPts)
      {
        b.clear();
        for (list<RingHom>::const_iterator it = EvalAtPts.begin(); it != EvalAtPts.end(); ++it)
          b.push_back((*it)(f));
      }


      void FirstOrderEval(vector<RingElem>& V,
                          ConstRefRingElem t,
                          const list<RingHom>& EvalAtPts,
                          const vector<RingElem>& tolerance,
                          const PolyRing& PErr)
      {
        RingHom phi = CoeffEmbeddingHom(PErr);
        const size_t dim = tolerance.size();
        const size_t NumPts = EvalAtPts.size();
        vector<RingElem> tmp;
        const vector<RingElem>& e = indets(PErr);

        V.assign(NumPts, zero(PErr));
        for (size_t i=0; i < dim; ++i)
        {
          if (!IsZero(tolerance[i]))
          {
            Evaluation(tmp, deriv(t,i), EvalAtPts);
            for (size_t j=0; j < NumPts; ++j)
              V[j] += phi(tmp[j]) * e[j*dim + i];
          }
        }
      }


      // Solve the LS problem A*x = b, A orthogonal matrix
      // implemented without deleting the denominators
      void LeastSquares(vector<RingElem>& x,
                        vector<RingElem>& Rho,
                        ConstMatrixView A,
                        ConstMatrixView DiagA,
                        const vector<RingElem>& b)
      {
        ring R = owner(b.front());
        const size_t n = b.size();
        vector<RingElem> RhoRho(n, zero(R));

        x = vector<RingElem>(NumRows(A),zero(R));
        A->myMulByCol(x,b);
        for (size_t i=0; i < n; ++i)
          if (DiagA(i,i) != 0)
            x[i] /= DiagA(i,i);
          else
            x[i] = zero(R);
        A->myMulByRow(RhoRho,x);
        Rho = b;
        for (size_t i = 0; i < n; ++i)
          Rho[i] -= RhoRho[i];
      }


      // First order least squares algorithm
      void FirstOrderLS(vector<RingElem>& Rho1,
                        vector<RingElem>& x1,
                        ConstMatrixView A0,
                        ConstMatrixView A1,
                        ConstMatrixView DiagA0,
                        ConstMatrixView A0A1,
                        const vector<RingElem>& b0,
                        const vector<RingElem>& b1,
                        const vector<RingElem>& x0)
      {
        SparsePolyRing P = AsSparsePolyRing(owner(b1.front()));
        RingHom phi = CoeffEmbeddingHom(P);
        matrix A0InP = apply(phi, A0);
        const vector<RingElem> x0InP = apply(phi, x0);//VectorEmbedding(x0, P);

        // x1 = A0*b1;
        x1.assign(NumRows(A0), zero(P));
        A0InP->myMulByCol(x1,b1);
        // B = A1*b0;
        vector<RingElem> B(NumRows(A1), zero(P));
        A1->myMulByCol(B, apply(phi, b0));
        // C = A0A1*x0;
        vector<RingElem> C(NumRows(A0A1), zero(P));
        A0A1->myMulByCol(C,x0InP);
        for (size_t i=0; i < b1.size(); ++i)
        {
          if (IsZero(DiagA0(i,i)))
            x1[i] = zero(P);
          else
          {
            x1[i] += (B[i] - C[i]);
            x1[i] *= phi(1/DiagA0(i,i));
          }
        }
        Rho1 = b1;
        // D = A0*x1;
        vector<RingElem> D(NumRows(A0A1), zero(P));
        A0InP->myMulByRow(D,x1);
        // E =  A1*x0;
        vector<RingElem> E(NumRows(A1), zero(P));
        A1->myMulByRow(E,x0InP);
        for (size_t i=0; i < b1.size(); ++i)
          Rho1[i] -= (D[i] + E[i]);
      }


      // Compute the matrix C that represents the vector V
      void ErrorMatrix(MatrixView& C, const vector<RingElem>& V)
      {
        SparsePolyRing P = AsSparsePolyRing(owner(V.front()));
        RingHom PToR = EvalHom(P, vector<RingElem>(NumIndets(P), zero(CoeffRing(P))));
        for (size_t i=0; i < V.size(); ++i)
          for (size_t j=0; j < NumIndets(P); ++j)
            SetEntry(C,i,j,PToR(deriv(V[i],j)));
      }


// //Compute an estimate of the squared minimal singular value of a matrix
// void LowerBoundMinSingValue(RingElem& LowerBound,
//                             ConstMatrixView A)
// {
//   list<RingElem> listOfNorms;
//   RingElem norm(BaseRing(A));
//   inftyMatrixNorm(norm, A);
//   listOfNorms.push_back(1/(NumRows(A)*norm*norm));
//   GlobalOutput() << "LB1 = " << 1/(NumRows(A)*norm*norm) << endl;
//   inftyMatrixNorm(norm, transpose(A));
//   listOfNorms.push_back(1/(NumCols(A)*norm*norm));
//   GlobalOutput() << "LB2 = " << 1/(NumCols(A)*norm*norm) << endl;
//   FrobeniusNorm(norm, A);
//   listOfNorms.push_back(1/norm);
//   GlobalOutput() << "LB3 = " << 1/norm << endl;
//   listOfNorms.sort();
//   list<RingElem>::iterator it = listOfNorms.end();
//   LowerBound = *(--it);
//   GlobalOutput() << "Lower Bound is = " << LowerBound << endl;
// }



      // Compute the vector of rows of the matrix A such that the minimal
      // singular value of the corresp submatrix is greater than ||Eps||_2
      void WellIndepRows(vector<size_t>& rows,
                         ConstRefRingElem Eps_square,
                         ConstMatrixView A)
      {
        const ring R = BaseRing(A);
        vector<size_t> cols;
        for (size_t j=0; j < NumCols(A); ++j)
          cols.push_back(j);
        rows.empty();
        for (size_t i=0; i < NumRows(A); ++i)
        {
          rows.push_back(i);
          ConstMatrixView B = submat(A, rows, cols);
          // Check lower bound for smallest singular value of B
          if (rank(B) < NumRows(B) || 1/FrobeniusNorm2(PseudoInverse(B)) < Eps_square)
          {
            rows.pop_back();
          }
        }
      }



      // Compute minimal 2-norm solution of a Full Rank Underdetermined System
      void SolveFRUDS(vector<RingElem>& x,
                      ConstMatrixView A,
                      const vector<RingElem>& b)
      {
        ConstMatrixView PseudoInvA = PseudoInverse(A);
        // x = PseudoInvA*b;
        x.assign(NumRows(PseudoInvA), zero(BaseRing(A)));
        PseudoInvA->myMulByCol(x, b);
      }



      // Add to QB and update
      void AddToQBAndUpdate(const PPMonoidElem& t,
                            QBGenerator& QBG,
                            const vector<RingElem>& Alpha0,
                            vector<RingElem>& S)
      {
        SparsePolyRing P = AsSparsePolyRing(owner(S.front()));
        RingHom RToP = CoeffEmbeddingHom(P);
        const RingElem tAsPoly = monomial(P, one(CoeffRing(P)), t);

        GlobalLogput() << "The term " << t << " is added to QB" << endl;
        const size_t lenQB = QBG.myQB().size();
        S.push_back(tAsPoly);
        for (size_t i=0; i < lenQB; ++i)
          S[lenQB] -= RToP(Alpha0[i])*S[i];
        QBG.myCornerPPIntoQB(t);
      }


      // Update evaluation matrices
      void UpdateMatrices(const QBGenerator& QBG,
                          MatrixView& A0,
                          MatrixView& A1,
                          MatrixView& DiagA0,
                          MatrixView& A0A1,
                          const vector<RingElem>& Rho0,
                          const vector<RingElem>& Rho1)
      {
        PolyRing PErr = AsPolyRing(BaseRing(A1));
        RingHom coeff = CoeffEmbeddingHom(PErr);
        const size_t k = QBG.myQB().size();
        SetEntry(DiagA0, k, k, SquareNorm(Rho0));
        for (size_t j=0; j < Rho0.size(); ++j)
        {
          SetEntry(A0, k, j, Rho0[j]);
          SetEntry(A1, k, j, Rho1[j]);
          SetEntry(A0A1, k, k, A0A1(k, k) + 2*Rho1[j]*coeff(Rho0[j]));
        }
      }



      // Update evaluation matrices
      void UpdateMatricesNBM(const QBGenerator& QBG,
                             MatrixView& M,
                             MatrixView& A,
                             MatrixView& DiagA,
                             const vector<RingElem>& V,
                             const vector<RingElem>& Rho)
      {
        const size_t k = QBG.myQB().size();
        SetEntry(DiagA, k, k, SquareNorm(Rho));
        for (size_t j=0; j < Rho.size(); ++j)
        {
          SetEntry(M, k, j, V[j]);
          SetEntry(A, k, j, Rho[j]);
        }
      }



      // Add to Corner set and update
      void AddToCornerAndUpdate(const PPMonoidElem& t,
                                QBGenerator& QBG,
                                const vector<RingElem>& Alpha0,
                                const vector<RingElem>& S,
                                vector<RingElem>& J)
      {
        SparsePolyRing P = AsSparsePolyRing(owner(S.front()));
        RingHom RToP = CoeffEmbeddingHom(P);
        const RingElem tAsPoly = monomial(P, one(CoeffRing(P)), t);

        GlobalLogput() << "The term " << t << " is added to Corner" << endl;
        RingElem f(tAsPoly);
        for (size_t i=0; i < S.size(); ++i)
          f -= RToP(Alpha0[i])*S[i];
        J.push_back(f);
        QBG.myCornerPPIntoAvoidSet(t);
      }



      // Compute the (first order) border of an order ideal
      void ComputeBorderPPs(list<PPMonoidElem>& BorderPPs,
                            const QBGenerator& QBG)
      {
        const vector<PPMonoidElem> QB = QBG.myQB();
        BorderPPs.clear();
        if (QB.empty()) // can this ever happen???
        {
          BorderPPs.push_back(one(PPM(QBG)));
          return;
        }
        const vector<PPMonoidElem>& X = indets(PPM(QBG));
        const size_t NumIndets = X.size();
        for (size_t i=0; i < QB.size(); ++i)
          for (size_t j=0; j < NumIndets; ++j)
            BorderPPs.push_back(QB[i]*X[j]);
        BorderPPs.sort();
        BorderPPs.unique();
        for (size_t i=0; i < QB.size(); ++i)
          BorderPPs.remove(QB[i]);
      }


      // Compute the poly f = t-sum x_1 S_i
      void ComputeBorderPoly(RingElem& f,
                             const PPMonoidElem& t,
                             ConstMatrixView A,
                             ConstMatrixView DiagA,
                             const list<RingHom>& EvalAtPts,
                             const vector<RingElem>& S)
      {
        SparsePolyRing P = AsSparsePolyRing(owner(S.front()));
        RingHom phi = CoeffEmbeddingHom(P);
        f = monomial(P, 1, t);
        vector<RingElem> b, x, Rho;
        Evaluation(b, f, EvalAtPts);
        LeastSquares(x, Rho, A, DiagA, b);
        for (size_t i=0; i < S.size(); ++i)
          f -= phi(x[i])*S[i];
      }



      //------------------------------------------------------------------
      // Auxiliary functions for NBM

      // Build the matrix of absolute values
      matrix AbsMatrix(ConstMatrixView M)
      {
        matrix AbsM = NewDenseMat(BaseRing(M), NumRows(M), NumCols(M));
        for (size_t i=0; i < NumRows(M); ++i)
          for (size_t j=0; j < NumCols(M); ++j)
            SetEntry(AbsM, i, j, abs(M(i,j)));
        return AbsM;
      }


      // Build the vector of absolute values
      vector<RingElem> AbsVector(const vector<RingElem>& V)
      {
        vector<RingElem> AbsV;
        for (size_t i=0; i < V.size(); ++i)
          AbsV.push_back(abs(V[i]));
        return AbsV;
      }


      // Computing the bound for the sufficient condition...
      void NBMBound(vector<RingElem>& Bound,
                    ConstMatrixView M,
                    const vector<RingElem>& tolerance,
                    const RingElem& g,
                    const list<RingHom>& EvalAtPts)
      {
        const size_t NumPts = EvalAtPts.size();
        ring R = BaseRing(M);
        RingHom phi = CanonicalHom(R, owner(g));
        matrix id = NewDenseMat(IdentityMat(R, NumPts));
        matrix H = transpose(M)*PseudoInverse(transpose(M));
        matrix C = NewDenseMat(IdentityMat(R, NumPts));
        for (size_t i=0; i < NumPts; ++i)
          for (size_t j=0; j < NumPts; ++j)
            SetEntry(C, i, j, id(i,j)-H(i,j));
        matrix CAbs = AbsMatrix(C);

        vector<RingElem> D(NumPts, zero(R));
        for (size_t i=0; i < tolerance.size(); ++i)
        {
          vector<RingElem> V;
          Evaluation(V, phi(tolerance[i])*deriv(g,i), EvalAtPts);
          vector<RingElem> AbsV = AbsVector(V);
          for (size_t j=0; j < NumPts; ++j)
            D[j] += AbsV[j];
        }
        Bound.assign(NumPts, zero(R));
        CAbs->myMulByCol(Bound, D);
      }


    } // end of anonymous namespace





    void SOI(std::vector<PPMonoidElem>& QB,
             std::vector<RingElem>& BBasis,
             std::vector<RingElem>& AlmostVanishing,
             const std::vector<PointR>& pts,
             const std::vector<RingElem>& tolerance,
             ConstRefRingElem gamma)
    {
      ring R = owner(pts[0][0]);                      // R must be an ordered field.
      CoCoA_ASSERT(IsOrderedDomain(R) && IsField(R)); //
      const size_t dim = tolerance.size();
      const size_t NumPts = pts.size();

      // Build poly rings P=QQ[x_1...x_n] and
      // PErr = QQ[e[1,1],...,e[NumPts,dim]]
      PPOrdering PPO = NewStdDegRevLexOrdering(dim);
      PPMonoid PPM = NewPPMonoid(SymbolRange("x", 0, dim-1), PPO);
      SparsePolyRing P = NewPolyRing(R, PPM);
      SparsePolyRing PErr = NewPolyRing(R, SymbolRange(symbol("e",1,1), symbol("e",NumPts,dim)));
      // Build evaluation homomorphisms at Pts
      list<RingHom> EvalAtPts;///  EvalAtPts.reserve(NumPts);
      for (size_t i=0; i < NumPts; ++i)
        EvalAtPts.push_back(EvalHom(P, pts[i]));

      //   --------------     INITIALIZE   ---------------
      size_t IterNum = 1;
      const RingElem Eps_square = SquareNorm(tolerance);
      const RingElem kappa = gamma+1;
      const RingElem Bound = kappa*kappa*NumPts*Eps_square;
      vector<RingElem> J, S;     //J = almost vanishing polys

      QBGenerator QBG(PPM);      //to handle quotient bases
      QBG.myCornerPPIntoQB(one(PPM));
      S.push_back(one(P));
      list<PPMonoidElem> L = QBG.myCorners();  //L = list of PP

      // Build and initialize matrices A0, A1 DiagA0, A0A1, CErr
      matrix A0(NewDenseMat(R, NumPts, NumPts));   // A0 = orthogonal
      for (size_t i=0; i < NumPts; ++i)
        SetEntry(A0, 0, i, 1);

      matrix A1(NewDenseMat(PErr, NumPts, NumPts));

      vector<RingElem> DiagA0Elems(NumPts, zero(R));
      MatrixView DiagA0 = DiagMat(DiagA0Elems); // DiagA0 = A0*A0'
      SetEntry(DiagA0, 0, 0, NumPts);

      vector<RingElem> A0A1Elems(NumPts, zero(PErr));
      MatrixView A0A1 = DiagMat(A0A1Elems); // A0A1 = 2*A0*A1'


      // ------------   MAIN CYCLE   -----------------
      while (!L.empty())
      {
        GlobalLogput() << "---------------------------------------" << endl
                       << "Iteration " << IterNum << endl
                       << "Quotient Basis = " << QBG.myQB() << endl
                       << "List of power products = " << L << endl;

        const PPMonoidElem t = L.front(); // PP considered in the current step
        GlobalLogput() << "Next PP = " << t << endl;
        const RingElem tAsPoly = monomial(P, one(R), t);

        // Compute eval V0=t(Pts) and solve LS problem A0'*b = V0
        vector<RingElem> V0, V1, Alpha0, Alpha1, Rho0, Rho1;
        Evaluation(V0, tAsPoly, EvalAtPts);
        LeastSquares(Alpha0, Rho0, A0, DiagA0, V0);
        // First trivial check
        if (IsZero(Rho0))
        {
          AddToCornerAndUpdate(t, QBG, Alpha0, S, J);
          GlobalLogput() << "Truly vanishing polynomial f = " << J.back() << endl;
          ++IterNum;
          L = QBG.myCorners();
          continue;
        }

        // Compute f.o.eval V1, solve (f.o.error anal.)LS problem
        FirstOrderEval(V1, tAsPoly, EvalAtPts, tolerance, PErr);
        FirstOrderLS(Rho1, Alpha1, A0, A1, DiagA0, A0A1, V0, V1, Alpha0);
        // Compute CErr such that Rho1 = CErr*e
        matrix CErr(NewDenseMat(R, NumPts, NumPts*dim));//CErr=Rho1(matrix)
        ErrorMatrix(CErr, Rho1);
        // Compute the strongly independent rows of CErr
        vector<size_t> rows;
        WellIndepRows(rows, Eps_square, CErr);
        GlobalLogput() << "Well Independent Rows =" << rows << endl;

        // if, for each i, ||CErr[i]|| < Eps ...
        if (rows.empty())
        {
          if (SquareNorm(Rho0) > Bound*Eps_square)
          {
            UpdateMatrices(QBG, A0, A1, DiagA0, A0A1,Rho0,Rho1);
            AddToQBAndUpdate(t, QBG, Alpha0, S);
          }
          else
          {
            AddToCornerAndUpdate(t, QBG, Alpha0, S, J);
            GlobalLogput() << "Almost vanishing polynomial: f = " << J.back() << endl;
          }
        }
        else
        {
          // Compute the submatrix of CErr with rows in rows
          vector<size_t> cols;
          for (size_t j=0; j < NumCols(CErr); ++j)
            cols.push_back(j);
          ConstMatrixView CErr1 = submat(CErr, rows, cols);
          vector<RingElem> MinusRho01;
          for (size_t i=0; i < rows.size(); ++i)
            MinusRho01.push_back(-Rho0[rows[i]]);
          vector<RingElem> eLS1;
          // Compute the minimal 2-norm solution of the
          // underdetermined system CErr1 * eLS1 = MinusRho01
          SolveFRUDS(eLS1, CErr1, MinusRho01);
          GlobalLogput() << "Square norm eLS1= " << SquareNorm(eLS1) << endl
                         << "Bound = " << Bound << endl;

          //Check and update the variables
          if (SquareNorm(eLS1) >= Bound)
          {
            UpdateMatrices(QBG, A0, A1, DiagA0, A0A1,Rho0,Rho1);
            AddToQBAndUpdate(t, QBG, Alpha0, S);
          }
          else
          {
            AddToCornerAndUpdate(t, QBG, Alpha0, S, J);
            GlobalLogput() << "Almost vanishing polynomial f = " << J.back() << endl;
          }
        }
        ++IterNum;
        L = QBG.myCorners();
      }

      GlobalLogput() << "==========================================" << endl
                     << endl
                     << "All elements of the SOI have been found" << endl;

      // Computation of the Border Basis and printing
      BBasis.clear();
      std::swap(AlmostVanishing, J); // morally AlmostVanishing = J;
      QB.clear(); QB = QBG.myQB(); // cannot use swap trick because myQB is read only
      if (QB.size() < NumPts)
      {
        GlobalLogput() << "Found " << QB.size() << " elements of QB" << endl
                       << "A finer tolerance should be used!" << endl
                       << endl
                       << "Almost vanishing polys = " << AlmostVanishing << endl;
        return;
      }

      list<PPMonoidElem> BorderPPs;
      ComputeBorderPPs(BorderPPs, QBG);
      for (list<PPMonoidElem>::const_iterator it=BorderPPs.begin(); it != BorderPPs.end(); ++it)
      {
        RingElem BBElem(P);
        ComputeBorderPoly(BBElem, *it, A0, DiagA0, EvalAtPts, S);
        BBasis.push_back(zero(P)); swap(BBasis.back(), BBElem); // morally  BBasis.push_back(f);
      }
    }


    void SOITwinFloat(std::vector<PPMonoidElem>& QB,
                      std::vector<RingElem>& BBasis,
                      std::vector<RingElem>& AlmostVanishing,
                      const std::vector<PointR>& OrigPts,
                      const std::vector<RingElem>& OrigTolerance,
                      ConstRefRingElem OrigGamma)
    {
      for (size_t BitPrec = 128; ; BitPrec *= 2)
      {
        try
        {
          ring RR = NewRingTwinFloat(BitPrec);
          vector<RingElem> tolerance;
          vector<PointR> pts;
          RingElem gamma(RR);
          ConversionToRR(OrigTolerance, OrigPts, OrigGamma, tolerance, pts, gamma, RR);
          SOI(QB, BBasis, AlmostVanishing, pts, tolerance, gamma);
          GlobalLogput() << "BBasis successfully computed with " << BitPrec << " bit precision " << endl;
          return;
        }
        catch (const RingTwinFloat::InsufficientPrecision&)
        {
          // If verbose mode, inform the user about the failure...
          GlobalLogput() << endl << "----------------------------------"
                         << endl << "A bit precision of " << BitPrec
                         << " was not sufficient." << endl
                         << "------------------------------------------"
                         << endl << endl;
        }
      }
    }


    void SOI(std::vector<PPMonoidElem>& QB,
             std::vector<RingElem>& BBasis,
             std::vector<RingElem>& AlmostVanishing,
             const std::vector<PointDbl>& OrigPts,
             const std::vector<double>& OrigTolerance,
             double OrigGamma)
    {
      ring Q = RingQ();
      vector<RingElem> tolerance;
      vector<PointR> pts;
      RingElem gamma(Q);
      ConversionToRR(OrigTolerance, OrigPts, OrigGamma, tolerance, pts, gamma, Q);
      SOITwinFloat(QB, BBasis, AlmostVanishing, pts, tolerance, gamma);
    }




//--------------  NBM  ---------------------------------------------



    void NBM(std::vector<PPMonoidElem>& QB,
             std::vector<RingElem>& BBasis,
             std::vector<RingElem>& AlmostVanishing,
             const std::vector<PointR>& pts,
             const std::vector<RingElem>& tolerance)
    {
      ring R = owner(pts[0][0]);                      // R must be an ordered field
      CoCoA_ASSERT(IsOrderedDomain(R) && IsField(R)); //
      const size_t dim = tolerance.size();
      const size_t NumPts = pts.size();

      // Build poly ring P=Q[x_1...x_n]
      PPOrdering PPO = NewStdDegRevLexOrdering(dim);
      PPMonoid PPM = NewPPMonoid(SymbolRange("x", 0, dim-1), PPO);
      SparsePolyRing P = NewPolyRing(R, PPM);
      RingHom coeff = CoeffEmbeddingHom(P);

      // Build evaluation homomorphisms at Pts
      list<RingHom> EvalAtPts;
      for (size_t i=0; i < NumPts; ++i)
        EvalAtPts.push_back(EvalHom(P, pts[i]));

      //   --------------     INITIALIZE   ---------------
      size_t IterNum = 1;
      vector<RingElem> J, S;     // J = almost vanishing polys

      QBGenerator QBG(PPM);      // to handle quotient bases
      QBG.myCornerPPIntoQB(one(PPM));
      S.push_back(one(P));
      list<PPMonoidElem> L = QBG.myCorners();  // L = list of corner PPs

      // Build and initialize evaluation matrix M (by rows),
      // its orthogonal part A, and DiagA = A*A'
      matrix A(NewDenseMat(R, NumPts, NumPts));
      for (size_t i=0; i < NumPts; ++i)
        SetEntry(A, 0, i, 1);

      matrix M(NewDenseMat(R, NumPts, NumPts));
      for (size_t i=0; i < NumPts; ++i)
        SetEntry(M, 0, i, 1);

      vector<RingElem> DiagAElems(NumPts, zero(R));
      MatrixView DiagA(DiagMat(DiagAElems));//DiagA=A*A'
      SetEntry(DiagA, 0, 0, NumPts);


      // ------------   MAIN CYCLE   -----------------
      while (!L.empty())
      {
        GlobalLogput() << "---------------------------------------" << endl
                       << "Iteration " << IterNum << endl
                       << "Quotient Basis = " << QBG.myQB() << endl
                       << "List of power products = " << L << endl;

        const PPMonoidElem t = L.front(); // PP considered in the current step
        GlobalLogput() << "Next PP = " << t << endl;
        const RingElem tAsPoly = monomial(P, one(R), t);

        // Compute eval V=t(Pts) and solve LS problem A'*b = V
        vector<RingElem> V, Alpha, Rho;
        Evaluation(V, tAsPoly, EvalAtPts);
        LeastSquares(Alpha, Rho, A, DiagA, V);

        // First trivial check
        if (IsZero(Rho))
        {
          AddToCornerAndUpdate(t, QBG, Alpha, S, J);
          GlobalLogput() << "Truly vanishing polynomial f = " << J.back() << endl;
          ++IterNum;
          L = QBG.myCorners();
          continue;
        }

        // build the poly g = t - sum(alpha_i s_i)
        RingElem g(tAsPoly);
        for (size_t i=0; i < S.size(); ++i)
          g -= coeff(Alpha[i])*S[i];

        vector<size_t> cols, rows;
        for (size_t i=0; i < QBG.myQB().size(); ++i)
          rows.push_back(i);
        for (size_t j=0; j < NumCols(M); ++j)
          cols.push_back(j);
        ConstMatrixView SubM = submat(M, rows, cols);

        // computing the bound for NBM
        vector<RingElem> Bound;
        NBMBound(Bound, SubM, tolerance, g, EvalAtPts);

        RingElem Diff(R);
        vector<RingElem> AbsRho = AbsVector(Rho);
        for (size_t i=0; i < NumPts; ++i)
        {
          Diff = AbsRho[i]-Bound[i];
          if (Diff > 0)
            break;
        }

        // if Diff = |Rho[i]|-Bound[i]>0 then ...
        if (Diff > 0)
        {
          UpdateMatricesNBM(QBG, M, A, DiagA, V, Rho);
          AddToQBAndUpdate(t, QBG, Alpha, S);
        }
        else
        {
          AddToCornerAndUpdate(t, QBG, Alpha, S, J);
          GlobalLogput() << "Almost vanishing polynomial f = " << J.back() << endl;
        }

        ++IterNum;
        L = QBG.myCorners();
      }

      GlobalLogput() << "==========================================" << endl
                     << endl
                     << "All elements of the order ideal have been found" << endl;

      // Computation of the Border Basis and printing
      BBasis.clear();
      std::swap(AlmostVanishing, J); // morally AlmostVanishing = J;
      QB.clear(); QB = QBG.myQB();
      if (QB.size() < NumPts)
      {
        GlobalLogput() << "Found " << QBG.myQB().size() << " elements of QB" << endl
                       << "A finer tolerance should be used!" << endl;
        return;
      }

      list<PPMonoidElem> BorderPPs;
      ComputeBorderPPs(BorderPPs, QBG);
      for (list<PPMonoidElem>::const_iterator it=BorderPPs.begin(); it != BorderPPs.end(); ++it)
      {
        RingElem BBElem(P);
        ComputeBorderPoly(BBElem, *it, A, DiagA, EvalAtPts, S);
        BBasis.push_back(zero(P)); swap(BBasis.back(), BBElem); // morally  BBasis.push_back(BBElem);
      }
    }



    void NBMTwinFloat(std::vector<PPMonoidElem>& QB,
                      std::vector<RingElem>& BBasis,
                      std::vector<RingElem>& AlmostVanishing,
                      const std::vector<PointR>& OrigPts,
                      const std::vector<RingElem>& OrigTolerance)
    {
      for (size_t BitPrec = 128; ; BitPrec *= 2)
      {
        try
        {
          ring RR = NewRingTwinFloat(BitPrec);
          vector<RingElem> tolerance;
          vector<PointR> pts;
          RingElem gamma(zero(RR));
          ConversionToRR(OrigTolerance, OrigPts, zero(owner(OrigPts[0][0])), tolerance, pts, gamma, RR);
          NBM(QB, BBasis, AlmostVanishing, pts, tolerance);
          GlobalLogput() << "BBasis successfully computed with " << BitPrec << " bit precision " << endl;
          return;
        }
        catch (const RingTwinFloat::InsufficientPrecision&)
        {
          // Inform the user about the failure...
          GlobalLogput() << endl << "----------------------------------"
                         << endl << "A bit precision of " << BitPrec
                         << " was not sufficient." << endl
                         << "------------------------------------------"
                         << endl << endl;
        }
      }
    }


    void NBM(std::vector<PPMonoidElem>& QB,
             std::vector<RingElem>& BBasis,
             std::vector<RingElem>& AlmostVanishing,
             const std::vector<PointDbl>& OrigPts,
             const std::vector<double>& OrigTolerance)
    {
      ring Q = RingQ();
      vector<RingElem> tolerance;
      vector<PointR> pts;
      RingElem gamma(Q);
      ConversionToRR(OrigTolerance, OrigPts, 0.0, tolerance, pts, gamma, Q);
      SOITwinFloat(QB, BBasis, AlmostVanishing, pts, tolerance, gamma);
    }


  } // end of namespace ApproxPts

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ApproxPts2.C,v 1.5 2009/12/11 11:46:32 abbott Exp $
// $Log: ApproxPts2.C,v $
// Revision 1.5  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.4  2009/09/24 12:38:27  abbott
// Added explicit include directive for system header <algorithm> as a consequence
// of removing <algorithm> from SmartPtrIRC.H.
// Shortened some very long lines (by breaking them into several lines).
//
// Revision 1.3  2008/12/17 12:14:21  abbott
// Tidied up a printing command.
//
// Revision 1.2  2008/11/24 17:13:11  abbott
// Final tidying: removed some dead code, removed some useless includes.
//
// Revision 1.1  2008/11/23 18:58:32  abbott
// Major overhaul to preprocessing and SOI/NBM code.
// Split SOI/NBM off into a separate file.
// Preprocessing is now "rational" (but internally guided by C++ doubles).
// SOI/NBM now each have 3 similar interfaces: one purely rational, one for
// input which is represented as doubles, and one which converts the input
// to RingTwinFloat values and produces a result which is over some RingTwinFloat
// (the precision is increased automatically until an answer is obtained).
//
// Revision 1.8  2008/10/07 15:45:22  abbott
// Changed ErrorInfo objects so they include the name of their own error ID.
// Changed catch statements to catch const objects.
// Removed calls to the member fn which accessed the error ID member of an
// ErrorInfo; now you simply compare directly with the error ID (makes the
// code easier to read).
//
// Revision 1.7  2008/09/12 13:28:43  bigatti
// -- new: NBM implementation
//
// Revision 1.6  2008/06/04 18:27:37  abbott
// Modified the server interface for "SOI": it now accepts a 3rd arg (gamma).
//
// Revision 1.5  2008/05/30 14:20:43  abbott
// SOI now returns also the "almost vanishing" polynomials.
//
// Revision 1.4  2008/05/30 12:51:08  abbott
// Added a comment.
//
// Revision 1.3  2008/05/29 15:46:29  bigatti
// -- added Approximate Border Basis (by Abbott,Torrente)
//
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.8  2007/03/08 18:22:30  cocoa
// Just whitespace cleaning.
//
// Revision 1.7  2006/12/21 13:48:33  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.6  2006/11/22 14:43:32  cocoa
// -- minor cleaning (indicated by Intel compiler)
//
// Revision 1.5  2006/11/20 15:53:38  cocoa
// Minor cleaning.  Improved comments.
//
// Revision 1.4  2006/10/31 14:26:03  cocoa
// Added some missing consts; should make the code a little easier to comprehend.
//
// Revision 1.3  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/06/21 17:05:47  cocoa
// Major overhaul of approx point preprocessing algms.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/29 16:16:47  cocoa
// Added "disj" preprocessing algorithm.
//
// Revision 1.2  2006/05/22 15:52:16  cocoa
// Added preprocess-disg algorithm to ApproxPts.
// Sundry minor improvements.
//
// Revision 1.1  2006/05/12 13:16:30  cocoa
// Added functions for preprocessing approximate points.
//
//
