//   Copyright (c)  2005-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for classes DenseMatBase and DenseMatImpl

#include "CoCoA/DenseMatrix.H"
#include "CoCoA/MatrixArith.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/ring.H"


//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;


namespace CoCoA
{

  class DenseMatBase: public MatrixBase
  {
  protected: // data members
    const ring myR;
    size_t myNumRowsValue;
    size_t myNumColsValue;
  protected:
    DenseMatBase(const ring& R, size_t nrows, size_t ncols);
  };



  class DenseMatImpl: public DenseMatBase
  {
  public:
    DenseMatImpl(const ring& R, size_t nrows, size_t ncols); // creates zero matrix
    ~DenseMatImpl();
  private: // implementation detail
    void myDtorBody();

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual DenseMatImpl* myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const;

    virtual void myAssignZero(); // sets all entries to zero
    bool myIsWritable(std::size_t /*i*/, std::size_t /*j*/) { return true; }
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);
// Use generic versions of the following four functions.
//     virtual void myMulByRow(vec& lhs, const vec& v) const;
//     virtual void myMulByCol(vec& lhs, const vec& v) const;
//     virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
//     virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
    virtual void myRowMul(std::size_t i, ConstRefRingElem c); ///< row(i) = c*row(i)
    virtual void myColMul(std::size_t j, ConstRefRingElem c); ///< col(j) = c*col(j)
    virtual void myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c); ///< row(i1) += c*row(i2)
    virtual void myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c); ///< col(j1) += c*col(j2)
    virtual void mySwapRows(std::size_t i1, std::size_t i2);
    virtual void mySwapCols(std::size_t j1, std::size_t j2);
    virtual void myResize(std::size_t NumRows, std::size_t NumCols);
    virtual void myDet(RefRingElem d) const;
    virtual size_t myRank() const;

   private: // data members
    vector< vector< RingElemRawPtr > > myEntries;
  };



  /////////////////////////////////////////////////////////////////////////////
  // Inline functions

  inline DenseMatBase::DenseMatBase(const ring& R, size_t NumRows, size_t NumCols):
      myR(R),
      myNumRowsValue(NumRows),
      myNumColsValue(NumCols)
  {}


//----------------------------------------------------------------------

  DenseMatImpl::DenseMatImpl(const ring& R, size_t NumRows, size_t NumCols):
      DenseMatBase(R, NumRows, NumCols)
  {
    // Making this exception safe turns out to be a bit complicated.
    try
    {
      myEntries.reserve(myNumRows()); // not necessary but might help
      for (size_t i=0; i < myNumRows(); ++i)
      {
        myEntries.push_back(vector<RingElemRawPtr>());
        myEntries[i].reserve(myNumCols()); // not necessary but might help
        for (size_t j=0; j < myNumCols(); ++j)
        {
          myEntries[i].push_back(myR->myNew());
        }
      }
    }
    catch (...)
    {
      myDtorBody();
    }
  }


  DenseMatImpl::~DenseMatImpl()
  {
    myDtorBody();
  }


  // Ignore the values of myNumRows and myNumCols because this procedure
  // may be called on a partially constructed DenseMat if an exception
  // is thrown during the ctor.
  void DenseMatImpl::myDtorBody()
  {
    // These two loops are slightly funny because i,j are unsigned.
    // Written this way, we destroy in reverse order of construction.
    for (size_t i=myEntries.size(); i > 0; --i)
    {
      for (size_t j=myEntries[i-1].size(); j > 0; --j)
        myR->myDelete(myEntries[i-1][j-1]);
    }
  }


  const ring& DenseMatImpl::myBaseRing() const
  {
    return myR;
  }


  size_t DenseMatImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  size_t DenseMatImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  inline ConstRefRingElem DenseMatImpl::myEntry(size_t i, size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    return ConstRefRingElem(myR, myEntries[i][j]);
  }


  DenseMatImpl* DenseMatImpl::myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const
  {
    return new DenseMatImpl(R, NumRows, NumCols);
  }


  void DenseMatImpl::myAssignZero()
  {
    // This fn is probably not exception safe -- can RingBase::myAssignZero throw?
    for (size_t i=0; i < myNumRows(); ++i)
      for (size_t j=0; j < myNumCols(); ++j)
        myR->myAssignZero(myEntries[i][j]);
  }


  RefRingElem DenseMatImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    return RefRingElem(myR, myEntries[i][j]);
  }


  void DenseMatImpl::mySetEntry(size_t i, size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(owner(r) == myR);
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    myR->myAssign(myEntries[i][j], raw(r));
  }


  void DenseMatImpl::mySetEntry(size_t i, size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    myR->myAssign(myEntries[i][j], n);
  }


  void DenseMatImpl::mySetEntry(size_t i, size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    myR->myAssign(myEntries[i][j], N);
  }


//   MOVED UP TO GENERIC IMPLEMENTATIONS IN ConstMatrixBase
//   void DenseMatImpl::myMulByRow(vec& lhs, const vec& v) const
//   {
//     CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
//     vector<RingElem> ans;
//     ans.resize(myNumCols(), zero(myR));
//     for (size_t j=0; j < myNumCols(); ++j)
//       for (size_t i=0; i < myNumRows(); ++i)
//         ans[j] += v[i]*myEntry(i, j);
//     // We have successfully computed the answer, now swap it in.
//     swap(lhs, ans);
// //     for (size_t j=0; j < myNumCols(); ++j)
// //       swap(lhs[j], ans[j]);
//   }


//   void DenseMatImpl::myMulByCol(vec& lhs, const vec& v) const
//   {
//     CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
//     vector<RingElem> ans;
//     ans.resize(myNumRows(), zero(myR));
//     for (size_t i=0; i < myNumRows(); ++i)
//       for (size_t j=0; j < myNumCols(); ++j)
//         ans[i] += myEntry(i, j)*v[j];

//     // We have successfully computed the answer, now swap it in.
//     swap(lhs, ans);
// //     for (size_t i=0; i < myNumRows(); ++i)
// //       swap(lhs[i], ans[i]);
//   }


//   bool DenseMatImpl::myIsZeroRow(std::size_t i) const
//   {
//     if (i >= myNumRows())
//       CoCoA_ERROR(ERR::BadRowIndex, "DenseMatImpl::myIsZeroRow");
//     for (size_t j=0; j < myNumCols(); ++j)
//       if (!myR->myIsZero(myEntries[i][j])) return false;
//     return true;
//   }


//   bool DenseMatImpl::myIsZeroCol(std::size_t j) const
//   {
//     if (j >= myNumCols())
//       CoCoA_ERROR(ERR::BadColIndex, "DenseMatImpl::myIsZeroCol");
//     for (size_t i=0; i < myNumRows(); ++i)
//       if (!myR->myIsZero(myEntries[i][j])) return false;
//     return true;
//   }


  void DenseMatImpl::myRowMul(std::size_t i, ConstRefRingElem c)
  {
    const char* const FnName = "DenseMatImpl::myRowMul";
    myCheckRowIndex(i, FnName);
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, FnName);
    if (IsOne(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumCols(), zero(myR));
    for (size_t j = 0; j < myNumCols(); ++j)
      myR->myMul(raw(ans[j]), raw(c), myEntries[i][j]);
    // Answer successfully computed in ans, swap it into the i-th row.
    for (size_t j = 0; j < myNumCols(); ++j)
      myR->mySwap(myEntries[i][j], raw(ans[j]));
  }


  void DenseMatImpl::myColMul(std::size_t j, ConstRefRingElem c)
  {
    const char* const FnName = "DenseMatImpl::myColMul";
    myCheckColIndex(j, FnName);
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, FnName);
    if (IsOne(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumRows(), zero(myR));
    for (size_t i = 0; i < myNumRows(); ++i)
      myR->myMul(raw(ans[i]), raw(c), myEntries[i][j]);
    // Answer successfully computed in ans, swap it into the i-th row.
    for (size_t i = 0; i < myNumRows(); ++i)
      myR->mySwap(myEntries[i][j], raw(ans[i]));
  }


  void DenseMatImpl::myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c)
  {
    const char* const FnName = "DenseMatImpl::myAddRowMul";
    myCheckRowIndex(i1, FnName);
    myCheckRowIndex(i2, FnName);
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, FnName);
    if (IsZero(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumCols(), zero(myR));
    for (size_t j = 0; j < myNumCols(); ++j)
    {
      myR->myAssign(raw(ans[j]),myEntries[i1][j]);
      myR->myIsZeroAddMul(raw(ans[j]), raw(c), myEntries[i2][j]);
    }
    // Answer successfully computed in ans, swap it into the i-th row
    for (size_t j = 0; j < myNumCols(); ++j)
      myR->mySwap(raw(ans[j]), myEntries[i1][j]);
  }


  void DenseMatImpl::myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c)
  {
    const char* const FnName = "DenseMatImpl::myAddColMul";
    myCheckColIndex(j1, FnName);
    myCheckColIndex(j2, FnName);
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, FnName);
    if (IsZero(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumRows(), zero(myR));
    for (size_t i = 0; i < myNumRows(); ++i)
    {
      myR->myAssign(raw(ans[i]),myEntries[i][j1]);
      myR->myIsZeroAddMul(raw(ans[i]), raw(c), myEntries[i][j2]);
    }
    // Answer successfully computed in ans, swap it into the j-th col
    for (size_t i = 0; i < myNumRows(); ++i)
      myR->mySwap(raw(ans[i]), myEntries[i][j1]);
  }


  void DenseMatImpl::mySwapRows(std::size_t i1, std::size_t i2)
  {
    const char* const FnName = "DenseMatImpl::mySwapRows";
    myCheckRowIndex(i1, FnName);
    myCheckRowIndex(i2, FnName);
    if (i1 == i2) return;
//??? std::swap(myEntries[i1], myEntries[i2]); // trouble with iterators???
    for (size_t j=0; j < myNumCols(); ++j)
      myR->mySwap(myEntries[i1][j], myEntries[i2][j]);
  }


  void DenseMatImpl::mySwapCols(std::size_t j1, std::size_t j2)
  {
    const char* const FnName = "DenseMatImpl::mySwapCols";
    myCheckColIndex(j1, FnName);
    myCheckColIndex(j2, FnName);
    if (j1 == j2) return;
    // Must do this the "hard way".
    for (size_t i=0; i < myNumRows(); ++i)
      myR->mySwap(myEntries[i][j1], myEntries[i][j2]);
  }


  // NOT FULLY EXCEPTION CLEAN (do not see how without taking a performance hit, or making it even harder to read)
  void DenseMatImpl::myResize(std::size_t NumRows, std::size_t NumCols)
  {
    // First remove any excess rows
    if (NumRows < myNumRowsValue)
    {
      for (size_t i=NumRows; i < myNumRowsValue; ++i)
        for (size_t j=0; j < myNumColsValue; ++j)
          myR->myDelete(myEntries[i][j]);
      myEntries.resize(NumRows);
    }

    // Now resize each remaining row.
    // Three cases: lengthen, shorten, or no change.
    if (NumCols > myNumColsValue)
    {
      for (size_t i=0; i < myEntries.size(); ++i)
      {
        for (size_t j=myNumColsValue; j < NumCols; ++j)
          myEntries[i].push_back(myR->myNew()); // could throw std::bad_alloc
      }
    }
    if (NumCols < myNumColsValue)
    {
      for (size_t i=0; i < myEntries.size(); ++i)
      {
        for (size_t j=NumCols; j < myNumColsValue; ++j)
          myR->myDelete(myEntries[i][j]);
        myEntries.resize(NumCols);
      }
    }
    // Add new zero rows if number of rows is to increase.
    if (NumRows > myNumRowsValue)
    {
      for (size_t i=myNumRowsValue; i < NumRows; ++i)
      {
        myEntries.push_back(vector<RingElemRawPtr>());
        for (size_t j=0; j < NumCols; ++j)
          myEntries[i].push_back(myR->myNew()); // could throw std::bad_alloc
      }
    }
    myNumRowsValue = NumRows;
    myNumColsValue = NumCols;
  }

  void DenseMatImpl::myDet(RefRingElem d) const
  {
    // ??? this function works only within integral domains!
    DetByGauss(d, ConstMatrixView(this));
  }


  size_t DenseMatImpl::myRank() const
  {
    vector<size_t> discard;
    return RankByGauss(discard, ConstMatrixView(this));
  }


  matrix NewDenseMat(const ring& R, size_t NumRows, size_t NumCols)
  {
    return matrix(new DenseMatImpl(R, NumRows, NumCols));
  }

  matrix NewDenseMat(const ring& R, std::size_t NumRows, std::size_t NumCols, const std::vector< std::vector<int> >& M)
  {
    // Verify that M has the right shape.
    if (M.size() != NumRows)
      CoCoA_ERROR(ERR::BadRowIndex, "NewDenseMat(R,n,m,M)");
    for (size_t i=0; i < NumRows; ++i)
      if (M[i].size() != NumCols)
        CoCoA_ERROR(ERR::BadColIndex, "NewDenseMat(R,n,m,M)");
    matrix ans(new DenseMatImpl(R, NumRows, NumCols));
    for (size_t i=0; i < NumRows; ++i)
      for (size_t col=0; col < NumCols; ++col)
        SetEntry(ans, i, col, M[i][col]);
    return ans;
  }


  // I bet this makes lots of wasteful copies...
  matrix NewDenseMat(ConstMatrixView M)
  {
    const size_t r = NumRows(M);
    const size_t c = NumCols(M);
    matrix ans(new DenseMatImpl(BaseRing(M), r, c));
    for (size_t i=0; i < r; ++i)
      for (size_t j=0; j < c; ++j)
        SetEntry(ans, i, j, M(i, j));
    return ans;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/DenseMatrix.C,v 1.7 2009/09/24 13:41:21 abbott Exp $
// $Log: DenseMatrix.C,v $
// Revision 1.7  2009/09/24 13:41:21  abbott
// Removed useless include directives.
// Removed unnecessary "std::" prefixes.
//
// Revision 1.6  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.5  2008/04/18 15:35:57  abbott
// (long overdue) Major revision to matrices
//
// Revision 1.4  2008/04/16 17:24:17  abbott
// Further cleaning of the new matrix code.  Updated documentation too.
//
// Revision 1.3  2008/04/08 15:26:42  abbott
// Major revision to matrix implementation: added matrix views.
// Lots of changes.
//
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/08 18:22:30  cocoa
// Just whitespace cleaning.
//
// Revision 1.5  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.4  2006/11/27 16:18:33  cocoa
// -- moved classes declarations from .H to .C (DenseMatrix, DiagMatrix,
//    FieldIdeal, SpecialMatrix)
//
// Revision 1.3  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.2  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.2  2006/01/25 15:58:19  cocoa
// -- fixed: size condition in mul(matrix&, ConstMatrix, ConstMatrix), by Kaspar
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.2  2005/05/06 11:57:15  cocoa
// Fixed bug in ctor for DenseMatrixImpl.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.9  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.8  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.7  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.6  2005/03/29 17:26:29  cocoa
// -- changed: myDet, myRank call MatrixArith functions
//
// Revision 1.5  2005/03/11 16:44:18  cocoa
// New abstract class structure for matrices.
// New types of special matrix.
//
// Revision 1.4  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.3  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.2  2005/02/09 11:19:30  cocoa
// -- fixed: determinant over a GCDRing (through FrF)
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.6  2004/11/29 16:22:35  cocoa
// -- added function for computing adjoint and inverse for DenseMatrix
//    (so adjoint/inverse matrix is computed by OrdvArith and is no
//    longer needed by PPOrdering)
//
// Revision 1.5  2004/11/02 14:49:03  cocoa
// -- new code for matrix orderings
//
// Revision 1.4  2004/07/20 15:04:06  cocoa
// The next step in the new "ring element" conversion process:
// handling the case of creating a "const RefRingElem" object
// (since C++ refuses to do this properly itself).
//
// Revision 1.3  2004/03/20 17:46:11  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.2  2004/03/09 16:15:41  cocoa
// First version of matrices.  A simple exmaple compiles and runs,
// so I'm checking in.
//
// Revision 1.1  2004/01/28 16:29:17  cocoa
// First attempt at implementing matrices.
//
//
