//   Copyright (c)  2007,2009 Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for abstract class SparsePolyRing and friends

#include "CoCoA/DenseUPolyRing.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/QQ.H"
#include "CoCoA/QuotientRing.H"  // for IsQuotientRing/AsQuotientRing
#include "CoCoA/RingTwinFloat.H" // for IsRingTwinFloat
#include "CoCoA/RingFp.H" // for IsRingFp
#include "CoCoA/RingQ.H" // for IsQ
#include "CoCoA/RingZ.H" // for IsZ
#include "CoCoA/RingDenseUPolyClean.H" // for NewPolyRing_DUP
#include "CoCoA/assert.H"
#include "CoCoA/degree.H"
#include "CoCoA/error.H"
#include "CoCoA/symbol.H"
#include "CoCoA/io.H"  // ONLY for debugging

//#include <vector>
using std::vector;
//#include <cstddef>
using std::size_t;
#include <algorithm>
using std::max;
using std::sort;


namespace CoCoA
{

  DenseUPolyRing NewPolyRing(const ring& CoeffRing)
  {
    return NewPolyRing_DUP(CoeffRing);
  }

  DenseUPolyRing NewPolyRing(const ring& CoeffRing, const symbol& IndetName)
  {
    return NewPolyRing_DUP(CoeffRing, IndetName);
  }

  bool IsGoodIndetName(const ring& CoeffRing, const symbol& IndetName)
  {
    vector<symbol> syms = symbols(CoeffRing);
    const size_t NumSyms = syms.size();
    for (size_t i=0; i < NumSyms; ++i)
    {
      if (syms[i] == IndetName) return false;
      if (head(syms[i]) == head(IndetName) &&
          NumIndices(syms[i]) != NumIndices(IndetName))
        return false;
    }
    return true;
  }
  

  //---- Functions for creating/building polynomials

  RingElem monomial(const DenseUPolyRing& P, ConstRefRingElem c, const unsigned long exp)
  {
    if (owner(c) != CoeffRing(P)) CoCoA_ERROR(ERR::MixedCoeffRings, "monomial(P,c,d)");
    if (IsZero(c)) return zero(P);
    return P->myMonomial(raw(c), exp);
  }
  

  RingElem monomial(const DenseUPolyRing& P, const ZZ& N, const unsigned long exp)
  {
    return monomial(P, RingElem(CoeffRing(P), N), exp);
  }
  

  RingElem monomial(const DenseUPolyRing& P, MachineInteger n, const unsigned long exp)
  {
    return monomial(P, RingElem(CoeffRing(P), n), exp);
  }


  /*----------------------------------------------------------------------
    Member functions every concrete DenseUPolyRing implementation
    must have in addition to those of PolyRingBase.
    ----------------------------------------------------------------------*/

  bool DenseUPolyRingBase::myIsValid(ConstRawPtr rawf) const
  {
    if (myDegPlus1(rawf)>mySize(rawf)) return false;
    for (size_t i=myDegPlus1(rawf); i<mySize(rawf); ++i)
      if (!IsZero(myCoeff(rawf, i))) return false;
    if (myIsZero(rawf)) return true;
    if (IsZero(myLC(rawf))) return false;
    return true;
  }


  long DenseUPolyRingBase::myStdDeg(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) CoCoA_ERROR(ERR::ZeroRingElem, "myStdDeg(rawf)");
    return myDegPlus1(rawf) - 1;
  }


  long DenseUPolyRingBase::myMaxExponent(ConstRawPtr rawf, size_t index) const
  {
    if (index!=0) CoCoA_ERROR(ERR::BadIndetIndex, "myMaxExponent(rawf, index)");
    return myStdDeg(rawf);
  }


  ConstRefRingElem DenseUPolyRingBase::myLC(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) CoCoA_ERROR(ERR::ZeroRingElem, "DenseUPolyRingBase::myLC(rawf)");
    return myCoeff(rawf, myStdDeg(rawf));
  }
  

  void DenseUPolyRingBase::myContent(RawPtr rawcontent, ConstRawPtr rawf) const
  {
    CoCoA_ASSERT(IsGCDDomain(myCoeffRing()));
    const ring& R = myCoeffRing();
    // Compute answer in local var to avoid aliasing problems; also exception clean.
    RingElem ans(R);
    for (size_t i=0; i<myDegPlus1(rawf); ++i)
      if (!IsZero(myCoeff(rawf,i)))
        {
          R->myGcd(raw(ans), raw(ans), raw(myCoeff(rawf, i))); // ans = GCD(ans, coeff(i));
          if (IsOne(ans)) break;
        }
    // Finally, swap answer into rawcontent.
    R->mySwap(rawcontent, raw(ans));
  }


  void DenseUPolyRingBase::myCommonDenom(RawPtr rawcontent, ConstRawPtr rawf) const
  {
    CoCoA_ASSERT(IsFractionField(myCoeffRing()) && IsGCDDomain(BaseRing(AsFractionField(myCoeffRing()))));
    const ring& R = BaseRing(AsFractionField(myCoeffRing()));
    // Compute answer in local var to avoid aliasing problems; also exception clean.
    RingElem ans = one(R);
    for (size_t i=0; i<myDegPlus1(rawf); ++i)
    {
      ConstRefRingElem C = myCoeff(rawf, i);
      if (!IsZero(C))
      {
        const RingElem D = den(C);
        ans *= D/gcd(D, ans);
      }
    }
    // Finally, swap answer into rawcontent.
    R->mySwap(rawcontent, raw(ans));
  }


  void DenseUPolyRingBase::myRemoveBigContent(RawPtr rawf) const
  {
    CoCoA_ASSERT(IsGCDDomain(myCoeffRing()));
    CoCoA_ASSERT(!myIsZero(rawf));
    RingElem cont(myCoeffRing());
    myContent(raw(cont), rawf);
    myDivByCoeff(rawf, raw(cont));
  }


  /*----------------------------------------------------------------------
    Member functions inherited from ring with a single implementation
    for all DenseUPolyRing implementations
    ----------------------------------------------------------------------*/

  namespace
  {
    // only if CoeffRing is a field
    void mod(RefRingElem a, ConstRefRingElem b)  // a = a%b
    {
      DenseUPolyRing P = AsDenseUPolyRing(owner(a));
      CoCoA_ASSERT(P == owner(b));
      CoCoA_ASSERT(!IsZero(b));  // this should be an error
      while (!IsZero(a) && deg(a) >= deg(b))
        P->myAddMul(raw(a), raw(-LC(a)/LC(b)), deg(a)-deg(b), raw(b));
    }
    
    void DivMod(RefRingElem q, RefRingElem a, ConstRefRingElem b)  // q = a/b;  a = a%b
    {
      // pointers must be different
      DenseUPolyRing P = AsDenseUPolyRing(owner(a));
      CoCoA_ASSERT(P == owner(b));
      CoCoA_ASSERT(!IsZero(b));  // this should be an error
      q = zero(owner(q));
      if (IsZero(a) || (deg(a)<deg(b)) ) return;
      P->myResize(raw(q), deg(a)-deg(b)+1);
      while (!IsZero(a) && deg(a) >= deg(b))
      {
        // q += monomial(P, -LC(a)/LC(b), deg(a)-deg(b));
        P->myAssignNonZeroCoeff(raw(q), raw(LC(a)/LC(b)), deg(a)-deg(b));
        P->myAddMul(raw(a), raw(-LC(a)/LC(b)), deg(a)-deg(b), raw(b));
      }
    }
    
    void ExgcdAux(RefRingElem acofac, RefRingElem bcofac, RefRingElem a, RefRingElem b) // gcd is acofac*orig_a + bcofac*orig_b;  terminates with a=gcd and b=0
    {
      // all pointers must be different
      CoCoA_ASSERT(owner(a) == owner(b));
      CoCoA_ASSERT(!IsZero(b));  // this should be an error
      acofac = 1;
      bcofac = 0;  // i.e. a = acofac*a + bcofac*b
      RingElem q(owner(b));
      RingElem b_acofac(owner(b));
      RingElem b_bcofac(one(owner(b)));  // i.e. b = b_acofac*a + b_bcofac*b
      while (!IsZero(b))
      {
        // q = div(a, b);  a = a - b*d;
        DivMod(q, a, b);
        acofac -= b_acofac*q;
        bcofac -= b_bcofac*q;
        swap(a, b);
        swap(acofac, b_acofac);
        swap(bcofac, b_bcofac);
      }
  }


  }  // anonymous namespace


  void DenseUPolyRingBase::myMul(RawPtr rawlhs, ConstRawPtr rawf, ConstRawPtr rawg) const
  {
    DenseUPolyRing P(AsDenseUPolyRing(ring(this)));
    ConstRefRingElem g(P, rawg);

    RingElem ans(P);
    for (size_t i=0; i<myDegPlus1(rawf); ++i)
      myAddMul(raw(ans), raw(myCoeff(rawf,i)), i, rawg);
    mySwap(raw(ans), rawlhs); // really an assignment
  }


  bool DenseUPolyRingBase::myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    DenseUPolyRing P(AsDenseUPolyRing(ring(this)));
    RingElem r = ConstRefRingElem(P, rawx);
    RingElem ans(P);
    DivMod(ans, r, ConstRefRingElem(P, rawy));
    mySwap(raw(ans), rawlhs); // really an assignment
    return IsZero(r);
  }


  // code for R[x]: compute gcd in FractionField(R)[x]
  // see also myExgcd
  void DenseUPolyRingBase::myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    RingElem a = ConstRefRingElem(ring(this), rawx);
    if (myIsZero(rawy))
    {
      mySwap(rawlhs, raw(a));
      return;
    }
    RingElem b = ConstRefRingElem(ring(this), rawy);
    if (myIsZero(rawx))
    {
      mySwap(rawlhs, raw(b));
      return;
    }
    if (!IsField(myCoeffRing()))
      CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::myGcd(RawPtr, ConstRawPtr, ConstRawPtr) const");
    // finally Euclid's Algorithm
    mod(a, b);  // a = remainder(a, b)
    while (!IsZero(a))
    {
      mySwap(raw(a), raw(b));
      mod(a, b);  // a = remainder(a, b)
    }
    mySwap(rawlhs, raw(b));
  }


  // See also myGcd
  void DenseUPolyRingBase::myExgcd(RawPtr rawlhs, RawPtr rawxcofac, RawPtr rawycofac, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    ring P(ring(this));
    RingElem ZeroP(P);
    RingElem OneP(one(P));
    RingElem a(ConstRefRingElem(P, rawx));
    if (myIsZero(rawy))  // = raw(b)
    {
      mySwap(rawlhs, raw(a));
      mySwap(rawxcofac, raw(OneP));
      mySwap(rawycofac, raw(ZeroP));
      return;
    }
    RingElem b(ConstRefRingElem(P, rawy));
    if (myIsZero(rawx))  // = raw(a)
    {
      mySwap(rawlhs, raw(b));
      mySwap(rawxcofac, raw(ZeroP));
      mySwap(rawycofac, raw(OneP));
      return;
    }
    if (!IsField(myCoeffRing()))
      CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::myGcd(RawPtr, ConstRawPtr, ConstRawPtr) const");
    // finally Euclid's Algorithm
    RingElem res(P), acofac(P), bcofac(P);
    ExgcdAux(acofac, bcofac, a, b);
    mySwap(rawlhs, raw(a));
    mySwap(rawxcofac, raw(acofac));
    mySwap(rawycofac, raw(bcofac));
  }


  void DenseUPolyRingBase::mySymbols(std::vector<symbol>& SymList) const
  {
    myCoeffRing()->mySymbols(SymList);
    SymList.push_back(myIndetName());
  }


  void DenseUPolyRingBase::myOutput(std::ostream& out, ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) { out << "0"; return; }

    const ring& R = myCoeffRing();
    bool PrintStar, IsFirstCoeff=true;
    bool IsQuotientOfZ = IsQuotientRing(R) && IsZ(BaseRing(AsQuotientRing(R)));
    //bool IsQuotientOfZ = IsRingFp(R);
    bool IsNumberRing = IsZ(R) || IsQuotientOfZ || IsQ(R) || IsRingTwinFloat(R);

    for (long d=myStdDeg(rawf) ; d != 0 ; --d)
    {
      PrintStar = true;
      ConstRefRingElem c = myCoeff(rawf, d);
      if (IsZero(c)) continue;
      //----------------------------------------------------------------------
      // coefficient
      //----------------------------------------------------------------------

      if (!IsNumberRing)
      {
        if (!IsFirstCoeff) out << " +";
        out << '(' << c << ')';
        goto EndCoeffPrint;
      }
      // assuming Q is printed with positive denominator
      if (!IsFirstCoeff)
      {
        if (R->myIsPrintedWithMinus(raw(c))) out << " ";
        else out <<" +";
      }
      if (IsOne(c)) // do not print "1 * "
      {
        PrintStar = false;
        goto EndCoeffPrint;
      }
      if ( !IsQuotientOfZ && IsMinusOne(c) ) // do not print "-1 * "
      {
        PrintStar = false;
        out << "-";
        goto EndCoeffPrint;
      }
      out << c;
      
    EndCoeffPrint:
      
      //----------------------------------------------------------------------
      // PP
      //----------------------------------------------------------------------
      if (PrintStar)  out << "*";
      out << myIndetName();
      if (d > 1) out << "^" << d;

      IsFirstCoeff = false;
    }      
      //----------------------------------------------------------------------
      // coefficient
      //----------------------------------------------------------------------

    ConstRefRingElem c = myCoeff(rawf, 0);
    if (IsZero(c)) return;
    if (!IsNumberRing)
    {
      if (!IsFirstCoeff) out << " +";
      out << '(' << c << ')';
      return;
    }
      // assuming Q is printed with positive denominator
    if (!IsFirstCoeff)
    {
      if (R->myIsPrintedWithMinus(raw(c))) out << " ";
      else out <<" +";
    }
    out << c;
  }


  bool DenseUPolyRingBase::myIsZero(ConstRawPtr rawx) const
  {
    return myDegPlus1(rawx)==0;
  }


  bool DenseUPolyRingBase::myIsPrintAtom(ConstRawPtr rawx) const
  {
    if (!myIsConstant(rawx)) return false;
    // same as in "output"
    const ring& R = myCoeffRing();

    if (IsZ(R) || IsQ(R)) return R->myIsPrintAtom(raw(myLC(rawx)));
    return true;
  }


  void DenseUPolyRingBase::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("cocoa", "DUPCoeffs");
    // ??? if (IsZero(rawx)) .....
    OMOut << myStdDeg(rawx);

    for (long i=0 ; i <= myStdDeg(rawx) ; ++i)
      OMOut << myCoeff(rawx, i);
    OMOut->mySendApplyEnd();
  }
  

  bool DenseUPolyRingBase::myIsOne(ConstRawPtr rawf) const
  {
    if (myDegPlus1(rawf)!=1) return false;
    return IsOne(myCoeff(rawf, 0));
  }


  bool DenseUPolyRingBase::myIsMinusOne(ConstRawPtr rawf) const
  {
    if (myDegPlus1(rawf)!=1) return false;
    return IsMinusOne(myCoeff(rawf, 0));
  }


  bool DenseUPolyRingBase::myIsConstant(ConstRawPtr rawf) const
  {
    if (myDegPlus1(rawf)>1) return false;
    return true;
  }


  bool DenseUPolyRingBase::myIsIndet(std::size_t& IndetIndex, ConstRawPtr rawf) const
  {
    if (myDegPlus1(rawf)!=2) return false;
    if (!(IsOne(myCoeff(rawf,1)) && IsZero(myCoeff(rawf,0))))
      return false;
    IndetIndex = 0;
    return true;
  }


  bool DenseUPolyRingBase::myIsMonomial(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) return true;
    for (long i=0; i<myStdDeg(rawf); ++i)
      if (!myIsZero(raw(myCoeff(rawf,i)))) return false;
    return true;
  }


  void DenseUPolyRingBase::myIndetPower(RawPtr rawf, std::size_t var, unsigned int exp) const
  {
    CoCoA_ASSERT(var==0);
    RingElem ans(ring(this));
    myResize(raw(ans), exp+1);
    myAssignNonZeroCoeff(raw(ans), raw(one(myCoeffRing())), exp);
    mySwap(raw(ans), rawf); // do it this way to be exception clean
  }


  std::size_t DenseUPolyRingBase::myNumTerms(ConstRawPtr rawf) const
  {
    size_t nt = 0;
    for (size_t i=0; i<myDegPlus1(rawf); ++i) 
      if (!myCoeffRing()->myIsZero(raw(myCoeff(rawf,i)))) ++nt;
    return nt;
  }
  
  RingElem DenseUPolyRingBase::myMonomial(ConstRawPtr rawc, unsigned long exp) const
  {
    CoCoA_ASSERT(!myCoeffRing()->myIsZero(rawc));
    RingElem ans(ring(this));
    myResize(raw(ans), exp+1);
    myAssignNonZeroCoeff(raw(ans), rawc, exp);
    return ans;
  }


  bool DenseUPolyRingBase::myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    if (myIsZero(rawx))
    {
      if (myIsZero(rawy)) return true;
      else return false;
    }
    if (myDegPlus1(rawx) != myDegPlus1(rawy)) return false;
    for (size_t i=0; i<myDegPlus1(rawx); ++i) 
      if (myCoeff(rawx,i) != myCoeff(rawy,i))
        return false;
    return true;
  }


  bool DenseUPolyRingBase::myIsInteger(ZZ& N, ConstRawPtr rawx) const
  {
    if (myIsZero(rawx))
    {
      N = 0;
      return true;
    }
    if (myIsConstant(rawx))
      return myCoeffRing()->myIsInteger(N, raw(myCoeff(rawx, 0)));
    return false;
  }


  bool DenseUPolyRingBase::myIsRational(QQ& Q, ConstRawPtr rawx) const
  {
    if (myIsZero(rawx)) { Q = 0; return true; }
    if (!myIsConstant(rawx)) return false;
    return IsRational(Q, myCoeff(rawx,0));
  }


//   bool DenseUPolyRingBase::myIsHomog(ConstRawPtr rawf) const
//   {
//     if (myIsConstant(rawf)) { return true; }
//     CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::myIsHomog(ConstRawPtr)");
//     return true;
//   }


  bool DenseUPolyRingBase::myIsInvertible(ConstRawPtr rawx) const
  {
    return (!myIsZero(rawx)) && myIsConstant(rawx) && IsInvertible(myLC(rawx));
  }


  void DenseUPolyRingBase::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(n > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    size_t NoUse;
    if (myIsIndet(NoUse, rawx))  //??? is this the right place for this check??
      myIndetPower(rawlhs, 0, n);
    else
      myBinaryPower(rawlhs, rawx, n);
    //    mySequentialPower(rawlhs, rawx, n); //??? BUG/SLUG myBinaryPower better if univariate or coeffs are finite field
  }


  //---- Special functions on RingElem owned by DenseUPolyRing


  //-- IdealImpl ----------------------------------------

  ideal DenseUPolyRingBase::myIdealCtor(const std::vector<RingElem>& gens) const
  {
    return ideal(new IdealImpl(DenseUPolyRing(this), gens)); //??? ugly ???
  }


  DenseUPolyRingBase::IdealImpl::IdealImpl(const DenseUPolyRing& P, const std::vector<RingElem>& gens):
    myP(P),
    myGensValue(gens)
      //      myGBasisIsValid(false)
  {
    myTidyGensIsValid = false;
    if (!IsField(CoeffRing(P)))
      CoCoA_ERROR("NYI ideal of polynomials with coeffs not in a field", "ideal(DenseUPolyRing, gens)");//???
    for (size_t i=0; i < gens.size(); ++i)
      if (owner(gens[i]) != myP) CoCoA_ERROR(ERR::MixedRings, "DenseUPolyRingBase::IdealImpl ctor");
  }


  IdealBase* DenseUPolyRingBase::IdealImpl::clone() const
  {
    return new IdealImpl(*this);
  }


  bool DenseUPolyRingBase::IdealImpl::IamZero() const
  {
    if (myTidyGensIsValid) return myTidyGensValue.empty();
    const vector<RingElem>& g = myGensValue;
    for (size_t i=0; i < g.size(); ++i)
      if (!IsZero(g[i])) return false;
    return true;
  }


  void DenseUPolyRingBase::IdealImpl::myMaximalTest() const
  {
    //  CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::myMaximalTest()");
    myPrimeTest();
  }


  void DenseUPolyRingBase::IdealImpl::myPrimeTest() const
  {
    // CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::myPrimeTest()");
    if (!IsField(CoeffRing(myP)))
      CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::myPrimeTest() non-field CoeffRing");
    if (IamZero()) { IamPrimeFlag = true; IamMaximalFlag = false; return; }//!!! Needs IsIntegralDomain(CoeffRing(myAmbientRing))!!!
    // The ring is a univariate poly ring, thus a PID; hence ideale has a single generator.
    ConstRefRingElem f = myTidyGens()[0];  // ???
    if (!IsConstant(f) && IsIrred(f))
      IamPrimeFlag = bool3::DefinitelyTrue;
    else
      IamPrimeFlag = bool3::DefinitelyFalse;
    IamMaximalFlag = IamPrimeFlag;
  }


  void DenseUPolyRingBase::IdealImpl::myReduceMod(RingElemRawPtr rawf) const
  {
    if (myP->myIsZero(rawf)) return;
    if (IamZero()) return;
    RingElem g = ConstRefRingElem(myP, rawf);
    mod(g, (myTidyGens())[0]);
    myP->mySwap(rawf, raw(g));
  }


  bool DenseUPolyRingBase::IdealImpl::contains(RingElemConstRawPtr rawf) const
  {
    RingElem g = ConstRefRingElem(myP, rawf);
    myReduceMod(raw(g));
    return IsZero(g);
  }


  inline const DenseUPolyRingBase::IdealImpl* DenseUPolyRingBase::IdealImpl::GetPtr(const ideal& I)
  {
    return dynamic_cast<const DenseUPolyRingBase::IdealImpl*>(I.myIdealPtr());
  }


  void DenseUPolyRingBase::IdealImpl::add(const ideal& Jin)
  {
    const IdealImpl* const J = GetPtr(Jin);
    copy(J->myGensValue.begin(), J->myGensValue.end(), back_inserter(myGensValue));
    myTidyGensIsValid = false;
    myTidyGensValue.clear();
  }


  void DenseUPolyRingBase::IdealImpl::mul(const ideal& Jin)
  {
    CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::mul");
  }


  void DenseUPolyRingBase::IdealImpl::intersect(const ideal& J)
  {
    if (IamZero()) return;
    if (IsZero(J))
    {
      myGensValue.clear();
      myTidyGensValue.clear();
      myTidyGensIsValid = true;
      return;
    }
    if (!IsField(CoeffRing(myP)))
      CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::intersect non-field CoeffRing");
    RingElem f = myTidyGens()[0];
    RingElem g = TidyGens(J)[0];
    //    ideal I(myP, (g*f)/gcd(f,g));
    RingElem h = (g*f)/gcd(f,g);
    myGensValue.clear();
    myGensValue.push_back(h);
    myTidyGensValue.clear();
    myTidyGensIsValid = false;
  }


  void DenseUPolyRingBase::IdealImpl::colon(const ideal& J)
  {
    if (!IsField(CoeffRing(myP)))
      CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::colon non-field CoeffRing");
    if (IamZero()) return;
    if (IsZero(J))
    {
      myGensValue.clear();
      myGensValue.push_back(one(myP));
      myTidyGensValue.clear();
      myTidyGensIsValid = false;
      return;
    }
    RingElem f = myTidyGens()[0];
    RingElem g = TidyGens(J)[0];
    RingElem h = f/gcd(f,g);
    myGensValue.clear();
    myGensValue.push_back(h);
    myTidyGensValue.clear();
    myTidyGensIsValid = false;
  }


  void DenseUPolyRingBase::IdealImpl::myDivMod(RingElemRawPtr /*rawlhs*/, RingElemConstRawPtr /*rawnum*/, RingElemConstRawPtr /*rawden*/) const
  {
    CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::myDivMod");
  }


  const std::vector<RingElem>& DenseUPolyRingBase::IdealImpl::myTidyGens() const
  {
    if (!IsField(CoeffRing(myP)))
      CoCoA_ERROR(ERR::NYI, "DenseUPolyRingBase::IdealImpl::myTidyGens non-field CoeffRing");
    if (myTidyGensIsValid) return myTidyGensValue;
    CoCoA_ASSERT(myTidyGensValue.empty());
    RingElem g(myP);
    for (size_t i=0; i < myGensValue.size(); ++i) // break out if g=1???
      myP->myGcd(raw(g), raw(g), raw(myGensValue[i]));
    if (!IsZero(g))  myTidyGensValue.push_back(g);
    myTidyGensIsValid = true;
    return myTidyGensValue;
  }


  //-- HomImpl ----------------------------------------

  DenseUPolyRingBase::HomImpl::HomImpl(const DenseUPolyRing& domain, const ring& codomain, const RingHom& CoeffHom, ConstRefRingElem IndetImage):
      RingHomBase(domain, codomain),
      myCoeffHom(CoeffHom),
      myIndetImage(IndetImage)
  {
    // No need to check anything: checks already made when CoeffHom was built.
  }

namespace
{
  // ??? this is essentially the same as ApplyGeneral: it should be improved
  // assume image==0
  void ApplyDPRCodomain(RefRingElem image, ConstRefRingElem arg, const RingHom CoeffHom, ConstRefRingElem IndetImage)
  {
    DenseUPolyRing DPR = AsDenseUPolyRing(owner(image));
    RingElem g(DPR);
    if (!IsZero(arg))
      for (long i=0; i<=deg(arg); ++i) 
      {
        RingElem SummandImage = CoeffHom(coeff(arg,i));
        CoCoA_ASSERT(owner(SummandImage) == DPR);
        if (IsZero(SummandImage)) continue; // efficiency hack????
        if (i != 0) 
          SummandImage *= power(IndetImage, i);
        g += SummandImage;
      }
    swap(image, g);
  }


  void ApplyGeneral(RefRingElem image, ConstRefRingElem arg, const RingHom CoeffHom, ConstRefRingElem IndetImage)
  {
    ring R = owner(image);
    RingElem g(R);
    if (!IsZero(arg))
      for (long i=0; i<=deg(arg); ++i) 
      {
        RingElem SummandImage = CoeffHom(coeff(arg,i));
        CoCoA_ASSERT(owner(SummandImage) == R);
        if (IsZero(SummandImage)) continue; // efficiency hack????
        if (i != 0) 
          SummandImage *= power(IndetImage, i);
        g += SummandImage;
      }
    swap(image, g);
  }
}  // end of anonymous namespace

  void DenseUPolyRingBase::HomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    RingElem ans(myCodomain);  // Putting result into ans is exception safe and avoids aliasing problems.
    if ( IsDenseUPolyRing(myCodomain) )
      ApplyDPRCodomain(ans, ConstRefRingElem(myDomain, rawarg), myCoeffHom, myIndetImage);
    else
      ApplyGeneral(ans, ConstRefRingElem(myDomain, rawarg), myCoeffHom, myIndetImage);
    myCodomain->mySwap(rawimage, raw(ans));
  }


  void DenseUPolyRingBase::HomImpl::myOutputSelfDetails(std::ostream& out) const
  {
    DenseUPolyRing P = AsDenseUPolyRing(myDomain);
    if (NumIndets(P) == 0) return;
    out << " sending "
        << "(" << indet(P, 0) << " |--> " << myIndetImage << ")";
  }


  RingHom DenseUPolyRingBase::myHomCtor(const ring& codomain, const RingHom& CoeffHom, const std::vector<RingElem>& IndetImages) const
  {
    // Args already sanity checked by PolyRingHom (see PolyRing.C)
// DON'T KNOW IF I REALLY WANT TO MAKE THIS CHECK...
//       // Check to see if we're building an identity homomorphism
//       if (ring(this) == codomain && IsIdentity(CoeffHom))
//       {
//         bool IndetsFixed = true;
//         for (size_t i=0; i < myNumIndetsValue; ++i)
//           IndetsFixed &= (myIndetVector[i] == IndetImages[i]);
//         if (IndetsFixed) return IdentityHom(ring(this));
//       }
      // General case
    return RingHom(new HomImpl(DenseUPolyRing(this), codomain, CoeffHom, IndetImages[0]));
  }


  RingHom DenseUPolyRingBase::myCompose(const RingHom& phi, const RingHom& theta) const
  {
    vector<RingElem> IndetImages;
    IndetImages.push_back(phi(theta(myIndets()[0])));

    return myHomCtor(codomain(phi), phi(theta(myCoeffEmbeddingHom())), IndetImages);
  }


  //-- CoeffEmbeddingHomImpl ----------------------------------------

  //---------------------------------------------------------------------------
  // Functions for the class DenseUPolyRingBase::CoeffEmbeddingHomImpl


  DenseUPolyRingBase::CoeffEmbeddingHomImpl::CoeffEmbeddingHomImpl(const DenseUPolyRing& P):
    RingHomEmbeddingBase(CoeffRing(P), P)
  {}


  void DenseUPolyRingBase::CoeffEmbeddingHomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    DenseUPolyRing P(AsDenseUPolyRing(myCodomain));
    RingElem ans(P);  // don't use image here for aliasing
    // ??? ANNA profile this:  (probably better to have myMonomial)
    if (!myDomain->myIsZero(rawarg))
      ans = monomial(P, ConstRefRingElem(myDomain, rawarg), 0);
    P->mySwap(rawimage, raw(ans));
  }


} // end of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/DenseUPolyRing.C,v 1.19 2010/02/04 10:12:04 bigatti Exp $
// $Log: DenseUPolyRing.C,v $
// Revision 1.19  2010/02/04 10:12:04  bigatti
// -- added "mul" for ideal (implemented only for SparsePolyRing)
//
// Revision 1.18  2009/10/02 13:27:26  bigatti
// -- unique implementation of myDiv in PolyRing.C
//
// Revision 1.17  2009/09/24 16:24:19  abbott
// Added include directive (after removing it from RingFp.H).
//
// Revision 1.16  2009/07/24 12:26:43  abbott
// Added CommonDenom function for polynomials.
//
// Revision 1.15  2009/07/02 16:32:11  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.14  2009/05/22 10:24:12  bigatti
// -- updated def of IsQuotientOfZ in myOutput
//
// Revision 1.13  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.12  2008/12/12 15:00:53  bigatti
// -- added { } to disambiguate "if"
//
// Revision 1.11  2008/04/21 12:55:39  abbott
// Fixed some minor bugs, and improved code layout slightly.
//
// Revision 1.10  2008/03/12 16:47:14  bigatti
// -- added: myExgcd, myPrimeTest, myMaximalTest
//
// Revision 1.9  2007/12/21 12:29:08  bigatti
// -- abstract implementation in DenseUPolyRing of myDiv, myIsDivisible, myIsInvertible, myGcd
// -- abstract implementation in DenseUPolyRing of ideal operations
// -- some cleaning
//
// Revision 1.8  2007/12/07 15:27:01  bigatti
// -- default implementation of "IamOne" in ideal.C
//
// Revision 1.7  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.6  2007/12/04 14:13:23  bigatti
// -- added forgotten declaration.... sorry
//
// Revision 1.5  2007/12/04 13:44:50  bigatti
// -- fixed problem with myMinCapacity in constructor
// -- commented out myPowerSmallExp in RingDenseUPolyCleanImpl
// -- refined myPowerSmallExp in DenseUPolyRing
//
// Revision 1.4  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/10/15 12:40:31  bigatti
// -- new implementation for myAddMul
// -- new: coeffs between deg+1 and size are now guaranteed to be 0
//
// Revision 1.2  2007/10/10 14:02:37  bigatti
// -- added myMulBy1MinusXExp
// -- fixed a few little bugs
//
// Revision 1.1  2007/10/05 15:28:56  bigatti
// -- added abstract class DenseUPolyRing for representing dense
//    univariate polynomials
// -- added concrete class RingDenseUPolyClean, the cleanest
//    implementation
// -- just for testing, still horribly inefficient and incomplete
//
