//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Implementation file for the class FreeModule


#include "CoCoA/FreeModule.H"
#include "CoCoA/assert.H"
#include "CoCoA/degree.H"
#include "CoCoA/module.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/ModuleTermOrdering.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/SparsePolyRing.H"

#include <cstddef>
using std::size_t;
#include <vector>
using std::vector;
#include <algorithm>
//using std::swap;
#include <iostream>
using std::ostream;
#include <memory>
using std::auto_ptr;
#include <new>
//for placement new

namespace CoCoA
{

  namespace // anonymous namespace
  {
    inline const RingElem* import(ModuleBase::ConstRawPtr rawx)
    {
      return static_cast<const RingElem*>(rawx.ptr);
    }

    inline RingElem* import(ModuleBase::RawPtr& rawx)
    {
      return static_cast<RingElem*>(rawx.ptr);
    }


    // This is an exception clean "constructor" for module elements
    void CreateZeroVector(void* ptr, std::size_t NumCompts, ring R)
    {
      RingElem* array = static_cast<RingElem*>(ptr);
      size_t i;
      try
      {
        for (i=0; i < NumCompts; ++i)
          new(&array[i]) RingElem(R);  // placement new
      }
      catch(...)
      {
        while (i-- > 0)
          array[i].~RingElem();
        throw;
      }
    }

    // This is an exception clean "constructor" for module elements
    void CopyVector(void* ptr, std::size_t NumCompts, const void* copyme)
    {
      RingElem* array = static_cast<RingElem*>(ptr);
      const RingElem* rhs = static_cast<const RingElem*>(copyme);
      size_t i;
      try
      {
        for (i=0; i < NumCompts; ++i)
          new(&array[i]) RingElem(rhs[i]);  // placement new
      }
      catch(...)
      {
        while (i-- > 0)
          array[i].~RingElem();
      }
    }

    void DeleteVector(void* ptr, std::size_t NumCompts)
    {
      RingElem* array = static_cast<RingElem*>(ptr);
      for (size_t i=NumCompts; i-- > 0;)
        array[i].~RingElem();
    }

  } // end of anonymous namespace


  class FreeModuleImpl: public FreeModuleBase
  {
    // Two typedefs to save typing.
    typedef FGModuleBase::RawPtr RawPtr;
    typedef FGModuleBase::ConstRawPtr ConstRawPtr;

  protected:
    friend FreeModule NewFreeModule(const ring& R, MachineInteger NumCompts);
    FreeModuleImpl(const ring& R, std::size_t NumCompts);
//???    ~FreeModuleImpl();  // can be handy for debugging
  public:
    size_t myNumCompts() const;
    const ring& myBaseRing() const;
    const FreeModule& myAmbientFreeModule() const;
    const std::vector<ModuleElem>& myGens() const;
    const std::vector<ModuleElem>& myTidyGens() const;

    const ModuleElem& myZero() const;
    void myNew(RawPtr& rawv) const;
    void myNew(RawPtr& rawv, ConstRawPtr rawt) const;
    void myDelete(RawPtr& rawv) const;                                              ///< destroys v (incl all resources)
    void mySwap(RawPtr& rawv, RawPtr& raww) const;                                ///< swap(v, w)
    void myAssign(RawPtr& rawlhs, ConstRawPtr rawv) const;                        ///< lhs = v
    ConstRefRingElem myCompt(const RawPtr& rawv, size_t pos) const;               ///< v[pos]
    RefRingElem myCompt(RawPtr& rawv, size_t pos) const;                          ///< v[pos]
    void myNegate(RawPtr& rawlhs, ConstRawPtr rawv) const;                        ///< lhs = -v
    void myAdd(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const;         ///< lhs = v+w
    void mySub(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const;         ///< lhs = v-w

    void myMul(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const; ///< lhs = r*v
    void myDiv(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const; ///< lhs = (1/r)*v
    void myOutput(std::ostream& out, ConstRawPtr rawv) const;                     ///< out << v
    void myOutputSelf(std::ostream& out) const;                                   ///< out << M
    void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawv) const;                 ///< OMOut << v
    void myOutputSelf(OpenMathOutput& OMOut) const;                               ///< OMOut << M
    bool myIsZero(ConstRawPtr rawv) const;                                        ///< v == 0
//???    bool IsZeroAddMul(RawPtr& rawlhs, RingElemConstRawPtr rawy, ConstRawPtr rawz) const;  // lhs += y*z, result says whether lhs == 0.
    bool myIsEqual(ConstRawPtr, ConstRawPtr) const;
  ///  void convert(string&, RawPtr) const;

  protected: // data members
    const ring myR;
    const size_t myNumComptsValue;
    mutable MemPool myMemMgr;               // This must come before myZeroValue.
    std::auto_ptr<ModuleElem> myZeroValue;  // This must come after myMemMgr.
    std::vector<ModuleElem> myGensArray;
    const FreeModule myM; // so that myAmbientFreeModule can work
  };



  FreeModuleImpl::FreeModuleImpl(const ring& R, std::size_t n):
      myR(R),
      myNumComptsValue(n),
      myMemMgr(n*sizeof(RingElem), "FreeModuleImpl.myMemMgr"), // MUST HAVE n != 0
      myZeroValue(new ModuleElem(module(this))),
      myM(this)
  {
    CoCoA_ASSERT("FreeModule must have at least one compt" && n > 0);
    CoCoA_ASSERT("FreeModule dim ludicrous" && n < 65536);
    for (size_t i=0; i < myNumComptsValue; ++i)
    {
      myGensArray.push_back(myZero());
      import(raw(myGensArray[i]))[i] = 1;
    }
    myRefCountZero();
  }


//    FreeModuleImpl::~FreeModuleImpl()
//    {}


  std::size_t FreeModuleImpl::myNumCompts() const
  {
    return myNumComptsValue;
  }


  const ring& FreeModuleImpl::myBaseRing() const
  {
    return myR;
  }


  const FreeModule& FreeModuleImpl::myAmbientFreeModule() const
  {
    return myM;
  }


  const std::vector<ModuleElem>& FreeModuleImpl::myGens() const
  {
    return myGensArray;
  }


  const std::vector<ModuleElem>& FreeModuleImpl::myTidyGens() const
  {
    return myGensArray;
  }


  const ModuleElem& FreeModuleImpl::myZero() const
  {
    return *myZeroValue;
  }


  void FreeModuleImpl::myNew(RawPtr& rawv) const
  {
    rawv.ptr = myMemMgr.alloc();
    CreateZeroVector(rawv.ptr, myNumComptsValue, myR);
  }


  void FreeModuleImpl::myNew(RawPtr& rawv, ConstRawPtr rawcopy) const
  {
    rawv.ptr = myMemMgr.alloc();
    CopyVector(rawv.ptr, myNumComptsValue, rawcopy.ptr);
  }


  void FreeModuleImpl::myDelete(RawPtr& rawv) const
  {
    // Kill elements in reverse order (might scramble the MemPool less?)
    DeleteVector(rawv.ptr, myNumComptsValue);
    myMemMgr.free(rawv.ptr);
  }


  void FreeModuleImpl::mySwap(RawPtr& rawv, RawPtr& raww) const
  {
    std::swap(rawv.ptr, raww.ptr);
  }


  ConstRefRingElem FreeModuleImpl::myCompt(const RawPtr& rawv, size_t pos) const
  {
    const RingElem* vv = import(rawv);
    return vv[pos];
  }


  RefRingElem FreeModuleImpl::myCompt(RawPtr& rawv, size_t pos) const
  {
    RingElem* vv = import(rawv);
    return vv[pos];
  }


  void FreeModuleImpl::myAssign(RawPtr& rawlhs, ConstRawPtr rawv) const
  {
    RingElem* lvec = import(rawlhs);
    const RingElem* vec = import(rawv);
    for (size_t i=0; i < myNumComptsValue; ++i)
      lvec[i] = vec[i];
  }


  void FreeModuleImpl::myNegate(RawPtr& rawlhs, ConstRawPtr rawv) const
  {
    RingElem* lvec= import(rawlhs);
    const RingElem* vec = import(rawv);
    for (size_t i=0; i < myNumComptsValue; ++i)
      lvec[i] = -vec[i];
  }


  void FreeModuleImpl::myAdd(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const
  {
    RingElem* lvec = import(rawlhs);
    const RingElem* vec1 = import(rawv);
    const RingElem* vec2 = import(raww);
    for (size_t i=0; i < myNumComptsValue; ++i)
      lvec[i] = vec1[i] + vec2[i];
  }


  void FreeModuleImpl::mySub(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const
  {
    RingElem* lvec = import(rawlhs);
    const RingElem* vec1 = import(rawv);
    const RingElem* vec2 = import(raww);
    for (size_t i=0; i < myNumComptsValue; ++i)
      lvec[i] = vec1[i] - vec2[i];
  }


  void FreeModuleImpl::myMul(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const
  {
    RingElem* lvec = import(rawlhs);
    const RingElem* vec = import(rawv);
    for (size_t i=0; i < myNumComptsValue; ++i)
      myR->myMul(raw(lvec[i]), rawx, raw(vec[i]));
  }


  void FreeModuleImpl::myDiv(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const
  {
    RingElem* lvec = import(rawlhs);
    const RingElem* vec = import(rawv);
    for (size_t i=0; i < myNumComptsValue; ++i)
      myR->myDiv(raw(lvec[i]), rawx, raw(vec[i]));
  }


  void FreeModuleImpl::myOutput(std::ostream& out, ConstRawPtr rawv) const
  {
    // Guaranteed that myNumComptsValue > 0
    const RingElem* vec = import(rawv);
    out << "[" << vec[0];
    for (size_t i=1; i < myNumComptsValue; ++i)
    {
      out << ", " << vec[i];
    }
    out << "]";
  }


  void FreeModuleImpl::myOutputSelf(std::ostream& out) const
  {
    out << "FreeModule(" << myR << ", " << myNumComptsValue << ")";
  }


  void FreeModuleImpl::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawv) const
  {
    // Guaranteed that myNumComptsValue > 0
    const RingElem* vec = import(rawv);
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("???", "FreeModuleElement"); // BUG: what should this OMSymbol be???
    myOutputSelf(OMOut);
    for (size_t i=0; i < myNumComptsValue; ++i)
    {
      OMOut << vec[i];
    }
    OMOut->mySendApplyEnd();
  }


  void FreeModuleImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("???", "FreeModule"); // BUG: what should this OMSymbol be???
    OMOut << myR << myNumComptsValue;
    OMOut->mySendApplyEnd();
  }


  bool FreeModuleImpl::myIsZero(ConstRawPtr rawv) const
  {
    const RingElem* vec = import(rawv);
    for (size_t i=0; i < myNumComptsValue; ++i)
      if (!IsZero(vec[i])) return false;
    return true;
  }


///???    bool IsZeroAddMul(RawPtr& rawlhs, RingElemConstRawPtr rawy, ConstRawPtr rawz) const;  // lhs += y*z, result says whether lhs == 0.


  bool FreeModuleImpl::myIsEqual(ConstRawPtr rawv, ConstRawPtr raww) const
  {
    const RingElem* vec1 = import(rawv);
    const RingElem* vec2 = import(raww);
    for (size_t i=0; i < myNumComptsValue; ++i)
      if (vec1[i] != vec2[i]) return false;
    return true;
  }


  //----------------------------------------------------------------------
  // non-member functions

  FreeModule NewFreeModule(const ring& R, MachineInteger NumCompts)
  {
    if (IsNegative(NumCompts))
      CoCoA_ERROR(ERR::NotNonNegative, "NewFreeModule(R,n)");

    if (IsSparsePolyRing(R))
      return NewGradedFreeModule(AsSparsePolyRing(R), NumCompts);
    return FreeModule(new FreeModuleImpl(R, AsUnsignedLong(NumCompts)));
  }


  bool IsFreeModule(const module& M)
  {
    return dynamic_cast<const FreeModuleImpl*>(M.myRawPtr()) != 0;
  }


  FreeModule AsFreeModule(const module& M)
  {
    if (!IsFreeModule(M))
      CoCoA_ERROR(ERR::NotFreeModule, "AsFreeModule(module)");
    return FreeModule(static_cast<const FreeModuleImpl*>(M.myRawPtr()));
  }


//   FreeModule::FreeModule(const FreeModuleBase* MPtr):
//       FGModule(MPtr)
//   {}


//   FreeModule::~FreeModule()
//   {}


  size_t FirstNonZeroPos(const ModuleElem& v)
  {
    for (size_t i=0 ; i<NumCompts(AsFreeModule(owner(v))) ; ++i )
      if ( !IsZero(v[i]) ) return i;
    CoCoA_ERROR("vector is zero", "FirstNonZeroPos(vector)");
    return 0; // just to keep the compiler quiet
  }




  class GradedFreeModuleImpl: public FreeModuleImpl
  {
    // should I use RawPtrs? it would make the code uglier
    //    typedef FGModuleBase::RawPtr RawPtr;
    //    typedef FGModuleBase::ConstRawPtr ConstRawPtr;

    class CmpBase
    {
    public:
      CmpBase(const std::vector<degree>& shifts);
      virtual ~CmpBase() {};
    public:
      virtual std::size_t myLPos(const ModuleElem& v) const =0;
      virtual int myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const =0;
      void myWDeg(degree& d, const ModuleElem& v) const;
      int myCmpWDeg(const ModuleElem& v1, const ModuleElem& v2) const;
    protected:
      std::vector<degree> myShiftsValue;
    };

    // ???ANNA: this is NOT degree compatible
    class PosWDegTOImpl: public CmpBase
    {
    public:
      PosWDegTOImpl(const std::vector<degree>& shifts);
      virtual ~PosWDegTOImpl() {};
      virtual std::size_t myLPos(const ModuleElem& v) const;
      virtual int myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const;
      int myCmpPos(const size_t pos1, const size_t pos2) const;
    private:
      std::vector<std::size_t> myComptsOrd;
    };

    class WDegTOPosImpl: public CmpBase
    {
    public:
      WDegTOPosImpl(const std::vector<degree>& shifts);
      virtual ~WDegTOPosImpl() {};
      virtual std::size_t myLPos(const ModuleElem& v) const;
      virtual int myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const;
      int myCmpPos(const size_t pos1, const size_t pos2) const;
    private:
      std::vector<std::size_t> myComptsOrd;
    };


    class WDegPosTOImpl: public CmpBase
    {
    public:
      WDegPosTOImpl(const std::vector<degree>& shifts);
      virtual ~WDegPosTOImpl() {};
      virtual std::size_t myLPos(const ModuleElem& v) const;
      virtual int myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const;
      int myCmpPos(const size_t pos1, const size_t pos2) const;
    private:
      std::vector<std::size_t> myComptsOrd;
    };



    //--- end of CmpBase ------------------------------------------------------------

  private:
    GradedFreeModuleImpl(const SparsePolyRing& P, const ModuleTermOrdering& O);
//???    ~GradedFreeModuleImpl();  // can be handy for debugging
    friend FreeModule NewGradedFreeModule(const SparsePolyRing& P, MachineInteger NumCompts);
    friend FreeModule NewFreeModule(const ring& P, const ModuleTermOrdering& O);

  public:
    const ModuleTermOrdering& myModuleTermOrdering() const;
    std::size_t myLPos(const ModuleElem&  v) const;
    int myCmpLT(const ModuleElem&  v1, const ModuleElem&  v2) const;
    void myWDeg(degree& d, const ModuleElem&  v) const;
    int myCmpWDeg(const ModuleElem&  v1, const ModuleElem&  v2) const;

  private: // data members
    const ModuleTermOrdering myModuleTermOrd; ///< abstract description of the ordering
    std::auto_ptr<CmpBase> myOrdPtr; ///< actual implementation of the ordering [should be const???]
  };


  class GradedFreeModule: public FreeModule  // copied from SparsePolyRing
  {
  public:
    explicit GradedFreeModule(const GradedFreeModuleImpl* MPtr);
    // Default copy ctor works fine.
    // Default dtor works fine.
  private: // disable assignment
    GradedFreeModule& operator=(const GradedFreeModule& rhs); // NEVER DEFINED -- assignment disabled
  public:
    const GradedFreeModuleImpl* operator->() const; // allow member fns to be called
  };


  inline GradedFreeModule::GradedFreeModule(const GradedFreeModuleImpl* MPtr):
    FreeModule(MPtr)
  {}


  inline const GradedFreeModuleImpl* GradedFreeModule::operator->() const
  {
    return static_cast<const GradedFreeModuleImpl*>(myRawPtr());
  }


  bool IsGradedFreeModule(const module& M)
  {
    return dynamic_cast<const GradedFreeModuleImpl*>(M.myRawPtr()) != 0;
  }


  inline GradedFreeModule AsGradedFreeModule(const module& M)
  {
    if (!IsGradedFreeModule(M))
      CoCoA_ERROR(ERR::NotGradedModule, "AsGradedFreeModule(module)");
    return GradedFreeModule(static_cast<const GradedFreeModuleImpl*>(M.myRawPtr()));
  }


  //--- GradedFreeModuleImpl

  GradedFreeModuleImpl::GradedFreeModuleImpl(const SparsePolyRing& P, const ModuleTermOrdering& O):
      FreeModuleImpl(P, NumComponents(O)),
      myModuleTermOrd(O),
      myOrdPtr()
  {
    if ( IsWDegPosTO(O) )
      myOrdPtr.reset(new WDegPosTOImpl(shifts(O)));
    else if ( IsWDegTOPos(O) )
      myOrdPtr.reset(new WDegTOPosImpl(shifts(O)));
    else
      CoCoA_ERROR("ModuleTermOrdering not implemented", "GradedModuleCmpBase");
  }


  FreeModule NewFreeModule(const ring& P, const ModuleTermOrdering& O)
  {
    if (!IsSparsePolyRing(P))
      CoCoA_ERROR(ERR::NotSparsePolyRing, "NewFreeModule(R, Ord)");
    return FreeModule(new GradedFreeModuleImpl(AsSparsePolyRing(P), O));
  }


  FreeModule NewGradedFreeModule(const SparsePolyRing& P, MachineInteger NumCompts)
  {
    if (IsNegative(NumCompts))
      CoCoA_ERROR(ERR::NotNonNegative, "NewGradedFreeModule(P,n)");
    return FreeModule(new GradedFreeModuleImpl(P, NewWDegTOPos(ordering(PPM(P)), AsUnsignedLong(NumCompts))));
  }


  const ModuleTermOrdering& GradedFreeModuleImpl::myModuleTermOrdering() const
  { return myModuleTermOrd; }


  std::size_t GradedFreeModuleImpl::myLPos(const ModuleElem&  v) const
  { return myOrdPtr->myLPos(v); }


  int GradedFreeModuleImpl::myCmpLT(const ModuleElem&  v1, const ModuleElem&  v2) const
  { return myOrdPtr->myCmpLT(v1, v2); }


  void GradedFreeModuleImpl::myWDeg(degree& d, const ModuleElem&  v) const
  { return myOrdPtr->myWDeg(d, v); }


  int GradedFreeModuleImpl::myCmpWDeg(const ModuleElem&  v1, const ModuleElem&  v2) const
  { return myOrdPtr->myCmpWDeg(v1, v2); }


  //--- non-member fns

  const ModuleTermOrdering& ordering(const FreeModule& M)
  {
    if (!IsGradedFreeModule(M))
      CoCoA_ERROR(ERR::NotGradedModule, "ordering(M)");
   return dynamic_cast<const GradedFreeModuleImpl*>(M.myRawPtr())->myModuleTermOrdering();
  }


  const std::vector<degree>& shifts(const FreeModule& M)
  {
    if (!IsGradedFreeModule(M))
      CoCoA_ERROR(ERR::NotGradedModule, "shifts(M)");
    return shifts(ordering(M));
  }


  //----------------------------------------------------------------------
  //--- CmpBase


  GradedFreeModuleImpl::CmpBase::CmpBase(const std::vector<degree>& shifts):
    myShiftsValue(shifts)
  {}


  void GradedFreeModuleImpl::CmpBase::myWDeg(degree& d, const ModuleElem& v) const
  {
    const size_t pos = myLPos(v);
    d = wdeg(v[pos]) + myShiftsValue[pos];
  }


  int GradedFreeModuleImpl::CmpBase::myCmpWDeg(const ModuleElem& v1, const ModuleElem& v2) const
  {
    return CmpWDeg(v1[LPos(v1)], v2[LPos(v2)]);
  }


  //----------------------------------------------------------------------
  //--- WDegTOPosImpl

  GradedFreeModuleImpl::WDegTOPosImpl::WDegTOPosImpl(const std::vector<degree>& shifts):
    CmpBase(shifts)
  {
    for ( size_t i=0 ; i<shifts.size() ; ++i )
      myComptsOrd.push_back(i);
  }


  std::size_t GradedFreeModuleImpl::WDegTOPosImpl::myLPos(const ModuleElem& v) const
  {
    size_t i;
    for ( i=0 ; i<myComptsOrd.size() ; ++i )
      if (!IsZero(v[myComptsOrd[i]]))
        break;
    CoCoA_ASSERT( i!=myComptsOrd.size() );
    size_t MaxPos = myComptsOrd[i];
    degree MaxWDeg(wdeg(v[MaxPos])+myShiftsValue[MaxPos]);
    for ( ++i ; i<myComptsOrd.size() ; ++i )
      if ( !IsZero(v[myComptsOrd[i]]))
      {
        size_t pos = myComptsOrd[i];
        int CmpWDegree = cmp(MaxWDeg, wdeg(v[pos])+myShiftsValue[pos]);
        if ( CmpWDegree==-1 || (CmpWDegree==0 && LPP(v[MaxPos])<LPP(v[pos])) )
        {
          MaxPos = pos;
          MaxWDeg = wdeg(v[MaxPos]) + myShiftsValue[MaxPos];
        }
      }
    return MaxPos;
  }


  int GradedFreeModuleImpl::WDegTOPosImpl::myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const
  {
    size_t LPos1 = myLPos(v1), LPos2 = myLPos(v2);
    int CmpFlag = cmp( wdeg(v1[LPos1]) + myShiftsValue[LPos1],
                       wdeg(v2[LPos2]) + myShiftsValue[LPos2] );
    if ( CmpFlag !=0 ) return CmpFlag;
    CmpFlag = cmp(LPP(v1[LPos1]), LPP(v2[LPos2]) );
    if ( CmpFlag !=0 ) return CmpFlag;
    return myCmpPos(LPos1, LPos2);
  }


  int GradedFreeModuleImpl::WDegTOPosImpl::myCmpPos(const size_t pos1, const size_t pos2) const
  {
    CoCoA_ASSERT( /*pos1>=0 &&*/ pos1<myComptsOrd.size() );
    CoCoA_ASSERT( /*pos2>=0 &&*/ pos2<myComptsOrd.size() );
    if ( pos1 == pos2 ) return 0;
    for ( size_t i=0 ; i<myComptsOrd.size() ; ++i )
    {
      if ( myComptsOrd[i]==pos1 ) return 1;
      if ( myComptsOrd[i]==pos2 ) return -1;
    }
    return 0;  // just to keep the compiler quiet
  }

  //----------------------------------------------------------------------
  //--- PosWDegTOImpl

  GradedFreeModuleImpl::PosWDegTOImpl::PosWDegTOImpl(const std::vector<degree>& shifts):
    CmpBase(shifts)
  {
    for ( size_t i=0 ; i<shifts.size() ; ++i )
      myComptsOrd.push_back(i);
  }


  std::size_t GradedFreeModuleImpl::PosWDegTOImpl::myLPos(const ModuleElem& v) const
  {
    size_t i;
    for ( i=0 ; i<myComptsOrd.size() ; ++i )
      if (!IsZero(v[myComptsOrd[i]]))
        break;
    return i;
  }


  int GradedFreeModuleImpl::PosWDegTOImpl::myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const
  {
    size_t LPos1 = myLPos(v1), LPos2 = myLPos(v2);
    int CmpFlag = cmp( wdeg(v1[LPos1]) + myShiftsValue[LPos1],
                       wdeg(v2[LPos2]) + myShiftsValue[LPos2] );
    if ( CmpFlag !=0 ) return CmpFlag;
    CmpFlag = cmp(LPP(v1[LPos1]), LPP(v2[LPos2]) );
    if ( CmpFlag !=0 ) return CmpFlag;
    return myCmpPos(LPos1, LPos2);
  }


  int GradedFreeModuleImpl::PosWDegTOImpl::myCmpPos(const size_t pos1, const size_t pos2) const
  {
    CoCoA_ASSERT( /*pos1>=0 &&*/ pos1<myComptsOrd.size() );
    CoCoA_ASSERT( /*pos2>=0 &&*/ pos2<myComptsOrd.size() );
    if ( pos1 == pos2 ) return 0;
    for ( size_t i=0 ; i<myComptsOrd.size() ; ++i )
    {
      if ( myComptsOrd[i]==pos1 ) return 1;
      if ( myComptsOrd[i]==pos2 ) return -1;
    }
    return 0;  // just to keep the compiler quiet
  }

  //----------------------------------------------------------------------
  //--- WDegPosTOImpl

  GradedFreeModuleImpl::WDegPosTOImpl::WDegPosTOImpl(const std::vector<degree>& shifts):
    CmpBase(shifts)
  {
    for ( size_t i=0 ; i<shifts.size() ; ++i )
      myComptsOrd.push_back(i);
  }


  std::size_t GradedFreeModuleImpl::WDegPosTOImpl::myLPos(const ModuleElem& v) const
  {
    size_t i;
    for ( i=0 ; i<myComptsOrd.size() ; ++i )
      if (!IsZero(v[myComptsOrd[i]]))
        break;
    CoCoA_ASSERT( i!=myComptsOrd.size() );
    size_t MaxPos = myComptsOrd[i];

    degree MaxWDeg(wdeg(v[MaxPos]) + myShiftsValue[MaxPos]);
    for ( ++i ; i<myComptsOrd.size() ; ++i )
      if ( !IsZero(v[myComptsOrd[i]]))
      {
        int CmpWDegree = cmp(MaxWDeg,
                             wdeg(v[myComptsOrd[i]])+myShiftsValue[myComptsOrd[i]]);
        if ( CmpWDegree==-1 )
        {
          MaxPos = myComptsOrd[i];
          MaxWDeg = wdeg(v[MaxPos]) + myShiftsValue[MaxPos];
        }
      }
    return MaxPos;
  }


  int GradedFreeModuleImpl::WDegPosTOImpl::myCmpLT(const ModuleElem& v1, const ModuleElem& v2) const
  {
    size_t LPos1 = myLPos(v1), LPos2 = myLPos(v2);
    int CmpFlag = cmp( wdeg(v1[LPos1]) + myShiftsValue[LPos1],
                       wdeg(v2[LPos2]) + myShiftsValue[LPos2] );
    if ( CmpFlag !=0 ) return CmpFlag;
    if ( LPos1 != LPos2 ) return ( LPos1>LPos2 ? 1 : -1 );
    return cmp(LPP(v1[LPos1]), LPP(v2[LPos2]));
  }


  int GradedFreeModuleImpl::WDegPosTOImpl::myCmpPos(const size_t pos1, const size_t pos2) const
  {
    CoCoA_ASSERT( /*pos1>=0 &&*/ pos1<myComptsOrd.size() );
    CoCoA_ASSERT( /*pos2>=0 &&*/ pos2<myComptsOrd.size() );
    if ( pos1 == pos2 ) return 0;
    for ( size_t i=0 ; i<myComptsOrd.size() ; ++i )
    {
      if ( myComptsOrd[i]==pos1 ) return 1;
      if ( myComptsOrd[i]==pos2 ) return -1;
    }
    return 0;  // just to keep the compiler quiet
  }

  //----------------------------------------------------------------------


  //  ConstRefPPMonoidElem LPP(const ModuleElem& v);

  std::size_t LPos(const ModuleElem& v)
  {
    if ( !IsGradedFreeModule(owner(v)) )
      CoCoA_ERROR(ERR::NotGradedModule, "LPos(v)");
    return AsGradedFreeModule(owner(v))->myLPos(v);
  }


  ConstRefPPMonoidElem LPP(const ModuleElem& v)
  {
    if ( !IsGradedFreeModule(owner(v)) )
      CoCoA_ERROR(ERR::NotGradedModule, "LPP(v)");
    return LPP(v[AsGradedFreeModule(owner(v))->myLPos(v)]);
  }


  degree wdeg(const ModuleElem& v)
  {
    if ( !IsGradedFreeModule(owner(v)) )
      CoCoA_ERROR(ERR::NotGradedModule, "wdeg(v)");
    degree ans(GradingDim(ordering(AsGradedFreeModule(owner(v)))));
    AsGradedFreeModule(owner(v))->myWDeg(ans, v);
    return ans;
  }


  int CmpWDeg(const ModuleElem& v1, const ModuleElem& v2)
  {
    if ( owner(v1) != owner(v2) )
      CoCoA_ERROR(ERR::MixedModules, "CmpWDeg(v1, v2)");
    if ( !IsGradedFreeModule(owner(v1)) )
      CoCoA_ERROR(ERR::NotGradedModule, "CmpWDeg(v1, v2)");
    return AsGradedFreeModule(owner(v1))->myCmpWDeg(v1, v2);
  }


  bool IsHomog(const ModuleElem& v)
  {
    if ( !IsGradedFreeModule(owner(v)) )
      CoCoA_ERROR(ERR::NotGradedModule, "IsHomog(v)");
    if (IsZero(v)) return true;
    const FreeModule FM=AsFreeModule(owner(v));
    const std::vector<degree>& s=shifts(FM);
    size_t FNZP=FirstNonZeroPos(v);
    degree d(wdeg(v[FNZP])+s[FNZP]);
    for (size_t i=FNZP ; i<NumCompts(FM) ; ++i )
    {
      if (!IsZero(v[i]))
        if (!IsHomog(v[i]) || wdeg(v[i])+s[i] != d)
          return false;
    }
    return true;
  }

  //---------------------------------------------------------------------------

//   ConstRefRingElem FreeModuleElem::operator[](MachineInteger pos) const
//   {
// //     if (!IsFGModule(owner(*this)))
// //       CoCoA_ERROR(ERR::NotFGModule, "FreeModuleElem[pos]");
//     const unsigned long i = AsUnsignedLong(pos);
//     if (IsNegative(pos) || i >= NumCompts(AsFGModule(owner(*this))))
//       CoCoA_ERROR(ERR::BadComptIndex, "FreeModuleElem[pos]");
//     return AsFGModule(owner(*this))->myCompt(raw(*this), i);
//   }


//   RefRingElem FreeModuleElem::operator[](MachineInteger pos)
//   {
// //     if (!IsFGModule(owner(*this)))
// //       CoCoA_ERROR(ERR::NotFGModule, "FreeModuleElem[pos]");
//     const unsigned long i = AsUnsignedLong(pos);
//     if (IsNegative(pos) || i >= NumCompts(AsFGModule(owner(*this))))
//       CoCoA_ERROR(ERR::BadComptIndex, "FreeModuleElem[pos]");
//     return AsFGModule(owner(*this))->myCompt(raw(*this), i);
//   }





}  // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/FreeModule.C,v 1.5 2009/12/03 17:26:34 abbott Exp $
// $Log: FreeModule.C,v $
// Revision 1.5  2009/12/03 17:26:34  abbott
// Renamed EFGModule to FGModule.
// Renamed ModuleBase member fns  myInit -> myNew, myKill -> myDelete.
// Removed some cruft (old code that was not used by anyone).
//
// Revision 1.4  2008/04/21 12:57:04  abbott
// Changed a size_t arg into MachineInteger.  Added defn of myCompts memfn.
// Improved layout of some comments.
//
// Revision 1.3  2008/03/12 16:35:18  bigatti
// -- changed: IsHomogeneous --> IsHomog
// -- changed: ERR:ZeroPoly --> ERR::ZeroRingElem
//
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.15  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.14  2007/01/15 13:37:51  cocoa
// -- added prefix "raw" to RawPtr arguments names
//
// Revision 1.13  2006/12/07 17:27:29  cocoa
// -- for compilation with _Wextra: commented out useless comparisons
//
// Revision 1.12  2006/11/24 17:01:43  cocoa
// -- reorganized includes of header files
//
// Revision 1.11  2006/11/20 14:57:17  cocoa
// -- added: (standard) sugar for modules
// -- fixed: non-homogeneous sysygies
// -- minor fixes     [M.Caboara]
//
// Revision 1.10  2006/11/14 17:46:20  cocoa
// -- changed: reference counting in modules now uses SmartPtrIRC
//
// Revision 1.9  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.8  2006/10/19 13:56:24  cocoa
// Added #include<new> whenever needed (i.e. for placement new).
//
// Revision 1.7  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.6  2006/10/06 16:43:54  cocoa
// -- added PosWDegTOImpl (warning: non-degree compatible ordering)
//
// Revision 1.5  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.4  2006/10/06 10:48:34  cocoa
// Removed #include references to GradedFreeModule.H
//
// Revision 1.3  2006/10/06 09:53:06  cocoa
// Moved definition of FreeModuleImpl out of the header file.
// Merged GradedFreeModule code into here so it can access FreeModuleImpl
// (which is no longer visible in the header file).
//
// Revision 1.2  2006/07/18 10:51:25  cocoa
// -- added: FirstNonZeroPos
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.8  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.7  2006/05/04 14:25:16  cocoa
// -- major cleaning of FreeModule: created GradedFreeModule and moved
//    some code around
//
// Revision 1.6  2006/04/21 14:56:33  cocoa
// Changed return type of myCompt member function: now it returns a
// ConstRefRingElem instead of a RingElem (i.e. a copy).
//
// Revision 1.5  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.4  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
// Revision 1.3  2005/11/29 13:04:47  cocoa
// -- added "const" to myCompt argument
//
// Revision 1.2  2005/11/24 16:09:38  cocoa
// -- added operator[] for ModuleElem
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.4  2005/09/28 12:06:40  cocoa
// -- fixed "n undeclared" when compiled with Debug on (inside CoCoA_ASSERT)
//
// Revision 1.3  2005/09/28 11:50:34  cocoa
// -- new code for graded modules
//
// Revision 1.2  2005/06/22 14:42:16  cocoa
// Renamed MemPool data member to myMemMgr
// (seems more sensible than myMemory).
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.4  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.3  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.3  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.2  2004/01/28 15:56:49  cocoa
// "Old style" code, brought in alignment with the new coding conventions.
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
