//   Copyright (c)  2005,2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/GMPAllocator.H"

#include "CoCoA/MemPool.H"
#include "CoCoA/error.H"
#include "CoCoA/assert.H"

#include "gmp.h"

#include <cstring>
using std::memcpy;
#include <algorithm>
using std::min;

namespace // anonymous namespace for file local globals
{

  size_t GMPPoolSliceSize = 0; // size in bytes of slices in the MemPool (compile-time constant)
  CoCoA::MemPool* GMPPoolPtr = 0;

} // end of anonymous namespace


void* CoCoA_GMP_alloc(size_t sz)
{
  CoCoA_ASSERT(GMPPoolSliceSize != 0);
  if (sz <= GMPPoolSliceSize)
    return GMPPoolPtr->alloc();
  return malloc(sz);
}


void* CoCoA_GMP_realloc(void* ptr, size_t oldsz, size_t newsz)
{
  CoCoA_ASSERT(GMPPoolSliceSize != 0);
  if (oldsz <= GMPPoolSliceSize && newsz <= GMPPoolSliceSize)
    return ptr;
  const size_t n = min(oldsz, newsz);
  void* dest=CoCoA_GMP_alloc(newsz);
  memcpy(dest, ptr, n);
  CoCoA_GMP_free(ptr, oldsz);
  return dest;
}


void CoCoA_GMP_free(void* ptr, size_t sz)
{
  CoCoA_ASSERT(GMPPoolSliceSize != 0);
  if (sz <= GMPPoolSliceSize)
    GMPPoolPtr->free(ptr);
  else
    free(ptr);
}


namespace CoCoA
{

  GMPAllocator::GMPAllocator(std::size_t SliceSize /* default = 16*/):
      myMemPool(SliceSize, "GMPAllocator.myMemPool")
  {
    if (GMPPoolPtr != 0)
      CoCoA_ERROR("Cannot change global GMPAllocator", "GMPAllocator ctor");
    GMPPoolPtr = &myMemPool;
    GMPPoolSliceSize = SliceSize;
    mp_set_memory_functions(&CoCoA_GMP_alloc, &CoCoA_GMP_realloc, &CoCoA_GMP_free);
  }

  GMPAllocator::~GMPAllocator()
  {
    if (GMPPoolPtr == 0)
      CoCoA_ERROR(ERR::SERIOUS, "GMPAllocator dtor: global GMPAllocator already killed");
    GMPPoolPtr = 0;
// See comments in GMPAllocator documentation before uncommenting the line below.
//    mp_set_memory_functions(&std::malloc, &std::realloc, &std::free); // need to #include <cstdlib>
  }

} // end of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/GMPAllocator.C,v 1.4 2010/03/23 11:57:31 abbott Exp $
// $Log: GMPAllocator.C,v $
// Revision 1.4  2010/03/23 11:57:31  abbott
// Added comment about default value of ctor arg.
//
// Revision 1.3  2008/10/07 15:46:10  abbott
// Added missing log/history lines at the ends of a few files.
//
//
