//   Copyright (c)  2007-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/MachineInteger.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/utils.H"

#include <iostream>

namespace CoCoA
{

  namespace
  {
    // Compute  round(n/d)  rounding halves up.
    // NB cannot simply do  (n+d/2)/d  because of possible overflow!
    unsigned long uround_half_up(unsigned long n, unsigned long d)
    {
      CoCoA_ASSERT(d != 0);
      const unsigned long q = n/d;
      if (n-q*d <= (d-1)/2) return q;
      return q+1; // cannot overflow
    }

    // Compute  round(n/d)  rounding halves down.
    // NB cannot simply do  (n+(d-1)/2)/d  because of possible overflow!
    unsigned long uround_half_down(unsigned long n, unsigned long d)
    {
      CoCoA_ASSERT(d != 0);
      const unsigned long q = n/d;
      if (n-q*d <= d/2) return q;
      return q+1; // cannot overflow
    }
  }


  std::ostream& operator<<(std::ostream& out, const MachineInteger& n)
  {
    using std::operator<<; // need this to avoid infinite recursion as current fn hides std::operator<<
    if (IsNegative(n))
      return out << AsSignedLong(n);
    return out << AsUnsignedLong(n);
  }


  int cmp(const MachineInteger& n1, const MachineInteger& n2)
  {
    if (IsNegative(n1))
    {
      if (IsNegative(n2))
        return cmp<long,long>(AsSignedLong(n1), AsSignedLong(n2));
      else
        return -1;
    }
    // n1 is non-negative
    if (IsNegative(n2)) return 1;
    // Both n1 and n2 are non-negative
    return cmp<unsigned long, unsigned long>(AsUnsignedLong(n1), AsUnsignedLong(n2));
  }


  long RoundDiv(const MachineInteger& n, const MachineInteger& d)
  {
    if (IsZero(d))
      CoCoA_ERROR(ERR::DivByZero, "RoundDiv(n,d)");
    if (IsZero(n)) return 0;
    // n and d are both non-zero
    if (IsNegative(d))
    {
      if (IsNegative(n))
      {
        // Here n < 0 and d < 0
        // Can overflow if d == -1 and n == numeric_limits<long>::min()
        return NumericCast<long>(uround_half_up(abs(n), abs(d)));
      }
      // Here n > 0 and d < 0
      // Can overflow if d == -1 and n > 1+numeric_limits<long>::max()
      return -NumericCast<long>(uround_half_down(AsUnsignedLong(n), abs(d)));
    }
    // Here we know d > 0
    if (IsNegative(n))
    {
      // Here n < 0 and d > 0
      return -static_cast<long>(uround_half_down(abs(n), AsUnsignedLong(d))); // never overflows
    }
    // Both n > 0 and d > 0
    return NumericCast<long>(uround_half_up(AsUnsignedLong(n), AsUnsignedLong(d)));
  }


  long SmallPower(const MachineInteger& base, const MachineInteger& exponent)
  {
    if (IsNegative(exponent)) CoCoA_ERROR(ERR::NegExp, "SmallPower");
    if (IsZero(base) && IsZero(exponent)) CoCoA_ERROR(ERR::BadPwrZero, "SmallPower");
    if (!IsSignedLong(base)) CoCoA_ERROR(ERR::BadArg, "SmallPower: base is too big to fit into a long");
    if (IsZero(base)) return 0;
    if (IsZero(exponent)) return 1;
    unsigned long e = abs(exponent);
    long b = abs(base);
    const bool negative = (e&1) && IsNegative(base);
    long ans = 1;
    while (e != 1)
    {
      if (e&1) ans *= b;
      e >>= 1;
      b *= b;
    }
    ans *= b;
    if (negative) return -ans;
    return ans;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/MachineInteger.C,v 1.7 2010/03/05 18:39:49 abbott Exp $
// $Log: MachineInteger.C,v $
// Revision 1.7  2010/03/05 18:39:49  abbott
// Added SmallPower function -- currently undefined behaviour if overflow occurs!!
//
// Revision 1.6  2009/12/23 18:53:52  abbott
// Major change to conversion functions:
//   convert(..) is now a procedure instead of a function
//   IsConvertible(..) replaces the former convert(..) function
//   Added new NumericCast conversion function (placeholder for BOOST feature)
//   Consequent changes in code which uses these features.
//
// Revision 1.5  2009/10/08 13:39:47  abbott
// Renamed "round" into "RoundDiv".
// Added some new versions of "RoundDiv".
// Added a test for "RoundDiv".
//
// Revision 1.4  2009/06/05 12:14:55  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
// Revision 1.3  2008/12/11 10:47:36  abbott
// Fixed bug in IsZero (it appeared only when CoCoA_DEBUG was set).
// Some cleaning.
//
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1  2007/05/21 12:57:28  abbott
// New class for passing machine integers as args; includes some simple
// operations on machine integers (cmp, gcd, IsNegative,...).  Operations
// between ZZ and machine integers modified to use the new class.  Inexact
// integer division (of a ZZ) by a negative value now triggers an error;
// new error for reporting inexact integer division by a negative value.
//
//
