//   Copyright (c)  2008  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/MatrixForOrdering.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/MatrixArith.H"
#include "CoCoA/MatrixViews.H"
#include "CoCoA/RingZ.H"

#include <iostream>
using std::ostream;
//#include <cstddef>
using std::size_t;
#include <vector>
using std::vector;

namespace CoCoA
{



  matrix NewMatHElim(ConstMatrixView GradM, const vector<size_t>& IndetsToElim);
  matrix NewMatAElim(ConstMatrixView GradM, const vector<size_t>& IndetsToElim);
  matrix NewMat0Elim(ConstMatrixView GradM, const vector<size_t>& IndetsToElim);


  bool IsTermOrdering(ConstMatrixView M)
  {
    if  (!IsZ(BaseRing(M)))
      CoCoA_ERROR("matrix must be ove RingZ()", "IsTermOrdering");
    for (size_t col=0; col<NumCols(M); ++col)
    {
      size_t row=0;
      for (; row<NumRows(M); ++row)
      {
        if ( IsZero(M(row,col)) ) continue;
        if ( sign(M(row,col))<0 ) return false;
        else break;
      }
      if (row==NumRows(M)) return false;
    }
    return true;
  }


/**
    expects a matrix with entries in an ordered ring: very likely only Z (maybe Q?)
    returns a matrix with positive entries which defines an equivalent ordering
*/

  matrix NewPositiveMat(ConstMatrixView M)
  {
    if  (!IsZ(BaseRing(M)))
      CoCoA_ERROR("matrix must be ove RingZ()", "NewPositiveMatrix");
    CoCoA_ASSERT( IsZ(BaseRing(M)) );
    if ( !IsTermOrdering(M) )
      CoCoA_ERROR(ERR::NotTermOrdering, "PositiveMatrix");

    matrix PM(NewDenseMat(M));
    const RingElem& unity(one(BaseRing(PM)));
    for (size_t row=0; row < NumRows(PM); ++row)
      for (size_t col=0; col < NumCols(PM); ++col)
        if (sign(PM(row,col)) < 0)
        {
          size_t PosRow=0;
          while (IsZero(PM(PosRow,col))) ++PosRow; // find positive entry in col
          while (sign(PM(row,col)) < 0) PM->myAddRowMul(row, PosRow, unity);
        }
    return PM;
  }

/**
    expects a matrix with entries in an ordered ring
*/

  bool IsPositiveGrading(ConstMatrixView M, size_t GradingDim)
  {
    if (!IsTermOrdering(M))  return false; // seems excessive but it isn't
    // check there are no null columns
    size_t i;
    for (size_t j=0; j < NumCols(M); ++j)
    {
      for (i=0; i < GradingDim; ++i) if (M(i,j) > 0) break;
      if (i==GradingDim) return false;
    }
    return true;
  }


  matrix NewMatMinimize(ConstMatrixView M)
  {
    if  (!IsZ(BaseRing(M)))
      CoCoA_ERROR("matrix must be ove RingZ()", "NewMatMinimize");
    if ( NumRows(M)<NumCols(M) )
      CoCoA_ERROR(ERR::BadMatrixSize, "MatrixMinimize");
    if ( rank(M)!=NumCols(M) )
      CoCoA_ERROR(ERR::NotFullRank, "MatrixMinimize");

    matrix MM(NewDenseMat(BaseRing(M), NumCols(M), NumCols(M)));
    size_t row=0;
    for (size_t r=0; r<NumRows(M); ++r)
    {
      for (size_t col=0; col<NumCols(M); ++col)
        SetEntry(MM, row, col, M(r, col));
      if (rank(MM) > row) ++row;
      if (row==NumRows(MM)) return MM;
    }
    CoCoA_ERROR(ERR::SERIOUS, "MatrixMinimize: should never get here");
    return MM;
  }


  matrix NewDenseMatRevLex(size_t n)
  {
    matrix RL(NewDenseMat(RingZ(),n,n));
    for (size_t r=0; r<n; ++r)
      SetEntry(RL, r, n-r-1, -1);
    return RL;
  }


  matrix NewMatCompleteOrd(ConstMatrixView M)
  {
    if  (!IsZ(BaseRing(M)))
      CoCoA_ERROR("matrix must be ove RingZ()", "NewMatCompleteOrd");
    matrix MC = NewDenseMat(ConcatVer(M, NewDenseMatRevLex(NumCols(M))));
    return NewMatMinimize(MC);
  }


  matrix NewMatElim(ConstMatrixView GradM,
                       const std::vector<size_t>& IndetsToElim,
		       const bool IsHomog)
  {
    if (NumRows(GradM)==0)
      return NewMat0Elim(GradM,IndetsToElim);
    if (IsHomog)
      return NewMatHElim(GradM,IndetsToElim);
    return  NewMatAElim(GradM,IndetsToElim);
  }


  matrix NewMatElim(size_t NumIndets, const std::vector<size_t>& IndetsToElim)
  {
    matrix ElimRows(NewDenseMat(RingZ(), 2, NumIndets));

    for (size_t i=0; i < IndetsToElim.size(); ++i)
      SetEntry(ElimRows, 0, IndetsToElim[i], 1);
    for (size_t i=0; i < NumIndets; ++i)
      if (ElimRows(0,i) == 0)  SetEntry(ElimRows, 1, i, 1);
    return NewMatCompleteOrd(ElimRows);
  }


  // this should be fixed: it may give a non-term-ordering
  matrix NewMat0Elim(ConstMatrixView GradM, const vector<size_t>& IndetsToElim)
  {
    if  (!IsZ(BaseRing(GradM)))
      CoCoA_ERROR("matrix must be ove RingZ()", "NewMat0Elim");
    matrix ElimRow(NewDenseMat(BaseRing(GradM), 1, NumCols(GradM)));
    for (size_t i=0; i < IndetsToElim.size(); ++i)
      SetEntry(ElimRow, 0, IndetsToElim[i], 1);

    matrix OneRow(NewDenseMat(BaseRing(GradM), 1, NumCols(GradM)));
    for (size_t i=0; i < NumCols(GradM); ++i)
      SetEntry(OneRow, 0, i, 1);
    return NewMatCompleteOrd(ConcatVer(ElimRow,OneRow));
  }

  matrix NewMatHElim(ConstMatrixView GradM, const vector<size_t>& IndetsToElim)
  {
    if  (!IsZ(BaseRing(GradM)))
      CoCoA_ERROR("matrix must be ove RingZ()", "NewMatHElim");
    matrix ElimRow(NewDenseMat(BaseRing(GradM), 1, NumCols(GradM)));
    for (size_t i=0; i < IndetsToElim.size(); ++i)
      SetEntry(ElimRow, 0, IndetsToElim[i], 1);
    return NewMatCompleteOrd(ConcatVer(GradM, ElimRow));
  }

  matrix NewMatAElim(ConstMatrixView GradM, const vector<size_t>& IndetsToElim)
  {
    if  (!IsZ(BaseRing(GradM)))
      CoCoA_ERROR("matrix must be ove RingZ()", "NewMatAElim");
    matrix ElimRow(NewDenseMat(BaseRing(GradM), 1, NumCols(GradM)));

    for (size_t i=0; i < IndetsToElim.size(); ++i)
      SetEntry(ElimRow, 0, IndetsToElim[i], 1);
    return NewMatCompleteOrd(ConcatVer(ElimRow,GradM));
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/MatrixForOrdering.C,v 1.4 2009/12/03 17:40:36 abbott Exp $
// $Log: MatrixForOrdering.C,v $
// Revision 1.4  2009/12/03 17:40:36  abbott
// Added include directives for ZZ.H (as a consequence of removing
// the directive from ring.H).
//
// Revision 1.3  2009/09/22 13:35:55  bigatti
// -- following coding conventions in function names Matrix --> Mat
// -- forced all matrices to be over RingZ
//
// Revision 1.2  2008/05/30 12:44:14  abbott
// Moved "ordering matrices" into their ownn special file.
//
// Revision 1.1  2008/04/21 11:23:11  abbott
// Separated functions dealing with matrices and PPOrderings into a new file.
// Added matrix norms, and completed adjoint.
//
//
