//   Copyright (c)  2005,2008  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/MatrixViews.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/ring.H"


//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;
#include <algorithm>
using std::min;

namespace CoCoA
{

  /***************************************************************************/
  // ZeroMatImpl -- zero matrices
  // Derived from MatrixViewBase rather than ConstMatrixViewBase so that BlockMatrix works well.

  class ZeroMatImpl: public MatrixViewBase
  {
  private:
    friend MatrixView ZeroMat(const ring&, std::size_t, std::size_t); // pseudo-ctor
    ZeroMatImpl(const ring& R, std::size_t nrows, std::size_t ncols);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ZeroMatImpl(const ZeroMatImpl&);            // NEVER DEFINED -- copy ctor disabled
    ZeroMatImpl& operator=(const ZeroMatImpl&); // NEVER DEFINED -- copy ctor disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
  };


  /******************************************************************/
  // Functions for ZeroMatImpl

  ZeroMatImpl::ZeroMatImpl(const ring& R, std::size_t NumRows, std::size_t NumCols):
      MatrixViewBase(),
      myR(R),
      myNumRowsValue(NumRows),
      myNumColsValue(NumCols)
  {}


  const ring& ZeroMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ZeroMatImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ZeroMatImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem ZeroMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return zero(myR);
  }


  // This is exception safe only if "lhs[i]=0;" cannot throw.
  void ZeroMatImpl::myMulByRow(vec& lhs, const vec& v) const
  {
// BUG: should check all elems of v belong to the right ring!
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    for (size_t i=0; i < myNumCols(); ++i)
      lhs[i] = 0;
  }


  // This is exception safe only if "lhs[i]=0;" cannot throw.
  void ZeroMatImpl::myMulByCol(vec& lhs, const vec& v) const
  {
// BUG: should check all elems of v belong to the right ring!
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    for (size_t i=0; i < myNumRows(); ++i)
      lhs[i] = 0;
  }


  bool ZeroMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return true;
  }


  bool ZeroMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return true;
  }


  void ZeroMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    CoCoA_ASSERT(myNumRows() == myNumCols());
    if (myNumRows() == 0)
      d = 1;
    else
      d = 0;
  }



  size_t ZeroMatImpl::myRank() const
  {
    return 0;
  }


  // Do nothing - matrix is already full of zeros!
  void ZeroMatImpl::myAssignZero()
  {}


  bool ZeroMatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return false;
  }


  RefRingElem ZeroMatImpl::myRefEntry(std::size_t /*i*/, std::size_t /*j*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "ZeroMat->myRefEntry(i,j)");
    return RingElem(myR); // never reached, just to keep compiler quiet
  }


  void ZeroMatImpl::mySetEntry(std::size_t /*i*/, std::size_t /*j*/, ConstRefRingElem /*r*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "SetEntry(ZeroMat,i,j,RingElem)");
  }

  void ZeroMatImpl::mySetEntry(std::size_t /*i*/, std::size_t /*j*/, const MachineInteger& /*n*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "SetEntry(ZeroMat,i,j,n)");
  }

  void ZeroMatImpl::mySetEntry(std::size_t /*i*/, std::size_t /*j*/, const ZZ& /*N*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "SetEntry(ZeroMat,i,j,ZZ)");
  }


  MatrixView ZeroMat(const ring& R, std::size_t NumRows, std::size_t NumCols)
  {
    return MatrixView(new ZeroMatImpl(R, NumRows, NumCols));
  }




  /***************************************************************************/
  // IdentityMatImpl -- identity matrices (necessarily square)
  // Derived from MatrixViewBase rather than ConstMatrixViewBase so that BlockMatrix works well.

  class IdentityMatImpl: public MatrixViewBase
  {
  private:
    friend MatrixView IdentityMat(const ring& R, std::size_t dim); // pseudo-ctor
    IdentityMatImpl(const ring& R, std::size_t dim);
    // default dtor is fine
  private: // disable default copy ctor and assignment
    IdentityMatImpl(const IdentityMatImpl&);            // NEVER DEFINED -- copy ctor disabled
    IdentityMatImpl& operator=(const IdentityMatImpl&); // NEVER DEFINED -- copy ctor disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    const ring myR;
    size_t myDim;
  };


  /***************************************************************************/
  // Functions for IdentityMatImpl

  IdentityMatImpl::IdentityMatImpl(const ring& R, size_t dim):
      MatrixViewBase(),
      myR(R),
      myDim(dim)
  {}


  const ring& IdentityMatImpl::myBaseRing() const
  {
    return myR;
  }


  size_t IdentityMatImpl::myNumRows() const
  {
    return myDim;
  }


  size_t IdentityMatImpl::myNumCols() const
  {
    return myDim;
  }


  ConstRefRingElem IdentityMatImpl::myEntry(size_t i, size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i == j) return one(myR);
    return zero(myR);
  }


  //BUG: should use FreeModule elems!
  void IdentityMatImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    CoCoA_ASSERT(myNumRows() == 0 || myBaseRing() == owner(v[0]));
    CoCoA_ASSERT(myNumRows() == 0 || myBaseRing() == owner(lhs[0]));
    lhs = v;
  }


  void IdentityMatImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
//BUG???    CoCoA_ASSERT(myBaseRing() == AmbientRing(v));
    lhs = v;
  }


  bool IdentityMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return false;
  }


  bool IdentityMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return false;
  }


  void IdentityMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    d = 1; // NB correct even for 0x0 matrix.
  }


  size_t IdentityMatImpl::myRank() const
  {
    return myNumRows();
  }


  // Always gives error!
  void IdentityMatImpl::myAssignZero()
  {
//???    if (myDim == 0) return;
    CoCoA_ERROR(ERR::ConstMatEntry, "IdentityMat->myAssignZero()");
  }


  bool IdentityMatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return false;
  }


  RefRingElem IdentityMatImpl::myRefEntry(std::size_t /*i*/, std::size_t /*j*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "IdentityMat->myRefEntry(i,j)");
    return RingElem(myR); // never reached, just to keep compiler quiet
  }


  void IdentityMatImpl::mySetEntry(std::size_t /*i*/, std::size_t /*j*/, ConstRefRingElem /*r*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "SetEntry(IdentityMat,i,j,RingElem)");
  }

  void IdentityMatImpl::mySetEntry(std::size_t /*i*/, std::size_t /*j*/, const MachineInteger& /*n*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "SetEntry(IdentityMat,i,j,n)");
  }

  void IdentityMatImpl::mySetEntry(std::size_t /*i*/, std::size_t /*j*/, const ZZ& /*N*/)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "SetEntry(IdentityMat,i,j,ZZ)");
  }


  MatrixView IdentityMat(const ring& R, std::size_t dim)
  {
    return MatrixView(new IdentityMatImpl(R, dim));
  }



  /***************************************************************************/
  // ConstTrasposeMatImpl -- a transposed view of a const matrix

  class ConstTrasposeMatImpl: public ConstMatrixViewBase
  {
  private:
    friend ConstMatrixView transpose(ConstMatrixView M); // pseudo-ctor
    ConstTrasposeMatImpl(ConstMatrixView M);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ConstTrasposeMatImpl(const ConstTrasposeMatImpl&);            // NEVER DEFINED -- copy ctor disabled
    ConstTrasposeMatImpl& operator=(const ConstTrasposeMatImpl&); // NEVER DEFINED -- copy ctor disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

  private: // data members
    ConstMatrixView myM;
  };


  /***************************************************************************/
  // Functions for ConstTrasposeMatImpl -- all are simple one-liners

  ConstTrasposeMatImpl::ConstTrasposeMatImpl(ConstMatrixView M):
      myM(M)
  {}


  const ring& ConstTrasposeMatImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  std::size_t ConstTrasposeMatImpl::myNumRows() const
  {
    return NumCols(myM);
  }


  std::size_t ConstTrasposeMatImpl::myNumCols() const
  {
    return NumRows(myM);
  }


  ConstRefRingElem ConstTrasposeMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    return myM(j, i);
  }


  void ConstTrasposeMatImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    myM->myMulByCol(lhs, v);
  }


  void ConstTrasposeMatImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    myM->myMulByRow(lhs, v);
  }



  bool ConstTrasposeMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return myM->myIsZeroCol(i);
  }


  bool ConstTrasposeMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return myM->myIsZeroRow(j);
  }


  void ConstTrasposeMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    CoCoA_ASSERT(myNumRows() == myNumCols());
    myM->myDet(d);
  }


  size_t ConstTrasposeMatImpl::myRank() const
  {
    return myM->myRank();
  }


  ConstMatrixView transpose(ConstMatrixView M)
  {
    return ConstMatrixView(new ConstTrasposeMatImpl(M));
  }


  /***************************************************************************/
  // TrasposeMatImpl -- a transposed view of a (non-const) matrix

  class TrasposeMatImpl: public MatrixViewBase
  {
  private:
    friend MatrixView transpose(MatrixView M); // pseudo-ctor
    TrasposeMatImpl(MatrixView M);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    TrasposeMatImpl(const TrasposeMatImpl&);            // NEVER DEFINED -- copy ctor disabled
    TrasposeMatImpl& operator=(const TrasposeMatImpl&); // NEVER DEFINED -- copy ctor disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

    // non-const member fns
    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    MatrixView myM;
  };


  /***************************************************************************/
  // Functions for TrasposeMatImpl -- all are simple one-liners

  TrasposeMatImpl::TrasposeMatImpl(MatrixView M):
      myM(M)
  {}


  const ring& TrasposeMatImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  std::size_t TrasposeMatImpl::myNumRows() const
  {
    return NumCols(myM);
  }


  std::size_t TrasposeMatImpl::myNumCols() const
  {
    return NumRows(myM);
  }


  ConstRefRingElem TrasposeMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    return myM(j, i);
  }


  void TrasposeMatImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    myM->myMulByCol(lhs, v);
  }


  void TrasposeMatImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    myM->myMulByRow(lhs, v);
  }



  bool TrasposeMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return myM->myIsZeroCol(i);
  }


  bool TrasposeMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return myM->myIsZeroRow(j);
  }


  void TrasposeMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    CoCoA_ASSERT(myNumRows() == myNumCols());
    myM->myDet(d);
  }


  size_t TrasposeMatImpl::myRank() const
  {
    return myM->myRank();
  }


  void TrasposeMatImpl::myAssignZero()
  {
    myM->myAssignZero();
  }


  bool TrasposeMatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myM->myIsWritable(j,i);
  }


  RefRingElem TrasposeMatImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    return myM->myRefEntry(j,i);
  }


  void TrasposeMatImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myM->mySetEntry(j,i,r);
  }


  void TrasposeMatImpl::mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myM->mySetEntry(j,i,n);
  }


  void TrasposeMatImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myM->mySetEntry(j,i,N);
  }


  MatrixView transpose(MatrixView M)
  {
    return MatrixView(new TrasposeMatImpl(M));
  }


  /***************************************************************************/
  // SubmatImpl -- see only certain rows/cols of a matrix

  class ConstSubmatImpl: public ConstMatrixViewBase
  {
  private:
    friend ConstMatrixView submat(ConstMatrixView M, const std::vector<std::size_t>& RowIndices, const std::vector<std::size_t>& ColIndices); // pseudo-ctor
    ConstSubmatImpl(ConstMatrixView M, const std::vector<std::size_t>& rows, const std::vector<std::size_t>& cols);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ConstSubmatImpl(const ConstSubmatImpl&);            // NEVER DEFINED -- copy ctor disabled
    ConstSubmatImpl& operator=(const ConstSubmatImpl&); // NEVER DEFINED -- assignment disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use DEFAULT definitions of the following six functions.
//     virtual void myMulByRow(vec& lhs, const vec& v) const;
//     virtual void myMulByCol(vec& lhs, const vec& v) const;
//     virtual bool myIsZeroRow(std::size_t i) const;
//     virtual bool myIsZeroCol(std::size_t j) const;
//     virtual void myDet(RefRingElem d) const;
//     virtual std::size_t myRank() const;

  private: // data members
    ConstMatrixView myM;
    std::vector<std::size_t> myRowTable;
    std::vector<std::size_t> myColTable;
  };


  /***************************************************************************/
  // Functions for ConstSubmatImpl

  ConstSubmatImpl::ConstSubmatImpl(ConstMatrixView M, const std::vector<std::size_t>& rows, const std::vector<std::size_t>& cols):
      myM(M),
      myRowTable(rows),
      myColTable(cols)
  {
    // Args already checked by pseudo-ctor submat, the only
    // function which calls this ctor
  }


  const ring& ConstSubmatImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  std::size_t ConstSubmatImpl::myNumRows() const
  {
    return myRowTable.size();
  }


  std::size_t ConstSubmatImpl::myNumCols() const
  {
    return myColTable.size();
  }


  ConstRefRingElem ConstSubmatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myM(myRowTable[i], myColTable[j]);
  }


  ConstMatrixView submat(ConstMatrixView M, const std::vector<std::size_t>& rows, const vector<std::size_t>& cols)
  {
    // Check all row indices are within range.
    const size_t r = NumRows(M);
    for (size_t i=0; i < rows.size(); ++i)
      if (rows[i] >= r)
        CoCoA_ERROR(ERR::BadRowIndex, "submat(M,rows,cols)");

    // Check all col indices are within range.
    const size_t c = NumCols(M);
    for (size_t j=0; j < cols.size(); ++j)
      if (cols[j] >= c)
        CoCoA_ERROR(ERR::BadColIndex, "submat(M,rows,cols)");

    // Check all row indices are distinct.
    for (size_t i=0; i < rows.size(); ++i)
      for (size_t j=i+1; j < rows.size(); ++j)
        if (rows[i] == rows[j])
          CoCoA_ERROR(ERR::BadRowIndex, "submat(M,rows,cols):  repeated row indices");

    // Check all col indices are distinct.
    for (size_t i=0; i < cols.size(); ++i)
      for (size_t j=i+1; j < cols.size(); ++j)
        if (cols[i] == cols[j])
          CoCoA_ERROR(ERR::BadColIndex, "submat(M,rows,cols):  repeated col indices");

    return ConstMatrixView(new ConstSubmatImpl(M, rows, cols));
  }


  /***************************************************************************/
  // SubmatImpl -- see only certain rows/cols of a matrix

  class SubmatImpl: public MatrixViewBase
  {
  private:
    friend MatrixView submat(MatrixView M, const std::vector<std::size_t>& RowIndices, const std::vector<std::size_t>& ColIndices); // pseudo-ctor
    SubmatImpl(MatrixView M, const std::vector<std::size_t>& rows, const std::vector<std::size_t>& cols);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    SubmatImpl(const SubmatImpl&);            // NEVER DEFINED -- copy ctor disabled
    SubmatImpl& operator=(const SubmatImpl&); // NEVER DEFINED -- assignment disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use DEFAULT definitions of the following six functions.
//     virtual void myMulByRow(vec& lhs, const vec& v) const;
//     virtual void myMulByCol(vec& lhs, const vec& v) const;
//     virtual bool myIsZeroRow(std::size_t i) const;
//     virtual bool myIsZeroCol(std::size_t j) const;
//     virtual void myDet(RefRingElem d) const;
//     virtual std::size_t myRank() const;

    // non-const member fns
    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    MatrixView myM;
    std::vector<std::size_t> myRowTable;
    std::vector<std::size_t> myColTable;
  };


  /***************************************************************************/
  // Functions for SubmatImpl


  SubmatImpl::SubmatImpl(MatrixView M, const std::vector<std::size_t>& rows, const std::vector<std::size_t>& cols):
      myM(M),
      myRowTable(rows),
      myColTable(cols)
  {
    // Args already checked by pseudo-ctor submat, the only
    // function which calls this ctor
  }


  const ring& SubmatImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  std::size_t SubmatImpl::myNumRows() const
  {
    return myRowTable.size();
  }


  std::size_t SubmatImpl::myNumCols() const
  {
    return myColTable.size();
  }


  ConstRefRingElem SubmatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myM(myRowTable[i], myColTable[j]);
  }


  void SubmatImpl::myAssignZero()
  {
    for (size_t i=0; i < myNumRows(); ++i)
      for (size_t j=0; j < myNumCols(); ++j)
        mySetEntry(i,j,0);
  }


  bool SubmatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myM->myIsWritable(myRowTable[i], myColTable[j]);
  }


  RefRingElem SubmatImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i, j));
    return myM->myRefEntry(myRowTable[i], myColTable[j]);
  }


  void SubmatImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(owner(r) == myBaseRing());
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i, j));
    myM->mySetEntry(myRowTable[i], myColTable[j], r);
  }

  void SubmatImpl::mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i, j));
    myM->mySetEntry(myRowTable[i], myColTable[j], n);
  }

  void SubmatImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i, j));
    myM->mySetEntry(myRowTable[i], myColTable[j], N);
  }


  MatrixView submat(MatrixView M, const std::vector<std::size_t>& rows, const vector<std::size_t>& cols)
  {
    // Check all row indices are within range.
    const size_t r = NumRows(M);
    for (size_t i=0; i < rows.size(); ++i)
      if (rows[i] >= r)
        CoCoA_ERROR(ERR::BadRowIndex, "submat(M,rows,cols)");

    // Check all col indices are within range.
    const size_t c = NumCols(M);
    for (size_t j=0; j < cols.size(); ++j)
      if (cols[j] >= c)
        CoCoA_ERROR(ERR::BadColIndex, "submat(M,rows,cols)");

    // Check all row indices are distinct.
    for (size_t i=0; i < rows.size(); ++i)
      for (size_t j=i+1; j < rows.size(); ++j)
        if (rows[i] == rows[j])
          CoCoA_ERROR(ERR::BadRowIndex, "submat(M,rows,cols):  repeated row indices");

    // Check all col indices are distinct.
    for (size_t i=0; i < cols.size(); ++i)
      for (size_t j=i+1; j < cols.size(); ++j)
        if (cols[i] == cols[j])
          CoCoA_ERROR(ERR::BadColIndex, "submat(M,rows,cols):  repeated col indices");

    return MatrixView(new SubmatImpl(M, rows, cols));
  }


  /****************************************************************************/

  class ConstColMatImpl: public ConstMatrixViewBase
  {
  private:
    friend ConstMatrixView ColMat(const std::vector<RingElem>& v); // pseudo-ctor
    ConstColMatImpl(const std::vector<RingElem>& v);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ConstColMatImpl(const ConstColMatImpl&);            // NEVER DEFINED -- disable copy ctor
    ConstColMatImpl& operator=(const ConstColMatImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    const std::vector<RingElem>& myV; // NB this is a reference!!!
  };


  ConstColMatImpl::ConstColMatImpl(const std::vector<RingElem>& v):
      myR(owner(v[0])),
      myNumRowsValue(v.size()),
      myV(v)
  {
    CoCoA_ASSERT(!v.empty());
#ifdef CoCoA_DEBUG
    ring R = owner(v[0]);
    for (size_t i=1; i < v.size(); ++i)
      CoCoA_ASSERT(owner(v[i]) == R);
#endif
  }


  const ring& ConstColMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ConstColMatImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ConstColMatImpl::myNumCols() const
  {
    return 1;
  }


  ConstRefRingElem ConstColMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myV[i];
  }


  bool ConstColMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return IsZero(myV[i]);
  }

  bool ConstColMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    for (size_t i=0; i < myNumRows(); ++i)
      if (!IsZero(myV[i])) return false;
    return true;
  }


  void ConstColMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(myNumRows() == 1);
    d = myV[0];
  }


  std::size_t ConstColMatImpl::myRank() const
  {
    if (myIsZeroCol(0)) return 0;
    return 1;
  }


  ConstMatrixView ColMat(const std::vector<RingElem>& v)
  {
    if (v.empty())
      CoCoA_ERROR("empty vector of ring elems", "ColMat(v)");
    ring R = owner(v[0]);
    for (size_t i=1; i < v.size(); ++i)
      if (owner(v[i]) != R)
        CoCoA_ERROR(ERR::MixedRings, "ColMat(v)");
    return ConstMatrixView(new ConstColMatImpl(v));
  }

  /****************************************************************************/

  class ColMatImpl: public MatrixViewBase
  {
  private:
    friend MatrixView ColMat(std::vector<RingElem>& v); // pseudo-ctor
    ColMatImpl(std::vector<RingElem>& v);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ColMatImpl(const ColMatImpl&);            // NEVER DEFINED -- disable copy ctor
    ColMatImpl& operator=(const ColMatImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

    // non-const member fns
    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::vector<RingElem>& myV; // NB this is a reference!!!
  };


  ColMatImpl::ColMatImpl(std::vector<RingElem>& v):
      myR(owner(v[0])),
      myNumRowsValue(v.size()),
      myV(v)
  {
    CoCoA_ASSERT(!v.empty());
#ifdef CoCoA_DEBUG
    ring R = owner(v[0]);
    for (size_t i=1; i < v.size(); ++i)
      CoCoA_ASSERT(owner(v[i]) == R);
#endif
  }


  const ring& ColMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ColMatImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ColMatImpl::myNumCols() const
  {
    return 1;
  }


  ConstRefRingElem ColMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myV[i];
  }


  bool ColMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return IsZero(myV[i]);
  }

  bool ColMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    for (size_t i=0; i < myNumRows(); ++i)
      if (!IsZero(myV[i])) return false;
    return true;
  }


  void ColMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(myNumRows() == 1);
    d = myV[0];
  }


  std::size_t ColMatImpl::myRank() const
  {
    if (myIsZeroCol(0)) return 0;
    return 1;
  }


  void ColMatImpl::myAssignZero()
  {
    for (size_t i=0; i < myNumRows(); ++i)
      myV[i] = 0;
  }


  bool ColMatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return true;
  }

  RefRingElem ColMatImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myV[i];
  }


  void ColMatImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(owner(r) == myR);
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    myV[i] = r;
  }

  void ColMatImpl::mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    myV[i] = n;
  }

  void ColMatImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    myV[i] = N;
  }


  MatrixView ColMat(std::vector<RingElem>& v)
  {
    if (v.empty())
      CoCoA_ERROR("empty vector of ring elems", "ColMat(v)");
    ring R = owner(v[0]);
    for (size_t i=1; i < v.size(); ++i)
      if (owner(v[i]) != R)
        CoCoA_ERROR(ERR::MixedRings, "ColMat(v)");
    return MatrixView(new ColMatImpl(v));
  }


  /***************************************************************************/
  // Very simple definitions... but may not be very efficient.

  MatrixView RowMat(std::vector<RingElem>& v)
  {
    return transpose(ColMat(v));
  }

  ConstMatrixView RowMat(const std::vector<RingElem>& v)
  {
    return transpose(ColMat(v));
  }



  /***************************************************************************/
  // ConstDiagMatImpl -- diagonal matrices implemented in a pretty obvious manner

  class ConstDiagMatImpl: public ConstMatrixViewBase
  {
  public:
    ConstDiagMatImpl(const ring& R, const std::vector<RingElem>& DiagEntries);
    // Default dtor is OK

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    const std::size_t myDim;
    const std::vector<RingElem>& myDiagEntries; // NB reference to an external vector!!!
  };

  //----------------------------------------------------------------------

  ConstDiagMatImpl::ConstDiagMatImpl(const ring& R, const std::vector<RingElem>& DiagEntries):
      myR(R),
      myDim(DiagEntries.size()),
      myDiagEntries(DiagEntries)
  {
#ifdef CoCoA_DEBUG
    for (size_t i=0; i < myDim; ++i)
      CoCoA_ASSERT(owner(myDiagEntries[i]) == myR);
#endif
  }


  const ring& ConstDiagMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ConstDiagMatImpl::myNumRows() const
  {
    return myDim;
  }


  std::size_t ConstDiagMatImpl::myNumCols() const
  {
    return myDim;
  }


  ConstRefRingElem ConstDiagMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i == j)
      return myDiagEntries[i];
    else
      return zero(myR);
  }


  void ConstDiagMatImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    //BUG: Should also check owner of elements of v and lhs.
    vector<RingElem> ans(myDim, zero(myR));
    for (size_t i=0; i < myDim; ++i)
      ans[i] = v[i]*myDiagEntries[i];
    for (size_t i=0; i < myDim; ++i)
      swap(lhs[i], ans[i]);
//???    std::swap(lhs, ans); //??? does this clobber iterators on lhs???
  }


  void ConstDiagMatImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    //BUG: Should also check owner of elements of v and lhs.
    vector<RingElem> ans(myDim, zero(myR));
    for (size_t i=0; i < myDim; ++i)
      ans[i] = v[i]*myDiagEntries[i];
    for (size_t i=0; i < myDim; ++i)
      swap(lhs[i], ans[i]);
//???    std::swap(lhs, ans); //??? does this clobber iterators on lhs???
  }


  bool ConstDiagMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return IsZero(myDiagEntries[i]);
  }


  bool ConstDiagMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return IsZero(myDiagEntries[j]);
  }


  void ConstDiagMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myR);
    RingElem ans = one(myR);
    for (size_t i=0; i < myDim; ++i)
      ans *= myDiagEntries[i];
    swap(d, ans);
  }


  // ??? better to use "count"?
  size_t ConstDiagMatImpl::myRank() const
  {
    CoCoA_ASSERT(IsIntegralDomain(myR));
    size_t rk=0;
    for (size_t i = 0; i < myDim; ++i)
      if (!IsZero(myDiagEntries[i]))
        ++rk;
    return rk;
  }


  ConstMatrixView DiagMat(const std::vector<RingElem>& DiagEntries)
  {
    if (DiagEntries.empty())
      CoCoA_ERROR("Empty vector", "DiagMat(v)");
    ring R = owner(DiagEntries[0]);
    for (size_t i=1; i < DiagEntries.size(); ++i)
      if (owner(DiagEntries[i]) != R)
        CoCoA_ERROR(ERR::MixedRings, "DiagMat(v)");
    return ConstMatrixView(new ConstDiagMatImpl(R, DiagEntries));
  }


  /***************************************************************************/
  // DiagMatImpl -- diagonal matrices implemented in a pretty obvious manner

  class DiagMatImpl: public MatrixViewBase
  {
  public:
    DiagMatImpl(const ring& R, std::vector<RingElem>& DiagEntries);
    // Default dtor is OK

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    const ring myR;
    const std::size_t myDim;
    std::vector<RingElem>& myDiagEntries; // NB reference to an external vector!!!
  };

  //----------------------------------------------------------------------

  DiagMatImpl::DiagMatImpl(const ring& R, std::vector<RingElem>& DiagEntries):
      myR(R),
      myDim(DiagEntries.size()),
      myDiagEntries(DiagEntries)
  {
#ifdef CoCoA_DEBUG
    for (size_t i=0; i < myDim; ++i)
      CoCoA_ASSERT(owner(myDiagEntries[i]) == myR);
#endif
  }


  const ring& DiagMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t DiagMatImpl::myNumRows() const
  {
    return myDim;
  }


  std::size_t DiagMatImpl::myNumCols() const
  {
    return myDim;
  }


  ConstRefRingElem DiagMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i == j)
      return myDiagEntries[i];
    else
      return zero(myR);
  }


  void DiagMatImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    // Should also check owner of elements of v and lhs.
    vector<RingElem> ans(myDim, zero(myR));
    for (size_t i=0; i < myDim; ++i)
      ans[i] = v[i]*myDiagEntries[i];
    for (size_t i=0; i < myDim; ++i)
      swap(lhs[i], ans[i]);
//???    std::swap(lhs, ans); //??? does this clobber iterators on lhs???
  }


  void DiagMatImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    // Should also check owner of elements of v and lhs.
    vector<RingElem> ans(myDim, zero(myR));
    for (size_t i=0; i < myDim; ++i)
      ans[i] = v[i]*myDiagEntries[i];
    for (size_t i=0; i < myDim; ++i)
      swap(lhs[i], ans[i]);
//???    std::swap(lhs, ans); //??? would this clobber iterators on lhs???
  }


  bool DiagMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return IsZero(myDiagEntries[i]);
  }


  bool DiagMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return IsZero(myDiagEntries[j]);
  }


  void DiagMatImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myR);
    RingElem ans = one(myR);
    for (size_t i=0; i < myDim; ++i)
      ans *= myDiagEntries[i];
    swap(d, ans);
  }


  // ??? better to use "std::count"???
  size_t DiagMatImpl::myRank() const
  {
    CoCoA_ASSERT(IsIntegralDomain(myR));
    size_t rk=0;
    for (size_t i = 0; i < myDim; ++i)
      if (!IsZero(myDiagEntries[i]))
        ++rk;
    return rk;
  }


  // Exception clean only if the call to fill cannot throw
  void DiagMatImpl::myAssignZero()
  {
    fill(myDiagEntries.begin(), myDiagEntries.end(), zero(myR));
  }


  bool DiagMatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    return i==j;
  }

  RefRingElem DiagMatImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    return myDiagEntries[i];
  }


  void DiagMatImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myDiagEntries[i] = r;
  }


  void DiagMatImpl::mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myDiagEntries[i] = n;
  }


  void DiagMatImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myDiagEntries[i] = N;
  }


  MatrixView DiagMat(std::vector<RingElem>& DiagEntries)
  {
    if (DiagEntries.empty())
      CoCoA_ERROR("Empty vector", "DiagMat(v)");
    ring R = owner(DiagEntries[0]);
    for (size_t i=1; i < DiagEntries.size(); ++i)
      if (owner(DiagEntries[i]) != R)
        CoCoA_ERROR(ERR::MixedRings, "DiagMat(v)");
    return MatrixView(new DiagMatImpl(R, DiagEntries));
  }

  /***************************************************************************/

  class ConstBlockMatImpl: public ConstMatrixViewBase
  {
  private:
    friend ConstMatrixView BlockMat(ConstMatrixView A, ConstMatrixView B, ConstMatrixView C, ConstMatrixView D); // pseudo-ctor
    ConstBlockMatImpl(ConstMatrixView A, ConstMatrixView B, ConstMatrixView C, ConstMatrixView D);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ConstBlockMatImpl(const ConstBlockMatImpl&);            // NEVER DEFINED -- disable copy ctor
    ConstBlockMatImpl& operator=(const ConstBlockMatImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myDet(RefRingElem d) const;
//    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
    ConstMatrixView myA, myB, myC, myD;
  };


  /***************************************************************************/
  // Functions for ConstBlockMatImpl -- a matrix of the form  (A B)
  //                                                             (C D)
  // This implementation requires that the submatrices correspond
  // to two straight cuts -- several changes would be needed to handle
  // a more general assembly of submatrices.

  ConstBlockMatImpl::ConstBlockMatImpl(ConstMatrixView A, ConstMatrixView B, ConstMatrixView C, ConstMatrixView D):
      myR(BaseRing(A)),
      myNumRowsValue(NumRows(A)+NumRows(C)),
      myNumColsValue(NumCols(A)+NumCols(B)),
      myA(A),
      myB(B),
      myC(C),
      myD(D)
  {
    // Args are checked for compatibility by pseudo-ctor BlockMat, the
    // only function which calls this ctor.
  }


  const ring& ConstBlockMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ConstBlockMatImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ConstBlockMatImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem ConstBlockMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i < NumRows(myA) && j < NumCols(myA))
      return myA(i, j);
    if (i < NumRows(myB))
      return myB(i, j-NumCols(myA));
    if (j < NumCols(myC))
      return myC(i-NumRows(myA), j);
    return myD(i-NumRows(myA), j-NumCols(myA));
  }


  bool ConstBlockMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    if (i < NumRows(myA))
      return myA->myIsZeroRow(i) && myB->myIsZeroRow(i);
    return myC->myIsZeroRow(i-NumRows(myA)) && myD->myIsZeroRow(i-NumRows(myA));
  }


  bool ConstBlockMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    if (j < NumCols(myA))
      return myA->myIsZeroCol(j) && myC->myIsZeroCol(j);
    return myB->myIsZeroCol(j-NumCols(myA)) && myD->myIsZeroCol(j-NumCols(myA));
  }


  ConstMatrixView BlockMat(ConstMatrixView A, ConstMatrixView B, ConstMatrixView C, ConstMatrixView D)
  {
    {
      // Check that the rings are all the same
      const ring R = BaseRing(A);
      if (BaseRing(B) != R || BaseRing(C) != R || BaseRing(D) != R)
        CoCoA_ERROR(ERR::MixedRings, "BlockMat(A,B,C,D)");
    }
    // Check that the matrix sizes are suitably matched.
    if (NumRows(A) != NumRows(B) ||
        NumRows(C) != NumRows(D) ||
        NumCols(A) != NumCols(C) ||
        NumCols(B) != NumCols(D))
      CoCoA_ERROR(ERR::BadMatrixSize, "BlockMat(A,B,C,D)");
    return ConstMatrixView(new ConstBlockMatImpl(A, B, C, D));
  }


/***************************************************************************/

  class BlockMatImpl: public MatrixViewBase
  {
  private:
    friend MatrixView BlockMat(MatrixView A, MatrixView B, MatrixView C, MatrixView D); // pseudo-ctor
    BlockMatImpl(MatrixView A, MatrixView B, MatrixView C, MatrixView D);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    BlockMatImpl(const BlockMatImpl&);            // NEVER DEFINED -- disable copy ctor
    BlockMatImpl& operator=(const BlockMatImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myDet(RefRingElem d) const;
//    virtual std::size_t myRank() const;

    // Non-const member functions
    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
    MatrixView myA, myB, myC, myD;
  };


  /***************************************************************************/
  // Functions for BlockMatImpl -- a matrix of the form  (A B)
  //                                                        (C D)
  // This implementation requires that the submatrices correspond
  // to two straight cuts -- several changes would be needed to handle
  // a more general assembly of submatrices.

  BlockMatImpl::BlockMatImpl(MatrixView A, MatrixView B, MatrixView C, MatrixView D):
      myR(BaseRing(A)),
      myNumRowsValue(NumRows(A)+NumRows(C)),
      myNumColsValue(NumCols(A)+NumCols(B)),
      myA(A),
      myB(B),
      myC(C),
      myD(D)
  {
    // Args are checked for compatibility by pseudo-ctor BlockMat, the
    // only function which calls this ctor.
  }


  const ring& BlockMatImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t BlockMatImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t BlockMatImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem BlockMatImpl::myEntry(std::size_t i, std::size_t j) const
  {
// This one-liner doesn't work:    return myRefEntry(i,j);
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i < NumRows(myA) && j < NumCols(myA))
      return myA(i, j);
    if (i < NumRows(myB))
      return myB(i, j-NumCols(myA));
    if (j < NumCols(myC))
      return myC(i-NumRows(myA), j);
    return myD(i-NumRows(myA), j-NumCols(myA));
  }


  bool BlockMatImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    if (i < NumRows(myA))
      return myA->myIsZeroRow(i) && myB->myIsZeroRow(i);
    return myC->myIsZeroRow(i-NumRows(myA)) && myD->myIsZeroRow(i-NumRows(myA));
  }


  bool BlockMatImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    if (j < NumCols(myA))
      return myA->myIsZeroCol(j) && myC->myIsZeroCol(j);
    return myB->myIsZeroCol(j-NumCols(myA)) && myD->myIsZeroCol(j-NumCols(myA));
  }


  bool BlockMatImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (j >= NumCols(myA))
    {
      if (i >= NumRows(myB))
        return myD->myIsWritable(i-NumRows(myA), j-NumCols(myB));
      else
        return myB->myIsWritable(i, j-NumCols(myA));
    }
    if (i >= NumRows(myA))
      return myC->myIsWritable(i-NumRows(myA), j);
    else
      return myA->myIsWritable(i, j);
  }


  void BlockMatImpl::myAssignZero()
  {
    myA->myAssignZero();
    myB->myAssignZero();
    myC->myAssignZero();
    myD->myAssignZero();
  }


  RefRingElem BlockMatImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i < NumRows(myA) && j < NumCols(myA))
      return myA->myRefEntry(i, j);
    if (i < NumRows(myB))
      return myB->myRefEntry(i, j-NumCols(myA));
    if (j < NumCols(myC))
      return myC->myRefEntry(i-NumRows(myA), j);
    return myD->myRefEntry(i-NumRows(myA), j-NumCols(myA));
  }


  void BlockMatImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(myR == owner(r));
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myRefEntry(i,j) = r;
  }

  void BlockMatImpl::mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myRefEntry(i,j) = n;
  }

  void BlockMatImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myRefEntry(i,j) = N;
  }


  MatrixView BlockMat(MatrixView A, MatrixView B, MatrixView C, MatrixView D)
  {
    {
      // Check that the rings are all the same
      const ring R = BaseRing(A);
      if (BaseRing(B) != R || BaseRing(C) != R || BaseRing(D) != R)
        CoCoA_ERROR(ERR::MixedRings, "BlockMat(A,B,C,D)");
    }
    // Check that the matrix sizes are suitably matched.
    if (NumRows(A) != NumRows(B) ||
        NumRows(C) != NumRows(D) ||
        NumCols(A) != NumCols(C) ||
        NumCols(B) != NumCols(D))
      CoCoA_ERROR(ERR::BadMatrixSize, "BlockMat(A,B,C,D)");
    return MatrixView(new BlockMatImpl(A, B, C, D));
  }


  /****************************************************************************/

  class ConstConcatVerImpl: public ConstMatrixViewBase
  {
  private:
    friend ConstMatrixView ConcatVer(ConstMatrixView A, ConstMatrixView B); // pseudo-ctor
    ConstConcatVerImpl(ConstMatrixView A, ConstMatrixView B);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ConstConcatVerImpl(const ConstConcatVerImpl&);            // NEVER DEFINED -- disable copy ctor
    ConstConcatVerImpl& operator=(const ConstConcatVerImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myDet(RefRingElem d) const;
//    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
    ConstMatrixView myA, myB;
  };



  ConstConcatVerImpl::ConstConcatVerImpl(ConstMatrixView A, ConstMatrixView B):
      myR(BaseRing(A)),
      myNumRowsValue(NumRows(A)+NumRows(B)),
      myNumColsValue(NumCols(A)),
      myA(A),
      myB(B)
  {
    // Args are checked for compatibility by ConcatVer, the
    // only function which calls this ctor.
  }


  const ring& ConstConcatVerImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ConstConcatVerImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ConstConcatVerImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem ConstConcatVerImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i < NumRows(myA))
      return myA(i, j);
    return myB(i-NumRows(myA), j);
  }


  bool ConstConcatVerImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    if (i < NumRows(myA))
      return myA->myIsZeroRow(i);
    return myB->myIsZeroRow(i-NumRows(myA));
  }


  bool ConstConcatVerImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return myA->myIsZeroCol(j) && myB->myIsZeroCol(j);
  }


  ConstMatrixView ConcatVer(ConstMatrixView A, ConstMatrixView B)
  {
    if (BaseRing(A) != BaseRing(B))
      CoCoA_ERROR(ERR::MixedRings, "ConcatVer(A,B)");
    if (NumCols(A) != NumCols(B))
      CoCoA_ERROR(ERR::BadMatrixSize, "ConcatVer(A,B)");

    return ConstMatrixView(new ConstConcatVerImpl(A, B));
  }


  /***************************************************************************/

  class ConcatVerImpl: public MatrixViewBase
  {
  private:
    friend MatrixView ConcatVer(MatrixView A, MatrixView B); // pseudo-ctor
    ConcatVerImpl(MatrixView A, MatrixView B);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ConcatVerImpl(const ConcatVerImpl&);            // NEVER DEFINED -- disable copy ctor
    ConcatVerImpl& operator=(const ConcatVerImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const;
    virtual bool myIsZeroCol(std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myDet(RefRingElem d) const;
//    virtual std::size_t myRank() const;

    // Non-const member functions
    virtual void myAssignZero();
    virtual bool myIsWritable(std::size_t i, std::size_t j);
    virtual RefRingElem myRefEntry(std::size_t i, std::size_t j);
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
    MatrixView myA, myB;
  };


  /***************************************************************************/
  // Functions for ConcatVerImpl -- a matrix of the form  (A)
  //                                                      (B)
  // This implementation requires that the submatrices correspond
  // to two straight cuts -- several changes would be needed to handle
  // a more general assembly of submatrices.

  ConcatVerImpl::ConcatVerImpl(MatrixView A, MatrixView B):
      myR(BaseRing(A)),
      myNumRowsValue(NumRows(A)+NumRows(B)),
      myNumColsValue(NumCols(A)),
      myA(A),
      myB(B)
  {
    // Args are checked for compatibility by ConcatVer, the
    // only function which calls this ctor.
  }


  const ring& ConcatVerImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ConcatVerImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ConcatVerImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem ConcatVerImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i < NumRows(myA))
      return myA(i, j);
    return myB(i-NumRows(myA), j);
  }


  bool ConcatVerImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    if (i < NumRows(myA))
      return myA->myIsZeroRow(i);
    return myB->myIsZeroRow(i-NumRows(myA));
  }


  bool ConcatVerImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return myA->myIsZeroCol(j) && myB->myIsZeroCol(j);
  }


  bool ConcatVerImpl::myIsWritable(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i >= NumRows(myA))
      return myB->myIsWritable(i-NumRows(myA), j);
    else
      return myA->myIsWritable(i, j);
  }


  void ConcatVerImpl::myAssignZero()
  {
    myA->myAssignZero();
    myB->myAssignZero();
  }


  RefRingElem ConcatVerImpl::myRefEntry(std::size_t i, std::size_t j)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    if (i >= NumRows(myA))
      return myB->myRefEntry(i-NumRows(myA), j);
    else
      return myA->myRefEntry(i, j);
  }


  void ConcatVerImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(myR == owner(r));
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myRefEntry(i,j) = r;
  }

  void ConcatVerImpl::mySetEntry(std::size_t i, std::size_t j, const MachineInteger& n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myRefEntry(i,j) = n;
  }

  void ConcatVerImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    CoCoA_ASSERT(myIsWritable(i,j));
    myRefEntry(i,j) = N;
  }


  MatrixView ConcatVer(MatrixView A, MatrixView B)
  {
    {
      // Check that the base rings are the same.
      if (BaseRing(A) != BaseRing(B))
        CoCoA_ERROR(ERR::MixedRings, "ConcatVer(A,B)");
    }
    if (NumCols(A) != NumCols(B))
      CoCoA_ERROR(ERR::BadMatrixSize, "ConcatVer(A,B)");
    return MatrixView(new ConcatVerImpl(A, B));
  }


  /***************************************************************************/
  // ConcatHor implemented in terms of ConcatVer.
  // Very simple implementations... but may not be so efficient.

  ConstMatrixView ConcatHor(ConstMatrixView M1, ConstMatrixView M2)
  {
    if (BaseRing(M1) != BaseRing(M2))
      CoCoA_ERROR(ERR::MixedRings, "ConcatHor(A,B)");
    if (NumRows(M1) != NumRows(M2))
      CoCoA_ERROR(ERR::BadMatrixSize, "ConcatHor(A,B)");

    return transpose(ConcatVer(transpose(M1), transpose(M2)));
  }

  MatrixView ConcatHor(MatrixView M1, MatrixView M2)
  {
    if (BaseRing(M1) != BaseRing(M2))
      CoCoA_ERROR(ERR::MixedRings, "ConcatHor(A,B)");
    if (NumRows(M1) != NumRows(M2))
      CoCoA_ERROR(ERR::BadMatrixSize, "ConcatHor(A,B)");

    return transpose(ConcatVer(transpose(M1), transpose(M2)));
  }


  /***************************************************************************/

  MatrixView ConcatDiag(MatrixView M1, MatrixView M2)
  {
    ring R = BaseRing(M1);
    if (BaseRing(M2) != R)
      CoCoA_ERROR(ERR::MixedRings, "ConcatDiag(A,B)");

    return BlockMat(M1,
                    ZeroMat(R, NumRows(M1), NumCols(M2)),
                    ZeroMat(R, NumRows(M2), NumCols(M1)),
                    M2);
  }

  ConstMatrixView ConcatDiag(ConstMatrixView M1, ConstMatrixView M2)
  {
    ring R = BaseRing(M1);
    if (BaseRing(M2) != R)
      CoCoA_ERROR(ERR::MixedRings, "ConcatDiag(A,B)");

    return BlockMat(M1,
                    ZeroMat(R, NumRows(M1), NumCols(M2)),
                    ZeroMat(R, NumRows(M2), NumCols(M1)),
                    M2);
  }


  MatrixView ConcatAntiDiag(MatrixView M1, MatrixView M2)
  {
    ring R = BaseRing(M1);
    if (BaseRing(M2) != R)
      CoCoA_ERROR(ERR::MixedRings, "ConcatAntiDiag(A,B)");

    return BlockMat(ZeroMat(R, NumRows(M1), NumCols(M2)),
                    M1,
                    M2,
                    ZeroMat(R, NumRows(M2), NumCols(M1)));
  }

  ConstMatrixView ConcatAntiDiag(ConstMatrixView M1, ConstMatrixView M2)
  {
    ring R = BaseRing(M1);
    if (BaseRing(M2) != R)
      CoCoA_ERROR(ERR::MixedRings, "ConcatAntiDiag(A,B)");

    return BlockMat(ZeroMat(R, NumRows(M1), NumCols(M2)),
                    M1,
                    M2,
                    ZeroMat(R, NumRows(M2), NumCols(M1)));
  }



} // End of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/MatrixViews.C,v 1.4 2009/12/03 17:40:36 abbott Exp $
// $Log: MatrixViews.C,v $
// Revision 1.4  2009/12/03 17:40:36  abbott
// Added include directives for ZZ.H (as a consequence of removing
// the directive from ring.H).
//
// Revision 1.3  2009/06/25 16:59:42  abbott
// Minor improvement to some error messages (better coherence & comprehensibility).
//
// Revision 1.2  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.1  2008/04/18 15:35:57  abbott
// (long overdue) Major revision to matrices
//
//
