//   Copyright (c)  2005  John Abbott, Anna Bigatti and Massimo Caboara

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ModuleTermOrdering.H"
#include "CoCoA/error.H"
#include "CoCoA/assert.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/io.H"  // for template to output a vector

#include <iostream>
using std::ostream;
//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;

namespace CoCoA
{

  ModuleTermOrderingBase::ModuleTermOrderingBase(const PPOrdering& PPO, const std::vector<degree>& shifts):
    IntrusiveReferenceCount(),
    myPPO(PPO)
  {
    for ( size_t i=0 ; i<shifts.size() ; ++i )
      if ( GradingDim(PPO)!=GradingDim(shifts[i]) )
        CoCoA_ERROR("GradingDim(PPO)!=GradingDim(shifts[i])", "ModuleTermOrderingBase");
    myShiftsValue = shifts;
  }


  ModuleTermOrderingBase::ModuleTermOrderingBase(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<std::size_t>& perm):
    IntrusiveReferenceCount(),
    myPPO(PPO)
  {
    for ( size_t i=0 ; i<shifts.size() ; ++i )
      if ( GradingDim(PPO)!=GradingDim(shifts[i]) )
        CoCoA_ERROR("GradingDim(PPO)!=GradingDim(shifts[i])", "ModuleTermOrderingBase");
    myShiftsValue = shifts;
    if ( perm.size() != shifts.size() )
      CoCoA_ERROR("Shift list and permutation must have the same cardinality", "ModuleTermOrderingBase");
    myPermutationValue = perm;
  }


  ModuleTermOrderingBase::~ModuleTermOrderingBase()
  {}


  //---------------------------------------------------------------------------


  const std::vector<degree>& ModuleTermOrderingBase::myShifts() const
  { return myShiftsValue; }


  const std::vector<size_t>& ModuleTermOrderingBase::myPerm() const
  { return myPermutationValue; }


  const PPOrdering& ModuleTermOrderingBase::myPPOrdering() const
  { return myPPO; }


  void ModuleTermOrderingBase::myOutputSelf(std::ostream& out) const
  {
    myOutputName(out);
    out << "(" << myPPO << ", " << myShiftsValue;
    if ( !myPermutationValue.empty() )
      out << ", " << myPermutationValue;
    out << ")";
  }


  void ModuleTermOrderingBase::myOutputSelf(OpenMathOutput& OMOut) const
  {
    // missing shifts and permutation
    OMOut->mySendApplyStart();
    myOutputName(OMOut);
    OMOut << myPPO;
    OMOut->mySendApplyEnd();
  }


  //---------------------------------------------------------------------------
  //  (ex-inline) non-member functions

  std::ostream& operator<<(std::ostream& out, const ModuleTermOrdering& MTO)
  { MTO->myOutputSelf(out); return out; }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const ModuleTermOrdering& MTO)
  { MTO->myOutputSelf(OMOut); return OMOut; }


  const std::vector<degree>& shifts(const ModuleTermOrdering& MTO)
  { return MTO->myShifts(); }


  const PPOrdering& ModPPOrdering(const ModuleTermOrdering& MTO)
  { return (MTO->myPPOrdering()); }


  std::size_t NumComponents(const ModuleTermOrdering& MTO)
  { return shifts(MTO).size(); }


  std::size_t GradingDim(const ModuleTermOrdering& MTO)
  { return GradingDim(ModPPOrdering(MTO)); }


  // ----------------- concrete classes ---------------------
  namespace ModuleTermOrd
  {

    class PosWDegTOImpl: public ModuleTermOrderingBase
    {
    private:
      friend ModuleTermOrdering CoCoA::NewPosWDegTO(const PPOrdering& PPO, std::size_t NumComponents);
      friend ModuleTermOrdering CoCoA::NewPosWDegTO(const PPOrdering& PPO, const std::vector<degree>& shifts);
    private:
      PosWDegTOImpl(const PPOrdering& PPO, const std::vector<degree>& shifts);
      PosWDegTOImpl(const PosWDegTOImpl&);            ///< NEVER DEFINED -- copy ctor disabled
      PosWDegTOImpl& operator=(const PosWDegTOImpl&); ///< NEVER DEFINED -- assignment disabled
    public:
      virtual void myOutputName(std::ostream& out) const;
      virtual void myOutputName(OpenMathOutput& OMOut) const;
    };


    class WDegTOPosImpl: public ModuleTermOrderingBase
    {
    private:
      friend ModuleTermOrdering CoCoA::NewWDegTOPos(const PPOrdering& PPO, std::size_t NumComponents);
      friend ModuleTermOrdering CoCoA::NewWDegTOPos(const PPOrdering& PPO, const std::vector<degree>& shifts);
      friend ModuleTermOrdering CoCoA::NewWDegTOPos(const PPOrdering& PPO, const std::vector<size_t>& perm);
      friend ModuleTermOrdering CoCoA::NewWDegTOPos(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm);
    private:
      WDegTOPosImpl(const PPOrdering& PPO, const std::vector<degree>& shifts);
      WDegTOPosImpl(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm);
      WDegTOPosImpl(const WDegTOPosImpl&);            ///< NEVER DEFINED -- copy ctor disabled
      WDegTOPosImpl& operator=(const WDegTOPosImpl&); ///< NEVER DEFINED -- assignment disabled
    public:
      virtual void myOutputName(std::ostream& out) const;
      virtual void myOutputName(OpenMathOutput& OMOut) const;
    };


    class WDegPosTOImpl: public ModuleTermOrderingBase
    {
    private:
      friend ModuleTermOrdering CoCoA::NewWDegPosTO(const PPOrdering& PPO, std::size_t NumComponents);
      friend ModuleTermOrdering CoCoA::NewWDegPosTO(const PPOrdering& PPO, const std::vector<degree>& shifts);
      friend ModuleTermOrdering CoCoA::NewWDegPosTO(const PPOrdering& PPO, const std::vector<size_t>& perm);
      friend ModuleTermOrdering CoCoA::NewWDegPosTO(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm);
    private:
      WDegPosTOImpl(const PPOrdering& PPO, const std::vector<degree>& shifts);
      WDegPosTOImpl(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm);
      WDegPosTOImpl(const PPOrdering& PPO, const WDegPosTOImpl&);            ///< NEVER DEFINED -- copy ctor disabled
      WDegPosTOImpl& operator=(const WDegPosTOImpl&); ///< NEVER DEFINED -- assignment disabled
    public:
      virtual void myOutputName(std::ostream& out) const;
      virtual void myOutputName(OpenMathOutput& OMOut) const;
    };


    //---------- PosWDegTOImpl ----------------------------------------

    PosWDegTOImpl::PosWDegTOImpl(const PPOrdering& PPO, const std::vector<degree>& shifts):
      ModuleTermOrderingBase(PPO, shifts)
    {}


    void PosWDegTOImpl::myOutputName(std::ostream& out) const
    {  out << "ModuleTermOrderingWDegTOPos"; }

    void PosWDegTOImpl::myOutputName(OpenMathOutput& OMOut) const
    {  OMOut << OpenMathSymbol("cocoa", "ModuleTermOrderingWDegTOPos"); }


    //---------- WDegTOPosImpl ----------------------------------------

    WDegTOPosImpl::WDegTOPosImpl(const PPOrdering& PPO, const std::vector<degree>& shifts):
      ModuleTermOrderingBase(PPO, shifts)
    {}


    WDegTOPosImpl::WDegTOPosImpl(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm):
      ModuleTermOrderingBase(PPO, shifts, perm)
    {}


    void WDegTOPosImpl::myOutputName(std::ostream& out) const
    {  out << "ModuleTermOrderingWDegTOPos"; }

    void WDegTOPosImpl::myOutputName(OpenMathOutput& OMOut) const
    {  OMOut << OpenMathSymbol("cocoa", "ModuleTermOrderingWDegTOPos"); }


    //---------- WDegPosTOImpl ----------------------------------------

    WDegPosTOImpl::WDegPosTOImpl(const PPOrdering& PPO, const std::vector<degree>& shifts):
      ModuleTermOrderingBase(PPO, shifts)
    {}


    WDegPosTOImpl::WDegPosTOImpl(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm):
      ModuleTermOrderingBase(PPO, shifts, perm)
    {}


    void WDegPosTOImpl::myOutputName(std::ostream& out) const
    {  out << "ModuleTermOrderingWDegTOPos"; }


    void WDegPosTOImpl::myOutputName(OpenMathOutput& OMOut) const
    {  OMOut << OpenMathSymbol("cocoa", "ModuleTermOrderingWDegTOPos"); }


    //------------------------------------------------------------//


  } // end of namespace ModuleTermOrd


  ModuleTermOrdering NewPosWDegTO(const PPOrdering& PPO, std::size_t NumComponents)
  {
    return ModuleTermOrdering(new ModuleTermOrd::PosWDegTOImpl(PPO, vector<degree>(NumComponents, degree(GradingDim(PPO)))));
  }

  ModuleTermOrdering NewWDegTOPos(const PPOrdering& PPO, std::size_t NumComponents)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegTOPosImpl(PPO, vector<degree>(NumComponents, degree(GradingDim(PPO)))));
  }


  ModuleTermOrdering NewWDegPosTO(const PPOrdering& PPO, std::size_t NumComponents)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegPosTOImpl(PPO, vector<degree>(NumComponents, degree(GradingDim(PPO)))));
  }


  ModuleTermOrdering NewPosWDegTO(const PPOrdering& PPO, const std::vector<degree>& shifts)
  {
    return ModuleTermOrdering(new ModuleTermOrd::PosWDegTOImpl(PPO, shifts));
  }


  ModuleTermOrdering NewWDegTOPos(const PPOrdering& PPO, const std::vector<degree>& shifts)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegTOPosImpl(PPO, shifts));
  }


  ModuleTermOrdering NewWDegPosTO(const PPOrdering& PPO, const std::vector<degree>& shifts)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegPosTOImpl(PPO, shifts));
  }


  ModuleTermOrdering NewWDegTOPos(const PPOrdering& PPO, const std::vector<size_t>& perm)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegTOPosImpl(PPO, vector<degree>(perm.size(), degree(GradingDim(PPO))), perm));
  }


  ModuleTermOrdering NewWDegPosTO(const PPOrdering& PPO, const std::vector<size_t>& perm)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegPosTOImpl(PPO, vector<degree>(perm.size(), degree(GradingDim(PPO))), perm));
  }


  ModuleTermOrdering NewWDegTOPos(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegTOPosImpl(PPO, shifts, perm));
  }


  ModuleTermOrdering NewWDegPosTO(const PPOrdering& PPO, const std::vector<degree>& shifts, const std::vector<size_t>& perm)
  {
    return ModuleTermOrdering(new ModuleTermOrd::WDegPosTOImpl(PPO, shifts, perm));
  }


  bool IsPosWDegTO(const ModuleTermOrdering& MTO)
  {
    if (dynamic_cast<const ModuleTermOrd::PosWDegTOImpl*>(MTO.myRawPtr())) return true;
    // must decide whether the matrix is PosWDeg..., possibly in disguise
    return false;
  }


  bool IsWDegTOPos(const ModuleTermOrdering& MTO)
  {
    if (dynamic_cast<const ModuleTermOrd::WDegTOPosImpl*>(MTO.myRawPtr())) return true;
    // must decide whether the matrix is WDeg..., possibly in disguise
    return false;
  }


  bool IsWDegPosTO(const ModuleTermOrdering& MTO)
  {
    if (dynamic_cast<const ModuleTermOrd::WDegPosTOImpl*>(MTO.myRawPtr())) return true;
    // must decide whether the matrix is WDeg..., possibly in disguise
    return false;
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ModuleTermOrdering.C,v 1.2 2007/10/30 17:14:08 abbott Exp $
// $Log: ModuleTermOrdering.C,v $
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.10  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.9  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.8  2007/01/18 15:34:04  cocoa
// -- changed namespace MTO into ModuleTermOrd
//
// Revision 1.7  2006/11/24 17:14:10  cocoa
// -- reorganized includes of header files
// -- SmartPtrIRC for reference counting
//
// Revision 1.6  2006/11/10 13:30:57  cocoa
// -- fixed: "const &" to PPOrdering arguments
// -- some more documentation
//
// Revision 1.5  2006/11/10 13:06:03  cocoa
// -- cleaned code
//
// Revision 1.4  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.3  2006/10/06 15:11:01  cocoa
// -- removed commented out code
//
// Revision 1.2  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.8  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.7  2006/05/11 15:59:23  cocoa
// -- changed reference count is done using SmartPtrIRC
//
// Revision 1.6  2006/05/04 14:19:02  cocoa
// -- moved some code from .H to .C
//
// Revision 1.5  2006/04/28 11:32:16  cocoa
// -- moved concrete class definition from .H to .C
//
// Revision 1.4  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.3  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
// Revision 1.2  2005/11/24 17:06:40  cocoa
// -- fixed IsWDegTOPos, IsWDegPosTO, IsWDegTOPermPos, IsWDegPermPosTO
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1  2005/09/28 11:50:34  cocoa
// -- new code for graded modules
//
