//   Copyright (c)  1999,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/NumTheory.H"
#include "CoCoA/utils.H"

#include <algorithm>
using std::swap;
#include <limits>
using std::numeric_limits;
#include <vector>
using std::vector;

namespace CoCoA
{

  // Anonymous namespace for file local "static" variables.
  namespace
  {
    // mask[n%30] == 0 iff n is divisible by at least one of 2, 3 and 5.
    // n+skip[n%30] is the next number after n not divisible by 2, 3 or 5
    // n-fall[n%30] is the largest number below n not divisible by 2, 3 or 5
//???NOT USED???    const unsigned int mask[30] = {0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,1};
    const unsigned int skip[30] = {1,6,5,4,3,2,1,4,3,2,1,2,1,4,3,2,1,2,1,4,3,2,1,6,5,4,3,2,1,2};
    const unsigned int fall[30] = {1,2,1,2,3,4,5,6,1,2,3,4,1,2,1,2,3,4,1,2,1,2,3,4,1,2,3,4,5,6};


    // Function to compute base^exp mod modulus; assumes 0 <= base < modulus.
    // Unclever for exp >= modulus.  modulus need not be prime, but (modulus-1)^2
    // must fit inside an unsigned long.
    // This is an iterative implementation of binary powering; seems to be
    // usefully faster than the more obvious recursive implementation.
    // If modulus=1, always produces 0.  Produces 1 when asked to compute 0^0.
    unsigned long PowerModSmallModulus(unsigned long base, unsigned long exp, unsigned long modulus)
    {
      CoCoA_ASSERT(modulus != 0);
      CoCoA_ASSERT(modulus-1 <= NumericLimits<unsigned long>::MaxSquarableInteger());

///???      if (modulus == 1) return 0;
      if (exp == 0 || base == 1) return 1;
      if (exp == 1 || base == 0) return base;
    
      unsigned long ans = 1;
      while (exp > 1)
      {
        if (exp&1) ans = (ans*base)%modulus;
        exp /= 2;
        base = (base*base)%modulus;
      }
      return (ans*base)%modulus;
    }

    // This handles case of large modulus or large exponent.
    ZZ PowerModLargeModulus(const ZZ& base, const ZZ& exponent, const ZZ& modulus)
    {
///???      CoCoA_ASSERT(modulus > NumericLimits<unsigned long>::MaxSquarableInteger());
      if (modulus < 2)
        CoCoA_ERROR(ERR::BadArg, "PowerMod(base, exp, modulus):  modulus must be >= 2");
      if (exponent < 0)
        CoCoA_ERROR(ERR::NegExp, "PowerMod(base, exp, modulus)");

///      if (modulus == 1) return 0;
      if (exponent == 1 || base == 1 || base == 0) return base;
      if (exponent == 0) return ZZ(1);
    
      ZZ ans;
      mpz_powm(mpzref(ans), mpzref(base), mpzref(exponent), mpzref(modulus));
      return ans;
    }


    // Miller-Rabin strong pseudo-prime test (for machine integers).
    // Tests whether n is a strong pseudo-prime to base b.
    // Assumes 1 < b < n
    bool StrongPseudoPrime(unsigned long b, unsigned long n)
    {
      CoCoA_ASSERT(n <= NumericLimits<unsigned long>::MaxSquarableInteger());
      CoCoA_ASSERT(1 < b && b < n);
///???      if (n < 2 || b%n == 0) return false;

      // Compute q,r such that  n-1 = q*2^r  with q odd.
      unsigned long q = n-1;
      int r = 0;
      while ((q&1) == 0)
      {
        ++r;
        q /= 2;
      }

      // Main loop
      const unsigned long n1 = n-1;
      unsigned long power = PowerMod(b, q, n);
      if (power == n1 || power == 1) return true;
      for (int i=0; i < r; i++)
      {
        power = (power*power)%n;
        if (power == n1) return true;
        if (power == 1) return false;
      }
      return false;
    }

    // This definition is for 64-bit machines, but remains valid for 32-bitters
    // ASSUMES n is less than sqrt of max unsigned long.
    // I do not recall the reference for this definition.
    bool IsSmallPrime(unsigned long n)
    {
      CoCoA_ASSERT(n != 0);
      CoCoA_ASSERT(n <= NumericLimits<unsigned long>::MaxSquarableInteger());
      if (n == 2 || n == 3 || n == 5 || n == 7) return true;
      if (n < 11) return false;
      // Exclude the 4 composite pseudo-primes (up to 2^32).
      if (n ==  746331041UL ||
          n == 2840871041UL ||
          n == 3014101261UL ||
          n == 3215031751UL)
        return false;
      return StrongPseudoPrime(2, n) &&
        StrongPseudoPrime(5, n) &&
        StrongPseudoPrime(7, n);
    }


    // Assumes N >> 0.
    // Returns true if N is prime, and false otherwise.
    // Maybe slow for those large N where N-1 is hard to factorize.
    bool LucasTest(const ZZ& N)
    {
      if (N <= 0) CoCoA_ERROR(ERR::BadArg, "LucasTest(N):  N must be positive");
      if (N == 1) return false;
      if (N == 2) return true;
      const factorization<ZZ> facpows = factor(N-1);
      const vector<ZZ>& primes = facpows.myFactors;
      const int NumPrimes = primes.size();
      const int Amax = static_cast<int>(floor(2*log(N)*log(N))); // experimentally checked up to 10^9: a better bound is 2.62*log(N)*log(log(N)) for N > 3
      for (int a=2; a <= Amax; ++a)
      {
        if (a==4 || a==8 || a==9 || a==16) continue; // skip some pure powers
        ZZ pwr;
        for (int i=0; i < NumPrimes; ++i)
        {
          pwr = PowerMod(a, (N-1)/primes[i], N);
          if (pwr == 1) break;
        }
        if (pwr != 1) return true;
      }
      return false;
    }


    // This fn may assume that abs(N) > NumericLimits<unsigned long>::MaxSquarableInteger().
    bool IsBigPrime(const ZZ& N)
    {
      CoCoA_ASSERT(N > NumericLimits<unsigned long>::MaxSquarableInteger());
      if (!IsProbPrime(abs(N))) return false;
      return LucasTest(abs(N));
    }

  } // end of anonymous namespace


  // Compute (base^exp)%modulus  with exp >= 0  and modulus >= 2
  // The case 0^0 yields 1.
  unsigned long PowerMod(const MachineInteger& base, const MachineInteger& exponent, const MachineInteger& modulus)
  {
    if (IsNegative(modulus) || abs(modulus) < 2)
      CoCoA_ERROR(ERR::BadArg, "PowerMod(base,exponent,modulus):  modulus must be >= 2");
    if (IsNegative(exponent))
    {
      const unsigned long inv = InvMod(base, modulus);
      if (inv == 0) CoCoA_ERROR(ERR::DivByZero, "PowerMod: negative exponent & non invertible base");
      return PowerMod(inv, abs(exponent), modulus);
    }
    const unsigned long e = AsUnsignedLong(exponent);
    if (e == 0) return 1;
    const unsigned long m = AsUnsignedLong(modulus);
    unsigned long b = abs(base)%m;
    if (b == 0) return 0;
    if (IsNegative(base) && e%2 == 1)
      b = m-b;
    if (m <= NumericLimits<unsigned long>::MaxSquarableInteger())
      return PowerModSmallModulus(b,e,m); // Call the fn in the anon namespace above.
    // Square of modulus doesn't fit into unsigned long, so compute with ZZs and convert answer back to unsigned long.
    unsigned long ans;
    convert(ans, PowerMod(b,e,ZZ(m)));
    return ans;
  }

  unsigned long PowerMod(const MachineInteger& base, const ZZ& exponent, const MachineInteger& modulus)
  {
    // BUG/SLUG: horribly inefficient!!
    // What is the best way to implement this fn?????  Any volunteers?
    unsigned long ans;
    convert(ans, PowerMod(ZZ(base), exponent, ZZ(modulus)));
    return ans;
  }

  unsigned long PowerMod(const ZZ& base, const MachineInteger& exponent, const MachineInteger& modulus)
  {
    if (IsNegative(modulus) || abs(modulus) < 2)
      CoCoA_ERROR(ERR::BadArg, "PowerMod(base,exponent,modulus):  modulus must be >= 2");
    return PowerMod(base%modulus, exponent, modulus);
  }

  unsigned long PowerMod(const ZZ& base, const ZZ& exponent, const MachineInteger& modulus)
  {
    // BUG/SLUG: horribly inefficient!!
    // What is the best way to implement this fn?????  Any volunteers?
    unsigned long ans;
    convert(ans, PowerMod(base, exponent, ZZ(modulus)));
    return ans;
  }

  ZZ PowerMod(const MachineInteger& base, const MachineInteger& exponent, const ZZ& modulus)
  {
    return PowerMod(ZZ(base), ZZ(exponent), modulus);
  }

  ZZ PowerMod(const ZZ& base, const MachineInteger& exponent, const ZZ& modulus)
  {
    if (modulus < 2)
      CoCoA_ERROR(ERR::BadArg, "PowerMod(ZZ, MachineInteger, ZZ): modulus must be >= 2");
    if (IsNegative(exponent))
    {
      const ZZ inv = InvMod(base, modulus);
      if (inv == 0) CoCoA_ERROR(ERR::DivByZero, "PowerMod: negative exponent & non invertible base");
      return PowerMod(inv, abs(exponent), modulus);
    }
    ZZ ans;
    mpz_powm_ui(mpzref(ans), mpzref(base), AsUnsignedLong(exponent), mpzref(modulus));
    return ans;
  }


  ZZ PowerMod(const MachineInteger& base, const ZZ& exponent, const ZZ& modulus)
  {
    if (modulus < 2)
      CoCoA_ERROR(ERR::BadArg, "PowerMod(MachineInteger, ZZ, ZZ)");
    if (exponent < 0)
    {
      const ZZ inv = InvMod(base, modulus);
      if (inv == 0) CoCoA_ERROR(ERR::DivByZero, "PowerMod: negative exponent & non invertible base");
      return PowerMod(inv, abs(exponent), modulus);
    }
    return PowerMod(ZZ(base), exponent, modulus);
  }


  ZZ PowerMod(const ZZ& base, const ZZ& exponent, const ZZ& modulus)
  {
    if (modulus < 2)
      CoCoA_ERROR(ERR::ZeroModulus, "PowerMod(ZZ, ZZ, ZZ)");
    if (exponent < 0)
    {
      const ZZ inv = InvMod(base, modulus);
      if (inv == 0) CoCoA_ERROR(ERR::DivByZero, "PowerMod: negative exponent & non invertible base");
      return PowerMod(inv, abs(exponent), modulus);
    }
    ZZ ans;
    mpz_powm(mpzref(ans), mpzref(base), mpzref(exponent), mpzref(modulus));
    return ans;
  }



  // Non-negative gcd of machine integers.
  // Emphasis is on simplicity rather than utmost speed.
  unsigned long gcd(const MachineInteger& a, const MachineInteger& b)
  {
    unsigned long A = abs(a);
    unsigned long B = abs(b);
    // Dispose of the trivial cases first.
    if (A == 0) return B;
    if (B == 0) return A;
    if (A == 1 || B == 1) return 1;
    if (A == B) return A;

    // General case.
    if (A < B) std::swap(A, B);
    while (B != 0)
    {
      A %= B;
      std::swap(A, B);
    }
    return A;
  }

  const ZZ gcd(const ZZ& A, const MachineInteger& b)
  { if (IsZero(b)) return abs(A); else return ZZ(gcd(A%abs(b), b)); }

  const ZZ gcd(const MachineInteger& a, const ZZ& B)
  { if (IsZero(a)) return abs(B); else return ZZ(gcd(a, B%abs(a))); }

  const ZZ gcd(const ZZ& A, const ZZ& B)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_gcd(ans, mpzref(A), mpzref(B));
    return ans[0];
  }


// Do we want a procedural form???
//   void gcd(ZZ& lhs, const ZZ& a, const ZZ& b)
//   {
//     //  if (IsZero(a) && IsZero(b)) throw invalid_argument("gcd(ZZ,ZZ,ZZ): both input arguments are 0");
//     mpz_gcd(mpzref(lhs), mpzref(a), mpzref(b));
//   }


  // This fn is OK -- it is not obvious, but the longs cannot overflow: for
  // a proof see Theorem 4.3 & the comment after it in Shoup's book
  // "A Computational Introduction to Number Theory and Algebra".
  unsigned long ExtendedEuclideanAlg(long& CofacA, long& CofacB, unsigned long a, unsigned long b)
  {
    if (a < b) return ExtendedEuclideanAlg(CofacB, CofacA, b, a);

    // Now we are sure that a >= b.
    long m00 = 1;
    long m01 = 0;
    long m10 = 0;
    long m11 = 1;

    while (true)
    {
      if (b==0) { CofacA = m00; CofacB = m01; return a; }
      long q = a/b; // can overflow (harmlessly) only if b == 1 upon 1st iteration
      a -= q*b;
      m00 -= q*m10;
      m01 -= q*m11;

      if (a == 0) { CofacA = m10; CofacB = m11; return b; }
      q = b/a;
      b -= q*a;
      m10 -= q*m00;
      m11 -= q*m01;
    }
  }


  unsigned long ExtGcd(long& CofacA, long& CofacB, const MachineInteger& a, const MachineInteger& b)
  {
    const unsigned long g = ExtendedEuclideanAlg(CofacA, CofacB, abs(a), abs(b));
    if (IsNegative(a)) CofacA = -CofacA;
    if (IsNegative(b)) CofacB = -CofacB;
    return g;
  }

  ZZ ExtGcd(ZZ& CofacA, ZZ& CofacB, const ZZ& A, const ZZ& B)
  {
    ZZ ans;
    mpz_gcdext(mpzref(ans), mpzref(CofacA), mpzref(CofacB), mpzref(A), mpzref(B));
    // GMP guarantees only that abs(CofacA) < abs(B), but we want 2*abs(CofacA) <= abs(B)/ans
    // so the next few lines check, and if necessary tweak, the cofactors.
    // The code below is needed only for GMP-4.3.1; GMP-4.2.4 gave correctly reduced answers.
    if (A == 0 || B == 0 || abs(A) == abs(B)) return ans;
    const ZZ CofacAModulus = abs(B)/ans;
    const ZZ CofacBModulus = abs(A)/ans;
    const ZZ q = RoundDiv(abs(CofacA), CofacAModulus);
    if (q != 0)
    {
      if (CofacA < 0)
      { CofacA += q*CofacAModulus; if (sign(A) == sign(B)) CofacB -= q*CofacBModulus; else CofacB += q*CofacBModulus; }
      else
      { CofacA -= q*CofacAModulus; if (sign(A) == sign(B)) CofacB += q*CofacBModulus; else CofacB -= q*CofacBModulus; }
    }
    if (2*abs(CofacB) <= CofacBModulus) return ans; // NB already have 2*abs(CofacA) <= CofacAModulus
    if (CofacB < 0)
    { CofacB += CofacBModulus; if (sign(A) == sign(B)) CofacA -= CofacAModulus; else CofacA += CofacAModulus; }
    else
    { CofacB -= CofacBModulus; if (sign(A) == sign(B)) CofacA += CofacAModulus; else CofacA -= CofacAModulus; }
    return ans;
  }


  unsigned long InvMod(const MachineInteger& r, const MachineInteger& m)
  {
    if (IsNegative(m) || abs(m) < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(residue,modulus):  modulus must be >= 2");
    const unsigned long modulus = abs(m);
    unsigned long residue = abs(r)%modulus;
    if (residue==0) return residue;
    if (IsNegative(r)) residue = modulus-residue;
    long inverse, junk;
    if (ExtGcd(inverse,junk,residue,modulus) != 1) return 0;
    if (inverse > 0) return inverse;
    return modulus+inverse;
  }

  unsigned long InvMod(const ZZ& r, const MachineInteger& m)
  {
    if (IsNegative(m) || abs(m) < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(residue,modulus):  modulus must be >= 2");
    return InvMod(r%m, m);
  }

  ZZ InvMod(const MachineInteger& r, const ZZ& m)
  {
    if (m < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(residue,modulus):  modulus must be >= 2");
    const ZZ residue(r);
    ZZ ans;
    const int InverseExists = mpz_invert(mpzref(ans), mpzref(residue), mpzref(m));
    if (!InverseExists) ans = 0;
    return ans;
  }

  ZZ InvMod(const ZZ& r, const ZZ& m)
  {
    if (m < 2) CoCoA_ERROR(ERR::BadArg, "InvMod(residue,modulus):  modulus must be >= 2");
    ZZ ans;
    const int InverseExists = mpz_invert(mpzref(ans), mpzref(r), mpzref(m));
    if (!InverseExists) ans = 0;
    return ans;
  }


  unsigned long lcm(const MachineInteger& a, const MachineInteger& b)
  {
    if (IsZero(a) || IsZero(b)) return 0;
    const unsigned long AoverG = abs(a)/gcd(a,b);
    if (abs(b) > numeric_limits<unsigned long>::max()/AoverG)
      CoCoA_ERROR(ERR::ArgTooBig, "lcm(MachineInteger,MachineInteger)");
    return AoverG*abs(b);
  }

  const ZZ lcm(const ZZ& A, const MachineInteger& b)
  {
    return lcm(A, ZZ(b));
  }

  const ZZ lcm(const MachineInteger& a, const ZZ& B)
  {
    return lcm(ZZ(a), B);
  }

  const ZZ lcm(const ZZ& a, const ZZ& b)
  {
    mpz_t ans;
    mpz_init(ans);
    mpz_lcm(ans, mpzref(a), mpzref(b));
    return ans[0];
  }


// Do we want a procedural form???
//   void lcm(ZZ& lhs, const ZZ& a, const ZZ& b)
//   {
//     mpz_lcm(mpzref(lhs), mpzref(a), mpzref(b));
//   }



  // A dispatcher: for small values which fit into a machine integer.
  // We call IsSmallPrime if possible, otherwise we call IsBigPrime.
  bool IsPrime(const MachineInteger& mi)
  {
    if (IsZero(mi) || IsNegative(mi)) CoCoA_ERROR(ERR::BadArg, "IsPrime(n):  n must be strictly positive");

    const unsigned long n = abs(mi);
    if (n <= NumericLimits<unsigned long>::MaxSquarableInteger())
      return IsSmallPrime(n);
    return IsBigPrime(ZZ(n));
  }

  // A dispatcher: for potentially large values (with no limit).
  // We call IsSmallPrime if possible, otherwise we call IsBigPrime.
  bool IsPrime(const ZZ& N)
  {
    if (N <= 0) CoCoA_ERROR(ERR::BadArg, "IsPrime(N):  N must be strictly positive");
    long n;
    if (IsConvertible(n, N) && abs(n) <= NumericLimits<unsigned long>::MaxSquarableInteger())
      return IsSmallPrime(abs(n));
    return IsBigPrime(abs(N));
  }

  // "Probable prime" test.  Currently uses Miller-Rabin.
  bool IsProbPrime(const MachineInteger& n, const MachineInteger& NumIters /*= 25*/)
  {
    if (IsZero(n) || IsNegative(n)) CoCoA_ERROR(ERR::BadArg, "IsProbPrime(n,NumIters):  n must be strictly positive");
    if (IsZero(NumIters) || IsNegative(NumIters)) CoCoA_ERROR(ERR::BadArg, "IsProbPrime(n,NumIters):  NumIters must be strictly positive");
    // Just call the slow version -- at least will guarantee coherent behaviour
    return IsProbPrime(ZZ(n), NumIters);
  }

  bool IsProbPrime(const ZZ& N, const MachineInteger& NumIters /*= 25*/)
  {
    if (N <= 0) CoCoA_ERROR(ERR::BadArg, "IsProbPrime(N,NumIters):  N must be strictly positive");
    if (IsZero(NumIters) || IsNegative(NumIters)) CoCoA_ERROR(ERR::BadArg, "IsProbPrime(N,NumIters):  NumIters must be strictly positive");
    return mpz_probab_prime_p(mpzref(abs(N)), AsUnsignedLong(NumIters));
  }



  unsigned long NextPrime(const MachineInteger& mi)
  {
    if (IsNegative(mi) || IsZero(mi)) CoCoA_ERROR(ERR::BadArg, "NextPrime(n):  n must be strictly positive");
    unsigned long n = abs(mi);

    // Special cases if n < 5
    if (n < 2) return 2;
    if (n == 2) return 3;
    if (n < 5) return 5;

    unsigned long n30 = n%30;
    while (true)
    {
      const unsigned long delta = skip[n30];
      if (n+delta < n) break; // break if n overflows
      n += delta;
      if (n30 == 29) n30=1;
      else n30 += delta;
      if (IsPrime(n)) return n;
    }
    // Reach here only if "overflow" has occurred.
    return 0; // to signify "overflow"
  }


  unsigned long PrevPrime(const MachineInteger& mi)
  {
    if (IsNegative(mi) || IsZero(mi)) CoCoA_ERROR(ERR::BadArg, "PrevPrime(n):  n must be strictly positive");
    unsigned long n = abs(mi);

    // Special cases if n < 8
    if (n < 3) return 0; // 0 indicates no previous prime
    if (n == 3) return 2;
    if (n < 6) return 3;
    if (n < 8) return 5;

    unsigned long n30 = n%30;
    do
    {
      const unsigned long delta = fall[n30];
      n -= delta;
      if (n30 < delta) n30 += 30;
      n30 -= delta;
    } while (!IsPrime(n));
    return n;
  }


  ZZ NextProbPrime(ZZ N, const MachineInteger& NumIters /*= 25*/)
  {
    if (N <= 0) CoCoA_ERROR(ERR::BadArg, "NextProbPrime(N,NumIters):  N must be strictly positive");
    if (IsZero(NumIters) || IsNegative(NumIters)) CoCoA_ERROR(ERR::BadArg, "NextProbPrime(N,NumIters):  NumIters must be strictly positive");

    // if N is small use NextPrime
    unsigned long n;
    if (IsConvertible(n, N))
    {
      unsigned long ans = NextPrime(n);
      if (ans != 0) return ZZ(ans);
    }

    unsigned long N30 = N%30;
    do
    {
      const unsigned long delta = skip[N30];
      N += delta;
      if (N30 == 29) N30=1;
      else N30 += delta;
    } while (!IsProbPrime(N, NumIters));
    return N;
  }

  ZZ PrevProbPrime(ZZ N, const MachineInteger& NumIters /*= 25*/)
  {
    if (N <= 0) CoCoA_ERROR(ERR::BadArg, "PrevProbPrime(N,NumIters):  N must be strictly positive");
    if (IsZero(NumIters) || IsNegative(NumIters)) CoCoA_ERROR(ERR::BadArg, "PrevProbPrime(N,NumIters):  NumIters must be strictly positive");

    // if N is small use PrevPrime
    unsigned long n;
    if (IsConvertible(n, N))
      return ZZ(PrevPrime(n));

    unsigned long N30 = N%30;
    do
    {
      const unsigned long delta = fall[N30];
      N -= delta;
      if (N30 < delta) N30 += 30;
      N30 -= delta;
    } while (!IsProbPrime(N, NumIters));
    return N;
  }


// PrimeSource PrimeSourceCtor(void)
// {
//   PrimeSource ans;

//   ans = (PrimeSource)MALLOC(sizeof(struct PrimeSourceStruct));
//   ans->start = nextprime(MAX_PRIME/3); /* heuristic/guess */
//   ans->p = ans->start;
//   return ans;
// }

// void PrimeSourceDtor(PrimeSource PS)
// {
//   FREE(PS);
// }

// FFelem NextPrime(PrimeSource PS)
// {
//   if (PS->start > 0)
//   {
//     PS->p = nextprime(PS->p);
//     if (PS->p != 0) return PS->p;
//     PS->p = PS->start;
//     PS->start = 0;
//   }
//   PS->p = prevprime(PS->p);
//   if (PS->p == 0) JERROR(JERROR_PRIMES_EXHAUSTED);
//   return PS->p;
// }


  factorization<long> SmoothFactor(const MachineInteger& n, const MachineInteger& TrialLimit)
  {
    if (IsZero(n))
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(n,TrialLimit):  n must be non-zero");
    if (!IsSignedLong(TrialLimit) || AsSignedLong(TrialLimit) < 2)
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(n,TrialLimit):  TrialLimit must be at least 2 and fit into a machine long");
    // Below Pmax is unsigned long so that the code will work even if input TrialLimit is numeric_limits<long>::max()
    const unsigned long Pmax = AsUnsignedLong(TrialLimit);
    if (!IsSignedLong(n))
      CoCoA_ERROR(ERR::ArgTooBig, "SmoothFactor(n,TrialLimit):  number to be factorized must fit into a machine long");
    unsigned long RemainingFactor = abs(n);


    // Main loop: we simply do trial divisions by 2,3,5 & numbers not divisible by 2,3,5.
    vector<long> factors;
    vector<long> exponents;
    unsigned long p = 2;
    unsigned long p30 = 5;
    // NB in line below cannot test p*p <= RemainingFactor directly as the product may overflow.
    while (p <= Pmax && p <= RemainingFactor/p)
    {
      int exp = 0;
      unsigned long rem = RemainingFactor % p;
      while (rem == 0)
      {
        ++exp;
        RemainingFactor /= p;
        rem = RemainingFactor % p;
      }
      if (exp > 0)
      {
        factors.push_back(p);
        exponents.push_back(exp);
      }
      // Advance to next p (we avoid multiples of 2,3,5 but do not skip multiples of 7,11,...)
      // Restricting to prime p actually makes the code slower.
      if (p==2) { p=3; continue; }
      if (p==3) { p=5; continue; }

      const unsigned long delta = skip[p30];
      p += delta;
      if (p30 >= 29) p30=1;
      else p30 += delta;
    }
    // if RemainingFactor is below limit, add it to the list of factors found.
    if (RemainingFactor > 1 && RemainingFactor <= Pmax)
    {
      factors.push_back(RemainingFactor);
      exponents.push_back(1);
      RemainingFactor = 1;
    }
    if (IsNegative(n))
      return factorization<long>(factors, exponents, -static_cast<long>(RemainingFactor));
    return factorization<long>(factors, exponents, RemainingFactor);
  }


  // This is very similar to the function above -- but I don't see how to share code.
  factorization<ZZ> SmoothFactor(const ZZ& N, const MachineInteger& TrialLimit)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(N,TrialLimit):  N must be non-zero");
    if (!IsSignedLong(TrialLimit) || AsSignedLong(TrialLimit) < 2)
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(N,TrialLimit):  TrialLimit must be at least 2 and fit into a machine long");
    // Below Pmax is unsigned long so that the code will work even if input TrialLimit is numeric_limits<long>::max()
    const unsigned long Pmax = AsUnsignedLong(TrialLimit);
    ZZ RemainingFactor = abs(N);


    // Main loop: we simply do trial divisions by 2,3,5 & numbers not divisible by 2,3,5.
    vector<ZZ> factors;
    vector<long> exponents;
    unsigned long p = 2;
    // NB in line below cannot test p*p <= RemainingFactor directly as the product may overflow.
    while (p <= Pmax && p <= RemainingFactor/p)
    {
      int exp = 0;
      ZZ quo,rem;
      quorem(quo,rem, RemainingFactor, p);
      while (rem == 0)
      {
        ++exp;
        RemainingFactor = quo;
        quorem(quo,rem, RemainingFactor, p);
      }
      if (exp > 0)
      {
        factors.push_back(ZZ(p));
        exponents.push_back(exp);
      }
      p = NextPrime(p);
    }
    // if RemainingFactor is below limit, add it to the list of factors found.
    if (RemainingFactor > 1 && RemainingFactor <= Pmax)
    {
      factors.push_back(RemainingFactor);
      exponents.push_back(1);
      RemainingFactor = 1;
    }
    if (N < 0)
      return factorization<ZZ>(factors, exponents, -RemainingFactor);
    return factorization<ZZ>(factors, exponents, RemainingFactor);
  }

  factorization<ZZ> SmoothFactor(const ZZ& N, const ZZ& TrialLimit)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(N,TrialLimit):  N must be non-zero");
    if (TrialLimit < 2)
      CoCoA_ERROR(ERR::BadArg, "SmoothFactor(N,TrialLimit):  TrialLimit must be at least 2");
    
    // Not implemented for large TrialLimit because it would be hideously slow...
    // A naive implementation could simply copy code from SmoothFactor(N,pmax) above.

    long pmax;
    if (!IsConvertible(pmax, TrialLimit))
      CoCoA_ERROR(ERR::NYI, "SmoothFactor(N,TrialLimit) with TrialLimit greater than largest signed long");
    return SmoothFactor(N, pmax);
  }


  factorization<long> factor(const MachineInteger& N)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::BadArg, "factor(n):  n must be non-zero");
    if (!IsSignedLong(N))
      CoCoA_ERROR(ERR::ArgTooBig, "factor(n):  n must fit into a signed long");
    // Simple rather than efficient.
    if (abs(N) < 2) return SmoothFactor(N,2);
    return SmoothFactor(N,N);
  }

  factorization<ZZ> factor(const ZZ& N)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::BadArg, "factor(N):  N must be non-zero");
    const long PrimeLimit = 10000000; // somewhat arbitrary choice of limit for trial division
    factorization<ZZ> ans = SmoothFactor(N, PrimeLimit);
    if (abs(ans.myRemainingFactor) != 1 && (abs(ans.myRemainingFactor) < power(PrimeLimit,2) || IsPrime(abs(ans.myRemainingFactor))))
    {
      ans.myFactors.push_back(ans.myRemainingFactor);
      ans.myExponents.push_back(1);
      ans.myRemainingFactor = sign(ans.myRemainingFactor);
      return ans;
    }
    if (abs(ans.myRemainingFactor) != 1)
      CoCoA_ERROR(ERR::NYI, "factor(N) unimplemented in this case -- too many large factors");
    return ans;
  }


  unsigned long EulerPhi(const MachineInteger& n)
  {
    if (IsZero(n) || IsNegative(n))
      CoCoA_ERROR(ERR::BadArg, "EulerPhi(n):  n must be strictly positive");
    const factorization<long> facpows = factor(n);
    const vector<long>& primes = facpows.myFactors;
    unsigned long ans = abs(n);
    int NumPrimes = primes.size();
    for (int i=0; i < NumPrimes; ++i)
    {
      const unsigned long p = primes[i];
      ans = (ans/p)*(p-1);
    }
    return ans;
  }

  ZZ EulerPhi(const ZZ& N)
  {
    if (N <= 0)
      CoCoA_ERROR(ERR::BadArg, "EulerPhi(N):  N must be strictly positive");
    const factorization<ZZ> facpows = factor(N);
    const vector<ZZ>& primes = facpows.myFactors;

    ZZ ans = abs(N);
    int NumPrimes = primes.size();
    for (int i=0; i < NumPrimes; ++i)
    {
      const ZZ p = primes[i];
      ans = (ans/p)*(p-1);
    }
    return ans;
  }



  // Taken from Cohen's book (page 24).
  long MultiplicativeOrderModPrime(unsigned long residue, unsigned long p)
  {
    CoCoA_ASSERT(IsPrime(p));
    CoCoA_ASSERT(p-1 <= NumericLimits<unsigned long>::MaxSquarableInteger());
    CoCoA_ASSERT(gcd(residue, p) == 1);

    if (p ==  2) return 1;
    const factorization<long> facpows = factor(p-1);
    const vector<long>& q = facpows.myFactors;
    const vector<long>& pwr = facpows.myExponents;
    const int n = q.size();
    long e = p-1;
    for (int i=0; i < n; ++i)
    {
      e /= SmallPower(q[i], pwr[i]);
      unsigned long rpower = PowerMod(residue, e, p);
      while (rpower != 1)
      {
        rpower = PowerMod(rpower, q[i], p);
        e *= q[i];
      }
    }
    return e;
  }

  long MultiplicativeOrderModPrimePower(unsigned long residue, unsigned long p, int e)
  {
    CoCoA_ASSERT(e > 0);
    CoCoA_ASSERT(residue >= 1 && residue < p);
    CoCoA_ASSERT(IsPrime(p));
    CoCoA_ASSERT(p-1 <= NumericLimits<unsigned long>::MaxSquarableInteger());
    long ord =  MultiplicativeOrderModPrime(residue, p);
    if (e == 1) return ord;
    const unsigned long q = SmallPower(p,e);
    unsigned long rpower = PowerMod(residue, ord, q);
    while (rpower != 1)
    {
      rpower = PowerMod(rpower, p, q);
      ord *= p;
    }
    return ord;
  }


  // Taken from Cohen's book (page 24).
  ZZ MultiplicativeOrderModPrime(const ZZ& residue, const ZZ& p)
  {
    CoCoA_ASSERT(IsPrime(p));
    CoCoA_ASSERT(gcd(residue, p) == 1);

// ???BUG NYI: use machine int code if possible???
    if (p == 2) return ZZ(1);
    const factorization<ZZ> facpows = factor(p-1);
    const vector<ZZ>& q = facpows.myFactors;
    const vector<long>& pwr = facpows.myExponents;
    const int n = q.size();
    ZZ e = p-1;
    for (int i=0; i < n; ++i)
    {
      e /= power(q[i], pwr[i]);
      ZZ rpower = PowerMod(residue, e, p);
      while (rpower != 1)
      {
        rpower = PowerMod(rpower, q[i], p);
        e *= q[i];
      }
    }
    return e;
  }

  ZZ MultiplicativeOrderModPrimePower(const ZZ& residue, const ZZ& p, long e)
  {
    CoCoA_ASSERT(e > 0);
    CoCoA_ASSERT(residue >= 1 && residue < p);
    CoCoA_ASSERT(IsPrime(p));
    ZZ ord =  MultiplicativeOrderModPrime(residue, p);
    if (e == 1) return ord;
    const ZZ q = power(p,e);
    ZZ rpower = PowerMod(residue, ord, q);
    while (rpower != 1)
    {
      rpower = PowerMod(rpower, p, q);
      ord *= p;
    }
    return ord;
  }


  long MultiplicativeOrderMod(const MachineInteger& residue, const MachineInteger& modulus)
  {
    if (IsNegative(modulus) || abs(modulus) < 2)
      CoCoA_ERROR(ERR::BadArg, "MultiplicativeOrderMod: must have modulus >= 2");
    if (gcd(residue, modulus) != 1)
      CoCoA_ERROR(ERR::BadArg, "MultiplicativeOrderMod: residue must be coprime to modulus");
    const unsigned long m = abs(modulus);
    if (m-1 > NumericLimits<unsigned long>::MaxSquarableInteger())
    {
      // return ConvertTo<long>(MultiplicativeOrderMod(residue, ZZ(modulus)));
      long ans;
      convert(ans, MultiplicativeOrderMod(residue, ZZ(modulus)));
      return ans;
    }

    unsigned long r = abs(residue)%m;
    if (r != 0 && IsNegative(residue)) r = m-r;

    const factorization<long> mfacs = factor(m);
    const vector<long>& factor = mfacs.myFactors;
    const vector<long>& exponent = mfacs.myExponents;
    const int n = factor.size();
    long ord = 1;
    for (int i=0; i < n; ++i)
    {
      unsigned long facpow = SmallPower(factor[i], exponent[i]);
      ord = lcm(ord, MultiplicativeOrderModPrimePower(r%facpow, factor[i], exponent[i]));
    }
    return ord;
  }

  long MultiplicativeOrderMod(const ZZ& residue, const MachineInteger& modulus)
  {
    return MultiplicativeOrderMod(residue%abs(modulus), modulus);
  }

  ZZ MultiplicativeOrderMod(const MachineInteger& residue, const ZZ& modulus)
  {
    return MultiplicativeOrderMod(ZZ(residue), modulus);
  }

  ZZ MultiplicativeOrderMod(const ZZ& residue, const ZZ& modulus)
  {
    if (modulus < 2)
      CoCoA_ERROR(ERR::BadArg, "MultiplicativeOrderMod: must have modulus >= 2");
    if (gcd(residue, modulus) != 1)
      CoCoA_ERROR(ERR::BadArg, "MultiplicativeOrderMod: residue must be coprime to modulus");
    unsigned long m;
    if (IsConvertible(m, modulus) && m-1 <= NumericLimits<unsigned long>::MaxSquarableInteger())
    {
      return ZZ(MultiplicativeOrderMod(residue%m, m));
    }

    const factorization<ZZ> mfacs = factor(modulus);
    const vector<ZZ>& factor = mfacs.myFactors;
    const vector<long>& exponent = mfacs.myExponents;
    const int n = factor.size();
    ZZ ord(1);
    for (int i=0; i < n; ++i)
    {
      ZZ facpow = power(factor[i], exponent[i]);
      ord = lcm(ord, MultiplicativeOrderModPrimePower(residue%facpow, factor[i], exponent[i]));
    }
    return ord;
  }

  // ZZ version of MultiplicativeOrder???  Could be VERY SLOW!!!

  int PrimitiveRoot(const MachineInteger& pp)
  {
    unsigned long p = abs(pp);
    if (IsZero(pp) || IsNegative(pp) || !IsPrime(p))
      CoCoA_ERROR(ERR::BadArg, "PrimitiveRoot(p):  p must be a (positive) prime");

    if (p == 2) return 1;
    const factorization<long> facpows = factor(p-1);
    const vector<long>& primes = facpows.myFactors;
    for (unsigned long root = 2; /*empty*/; ++root)
    {
      CoCoA_ASSERT(root < p);
      if (root == 4 || root == 8 || root == 9) continue; // skip the first few prime powers
      bool failed = false;
      for (vector<long>::const_iterator it = primes.begin(); it != primes.end(); ++it)
      {
	if (PowerMod(root, (p-1)/(*it), p) == 1)  // need brackets around *it to avoid "start of comment"
        { failed = true; break; } // effectively a "continue;" for the outer loop
      }
      if (!failed) return root;
    }
  }

  // Essentially identical to the function above.
  // !!!WARNING: calls factor(P-1) so could be VERY SLOW!!!
  int PrimitiveRoot(const ZZ& P)
  {
    if (P <= 0 || !IsPrime(P))
      CoCoA_ERROR(ERR::BadArg, "PrimitiveRoot(P):  P must be a (positive) prime");

    if (P == 2) return 1;
    const factorization<ZZ> facpows = factor(P-1);
    const vector<ZZ>& primes = facpows.myFactors;
    for (unsigned long root = 2; /*empty*/; ++root)
    {
      CoCoA_ASSERT(root < P);
      if (root == 4 || root == 8 || root == 9) continue; // skip the first few prime powers
      bool failed = false;
      for (vector<ZZ>::const_iterator it = primes.begin(); it != primes.end(); ++it)
      {
	if (PowerMod(root, (P-1)/(*it), P) == 1)  // need brackets around *it to avoid "start of comment"
        { failed = true; break; } // effectively a "continue;" for the outer loop
      }
      if (!failed) return root;
    }
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/NumTheory.C,v 1.12 2010/03/05 21:37:58 abbott Exp $
// $Log: NumTheory.C,v $
// Revision 1.12  2010/03/05 21:37:58  abbott
// Completed implementation of MultiplicativeOrderMod.
//
// Revision 1.11  2010/03/03 15:02:09  abbott
// Corrected bug in PowerMod: forgot to give error for negative power of non-invertible base.
// Added tests of PowerMod to test-NumTheory1.C (and consequent change to expected output).
//
// Revision 1.10  2010/03/03 10:43:34  abbott
// Added PrimitiveRoot for big primes (might be very slow).
// Added MultiplicativeOrderMod (currently very SLUGGY implementation).
//
// Revision 1.9  2009/12/29 22:44:32  abbott
// Removed buggy proxy class ZZ::rtn.
// Consequent changes for function prototypes also in NumTheory.
// Corrected some minor buglets in NumTheory.
//
// Revision 1.8  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.7  2009/12/03 17:41:47  abbott
// Minor correction to a comment.
//
// Revision 1.6  2009/10/08 13:39:47  abbott
// Renamed "round" into "RoundDiv".
// Added some new versions of "RoundDiv".
// Added a test for "RoundDiv".
//
// Revision 1.5  2009/09/29 12:26:13  abbott
// Added patch/workaround for a problem in gmp-4.3.1, so that exgcd gives
// the right answers (gmp-4.3.1 no longer produces reduced cofactors).
//
// Revision 1.4  2009/07/06 12:30:48  abbott
// Fixed incorrect assertion in IsBigPrime.
// Corrected a bug in SmoothFactor (had > instead of >=), and tidied up a little.
//
// Revision 1.3  2009/07/02 16:28:10  abbott
// Fairly extensive change to NumTheory (at least internally and philosophically).
// Improved and cleaned NumTheory.  Added documentation.
// Clarified the exact behaviour of most functions.
//
// Revision 1.2  2009/06/11 14:10:58  abbott
// Added commented out procedural forms for gcd/lcm, in case we should
// later want to activate them.
//
// Revision 1.1  2009/06/05 12:14:55  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
//
