//   Copyright (c)  2001-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "CoCoA/PPMonoid.H"

#include "CoCoA/OpenMath.H"
#include "CoCoA/PPMonoidEvOv.H" //  for NewPPMonoid
#include "CoCoA/convert.H"
#include "CoCoA/degree.H"
#include "CoCoA/error.H"
#include "CoCoA/symbol.H"

// #include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
// #include <vector>
using std::vector;

namespace CoCoA
{

  namespace // for functions local to this file/compilation unit.
  {
    inline void CheckCompatible(const ConstRefPPMonoidElem& pp1, const ConstRefPPMonoidElem& pp2, const char* const fn)
    {
      //      std::clog << "owner(pp1) = " << owner(pp1) << " and owner(pp2) = " << owner(pp2) << std::endl;
      if (owner(pp1) != owner(pp2))
        CoCoA_ERROR(ERR::MixedPPMs, fn);
    }
  }


  inline PPMonoidElemRawPtr PPMonoidBase::myNewCheckVecSize(const std::vector<long>& v) const
  {
    if (v.size() != myNumIndets())
      CoCoA_ERROR(ERR::BadArraySize, "PPMonoidElem(PPM, ExpVec)  or  PPMonoidBase::myNewCheckVecSize");
    return myNew(v);
  }


  // Default definition of a virtual function.
  bool PPMonoidBase::myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myCmp(rawpp1, rawpp2) == 0;
  }


  std::ostream& operator<<(std::ostream& out, const PPMonoid& PPM)
  {
    PPM->myOutputSelf(out);
    return out;
  }


  // Default impl, uses small exp version or throws ERR::ArgTooBig
  void PPMonoidBase::myMulIndetPower(RawPtr rawpp, std::size_t var, const ZZ& EXP) const
  {
    CoCoA_ASSERT(EXP >= 0);
    unsigned long exp;
    if (!IsConvertible(exp, EXP))
      CoCoA_ERROR(ERR::ArgTooBig, "IndetPower/myMulIndetPower");
    myMulIndetPower(rawpp, var, exp);
  }


  // Assume inputs are mathematically valid (i.e. exp >= 0).
  // Deal with all trivial cases; pass other cases to myPowerSmallExp.
  void PPMonoidBase::myPower(RawPtr rawpp, ConstRawPtr rawpp1, unsigned long exp) const
  {
    if (exp==0 || myIsOne(rawpp1))
    {
      myAssignOne(rawpp);
      return;
    }
    myPowerSmallExp(rawpp, rawpp1, exp);
  }

  // Assume inputs are mathematically valid (i.e. exp >= 0).
  // Deal with all trivial cases; pass other cases to myPowerSmallExp or myPowerBigExp.
  void PPMonoidBase::myPower(RawPtr rawpp, ConstRawPtr rawpp1, const ZZ& EXP) const
  {
    if (IsZero(EXP) || myIsOne(rawpp1))
    {
      myAssignOne(rawpp);
      return;
    }
    unsigned long exp;
    if (IsConvertible(exp, EXP))
      myPowerSmallExp(rawpp, rawpp1, exp);
    else
      myPowerBigExp(rawpp, rawpp1, EXP);
  }


  // Default defn of virtual fn: just gives an error.
  void PPMonoidBase::myPowerBigExp(RawPtr /*rawpp*/, ConstRawPtr /*rawpp1*/, const ZZ& /*EXP*/) const
  {
    CoCoA_ERROR(ERR::ExpTooBig, "power(pp,N)");
  }


  // Generic implementation -- default defn of virtual function.
  bool PPMonoidBase::myIsIndetPosPower(std::size_t& indet, ZZ& EXP, ConstRawPtr rawpp) const
  {
    CoCoA_ASSERT(!myIsOne(rawpp));
    const size_t NumIndets = myNumIndets();
    vector<long> expv(NumIndets);  // SLUG wasteful new+delete ???
    myExponents(expv, rawpp);
    // ??? it should be myZZExponents 2010-02-02
    indet = 0;
    for (size_t i = 1; i < NumIndets; ++i)
    {
      if (expv[i] == 0) continue;
      if (expv[indet] != 0) return false;
      indet = i;
    }
    EXP = expv[indet];
    return true;
  }


  bool PPMonoidBase::myIsIndetPosPower(std::size_t& indet, long& pow, ConstRawPtr rawpp) const
  {
    CoCoA_ASSERT(!myIsOne(rawpp));
    const size_t NumIndets = myNumIndets();
    vector<long> expv(NumIndets);  // SLUG wasteful new+delete ???
    myExponents(expv, rawpp);
    indet = 0;
    for (size_t i = 1; i < NumIndets; ++i)
    {
      if (expv[i] == 0) continue;
      if (expv[indet] != 0) return false;
      indet = i;
    }
    pow = expv[indet];
    return true;
  }


  // Generic PP printing routine.
  void PPMonoidBase::myOutput(std::ostream& out, ConstRawPtr rawpp) const
  {
    bool all0 = true;
    ZZ d;
    for (size_t indet=0; indet < myNumIndets(); ++indet)
    {
      myZZExponent(d, rawpp, indet); // Genericity is more important than efficiency here.
      if (IsZero(d)) continue;
      if (!all0) out << "*";
      all0 = false;
      out << myIndetName(indet);
      if (d > 1) out << "^" << d;
      else if (d < 0) out << "^(" << d << ")"; // ...in case we ever allow negative exponents.
    }
    if (all0) out << "1";
  }


  void PPMonoidBase::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawpp) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("cocoa", "PPExponents");
    ZZ d;
    for (size_t indet=0; indet < myNumIndets(); ++indet)
    {
      myZZExponent(d, rawpp, indet); // Genericity is more important than efficiency here.
      OMOut << d;
    }
    OMOut->mySendApplyEnd();
  }


  // Makes a wasteful copy of its result.
  std::vector<symbol> symbols(const PPMonoid& PPM)
  {
    vector<symbol> ans;
    PPM->mySymbols(ans);
    return ans;
  }


  //----------------------------------------------------------------------
  // For fairly obvious reasons assignment (i.e. operator=) does not permit
  // polymorphism on the destination type (e.g. to avoid slicing); so I must
  // define these operators in addition to those for PPMonoidElem.
  // All four operators should have identical definitions: bear this in mind
  // should you ever want to change one of them!

  RefPPMonoidElem& RefPPMonoidElem::operator=(const RefPPMonoidElem& rhs)
  {
    CheckCompatible(*this, rhs, "PPMonoidElem::operator=");
    // if (this == &rhs) return *this;  // trivial check not really needed
    myPPM->myAssign(myPPPtr, raw(rhs));
    return *this;
  }


  RefPPMonoidElem& RefPPMonoidElem::operator=(const ConstRefPPMonoidElem& rhs)
  {
    CheckCompatible(*this, rhs, "PPMonoidElem::operator=");
    // if (this == &rhs) return *this;  // trivial check not really needed
    myPPM->myAssign(myPPPtr, raw(rhs));
    return *this;
  }


  //----------------------------------------------------------------------
  // Implementation of class PPMonoidElem below...

  //-------------------- constructors & destructor --------------------//


  PPMonoidElem::PPMonoidElem(const PPMonoid& PPM):
      RefPPMonoidElem(PPM, PPM->myNew())
  {}


  PPMonoidElem::PPMonoidElem(const PPMonoid& PPM, const std::vector<long>& v):
      RefPPMonoidElem(PPM, PPM->myNewCheckVecSize(v))
  {}



  // This function assumes onwership of the value pointed to by ToBeOwned
  PPMonoidElem::PPMonoidElem(const PPMonoid& PPM, PPMonoidElemRawPtr rawToBeOwned):
      RefPPMonoidElem(PPM, rawToBeOwned)
  {}


  // This is a copy ctor.
  PPMonoidElem::PPMonoidElem(const PPMonoidElem& copy):
      RefPPMonoidElem(owner(copy), owner(copy)->myNew(raw(copy)))
  {}


  // This is a copy ctor.
  PPMonoidElem::PPMonoidElem(const ConstRefPPMonoidElem& copy):
      RefPPMonoidElem(owner(copy), owner(copy)->myNew(raw(copy)))
  {}


  // These assignment operators are replicated for RefPPMonoidElem because
  // C++ cannot safely allow polymorphism in the lhs (e.g. slicing troubles).
  // All four operators should have identical definitions: bear this in mind
  // should you ever want to change one of them!

  PPMonoidElem& PPMonoidElem::operator=(const PPMonoidElem& rhs)
  {
    CheckCompatible(*this, rhs, "PPMonoidElem::operator=");
    // if (this == &rhs) return *this;  // trivial check not really needed
    myPPM->myAssign(myPPPtr, raw(rhs));
    return *this;
  }

  PPMonoidElem& PPMonoidElem::operator=(const ConstRefPPMonoidElem& rhs)
  {
    CheckCompatible(*this, rhs, "PPMonoidElem::operator=");
    // if (this == &rhs) return *this;  // trivial check not really needed
    myPPM->myAssign(myPPPtr, raw(rhs));
    return *this;
  }


  PPMonoidElem::~PPMonoidElem()
  {
    myPPM->myDelete(myPPPtr);
  }



  long StdDeg(ConstRefPPMonoidElem pp)
  {
    return owner(pp)->myStdDeg(raw(pp));
  }

  long deg(ConstRefPPMonoidElem pp)
  {
    return owner(pp)->myStdDeg(raw(pp));
  }


  degree wdeg(ConstRefPPMonoidElem pp)
  {
    degree ans(GradingDim(ordering(owner(pp))));
    owner(pp)->myWDeg(ans, raw(pp));
    return ans;
  }


  int CmpWDeg(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)
  {
    CheckCompatible(pp1, pp2, "CmpWDeg(PP,PP)");
    return owner(pp1)->myCmpWDeg(raw(pp1), raw(pp2));
  }


  int CmpWDegPartial(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2, std::size_t PartialGradingDim)
  {
    CheckCompatible(pp1, pp2, "CmpWDeg(PP,PP)");
    if ( PartialGradingDim > GradingDim(owner(pp1)) )
      CoCoA_ERROR("PartialGradingDim > GradingDim", "CmpWDegPartial");
    return owner(pp1)->myCmpWDegPartial(raw(pp1), raw(pp2), PartialGradingDim);
  }


  long exponent(ConstRefPPMonoidElem pp, std::size_t indet) // degree in indet
  {
    if (indet >= NumIndets(owner(pp)))
      CoCoA_ERROR(ERR::BadIndetIndex, "exponent(PP)");
    return owner(pp)->myExponent(raw(pp), indet);
  }


  ZZ ZZExponent(ConstRefPPMonoidElem pp, std::size_t indet)
  {
    if (indet >= NumIndets(owner(pp)))
      CoCoA_ERROR(ERR::BadIndetIndex, "ZZExponent(PP)");
    ZZ ans;
    owner(pp)->myZZExponent(ans, raw(pp), indet);
    return ans;
  }


  void exponents(std::vector<long>& expv, ConstRefPPMonoidElem pp)
  {
    expv.resize(NumIndets(owner(pp)));
//  Should either resize expv, or make sure expv is of the right size.
//    if (expv.size() != NumIndets(owner(pp)))
//      CoCoA_ERROR(ERR::BadArraySize, "exponents(expv, pp)");
    owner(pp)->myExponents(expv, raw(pp));
  }


  void swap(RefPPMonoidElem pp1, RefPPMonoidElem pp2) // swap(pp1, pp2);
  {
    CheckCompatible(pp1, pp2, "swap(PP,PP)");
    owner(pp1)->mySwap(raw(pp1), raw(pp2));
  }


  bool IsOne(ConstRefPPMonoidElem pp)
  {
    return owner(pp)->myIsOne(raw(pp));
  }


  bool IsIndet(std::size_t& index, ConstRefPPMonoidElem pp)
  {
    return owner(pp)->myIsIndet(index, raw(pp));
  }


  bool IsIndetPosPower(std::size_t& index, ZZ& EXP, ConstRefPPMonoidElem pp)
  {
    if (IsOne(pp)) CoCoA_ERROR(ERR::BadArg, "IsIndetPosPower(var, EXP, pp)");
    return owner(pp)->myIsIndetPosPower(index, EXP, raw(pp));
  }


  // ------------------------------ Comparisons ------------------------------

  int cmp(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)                  // <0, =0, >0 as pp1 < = > pp2
  {
    CheckCompatible(pp1, pp2, "cmp(PP,PP)");
    return owner(pp1)->myCmp(raw(pp1), raw(pp2));
  }


  bool operator==(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)          // pp1 == pp2
  {
    CheckCompatible(pp1, pp2, "PP == PP");
    return owner(pp1)->myIsEqual(raw(pp1), raw(pp2));
  }


  bool operator!=(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)          // pp1 != pp2
  {
    CheckCompatible(pp1, pp2, "PP != PP");
    return !owner(pp1)->myIsEqual(raw(pp1), raw(pp2));
  }


  bool operator<(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)           // pp1 < pp2
  {
    CheckCompatible(pp1, pp2, "PP < PP");
    return owner(pp1)->myCmp(raw(pp1), raw(pp2)) < 0;
  }


  bool operator<=(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)          // pp1 <= pp2
  {
    CheckCompatible(pp1, pp2, "PP <= PP");
    return owner(pp1)->myCmp(raw(pp1), raw(pp2)) <= 0;
  }


  bool operator>(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)           // pp1 > pp2
  {
    CheckCompatible(pp1, pp2, "PP > PP");
    return owner(pp1)->myCmp(raw(pp1), raw(pp2)) > 0;
  }


  bool operator>=(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)          // pp1 >= pp2
  {
    CheckCompatible(pp1, pp2, "PP >= PP");
    return owner(pp1)->myCmp(raw(pp1), raw(pp2)) >= 0;
  }


// ------------------------------ Arithmetic ------------------------------

  PPMonoidElem operator*(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)   //  pp1*pp2;
  {
    CheckCompatible(pp1, pp2, "PP * PP");
    PPMonoidElem ans(owner(pp1));
    owner(pp1)->myMul(raw(ans), raw(pp1), raw(pp2));
    return ans;
  }


  PPMonoidElem operator/(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)   // pp1/pp2;
  {
    CheckCompatible(pp1, pp2, "PP / PP");
    if (!owner(pp1)->myIsDivisible(raw(pp1), raw(pp2)))
      CoCoA_ERROR(ERR::BadQuot, "PP / PP");
    PPMonoidElem ans(owner(pp1));
    owner(pp1)->myDiv(raw(ans), raw(pp1), raw(pp2));
    return ans;
  }


  PPMonoidElem colon(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)       //  pp1:pp2;
  {
    CheckCompatible(pp1, pp2, "colon(PP, PP)");
    PPMonoidElem ans(owner(pp1));
    owner(pp1)->myColon(raw(ans), raw(pp1), raw(pp2));
    return ans;
  }


  PPMonoidElem gcd(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)         // gcd(pp1,pp2);
  {
    CheckCompatible(pp1, pp2, "gcd(PP, PP)");
    PPMonoidElem ans(owner(pp1));
    owner(pp1)->myGcd(raw(ans), raw(pp1), raw(pp2));
    return ans;
  }


  PPMonoidElem lcm(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)         // lcm(pp1,pp2);
  {
    CheckCompatible(pp1, pp2, "lcm(PP, PP)");
    PPMonoidElem ans(owner(pp1));
    owner(pp1)->myLcm(raw(ans), raw(pp1), raw(pp2));
    return ans;
  }


  PPMonoidElem radical(ConstRefPPMonoidElem pp)         // radical(pp)
  {
    PPMonoidElem ans(owner(pp));
    owner(pp)->myRadical(raw(ans), raw(pp));
    return ans;
  }


  PPMonoidElem power(ConstRefPPMonoidElem pp, long exp)                        // pp^exp
  {
    if (exp < 0)
      CoCoA_ERROR(ERR::NegExp, "power(pp, n) and n < 0");
    PPMonoidElem ans(owner(pp));
    owner(pp)->myPower(raw(ans), raw(pp), exp);
    return ans;
  }


  PPMonoidElem power(ConstRefPPMonoidElem pp, const ZZ& EXP)                 // pp^EXP
  {
    if (EXP < 0)
      CoCoA_ERROR(ERR::NegExp, "power(pp, N) and N < 0");
    PPMonoidElem ans(owner(pp));
    owner(pp)->myPower(raw(ans), raw(pp), EXP);
    return ans;
  }


  bool IsCoprime(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)  // are pp1, pp2 coprime?
  {
    CheckCompatible(pp1, pp2, "IsCoprime(PP, PP)");
    return owner(pp1)->myIsCoprime(raw(pp1), raw(pp2));
  }


  bool IsDivisible(ConstRefPPMonoidElem pp1, ConstRefPPMonoidElem pp2)
  {
    CheckCompatible(pp1, pp2, "IsDivisible(PP, PP)");
    return owner(pp1)->myIsDivisible(raw(pp1), raw(pp2));
  }


  bool IsRadical(ConstRefPPMonoidElem pp)
  {
    return owner(pp)->myIsRadical(raw(pp));
  }


  void AssignOne(RefPPMonoidElem dest)
  {
    owner(dest)->myAssignOne(raw(dest));
  }


  RefPPMonoidElem operator*=(RefPPMonoidElem pp, ConstRefPPMonoidElem pp1)
  {
    CheckCompatible(pp, pp1, "PP *= PP");
    owner(pp)->myMul(raw(pp), raw(pp), raw(pp1));
    return pp;
  }


  RefPPMonoidElem operator/=(RefPPMonoidElem pp, ConstRefPPMonoidElem pp1)
  {
    CheckCompatible(pp, pp1, "PP /= PP");
    owner(pp)->myDiv(raw(pp), raw(pp), raw(pp1));
    return pp;
  }


  ConstRefPPMonoidElem indet(const PPMonoid& M, std::size_t index)
  {
    if (index >= NumIndets(M)) CoCoA_ERROR(ERR::BadIndetIndex, "indet(PPMonoid, var)");
    return indets(M)[index];
  }


  PPMonoidElem IndetPower(const PPMonoid& M, std::size_t indet, long exp)
  {
    if (exp < 0) CoCoA_ERROR(ERR::NegExp, "IndetPower(PPMonoid, indet, exp)");
    if (indet >= NumIndets(M)) CoCoA_ERROR(ERR::BadIndetIndex, "IndetPower(PPMonoid, indet, exp)");
    PPMonoidElem ans(M);
    M->myMulIndetPower(raw(ans), indet, exp); // exp becomes unsigned long in this call.
    return ans;
  }

  PPMonoidElem IndetPower(const PPMonoid& M, std::size_t indet, const ZZ& EXP)
  {
    if (EXP < 0) CoCoA_ERROR(ERR::NegExp, "IndetPower(PPMonoid, indet, EXP)");
    if (indet >= NumIndets(M)) CoCoA_ERROR(ERR::BadIndetIndex, "IndetPower(PPMonoid, indet, EXP)");
    PPMonoidElem ans(M);
    M->myMulIndetPower(raw(ans), indet, EXP);
    return ans;
  }


//   // Possible default implementation -- would wastefully alloc/free though
//   void PPMonoidElem::myComputeDivMask(DivMask::value& dm, const DivMask::base& DivMaskImpl, const RawPtr& pp) const
//   {
//     std::vector<long> v(myNumIndets);

//     myExponents(v, pp);
//     DivMaskImpl.myAssignFromExpv(dm, &v[0], myNumIndets());
//   }

// ------------------------------ Input/Output ------------------------------

  std::ostream& operator<<(std::ostream& out, ConstRefPPMonoidElem pp)
  {
    owner(pp)->myOutput(out, raw(pp));
    return out;
  }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, ConstRefPPMonoidElem pp)
  {
    owner(pp)->myOutput(OMOut, raw(pp));
    return OMOut;
  }


  PPMonoid NewPPMonoid(const std::vector<symbol>& IndetNames, const PPOrdering& ord)
  {
    return NewPPMonoidEvOv(IndetNames, ord); // use EvOv by default
  }

  PPMonoid NewPPMonoid(const std::vector<symbol>& IndetNames, const PPOrderingCtor& ord)
  {
    return NewPPMonoidEvOv(IndetNames, ord.myCtor(IndetNames.size())); // use EvOv by default
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/PPMonoid.C,v 1.14 2010/02/03 16:13:52 abbott Exp $
// $Log: PPMonoid.C,v $
// Revision 1.14  2010/02/03 16:13:52  abbott
// Added new single word tags for specifying the ordering in PPMonoid
// pseudo-ctors.
//
// Revision 1.13  2010/02/03 14:12:34  bigatti
// -- added IsIndetPosPower with exponent of type long
//
// Revision 1.12  2010/02/02 16:44:31  abbott
// Added radical & IsRadical (via mem fns myRadical & myIsRadical)
// for PPMonoidElems.
//
// Revision 1.11  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.10  2009/10/29 18:31:50  abbott
// Changed order of include directives (now alphabetical by file name).
//
// Revision 1.9  2009/09/22 14:01:33  bigatti
// -- added myCmpWDegPartial (ugly name, I know....)
// -- cleaned up and realigned code in PPMonoid*.C files
//
// Revision 1.8  2008/10/07 12:23:43  abbott
// Added deg function (same as StdDeg).
//
// Revision 1.7  2008/09/16 14:03:39  bigatti
// -- added comment
//
// Revision 1.6  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.5  2007/12/04 14:27:07  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.4  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/09/24 14:21:29  abbott
// Added default defn of myMulIndetPower.
// Added IndetPower.
// Renamed myIsIndetPower to myIsIndetPosPower.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.10  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.9  2007/03/08 17:43:11  cocoa
// Swapped order of args to the NewPPMonoid pseudo ctors.
//
// Revision 1.8  2007/03/07 13:44:49  bigatti
// -- minor cleanup for -Wextra
//
// Revision 1.7  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.6  2006/12/06 17:34:51  cocoa
// -- style: RawPtr args are now called "raw.."
//
// Revision 1.5  2006/11/24 17:04:32  cocoa
// -- reorganized includes of header files
//
// Revision 1.4  2006/11/23 17:35:48  cocoa
// -- changed: PPMonoid is now a class (instead of typedef)
//
// Revision 1.3  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.2  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.11  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.10  2006/04/21 14:59:22  cocoa
// Changed return type of indet(PPM, i) to ConstRefPPMonoidElem
// instead of PPMonoidElem (which made a copy).
//
// Revision 1.9  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.8  2006/03/14 17:21:18  cocoa
// Moved concrete PPMonoid impls entirely into their respective .C files.
// Now the corresponding .H files are very compact.
//
// Revision 1.7  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.6  2006/03/07 10:08:48  cocoa
// -- fixed: PPMonoidElem(const ConstRefPPMonoidElem& copy);  [added const]
//
// Revision 1.5  2006/02/20 22:41:20  cocoa
// All forms of the log function for power products now return SmallExponent_t
// (instead of int).  exponents now resizes the vector rather than requiring
// the user to pass in the correct size.
//
// Revision 1.4  2006/02/13 12:08:04  cocoa
// Fixed a problem with some missing assignment ops for certain PPMonoidElems.
// Fixed a bug in RingDistrMPoly::myIndetPower.
//
// Revision 1.3  2006/02/01 16:56:13  cocoa
// Added some missing assignment operators for (Ref)PPMonoidElems.
//
// Revision 1.2  2005/11/17 15:43:04  cocoa
// -- added ctor PPMonoidElem(ConstRefPPMonoidElem& copy);
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.7  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.6  2005/07/08 15:09:28  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//
// Revision 1.5  2005/07/01 16:08:15  cocoa
// Friday check-in.  Major change to structure under PolyRing:
// now SparsePolyRing and DUPolyRing are separated (in preparation
// for implementing iterators).
//
// A number of other relatively minor changes had to be chased through
// (e.g. IndetPower).
//
// Revision 1.4  2005/06/27 14:55:24  cocoa
// Cleaned up some more PPMonoid code.
//
// Revision 1.3  2005/06/23 15:42:41  cocoa
// Fixed typo in GNU fdl -- all doc/*.txt files affected.
// Minor corrections to PPMonoid (discovered while writing doc).
//
// Revision 1.2  2005/06/22 14:47:56  cocoa
// PPMonoids and PPMonoidElems updated to mirror the structure
// used for rings and RingElems.  Many consequential changes.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.3  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.2  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.11  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.10  2004/11/11 13:40:39  cocoa
// -- minor changes for doxygen
//
// Revision 1.9  2004/11/02 14:56:33  cocoa
// -- changed *Print* into *Output* (myPrint --> myOutput)
// -- changed *Var* into *Indet* (myPrintVarName --> myOutputIndetName)
// -- removed suffix "IgnoreDivMask"
// -- added myComputeDivMask
// -- improved storing of IndetNames
// -- changed ExpvElem into SmallExponent_t
//
// Revision 1.8  2004/10/21 17:16:37  cocoa
// Fairly major change: new OrdvArith namspace with various members,
//   new global typedef  SmallExponent_t (defined in config.H).
//
// Revision 1.7  2004/06/29 17:10:22  cocoa
// Partially tidied use of "protected" and "private" in various
// base classes.  Checking in at the end of the day -- it works,
// and I wouldn't want it to be lost next time point's disk
// misbehaves.
//
// Revision 1.6  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.5  2004/05/24 15:52:14  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.4  2004/03/20 17:46:10  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.3  2004/01/28 15:20:37  cocoa
// Added CmpDeg function and myCmpDeg member function.
// Aligned name of data member of PPMonoid with the convention used
// for the data member of ring.
//
// Revision 1.2  2003/10/01 10:35:31  cocoa
// - applied "my" coding convention to PPMonoid and PPOrdering
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
// Revision 1.17  2003/09/23 17:01:51  bigatti
// - added #include "CoCoA/PPMonoidDiv.H"
// - changed default monoid to PPMonoidDiv
//
// Revision 1.16  2003/09/22 13:04:24  abbott
// Improved definition of operator!= and tidied a few comments.
// Something odd had happened: the corrected version was write
// protected but not checked in.
//
// Revision 1.15  2003/06/23 16:57:08  abbott
// Minor cleaning prior to public release.
//
// Revision 1.14  2003/05/14 16:41:41  abbott
// Added new functions: IsOne and PPMonoidBase::IsEqualIgnoreDivMask
// (default definition of a virtual functions).
// Revitalized the function exponents.
//
// Revision 1.13  2003/04/29 13:59:37  abbott
// Uh oh.  Lots of changes.  Checking in prior to restructuring
// to align with new PPOrdering code.
//
// Revision 1.12  2003/01/15 16:03:50  abbott
// Checking in prior to major change.
// Minor changes:
//   new definition of deg (to go with new signature),
//   better error messages in fns on PPMonoid::elems,
//   better exception safety in fns on PPMonoid::elems.
//
// Revision 1.11  2002/11/15 17:07:21  bigatti
// - added check of myDivMask in operators == and !=
//
// Revision 1.10  2002/11/15 16:25:41  abbott
// MAJOR CHANGE reflecting a new "philosophy".
// The new philosophy is that PPMonoid::elems will be "slow and safe" since
// they are now higly decoupled from the internal representation of
// polynomials (which use "order vectors").
// Eliminated HalfPP, alias and temp subclasses.
// Every RawPP now contains its own "divisibility mask" because Anna needs
// the DivMask check to be inline.
//
// Revision 1.9  2002/06/22 17:16:16  abbott
// Numerous changes to accommodate the new idea of PPOrdering.
// And some other changes (e.g. NumVars is now NumIndets).
//
// Revision 1.8  2002/03/06 11:24:19  abbott
// Un-inlined the destructor for PPMonoid::elem (but shouldn't
// it be inline in the header file?).
//
// Revision 1.7  2002/02/08 12:07:48  abbott
// MAJOR CHANGE: added two new proxy classes "alias" and "temp" (see PP.txt)
// Numerous consequential changes.  Changed order of some definitions.
//
// Revision 1.6  2002/01/30 14:21:06  abbott
// Tidied up use of "using" declaration.
// Added definition of PPMonoid::print -- all concrete derived classes had
// effectively the same implementation.  This is the only non-pure virtual
// function in the class PPMonoid.
//
// Revision 1.5  2001/12/04 19:53:32  abbott
// Names changed in accordance with the new coding conventions.
//
// Revision 1.4  2001/11/19 15:25:47  abbott
// Corrected an imcompatibility between PPmonoid::cmp and
// bool operator<(const PPmonoid::elem&, const PPmonoid::elem&).
// The latter expected cmp to return one of -1,0 or +1 whereas
// cmp actually returned negative, zero, or positive.
//
// Revision 1.3  2001/11/16 18:41:58  bigatti
// added:   using namespace std;
// for compatibility with gcc-3
//
// Revision 1.2  2001/10/31 20:42:57  abbott
// Pseudo copy constructor for PPmonoid::elem now takes a const copy as
// argument (previous const was omitted).
//
// Revision 1.1  2001/10/25 17:54:55  abbott
// Initial revision
//

