//   Copyright (c)  2004-2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Implementation of class PPMonoidEvImpl

#include "CoCoA/PPMonoidEv.H"
#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/degree.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/symbol.H"
#include "CoCoA/DivMask.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/PPOrdering.H"
#include "CoCoA/ZZ.H"

#include <algorithm>
using std::min;
using std::max;
//using std::swap;
using std::back_inserter;
#include <cstddef>
using std::size_t;
#include <cstring>
using std::memcpy;
#include <iostream>
using std::ostream;
#include<limits>
using std::numeric_limits;
#include <memory>
using std::auto_ptr;
#include <vector>
using std::vector;


namespace CoCoA
{

  /*-- class PPMonoidEvImpl ---------------------------------------*/
  /**

  \brief Implementation power product monoid with exponent vector

  PPMonoidEvImpl implements a power product monoid for safe testing.
  It stores the exponents as a C vector of SmallExponent_t.
  It is not designed to be efficient (but it might be ;-)

  */
  /*-----------------------------------------------------------------*/

  class PPMonoidEvImpl: public PPMonoidBase
  {
    typedef PPMonoidElemRawPtr RawPtr;           // just to save typing
    typedef PPMonoidElemConstRawPtr ConstRawPtr; // just to save typing

    class CmpBase
    {
    public:
      CmpBase(std::size_t NumIndets, std::size_t GradingDim);
      virtual ~CmpBase() {};
    public:
      virtual int myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const =0;
      virtual void myWDeg(degree& d, const SmallExponent_t* v) const =0;
      //      virtual int myCmpWDeg(const SmallExponent_t* v1, const SmallExponent_t* v2) const =0;
      virtual int myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t) const =0;
      int myCmpWDeg(const SmallExponent_t* v1, const SmallExponent_t* v2) const { return myCmpWDegPartial(v1, v2, myGradingDimValue); }
      
    protected:
      std::size_t myNumIndetsValue;
      std::size_t myGradingDimValue;
    };

    class LexImpl: public CmpBase
    {
    public:
      LexImpl(std::size_t NumIndets);
      virtual int myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const;
      virtual void myWDeg(degree& d, const SmallExponent_t* v) const;
      virtual int myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t) const;
    };


    class StdDegLexImpl: public CmpBase
    {
    public:
      StdDegLexImpl(std::size_t NumIndets);
      virtual int myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const;
      virtual void myWDeg(degree& d, const SmallExponent_t* v) const;
      virtual int myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t) const;
    };


    class StdDegRevLexImpl: public CmpBase
    {
    public:
      StdDegRevLexImpl(std::size_t NumIndets);
      virtual int myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const;
      virtual void myWDeg(degree& d, const SmallExponent_t* v) const;
      virtual int myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t) const;
    };


    class MatrixOrderingImpl: public CmpBase
    {
    public:
      MatrixOrderingImpl(std::size_t NumIndets, std::size_t GradingDim, const matrix& OrderMatrix);
      virtual int myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const;
      virtual void myWDeg(degree& d, const SmallExponent_t* v) const;
      virtual int myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t) const;
    private:
      std::vector< std::vector<ZZ> > myOrderMatrix;
    };


  public:
    PPMonoidEvImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord);
    ~PPMonoidEvImpl();
  private: // disable copy ctor and assignment
    explicit PPMonoidEvImpl(const PPMonoidEvImpl& copy);  // NEVER DEFINED -- copy ctor disabled
    PPMonoidEvImpl& operator=(const PPMonoidEvImpl& rhs); // NEVER DEFINED -- assignment disabled

  public:
    void contents() const; // FOR DEBUGGING ONLY

    std::size_t myNumIndets() const;
    const std::vector<PPMonoidElem>& myIndets() const;               ///< std::vector whose n-th entry is n-th indet as PPMonoidElem
    void mySymbols(std::vector<symbol>& SymList) const;              ///< appends PPM's symbols to SymList
    const symbol& myIndetName(std::size_t indet) const;

    // The functions below are operations on power products owned by PPMonoidEvImpl
    const PPMonoidElem& myOne() const;
    PPMonoidElemRawPtr myNew() const;                                ///< ctor from nothing
    PPMonoidElemRawPtr myNew(PPMonoidElemConstRawPtr rawpp) const;   ///< ctor by assuming ownership
    PPMonoidElemRawPtr myNew(const std::vector<long>& expv) const;   ///< ctor from exp vector
//NYI    PPMonoidElemRawPtr myNew(const std::vector<ZZ>& EXPV) const;///< ctor from exp vector
    void myDelete(RawPtr rawpp) const;                               ///< dtor, frees pp
    void mySwap(RawPtr rawpp1, RawPtr rawpp2) const;                 ///< swap(pp1, pp2);
    void myAssignOne(RawPtr rawpp) const;                            ///< pp = 1
    void myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const;           ///< p = pp1
    void myAssign(RawPtr rawpp, const std::vector<long>& expv) const;///< pp = expv (assign from exp vector)

    void myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = pp1*pp2
    void myMulIndetPower(RawPtr rawpp, std::size_t indet, unsigned long exp) const;///< pp *= indet^exp
    void myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = pp1/pp2
    void myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1/gcd(pp1,pp2)
    void myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = gcd(pp1,pp2)
    void myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = lcm(pp1,pp2)
    void myRadical(RawPtr rawpp, ConstRawPtr rawpp1) const;                   ///< pp = radical(pp1)
    void myPowerSmallExp(RawPtr rawpp, ConstRawPtr rawpp1, unsigned long exp) const;///< pp = pp1^exp;

    bool myIsOne(ConstRawPtr rawpp) const;                                   ///< is pp = 1?
    bool myIsIndet(std::size_t& index, ConstRawPtr rawpp) const;             ///< true iff pp is an indet
    bool myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;          ///< are pp1 & pp2 coprime?
    bool myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;            ///< is pp1 equal to pp2?
    bool myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;        ///< does pp2 divide pp1?
    bool myIsRadical(ConstRawPtr rawpp) const;                               ///< is pp equal to its radical?

    int myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;                 ///< -1,0,1 as pp1 < = > pp2
    long myStdDeg(ConstRawPtr rawpp) const;                                  ///< standard degree of pp
    void myWDeg(degree& d, ConstRawPtr rawpp) const;                         ///< d = grading(pp)
    int myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;             ///< <0, =0, >0 as wdeg(pp1) < = > wdeg(pp2)
    int myCmpWDegPartial(ConstRawPtr rawpp1, ConstRawPtr rawpp2, std::size_t) const; ///< as myCmpWDeg wrt the first weights
    long myExponent(ConstRawPtr rawpp, std::size_t indet) const;             ///< exponent of indet in pp
    void myZZExponent(ZZ& EXP, ConstRawPtr rawpp, std::size_t indet) const;  ///< EXP = exponent of indet in pp
    void myExponents(std::vector<long>& expv, ConstRawPtr rawpp) const;      ///< expv[i] = exponent(pp,i)
    void myOutputSelf(std::ostream& out) const;                              ///< out << PPM
    // INHERITED DEFINITION of virtual  void myOutput(std::ostream& out, ConstRawPtr rawpp) const;
    void myDebugPrint(std::ostream& out, ConstRawPtr rawpp) const;           ///< print pp in debugging format???


  private: // auxiliary functions
    SmallExponent_t* myExpv(RawPtr) const;
    const SmallExponent_t* myExpv(ConstRawPtr) const;

    void myComputeDivMask(DivMask& dm, const DivMaskRule& DivMaskImpl, ConstRawPtr rawpp) const; ///< used by PPWithMask
    bool myCheckExponents(const std::vector<long>& expv) const;

  private: // data members
    ///@name Class members
    //@{
    const size_t myNumIndetsValue;
    const size_t myEntrySize;     ///< size in ?bytes???
    mutable MemPool myMemMgr;     // IMPORTANT: this must come *before* myIndetVector and myOnePtr.
    vector<PPMonoidElem> myIndetVector; ///< the indets as PPMonoidElems
    vector<symbol> myIndetNames;        ///< the names of the indeterminates
    auto_ptr<CmpBase> myOrdPtr;         ///< actual implementation of the ordering [should be const???]
    auto_ptr<PPMonoidElem> myOnePtr;
    //@}
  };


  // File local inline functions

  inline SmallExponent_t* PPMonoidEvImpl::myExpv(RawPtr rawpp) const
  {
    return static_cast<SmallExponent_t*>(rawpp.myRawPtr());
  }


  inline const SmallExponent_t* PPMonoidEvImpl::myExpv(ConstRawPtr rawpp) const
  {
    return static_cast<const SmallExponent_t*>(rawpp.myRawPtr());
  }


  bool PPMonoidEvImpl::myCheckExponents(const std::vector<long>& expv) const
  {
    // Check expv.size == myNumIndetsValue.
    // Check exps are non-neg and not too big.
    if (expv.size() != myNumIndetsValue) return false;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      if (expv[i] < 0 || static_cast<unsigned long>(expv[i]) > numeric_limits<SmallExponent_t>::max()) return false;
    return true;
  }


  //----   Constructors & destructor   ----//

  PPMonoidEvImpl::PPMonoidEvImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord):
      PPMonoidBase(ord),
      myNumIndetsValue(NumIndets(ord)),
      myEntrySize(sizeof(SmallExponent_t)*NumIndets(ord)),
      myMemMgr(myEntrySize, "PPMonoidEvImpl.myMemMgr"),
      myIndetVector(),
      myIndetNames(IndetNames)
  {
    // Sanity check on the indet names given.
    CoCoA_ASSERT(myNumIndetsValue == myIndetNames.size());
    CoCoA_ASSERT(AreDistinct(myIndetNames));
    CoCoA_ASSERT(AreArityConsistent(myIndetNames));

    // Put the correct implementation in myOrdPtr...  [VERY UGLY CODE!!!]
    if (IsLex(ord))
      myOrdPtr.reset(new LexImpl(myNumIndetsValue));
    else if (IsStdDegLex(ord))
      myOrdPtr.reset(new StdDegLexImpl(myNumIndetsValue));
    else if (IsStdDegRevLex(ord))
      myOrdPtr.reset(new StdDegRevLexImpl(myNumIndetsValue));
    else
      myOrdPtr.reset(new MatrixOrderingImpl(myNumIndetsValue, GradingDim(ord), GetMatrix(ord)));

    myOnePtr.reset(new PPMonoidElem(PPMonoid(this)));
    {
      // IMPORTANT: this block destroys pp *before* the call to myRefCountZero.
      PPMonoidElem pp(PPMonoid(this));
      vector<long> expv(myNumIndetsValue);
      for (size_t i=0; i < myNumIndetsValue; ++i)
      {
        expv[i] = 1;
        myAssign(raw(pp), expv);
        myIndetVector.push_back(pp);
        expv[i] = 0;
      }
    }
    myRefCountZero();
  }


  PPMonoidEvImpl::~PPMonoidEvImpl()
  {}

/////////////////////////////////////////////////////////////////////////////



  std::size_t PPMonoidEvImpl::myNumIndets() const
  {
    return myNumIndetsValue;
  }


  const std::vector<PPMonoidElem>& PPMonoidEvImpl::myIndets() const
  {
    return myIndetVector;
  }


  void PPMonoidEvImpl::mySymbols(std::vector<symbol>& SymList) const
  {
    std::copy(myIndetNames.begin(), myIndetNames.end(), back_inserter(SymList));
  }


  const symbol& PPMonoidEvImpl::myIndetName(std::size_t indet) const
  {
    return myIndetNames[indet];
  }


  const PPMonoidElem& PPMonoidEvImpl::myOne() const
  {
    return *myOnePtr;
  }


  PPMonoidElemRawPtr PPMonoidEvImpl::myNew() const
  {
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    myAssignOne(rawpp); // cannot throw
    return rawpp;
  }

  PPMonoidElemRawPtr PPMonoidEvImpl::myNew(PPMonoidElemConstRawPtr rawcopypp) const
  {
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    myAssign(rawpp, rawcopypp); // cannot throw
    return rawpp;
  }


  PPMonoidElemRawPtr PPMonoidEvImpl::myNew(const std::vector<long>& expv) const
  {
    CoCoA_ASSERT(myCheckExponents(expv));
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    myAssign(rawpp, expv); // cannot throw
    return rawpp;
  }


  void PPMonoidEvImpl::myAssignOne(RawPtr rawpp) const
  {
    SmallExponent_t* const expv = myExpv(rawpp);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = 0;
  }


  void PPMonoidEvImpl::myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const
  {
    if (rawpp == rawpp1) return;

//     SmallExponent_t* const expv = myExpv(rawpp);
//     const SmallExponent_t* const expv1 = myExpv(rawpp1);
//     std::copy(expv1, expv1+myNumIndetsValue, expv);
    memcpy(myExpv(rawpp), myExpv(rawpp1), myNumIndetsValue*sizeof(SmallExponent_t));
  }

  void PPMonoidEvImpl::myAssign(RawPtr rawpp, const vector<long>& expv1) const
  {
    CoCoA_ASSERT(myCheckExponents(expv1));

    SmallExponent_t* const expv = myExpv(rawpp);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = NumericCast<SmallExponent_t>(expv1[i]);
  }


  void PPMonoidEvImpl::myDelete(RawPtr rawpp) const
  {
    myMemMgr.free(rawpp.myRawPtr());
  }


  void PPMonoidEvImpl::mySwap(RawPtr rawpp1, RawPtr rawpp2) const
  {
    if (rawpp1 == rawpp2) return;
    SmallExponent_t* const expv1 = myExpv(rawpp1);
    SmallExponent_t* const expv2 = myExpv(rawpp2);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      std::swap(expv1[i], expv2[i]);
  }


  void PPMonoidEvImpl::myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
    {
      CoCoA_ASSERT("Exponent Overflow" && expv1[i] <= std::numeric_limits<SmallExponent_t>::max()-expv2[i]);
      expv[i] = expv1[i] + expv2[i];
    }
  }


  void PPMonoidEvImpl::myMulIndetPower(RawPtr rawpp, std::size_t indet, unsigned long exp) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    SmallExponent_t* const expv = myExpv(rawpp);
    // If CoCoA_DEBUG active, check for exponent overflow...
    CoCoA_ASSERT("Exponent Overflow" && static_cast<unsigned long>(numeric_limits<SmallExponent_t>::max()-expv[indet]) >= exp);
    expv[indet] += static_cast<SmallExponent_t>(exp);  // cast to keep M$ compiler quiet
  }


  void PPMonoidEvImpl::myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
    {
      CoCoA_ASSERT("Exponent Underflow" && expv1[i] >= expv2[i]);
      expv[i] = expv1[i] - expv2[i];
    }
  }


  void PPMonoidEvImpl::myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] > expv2[i])
        expv[i] = expv1[i] - expv2[i];
      else
        expv[i] = 0;
  }


  void PPMonoidEvImpl::myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = min(expv1[i], expv2[i]);
  }


  void PPMonoidEvImpl::myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = max(expv1[i], expv2[i]);
  }


  void PPMonoidEvImpl::myRadical(RawPtr rawpp, ConstRawPtr rawpp1) const
  {
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = (expv1[i] > 0);
  }


  void PPMonoidEvImpl::myPowerSmallExp(RawPtr rawpp, ConstRawPtr rawpp1, unsigned long exp) const
  {
    SmallExponent_t* const expv = myExpv(rawpp);
    const SmallExponent_t* const expv1 = myExpv(rawpp1);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
    {
      // If CoCoA_DEBUG active, check for exponent overflow...
      CoCoA_ASSERT("Exponent Overflow" && (expv1[i]==0 || static_cast<unsigned long>(numeric_limits<SmallExponent_t>::max()/expv1[i]) >= exp));
      expv[i] = expv1[i] * static_cast<SmallExponent_t>(exp); // cast to keep M$ compiler quiet
    }
  }


  bool PPMonoidEvImpl::myIsOne(ConstRawPtr rawpp) const
  {
    const SmallExponent_t* const expv = myExpv(rawpp);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv[i] != 0) return false;

    return true;
  }


  bool PPMonoidEvImpl::myIsIndet(std::size_t& index, ConstRawPtr rawpp) const
  {
    const SmallExponent_t* const expv = myExpv(rawpp);
    size_t j = myNumIndetsValue;
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
    {
      if (expv[i] == 0) continue;
      if (j != myNumIndetsValue || expv[i] != 1) return false;
      j = i;
    }
    if (j == myNumIndetsValue) return false;
    index = j;
    return true;
  }


  bool PPMonoidEvImpl::myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] != 0 && expv2[i] != 0) return false;

    return true;
  }


  bool PPMonoidEvImpl::myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] != expv2[i]) return false;
    return true;
  }


  bool PPMonoidEvImpl::myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const SmallExponent_t* const expv1 = myExpv(rawpp1);
    const SmallExponent_t* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] < expv2[i]) return false;

    return true;
  }


  bool PPMonoidEvImpl::myIsRadical(ConstRawPtr rawpp) const
  {
    const SmallExponent_t* const expv = myExpv(rawpp);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv[i] > 1) return false;

    return true;
  }


  int PPMonoidEvImpl::myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myOrdPtr->myCmpExpvs(myExpv(rawpp1), myExpv(rawpp2));
  }


  long PPMonoidEvImpl::myStdDeg(ConstRawPtr rawpp) const
  {
    const SmallExponent_t* const expv = myExpv(rawpp);
    size_t d=0;
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
      d += expv[i];
    return d;
  }


  void PPMonoidEvImpl::myWDeg(degree& d, ConstRawPtr rawpp) const
  {
    myOrdPtr->myWDeg(d, myExpv(rawpp));
  }


  int PPMonoidEvImpl::myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myOrdPtr->myCmpWDeg(myExpv(rawpp1), myExpv(rawpp2));
  }


  int PPMonoidEvImpl::myCmpWDegPartial(ConstRawPtr rawpp1, ConstRawPtr rawpp2, std::size_t i) const
  {
    return myOrdPtr->myCmpWDegPartial(myExpv(rawpp1), myExpv(rawpp2), i);
  }


  long PPMonoidEvImpl::myExponent(ConstRawPtr rawpp, std::size_t indet) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    return NumericCast<long>(myExpv(rawpp)[indet]);
  }

  void PPMonoidEvImpl::myZZExponent(ZZ& EXP, ConstRawPtr rawpp, std::size_t indet) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    EXP = myExpv(rawpp)[indet];
  }


  void PPMonoidEvImpl::myExponents(vector<long>& v, ConstRawPtr rawpp) const
  {
    CoCoA_ASSERT(v.size() == myNumIndetsValue);
    const SmallExponent_t* const expv = myExpv(rawpp);
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
      v[i] = NumericCast<long>(expv[i]);
  }


  void PPMonoidEvImpl::myComputeDivMask(DivMask& dm, const DivMaskRule& DivMaskImpl, ConstRawPtr rawpp) const
  {
    DivMaskImpl->myAssignFromExpv(dm, myExpv(rawpp), myNumIndetsValue);
  }


  void PPMonoidEvImpl::myOutputSelf(std::ostream& out) const
  {
    out << "PPMonoidEv(" << myNumIndetsValue << ", " << myOrd <<")";
  }


  void PPMonoidEvImpl::myDebugPrint(std::ostream& out, ConstRawPtr rawpp) const
  {
    out << "DEBUG PP: myNumIndetsValue=" << myNumIndetsValue << ", exps=[";
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
      out << myExponent(rawpp, i) << " ";
    out << "]" << std::endl;
  }


  PPMonoid NewPPMonoidEv(const std::vector<symbol>& IndetNames, const PPOrdering& ord)
  {
    // Sanity check on the indet names given.
    const size_t nvars = NumIndets(ord);

    if (IndetNames.size() != nvars)
      CoCoA_ERROR(ERR::BadNumIndets, "NewPPMonoidEv(IndetNames,ord)");
    if (!AreDistinct(IndetNames))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidEv(IndetNames,ord)");
    if (!AreArityConsistent(IndetNames))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidEv(IndetNames,ord)");

    // Inefficient quadratic loop -- speed is probably not important.
    for (size_t i=0; i < nvars; ++i)
      for (size_t j=i+1; j < nvars; ++j)
        if (IndetNames[i] == IndetNames[j])
          CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidEvOv(IndetNames,ord)");

    return PPMonoid(new PPMonoidEvImpl(IndetNames, ord));
  }

  PPMonoid NewPPMonoidEv(const std::vector<symbol>& IndetNames, const PPOrderingCtor& ord)
  {
    return NewPPMonoidEv(IndetNames, ord.myCtor(IndetNames.size()));
  }


//--- orderings ----------------------------------------

  PPMonoidEvImpl::CmpBase::CmpBase(std::size_t NumIndets, size_t GradingDim):
    myNumIndetsValue(NumIndets),
    myGradingDimValue(GradingDim)
  {
    CoCoA_ASSERT(NumIndets > 0);
    CoCoA_ASSERT(NumIndets < 1000000); // complain about ridiculously large number of indets
    CoCoA_ASSERT(GradingDim <= NumIndets);
  }


//--- LexImpl

  PPMonoidEvImpl::LexImpl::LexImpl(std::size_t NumIndets):
    CmpBase(NumIndets, 0)
  {
  }


  int PPMonoidEvImpl::LexImpl::myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const
  {
    for (size_t i=0; i<myNumIndetsValue; ++i)
      if (v1[i] != v2[i])
      {
        if (v1[i] > v2[i]) return 1; else return -1;
      }
    return 0;
  }


  void PPMonoidEvImpl::LexImpl::myWDeg(degree& /*d*/, const SmallExponent_t* /*v*/) const
  {
    // deliberately does nothing because GradingDim=0
    //??? should it assign:  d = []
  }


  int PPMonoidEvImpl::LexImpl::myCmpWDegPartial(const SmallExponent_t* /*v1*/, const SmallExponent_t* /*v2*/, std::size_t /*i*/) const
  {
    return 0; // GradingDim=0, and all degrees in Z^0 are equal
  }

//--- StdDegLexImpl

  PPMonoidEvImpl::StdDegLexImpl::StdDegLexImpl(std::size_t NumIndets):
    CmpBase(NumIndets, 1)
  {
  }


  void PPMonoidEvImpl::StdDegLexImpl::myWDeg(degree& d, const SmallExponent_t* v) const
  {
    long deg=0;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      deg += v[i];
    SetComponent(d, 0, deg);
  }


  int PPMonoidEvImpl::StdDegLexImpl::myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const
  {
    long deg1=0, deg2=0;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2)
    {
      if (deg1 > deg2) return 1; else return -1;
    }

    for (size_t i=0; i < myNumIndetsValue; ++i)
      if (v1[i] != v2[i])
      {
        if (v1[i] > v2[i]) return 1; else return -1;
      }
    return 0;
  }


  int PPMonoidEvImpl::StdDegLexImpl::myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t i) const
  {
    if (i==0) return 0; // ie: wrt to 0-grading
    long deg1=0, deg2=0;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2)
    {
      if (deg1 > deg2) return 1; else return -1;
    }
    return 0;
  }


//--- StdDegRevLexImpl

  PPMonoidEvImpl::StdDegRevLexImpl::StdDegRevLexImpl(std::size_t NumIndets):
    CmpBase(NumIndets, 1)
  {
  }


  void PPMonoidEvImpl::StdDegRevLexImpl::myWDeg(degree& d, const SmallExponent_t* v) const
  {
    long deg=0;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      deg += v[i];
    SetComponent(d, 0, deg);
  }


  int PPMonoidEvImpl::StdDegRevLexImpl::myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const
  {
    long deg1=0, deg2=0;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2)
    {
      if (deg1 > deg2) return 1; else return -1;
    }

    for (size_t i = myNumIndetsValue-1; i>0; --i)
      if (v1[i] != v2[i])
      {
        if (v1[i] > v2[i]) return -1; else return 1;
      }
    return 0;
  }


  int PPMonoidEvImpl::StdDegRevLexImpl::myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, size_t i) const
  {
    if (i==0) return 0; // ie: wrt to 0-grading
    long deg1=0, deg2=0;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2)
    {
      if (deg1 > deg2) return 1; else return -1;
    }
    return 0;
  }


//--- MatrixOrderingImpl

  PPMonoidEvImpl::MatrixOrderingImpl::MatrixOrderingImpl(std::size_t NumIndets, std::size_t GradingDim, const matrix& OrderMatrix):
    CmpBase(NumIndets, GradingDim)
  {
    CoCoA_ASSERT(NumRows(OrderMatrix) == NumIndets);
    CoCoA_ASSERT(NumCols(OrderMatrix) == NumIndets);

    myOrderMatrix.resize(NumIndets, vector<ZZ>(NumIndets));
    for (size_t i=0; i < NumIndets; ++i)
      for (size_t j=0; j < NumIndets; ++j)
        IsInteger(myOrderMatrix[i][j], OrderMatrix(i,j));
  }


  void PPMonoidEvImpl::MatrixOrderingImpl::myWDeg(degree& d, const SmallExponent_t* v) const
  {
    ZZ deg;
    for (size_t g=0; g < myGradingDimValue; ++g)
    {
      deg = 0;
      for (size_t i=0; i < myNumIndetsValue; ++i)
        deg += v[i] * myOrderMatrix[g][i];
      SetComponent(d, g, deg);
    }
  }


  static int MatrixOrderingCmpArrays(const vector<vector<ZZ> >& M, const SmallExponent_t* v1, const SmallExponent_t* v2, size_t UpTo)
  {
    ZZ s1, s2; // automatically set to 0
    for (size_t r=0; r < UpTo; ++r)
    {
      for (size_t i=0; i < M[1].size(); ++i)
      {
        s1 += v1[i]*M[r][i];
        s2 += v2[i]*M[r][i];
      }
      if (s1 != s2)
      {
        if (s1 > s2) return 1; else return -1;
      }
      s1 = 0;
      s2 = 0;
    }
    return 0;
  }


  int PPMonoidEvImpl::MatrixOrderingImpl::myCmpExpvs(const SmallExponent_t* v1, const SmallExponent_t* v2) const
  {
    return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, myNumIndetsValue);
  }


//   int PPMonoidEvImpl::MatrixOrderingImpl::myCmpWDeg(const SmallExponent_t* v1, const SmallExponent_t* v2) const
//   {
//     return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, myGradingDimValue);
//   }


  int PPMonoidEvImpl::MatrixOrderingImpl::myCmpWDegPartial(const SmallExponent_t* v1, const SmallExponent_t* v2, std::size_t i) const
  {
    return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, i);
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/PPMonoidEv.C,v 1.15 2010/02/03 16:13:52 abbott Exp $
// $Log: PPMonoidEv.C,v $
// Revision 1.15  2010/02/03 16:13:52  abbott
// Added new single word tags for specifying the ordering in PPMonoid
// pseudo-ctors.
//
// Revision 1.14  2010/02/02 16:44:31  abbott
// Added radical & IsRadical (via mem fns myRadical & myIsRadical)
// for PPMonoidElems.
//
// Revision 1.13  2009/12/23 18:53:52  abbott
// Major change to conversion functions:
//   convert(..) is now a procedure instead of a function
//   IsConvertible(..) replaces the former convert(..) function
//   Added new NumericCast conversion function (placeholder for BOOST feature)
//   Consequent changes in code which uses these features.
//
// Revision 1.12  2009/09/22 14:01:33  bigatti
// -- added myCmpWDegPartial (ugly name, I know....)
// -- cleaned up and realigned code in PPMonoid*.C files
//
// Revision 1.11  2009/07/02 16:34:18  abbott
// Minor change to keep the compiler quiet.
//
// Revision 1.10  2008/03/26 16:52:04  abbott
// Added exponent overflow checks (also for ordvs) when CoCoA_DEBUG is active.
//
// Revision 1.9  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.8  2007/12/04 14:27:07  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.7  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.6  2007/09/25 16:32:30  abbott
// Several minor changes to silence gcc-4.3:
//    more #includes,
//    and fixed a template problemm in RegisterServerOps.C
//
// Revision 1.5  2007/05/31 14:54:31  bigatti
// -- now using AreDistinct and AreArityConsistent for sanity check on
//    indet names
//
// Revision 1.3  2007/05/03 10:35:23  abbott
// Added new PPMonoidEvZZ with (virtually) unlimited exponents.
// Modified test-PPMonoid1.C accordingly.
// Added warning in doc about silent/unchecked exponent overflow in other
// PPMonoids.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.15  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.14  2007/03/08 17:43:11  cocoa
// Swapped order of args to the NewPPMonoid pseudo ctors.
//
// Revision 1.13  2007/03/08 11:07:12  cocoa
// Made pseudo ctors for polynomial rings more uniform.  This allowed me to
// remove an include of CoCoA/symbol.H  from the RingDistrM*.H files, but then
// I had to put the include in several .C files.
//
// Revision 1.12  2006/12/07 17:23:46  cocoa
// -- for compilation with _Wextra: commented out names of unused arguments
//
// Revision 1.11  2006/12/06 17:35:58  cocoa
// -- style: RawPtr args are now called "raw.."
//
// Revision 1.10  2006/11/27 13:06:23  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.9  2006/11/24 17:04:32  cocoa
// -- reorganized includes of header files
//
// Revision 1.8  2006/11/23 17:39:11  cocoa
// -- added #include
//
// Revision 1.7  2006/11/16 11:27:20  cocoa
// -- reinserted myRefCountZero(): sometimes really necessary, in general safe
//
// Revision 1.6  2006/11/14 17:29:20  cocoa
// -- commented out myRefCountZero() (not necessary???)
//
// Revision 1.5  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.4  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.3  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.2  2006/06/21 17:07:10  cocoa
// Fixed IsIndet bug -- why are there three almost identical copies of code?
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.9  2006/04/27 13:45:30  cocoa
// Changed name of NewIdentityRingHom to NewIdentityHom.
// Changed name of member functions which print out their own object
// into myOutputSelf (to distinguish from "transitive" myOutput fns).
//
// Revision 1.8  2006/03/27 12:21:25  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.7  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.6  2006/03/14 17:21:18  cocoa
// Moved concrete PPMonoid impls entirely into their respective .C files.
// Now the corresponding .H files are very compact.
//
// Revision 1.5  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.4  2006/02/20 22:41:20  cocoa
// All forms of the log function for power products now return SmallExponent_t
// (instead of int).  exponents now resizes the vector rather than requiring
// the user to pass in the correct size.
//
// Revision 1.3  2006/01/17 10:23:08  cocoa
// Updated DivMask; many consequential changes.
// A few other minor fixes.
//
// Revision 1.2  2005/11/17 13:25:13  cocoa
// -- "unsigned int" --> "SmallExponent_t" in ctor PPMonoidEvImpl(o, IndetNames)
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.9  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.8  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.7  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.6  2005/07/08 15:09:28  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//
// Revision 1.5  2005/07/01 16:08:15  cocoa
// Friday check-in.  Major change to structure under PolyRing:
// now SparsePolyRing and DUPolyRing are separated (in preparation
// for implementing iterators).
//
// A number of other relatively minor changes had to be chased through
// (e.g. IndetPower).
//
// Revision 1.4  2005/06/23 15:42:41  cocoa
// Fixed typo in GNU fdl -- all doc/*.txt files affected.
// Minor corrections to PPMonoid (discovered while writing doc).
//
// Revision 1.3  2005/06/22 14:47:56  cocoa
// PPMonoids and PPMonoidElems updated to mirror the structure
// used for rings and RingElems.  Many consequential changes.
//
// Revision 1.2  2005/05/04 16:50:31  cocoa
// -- new code for MatrixOrderingImpl
// -- changed: CmpBase also requires GradingDim
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.4  2005/04/29 15:42:02  cocoa
// Improved documentation for GMPAllocator.
// Added example program for GMPAllocator.
// Added example program for simple ops on polynomials.
// Added two new ctors for (principal) ideals (from long, and from ZZ).
// Added (crude) printing for PPMonoids.
// Updated library.H (#included GMPAllocator.H).
//
// Revision 1.3  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.2  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.3  2004/11/25 16:14:21  cocoa
// (1) Fixed definition of specialization of std::swap template function
//     so that it compiles with gcc 3.4.3
// (2) Implemented monomial function for polynomial rings.
// (3) Added one(PPM) and PPM->myOne() functions.
//
// Revision 1.2  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.1  2004/11/11 13:45:22  cocoa
// -- renamed: was PPMonoidSafe
//
// Revision 1.3  2004/11/03 17:52:05  cocoa
// -- added just a warning to remind me to implement matrix ordering
//
// Revision 1.2  2004/11/02 14:56:33  cocoa
// -- changed *Print* into *Output* (myPrint --> myOutput)
// -- changed *Var* into *Indet* (myPrintVarName --> myOutputIndetName)
// -- removed suffix "IgnoreDivMask"
// -- added myComputeDivMask
// -- improved storing of IndetNames
// -- changed ExpvElem into SmallExponent_t
//
// Revision 1.1  2004/10/29 15:31:25  cocoa
// -- new PPMonoid for compatibility with OrdvArith (without DivMask)
//
