//   Copyright (c)  2007-2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Implementation of class PPMonoidEvZZImpl

#include "CoCoA/PPMonoidEvZZ.H"
#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/degree.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/symbol.H"
#include "CoCoA/DivMask.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/PPOrdering.H"
#include "CoCoA/ZZ.H"

#include <algorithm>
using std::min;
using std::max;
//using std::swap;
using std::back_inserter;
#include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
#include<limits>
using std::numeric_limits;
#include <memory>
using std::auto_ptr;
#include <vector>
using std::vector;


namespace CoCoA
{

  /*-- class PPMonoidEvZZImpl ---------------------------------------*/
  /**

  \brief Implementation power product monoid with exponent vector

  PPMonoidEvZZImpl implements a power product monoid for safe testing.
  It stores the exponents as a std::vector<ZZ>.
  It is not designed to be efficient.

  */
  /*-----------------------------------------------------------------*/

  class PPMonoidEvZZImpl: public PPMonoidBase
  {
    typedef PPMonoidElemRawPtr RawPtr;           // just to save typing
    typedef PPMonoidElemConstRawPtr ConstRawPtr; // just to save typing

    class CmpBase
    {
    public:
      CmpBase(std::size_t NumIndets, std::size_t GradingDim);
      virtual ~CmpBase() {};
    public:
      virtual int myCmpExpvs(const ZZ* v1, const ZZ* v2) const =0;
      virtual void myWDeg(degree& d, const ZZ* v) const =0;
      virtual int myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t) const =0;
      //      virtual int myCmpWDeg(const ZZ* v1, const ZZ* v2) const =0;
      int myCmpWDeg(const ZZ* v1, const ZZ* v2) const { return myCmpWDegPartial(v1, v2, myGradingDimValue); }
    protected:
      std::size_t myNumIndetsValue;
      std::size_t myGradingDimValue;
    };

    class LexImpl: public CmpBase
    {
    public:
      LexImpl(std::size_t NumIndets);
      virtual int myCmpExpvs(const ZZ* v1, const ZZ* v2) const;
      virtual void myWDeg(degree& d, const ZZ* v) const;
      virtual int myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t) const;
    };


    class StdDegLexImpl: public CmpBase
    {
    public:
      StdDegLexImpl(std::size_t NumIndets);
      virtual int myCmpExpvs(const ZZ* v1, const ZZ* v2) const;
      virtual void myWDeg(degree& d, const ZZ* v) const;
      virtual int myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t) const;
    };


    class StdDegRevLexImpl: public CmpBase
    {
    public:
      StdDegRevLexImpl(std::size_t NumIndets);
      virtual int myCmpExpvs(const ZZ* v1, const ZZ* v2) const;
      virtual void myWDeg(degree& d, const ZZ* v) const;
      virtual int myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t) const;
    };


    class MatrixOrderingImpl: public CmpBase
    {
    public:
      MatrixOrderingImpl(std::size_t NumIndets, std::size_t GradingDim, const matrix& OrderMatrix);
      virtual int myCmpExpvs(const ZZ* v1, const ZZ* v2) const;
      virtual void myWDeg(degree& d, const ZZ* v) const;
      virtual int myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t) const;
    private:
      std::vector< std::vector<ZZ> > myOrderMatrix;
    };


  public:
    PPMonoidEvZZImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord);
    ~PPMonoidEvZZImpl();
  private: // disable copy ctor and assignment
    explicit PPMonoidEvZZImpl(const PPMonoidEvZZImpl& copy);  // NEVER DEFINED -- copy ctor disabled
    PPMonoidEvZZImpl& operator=(const PPMonoidEvZZImpl& rhs); // NEVER DEFINED -- assignment disabled

  public:
    void contents() const; // FOR DEBUGGING ONLY

    std::size_t myNumIndets() const;
    const std::vector<PPMonoidElem>& myIndets() const;               ///< std::vector whose n-th entry is n-th indet as PPMonoidElem
    void mySymbols(std::vector<symbol>& SymList) const;              ///< appends PPM's symbols to SymList
    const symbol& myIndetName(std::size_t indet) const;

    // The functions below are operations on power products owned by PPMonoidEvZZImpl
    const PPMonoidElem& myOne() const;
    PPMonoidElemRawPtr myNew() const;                                ///< ctor from nothing
    PPMonoidElemRawPtr myNew(PPMonoidElemConstRawPtr rawpp) const;   ///< ctor by assuming ownership
    PPMonoidElemRawPtr myNew(const std::vector<long>& expv) const;   ///< ctor from exp vector
//NYI    PPMonoidElemRawPtr myNew(const std::vector<ZZ>& EXPV) const;///< ctor from exp vector
    void myDelete(RawPtr rawpp) const;                               ///< dtor, frees pp
    void mySwap(RawPtr rawpp1, RawPtr rawpp2) const;                 ///< swap(pp1, pp2);
    void myAssignOne(RawPtr rawpp) const;                            ///< pp = 1
    void myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const;           ///< p = pp1
    void myAssign(RawPtr rawpp, const std::vector<long>& expv) const;///< pp = expv (assign from exp vector)

    void myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = pp1*pp2
    void myMulIndetPower(RawPtr rawpp, std::size_t indet, unsigned long exp) const;///< pp *= indet^exp
    void myMulIndetPower(RawPtr rawpp, std::size_t indet, const ZZ& EXP) const;///< pp *= indet^EXP
    void myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = pp1/pp2
    void myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1/gcd(pp1,pp2)
    void myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = gcd(pp1,pp2)
    void myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = lcm(pp1,pp2)
    void myRadical(RawPtr rawpp, ConstRawPtr rawpp1) const;                   ///< pp = radical(pp1)
    void myPowerSmallExp(RawPtr rawpp, ConstRawPtr rawpp1, unsigned long exp) const;///< pp = pp1^exp;

    bool myIsOne(ConstRawPtr rawpp) const;                              ///< is pp = 1?
    bool myIsIndet(std::size_t& index, ConstRawPtr rawpp) const;        ///< true iff pp is an indet
    bool myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;     ///< are pp1 & pp2 coprime?
    bool myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;       ///< is pp1 equal to pp2?
    bool myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< does pp2 divide pp1?
    bool myIsRadical(ConstRawPtr rawpp) const;                          ///< is pp equal to its radical?

    int myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;            ///< -1,0,1 as pp1 < = > pp2
    long myStdDeg(ConstRawPtr rawpp) const;                      ///< standard degree of pp
    void myWDeg(degree& d, ConstRawPtr rawpp) const;                    ///< d = grading(pp)
    int myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;        ///< <0, =0, >0 as wdeg(pp1) < = > wdeg(pp2)
    int myCmpWDegPartial(ConstRawPtr rawpp1, ConstRawPtr rawpp2, std::size_t) const; ///< as myCmpWDeg wrt the first weights
    long myExponent(ConstRawPtr rawpp, std::size_t indet) const;             ///< exponent of indet in pp
    void myZZExponent(ZZ& EXP, ConstRawPtr rawpp, std::size_t indet) const;  ///< EXP = exponent of indet in pp
    void myExponents(std::vector<long>& expv, ConstRawPtr rawpp) const;      ///< expv[i] = exponent(pp,i)
    void myOutputSelf(std::ostream& out) const;                      ///< out << PPM
    // INHERITED DEFINITION of virtual  void myOutput(std::ostream& out, ConstRawPtr rawpp) const;
    void myDebugPrint(std::ostream& out, ConstRawPtr rawpp) const;   ///< print pp in debugging format???


  private: // auxiliary functions
    ZZ* myExpv(RawPtr) const;
    const ZZ* myExpv(ConstRawPtr) const;

    void myComputeDivMask(DivMask& dm, const DivMaskRule& DivMaskImpl, ConstRawPtr rawpp) const; ///< used by PPWithMask
    bool myCheckExponents(const std::vector<long>& expv) const;

  private: // data members
    ///@name Class members
    //@{
    const std::size_t myNumIndetsValue;
    const size_t myEntrySize;     ///< size in ????
    mutable MemPool myMemMgr;     // IMPORTANT: this must come *before* myIndetVector and myOnePtr.
    std::vector<PPMonoidElem> myIndetVector; ///< the indets as PPMonoidElems
    std::vector<symbol> myIndetNames;        ///< the names of the indeterminates
    std::auto_ptr<CmpBase> myOrdPtr;         ///< actual implementation of the ordering [should be const???]
    std::auto_ptr<PPMonoidElem> myOnePtr;
    //@}
  };


  // File local inline functions

  inline ZZ* PPMonoidEvZZImpl::myExpv(RawPtr rawpp) const
  {
    return static_cast<ZZ*>(rawpp.myRawPtr());
  }


  inline const ZZ* PPMonoidEvZZImpl::myExpv(ConstRawPtr rawpp) const
  {
    return static_cast<const ZZ*>(rawpp.myRawPtr());
  }


  bool PPMonoidEvZZImpl::myCheckExponents(const std::vector<long>& expv) const
  {
    // Check expv.size == myNumIndetsValue.
    // Check exps are non-neg and not too big.
    if (expv.size() != myNumIndetsValue) return false;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      if (expv[i] < 0) return false;
    return true;
  }


  //----   Constructors & destructor   ----//

  PPMonoidEvZZImpl::PPMonoidEvZZImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord):
      PPMonoidBase(ord),
      myNumIndetsValue(NumIndets(ord)),
      myEntrySize(sizeof(ZZ)*NumIndets(ord)),
      myMemMgr(myEntrySize, "PPMonoidEvZZImpl.myMemMgr"),
      myIndetVector(),
      myIndetNames(IndetNames)
  {
    // Sanity check on the indet names given.
    CoCoA_ASSERT(myNumIndetsValue == myIndetNames.size());
    CoCoA_ASSERT(AreDistinct(myIndetNames));
    CoCoA_ASSERT(AreArityConsistent(myIndetNames));

    // Put the correct implementation in myOrdPtr...  [VERY UGLY CODE!!!]
    if (IsLex(ord))
      myOrdPtr.reset(new LexImpl(myNumIndetsValue));
    else if (IsStdDegLex(ord))
      myOrdPtr.reset(new StdDegLexImpl(myNumIndetsValue));
    else if (IsStdDegRevLex(ord))
      myOrdPtr.reset(new StdDegRevLexImpl(myNumIndetsValue));
    else
      myOrdPtr.reset(new MatrixOrderingImpl(myNumIndetsValue, GradingDim(ord), GetMatrix(ord)));

    myOnePtr.reset(new PPMonoidElem(PPMonoid(this)));
    {
      // IMPORTANT: this block destroys pp *before* the call to myRefCountZero.
      PPMonoidElem pp(PPMonoid(this));
      vector<long> expv(myNumIndetsValue);
      for (size_t i=0; i < myNumIndetsValue; ++i)
      {
        expv[i] = 1;
        myAssign(raw(pp), expv);
        myIndetVector.push_back(pp);
        expv[i] = 0;
      }
    }
    myRefCountZero();
  }


  PPMonoidEvZZImpl::~PPMonoidEvZZImpl()
  {}

/////////////////////////////////////////////////////////////////////////////



  std::size_t PPMonoidEvZZImpl::myNumIndets() const
  {
    return myNumIndetsValue;
  }


  const std::vector<PPMonoidElem>& PPMonoidEvZZImpl::myIndets() const
  {
    return myIndetVector;
  }


  void PPMonoidEvZZImpl::mySymbols(std::vector<symbol>& SymList) const
  {
    std::copy(myIndetNames.begin(), myIndetNames.end(), back_inserter(SymList));
  }


  const symbol& PPMonoidEvZZImpl::myIndetName(std::size_t indet) const
  {
    return myIndetNames[indet];
  }


  const PPMonoidElem& PPMonoidEvZZImpl::myOne() const
  {
    return *myOnePtr;
  }


  PPMonoidElemRawPtr PPMonoidEvZZImpl::myNew() const
  {
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    ZZ* expv = myExpv(rawpp);
    for (size_t i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv[i]) ZZ;
    myAssignOne(rawpp); // cannot throw
    return rawpp;
  }

  PPMonoidElemRawPtr PPMonoidEvZZImpl::myNew(PPMonoidElemConstRawPtr rawcopypp) const
  {
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    ZZ* expv = myExpv(rawpp);
    for (size_t i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv[i]) ZZ;
    myAssign(rawpp, rawcopypp); // cannot throw
    return rawpp;
  }


  PPMonoidElemRawPtr PPMonoidEvZZImpl::myNew(const std::vector<long>& expv) const
  {
    CoCoA_ASSERT(myCheckExponents(expv));
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    ZZ* expv1 = myExpv(rawpp);
    for (size_t i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv1[i]) ZZ;
    myAssign(rawpp, expv); // cannot throw
    return rawpp;
  }


  void PPMonoidEvZZImpl::myAssignOne(RawPtr rawpp) const
  {
    ZZ* const expv = myExpv(rawpp);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = 0;
  }


  void PPMonoidEvZZImpl::myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const
  {
    if (rawpp == rawpp1) return;

    ZZ* expv = myExpv(rawpp);
    const ZZ* expv1 = myExpv(rawpp1);
    std::copy(&expv1[0], &expv1[myNumIndetsValue], &expv[0]);
  }

  void PPMonoidEvZZImpl::myAssign(RawPtr rawpp, const vector<long>& expv1) const
  {
    CoCoA_ASSERT(myCheckExponents(expv1));

    ZZ* const expv = myExpv(rawpp);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = expv1[i];
  }


  void PPMonoidEvZZImpl::myDelete(RawPtr rawpp) const
  {
    ZZ* expv = myExpv(rawpp);;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      expv[i].~ZZ();
    myMemMgr.free(rawpp.myRawPtr());
  }


  void PPMonoidEvZZImpl::mySwap(RawPtr rawpp1, RawPtr rawpp2) const
  {
    if (rawpp1 == rawpp2) return;
    ZZ* expv1 = myExpv(rawpp1);
    ZZ* expv2 = myExpv(rawpp2);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      std::swap(expv1[i], expv2[i]);
  }


  void PPMonoidEvZZImpl::myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    ZZ* expv = myExpv(rawpp);
    const ZZ* expv1 = myExpv(rawpp1);
    const ZZ* expv2 = myExpv(rawpp2);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = expv1[i] + expv2[i];
  }


  void PPMonoidEvZZImpl::myMulIndetPower(RawPtr rawpp, std::size_t indet, unsigned long exp) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    ZZ* expv = myExpv(rawpp);
    expv[indet] += exp;
  }


  // This fn overrides the default defn in PPMonoid.C (which throws for big exps)
  void PPMonoidEvZZImpl::myMulIndetPower(RawPtr rawpp, std::size_t indet, const ZZ& EXP) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    CoCoA_ASSERT(EXP >= 0);
    ZZ* expv = myExpv(rawpp);
    expv[indet] += EXP;
  }


  void PPMonoidEvZZImpl::myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    ZZ* expv = myExpv(rawpp);
    const ZZ* expv1 = myExpv(rawpp1);
    const ZZ* expv2 = myExpv(rawpp2);
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = expv1[i] - expv2[i]; // assert that result is positive?
  }


  void PPMonoidEvZZImpl::myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    ZZ* const expv = myExpv(rawpp);
    const ZZ* const expv1 = myExpv(rawpp1);
    const ZZ* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] > expv2[i])
        expv[i] = expv1[i] - expv2[i];
      else
        expv[i] = 0;
  }


  void PPMonoidEvZZImpl::myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    ZZ* const expv = myExpv(rawpp);
    const ZZ* const expv1 = myExpv(rawpp1);
    const ZZ* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = min(expv1[i], expv2[i]);
  }


  void PPMonoidEvZZImpl::myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    ZZ* const expv = myExpv(rawpp);
    const ZZ* const expv1 = myExpv(rawpp1);
    const ZZ* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = max(expv1[i], expv2[i]);
  }


  void PPMonoidEvZZImpl::myRadical(RawPtr rawpp, ConstRawPtr rawpp1) const
  {
    ZZ* const expv = myExpv(rawpp);
    const ZZ* const expv1 = myExpv(rawpp1);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      expv[i] = (expv1[i] > 0);
  }


  void PPMonoidEvZZImpl::myPowerSmallExp(RawPtr rawpp, ConstRawPtr rawpp1, unsigned long exp) const
  {
    ZZ* const expv = myExpv(rawpp);
    const ZZ* const expv1 = myExpv(rawpp1);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
    {
      expv[i] = expv1[i]*exp;
    }
  }


  bool PPMonoidEvZZImpl::myIsOne(ConstRawPtr rawpp) const
  {
    const ZZ* const expv = myExpv(rawpp);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (!IsZero(expv[i])) return false;

    return true;
  }


  bool PPMonoidEvZZImpl::myIsIndet(std::size_t& index, ConstRawPtr rawpp) const
  {
    const ZZ* expv = myExpv(rawpp);
    size_t j = myNumIndetsValue;
    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
    {
      if (IsZero(expv[i])) continue;
      if (j != myNumIndetsValue || IsOne(expv[i])) return false;
      j = i;
    }
    if (j == myNumIndetsValue) return false;
    index = j;
    return true;
  }


  bool PPMonoidEvZZImpl::myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const ZZ* const expv1 = myExpv(rawpp1);
    const ZZ* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (!IsZero(expv1[i]) && !IsZero(expv2[i])) return false;

    return true;
  }


  bool PPMonoidEvZZImpl::myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const ZZ* const expv1 = myExpv(rawpp1);
    const ZZ* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] != expv2[i]) return false;
    return true;
  }


  bool PPMonoidEvZZImpl::myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const ZZ* const expv1 = myExpv(rawpp1);
    const ZZ* const expv2 = myExpv(rawpp2);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] < expv2[i]) return false;

    return true;
  }


  bool PPMonoidEvZZImpl::myIsRadical(ConstRawPtr rawpp) const
  {
    const ZZ* const expv = myExpv(rawpp);

    for (std::size_t i = 0; i < myNumIndetsValue; ++i)
      if (expv[i] > 1) return false;

    return true;
  }


  int PPMonoidEvZZImpl::myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myOrdPtr->myCmpExpvs(myExpv(rawpp1), myExpv(rawpp2));
  }


  long PPMonoidEvZZImpl::myStdDeg(ConstRawPtr rawpp) const
  {
    const ZZ* const expv = myExpv(rawpp);
    ZZ d;
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
      d += expv[i];
    size_t ans;
    if (!IsConvertible(ans, d))
      CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidEvZZImpl::myStdDeg");
    return ans;
  }


  void PPMonoidEvZZImpl::myWDeg(degree& d, ConstRawPtr rawpp) const
  {
    myOrdPtr->myWDeg(d, myExpv(rawpp));
  }


  int PPMonoidEvZZImpl::myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myOrdPtr->myCmpWDeg(myExpv(rawpp1), myExpv(rawpp2));
  }


  int PPMonoidEvZZImpl::myCmpWDegPartial(ConstRawPtr rawpp1, ConstRawPtr rawpp2, std::size_t i) const
  {
    return myOrdPtr->myCmpWDegPartial(myExpv(rawpp1), myExpv(rawpp2), i);
  }


  long PPMonoidEvZZImpl::myExponent(ConstRawPtr rawpp, std::size_t indet) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    long ans;
    if (!IsConvertible(ans, myExpv(rawpp)[indet]))
      CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidEvZZImpl::myExponent");
    return ans;
  }

  void PPMonoidEvZZImpl::myZZExponent(ZZ& EXP, ConstRawPtr rawpp, std::size_t indet) const
  {
    CoCoA_ASSERT(indet < myNumIndetsValue);
    EXP = myExpv(rawpp)[indet];
  }


  void PPMonoidEvZZImpl::myExponents(vector<long>& v, ConstRawPtr rawpp) const
  {
    CoCoA_ASSERT(v.size() == myNumIndetsValue);
    const ZZ* expv = myExpv(rawpp);
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
    {
      if (!IsConvertible(v[i], expv[i]))
        CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidEvZZImpl::myExponents");
    }
  }


  void PPMonoidEvZZImpl::myComputeDivMask(DivMask& /*dm*/, const DivMaskRule& /*DivMaskImpl*/, ConstRawPtr /*rawpp*/) const
  {
    CoCoA_ERROR(ERR::NYI, "PPMonoidEvZZImpl::myComputeDivMask");
// BUG BUG can't do this (yet)    DivMaskImpl->myAssignFromExpv(dm, myExpv(rawpp), myNumIndetsValue);
  }


  void PPMonoidEvZZImpl::myOutputSelf(std::ostream& out) const
  {
    out << "PPMonoidEvZZ(" << myNumIndetsValue << ", " << myOrd <<")";
  }


  void PPMonoidEvZZImpl::myDebugPrint(std::ostream& out, ConstRawPtr rawpp) const
  {
    out << "DEBUG PP: myNumIndetsValue=" << myNumIndetsValue << ", exps=[";
    for (std::size_t i=0; i < myNumIndetsValue; ++i)
      out << myExponent(rawpp, i) << " ";
    out << "]" << std::endl;
  }


  PPMonoid NewPPMonoidEvZZ(const std::vector<symbol>& IndetNames, const PPOrdering& ord)
  {
    // Sanity check on the indet names given.
    const size_t nvars = NumIndets(ord);

    if (IndetNames.size() != nvars)
      CoCoA_ERROR(ERR::BadNumIndets, "NewPPMonoidEvZZ(IndetNames,ord)");
    if (!AreDistinct(IndetNames))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidEvZZ(IndetNames,ord)");
    if (!AreArityConsistent(IndetNames))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidEvZZ(IndetNames,ord)");

    // Inefficient quadratic loop -- speed is probably not important.
    for (size_t i=0; i < nvars; ++i)
      for (size_t j=i+1; j < nvars; ++j)
        if (IndetNames[i] == IndetNames[j])
          CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidEvOv(IndetNames,ord)");

    return PPMonoid(new PPMonoidEvZZImpl(IndetNames, ord));
  }

  PPMonoid NewPPMonoidEvZZ(const std::vector<symbol>& IndetNames, const PPOrderingCtor& ord)
  {
    return NewPPMonoidEvZZ(IndetNames, ord.myCtor(IndetNames.size()));
  }


//--- orderings ----------------------------------------

  PPMonoidEvZZImpl::CmpBase::CmpBase(std::size_t NumIndets, size_t GradingDim):
    myNumIndetsValue(NumIndets),
    myGradingDimValue(GradingDim)
  {
    CoCoA_ASSERT(NumIndets > 0);
    CoCoA_ASSERT(NumIndets < 1000000); // complain about ridiculously large number of indets
    CoCoA_ASSERT(GradingDim <= NumIndets);
  }


//--- LexImpl

  PPMonoidEvZZImpl::LexImpl::LexImpl(std::size_t NumIndets):
    CmpBase(NumIndets, 0)
  {
  }


  int PPMonoidEvZZImpl::LexImpl::myCmpExpvs(const ZZ* v1, const ZZ* v2) const
  {
    for (size_t i=0; i<myNumIndetsValue; ++i)
      if (v1[i] != v2[i]) return (v1[i]>v2[i] ? 1 : -1 );
    return 0;
  }


  void PPMonoidEvZZImpl::LexImpl::myWDeg(degree& /*d*/, const ZZ* /*v*/) const
  {
    // deliberately does nothing because GradingDim=0
    //??? should it assign:  d = []
  }


  int PPMonoidEvZZImpl::LexImpl::myCmpWDegPartial(const ZZ* /*v1*/, const ZZ* /*v2*/, std::size_t /*i*/) const
  {
    return 0; // GradingDim=0, and all degrees in Z^0 are equal
  }

//--- StdDegLexImpl

  PPMonoidEvZZImpl::StdDegLexImpl::StdDegLexImpl(std::size_t NumIndets):
    CmpBase(NumIndets, 1)
  {
  }


  void PPMonoidEvZZImpl::StdDegLexImpl::myWDeg(degree& d, const ZZ* v) const
  {
    ZZ deg;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      deg += v[i];
    SetComponent(d, 0, deg);
  }


  int PPMonoidEvZZImpl::StdDegLexImpl::myCmpExpvs(const ZZ* v1, const ZZ* v2) const
  {
    ZZ deg1;
    ZZ deg2;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );

    for (size_t i=0; i < myNumIndetsValue; ++i)
      if (v1[i] != v2[i]) return (v1[i]>v2[i] ? 1 : -1 );
    return 0;
  }


  int PPMonoidEvZZImpl::StdDegLexImpl::myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t i) const
  {
    if (i==0) return 0; // ie: wrt to 0-grading
    ZZ deg1;
    ZZ deg2;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );
    return 0;
  }


//--- StdDegRevLexImpl

  PPMonoidEvZZImpl::StdDegRevLexImpl::StdDegRevLexImpl(std::size_t NumIndets):
    CmpBase(NumIndets, 1)
  {
  }


  void PPMonoidEvZZImpl::StdDegRevLexImpl::myWDeg(degree& d, const ZZ* v) const
  {
    ZZ deg;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      deg += v[i];
    SetComponent(d, 0, deg);
  }


  int PPMonoidEvZZImpl::StdDegRevLexImpl::myCmpExpvs(const ZZ* v1, const ZZ* v2) const
  {
    ZZ deg1;
    ZZ deg2;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );

    for (size_t i = myNumIndetsValue-1; i>0; --i)
      if (v1[i] != v2[i]) return (v1[i]>v2[i] ? -1 : 1 );
    return 0;
  }


  int PPMonoidEvZZImpl::StdDegRevLexImpl::myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t i) const
  {
    if (i==0) return 0; // ie: wrt to 0-grading
    ZZ deg1;
    ZZ deg2;
    for (size_t i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );
    return 0;
  }


//--- MatrixOrderingImpl

  PPMonoidEvZZImpl::MatrixOrderingImpl::MatrixOrderingImpl(std::size_t NumIndets, std::size_t GradingDim, const matrix& OrderMatrix):
    CmpBase(NumIndets, GradingDim)
  {
    CoCoA_ASSERT(NumRows(OrderMatrix) == NumIndets);
    CoCoA_ASSERT(NumCols(OrderMatrix) == NumIndets);

    myOrderMatrix.resize(NumIndets, vector<ZZ>(NumIndets));
    for (size_t i=0; i < NumIndets; ++i)
      for (size_t j=0; j < NumIndets; ++j)
        IsInteger(myOrderMatrix[i][j], OrderMatrix(i,j));
  }


  void PPMonoidEvZZImpl::MatrixOrderingImpl::myWDeg(degree& d, const ZZ* v) const
  {
    ZZ deg;
    for (size_t g=0; g < myGradingDimValue; ++g)
    {
      deg = 0;
      for (size_t i=0; i < myNumIndetsValue; ++i)
        deg += v[i] * myOrderMatrix[g][i];
      SetComponent(d, g, deg);
    }
  }


  static int MatrixOrderingCmpArrays(const vector<vector<ZZ> >& M, const ZZ* v1, const ZZ* v2, size_t UpTo)
  {
    ZZ s1, s2; // automatically set to 0
    for (size_t r=0; r < UpTo; ++r)
    {
      for (size_t i=0; i < M[1].size(); ++i)
      {
        s1 += v1[i]*M[r][i];
        s2 += v2[i]*M[r][i];
      }
      if (s1 != s2) return (s1>s2 ? 1 : -1 );
      s1 = 0;
      s2 = 0;
    }
    return 0;
  }


  int PPMonoidEvZZImpl::MatrixOrderingImpl::myCmpExpvs(const ZZ* v1, const ZZ* v2) const
  {
    return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, myNumIndetsValue);
  }


  int PPMonoidEvZZImpl::MatrixOrderingImpl::myCmpWDegPartial(const ZZ* v1, const ZZ* v2, std::size_t i) const
  {
    return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, i);
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/PPMonoidEvZZ.C,v 1.13 2010/02/03 16:13:52 abbott Exp $
// $Log: PPMonoidEvZZ.C,v $
// Revision 1.13  2010/02/03 16:13:52  abbott
// Added new single word tags for specifying the ordering in PPMonoid
// pseudo-ctors.
//
// Revision 1.12  2010/02/02 16:44:31  abbott
// Added radical & IsRadical (via mem fns myRadical & myIsRadical)
// for PPMonoidElems.
//
// Revision 1.11  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.10  2009/09/22 14:01:33  bigatti
// -- added myCmpWDegPartial (ugly name, I know....)
// -- cleaned up and realigned code in PPMonoid*.C files
//
// Revision 1.9  2008/04/21 12:58:17  abbott
// Commented out some formal parameters (so compiler doesn't give annoying warnings).
//
// Revision 1.8  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.7  2007/12/04 14:27:06  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.6  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.5  2007/09/24 14:22:12  abbott
// Added big exponent myMulIndetPower.
//
// Revision 1.4  2007/05/31 14:54:31  bigatti
// -- now using AreDistinct and AreArityConsistent for sanity check on
//    indet names
//
// Revision 1.2  2007/05/21 12:42:48  abbott
// Added missing NYI error message.
//
// Revision 1.1  2007/05/03 10:35:23  abbott
// Added new PPMonoidEvZZ with (virtually) unlimited exponents.
// Modified test-PPMonoid1.C accordingly.
// Added warning in doc about silent/unchecked exponent overflow in other
// PPMonoids.
//
