//   Copyright (c)  2005,2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Implementation of class PPMonoidSparseImpl

#include "CoCoA/PPMonoidSparse.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"
#include "CoCoA/symbol.H"
#include "CoCoA/DivMask.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/PPOrdering.H"

#include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
#include <new>
//for placement new
#include <string>
using std::string;
#include <utility>
using std::pair;
#include <vector>
using std::vector;


namespace CoCoA
{

  namespace // anonymous namespace for file local definitions
  {
    struct ExpListEntry
    {
    public:
      ExpListEntry(size_t indet, int exp);
    public:
      size_t myIndet;
      int myExp;
      ExpListEntry* myNextPtr;
    };

    ExpListEntry::ExpListEntry(size_t indet, int exp):
        myIndet(indet),
        myExp(exp),
        myNextPtr(0) // NULL ptr
    {}

    struct RawExpList
    {
      ExpListEntry* myFirstPtr; // ptr is NULL if list is empty
    };

    class ExpListMgr
    {
      ExpListMgr(MemPool& MemMgr);
      // Default copy ctor is OK.
      // Assignment fails because of reference data member.
      ~ExpListMgr();
      ExpListEntry* myNew(size_t indet, int exp);
      RawExpList myNew(const RawExpList& copy);
      void myDelete(RawExpList ToBeDeleted);
    private: // Data members
      mutable MemPool& myMemMgr;
    };

    inline ExpListMgr::ExpListMgr(MemPool& MemMgr):
        myMemMgr(MemMgr)
    {}

    inline ExpListMgr::~ExpListMgr()
    {}

    inline RawExpList ExpListMgr::myNew(size_t indet, int exp)
    {
      return new(myMemMgr.alloc()) ExpListEntry(indet, exp); // placement new
    }

    // Copies the whole list.
    RawExpList ExpListMgr::myNew(const RawExpList& copy)
    {
      ExpList ans(*this);
      RawExpList* ptr = &ans.myData;
      while (copy)
      {
        ExpListEntry* NewEntry = myNew(copy.myIndet, copy.myExp);
        *ptr = NewEntry;
        ptr = ptr->myNextPtr;
      }
      *ptr = 0; // NULL ptr
      return ans;
    }

    // Deletes the whole list.
    void ExpListMgr::myDelete(RawExpList ToBeDeleted)
    {
      ExpListEntry* curr = ToBeDeleted.myFirstPtr;
      while (curr)
      {
        ExpListEntry* next = curr->myNext;
        myMemMgr.free(curr);
        curr = next;
      }
    }

    class ExpList
    {
      explicit ExpList(const ExpListMgr& mgr);
      ExpList(const ExpListMgr& mgr, const RawExpList& data); // pseudo copy ctor
      ~ExpList();
    private:
      ExpListMgr myListMgr;
      RawExpList myData;
    };

    inline ExpList::~ExpList()
    {
      myListMgr.myDelete(myData);
    }

    inline ExpList::ExpList(const ExpListMgr& mgr, const RawExpList& data):
        myMemMgr(mgr),
        myData(myMemMgr.myNew(data))
    {}

    inline RawExpList ExpList::release()
    {
      RawExpList ans = myData;
      myData = 0; // NULL ptr
      return ans;
    }

//     inline void swap(ExpList& l1, ExpList& l2)
//     {
//       ExpListEntry* tmp = l1.myFirst;
//       l1.myFirst = l2.myFirst;
//       l2.myFirst = tmp;
//     }

  }


/////////////////////////////////////////////////////////////////////////////

  class PPMonoidSparseImpl
  {
  protected:
    PPMonoidSparseImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord);
    virtual ~PPMonoidSparseImpl() {};
  private: // disable copy ctor and assignment
    PPMonoidSparseImpl(const PPMonoidSparseImpl& copy);       ///< NEVER DEFINED -- disable default copy ctor
    PPMonoidSparseImpl& operator=(const PPMonoidSparseImpl&); ///< NEVER DEFINED -- disable assignment

  public:
    typedef PPMonoidElemRawPtr RawPtr;           ///< just to save typing
    typedef PPMonoidElemConstRawPtr ConstRawPtr; ///< just to save typing

    const PPOrdering& myOrdering() const;
    virtual std::size_t myNumIndets() const;                               ///< std::vector whose n-th entry is n-th indet as PPMonoidElem
    virtual const std::vector<PPMonoidElem>& myIndets() const;
    virtual void mySymbols(std::vector<symbol>& SymList) const;            ///< appends PPM's symbols to SymList
    virtual const symbol& myIndetName(std::size_t var) const;              ///< the symbol of the var-th indet

    virtual const PPMonoidElem& myOne() const;
    virtual PPMonoidElemRawPtr myNew() const;                              ///< ctor from nothing
    virtual PPMonoidElemRawPtr myNew(PPMonoidElemConstRawPtr rawpp) const;    ///< ctor from another pp
    virtual PPMonoidElemRawPtr myNew(const std::vector<long>& expv) const; ///< ctor from exp vector
//NYI    virtual PPMonoidElemRawPtr myNew(const std::vector<ZZ>& v) const;     ///< ctor from exp vector

    virtual void myDelete(RawPtr rawpp) const;                                ///< dtor, frees pp
    virtual void mySwap(RawPtr rawpp1, RawPtr rawpp2) const;                     ///< swap(pp1, pp2)
    virtual void myAssignOne(RawPtr rawpp) const;                             ///< pp = 1
    virtual void myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const;               ///< pp = pp1
    virtual void myAssign(RawPtr rawpp, const std::vector<long>& expv) const; ///< pp = v (assign from exp vector)
    virtual void myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1*pp2
    virtual void myMulIndetPower(RawPtr rawpp, std::size_t indet, long exp) const;///< pp *= indet^exp
    virtual void myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1/pp2
    virtual void myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;///< pp = pp1/gcd(pp1,pp2)
    virtual void myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = gcd(pp1,pp2)
    virtual void myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = lcm(pp1,pp2)
    virtual void myPower(RawPtr rawpp, ConstRawPtr rawpp1, long exp) const;      ///< pp = pp1^exp

    virtual bool myIsOne(ConstRawPtr rawpp) const;                            ///< is pp = 1?
    virtual bool myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;      ///< are pp1, pp2 coprime?
    virtual bool myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;            ///< is pp1 equal to pp2?
    virtual bool myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;    ///< is pp1 is divisible by pp2?

    virtual int myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;             ///< <0, =0, >0 as pp1 < = > pp2
    //    virtual int myHomogCmp(ConstRawPtr rawt1, ConstRawPtr rawt2) const;   ///< <0, =0, >0 as t1 < = > t2 assuming t1 and t2 have same multi-degree
    //    virtual int myHomogDegRevLex(ConstRawPtr rawt1, ConstRawPtr rawt2) const; ///< <0, =0, >0 as t1 < = > t2 ??? degrevlex assuming t1 and t2 have same multi-degree TO BE REMOVED

    virtual long myStdDeg(ConstRawPtr rawpp) const;                    ///< standard degree of pp
    virtual void myWDeg(degree& d, ConstRawPtr rawpp) const;                  ///< d = grading(pp)
    virtual int myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;         ///< <0, =0, >0 as wdeg(pp1) < = > wdeg(pp2)
    virtual long myExponent(ConstRawPtr rawpp, std::size_t indet) const;           ///< degree of pp in indet

    virtual void myExponents(std::vector<long>& expv, ConstRawPtr rawpp) const; ///< get exponents, SHOULD BE vector<ZZ> ????
    virtual void myComputeDivMask(DivMask& dm, const DivMaskRule& DivMaskImpl, ConstRawPtr rawpp) const; ///< computes the DivMask for pp according to DivMaskImpl
    virtual void myOutputSelf(std::ostream& out) const;                    ///< print value of PPMonoid
    virtual void myOutput(std::ostream& out, ConstRawPtr rawpp) const;   ///< NOT PURE!!
    virtual void myOutput(OpenMath::OutputChannel& OMOut, ConstRawPtr rawpp) const;///< NOT PURE!!

  protected: // Data members
    std::size_t myNumIndetsValue;
    mutable MemPool myMemMgr; // IMPORTANT: must come before myIndetVector.
    std::vector<PPMonoidElem> myIndetVector; ///< the indets as PPMonoidElems
    std::vector<symbol> myIndetNames;        ///< the names of the variables
  };


  PPMonoidSparseImpl(const PPOrdering& ord, const std::vector<symbol>& IndetNames):
      PPMonoidBase(ord),
      myNumIndetsValue(NumIndets(ord)),
      myIndetNames(IndetNames),
      myMemMgr(sizeof(ExpListEntry))
  {
    // Sanity check on the indet names given.
    CoCoA_ASSERT(myNumIndetsValue == myIndetNames.size());
    CoCoA_ASSERT(AreDistinct(myIndetNames));
    CoCoA_ASSERT(AreArityConsistent(myIndetNames));
  }


  std::size_t PPMonoidSparseImpl::myNumIndets() const
  {
    return myNumIndetsValue;
  }


  const std::vector<PPMonoidElem>& PPMonoidSparseImpl::myIndets() const
  {
    return myIndetVector;
  }


  void PPMonoidSparseImpl::mySymbols(std::vector<symbol>& SymList) const
  {
    std::copy(myIndetNames.begin(), myIndetNames.end(), back_inserter(SymList));
  }


  const symbol& PPMonoidSparseImpl::myIndetName(std::size_t var) const
  {
    return myIndetNames[var];
  }


  const PPMonoidElem& PPMonoidSparseImpl::myOne() const
  {
    return *myOnePtr;
  }


  PPMonoidElemRawPtr PPMonoidSparseImpl::myNew() const
  {
    return 0; // NULL pointer
  }


  //  THIS IS NOT EXCEPTION CLEAN AS WRITTEN HERE: will leak if MemPool::alloc throws.
  PPMonoidElemRawPtr PPMonoidSparseImpl::myNew(PPMonoidElemConstRawPtr rawpp) const
  {
    return myListMgr.myNew(import(rawpp));
  }


  PPMonoidElemRawPtr PPMonoidSparseImpl::myNew(const std::vector<long>& expv) const
  {
    ExpList ans(myListMgr);
    ExpListEntry* ptr = &ans.myData;
    for (size_t i=0; i < myNumIndetsValue; ++i)
      if (expv[i] != 0)
      {
        *ptr = myListMgr.myNew(i, expv[i]);
        ptr = ptr->myNextPtr;
      }
    return ans;
  }

//NYI    virtual PPMonoidElemRawPtr myNew(const std::vector<ZZ>& v) const;     ///< ctor from exp vector

  void PPMonoidSparseImpl::myDelete(RawPtr rawpp) const
  {
    myListMgr.myDelete(import(rawpp));
  }


  void PPMonoidSparseImpl::mySwap(RawPtr rawpp1, RawPtr rawpp2) const
  {
  }
    virtual void myAssignOne(RawPtr rawpp) const;                             ///< pp = 1
    virtual void myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const;               ///< pp = pp1
    virtual void myAssign(RawPtr rawpp, const std::vector<long>& expv) const; ///< pp = v (assign from exp vector)
    virtual void myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1*pp2
    virtual void myMulIndetPower(RawPtr rawpp, std::size_t indet, long exp) const;///< pp *= indet(var)^exp
    virtual void myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1/pp2
    virtual void myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;///< pp = pp1/gcd(pp1,pp2)
    virtual void myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = gcd(pp1,pp2)
    virtual void myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = lcm(pp1,pp2)
    virtual void myPower(RawPtr rawpp, ConstRawPtr rawpp1, long exp) const;///< pp = pp1^exp

    virtual bool myIsOne(ConstRawPtr rawpp) const;                            ///< is pp = 1?
    virtual bool myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;      ///< are pp1, pp2 coprime?
    virtual bool myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;            ///< is pp1 equal to pp2?
    virtual bool myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;    ///< is pp1 is divisible by pp2?

    virtual int myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;             ///< <0, =0, >0 as pp1 < = > pp2
    //    virtual int myHomogCmp(ConstRawPtr rawt1, ConstRawPtr rawt2) const;   ///< <0, =0, >0 as t1 < = > t2 assuming t1 and t2 have same multi-degree
    //    virtual int myHomogDegRevLex(ConstRawPtr rawt1, ConstRawPtr rawt2) const; ///< <0, =0, >0 as t1 < = > t2 ??? degrevlex assuming t1 and t2 have same multi-degree TO BE REMOVED

    virtual long myStdDeg(ConstRawPtr rawpp) const;                    ///< standard degree of pp
    virtual void myWDeg(degree& d, ConstRawPtr rawpp) const;                  ///< d = grading(pp)
    virtual int myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;         ///< <0, =0, >0 as wdeg(pp1) < = > wdeg(pp2)
    virtual long myExponent(ConstRawPtr rawpp, std::size_t var) const;  ///< exponent of indet in pp

    virtual void myExponents(std::vector<long>& v, ConstRawPtr rawpp) const; ///< get exponents, SHOULD BE vector<ZZ> ????
// CoCoA_ASSERT(v.size() == myNumIndetsValue);
    virtual void myComputeDivMask(DivMask& dm, const DivMaskRule& DivMaskImpl, ConstRawPtr rawpp) const; ///< computes the DivMask for pp according to DivMaskImpl
    virtual std::ostream& myOutput(std::ostream& out) const;                        ///< print value of PPMonoid
    virtual std::ostream& myOutput(std::ostream& out, ConstRawPtr rawpp) const;   ///< NOT PURE!!
    virtual OpenMath::OutputChannel& myOutput(OpenMath::OutputChannel& OMOut, ConstRawPtr rawpp) const;///< NOT PURE!!

  protected: // Data members
    std::size_t myNumIndetsValue;
    mutable MemPool myMemMgr; // IMPORTANT: must come before myIndetVector.
    std::vector<PPMonoidElem> myIndetVector; ///< the indets as PPMonoidElems
    std::vector<symbol> myIndetNames;        ///< the names of the variables
  };


  //----------------------------------------------------------------------
  // Pseudo-ctor for sparse PPMonoids

PPMonoid NewPPMonoidSparse(const std::vector<symbol>& IndetNames, const PPOrdering& ord)
{
  // Sanity check on the indet names given.
  const size_t nvars = NumIndets(ord);
  
  if (IndetNames.size() != nvars)
    CoCoA_ERROR(ERR::BadNumIndets, "NewPPMonoidSparse(IndetNames,ord)");
  if (!AreDistinct(IndetNames))
    CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidSparse(IndetNames,ord)");
  if (!AreArityConsistent(IndetNames))
    CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidSparse(IndetNames,ord)");
  
  return PPMonoid(new PPMonoidSparseImpl(IndetNames, ord));
}



}// end of namespace CoCoA




// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/PPMonoidSparse.C,v 1.7 2008/06/05 14:57:58 bigatti Exp $
// $Log: PPMonoidSparse.C,v $
// Revision 1.7  2008/06/05 14:57:58  bigatti
// -- added: include for MemPool and DivMask
//
// Revision 1.6  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.5  2007/12/04 14:27:06  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.4  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/05/31 14:54:31  bigatti
// -- now using AreDistinct and AreArityConsistent for sanity check on
//    indet names
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.10  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.9  2007/03/08 17:43:11  cocoa
// Swapped order of args to the NewPPMonoid pseudo ctors.
//
// Revision 1.8  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.7  2007/01/18 14:22:44  cocoa
// -- added "raw" to RawPtr arguments
//
// Revision 1.6  2007/01/15 13:39:54  cocoa
// -- added prefix "raw" to RawPtr arguments names
//
// Revision 1.5  2006/11/24 17:04:32  cocoa
// -- reorganized includes of header files
//
// Revision 1.4  2006/10/19 13:56:24  cocoa
// Added #include<new> whenever needed (i.e. for placement new).
//
// Revision 1.3  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.5  2006/04/28 16:33:51  cocoa
// Used SmartPtrIRC for PPOrderings.
//
// Revision 1.4  2006/02/20 22:41:20  cocoa
// All forms of the log function for power products now return SmallExponent_t
// (instead of int).  exponents now resizes the vector rather than requiring
// the user to pass in the correct size.
//
// Revision 1.3  2006/01/17 10:23:08  cocoa
// Updated DivMask; many consequential changes.
// A few other minor fixes.
//
// Revision 1.2  2005/10/24 15:22:12  cocoa
// Fixed a couple of buglets
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
