//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for abstract class PolyRing and friends

#include "CoCoA/PolyRing.H"
#include "CoCoA/error.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/TmpFactor.H"  // for IsIrredPoly

//#include <vector>
using std::vector;
//#include <cstddef>
using std::size_t;


namespace CoCoA
{

  bool PolyRingBase::myIsIrred(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) CoCoA_ERROR(ERR::ZeroRingElem, "RingBase::myIsIrred(rawx)");
    if (myIsInvertible(rawf)) CoCoA_ERROR(ERR::InvertibleRingElem, "RingBase::myIsIrred(rawx)");
    return IsIrredPoly(ConstRefRingElem(ring(this), rawf));
  }


  void PolyRingBase::myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    if (myIsZero(rawy)) // or CoCoA_ASSERT???
      CoCoA_ERROR(ERR::DivByZero,"PolyRingBase::myDiv");
    if (!myIsDivisible(rawlhs, rawx, rawy))
      CoCoA_ERROR("Inexact division", "PolyRingBase::myDiv");
  }
  
  

  const RingElem& indet(const PolyRing& P, std::size_t var)
  {
    if (var >= NumIndets(P))
      CoCoA_ERROR(ERR::BadIndetIndex, "indet(R, var)");
    return P->myIndets()[var];
  }


  RingElem IndetPower(const PolyRing& P, std::size_t var, unsigned int exp)
  {
    if (var >= NumIndets(P))
      CoCoA_ERROR(ERR::BadIndetIndex, "IndetPower(P, var, exp)");
    RingElem ans(P, 1);
    P->myIndetPower(raw(ans), var, exp);
    return ans;
  }


  std::size_t NumTerms(ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "NumTerms(f)");
    return AsPolyRing(owner(f))->myNumTerms(raw(f));
  }


  bool IsMonomial(ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "IsMonomial(f)");
    return AsPolyRing(owner(f))->myIsMonomial(raw(f));
  }


  bool IsConstant(ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "IsConstant(f)");
    return AsPolyRing(owner(f))->myIsConstant(raw(f));
  }


  bool IsIndet(ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "IsIndet(f)");
    size_t junk;
    return AsPolyRing(owner(f))->myIsIndet(junk, raw(f));
  }

  bool IsIndet(std::size_t& index, ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "IsIndet(index,f)");
    return AsPolyRing(owner(f))->myIsIndet(index, raw(f));
  }


  long MaxExponent(ConstRefRingElem f, std::size_t var)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "log(f,var)");
    if (IsZero(f))
      CoCoA_ERROR(ERR::LogZero, "log(f,var)");
    return AsPolyRing(owner(f))->myMaxExponent(raw(f), var);
  }


  RingElem content(ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "content(f)");
    const PolyRing Rx = AsPolyRing(owner(f));
    const ring R = CoeffRing(Rx);
    if (!IsGCDDomain(R))
      CoCoA_ERROR(ERR::NotElemGCDDomain, "content(f)");
//     if (IsField(R))  // see documentation (Bugs section)
//       CoCoA_ERROR(ERR::NotTrueGCDDomain, "content(f)");
    if (IsZero(f)) return zero(R);
    if (IsField(R)) return one(R);
    RingElem ans(R);
    Rx->myContent(raw(ans), raw(f));
    return ans;
  }


  RingElem CommonDenom(ConstRefRingElem f)
  {
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotElemPolyRing, "CommonDenom(f)");
    const PolyRing Qx = AsPolyRing(owner(f));
    const ring Q = CoeffRing(Qx);
    if (!IsFractionField(Q))
      CoCoA_ERROR(ERR::NotElemGCDDomain, "CoeffRing inside CommonDenom(f)");
    const ring R = BaseRing(AsFractionField(Q));
    if (!IsGCDDomain(R))
      CoCoA_ERROR(ERR::NotGCDDomain, "BaseRing of CoeffRing inside CommonDenom(f)");
//     if (IsField(R))  // see documentation (Bugs section)
//       CoCoA_ERROR(ERR::NotTrueGCDDomain, "content(f)");
    if (IsZero(f)) return one(R);

    RingElem ans(R);
    Qx->myCommonDenom(raw(ans), raw(f));
    return ans;
  }


  RingElem deriv(ConstRefRingElem f, ConstRefRingElem x)
  {
    PolyRing Rx = AsPolyRing(owner(f));
    if (owner(x) != Rx) CoCoA_ERROR(ERR::MixedRings, "deriv(f, x)");
    //??? Should check if LC = 1???
    if (NumTerms(x) != 1) CoCoA_ERROR("Nonsensical derivative", "deriv(f, x)");
    RingElem ans(Rx);
    Rx->myDeriv(raw(ans), raw(f), raw(x));
    return ans;
  }


  RingElem deriv(ConstRefRingElem f, std::size_t x) // here x is the index of the variable
  {
    PolyRing Rx = AsPolyRing(owner(f));
    if (x >= NumIndets(Rx)) CoCoA_ERROR(ERR::BadIndetIndex, "deriv(f, x)");
    return deriv(f, indet(Rx, x));
  }


  // Rx is the domain, S is the codomain
  RingHom PolyRingHom(const PolyRing& Rx, const ring& S, const RingHom& CoeffHom, const std::vector<RingElem>& IndetImages)
  {
    const char* const FnName = "PolyRingHom(Rx,S,CoeffHom,IndetImages)";
    if (domain(CoeffHom) != CoeffRing(Rx))
      CoCoA_ERROR(ERR::MixedCoeffRings, FnName);
    if (codomain(CoeffHom) != S)
      CoCoA_ERROR(ERR::BadCodomain, FnName);
    if (NumIndets(Rx) != IndetImages.size())
      CoCoA_ERROR(ERR::BadPolyRingHomImages, FnName);
    for (size_t i=0; i < NumIndets(Rx); ++i)
      if (owner(IndetImages[i]) != S)
        CoCoA_ERROR(ERR::BadPolyRingHomImages, FnName);

    return Rx->myHomCtor(S, CoeffHom, IndetImages);
  }


  RingHom EvalHom(const PolyRing& Rx, const std::vector<RingElem>& IndetImages)
  {
    const char* const FnName = "EvalHom(Rx,IndetImages)";
    const ring R = CoeffRing(Rx);
    if (NumIndets(Rx) != IndetImages.size())
      CoCoA_ERROR(ERR::BadPolyRingHomImages, FnName);
    for (size_t i=0; i < NumIndets(Rx); ++i)
      if (owner(IndetImages[i]) != R)
        CoCoA_ERROR(ERR::BadPolyRingHomImages, FnName);

    return Rx->myHomCtor(R, IdentityHom(R), IndetImages);
  }


  RingHom PolyAlgebraHom(const PolyRing& Rx, const ring& Ry, const std::vector<RingElem>& IndetImages)
  {
    const char* const FnName = "PolyAlgebraHom(Rx,Ry,IndetImages)";
    // Check that IndetImages are sensible...
    if (NumIndets(Rx) != IndetImages.size())
      CoCoA_ERROR(ERR::BadPolyRingHomImages, FnName);
    for (size_t i=0; i < NumIndets(Rx); ++i)
      if (owner(IndetImages[i]) != Ry)
        CoCoA_ERROR(ERR::BadPolyRingHomImages, FnName);
    // Special case: codomain is coeff ring.
    if (Ry == CoeffRing(Rx))
      return Rx->myHomCtor(Ry, IdentityHom(Ry), IndetImages);
    // General case: codomain must be a poly ring with same coeffs
    if (!IsPolyRing(Ry))
      CoCoA_ERROR(ERR::BadCodomain, FnName);
    if (CoeffRing(Rx) != CoeffRing(AsPolyRing(Ry)))
      CoCoA_ERROR(ERR::MixedCoeffRings, FnName);
    return Rx->myHomCtor(Ry, CoeffEmbeddingHom(AsPolyRing(Ry)), IndetImages);
  }


} // end of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/PolyRing.C,v 1.10 2009/10/02 13:27:26 bigatti Exp $
// $Log: PolyRing.C,v $
// Revision 1.10  2009/10/02 13:27:26  bigatti
// -- unique implementation of myDiv in PolyRing.C
//
// Revision 1.9  2009/07/24 12:26:43  abbott
// Added CommonDenom function for polynomials.
//
// Revision 1.8  2008/04/21 12:50:09  abbott
// Added EvalHom, an easy way to create evaluation homomorphisms for poly rings.
//
// Revision 1.7  2008/03/12 16:02:47  bigatti
// -- added: IsIrred
//
// Revision 1.6  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.5  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.4  2007/09/24 14:24:08  abbott
// Changed content to work if input is zero or if coeffs are in a field.
// Cleaned impl of PolyAlgebraHom to keep gcc-4.3 prerelease quiet.
//
// Revision 1.3  2007/05/31 15:26:23  bigatti
// -- default implementation for IamCommutative, IamIntegralDomain,
//    IamGCDDomain, IamField, myCharacteristic
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.7  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.6  2007/03/07 13:59:45  bigatti
// -- added PolyAlgebraHom(Rx, Ry, IndetImages)
//
// Revision 1.5  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.4  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.3  2007/02/28 13:51:59  bigatti
// -- added function IsMonomial
//
// Revision 1.2  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.2  2006/02/20 22:41:19  cocoa
// All forms of the log function for power products now return SmallExponent_t
// (instead of int).  exponents now resizes the vector rather than requiring
// the user to pass in the correct size.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.3  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.2  2005/07/01 16:08:15  cocoa
// Friday check-in.  Major change to structure under PolyRing:
// now SparsePolyRing and DUPolyRing are separated (in preparation
// for implementing iterators).
//
// A number of other relatively minor changes had to be chased through
// (e.g. IndetPower).
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.7  2005/04/29 15:42:02  cocoa
// Improved documentation for GMPAllocator.
// Added example program for GMPAllocator.
// Added example program for simple ops on polynomials.
// Added two new ctors for (principal) ideals (from long, and from ZZ).
// Added (crude) printing for PPMonoids.
// Updated library.H (#included GMPAllocator.H).
//
// Revision 1.6  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.5  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.4  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.3  2005/03/11 18:36:36  cocoa
// -- new: NewPolyRingHom Rx-->Ry ; code is still commented out
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.10  2004/12/16 13:35:01  cocoa
// Added new monomial functions (with more coherent signatures).
//
// Revision 1.9  2004/11/25 16:14:21  cocoa
// (1) Fixed definition of specialization of std::swap template function
//     so that it compiles with gcc 3.4.3
// (2) Implemented monomial function for polynomial rings.
// (3) Added one(PPM) and PPM->myOne() functions.
//
// Revision 1.8  2004/11/19 16:15:51  cocoa
// (a) Removed unused error message about degree of zero;
//     replaced it by a more general message about needing a
//     non-zero polynomial (for various ops such as LC, LPP).
// (b) Added some missing arg checking in LC, LPP and deg
//     (for elems of a PolyRing).
// (c) Updated some commented out code in GPair and GPoly.
//
// Revision 1.7  2004/11/19 15:44:27  cocoa
// Changed names of "casting" functions which convert a ring into
// one with a more special structure (e.g. FractionField).  These
// functions now have names starting with "As".  There were several
// consequential changes.
//
// Revision 1.6  2004/11/18 18:33:41  cocoa
// Now every ring know its own "one" element (as well as "zero").
// Several consequential changes.
//
// Revision 1.5  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.4  2004/10/29 16:05:03  cocoa
// -- changed PPOrdering::ExpvElem --> SmallExponent_t
//
// Revision 1.3  2004/05/24 15:52:14  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.2  2003/10/09 12:18:42  cocoa
// New coding convention for rings.
//
// Revision 1.5  2003/06/23 17:04:45  abbott
// Minor cleaning prior to public release.
//
// Revision 1.4  2003/05/30 12:25:12  abbott
// Corrected signature of IndetPower: previously allowed signed exponent.
//
// Revision 1.3  2003/05/14 16:53:05  abbott
// Numerous changes reflecting the renaming and restructuring
// mentioned in PolyRing.H (qv.).
//
// Revision 1.2  2002/11/15 15:15:55  abbott
// Revised according to the renaming in ring.H.
// Error messages improved.
//
// Revision 1.1  2002/06/22 16:54:40  abbott
// Initial revision
//
//

