//   Copyright (c)  2009-2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/QQ.H"

#include "CoCoA/NumTheory.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"

#include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
using std::istream;

namespace CoCoA
{
  
//????????  QQ::QQ(MP_RAT to_be_owned); // for efficient return values in operator+ etc.

  QQ::QQ()
  {
    mpq_init(myRep);
  }

  QQ::QQ(const MachineInteger& n)
  {
    mpq_init(myRep);
    if (IsNegative(n))
      mpq_set_si(myRep, AsSignedLong(n), 1);
    else
      mpq_set_ui(myRep, AsUnsignedLong(n), 1);
  }

  QQ::QQ(const ZZ& N)
  {
    mpq_init(myRep);
    mpq_set_z(myRep, mpzref(N));
  }

  QQ::QQ(const MachineInteger& n1, const MachineInteger& n2, NormalizeFlag status)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(n1,n2)");
    mpq_init(myRep);
    mpq_set_ui(myRep, abs(n1), abs(n2));
    if (status == NotNormalized)
      mpq_canonicalize(myRep);
    else
      CoCoA_ASSERT(!IsNegative(n2) && gcd(n1,n2) == 1);
    if ( (IsNegative(n1) && !IsNegative(n2)) ||
         (!(IsNegative(n1)) && IsNegative(n2)))
      mpq_neg(myRep, myRep);
  }

  QQ::QQ(const MachineInteger& n1, const ZZ& N2, NormalizeFlag status)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(n1,N2)");
    mpq_init(myRep);
    mpz_set_ui(mpq_numref(myRep), abs(n1));
    if (IsNegative(n1))
      mpz_neg(mpq_numref(myRep), mpq_numref(myRep));
    mpz_set(mpq_denref(myRep), mpzref(N2));
    if (status == NotNormalized)
      mpq_canonicalize(myRep);
    else
      CoCoA_ASSERT(N2 > 0 && gcd(n1,N2) == 1);
  }

  QQ::QQ(const ZZ& N1, const MachineInteger& n2, NormalizeFlag status)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(N1,n2)");
    mpq_init(myRep);
    mpz_set(mpq_numref(myRep), mpzref(N1));
    mpz_set_ui(mpq_denref(myRep), abs(n2));
    if (IsNegative(n2))
      mpz_neg(mpq_numref(myRep), mpq_numref(myRep));
    if (status == NotNormalized)
      mpq_canonicalize(myRep);
    else
      CoCoA_ASSERT(!IsNegative(n2) > 0 && gcd(N1,n2) == 1);
  }

  QQ::QQ(const ZZ& N1, const ZZ& N2, NormalizeFlag status)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(N1,N2)");
    mpq_init(myRep);
    mpz_set(mpq_numref(myRep), mpzref(N1));
    mpz_set(mpq_denref(myRep), mpzref(N2));
    if (status == NotNormalized)
      mpq_canonicalize(myRep);
    else
      CoCoA_ASSERT(N2 > 0 && gcd(N1,N2) == 1);
  }


  QQ::QQ(const std::string& str)
  {
    mpq_init(myRep);
    convert(*this, str);
  }

  QQ::QQ(const char* str)
  {
    mpq_init(myRep);
    if (str != 0)
      convert(*this, str);
  }


  QQ::QQ(const QQ& from)
  {
    mpq_init(myRep);
    mpq_set(myRep, from.myRep);
  }


// QQ& clone() const;

  QQ::~QQ()
  {
    mpq_clear(myRep);
  }

  QQ& QQ::operator=(const QQ& rhs)
  {
    mpq_set(myRep, rhs.myRep);
    return *this;
  }

    // -------- functions that modify at least one argument or `*this' ----------

  QQ& QQ::operator+=(const QQ& rhs)
  {
    mpq_add(myRep, myRep, rhs.myRep);
    return *this;
  }

  QQ& QQ::operator-=(const QQ& rhs)
  {
    mpq_sub(myRep, myRep, rhs.myRep);
    return *this;
  }

  QQ& QQ::operator*=(const QQ& rhs)
  {
//     if (mpz_sgn(mpq_numref(myRep)) == 0) return *this;
//     if (mpz_sgn(mpq_numref(rhs.myRep)) == 0) return operator=(rhs);
    mpq_mul(myRep, myRep, rhs.myRep);
    return *this;
  }

  QQ& QQ::operator/=(const QQ& rhs)
  {
    if (mpz_sgn(mpq_numref(rhs.myRep)) == 0)
      CoCoA_ERROR(ERR::DivByZero, "q1 /= q2");
    mpq_div(myRep, myRep, rhs.myRep);
    return *this;
  }
                        
    // Same but with RHS a ZZ...
  QQ& QQ::operator=(const ZZ& rhs)
  {
    mpq_set_z(myRep, mpzref(rhs));
    return *this;
  }


  // SLUG: impl is needlessly inefficient: makes useless copy in D
  QQ& QQ::operator+=(const ZZ& rhs)
  {
    ZZ D(CopyFromMPZ, mpq_denref(myRep));
    ZZ tmp = rhs*D;
    mpz_add(mpq_numref(myRep), mpq_numref(myRep), mpzref(tmp));
    // no need to call mpq_canonicalize
    return *this;
  }

  // SLUG: impl is needlessly inefficient: makes useless copy in D
  QQ& QQ::operator-=(const ZZ& rhs)
  {
    ZZ D(CopyFromMPZ, mpq_denref(myRep));
    ZZ tmp = rhs*D;
    mpz_sub(mpq_numref(myRep), mpq_numref(myRep), mpzref(tmp));
    // no need to call mpq_canonicalize
    return *this;
  }

  QQ& QQ::operator*=(const ZZ& rhs)
  {
    return operator*=(QQ(rhs));
  }

  QQ& QQ::operator/=(const ZZ& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "Q /= N");
    // Could be more efficient if "*this" is 0.
    return operator/=(QQ(rhs));
  }
                        
    // Same but with RHS a MachinInteger...
  QQ& QQ::operator= (const MachineInteger& rhs)
  {
    if (IsNegative(rhs))
      mpq_set_si(myRep, AsSignedLong(rhs), 1);
    else
      mpq_set_ui(myRep, AsUnsignedLong(rhs), 1);
    return *this;
  }

  QQ& QQ::operator+=(const MachineInteger& rhs)
  {
    return operator+=(ZZ(rhs));
  }

  QQ& QQ::operator-=(const MachineInteger& rhs)
  {
    return operator-=(ZZ(rhs));
  }

  QQ& QQ::operator*=(const MachineInteger& rhs)
  {
    return operator*=(ZZ(rhs));
  }

  QQ& QQ::operator/=(const MachineInteger& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "Q /= n");
    return operator/=(ZZ(rhs));
  }


  const QQ& QQ::operator++()
  {
    mpz_add(mpq_numref(myRep), mpq_numref(myRep), mpq_denref(myRep)); // no need to reduce
    return *this;
  }

  const QQ& QQ::operator--()
  {
    mpz_sub(mpq_numref(myRep), mpq_numref(myRep), mpq_denref(myRep));
    return *this;
  }

  const QQ QQ::operator++(int) // INEFFICIENT
  {
    QQ ans(*this);
    operator++();
    return ans;
  }

  const QQ QQ::operator--(int) // INEFFICIENT
  {
    QQ ans(*this);
    operator--();
    return ans;
  }



  // I/O FUNCTIONS

  std::ostream& operator<<(std::ostream& out, const QQ& Q)
  {
    out << num(Q);
    if (!IsOne(den(Q)))
      out << "/" << den(Q);
    return out;
  }

  std::istream& operator>>(std::istream& in, QQ& Q)
  {
    if (!in.good()) return in;
    ZZ N;
    in >> N;
    if (!in.good()) return in;
    const char slash = in.peek();
    if (slash != '/')
    {
      Q = N;
      return in;
    }
    ZZ D;
    in >> D;
    if (!in.good()) return in;
    QQ ans(N,D);
    swap(Q, ans); // really an assignment
    return in;
  }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const QQ& Q)
  {
    OMOut->mySendApplyStart();
    OMOut->mySendSymbol("nums1","rational");
    OMOut->mySend(num(Q));
    OMOut->mySend(den(Q));
    OMOut->mySendApplyEnd();
    return OMOut;
  }


  OpenMathInput& operator>>(OpenMathInput& OMIn, QQ& /*Q*/)
  {
    CoCoA_ERROR(ERR::NYI, "OpenMathInput fn for QQ");
    return OMIn;
  }


  // STANDARD ARITHMETIC OPERATIONS

  void swap(QQ& a, QQ& b)
  {
    mpq_swap(mpqref(a), mpqref(b));
  }

  const ZZ num(const QQ& Q)
  {
    return ZZ(CopyFromMPZ, mpq_numref(mpqref(Q)));
  }

  const ZZ den(const QQ& Q)
  {
    return ZZ(CopyFromMPZ, mpq_denref(mpqref(Q)));
  }

  const QQ abs(const QQ& Q)
  {
    QQ ans;
    mpq_abs(mpqref(ans), mpqref(Q));
    return ans;
  }

  const QQ operator-(const QQ& Q)
  {
    QQ ans;
    mpq_neg(mpqref(ans), mpqref(Q));
    return ans;
  }

  const QQ operator+(const QQ& Q1, const QQ& Q2)
  {
    QQ ans;
    mpq_add(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }

  const QQ operator-(const QQ& Q1, const QQ& Q2)
  {
    QQ ans;
    mpq_sub(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }

  const QQ operator*(const QQ& Q1, const QQ& Q2)
  {
    QQ ans;
    mpq_mul(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }

  const QQ operator/(const QQ& Q1, const QQ& Q2)
  {
    if (IsZero(Q2))
      CoCoA_ERROR(ERR::DivByZero, "Q1/Q2");
    QQ ans;
    mpq_div(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }


  const QQ operator+(const QQ& Q, const ZZ& N)
  {
    QQ ans = Q;
    return ans += N;
//   THE LINES BELOW SHOULD BE MORE EFFICIENT (but they're not very readable).
//     QQ ans;
//     mpz_mul(mpq_denref(ans), mpzref(N), mpq_denref(mpqref(Q)));
//     mpz_add(mpq_numref(mpqref(ans)), mpq_denref(mpqref(ans)), mpq_numref(mpqref(Q)));
//     mpz_set(mpq_denref(mpqref(ans)), mpq_denref(mpqref(Q)));
//     return ans;
  }

  const QQ operator-(const QQ& Q, const ZZ& N)
  {
    QQ ans(Q);
    return ans -= N;
  }

  const QQ operator*(const QQ& Q, const ZZ& N)
  {
    QQ ans(Q);
    return ans *= N;
  }

  const QQ operator/(const QQ& Q, const ZZ& N)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::DivByZero, "Q/N");
    QQ ans(Q);
    return ans /= N;
  }

  const QQ operator+(const ZZ& N, const QQ& Q)
  {
    return Q+N;
  }

  const QQ operator-(const ZZ& N, const QQ& Q)
  {
    QQ ans = Q-N;
    return -ans;
  }

  const QQ operator*(const ZZ& N, const QQ& Q)
  {
    return Q*N;
  }

  const QQ operator/(const ZZ& N, const QQ& Q)
  {
    if (IsZero(Q))
      CoCoA_ERROR(ERR::DivByZero, "N/Q");
    return QQ(N)/Q;
  }


  const QQ operator+(const QQ& Q, const MachineInteger& n)
  {
    return Q + QQ(n);
  }

  const QQ operator-(const QQ& Q, const MachineInteger& n)
  {
    return Q - QQ(n);
  }

  const QQ operator*(const QQ& Q, const MachineInteger& n)
  {
    return Q * QQ(n);
  }

  const QQ operator/(const QQ& Q, const MachineInteger& n)
  {
    if (IsZero(n))
      CoCoA_ERROR(ERR::DivByZero, "Q/n");
    return Q / QQ(n);
  }


  const QQ operator+(const MachineInteger& n, const QQ& Q)
  {
    return QQ(n) + Q;
  }

  const QQ operator-(const MachineInteger& n, const QQ& Q)
  {
    return QQ(n) - Q;
  }

  const QQ operator*(const MachineInteger& n, const QQ& Q)
  {
    return QQ(n) * Q;
  }

  const QQ operator/(const MachineInteger& n, const QQ& Q)
  {
    if (IsZero(Q))
      CoCoA_ERROR(ERR::DivByZero, "n/Q");
    return QQ(n) / Q;
  }

  const QQ power(const QQ& base, const ZZ& exponent)
  {
    return QQ(power(num(base), exponent), power(den(base), exponent)); // BUG??? we know there is no common factor
  }

  const QQ power(const QQ& base, const MachineInteger& exponent)
  {
    return QQ(power(num(base), exponent), power(den(base), exponent)); // BUG??? we know there is no common factor
  }



  bool IsZero(const QQ& Q)
  {
    return IsZero(num(Q));
  }


  bool IsOne(const QQ& Q)
  {
    return IsOne(num(Q)) && IsOne(den(Q));
  }


  bool IsMinusOne(const QQ& Q)
  {
    return IsMinusOne(num(Q)) && IsOne(den(Q));
  }


  bool IsOneDen(const QQ& Q)
  {
    return mpz_cmp_ui(mpq_denref(mpqref(Q)), 1) == 0;
  }


  int sign(const QQ& Q)
  {
    return mpq_sgn(mpqref(Q));
  }


  // COMPARISON FUNCTIONS

  int cmp(const QQ& Q1, const QQ& Q2)
  {
    return mpq_cmp(mpqref(Q1), mpqref(Q2));
  }


  int cmp(const QQ& Q, const ZZ& N)
  {
    return cmp(num(Q), N*den(Q));
  }


  int cmp(const ZZ& N, const QQ& Q)
  {
    return cmp(N*den(Q), num(Q));
  }


  int cmp(const QQ& Q, const MachineInteger& n)
  {
    if (IsNegative(n))
      return mpq_cmp_si(mpqref(Q), AsSignedLong(n),1);
    return mpq_cmp_ui(mpqref(Q), AsUnsignedLong(n),1);
  }


  int cmp(const MachineInteger& n, const QQ& Q)
  {
    return -cmp(Q, n);
  }


  bool operator==(const QQ& Q1, const QQ& Q2)
  {
    return mpq_equal(mpqref(Q1), mpqref(Q2));
  }

  bool operator!=(const QQ& Q1, const QQ& Q2)
  {
    return !(Q1 == Q2);
  }

  bool operator> (const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) > 0;
  }

  bool operator>=(const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) >= 0;
  }

  bool operator< (const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) < 0;
  }

  bool operator<=(const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) <= 0;
  }

                        
  bool operator==(const QQ& Q, const ZZ& N)
  {
    return IsOne(den(Q)) && num(Q) == N;
  }

  bool operator!=(const QQ& Q, const ZZ& N)
  {
    return !(Q == N);
  }

  bool operator> (const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) > 0;
  }

  bool operator>=(const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) >= 0;
  }

  bool operator< (const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) < 0;
  }

  bool operator<=(const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) <= 0;
  }

                        
  bool operator==(const ZZ& N, const QQ& Q)
  {
    return Q == N;
  }

  bool operator!=(const ZZ& N, const QQ& Q)
  {
    return !(Q == N);
  }

  bool operator> (const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) > 0;
  }

  bool operator>=(const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) >= 0;
  }

  bool operator< (const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) < 0;
  }

  bool operator<=(const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) <= 0;
  }

                        
  bool operator==(const QQ& Q, const MachineInteger& n)
  {
    return IsOne(den(Q)) && num(Q) == n;
  }

  bool operator!=(const QQ& Q, const MachineInteger& n)
  {
    return !(Q == n);
  }

  bool operator> (const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) > 0;
  }

  bool operator>=(const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) >= 0;
  }

  bool operator< (const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) < 0;
  }

  bool operator<=(const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) <= 0;
  }

                
  bool operator==(const MachineInteger& n, const QQ& Q)
  {
    return IsOne(den(Q)) && num(Q) == n;
  }

  bool operator!=(const MachineInteger& n, const QQ& Q)
  {
    return !(n == Q);
  }

  bool operator> (const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) > 0;
  }

  bool operator>=(const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) >= 0;
  }

  bool operator< (const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) < 0;
  }

  bool operator<=(const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) <= 0;
  }

                        

  // MISCELLANEOUS FUNCTIONS

  double mantissa(const QQ& /*Q*/)
  {
    CoCoA_ERROR(ERR::NYI, "mantissa(QQ)");
    return 0.0;
  }

  long exponent(const QQ& /*Q*/)
  {
    CoCoA_ERROR(ERR::NYI, "exponent(QQ)");
    return 0;
  }

  // BUG BUG BUG Is this log(Q)  or log(abs(Q))???
  double log(const QQ& Q)
  {
    if (IsZero(Q))
      CoCoA_ERROR(ERR::BadArg, "log(Q)");
    return log(num(Q)) - log(den(Q));
  }


  ZZ floor(const QQ& Q)
  {
    ZZ ans;
    mpz_fdiv_q(mpzref(ans), mpq_numref(mpqref(Q)), mpq_denref(mpqref(Q)));
    return ans;
  }

  ZZ ceil(const QQ& Q)
  {
    ZZ ans;
    mpz_cdiv_q(mpzref(ans), mpq_numref(mpqref(Q)), mpq_denref(mpqref(Q)));
    return ans;
  }


  ZZ round(const QQ& Q)
  {
    if (IsOneDen(Q)) return num(Q);
    return (2*num(Q)+den(Q))/(2*den(Q));
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/QQ.C,v 1.7 2010/03/22 11:49:28 abbott Exp $
// $Log: QQ.C,v $
// Revision 1.7  2010/03/22 11:49:28  abbott
// Added ctor from a string.
//
// Revision 1.6  2010/03/18 16:40:42  abbott
// Added missing include directive.
//
// Revision 1.5  2010/03/18 16:34:10  abbott
// Added new pseudo-ctors for QQ with optional flag to indicate that value is already normalized.
// Added OpenMath I/O operators.
//
// Revision 1.4  2009/10/26 15:39:24  bigatti
// -- added CopyFromMPZ in ZZ ctor
//
// Revision 1.3  2009/07/08 12:26:53  abbott
// Added floor and ceil functions for QQs.
// Added example program for QQs.
// Minor correction to convert.C; minor cleaning to ex-ZZ1.C
//
// Revision 1.2  2009/07/06 12:31:26  abbott
// Commented out two unused function arguments (to keep compiler quiet
// when compiling a debugging version).
//
// Revision 1.1  2009/07/02 16:29:42  abbott
// Added new class QQ to represent rational numbers.
// Consequent change to the Makefile.
//
//
