//   Copyright (c)  2003-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for class GeneralQuotientRingImpl

#include "CoCoA/QuotientRing.H"

#include "CoCoA/GlobalManager.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingFpDouble.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/ZZ.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"


#include <iostream>
using std::ostream;
#include <vector>
using std::vector;


namespace CoCoA
{

  class QuotientingHomImpl: public RingHomBase
  {
  public:
    QuotientingHomImpl(const ring& R, const QuotientRing& RmodI);
    virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
  private:
    virtual void myOutputSelfDetails(std::ostream& out) const;
  };


  RingHom QuotientRingBase::myQuotientingHomCtor() const
  {
    return RingHom(new QuotientingHomImpl(myBaseRing, QuotientRing(this)));
  }


  class GeneralQuotientRingImpl: public QuotientRingBase
  {
  private: // data members
    RingHom myQuotHom;

  public:
    GeneralQuotientRingImpl(const ring& R, const ideal& I);
    // Default copy ctor works fine.
    // Assignment disabled
    ~GeneralQuotientRingImpl();
  private:
    GeneralQuotientRingImpl& operator=(const GeneralQuotientRingImpl&); // NEVER DEFINED
  public: // functions that every ring must have
    virtual void myCharacteristic(ZZ& p) const;
    virtual bool IamCommutative() const;
    virtual bool IamIntegralDomain() const;
    virtual bool IamGCDDomain() const;
    virtual bool IamField() const;
    virtual ConstRefRingElem myZero() const;
    virtual ConstRefRingElem myOne() const;
    virtual RingElemRawPtr myNew() const;
    virtual RingElemRawPtr myNew(const MachineInteger& n) const;
    virtual RingElemRawPtr myNew(const ZZ& N) const;
    virtual RingElemRawPtr myNew(ConstRawPtr rawt) const;
    virtual void myDelete(RawPtr rawx) const;                      // destroys x (incl all resources)
    virtual void mySwap(RawPtr rawx, RawPtr rawy) const;                           // swap(x, y)
    virtual void myAssign(RawPtr rawlhs, ConstRawPtr rawx) const;                  // lhs = x
    virtual void myAssign(RawPtr rawlhs, const MachineInteger& n) const;           // lhs = n
    virtual void myAssign(RawPtr rawlhs, const ZZ& N) const;                       // lhs = N
    virtual void myAssignZero(RawPtr rawlhs) const;                                // lhs = 0
    virtual void myNegate(RawPtr rawlhs, ConstRawPtr rawx) const;                  // lhs = -x
    virtual void myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;   // lhs = x+y
    virtual void mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;   // lhs = x-y
    virtual void myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;   // lhs = x*y
    virtual void myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;   // lhs = x/y
    virtual bool myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;// lhs = x/y, if divisible
    virtual bool myIsInvertible(ConstRawPtr rawx) const;                           // true iff x is invertible
    virtual void myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;   // lhs = gcd(x,y) if GCDDomain; o/w error.
    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const;// lhs = x^n, n>1, x not -1,0,1
    virtual void mySymbols(std::vector<symbol>& SymList) const;                    // appends ring's symbols to SymList
    virtual void myOutput(std::ostream& out, ConstRawPtr rawx) const;              // out << x
    virtual bool myIsPrintAtom(ConstRawPtr rawx) const;                            ///< x^n may be printed without parentheses
    virtual bool myIsPrintedWithMinus(ConstRawPtr rawx) const;                     ///< first character of x printed is a minus sign
    virtual void myOutputSelf(std::ostream& out) const;                            // out << R
    virtual void myOutputSelf(OpenMathOutput& OMOut) const;                        // OMOut << R
    virtual void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const;          // OMOut << x
    virtual bool myIsZero(ConstRawPtr rawx) const;                                 // x == 0
    virtual bool myIsOne(ConstRawPtr rawx) const;                                  // x == 1
    virtual bool myIsMinusOne(ConstRawPtr rawx) const;                             // x == -1
    virtual bool myIsInteger(ZZ& N, ConstRawPtr rawx) const;                       // true iff x is integer
    virtual bool myIsRational(QQ& Q, ConstRawPtr rawx) const;                      // true iff x is rational
    virtual bool myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const;              // x == y
  //???  virtual void convert(string&, ConstRawPtr) const;

    virtual ideal myIdealCtor(const std::vector<RingElem>& gens) const;

    virtual RingHom myCompose(const RingHom& phi, const RingHom& theta) const; // phi(theta(...))

  public: // functions that every QuotientRing must have
    virtual RingElem myCanonicalRepr(ConstRawPtr rawx) const;
    virtual void myReduction(RawPtr rawimage, ConstRawPtr rawarg) const;
    virtual const RingHom& myQuotientingHom() const;
    virtual RingHom myInducedHomCtor(const RingHom& InducingHom) const; // NB domain(InducingHom) == myReprRing, and ker(InducingHom) contains myReducingIdeal.

  private:
    void myReduction(RawPtr rawx) const; // implementation detail


  private:
    class InducedHomImpl: public RingHomInducedBase
    {
    public:
      InducedHomImpl(const QuotientRing& RmodI, const RingHom& InducingHom);
      // Default copy ctor & assignment disabled in RingHomBase.
      // Default dtor works fine.
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    };

  private:
    class IdealImpl: public IdealBase
    {
    public:
      friend class GeneralQuotientRingImpl;
      IdealImpl(const QuotientRing& R, const vector<RingElem>& gens);
      ~IdealImpl();
      virtual IdealImpl* clone() const;
//???    virtual void swap(ideal& other);

      virtual const QuotientRing& myAmbientRing() const; // covariant return type!
      virtual bool IamZero() const;
      virtual bool IamOne() const;
      virtual void mySetMaximalFlag(bool b) const;    // virtual + default impl
      virtual void mySetPrimeFlag(bool b) const;      // virtual + default impl

      virtual void myReduceMod(RingElemRawPtr rawx) const; // r elem of R, where I is ideal in R
      virtual bool contains(RingElemConstRawPtr rawx) const;
      virtual void add(const ideal&);
      virtual void mul(const ideal&);
      virtual void intersect(const ideal&);
      virtual void colon(const ideal&);
      virtual void myDivMod(RingElemRawPtr rawlhs, RingElemConstRawPtr rawnum, RingElemConstRawPtr rawden) const; // lhs = num/den modulo the ideal  (lhs = 0 if quotient does not exist or ideal = ideal(1))

      virtual const std::vector<RingElem>& myGens() const; // gens as specified by user
      virtual const std::vector<RingElem>& myTidyGens() const; // tidier set of gens
//??? use default??      virtual void myOutputSelf(OpenMathOutput&) const;
    protected:
      virtual void myMaximalTest() const;
      virtual void myPrimeTest() const;

    private:
      QuotientRing myAmbientRingValue;
      vector<RingElem> myGensValue;
      vector<RingElem> myTidyGensValue;
      ideal myPreimageIdeal;
      void SetTidyGensFromPreimageIdeal(); // implementation detail
      static const IdealImpl* GetPtr(const ideal& I);
    };
  };




  QuotientRingBase::QuotientRingBase(const ring& R, const ideal& I):
      myBaseRing(R),
      myDefiningIdeal(I),
      myReprRing(IsQuotientRing(R)?ReprRing(AsQuotientRing(R)):R),
      myReducingIdeal(IsQuotientRing(R)?(I+ReducingIdeal(AsQuotientRing(R))):I)
  {
    CoCoA_ASSERT(AmbientRing(I) == R);
  }


  QuotientRing::QuotientRing(const QuotientRingBase* RingPtr):
      ring(RingPtr)
  {}


  QuotientingHomImpl::QuotientingHomImpl(const ring& R, const QuotientRing& RmodI):
      RingHomBase(R, RmodI)
  {}


  void QuotientingHomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    AsQuotientRing(myCodomain)->myReduction(rawimage, rawarg);
  }


  void QuotientingHomImpl::myOutputSelfDetails(std::ostream& out) const
  {
    out << " canonical quotient";
  }



  GeneralQuotientRingImpl::GeneralQuotientRingImpl(const ring& R, const ideal& I):
      QuotientRingBase(R, I),
      myQuotHom(new QuotientingHomImpl(R, QuotientRing(this)))
  {
    myRefCountZero();
  }


  GeneralQuotientRingImpl::~GeneralQuotientRingImpl()
  {}


  void GeneralQuotientRingImpl::myCharacteristic(ZZ& p) const
  {
    // Handle the simple case of Z/I
    if (IsZ(myReprRing))
    {
      if (IsZero(myReducingIdeal)) p = 0;
      else IsInteger(p, TidyGens(myReducingIdeal)[0]);
      return;
    }
    if (IsOne(myReducingIdeal)) { p = 1; return; } // trivial ring has char=1
    CoCoA_ERROR(ERR::NYI, "GeneralQuotientRingImpl::characteristic");
    myBaseRing->myCharacteristic(p); // just to keep compiler quiet
  }


  bool GeneralQuotientRingImpl::IamCommutative() const
  {
    return IsCommutative(myBaseRing); // BUG??? this could fail to recognize commutativity
  }


  bool GeneralQuotientRingImpl::IamIntegralDomain() const
  {
    return IsPrime(myReducingIdeal);
  }


//??? How to define this function correctly???
// we could be clever in some easy cases, but it is hard in general
  bool GeneralQuotientRingImpl::IamGCDDomain() const
  {
    if (IsZ(myReprRing)) return IsPrime(myReducingIdeal);
    return false; //??? BUG this is strictly wrong for Z[x]/ideal(x)
  }


  bool GeneralQuotientRingImpl::IamField() const
  {
    return IsMaximal(myReducingIdeal);
  }


  ConstRefRingElem GeneralQuotientRingImpl::myZero() const
  {
    return ConstRefRingElem(ring(this), raw(zero(myReprRing)));
  }


  ConstRefRingElem GeneralQuotientRingImpl::myOne() const
  {
    return ConstRefRingElem(ring(this), raw(one(myReprRing)));
  }


  RingElem GeneralQuotientRingImpl::myCanonicalRepr(ConstRawPtr rawx) const
  {
    RingElem ans(myReprRing);
    myReprRing->myAssign(raw(ans), rawx);
    return ans;
  }


  void GeneralQuotientRingImpl::myReduction(RawPtr rawimage, ConstRawPtr rawarg) const
  {
    myReprRing->myAssign(rawimage, rawarg);
    myReducingIdeal->myReduceMod(rawimage);
  }


  RingElemRawPtr GeneralQuotientRingImpl::myNew() const
  {
    return myReprRing->myNew();
  }


  RingElemRawPtr GeneralQuotientRingImpl::myNew(const MachineInteger& n) const
  {
    RingElemRawPtr rawans = myReprRing->myNew(n);
    myReduction(rawans);
    return rawans;
  }


  RingElemRawPtr GeneralQuotientRingImpl::myNew(const ZZ& N) const
  {
    RingElemRawPtr rawans = myReprRing->myNew(N);
    myReduction(rawans);
    return rawans;
  }


  RingElemRawPtr GeneralQuotientRingImpl::myNew(ConstRawPtr rawCopyMe) const
  {
    return myReprRing->myNew(rawCopyMe);
  }


  void GeneralQuotientRingImpl::myDelete(RawPtr rawx) const
  {
    myReprRing->myDelete(rawx);
  }


  void GeneralQuotientRingImpl::mySwap(RawPtr rawx, RawPtr rawy) const
  {
    myReprRing->mySwap(rawx, rawy);
  }


  void GeneralQuotientRingImpl::myAssign(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myReprRing->myAssign(rawlhs, rawx);
  }


  void GeneralQuotientRingImpl::myAssign(RawPtr rawlhs, const MachineInteger& n) const
  {
    myReprRing->myAssign(rawlhs, n);
    myReduction(rawlhs);
  }


  void GeneralQuotientRingImpl::myAssign(RawPtr rawlhs, const ZZ& N) const
  {
    myReprRing->myAssign(rawlhs, N);
    myReduction(rawlhs);
  }


  void GeneralQuotientRingImpl::myAssignZero(RawPtr rawlhs) const
  {
    myReprRing->myAssignZero(rawlhs); // assume zero is always reduced
  }


  void GeneralQuotientRingImpl::myNegate(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myReprRing->myNegate(rawlhs, rawx);
    myReduction(rawlhs);
  }


  void GeneralQuotientRingImpl::myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myReprRing->myAdd(rawlhs, rawx, rawy);
    myReduction(rawlhs);
  }


  void GeneralQuotientRingImpl::mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myReprRing->mySub(rawlhs, rawx, rawy);
    myReduction(rawlhs);
  }


  void GeneralQuotientRingImpl::myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myReprRing->myMul(rawlhs, rawx, rawy);
    myReduction(rawlhs);
  }


  void GeneralQuotientRingImpl::myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    RingElem X = myCanonicalRepr(rawx);
    RingElem Y = myCanonicalRepr(rawy);
    RingElem ans(myReprRing);
    myReducingIdeal->myDivMod(raw(ans), raw(X), raw(Y));
    myReduction(rawlhs, raw(ans));
  }


  bool GeneralQuotientRingImpl::myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    RingElem X = myCanonicalRepr(rawx);
    RingElem Y = myCanonicalRepr(rawy);
    RingElem ans(myReprRing);
    myReducingIdeal->myDivMod(raw(ans), raw(X), raw(Y));
    if (!myIsZero(rawx) && IsZero(ans)) return false;
    myReduction(rawlhs, raw(ans));
    return true;
  }


  // We could simply use the default definition of this function.
  bool GeneralQuotientRingImpl::myIsInvertible(ConstRawPtr rawx) const
  {
    ring R(myReprRing);
    RingElem xcopy(R);
    R->myAssign(raw(xcopy), rawx);
    return IsOne(myReducingIdeal + ideal(R, xcopy));
  }



  //  This seems hard to define in full generality; e.g. there are some GCD domains which are not euclidean.
  void GeneralQuotientRingImpl::myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
 {
   if (!IamGCDDomain())
     CoCoA_ERROR(ERR::NotGCDDomain, "gcd(lhs, x, y) in a quotient ring");
   if (IamField()) { myGcdInField(rawlhs, rawx, rawy); return; }

   //??? BUG INCOMPLETE BUG ???
   CoCoA_ERROR(ERR::NYI, "gcd in generic quotient ring which is not a field");
 }


  void GeneralQuotientRingImpl::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(n > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    myBinaryPower(rawlhs, rawx, n); //??? BUG/SLUG this is not always the best choice ???
  }


  void GeneralQuotientRingImpl::mySymbols(std::vector<symbol>& SymList) const
  {
    myReprRing->mySymbols(SymList);
  }


  void GeneralQuotientRingImpl::myOutput(std::ostream& out, ConstRawPtr rawx) const
  {
    // THIS PRINT FORMAT IS UGLY!  BUG???
    ConstRefRingElem x(myReprRing, rawx); // alias for value to be printed
    if (!IsZ(myReprRing))
    {
      out << "(" << x << ")";
      return;
    }
    if (!DefaultResiduesAreSymm())
    {
      out << x;
      return;
    }
    // Special case of Z/(nZ) and using symmetric residues...
    ZZ n; myCharacteristic(n);
    if (x > n/2)
      out << x-n;
    else
      out << x;
  }


  bool GeneralQuotientRingImpl::myIsPrintAtom(ConstRawPtr rawx) const
  {
    if (myIsPrintedWithMinus(rawx))  return false;
    return true;  // either positive int or between parentheses
  }


  bool GeneralQuotientRingImpl::myIsPrintedWithMinus(ConstRawPtr rawx) const
  {
    if (IsZ(myReprRing) && DefaultResiduesAreSymm() )
    {
      ConstRefRingElem x(myReprRing, rawx); // alias for value to be printed
      ZZ n; myCharacteristic(n);
      return x > n/2;
    }
    return false;    
  }
  

  void GeneralQuotientRingImpl::myOutputSelf(std::ostream& out) const
  {
    out << "QuotientRing(" << myBaseRing << ", " << myDefiningIdeal << ")";
  }


  void GeneralQuotientRingImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("cocoa", "QuotientRing");
    OMOut << myBaseRing;     // superfluous since the ideal already contains the ring info...
    OMOut << myDefiningIdeal;
    OMOut->mySendApplyEnd();
  }


  void GeneralQuotientRingImpl::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const
  {
    myReprRing->myOutput(OMOut, rawx);
  }


  bool GeneralQuotientRingImpl::myIsZero(ConstRawPtr rawx) const
  {
    return myReprRing->myIsZero(rawx);
  }


  bool GeneralQuotientRingImpl::myIsOne(ConstRawPtr rawx) const
  {
    return myReprRing->myIsOne(rawx) || IsOne(myReducingIdeal);
  }


  bool GeneralQuotientRingImpl::myIsMinusOne(ConstRawPtr rawx) const
  {
    if (myReprRing->myIsMinusOne(rawx) || IsOne(myReducingIdeal)) return true;
    RingElem tmp(myReprRing);
    myReprRing->myAdd(raw(tmp), raw(myOne()), rawx);
    return IsElem(tmp, myReducingIdeal);
  }


  bool GeneralQuotientRingImpl::myIsInteger(ZZ& N, ConstRawPtr rawx) const
  {
    return myReprRing->myIsInteger(N, rawx);
  }


  bool GeneralQuotientRingImpl::myIsRational(QQ& Q, ConstRawPtr rawx) const
  {
    return myReprRing->myIsRational(Q, rawx);
  }


  bool GeneralQuotientRingImpl::myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    return myReprRing->myIsEqual(rawx, rawy);
  }


  ideal GeneralQuotientRingImpl::myIdealCtor(const std::vector<RingElem>& gens) const
  {
    return ideal(new IdealImpl(QuotientRing(this), gens));
  }


  void GeneralQuotientRingImpl::myReduction(RawPtr rawx) const
  {
    myReducingIdeal->myReduceMod(rawx);
  }


  //---------------------------------------------------------------------------
  // Functions to do with homomorphisms

  const RingHom& GeneralQuotientRingImpl::myQuotientingHom() const
  {
    return myQuotHom;
  }


  RingHom GeneralQuotientRingImpl::myInducedHomCtor(const RingHom& InducingHom) const
  {
    CoCoA_ASSERT(IsInKer(myDefiningIdeal, InducingHom));
    return RingHom(new InducedHomImpl(QuotientRing(this), InducingHom));
  }


  RingHom GeneralQuotientRingImpl::myCompose(const RingHom& phi, const RingHom& theta) const
  {
    return myInducedHomCtor(phi(theta(myQuotientingHom())));
  }


  GeneralQuotientRingImpl::InducedHomImpl::InducedHomImpl(const QuotientRing& RmodI, const RingHom& InducingHom):
      RingHomInducedBase(RmodI, InducingHom)
  { /* Compatibility already checked in InducedHom in QuotientRing.C */ }


  void GeneralQuotientRingImpl::InducedHomImpl::myApply(RawPtr rawimage, ConstRawPtr rawarg) const
  {
    myInducingHom->myApply(rawimage, rawarg);
  }


  //---------------------------------------------------------------------------
  // Functions to do with GQRIdeals

  inline const GeneralQuotientRingImpl::IdealImpl* GeneralQuotientRingImpl::IdealImpl::GetPtr(const ideal& I)
  {
    return dynamic_cast<const IdealImpl*>(I.myIdealPtr());
  }


  GeneralQuotientRingImpl::IdealImpl::IdealImpl(const QuotientRing& RmodI, const vector<RingElem>& gens):
      myAmbientRingValue(RmodI),
      myGensValue(gens),
      myPreimageIdeal(ReducingIdeal(RmodI))
  {
    vector<RingElem> OverGens;
    for (size_t i=0; i < gens.size(); ++i)
      OverGens.push_back(RmodI->myCanonicalRepr(raw(gens[i])));
    myPreimageIdeal += ideal(ReprRing(RmodI), OverGens);
    SetTidyGensFromPreimageIdeal();
  }


  GeneralQuotientRingImpl::IdealImpl::~IdealImpl()
  {}


  GeneralQuotientRingImpl::IdealImpl* GeneralQuotientRingImpl::IdealImpl::clone() const
  {
    return new IdealImpl(*this);
  }


  // NB the fowarded call to myPreimageIdeal does all the consistency checking
  void GeneralQuotientRingImpl::IdealImpl::mySetPrimeFlag(bool b) const
  {
    myPreimageIdeal->mySetPrimeFlag(b);
    IamPrimeFlag = b;
  }


  // NB the fowarded call to myPreimageIdeal does all the consistency checking
  void GeneralQuotientRingImpl::IdealImpl::mySetMaximalFlag(bool b) const
  {
    myPreimageIdeal->mySetMaximalFlag(b);
    IamMaximalFlag = b;
  }


  const QuotientRing& GeneralQuotientRingImpl::IdealImpl::myAmbientRing() const
  {
    return myAmbientRingValue;
  }


  bool GeneralQuotientRingImpl::IdealImpl::IamZero() const
  {
    return myPreimageIdeal == ReducingIdeal(myAmbientRingValue);
  }


  bool GeneralQuotientRingImpl::IdealImpl::IamOne() const
  {
    return IsOne(myPreimageIdeal);
  }


  void GeneralQuotientRingImpl::IdealImpl::myMaximalTest() const
  {
    mySetMaximalFlag(IsMaximal(myPreimageIdeal));
  }


  void GeneralQuotientRingImpl::IdealImpl::myPrimeTest() const
  {
    mySetPrimeFlag(IsPrime(myPreimageIdeal));
  }


  void GeneralQuotientRingImpl::IdealImpl::myReduceMod(RawPtr rawx) const
  {
    if (IamZero()) return;
    myPreimageIdeal->myReduceMod(rawx);
  }


  bool GeneralQuotientRingImpl::IdealImpl::contains(RingElemConstRawPtr rawx) const
  {
    return myPreimageIdeal->contains(rawx);
  }


  void GeneralQuotientRingImpl::IdealImpl::add(const ideal& J)
  {
    copy(gens(J).begin(), gens(J).end(), back_inserter(myGensValue));
    myPreimageIdeal += GetPtr(J)->myPreimageIdeal;
    SetTidyGensFromPreimageIdeal();
    IamPrimeFlag = GetPtr(J)->IamPrimeFlag;
    IamMaximalFlag = GetPtr(J)->IamMaximalFlag;
  }


  void GeneralQuotientRingImpl::IdealImpl::mul(const ideal& J)
  {
    if (IamZero()) return;
    CoCoA_ERROR(ERR::NYI, "GeneralQuotientRingImpl::IdealImpl::mul");
  }


  void GeneralQuotientRingImpl::IdealImpl::intersect(const ideal& J)
  {
    MakeUnique(myPreimageIdeal)->intersect(GetPtr(J)->myPreimageIdeal);
    SetTidyGensFromPreimageIdeal();
    myGensValue = myTidyGensValue;
    IamPrimeFlag = GetPtr(J)->IamPrimeFlag;
    IamMaximalFlag = GetPtr(J)->IamMaximalFlag;
  }


  void GeneralQuotientRingImpl::IdealImpl::colon(const ideal& J)
  {
    MakeUnique(myPreimageIdeal)->colon(GetPtr(J)->myPreimageIdeal);
    SetTidyGensFromPreimageIdeal();
    myGensValue = myTidyGensValue;
    IamPrimeFlag = GetPtr(J)->IamPrimeFlag;
    IamMaximalFlag = GetPtr(J)->IamMaximalFlag;
    //??? CANNOT WORK OUT HOW TO USE THE ALGORITHM transform OR for_each :-(
    //transform(myPreimageIdeal->gens().begin(), myPreimageIdeal->gens().end(), back_inserter(myGens), myAmbientRing->myReduction);
  }


  void GeneralQuotientRingImpl::IdealImpl::myDivMod(RingElemRawPtr rawlhs, RingElemConstRawPtr rawnum, RingElemConstRawPtr rawden) const
  {
    const QuotientRing& RmodI = myAmbientRingValue;
    const ring& R = ReprRing(RmodI);
    RingElem N = RmodI->myCanonicalRepr(rawnum);
    RingElem D = RmodI->myCanonicalRepr(rawden);
    RingElem ans(R);
    myPreimageIdeal->myDivMod(raw(ans), raw(N), raw(D));
    RmodI->myReduction(rawlhs, raw(ans));
  }


  const std::vector<RingElem>& GeneralQuotientRingImpl::IdealImpl::myGens() const
  {
    return myGensValue;
  }


  const std::vector<RingElem>& GeneralQuotientRingImpl::IdealImpl::myTidyGens() const
  {
    return myTidyGensValue;
  }


  void GeneralQuotientRingImpl::IdealImpl::SetTidyGensFromPreimageIdeal()
  {
    vector<RingElem> NewTidyGens;
    const vector<RingElem>& OverGens = TidyGens(myPreimageIdeal);
    for (size_t i=0; i < OverGens.size(); ++i)
    {
      RingElem g(myAmbientRingValue);
      myAmbientRingValue->myReduction(raw(g), raw(OverGens[i]));
      if (!IsZero(g))
        NewTidyGens.push_back(g);
    }
    myTidyGensValue = NewTidyGens;
  }


  //---------------------------------------------------------------------------

  QuotientRing NewQuotientRing(const ring& R, const ideal& I)
  {
    if (R != AmbientRing(I)) CoCoA_ERROR(ERR::IdealNotInRing, "NewQuotientRing");
    if (IsOne(I)) CoCoA_ERROR(ERR::BadQuotRing, "NewQuotientRing");
//???    if (IsZero(I)) CoCoA_ERROR(ERR::BadQuotRing, "NewQuotientRing"); ???
    if (!IsZero(I) && IsZ(R))
    {
      // Try to make a RingFp first as it is fastest.
      try { return NewRingFp(I); }
      catch (const CoCoA::ErrorInfo& err) { if (err != ERR::BadSmallFpChar) throw; }
      // Couldn't make a RingFp, so try to make a RingFpDouble.
      try { return NewRingFpDouble(I); }
      catch (const CoCoA::ErrorInfo& err) { if (err != ERR::BadSmallFpChar) throw; }
      // Couldn't make RingFp or RingFpDouble (char not prime or too big),
      // so fall through to (relatively slow) generic implementation.
    }
    return QuotientRing(new GeneralQuotientRingImpl(R, I));
  }


  QuotientRing NewZmod(const MachineInteger& n)
  {
    return NewZmod(ZZ(n));
  }


  QuotientRing NewZmod(const ZZ& N)
  {
    if (N == 1 || N == -1) CoCoA_ERROR(ERR::BadQuotRing, "NewZmod");
    if (IsZero(N)) CoCoA_ERROR(ERR::BadQuotRing, "NewZmod"); //??? disallow N==0???
    return NewQuotientRing(RingZ(), ideal(RingZ(), N));
  }



  bool IsQuotientRing(const ring& R)
  {
    return dynamic_cast<const QuotientRingBase*>(R.myRawPtr()) != 0;
  }


  QuotientRing AsQuotientRing(const ring& R)
  {
    if (!IsQuotientRing(R))
      CoCoA_ERROR(ERR::NotQuotientRing, "QuotientRing(ring)");
    return QuotientRing(static_cast<const QuotientRingBase*>(R.myRawPtr()));
  }


  const RingHom& QuotientingHom(const QuotientRing& RmodI)
  {
    return RmodI->myQuotientingHom();
  }


  RingHom InducedHom(const QuotientRing& RmodI, const RingHom& phi)
  {
    if (domain(phi) != BaseRing(RmodI))
      CoCoA_ERROR(ERR::BadInducingHom, "InducedHom(QuotientRing,hom)");

    if (!IsInKer(DefiningIdeal(RmodI), phi))
      CoCoA_ERROR(ERR::BadInducingHomKer, "InducedHom(QuotientRing,hom)");

    return RmodI->myInducedHomCtor(phi);
  }


  RingElem CanonicalRepr(ConstRefRingElem r)
  {
    if (!IsQuotientRing(owner(r)))
      CoCoA_ERROR(ERR::NotElemQuotientRing, "CanonicalRepr(r)");
    return AsQuotientRing(owner(r))->myCanonicalRepr(raw(r));
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/QuotientRing.C,v 1.13 2010/02/04 09:57:11 bigatti Exp $
// $Log: QuotientRing.C,v $
// Revision 1.13  2010/02/04 09:57:11  bigatti
// -- added "mul" for ideals.  Implemented only for SparsePolyRing
//
// Revision 1.12  2009/12/03 17:40:36  abbott
// Added include directives for ZZ.H (as a consequence of removing
// the directive from ring.H).
//
// Revision 1.11  2009/09/24 15:36:00  abbott
// Removed some unnecessary "std::" prefixes.
// Removed some pointless include directives.
//
// Revision 1.10  2009/07/24 14:19:23  abbott
// Cleaned up include directives, and added fwd decls to the header file.
//
// Revision 1.9  2009/07/02 16:32:11  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.8  2009/05/22 10:31:41  bigatti
// -- added myIsPrintAtom, myIsPrintedWithMinus
// -- improved myOutput
//
// Revision 1.7  2009/05/14 09:39:29  abbott
// Added possibility to specify "symmetric" or "non-negative" residues
// in quotients of ZZ.  Affects printing of elements in quotients of ZZ
// (also changed printing of elements in general quotient rings).
// Consequent changes in several tests.
//
// Revision 1.6  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.5  2008/10/07 15:45:22  abbott
// Changed ErrorInfo objects so they include the name of their own error ID.
// Changed catch statements to catch const objects.
// Removed calls to the member fn which accessed the error ID member of an
// ErrorInfo; now you simply compare directly with the error ID (makes the
// code easier to read).
//
// Revision 1.4  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/09/25 16:32:30  abbott
// Several minor changes to silence gcc-4.3:
//    more #includes,
//    and fixed a template problemm in RegisterServerOps.C
//
// Revision 1.2  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.14  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.13  2007/03/08 10:23:29  bigatti
// -- CanonHom --> CanonicalHom
//
// Revision 1.12  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.11  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.10  2007/02/26 15:00:54  bigatti
// -- added placeholders for new syntax based on unique Z implementation
//
// Revision 1.9  2007/01/15 15:47:57  cocoa
// -- added prefix "raw" to RawPtr arguments names
// -- changed rhs into rawx, n, or N
//
// Revision 1.8  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.7  2006/12/21 13:48:32  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.6  2006/12/18 15:16:05  cocoa
// Added definition of CanonRepr.
//
// Revision 1.5  2006/12/07 11:57:53  cocoa
// -- style: RawPtr args are now called "raw.."
//
// Revision 1.4  2006/11/08 16:21:59  cocoa
// Structural cleaning of RingHom; many consequential changes.
//
// Revision 1.3  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.2  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.9  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.8  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.7  2006/04/21 15:01:36  cocoa
// Changed default implementation of RingBase::myGcd -- it now gives a SERIOUS
// error.  All fields must now handle a call to gcd explicitly: they can use
// the new myGcdInField function.  It's now cleaner than it was.
//
// Revision 1.6  2006/03/27 12:21:25  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.5  2006/03/21 09:43:14  cocoa
// Changed names of some member fns of ideals (dealing with setting and testing
// the flags for primeness and maximality).  Hope icc will complain less now.
//
// Revision 1.4  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.3  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.5  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.4  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.3  2005/09/30 15:03:39  cocoa
// Minor cleaning and tidying.
// DistrMPolyInlPP: use of summands now rather cleaner.
//
// Revision 1.2  2005/07/08 15:09:28  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.5  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.4  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.3  2005/04/01 16:18:20  cocoa
// Friday check-in.  Fixed a bug in the ctor for GeneralQuotientRingImpl.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.19  2004/11/19 15:44:27  cocoa
// Changed names of "casting" functions which convert a ring into
// one with a more special structure (e.g. FractionField).  These
// functions now have names starting with "As".  There were several
// consequential changes.
//
// Revision 1.18  2004/11/18 18:33:41  cocoa
// Now every ring know its own "one" element (as well as "zero").
// Several consequential changes.
//
// Revision 1.17  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.16  2004/11/09 15:50:54  cocoa
// -- changed GeneralQuotientRing --> GeneralQuotientRingImpl
// -- added: NewZmod(Z, N)  with N in ZZ
//
// Revision 1.15  2004/11/04 18:47:43  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.14  2004/07/27 16:03:39  cocoa
// Added IsCommutative test and IamCommutative member function
// to all rings.  Tidied geobuckets a little.
//
// Revision 1.13  2004/07/16 15:45:12  cocoa
// First stage of new RingElem implementation completed.
//
// Revision 1.12  2004/07/13 16:32:27  cocoa
// First stage of major revamp of ring elements.
// Implementation of RingFp has been split into "ring interface"
// and "algorithms plus data structures".
//
// Revision 1.11  2004/06/30 16:46:06  cocoa
// End of day check-in: mostly tidying up, and correcting some
// overly lax access permissions.
//
// Revision 1.10  2004/06/29 17:10:22  cocoa
// Partially tidied use of "protected" and "private" in various
// base classes.  Checking in at the end of the day -- it works,
// and I wouldn't want it to be lost next time point's disk
// misbehaves.
//
// Revision 1.9  2004/06/29 15:34:29  cocoa
// Tidied the functions dealing with bool3 values.
//
// Revision 1.8  2004/05/24 15:52:14  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.7  2004/04/08 15:33:34  cocoa
// Added function IsInteger, and the related RingBase::myIsInteger
// virtual function, plus all necessary implementations.
//
// Revision 1.6  2004/03/20 17:46:10  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.5  2004/02/03 16:16:21  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.4  2004/01/28 15:32:48  cocoa
// Tidied use of C++ std library components, and reordered some #includes.
//
// Revision 1.3  2003/10/17 10:51:06  cocoa
// Major cleaning, and new naming convention.
//
// Revision 1.2  2003/10/09 12:16:38  cocoa
// New coding convention for rings.
//
// Revision 1.2  2003/06/23 16:48:54  abbott
// Minor cleaning prior to public release.
// Mostly consequential changes.
// Fixed NewQuotientRing to work properly in limit cases (e.g. R/ideal(0)).
//
// Revision 1.1  2003/05/14 17:12:01  abbott
// Initial revision
//
//
