//   Copyright (c)  2006,2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomBitStream.H"
#include "CoCoA/error.H"

#include <iostream>
using std::ostream;
//#include <bitset>
using std::bitset;
#include <time.h>
//using std::time; // to randomize automatically when the seed is 0.

namespace CoCoA
{

  // Static const data member.
  const size_t RandomBitStream::ourBufferBits;  // Just to keep the compiler quiet.


  RandomBitStream::RandomBitStream(const MachineInteger& seed):
      myBitIndex(0),
      myBuffer(0),
      myCounter(0),
      mySeed((IsZero(seed)?time(0):abs(seed))),
      myState() // init in body
  {
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myFillBuffer();
  }

  RandomBitStream::RandomBitStream(const RandomBitStream& copy):
      myBitIndex(copy.myBitIndex),
      myBuffer(copy.myBuffer),
      myCounter(copy.myCounter),
      mySeed(copy.mySeed),
      myState() //init in body
  {
    gmp_randinit_set(myState, copy.myState);
  }

  RandomBitStream& RandomBitStream::operator=(const RandomBitStream& rhs)
  {
    if (this == &rhs) return *this; // this is necessary (because of call to gmp_randclear)
    myBitIndex = rhs.myBitIndex;
    myBuffer = rhs.myBuffer;
    myCounter = rhs.myCounter;
    mySeed = rhs.mySeed;
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //

    return *this;
  }

  RandomBitStream::~RandomBitStream()
  {
    gmp_randclear(myState);
  }


  // Better define this in case some idiot wants to use it...
  // Not inline, so it is as slow as possible >-}
  RandomBitStream RandomBitStream::operator++(int)
  {
    RandomBitStream ans(*this); // Probably not cheap!
    operator++();
    return ans; // Makes another copy upon returning!
  }


  size_t RandomBitStream::myIndex() const
  {
    return myCounter;
  }


  void RandomBitStream::myFillBuffer()
  {
    myBuffer = gmp_urandomb_ui(myState, ourBufferBits);
    myBitIndex = 0;
  }


//   // OBSOLESCENT !!!
//   bool SampleBool(RandomBitStream& RBS)
//   {
//     return sample(RBS);
//   }


//   // OBSOLESCENT !!!
//   long SampleLong(RandomBitStream& RBS)
//   {
//     if (RBS.myBitIndex > 0)
//     {
//       RBS.myCounter += RandomBitStream::ourBufferBits - RBS.myBitIndex;
//       RBS.myFillBuffer();
//     }
//     const unsigned long ans = RBS.myBuffer.to_ulong();
//     RBS.myCounter += RandomBitStream::ourBufferBits;
//     RBS.myFillBuffer();
//     return ans; // compiler casts to signed long
//   }


  // This would be cleaner if I were to use an unsigned long instead of a double.
  bool prob(double P, RandomBitStream& RBS)
  {
    if (P < 0 || P > 1) CoCoA_ERROR(ERR::BadProbability, "prob(P,RBS)");
    if (P == 1) return true;
    while (true)
    {
      if (P == 0) return false;
      const bool bit = (P >= 0.5);
      if (bit ^ sample(RBS)) return bit;
      P *= 2;
      if (bit) --P;
    }
  }


  ostream& operator<<(ostream& out, const RandomBitStream& RBS)
  {
    out << "RandomBitStream(seed=" << RBS.mySeed << ", counter=" << RBS.myCounter << ")";
    return out;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomBitStream.C,v 1.4 2010/02/16 10:20:58 abbott Exp $
// $Log: RandomBitStream.C,v $
// Revision 1.4  2010/02/16 10:20:58  abbott
// Added new fn sample; SampleBool & SampleLong are obsolescent.
//
// Revision 1.3  2010/02/01 22:41:02  abbott
// Removed myMPZ data field which is no longer needed.
// Changed a couple of calls to GMP, after havong found
// ones which do directly what I need.
//
// Revision 1.2  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1  2007/06/06 15:16:48  abbott
// Added new RandomBitStream class (now based upon GMP's random generator).
// Consequential changes to Makefiles, etc.  There's even doc and an example!
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.2  2006/10/27 19:01:17  cocoa
// Corrected a spelling mistake: Steam has become Stream :-)
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/04/27 13:45:30  cocoa
// Changed name of NewIdentityRingHom to NewIdentityHom.
// Changed name of member functions which print out their own object
// into myOutputSelf (to distinguish from "transitive" myOutput fns).
//
// Revision 1.2  2006/04/21 15:58:55  cocoa
// Substituted non-portable used of srandomdev (works only BSD flavours).
//
// Revision 1.1  2006/04/14 13:51:40  cocoa
// Added a class for generating random bits (RandomBitStream in the file random.H).
//
