//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomLongStream.H"
#include "CoCoA/error.H"

#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits; // only for myGenValue
#include <time.h>
//using std::time; // to randomize automatically when the seed is 0.

namespace CoCoA
{

  RandomLongStream::RandomLongStream(const MachineInteger& lwb, const MachineInteger& upb, const MachineInteger& seed):
      myLwb(AsSignedLong(lwb)),
      myUpb(AsSignedLong(upb)),
      myRange(1+static_cast<unsigned long>(myUpb)-static_cast<unsigned long>(myLwb)), // any overflow is harmless (see myGenValue)
      myValue(0), // initialized "properly" in the body
      myCounter(0),
      mySeed((IsZero(seed)?time(0):abs(seed))),
      myState() // init in body
  {
    if (myLwb >= myUpb)
      CoCoA_ERROR(ERR::BadArg, "RandomLongStream: must have lwb < upb");
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myGenValue();
  }

  RandomLongStream::RandomLongStream(const RandomLongStream& copy):
      myLwb(copy.myLwb),
      myUpb(copy.myUpb),
      myRange(copy.myRange),
      myValue(copy.myValue),
      myCounter(copy.myCounter),
      mySeed(copy.mySeed),
      myState() //init in body
  {
    gmp_randinit_set(myState, copy.myState);
  }

  RandomLongStream& RandomLongStream::operator=(const RandomLongStream& rhs)
  {
    if (this == &rhs) return *this; // this is necessary (because of call to gmp_randclear)
    myLwb = rhs.myLwb;
    myUpb = rhs.myUpb;
    myRange = rhs.myRange;
    myValue = rhs.myValue;
    myCounter = rhs.myCounter;
    mySeed = rhs.mySeed;
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //

    return *this;
  }

  RandomLongStream::~RandomLongStream()
  {
    gmp_randclear(myState);
  }


  // Better define this in case some idiot wants to use it...
  // Not inline, so it is as slow as possible >-}
  RandomLongStream RandomLongStream::operator++(int)
  {
    RandomLongStream ans(*this); // Probably not cheap!
    operator++();
    return ans; // Makes another copy upon returning!
  }


  size_t RandomLongStream::myIndex() const
  {
    return myCounter;
  }


  void RandomLongStream::myGenValue()
  {
    if (myRange > 0)
      myValue = myLwb + gmp_urandomm_ui(myState, myRange); // any overflow is harmless
    else
      // Get here only if myLwb = min_long & myUpb = max_long
      myValue = myLwb + gmp_urandomb_ui(myState, numeric_limits<unsigned long>::digits);
  }


  ostream& operator<<(ostream& out, const RandomLongStream& RLS)
  {
    out << "RandomLongStream(lwb=" << RLS.myLwb
        << ", upb=" << RLS.myUpb
        << ", seed=" << RLS.mySeed
        << ", counter=" << RLS.myCounter << ")";
    return out;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomLongStream.C,v 1.1 2010/02/16 10:19:29 abbott Exp $
// $Log: RandomLongStream.C,v $
// Revision 1.1  2010/02/16 10:19:29  abbott
// Added new class RandomLongStream with example and test.
//
//
