//   Copyright (c)  2007  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for RingDenseUPolyCleanImpl

#include "CoCoA/RingDenseUPolyClean.H"
#include "CoCoA/DenseUPolyRing.H"
#include "CoCoA/DenseUPolyClean.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/ZZ.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"
#include "CoCoA/io.H" // only for debugging
#include "CoCoA/ring.H"
#include "CoCoA/symbol.H"

// #include <cstddef> // from MemPool.H
using std::size_t;
#include <iostream>
using std::ostream;
// #include <memory>   // from MemPool.H
using std::auto_ptr;
#include <vector>
using std::vector;


namespace CoCoA
{
  class RingDenseUPolyCleanImpl: public DenseUPolyRingBase
  {
  private:
    typedef DenseUPolyClean value_t; // DenseUPolyClean is the actual type of the values in a RingDenseUPolyCleanImpl
    static value_t& import(RingElemRawPtr rawf);
    static const value_t& import(RingElemConstRawPtr rawf);

  public:
    RingDenseUPolyCleanImpl(const ring& R, const symbol& x, std::size_t MinCapacity);
    ~RingDenseUPolyCleanImpl();

  private: // Data members of RingDenseUPolyCleanImpl
    const ring myCoeffRingValue;  ///< the coefficient ring
    mutable MemPool myDUPPool; ///< memory manager for polynomials
    std::auto_ptr<RingElem> myZeroPtr;  ///< Every ring stores its own zero.
    std::auto_ptr<RingElem> myOnePtr;   ///< Every ring stores its own one.
    std::vector<RingElem> myIndetVector; ///< Vector for compatibility with SparsePolyRing ???
    symbol myIndetNameValue; ///< the indet name
    std::auto_ptr<RingHom> myCoeffEmbeddingHomPtr;
    std::size_t myMinCapacity;  // the minumum capacity for all coeff vectors

  public:  // functions which every ring must implement
    virtual void myCharacteristic(ZZ& p) const;
    virtual bool IamCommutative() const;
    virtual bool IamIntegralDomain() const;
    virtual bool IamGCDDomain() const;
    virtual bool IamField() const;
    virtual ConstRefRingElem myZero() const;
    virtual ConstRefRingElem myOne() const;
    virtual RingElemRawPtr myNew() const;
    virtual RingElemRawPtr myNew(const MachineInteger& n) const;
    virtual RingElemRawPtr myNew(const ZZ& N) const;
    virtual RingElemRawPtr myNew(ConstRawPtr rawcopy) const;
    virtual void myDelete(RawPtr rawx) const;                             // destroys x (incl all resources)
    virtual void mySwap(RawPtr rawx, RawPtr rawy) const;                  // swap(x, y)
    virtual void myAssign(RawPtr rawlhs, ConstRawPtr rawx) const;         // lhs = x
    virtual void myAssign(RawPtr rawlhs, const MachineInteger& n) const;  // lhs = n
    virtual void myAssign(RawPtr rawlhs, const ZZ& N) const;              // lhs = N
    virtual void myAssignZero(RawPtr rawlhs) const;                       // lhs = 0
    virtual void myNegate(RawPtr rawlhs, ConstRawPtr rawx) const;         // lhs = -x
    virtual void myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; // lhs = x+y
    virtual void mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; // lhs = x-y
    //    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const; // lhs = x^n, n>1, x not -1,0,1
    virtual void myOutputSelf(std::ostream& out) const;      // out << R
    virtual void myOutputSelf(OpenMathOutput& OMOut) const;  // OMOut << R
    //    virtual bool myIsZeroAddMul: use default definition
    void myAddMul(RawPtr rawf, ConstRawPtr rawc, std::size_t d, ConstRawPtr rawg) const;

    // functions which every PolyRing must implement
    virtual const ring& myCoeffRing() const;
    virtual const std::vector<RingElem>& myIndets() const;

    ///@name Simple functions on polynomials
    //@{
    virtual void myMulByCoeff(RawPtr rawf, ConstRawPtr rawc) const; ///< WEAK EXCEPTION GUARANTEE
    virtual bool myDivByCoeff(RawPtr rawf, ConstRawPtr rawc) const; ///< WEAK EXCEPTION GUARANTEE
    virtual void myDeriv(RawPtr rawlhs, ConstRawPtr rawf, ConstRawPtr rawx) const; ///< lhs = deriv(f, x)
    //@}
    const RingHom& myCoeffEmbeddingHom() const;

    //----------------------------------------------------------------------
    // Functions which every DenseUPolyRing must implement:
    //----------------------------------------------------------------------

    ///@name Member functions every concrete DenseUPolyRing implementation must have in addition to those of PolyRingBase.
    //@{
    // ANNA ??? some of these should be in PolyRing
    virtual const symbol& myIndetName() const;
    virtual std::size_t myDegPlus1(ConstRawPtr rawf) const;          ///<  standard degree+1 of f, 0 for zero poly
    virtual std::size_t mySize(ConstRawPtr rawf) const;              ///<  size of f
    virtual void myMulByXExp(RawPtr rawf, unsigned long n) const;
    virtual void myMulBy1MinusXExp(RawPtr rawf, unsigned long n) const;
    virtual void myResize(RawPtr rawf, std::size_t NewSize) const;
    virtual void myResetDeg(RawPtr rawf) const; ///< reset the correct value of deg (assumed to be less than the current value)
    virtual void myAssignZeroCoeff(RawPtr rawf, std::size_t d) const; ///< f_d = 0, no check on size nor degree
    virtual void myAssignNonZeroCoeff(RawPtr rawf, ConstRawPtr rawc, std::size_t d) const; ///< f_d = c, no check on size nor degree
    virtual ConstRefRingElem myCoeff(ConstRawPtr rawf, std::size_t d) const;
    //@}

    ///@name   Functions for creating/building polynomials
    //@{
    //@}

  private: // homomorphism class
    class HomImpl: public RingHomBase
    {
    public:
      HomImpl(const ring& codomain);
      void myApply(RawPtr rawimage, ConstRawPtr rawarg) const;
    };

  };


  //----------------------------------------------------------------------

  inline RingDenseUPolyCleanImpl::value_t& RingDenseUPolyCleanImpl::import(RingElemRawPtr rawf)
  {
    return *static_cast<value_t*>(rawf.myRawPtr());
  }
  
  inline const RingDenseUPolyCleanImpl::value_t& RingDenseUPolyCleanImpl::import(RingElemConstRawPtr rawf)
  {
    return *static_cast<const value_t*>(rawf.myRawPtr());
  }
  

  RingDenseUPolyCleanImpl::RingDenseUPolyCleanImpl(const ring& R, const symbol& x, std::size_t MinCapacity):
    myCoeffRingValue(R),
    myDUPPool(sizeof(DenseUPolyClean), "RingDenseUPolyCleanImpl::myDUPPool"),
    myIndetNameValue(x)
  {
    CoCoA_ASSERT(IsCommutative(myCoeffRingValue));
    myMinCapacity = MinCapacity;
    myZeroPtr.reset(new RingElem(ring(this)));
    myOnePtr.reset(new RingElem(ring(this), 1));
    myIndetVector.resize(1, *myZeroPtr);
    import(raw(myIndetVector[0])).myResize(2);  // deg+1
    import(raw(myIndetVector[0])).myAssignNonZeroCoeff(one(R), 1);
    myCoeffEmbeddingHomPtr.reset(new RingHom(new CoeffEmbeddingHomImpl(DenseUPolyRing(this))));
    myRefCountZero(); // otherwise it is 2 + NumIndets and won't be destroyed
  }


  RingDenseUPolyCleanImpl::~RingDenseUPolyCleanImpl()
  {}


  //----------------------------------------------------------------------
  // Functions which every ring must implement:
  //----------------------------------------------------------------------

  void RingDenseUPolyCleanImpl::myCharacteristic(ZZ& p) const
  {
    myCoeffRingValue->myCharacteristic(p);
  }


  bool RingDenseUPolyCleanImpl::IamCommutative() const
  {
    return true;  // we are assuming that the coeff ring is commutative
  }


  bool RingDenseUPolyCleanImpl::IamIntegralDomain() const
  {
    return IsIntegralDomain(myCoeffRingValue);
  }


  bool RingDenseUPolyCleanImpl::IamGCDDomain() const
  {
    return IsGCDDomain(myCoeffRingValue);
  }


  bool RingDenseUPolyCleanImpl::IamField() const
  {
    return false; //??? (myNumIndetsValue==0 && IsField(myCoeffRingValue))
  }


  ConstRefRingElem RingDenseUPolyCleanImpl::myZero() const
  {
    return *myZeroPtr;
  }


  ConstRefRingElem RingDenseUPolyCleanImpl::myOne() const
  {
    return *myOnePtr;
  }

  //---- RingDenseUPolyClean specific functions with RingElem

  RingElemRawPtr RingDenseUPolyCleanImpl::myNew() const
  {
    void* ptr = myDUPPool.alloc();
    new(ptr) DenseUPolyClean(myCoeffRingValue, myMinCapacity); // placement new
    return RingElemRawPtr(ptr);
  }


  RingElemRawPtr RingDenseUPolyCleanImpl::myNew(const MachineInteger& n) const
  {
    if (IsZero(n)) return myNew();  // not really necessary
    auto_ptr<DenseUPolyClean> ans(new(myDUPPool.alloc()) DenseUPolyClean(myCoeffRingValue, myMinCapacity)); // placement new
    *ans = n;
    return RingElemRawPtr(ans.release());
  }


  RingElemRawPtr RingDenseUPolyCleanImpl::myNew(const ZZ& N) const
  {
    if (N == 0) return myNew();  // not really necessary
    auto_ptr<DenseUPolyClean> ans(new(myDUPPool.alloc()) DenseUPolyClean(myCoeffRingValue, myMinCapacity)); // placement new
    *ans = N;
    return RingElemRawPtr(ans.release());
  }


  RingElemRawPtr RingDenseUPolyCleanImpl::myNew(ConstRawPtr rawcopy) const
  {
    auto_ptr<DenseUPolyClean> ans(new(myDUPPool.alloc()) DenseUPolyClean(import(rawcopy), myMinCapacity)); // placement new
    //    *ans = import(rawcopy);
    return RingElemRawPtr(ans.release());
  }


  void RingDenseUPolyCleanImpl::myDelete(RawPtr rawx) const
  {
    import(rawx).~DenseUPolyClean();
    myDUPPool.free(rawx.myRawPtr());
  }


  void RingDenseUPolyCleanImpl::mySwap(RawPtr rawx, RawPtr rawy) const
  {
    swap(import(rawx), import(rawy));
  }


  void RingDenseUPolyCleanImpl::myAssign(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    import(rawlhs) = import(rawx);
  }


  void RingDenseUPolyCleanImpl::myAssign(RawPtr rawlhs, const MachineInteger& n) const
  {
    import(rawlhs) = n;
  }


  void RingDenseUPolyCleanImpl::myAssign(RawPtr rawlhs, const ZZ& N) const
  {
    import(rawlhs) = N;
  }


  void RingDenseUPolyCleanImpl::myAssignZero(RawPtr rawlhs) const
  {
    import(rawlhs).myAssignZero();
  }


  void RingDenseUPolyCleanImpl::myNegate(RawPtr rawlhs, ConstRawPtr rawx) const
  {
//     rawlhs.clear();
//     transform(rawx.begin(), rawx.end(), back_inserter(rawlhs), myNegate());
    import(rawlhs) = import(rawx);
    import(rawlhs).myNegate();
  }


  void RingDenseUPolyCleanImpl::myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    add(import(rawlhs), import(rawx), import(rawy));
  }


  void RingDenseUPolyCleanImpl::mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    sub(import(rawlhs), import(rawx), import(rawy));
  }


//   void RingDenseUPolyCleanImpl::myDiv(RawPtr /*rawlhs*/, ConstRawPtr /*rawx*/, ConstRawPtr rawy) const
//   {
//     CoCoA_ASSERT(!myIsZero(rawy));
//     CoCoA_ERROR(ERR::NYI, "RingDenseUPolyCleanImpl::myDiv");
//   }


//   bool RingDenseUPolyCleanImpl::myIsDivisible(RawPtr /*rawlhs*/, ConstRawPtr /*rawx*/, ConstRawPtr rawy) const
//   {
//     CoCoA_ASSERT(!myIsZero(rawy));
//     CoCoA_ERROR(ERR::NYI, "RingDenseUPolyCleanImpl::myDiv");
//     return true;
//   }


//   void RingDenseUPolyCleanImpl::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
//   {
//     // Assert that we have a genuinely non-trivial case.
//     CoCoA_ASSERT(n > 1);
//     CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
//     size_t NoUse;
//     if (myIsIndet(NoUse, rawx))
//       myIndetPower(rawlhs, 0, n);
//     else
//       CoCoA_ERROR(ERR::NYI, "RingDenseUPolyCleanImpl::myPowerSmallExp");
//   }


  void RingDenseUPolyCleanImpl::myOutputSelf(std::ostream& out) const
  {
    out << "RingDenseUPolyCleanImpl(" << myCoeffRingValue << ")";
  }


  void RingDenseUPolyCleanImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("setname1", "RingDenseUPolyCleanImpl");
    OMOut << myCoeffRingValue;
    OMOut->mySendApplyEnd();
  }


  const symbol& RingDenseUPolyCleanImpl::myIndetName() const
  {
    return myIndetNameValue;
  }

    
  void RingDenseUPolyCleanImpl::myResize(RawPtr rawf, std::size_t NewSize) const
  {
    CoCoA_ASSERT(NewSize < 4094967295ul);  // to catch silly mistakes
    if (NewSize > myDegPlus1(rawf))  // ??? should this check be here?
    //  (NewSize > mySize())
      import(rawf).myResize(NewSize);
  }


  void RingDenseUPolyCleanImpl::myResetDeg(RawPtr rawf) const
  {
    import(rawf).myResetDeg();
  }


  std::size_t RingDenseUPolyCleanImpl::mySize(ConstRawPtr rawf) const
  {
    return import(rawf).mySize();
  }


  std::size_t RingDenseUPolyCleanImpl::myDegPlus1(ConstRawPtr rawf) const
  {
    return import(rawf).myDegPlus1();
  }


  void RingDenseUPolyCleanImpl::myAssignZeroCoeff(RawPtr rawf, std::size_t d) const
  {
    import(rawf).myAssignZeroCoeff(d);
  }


  void RingDenseUPolyCleanImpl::myAssignNonZeroCoeff(RawPtr rawf, ConstRawPtr rawc, std::size_t d) const
  {
    import(rawf).myAssignNonZeroCoeff(ConstRefRingElem(myCoeffRingValue,rawc), d);
  }


  ConstRefRingElem RingDenseUPolyCleanImpl::myCoeff(ConstRawPtr rawf, std::size_t d) const
  {
    return import(rawf).myCoeff(d);
  }


  //----------------------------------------------------------------------
  // Functions which every PolyRing must implement
  //----------------------------------------------------------------------

  const ring& RingDenseUPolyCleanImpl::myCoeffRing() const
  {
    return myCoeffRingValue;
  }


  const std::vector<RingElem>& RingDenseUPolyCleanImpl::myIndets() const
  {
    return myIndetVector;
  }


  void RingDenseUPolyCleanImpl::myAddMul(RawPtr rawf, ConstRawPtr rawc, std::size_t d, ConstRawPtr rawg) const
  {   // not exception clean  f += c*indet^d*g
    import(rawf).myAddMul(ConstRefRingElem(myCoeffRing(),rawc), d, import(rawg));
  }


  void RingDenseUPolyCleanImpl::myMulByXExp(RawPtr rawf, unsigned long n) const
  {
    import(rawf).myMulByXExp(n);
  }


  void RingDenseUPolyCleanImpl::myMulBy1MinusXExp(RawPtr rawf, unsigned long n) const
  {
    import(rawf).myMulBy1MinusXExp(n);
  }


  bool RingDenseUPolyCleanImpl::myDivByCoeff(RawPtr, ConstRawPtr) const
  {
    CoCoA_ERROR(ERR::NYI, "RingDenseUPolyCleanImpl");
    return true;
  }
  
  
  void RingDenseUPolyCleanImpl::myMulByCoeff(RawPtr rawf, ConstRawPtr rawc) const
  {
    import(rawf).myMulByCoeff(ConstRefRingElem(myCoeffRing(),rawc));
  }


  void RingDenseUPolyCleanImpl::myDeriv(RawPtr /*rawlhs*/, ConstRawPtr /*rawf*/, ConstRawPtr /*rawx*/) const
  {
    CoCoA_ERROR(ERR::NYI, "RingDenseUPolyCleanImpl");
  }


  const RingHom& RingDenseUPolyCleanImpl::myCoeffEmbeddingHom() const
  {
    return *myCoeffEmbeddingHomPtr;
  }


  //----------------------------------------------------------------------
  // Pseudo-ctors for polynomial rings.

  DenseUPolyRing NewPolyRing_DUP(const ring& CoeffRing, const symbol& IndetName, std::size_t MinCapacity)
  {
    if (!IsCommutative(CoeffRing))
      CoCoA_ERROR(ERR::NotCommutative, "NewPolyRing_DUP(R, x) pseudo ctor");
    if (!IsGoodIndetName(CoeffRing, IndetName))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPolyRing_DUP(R, x) pseudo ctor");
    return DenseUPolyRing(new RingDenseUPolyCleanImpl(CoeffRing, IndetName, MinCapacity));
  }

  DenseUPolyRing NewPolyRing_DUP(const ring& CoeffRing, const symbol& IndetName)
  {
    return NewPolyRing_DUP(CoeffRing, IndetName, 10);
  }

  DenseUPolyRing NewPolyRing_DUP(const ring& CoeffRing)
  {
    return NewPolyRing_DUP(CoeffRing, symbol("x"));
  }
  

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RingDenseUPolyClean.C,v 1.12 2009/12/03 17:40:36 abbott Exp $
// $Log: RingDenseUPolyClean.C,v $
// Revision 1.12  2009/12/03 17:40:36  abbott
// Added include directives for ZZ.H (as a consequence of removing
// the directive from ring.H).
//
// Revision 1.11  2009/10/02 13:47:07  bigatti
// -- myDivByCoeff now returns bool
// -- unique implementation of myDiv in PolyRing.C
//
// Revision 1.10  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.9  2008/06/30 17:15:03  abbott
// Commented out unused parameter names (to avoid annoying compiler warnings).
//
// Revision 1.8  2007/12/21 12:29:08  bigatti
// -- abstract implementation in DenseUPolyRing of myDiv, myIsDivisible, myIsInvertible, myGcd
// -- abstract implementation in DenseUPolyRing of ideal operations
// -- some cleaning
//
// Revision 1.7  2007/12/04 13:44:50  bigatti
// -- fixed problem with myMinCapacity in constructor
// -- commented out myPowerSmallExp in RingDenseUPolyCleanImpl
// -- refined myPowerSmallExp in DenseUPolyRing
//
// Revision 1.6  2007/11/29 17:42:21  bigatti
// -- fixed: myMinCapacity initialization moved up in RingDenseUPolyCleanImpl ctor
//
// Revision 1.5  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.4  2007/10/19 10:04:23  bigatti
// -- RingDenseUPolyClean now allow to specify the MinCapacity for all
//    coeff vectors (to avoid too many reallocations)
//
// Revision 1.3  2007/10/10 14:02:37  bigatti
// -- added myMulBy1MinusXExp
// -- fixed a few little bugs
//
// Revision 1.2  2007/10/05 16:00:55  bigatti
// -- changed argument of myMulByXExp to std::size_t fon compatibility,
//    but they all have to be reconsidered
//
// Revision 1.1  2007/10/05 15:28:55  bigatti
// -- added abstract class DenseUPolyRing for representing dense
//    univariate polynomials
// -- added concrete class RingDenseUPolyClean, the cleanest
//    implementation
// -- just for testing, still horribly inefficient and incomplete
//
