//   Copyright (c)  2005-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RingFpDouble.H"

#include "CoCoA/FieldIdeal.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/QQ.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/SmallFpDoubleImpl.H"
#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/ideal.H"


#include <algorithm>
//using std::swap;         // only in mySwap
#include <iostream>
using std::ostream;        // only in myOutput
//#include <limits>  ---  included in MachineInteger.H (included via QQ.H)
using std::numeric_limits; // only in ctor
//#include <memory>  ---  included in MemPool.H
using std::auto_ptr;
//#include <vector>  ---  included in ideal.H
using std::vector;


namespace CoCoA
{

  /*-----------------------------------------------------------------*/
  /** \include RingFpDouble.txt  */
  /*-----------------------------------------------------------------*/
  class RingFpDoubleImpl: public QuotientRingBase
  {
  private: // data members
    typedef SmallFpDoubleImpl::value_t value_t;
    const unsigned long myModulus;
    const SmallFpDoubleImpl myImpl;
    mutable MemPool myMemMgr;       // MemPool must come BEFORE myZeroPtr and myOnePtr
    auto_ptr<RingElem> myZeroPtr;  ///< Every ring stores its own zero.
    auto_ptr<RingElem> myOnePtr;   ///< Every ring stores its own one.
    auto_ptr<RingHom> myQuotientingHomPtr;

  private: // auxiliary functions
    static unsigned long PrincipalGen(const ideal& I); // used for arg checking in ctor
    static value_t& import(RingElemRawPtr rawx);
    static const value_t& import(RingElemConstRawPtr rawx);

  private:
    RingFpDoubleImpl(const ideal& I); // called only by NewRingFpDouble
    ~RingFpDoubleImpl();
    friend QuotientRing NewRingFpDouble(const MachineInteger& p);
    friend QuotientRing NewRingFpDouble(const ZZ& P);
    friend QuotientRing NewRingFpDouble(const ideal& I);
  private:
    RingFpDoubleImpl(const RingFpDoubleImpl&);            ///< NEVER DEFINED -- disallow copy construction
    RingFpDoubleImpl& operator=(const RingFpDoubleImpl&); ///< NEVER DEFINED -- disallow assignment
  public:

    // functions required by every ring
    virtual void myCharacteristic(ZZ& p) const;
    virtual bool IamCommutative() const;
    virtual bool IamIntegralDomain() const;
    virtual bool IamField() const;
    virtual ConstRefRingElem myZero() const;
    virtual ConstRefRingElem myOne() const;
    virtual RingElemRawPtr myNew() const;
    virtual RingElemRawPtr myNew(const MachineInteger& n) const;
    virtual RingElemRawPtr myNew(const ZZ& N) const;
    virtual RingElemRawPtr myNew(ConstRawPtr rawt) const;
    virtual void myDelete(RawPtr rawx) const;                             // destroys x (incl all resources)
    virtual void mySwap(RawPtr rawx, RawPtr rawy) const;                  // swap(x, y)
    virtual void myAssign(RawPtr rawlhs, ConstRawPtr rawx) const;         // lhs = x
    virtual void myAssign(RawPtr rawlhs, const MachineInteger& n) const;  // lhs = n
    virtual void myAssign(RawPtr rawlhs, const ZZ& N) const;              // lhs = N
    virtual void myAssignZero(RawPtr rawlhs) const;                       // lhs = 0
    virtual void myNegate(RawPtr rawlhs, ConstRawPtr rawx) const;         // lhs = -x
    virtual void myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x+y
    virtual void mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x-y
    virtual void myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x*y
    virtual void myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x/y
    virtual bool myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;// lhs = x/y, if divisible
    virtual bool myIsInvertible(ConstRawPtr rawx) const;                                // true iff x is invertible
    virtual void myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = gcd(x,y) in a field
    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const;// lhs = x^n, n>1, x not -1,0,1
    virtual void myPowerBigExp(RawPtr rawlhs, ConstRawPtr rawx, const ZZ& N) const;     // lhs = x^N, N big, x not -1,0,1
    virtual void myOutput(std::ostream& out, ConstRawPtr rawx) const;                   // out << x
    virtual bool myIsPrintAtom(ConstRawPtr rawx) const;
    virtual bool myIsPrintedWithMinus(ConstRawPtr rawx) const;
    virtual void myOutputSelf(std::ostream& out) const;                           // out << R
    virtual void myOutputSelf(OpenMathOutput& OMOut) const;                       // OMOut << R
    virtual void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const;         // OMOut << x
    virtual bool myIsZero(ConstRawPtr rawx) const;                                // x == 0
    virtual bool myIsOne(ConstRawPtr rawx) const;                                 // x == 1
    virtual bool myIsMinusOne(ConstRawPtr rawx) const;                            // x == -1
    virtual bool myIsInteger(ZZ& N, ConstRawPtr rawx) const;                      // true iff x is integer
    virtual bool myIsRational(QQ& Q, ConstRawPtr rawx) const;                     // true iff x is rational
    virtual bool myIsZeroAddMul(RawPtr rawlhs, ConstRawPtr rawy, ConstRawPtr rawz) const;// lhs += y*z, result says whether lhs == 0.
    virtual bool myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const;                // x == y

    virtual ideal myIdealCtor(const std::vector<RingElem>& gens) const;

    virtual RingHom myCompose(const RingHom& phi, const RingHom& theta) const; // phi(theta(...))

    // functions required for a QuotientRing
    virtual RingElem myCanonicalRepr(ConstRawPtr rawx) const; // result is element of myReprRing
    virtual void myReduction(RawPtr rawimage, ConstRawPtr rawarg) const;
    virtual const RingHom& myQuotientingHom() const;
    virtual RingHom myInducedHomCtor(const RingHom& InducingHom) const;


  private:
    class InducedHomImpl: public RingHomBase
    {
    public:
      InducedHomImpl(const QuotientRing& domain, const RingHom& InducingHom);
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    };

  };



  inline RingFpDoubleImpl::value_t& RingFpDoubleImpl::import(RingElemRawPtr rawx)
  {
    return *static_cast<value_t*>(rawx.myRawPtr());
  }

  inline const RingFpDoubleImpl::value_t& RingFpDoubleImpl::import(RingElemConstRawPtr rawx)
  {
    return *static_cast<const value_t*>(rawx.myRawPtr());
  }



  unsigned long RingFpDoubleImpl::PrincipalGen(const ideal& I)
  {
    if (IsZero(I)) return 0;
    unsigned long p;
    ZZ GenI;
    IsInteger(GenI, TidyGens(I)[0]);
    if (!IsConvertible(p, GenI))  // check that the value of the principal generator will fit
      CoCoA_ERROR(ERR::BadSmallFpChar, "RingFpDoubleImpl ctor");
    return p;
  }


  RingFpDoubleImpl::RingFpDoubleImpl(const ideal& I):
      QuotientRingBase(RingZ(), I),  // confirms that I is an ideal of R
      myModulus(PrincipalGen(I)),
      myImpl(myModulus),       // checks that myModulus is a small prime
      myMemMgr(SmallFpDoubleImpl::DatumSize, "RingFpDouble.myMemMgr")
  {
    myZeroPtr.reset(new RingElem(ring(this)));
    myOnePtr.reset(new RingElem(ring(this), 1));
    myQuotientingHomPtr.reset(new RingHom(myQuotientingHomCtor()));
    myRefCountZero();
  }


  RingFpDoubleImpl::~RingFpDoubleImpl()
  {}


  void RingFpDoubleImpl::myCharacteristic(ZZ& p) const
  {
    p = myModulus;
  }


  bool RingFpDoubleImpl::IamCommutative() const
  {
    return true;
  }


  bool RingFpDoubleImpl::IamIntegralDomain() const
  {
    return true;
  }


  bool RingFpDoubleImpl::IamField() const
  {
    return true;
  }


  ConstRefRingElem RingFpDoubleImpl::myZero() const
  {
    return *myZeroPtr;
  }


  ConstRefRingElem RingFpDoubleImpl::myOne() const
  {
    return *myOnePtr;
  }


  RingElemRawPtr RingFpDoubleImpl::myNew() const
  {
    value_t* ans = static_cast<value_t*>(myMemMgr.alloc());
    myImpl.myAssignZero(*ans);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingFpDoubleImpl::myNew(const MachineInteger& n) const
  {
    value_t* ans = static_cast<value_t*>(myMemMgr.alloc());
    myImpl.myAssign(*ans, n);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingFpDoubleImpl::myNew(const ZZ& N) const
  {
    value_t* ans = static_cast<value_t*>(myMemMgr.alloc());
    myImpl.myAssign(*ans, N);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingFpDoubleImpl::myNew(ConstRawPtr rawy) const
  {
    value_t* ans = static_cast<value_t*>(myMemMgr.alloc());
    myImpl.myAssign(*ans, import(rawy));
    return RingElemRawPtr(ans);
  }


  void RingFpDoubleImpl::myDelete(RawPtr rawx) const
  {
    myMemMgr.free(rawx.myRawPtr());
  }


  void RingFpDoubleImpl::mySwap(RawPtr rawx, RawPtr rawy) const
  {
    std::swap(import(rawx), import(rawy));
  }


  void RingFpDoubleImpl::myAssign(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myImpl.myAssign(import(rawlhs), import(rawx));
  }


  void RingFpDoubleImpl::myAssign(RawPtr rawlhs, const MachineInteger& n) const
  {
    myImpl.myAssign(import(rawlhs), n);
  }


  void RingFpDoubleImpl::myAssign(RawPtr rawlhs, const ZZ& N) const
  {
    myImpl.myAssign(import(rawlhs), N);
  }


  void RingFpDoubleImpl::myAssignZero(RawPtr rawlhs) const
  {
    myImpl.myAssignZero(import(rawlhs));
  }


  void RingFpDoubleImpl::myNegate(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myImpl.myNegate(import(rawlhs), import(rawx));
  }


  void RingFpDoubleImpl::myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myImpl.myAdd(import(rawlhs), import(rawx), import(rawy));
  }


  void RingFpDoubleImpl::mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myImpl.mySub(import(rawlhs), import(rawx), import(rawy));
  }


  void RingFpDoubleImpl::myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myImpl.myMul(import(rawlhs), import(rawx), import(rawy));
  }


  void RingFpDoubleImpl::myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    CoCoA_ASSERT(!myIsZero(rawy));
    myImpl.myDiv(import(rawlhs), import(rawx), import(rawy));
  }


  bool RingFpDoubleImpl::myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    return myImpl.myIsDivisible(import(rawlhs), import(rawx), import(rawy));
  }


  bool RingFpDoubleImpl::myIsInvertible(ConstRawPtr rawx) const
  {
    return !myIsZero(rawx);
  }


  void RingFpDoubleImpl::myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myGcdInField(rawlhs, rawx, rawy);
  }


  void RingFpDoubleImpl::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(n > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    myImpl.myPower(import(rawlhs), import(rawx), n);
  }

  void RingFpDoubleImpl::myPowerBigExp(RawPtr rawlhs, ConstRawPtr rawx, const ZZ& N) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(N > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    // Use Fermat's Little Theorem to reduce exponent...
    myImpl.myPower(import(rawlhs), import(rawx), N%(myModulus-1));
  }


  void RingFpDoubleImpl::myOutput(ostream& out, ConstRawPtr rawx) const
  {
    myImpl.myOutput(out, import(rawx));
  }


  bool RingFpDoubleImpl::myIsPrintAtom(ConstRawPtr rawx) const
  {
    return myImpl.myIsPrintAtom(import(rawx));
  }


  bool RingFpDoubleImpl::myIsPrintedWithMinus(ConstRawPtr rawx) const
  {
    return myImpl.myIsPrintedWithMinus(import(rawx));
  }


  void RingFpDoubleImpl::myOutputSelf(ostream& out) const
  {
    out << "FFp(" << myModulus << ")";
  }


  void RingFpDoubleImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("setname2", "GFp");
    OMOut << myModulus;
    OMOut->mySendApplyEnd();
  }


  void RingFpDoubleImpl::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const
  {
    myImpl.myOutput(OMOut, import(rawx));
  }


  bool RingFpDoubleImpl::myIsZero(ConstRawPtr rawx) const
  {
    return myImpl.myIsZero(import(rawx));
  }


  bool RingFpDoubleImpl::myIsOne(ConstRawPtr rawx) const
  {
    return myImpl.myIsOne(import(rawx));
  }


  bool RingFpDoubleImpl::myIsMinusOne(ConstRawPtr rawx) const
  {
    return myImpl.myIsMinusOne(import(rawx));
  }


  bool RingFpDoubleImpl::myIsInteger(ZZ& N, ConstRawPtr rawx) const
  {
    return myImpl.myIsInteger(N, import(rawx));
  }


  bool RingFpDoubleImpl::myIsRational(QQ& Q, ConstRawPtr rawx) const
  {
    ZZ tmp;
    bool ans =  myImpl.myIsInteger(tmp, import(rawx));
    if (ans) Q = tmp;
    return ans;
  }


  bool RingFpDoubleImpl::myIsZeroAddMul(RawPtr rawlhs, ConstRawPtr rawfact1, ConstRawPtr rawfact2) const
  {
    return myImpl.myIsZeroAddMul(import(rawlhs), import(rawfact1), import(rawfact2));
  }


  bool RingFpDoubleImpl::myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    return myImpl.myIsEqual(import(rawx), import(rawy));
  }




  ideal RingFpDoubleImpl::myIdealCtor(const std::vector<RingElem>& gens) const
  {
    return NewFieldIdeal(ring(this), gens);
  }


  RingHom RingFpDoubleImpl::myCompose(const RingHom& phi, const RingHom& theta) const
  {
    // No need to check compatibility -- it was checked when theta was built.
    return RingHom(new InducedHomImpl(QuotientRing(this), phi(theta(myQuotientingHom()))));
  }


  RingElem RingFpDoubleImpl::myCanonicalRepr(ConstRawPtr rawx) const
  {
    return RingElem(myReprRing, static_cast<unsigned long>(import(rawx)));
  }


  void RingFpDoubleImpl::myReduction(RawPtr rawimage, ConstRawPtr rawarg) const
  {
    ZZ tmp;
    CoCoA_ASSERT(myReprRing->myIsInteger(tmp, rawarg));
    myReprRing->myIsInteger(tmp, rawarg);
    myImpl.myAssign(import(rawimage), tmp);
  }


  const RingHom& RingFpDoubleImpl::myQuotientingHom() const
  {
    return *myQuotientingHomPtr;
  }


  RingHom RingFpDoubleImpl::myInducedHomCtor(const RingHom& InducingHom) const
  {
    // Compatibility has already been checked (see InducedHom in QuotientRing.C)
    CoCoA_ASSERT(IsZero(InducingHom(myModulus)));
    return RingHom(new InducedHomImpl(QuotientRing(this), InducingHom));
  }


  //---------------------------------------------------------------------------
  // Functions to do with ring homomorphisms


  RingFpDoubleImpl::InducedHomImpl::InducedHomImpl(const QuotientRing& domain, const RingHom& InducingHom):
      RingHomBase(domain, codomain(InducingHom))
  { /* Compatibility already checked in InducedHom in QuotientRing.C */  }


  void RingFpDoubleImpl::InducedHomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    ZZ tmp;  //??? wasteful new/delete
    CoCoA_ASSERT(myDomain->myIsInteger(tmp, rawarg));
    myDomain->myIsInteger(tmp, rawarg);  // must necessarily succeed
    myCodomain->myAssign(rawimage, tmp);
  }



  QuotientRing NewRingFpDouble(const MachineInteger& p)
  {
    return QuotientRing(new RingFpDoubleImpl(ideal(RingZ(), p)));
  }

  QuotientRing NewRingFpDouble(const ZZ& P)
  {
    return QuotientRing(new RingFpDoubleImpl(ideal(RingZ(), P)));
  }

  QuotientRing NewRingFpDouble(const ideal& I)
  {
    if (!IsZ(AmbientRing(I))) CoCoA_ERROR(ERR::IdealNotInRing, "NewRingFpDouble(I)");
    return QuotientRing(new RingFpDoubleImpl(I));
  }

} // end of namespace CoCoA


// RCS header/log
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RingFpDouble.C,v 1.11 2009/12/11 11:46:32 abbott Exp $
// $Log: RingFpDouble.C,v $
// Revision 1.11  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.10  2009/09/25 14:01:11  abbott
// Cleaned up include directives.
//
// Revision 1.9  2009/09/24 16:22:29  abbott
// Tidied up include directives.  Removed some unnecessary "std::" prefixes.
//
// Revision 1.8  2009/07/02 16:32:11  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.7  2009/06/05 12:08:28  abbott
// Changed return type of operator%(ZZ,MachineInteger); it is now unsigned long
// instead of ZZ.
//
// Revision 1.6  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.5  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.4  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.3  2007/05/21 12:44:25  abbott
// Changed impl of powering routine as a consequence of changed signature
// to the modulus operator for ZZ modulo machine integer.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.14  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.13  2007/03/08 10:23:29  bigatti
// -- CanonHom --> CanonicalHom
//
// Revision 1.12  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.11  2007/03/03 14:07:23  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.10  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.9  2007/01/17 12:32:39  cocoa
// Changed a few more "raw" variable names so that the code compiles fine
// also when CoCoA_DEBUG is set.
//
// Revision 1.8  2007/01/15 16:00:10  cocoa
// -- added prefix "raw" to RawPtr arguments names
// -- changed rhs into rawx, n, or N
//
// Revision 1.7  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.6  2006/11/08 16:21:59  cocoa
// Structural cleaning of RingHom; many consequential changes.
//
// Revision 1.5  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.4  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.3  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/10/06 10:15:52  cocoa
// In response to Susan's bug: a fiasco when compiling with CoCoA_MEMPOOL_DEBUG
// set wrongly.  Moved several implementation classes out of their header files
// into the implementation files.  Several functions had to be uninlined.
// Also corrected position of #include, etc.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.7  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.6  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.5  2006/04/21 15:01:36  cocoa
// Changed default implementation of RingBase::myGcd -- it now gives a SERIOUS
// error.  All fields must now handle a call to gcd explicitly: they can use
// the new myGcdInField function.  It's now cleaner than it was.
//
// Revision 1.4  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.3  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.3  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.2  2005/10/12 15:52:09  cocoa
// Completed test-RingFp1 and corrected/cleaned the SmallFp*
// and RingFp* files.
//
// Some minor tidying elsewhere.
//
// Revision 1.1  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
