//   Copyright (c)  2002-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RingFpLog.H"

#include "CoCoA/FieldIdeal.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/QQ.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/SmallFpLogImpl.H"
#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/ideal.H"


#include <algorithm>
//using std::swap;         // only in mySwap
#include <iostream>
using std::ostream;        // only in myOutput
// #include <limits>  ---  included in MachineInteger (included via QQ.H)
using std::numeric_limits; // only in ctor
// #include <memory>  ---  included in MemPool.H
using std::auto_ptr;
// #include <vector>  ---  included in ideal.H
using std::vector;


namespace CoCoA
{

  /*-----------------------------------------------------------------*/
  /** \include RingFpLog.txt  */
  /*-----------------------------------------------------------------*/
  class RingFpLogImpl: public QuotientRingBase
  {
  private: // data members
    typedef SmallFpLogImpl::value_t value_t;
    const value_t myModulus;
    const SmallFpLogImpl myImpl;
    mutable MemPool myMemMgr;       // MemPool must come BEFORE myZeroPtr and myOnePtr
    auto_ptr<RingElem> myZeroPtr;  ///< Every ring stores its own zero.
    auto_ptr<RingElem> myOnePtr;   ///< Every ring stores its own one.
    auto_ptr<RingHom> myQuotientingHomPtr;
    static value_t PrincipalGen(const ideal& I); // used for arg checking in ctor

  private:
    static value_t& import(RingElemRawPtr rawx);
    static const value_t& import(RingElemConstRawPtr rawx);

  private:
    RingFpLogImpl(const ideal& I); // called only by NewRingFpLog
    ~RingFpLogImpl();
    friend QuotientRing NewRingFpLog(const MachineInteger& p);
    friend QuotientRing NewRingFpLog(const ZZ& P);
    friend QuotientRing NewRingFpLog(const ideal& I);
  private:
    RingFpLogImpl(const RingFpLogImpl&);            ///< NEVER DEFINED -- disallow copy construction
    RingFpLogImpl& operator=(const RingFpLogImpl&); ///< NEVER DEFINED -- disallow assignment
  public:

    // functions required by every ring
    virtual void myCharacteristic(ZZ& p) const;
    virtual bool IamCommutative() const;
    virtual bool IamIntegralDomain() const;
    virtual bool IamField() const;
    virtual ConstRefRingElem myZero() const;
    virtual ConstRefRingElem myOne() const;
    virtual RingElemRawPtr myNew() const;
    virtual RingElemRawPtr myNew(const MachineInteger& n) const;
    virtual RingElemRawPtr myNew(const ZZ& N) const;
    virtual RingElemRawPtr myNew(ConstRawPtr rawt) const;
    virtual void myDelete(RawPtr rawx) const;                             // destroys x (incl all resources)
    virtual void mySwap(RawPtr rawx, RawPtr rawy) const;                  // swap(x, y)
    virtual void myAssign(RawPtr rawlhs, ConstRawPtr rawx) const;         // lhs = x
    virtual void myAssign(RawPtr rawlhs, const MachineInteger& n) const;  // lhs = n
    virtual void myAssign(RawPtr rawlhs, const ZZ& N) const;              // lhs = N
    virtual void myAssignZero(RawPtr rawlhs) const;                       // lhs = 0
    virtual void myNegate(RawPtr rawlhs, ConstRawPtr rawx) const;         // lhs = -x
    virtual void myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x+y
    virtual void mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x-y
    virtual void myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x*y
    virtual void myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = x/y
    virtual bool myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;// lhs = x/y, if divisible
    virtual bool myIsInvertible(ConstRawPtr rawx) const;                                // true iff x is invertible
    virtual void myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const;        // lhs = gcd(x,y) in a field
    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const;// lhs = x^n, n>1, x not -1,0,1
    virtual void myPowerBigExp(RawPtr rawlhs, ConstRawPtr rawx, const ZZ& N) const;     // lhs = x^N, N big, x not -1,0,1
    virtual void myOutput(std::ostream& out, ConstRawPtr rawx) const;                   // out << x
    virtual bool myIsPrintAtom(ConstRawPtr rawx) const;
    virtual bool myIsPrintedWithMinus(ConstRawPtr rawx) const;
    virtual void myOutputSelf(std::ostream& out) const;                           // out << R
    virtual void myOutputSelf(OpenMathOutput& OMOut) const;                       // OMOut << R
    virtual void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const;         // OMOut << x
    virtual bool myIsZero(ConstRawPtr rawx) const;                                // x == 0
    virtual bool myIsOne(ConstRawPtr rawx) const;                                 // x == 1
    virtual bool myIsMinusOne(ConstRawPtr rawx) const;                            // x == -1
    virtual bool myIsInteger(ZZ& N, ConstRawPtr rawx) const;                      // true iff x is integer
    virtual bool myIsRational(QQ& Q, ConstRawPtr rawx) const;                     // true iff x is rational
    virtual bool myIsZeroAddMul(RawPtr rawlhs, ConstRawPtr rawy, ConstRawPtr rawz) const;// lhs += y*z, result says whether lhs == 0.
    virtual bool myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const;                // x == y

    virtual ideal myIdealCtor(const std::vector<RingElem>& gens) const;

    virtual RingHom myCompose(const RingHom& phi, const RingHom& theta) const; // phi(theta(...))

    // functions required for a QuotientRing
    virtual RingElem myCanonicalRepr(ConstRawPtr rawx) const; // result is element of myReprRing
    virtual void myReduction(RawPtr rawimage, ConstRawPtr rawarg) const;
    virtual const RingHom& myQuotientingHom() const;
    virtual RingHom myInducedHomCtor(const RingHom& InducingHom) const;


  private:
    class InducedHomImpl: public RingHomBase
    {
    public:
      InducedHomImpl(const QuotientRing& domain, const RingHom& InducingHom);
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    };

  };



  inline RingFpLogImpl::value_t& RingFpLogImpl::import(RingElemRawPtr rawx)
  {
    return *static_cast<value_t*>(rawx.myRawPtr());
  }

  inline const RingFpLogImpl::value_t& RingFpLogImpl::import(RingElemConstRawPtr rawx)
  {
    return *static_cast<const value_t*>(rawx.myRawPtr());
  }



  RingFpLogImpl::value_t RingFpLogImpl::PrincipalGen(const ideal& I)
  {
    if (IsZero(I)) return 0;
    value_t p;
    ZZ GenI;
    IsInteger(GenI, TidyGens(I)[0]);
    if (!IsConvertible(p, GenI))  // check that the value of the principal generator will fit
      CoCoA_ERROR(ERR::BadSmallFpChar, "RingFpLogImpl ctor");
    return p;
  }


  RingFpLogImpl::RingFpLogImpl(const ideal& I):
      QuotientRingBase(RingZ(), I),  // confirms that I is an ideal of R
      myModulus(PrincipalGen(I)),
      myImpl(myModulus),       // checks that myModulus is a small prime
      myMemMgr(SmallFpLogImpl::DatumSize, "RingFpLog.myMemMgr")
  {
    myZeroPtr.reset(new RingElem(ring(this)));
    myOnePtr.reset(new RingElem(ring(this), 1));
    myQuotientingHomPtr.reset(new RingHom(myQuotientingHomCtor()));
    myRefCountZero();
  }


  RingFpLogImpl::~RingFpLogImpl()
  {}


  void RingFpLogImpl::myCharacteristic(ZZ& p) const
  {
    p = myModulus;
  }


  bool RingFpLogImpl::IamCommutative() const
  {
    return true;
  }


  bool RingFpLogImpl::IamIntegralDomain() const
  {
    return true;
  }


  bool RingFpLogImpl::IamField() const
  {
    return true;
  }


  ConstRefRingElem RingFpLogImpl::myZero() const
  {
    return *myZeroPtr;
  }


  ConstRefRingElem RingFpLogImpl::myOne() const
  {
    return *myOnePtr;
  }


  RingElemRawPtr RingFpLogImpl::myNew() const
  {
    SmallFpElem_t* ans = static_cast<SmallFpElem_t*>(myMemMgr.alloc());
    myImpl.myAssignZero(*ans);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingFpLogImpl::myNew(const MachineInteger& n) const
  {
    SmallFpElem_t* ans = static_cast<SmallFpElem_t*>(myMemMgr.alloc());
    myImpl.myAssign(*ans, n);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingFpLogImpl::myNew(const ZZ& N) const
  {
    SmallFpElem_t* ans = static_cast<SmallFpElem_t*>(myMemMgr.alloc());
    myImpl.myAssign(*ans, N);
    return RingElemRawPtr(ans);
  }


  RingElemRawPtr RingFpLogImpl::myNew(ConstRawPtr rawy) const
  {
    SmallFpElem_t* ans = static_cast<SmallFpElem_t*>(myMemMgr.alloc());
    myImpl.myAssign(*ans, import(rawy));
    return RingElemRawPtr(ans);
  }


  void RingFpLogImpl::myDelete(RawPtr rawx) const
  {
    myMemMgr.free(rawx.myRawPtr());
  }


  void RingFpLogImpl::mySwap(RawPtr rawx, RawPtr rawy) const
  {
    std::swap(import(rawx), import(rawy));
  }


  void RingFpLogImpl::myAssign(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myImpl.myAssign(import(rawlhs), import(rawx));
  }


  void RingFpLogImpl::myAssign(RawPtr rawlhs, const MachineInteger& n) const
  {
    myImpl.myAssign(import(rawlhs), n);
  }


  void RingFpLogImpl::myAssign(RawPtr rawlhs, const ZZ& N) const
  {
    myImpl.myAssign(import(rawlhs), N);
  }


  void RingFpLogImpl::myAssignZero(RawPtr rawlhs) const
  {
    myImpl.myAssignZero(import(rawlhs));
  }


  void RingFpLogImpl::myNegate(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myImpl.myNegate(import(rawlhs), import(rawx));
  }


  void RingFpLogImpl::myAdd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myImpl.myAdd(import(rawlhs), import(rawx), import(rawy));
  }


  void RingFpLogImpl::mySub(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myImpl.mySub(import(rawlhs), import(rawx), import(rawy));
  }


  void RingFpLogImpl::myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myImpl.myMul(import(rawlhs), import(rawx), import(rawy));
  }


  void RingFpLogImpl::myDiv(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    CoCoA_ASSERT(!myIsZero(rawy));
    myImpl.myDiv(import(rawlhs), import(rawx), import(rawy));
  }


  bool RingFpLogImpl::myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    return myImpl.myIsDivisible(import(rawlhs), import(rawx), import(rawy));
  }


  bool RingFpLogImpl::myIsInvertible(ConstRawPtr rawx) const
  {
    return !myIsZero(rawx);
  }


  void RingFpLogImpl::myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    myGcdInField(rawlhs, rawx, rawy);
  }


  void RingFpLogImpl::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(n > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    myImpl.myPower(import(rawlhs), import(rawx), n);
  }

  void RingFpLogImpl::myPowerBigExp(RawPtr rawlhs, ConstRawPtr rawx, const ZZ& N) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(N > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));
    // Use Fermat's Little Theorem to reduce exponent...
    myImpl.myPower(import(rawlhs), import(rawx), N%(myModulus-1));
  }


  void RingFpLogImpl::myOutput(ostream& out, ConstRawPtr rawx) const
  {
    myImpl.myOutput(out, import(rawx));
  }


  bool RingFpLogImpl::myIsPrintAtom(ConstRawPtr rawx) const
  {
    return myImpl.myIsPrintAtom(import(rawx));
  }


  bool RingFpLogImpl::myIsPrintedWithMinus(ConstRawPtr rawx) const
  {
    return myImpl.myIsPrintedWithMinus(import(rawx));
  }


  void RingFpLogImpl::myOutputSelf(ostream& out) const
  {
    out << "FFp(" << myModulus << ")";
  }


  void RingFpLogImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("setname2", "GFp");
    OMOut << myModulus;
    OMOut->mySendApplyEnd();
  }


  void RingFpLogImpl::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const
  {
    myImpl.myOutput(OMOut, import(rawx));
  }


  bool RingFpLogImpl::myIsZero(ConstRawPtr rawx) const
  {
    return myImpl.myIsZero(import(rawx));
  }


  bool RingFpLogImpl::myIsOne(ConstRawPtr rawx) const
  {
    return myImpl.myIsOne(import(rawx));
  }


  bool RingFpLogImpl::myIsMinusOne(ConstRawPtr rawx) const
  {
    return myImpl.myIsMinusOne(import(rawx));
  }


  bool RingFpLogImpl::myIsInteger(ZZ& N, ConstRawPtr rawx) const
  {
    return myImpl.myIsInteger(N, import(rawx));
  }


  bool RingFpLogImpl::myIsRational(QQ& Q, ConstRawPtr rawx) const
  {
    ZZ tmp;
    bool ans =  myImpl.myIsInteger(tmp, import(rawx));
    if (ans) Q = tmp;
    return ans;
  }


  bool RingFpLogImpl::myIsZeroAddMul(RawPtr rawlhs, ConstRawPtr rawfact1, ConstRawPtr rawfact2) const
  {
    return myImpl.myIsZeroAddMul(import(rawlhs), import(rawfact1), import(rawfact2));
  }


  bool RingFpLogImpl::myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    return myImpl.myIsEqual(import(rawx), import(rawy));
  }




  ideal RingFpLogImpl::myIdealCtor(const std::vector<RingElem>& gens) const
  {
    return NewFieldIdeal(ring(this), gens);
  }


  RingHom RingFpLogImpl::myCompose(const RingHom& phi, const RingHom& theta) const
  {
    // No need to check compatibility -- it was checked when theta and phi were built
    return RingHom(new InducedHomImpl(QuotientRing(this), phi(theta(myQuotientingHom()))));
  }


  RingElem RingFpLogImpl::myCanonicalRepr(ConstRawPtr rawx) const
  {
    return RingElem(myReprRing, import(rawx));
  }


  void RingFpLogImpl::myReduction(RawPtr rawimage, ConstRawPtr rawarg) const
  {
    ZZ tmp;
    CoCoA_ASSERT(myReprRing->myIsInteger(tmp, rawarg));
    myReprRing->myIsInteger(tmp, rawarg);
    myImpl.myAssign(import(rawimage), tmp);
  }


  const RingHom& RingFpLogImpl::myQuotientingHom() const
  {
    return *myQuotientingHomPtr;
  }


  RingHom RingFpLogImpl::myInducedHomCtor(const RingHom& InducingHom) const
  {
    // Compatibility has already been checked (see InducedHom in QuotientRing.C)
    CoCoA_ASSERT(IsZero(InducingHom(myModulus)));
    return RingHom(new InducedHomImpl(QuotientRing(this), InducingHom));
  }


  //---------------------------------------------------------------------------
  // Functions to do with ring homomorphisms


  RingFpLogImpl::InducedHomImpl::InducedHomImpl(const QuotientRing& domain, const RingHom& InducingHom):
      RingHomBase(domain, codomain(InducingHom))
  { /* Compatibility already checked in InducedHom in QuotientRing.C */  }


  void RingFpLogImpl::InducedHomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    ZZ tmp;  //??? wasteful new/delete
    CoCoA_ASSERT(myDomain->myIsInteger(tmp, rawarg));
    myDomain->myIsInteger(tmp, rawarg);  // must necessarily succeed
    myCodomain->myAssign(rawimage, tmp);
  }



  QuotientRing NewRingFpLog(const MachineInteger& p)
  {
    return QuotientRing(new RingFpLogImpl(ideal(RingZ(), p)));
  }

  QuotientRing NewRingFpLog(const ZZ& P)
  {
    return QuotientRing(new RingFpLogImpl(ideal(RingZ(), P)));
  }

  QuotientRing NewRingFpLog(const ideal& I)
  {
    if (!IsZ(AmbientRing(I))) CoCoA_ERROR(ERR::IdealNotInRing, "NewRingFpLog(I)");
    return QuotientRing(new RingFpLogImpl(I));
  }

} // end of namespace CoCoA


// RCS header/log
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RingFpLog.C,v 1.11 2009/12/11 11:46:32 abbott Exp $
// $Log: RingFpLog.C,v $
// Revision 1.11  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.10  2009/09/25 14:01:11  abbott
// Cleaned up include directives.
//
// Revision 1.9  2009/09/24 16:22:44  abbott
// Tidied up include directives.  Removed some unnecessary "std::" prefixes.
//
// Revision 1.8  2009/07/02 16:32:11  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.7  2009/06/05 12:08:27  abbott
// Changed return type of operator%(ZZ,MachineInteger); it is now unsigned long
// instead of ZZ.
//
// Revision 1.6  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.5  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.4  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.3  2007/05/21 12:44:25  abbott
// Changed impl of powering routine as a consequence of changed signature
// to the modulus operator for ZZ modulo machine integer.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.13  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.12  2007/03/08 10:23:29  bigatti
// -- CanonHom --> CanonicalHom
//
// Revision 1.11  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.10  2007/03/03 14:07:23  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.9  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.8  2007/01/15 16:11:41  cocoa
// -- added prefix "raw" to RawPtr arguments names
// -- changed rhs into rawx, n, or N
//
// Revision 1.7  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.6  2006/11/08 16:21:59  cocoa
// Structural cleaning of RingHom; many consequential changes.
//
// Revision 1.5  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.4  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.3  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/10/06 10:15:52  cocoa
// In response to Susan's bug: a fiasco when compiling with CoCoA_MEMPOOL_DEBUG
// set wrongly.  Moved several implementation classes out of their header files
// into the implementation files.  Several functions had to be uninlined.
// Also corrected position of #include, etc.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.7  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.6  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.5  2006/04/21 15:01:36  cocoa
// Changed default implementation of RingBase::myGcd -- it now gives a SERIOUS
// error.  All fields must now handle a call to gcd explicitly: they can use
// the new myGcdInField function.  It's now cleaner than it was.
//
// Revision 1.4  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.3  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.6  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.5  2005/10/12 15:52:09  cocoa
// Completed test-RingFp1 and corrected/cleaned the SmallFp*
// and RingFp* files.
//
// Some minor tidying elsewhere.
//
// Revision 1.4  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.3  2005/09/30 15:03:39  cocoa
// Minor cleaning and tidying.
// DistrMPolyInlPP: use of summands now rather cleaner.
//
// Revision 1.2  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.5  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.4  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.3  2005/02/11 16:45:24  cocoa
// Removed the useless and misleading functions myInit and myKill
// from the SmallFp*Impl classes; various consequential changes.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.16  2004/11/18 18:33:41  cocoa
// Now every ring know its own "one" element (as well as "zero").
// Several consequential changes.
//
// Revision 1.15  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.14  2004/11/09 15:47:29  cocoa
// -- changed myOutput: FF --> FFp
//
// Revision 1.13  2004/11/05 15:34:33  cocoa
// Consequential change following from the renaming of
// FieldIdealImpl and the introduction of the new pseudo-ctor.
//
// Revision 1.12  2004/11/04 18:47:43  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.11  2004/11/02 15:09:45  cocoa
// -- added name to MemPool
//
// Revision 1.10  2004/07/27 16:03:39  cocoa
// Added IsCommutative test and IamCommutative member function
// to all rings.  Tidied geobuckets a little.
//
// Revision 1.9  2004/07/14 16:40:42  cocoa
// Separated RingFpLog from its implementation which now resides in
// a new class: SmallFpLogImpl.  This is analogous to the change made
// to RingFp yesterday.
//
// Some tidying and other sundry minor changes.
//
// Revision 1.8  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.7  2004/04/08 15:33:34  cocoa
// Added function IsInteger, and the related RingBase::myIsInteger
// virtual function, plus all necessary implementations.
//
// Revision 1.6  2004/03/20 17:46:10  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.5  2004/02/03 16:16:20  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.4  2004/01/30 14:07:10  cocoa
// Tidied RingRawValue union: now it contains just two fields,
// and has no need of forward declarations of types used internally
// by the concrete rings -- it uses explicitly a void* instead.
//
// I have tidied the "import" functions used by most concrete rings.
//
// I have moved the choice of representation type for RingFp and RingFpLog
// into a typedef in config.H -- this is to recognise that different
// choices may work best on different platforms.
//
// Revision 1.3  2003/10/17 10:51:06  cocoa
// Major cleaning, and new naming convention.
//
// Revision 1.2  2003/10/09 12:15:45  cocoa
// New coding convention for rings.
//
// Revision 1.14  2003/06/23 16:55:43  abbott
// Minor cleaning prior to public release.
// Name change and alignment with RingBase.
//
// Revision 1.13  2003/04/23 10:16:46  abbott
// Numerous changes, almost all consequential from changes made to ring_FpLog.H
//  * ring_FpLog is now derived from QuotientRing (so has a different ctor,
//    and must implement the functions specific to QuotientRings)
//  * implementation of ring_FpLog::hom added
//  * uses the generic FieldIdeal for ideals
//  * now uses reset member function of auto_ptr instead of assignment from raw ptr
//  * ring_FpLog::power is now simpler (since exponent must be non-negative)
//
// Revision 1.12  2002/11/14 18:22:00  abbott
// Consequential changes following renaming in ring.H.
// Slightly improved some error messages.
// Changed division by zero check into an assertion.
//
// Revision 1.11  2002/07/05 15:24:12  abbott
// Added definition of member function IsDivisible.
//
// Revision 1.10  2002/06/28 12:10:06  abbott
// Now using C++ vectors and auto_ptr to achieve exception safety.
// Better argument checking in the constructor.
//
// Revision 1.9  2002/06/22 17:08:18  abbott
// Changed name of "equal" member function to "IsEqual" (as per new ring.H).
//
// Revision 1.8  2002/06/04 13:04:02  abbott
// Changed type of elements in the log/exp tables; now the type is FpTableElem
// (which is a typedef in ring_FpLog.H)
//
// Revision 1.7  2002/05/30 14:31:29  abbott
// Added definitions for member functions IsField and zero (as required by the new ring.H).
// Removed bogus function ngens.
//
// Revision 1.6  2002/05/15 15:43:47  abbott
// Fixed bug in power for negative exponents.
//
// Revision 1.5  2002/05/15 15:25:31  abbott
// Added definitions for new member functions of ring_FpLog:
//   characteristic, negate, and power.
// Other minor changes to stay in line with new ring.H
//
// Revision 1.4  2002/03/19 13:59:47  abbott
// Changed class name to ring_FpLog.  Added two "using" directives.
//
// Revision 1.3  2002/02/08 11:20:14  bigatti
// - changed syntax to IsZeroAddMul
//
// Revision 1.2  2002/01/31 11:01:37  abbott
// Corrected the constructor in the case of characteristic 2.
//
// Revision 1.1  2002/01/30 17:26:27  abbott
// Initial revision
//
