//   Copyright (c)  2005,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/SmallFpDoubleImpl.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/SmallPrime.H"
#include "CoCoA/NumTheory.H"
#include "CoCoA/error.H"
#include "CoCoA/convert.H"
#include "CoCoA/OpenMath.H"

#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits;
//#include <cmath>
using std::floor;

namespace CoCoA
{

  namespace // anonymous namespace
  {
    // JAA believes that this routine always returns a reduced value,
    // but does not have a proof of this currently.
    SmallFpDoubleImpl::value_t exgcd(SmallFpDoubleImpl::value_t r, SmallFpDoubleImpl::value_t m)
    {
      CoCoA_ASSERT(r != 0);  CoCoA_ASSERT(m != 0);
      SmallFpDoubleImpl::value_t p = m;
      SmallFpDoubleImpl::value_t cr = 1;
      SmallFpDoubleImpl::value_t cm = 0; // this is minus what you think it is!
      while (r != 0)
      {
	SmallFpDoubleImpl::value_t q = floor(m/r);
	m -= q*r;
	cm += q*cr;
	if (m == 0) break;
	q = floor(r/m);
	r -= q*m;
	cr += q*cm;
      }
      //???  if (r+m != 1) error("cannot invert zero-divisor");
      if (r == 0) return p-cm;
      return cr;
    }


  } // end of anonymous namespace


  // These two inline fns are used only in the ctors.
  inline SmallFpDoubleImpl::value_t SmallFpDoubleImpl::CalcDrop(value_t p)
  {
    return p*floor(MaxInt()/p/2); // Largest (integer) multiple of p <= MaxInt/2.
  }

  inline SmallFpDoubleImpl::value_t SmallFpDoubleImpl::CalcIterLimit(value_t p)
  {
    return floor(MaxInt()/(p-1)/(p-1)/2); // Max no. of unreduced products you can sum without exceeding MaxInt/2.
  }


  SmallFpDoubleImpl::SmallFpDoubleImpl(unsigned long p):
      myModulus(CheckCtorArg(p)),
      myResiduesAreSymm(DefaultResiduesAreSymm()),
      myDrop(CalcDrop(myModulus)),
      myIterLimit(CalcIterLimit(myModulus))
  {}


  SmallFpDoubleImpl::SmallFpDoubleImpl(const ZZ& P):
      myModulus(CheckCtorArg(P)),
      myResiduesAreSymm(DefaultResiduesAreSymm()),
      myDrop(CalcDrop(myModulus)),
      myIterLimit(CalcIterLimit(myModulus))
  {}


  void SmallFpDoubleImpl::myDiv(value_t& lhs, value_t x, value_t y) const
  {
    CoCoA_ASSERT(x == myReduceMod(x));
    CoCoA_ASSERT(y == myReduceMod(y));
    CoCoA_ASSERT(!myIsZero(y));
    value_t recip = exgcd(y, myModulus);
    myMul(lhs, x, recip);
  }


  void SmallFpDoubleImpl::myPower(value_t& lhs, value_t x, unsigned long n) const
  {
    CoCoA_ASSERT(!myIsZero(x) || n != 0); // Cannot compute 0^0
    if (myIsZero(x)) { lhs = 0; return; }
    if (myIsOne(x)) { lhs = 1; return; }
    n %= static_cast<long>(myModulus) - 1; // OK by fermat's little theorem.
    if (n == 0) { lhs = 1; return; }
    if (n == 1) { lhs = x; return; }
    // Here we know that n >= 2 and x is not 0 or 1.
    // Below is an iterative version of repeated squaring.
    unsigned long mask = 1;
    unsigned long quartern = n/4;
    while (mask <= quartern) mask <<= 1;
    value_t X = x; // copy x in case lhs aliases x
    lhs = x;
    for (; mask; mask >>= 1)
    {
      myMul(lhs, lhs, lhs);
      if (n & mask) myMul(lhs, lhs, X);
    }
  }


  void SmallFpDoubleImpl::myOutput(ostream& out, value_t x) const
  {
    CoCoA_ASSERT(x == myReduceMod(x));
    // Convert x into long to avoid any risk of a decimal point appearing.
    const long longx = static_cast<long>(x);
    if (!myIsPrintedWithMinus(x))
      out << longx;
    else
      out << longx-long(myModulus);
  }


  void SmallFpDoubleImpl::myOutput(OpenMathOutput& OMOut, value_t x) const
  {
    CoCoA_ASSERT(x == myReduceMod(x));
    OMOut << static_cast<long>(x); // Make sure no decimal point appears!
  }


  // These two fns check that the modulus is a suitably small prime.
  // If the arg is not suitable then an exception is thrown, otherwise
  // the return value is the suitably small prime (as unsigned long).
  SmallFpDoubleImpl::value_t SmallFpDoubleImpl::CheckCtorArg(unsigned long p)
  {
    const value_t MaxAddMul = value_t(p)*value_t(p+1);
    if (MaxAddMul > MaxInt() ||
        CalcIterLimit(p) == 0 ||
        !IsPrime(p))
      CoCoA_ERROR(ERR::BadSmallFpChar, "SmallFpDoubleImpl ctor");
    return p;
  }


  SmallFpDoubleImpl::value_t SmallFpDoubleImpl::CheckCtorArg(const ZZ& P)
  {
    // First convert to ulong, if successful then check value is small enough.
    unsigned long p;
    if (P <= 0 || !IsConvertible(p, P))
      CoCoA_ERROR(ERR::BadSmallFpChar, "SmallFpDoubleImpl ctor");
    return CheckCtorArg(p);
///???    return CheckCtorArg(ConvertTo<unsigned long>(P));
  }


  // MaxInt gives the largest integer N such that all integers from 1 to N
  // N can be represented exactly using the type SmallFpDoubleImpl::value_t.
  SmallFpDoubleImpl::value_t SmallFpDoubleImpl::MaxInt()
  {
    return 2 / numeric_limits<value_t>::epsilon();
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SmallFpDoubleImpl.C,v 1.8 2010/03/23 11:58:04 abbott Exp $
// $Log: SmallFpDoubleImpl.C,v $
// Revision 1.8  2010/03/23 11:58:04  abbott
// Simplified MaxInt -- removed static data member (troublesome for thread safety).
//
// Revision 1.7  2009/12/23 18:53:52  abbott
// Major change to conversion functions:
//   convert(..) is now a procedure instead of a function
//   IsConvertible(..) replaces the former convert(..) function
//   Added new NumericCast conversion function (placeholder for BOOST feature)
//   Consequent changes in code which uses these features.
//
// Revision 1.6  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.5  2009/06/05 12:14:55  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
// Revision 1.4  2009/05/14 09:39:29  abbott
// Added possibility to specify "symmetric" or "non-negative" residues
// in quotients of ZZ.  Affects printing of elements in quotients of ZZ
// (also changed printing of elements in general quotient rings).
// Consequent changes in several tests.
//
// Revision 1.3  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.4  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.3  2006/11/27 13:06:22  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.2  2006/11/02 13:25:43  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.4  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.3  2005/10/12 15:52:09  cocoa
// Completed test-RingFp1 and corrected/cleaned the SmallFp*
// and RingFp* files.
//
// Some minor tidying elsewhere.
//
// Revision 1.2  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.1  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
