//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/config.H"
#include "CoCoA/SmallPrime.H"

#include <limits>
using std::numeric_limits;
#include <cmath>
using std::sqrt;
#include <vector>
using std::vector;

namespace CoCoA
{

  namespace // anonymous namespace
  {
    // These tables are used by NextSmallPrime and PrevSmallPrime
    // mask[n%30] == 0 iff n is divisible by at least one of 2, 3 and 5.
    // n+skip[n%30] is the next number after n not divisible by 2, 3 or 5
    // n-fall[n%30] is the last number before n not divisible by 2, 3 or 5
    const unsigned int mask[30] = {0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,0,0,1};
    const unsigned int skip[30] = {1,6,5,4,3,2,1,4,3,2,1,2,1,4,3,2,1,2,1,4,3,2,1,6,5,4,3,2,1,2};
    const unsigned int fall[30] = {1,2,1,2,3,4,5,6,1,2,3,4,1,2,1,2,3,4,1,2,1,2,3,4,1,2,3,4,5,6};


    // These functions work only for (putative) primes not exceeding SmallPrimeLimit.
    // Note: SmallPrimeLimit itself will probably not be a prime!
    const unsigned long SmallPrimeLimit = static_cast<unsigned long>(sqrt(double(numeric_limits<SmallFpElem_t>::max())));

    // This recursive function is about 20% slower than the iterative equivalent.
    // // Function to compute a^n mod p; assumes a < p, unclever for n > p
    // // p need not be prime, but (p-1)^2 must fit inside an unsigned long.
    // unsigned long pwr(unsigned long a, unsigned long n, unsigned long p)
    // {
    //   unsigned long b;
    //
    //   if (n == 0) return 1;
    //   if (n == 1) return a;
    //   b = pwr((a*a)%p, n/2, p);
    //   if ((n&1)) b = (a*b)%p;
    //   return b;
    // }


    /// Function to compute a^n mod p; assumes a < p, unclever for n > p
    /// p need not be prime, but (p-1)^2 must fit inside an unsigned long.
    unsigned long pwr(unsigned long a, unsigned long n, unsigned long p)
    {
      unsigned long b = 1;

      if (n == 0) return b;
      while (n > 1)
      {
        if (n&1) b = (b*a)%p;
        n /= 2;
        a = (a*a)%p;
      }
      return (b*a)%p;
    }


    /// Function to determine if n is a strong pseudo-prime to base b
    /// using Miller-Rabin test.  n > 1 must be odd, and b < n.
    /// Also n*n must fit inside an unsigned long.
    bool StrongPseudoPrime(unsigned long b, unsigned long n)
    {
      const unsigned long n1 = n-1;
      // Compute in q the largest odd factor of n-1
      unsigned long q = n1;
      while ((q&1) == 0)
        q /= 2;

      unsigned long power = pwr(b, q, n);
      if (power == n1 || power == 1) return true;
      q *= 2;
      while (q < n1)
      {
        power = (power*power)%n;
        if (power == n1) return true;
        if (power == 1) return false;
        q *= 2;
      }
      return false;
    }


    /// This procedure finds all distinct prime factors of n (assuming n < 2^32).
    /// These primes are placed in the vector PrimeFacs (in ascending order).
    void Factorize32bits(std::vector<unsigned long>& PrimeFacs, unsigned long n)
    {
      PrimeFacs.clear();

      for (unsigned long p = 2;; p = NextSmallPrime(p))
      {
        if (p*p > n) break;
        if (n%p != 0) continue;

        PrimeFacs.push_back(p);
        do { n /= p; } while (n%p == 0);
      }
      if (n != 1) PrimeFacs.push_back(n);
    }

  } // end of anonymous namespace


  /// Function to test whether a small integer is prime;
  /// note that n*n must fit inside an unsigned long.
  /// This definition is for 64-bit machines, but remains valid for 32-bitters.
  bool IsSmallPrime(unsigned long n)
  {
    if (n > SmallPrimeLimit ||
	n == 746331041UL ||
	n == 2840871041UL ||
	n == 3014101261UL ||
	n == 3215031751UL)
      return false;
    if (n == 2 || n == 3 || n == 5 || n == 7) return true;
    if (n < 11) return false;
    return StrongPseudoPrime(2, n) &&
           StrongPseudoPrime(5, n) &&
           StrongPseudoPrime(7, n);
  }


  /// Returns the least SmallPrime strictly greater than n, or 0 if
  /// there is no SmallPrime greater than n.
  unsigned long NextSmallPrime(unsigned long n)
  {
    if (n > SmallPrimeLimit) return 0; // to indicate overflow
    if (n < 2) return 2;
    if (n == 2) return 3;
    if (n < 5) return 5;

    unsigned long n30 = n%30;
    do
    {
      unsigned long delta = skip[n30];
      if (n30 == 29) n30 = 1;
      else n30 += delta;
      n += delta;
      if (n > SmallPrimeLimit) return 0; // to indicate overflow
    } while (!IsSmallPrime(n));
    return n;
  }


  /// Returns the greatest SmallPrime strictly less than n, or 0 if
  /// there is no (positive) SmallPrime less than n.
  unsigned long PrevSmallPrime(unsigned long n)
  {
    if (n < 3) return 0;
    if (n == 3) return 2;
    if (n < 6) return 3;
    if (n < 8) return 5;

    unsigned long n30 = n%30;
    do
    {
      unsigned long delta = fall[n30];
      if (n30 < delta) n30 += 30;
      n30 -= delta;
      n -= delta;
    } while (!IsSmallPrime(n));
    return n;
  }



  SmallPrimeSource::SmallPrimeSource(unsigned long StartingFrom)
  {
    myCurrentPrime = NextSmallPrime(StartingFrom);
  }


  SmallPrimeSource& SmallPrimeSource::operator++()
  {
    if (myCurrentPrime == 0) return *this;
    if (myCurrentPrime == 2) { myCurrentPrime = 3; return *this; }
    myCurrentPrime = NextSmallPrime(myCurrentPrime);
    return *this;
  }



  /// This function finds the least positive primitive root modulo p;
  /// if p is not a small prime, the value returned is 0.
  unsigned long FindPrimRoot(unsigned long p)
  {
    if (p == 2) return 1;
    if (p > SmallPrimeLimit || !IsSmallPrime(p)) return 0; // failure

    vector<unsigned long> PrimeFacs; // all integers below 2^32 have <= 9 distinct prime factors
                                     // all integers below 2^64 have <= 15 distinct prime factors
    Factorize32bits(PrimeFacs, p-1);
    for (unsigned long root = 2; ; ++root)
    {
      if (root == 4 || root == 8 || root == 9) continue; // skip the first few prime powers
      bool failed = false;
      for (vector<unsigned long>::iterator it = PrimeFacs.begin(); it != PrimeFacs.end(); ++it)
      {
	if (pwr(root, (p-1)/(*it), p) == 1)  // need brackets around *it to avoid "start of comment"
        { failed = true; break; } // effectively a "continue;" for the outer loop
      }
      if (!failed) return root;
    }
  }


} // end of namespace CoCoA


// RCS header/log
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SmallPrime.C,v 1.4 2009/05/14 12:25:44 abbott Exp $
// $Log: SmallPrime.C,v $
// Revision 1.4  2009/05/14 12:25:44  abbott
// Fixed embarrassing bug (tried to make a negative unsigned long :-)
//
// Revision 1.3  2009/05/14 09:19:38  abbott
// Cleaned code (e.g. to use true/false instead of 1/0 inherited from C).
// This file will probably soon change name...
//
// Revision 1.2  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.3  2006/12/21 13:48:32  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.2  2006/12/06 17:28:59  cocoa
// -- added #include "config.H"
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.2  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.3  2004/11/12 16:11:05  cocoa
// Some missing semicolons... oops.
// Now it compiles OK.
//
// Revision 1.2  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
// Revision 1.2  2003/06/23 16:11:05  abbott
// Minor cleaning prior to public release.
//
// Revision 1.1  2002/01/30 18:06:43  abbott
// Initial revision
//
