//   Copyright (c)  2005,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for abstract class SparsePolyRing and friends

#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/PPWithMask.H"  // for monomial ideals
#include "CoCoA/QQ.H"
#include "CoCoA/QuotientRing.H" // for IsQuotientRing/AsQuotientRing
#include "CoCoA/ReductionCog.H"
#include "CoCoA/RingDistrMPoly.H" // for NewPolyRing_DMP
#include "CoCoA/RingFp.H" // for IsRingFp
#include "CoCoA/RingQ.H" // for IsQ
#include "CoCoA/RingTwinFloat.H" // for IsRingTwinFloat
#include "CoCoA/RingZ.H" // for IsZ
#include "CoCoA/TmpGOperations.H"  // for myGcd
#include "CoCoA/TmpPPVector.H"  // for interreduce(PPs)
#include "CoCoA/assert.H"
#include "CoCoA/degree.H"
#include "CoCoA/error.H"
#include "CoCoA/geobucket.H" // for myMul
#include "CoCoA/ideal.H"     // for myGcd
#include "CoCoA/module.H"    // for myGcd
#include "CoCoA/symbol.H"

#include <algorithm>
using std::max;     // for MaxExponent, StdDeg
using std::sort;    // for AreGoodIndetNames
//#include <cstddef>
using std::size_t;
#include <functional>
using std::not1;    // for IsMonomial
using std::ptr_fun; // for IsMonomial
#include <iostream>
// using std::ostream in SparsePolyRingBase::myOutput
//#include <vector>
using std::vector;
#include <string>
// using std::string;

namespace CoCoA
{

  SparsePolyRing NewPolyRing(const ring& CoeffRing, const PPMonoid& PPM)
  {
    return NewPolyRing_DMP(CoeffRing, PPM);
  }

  SparsePolyRing NewPolyRing(const ring& CoeffRing, const std::vector<symbol>& IndetNames, const PPOrdering& ord)
  {
    return NewPolyRing_DMP(CoeffRing, IndetNames, ord);
  }

  SparsePolyRing NewPolyRing(const ring& CoeffRing, const std::vector<symbol>& IndetNames, const PPOrderingCtor& ord)
  {
    return NewPolyRing_DMP(CoeffRing, IndetNames, ord);
  }

  SparsePolyRing NewPolyRing(const ring& CoeffRing, const std::vector<symbol>& IndetNames)
  {
    return NewPolyRing_DMP(CoeffRing, IndetNames);
  }

  SparsePolyRing NewPolyRing(const ring& CoeffRing, std::size_t NumIndets)
  {
    return NewPolyRing_DMP(CoeffRing, NumIndets);
  }

  SparsePolyRing NewPolyRing(const ring& CoeffRing, std::size_t NumIndets, const PPOrderingCtor& ord)
  {
    return NewPolyRing_DMP(CoeffRing, NumIndets, ord);
  }

  bool AreGoodIndetNames(const ring& CoeffRing, const std::vector<symbol>& IndetNames)
  {
    // inefficient: we might know that SymbolsCR and IndetNames are good
    // does it matter?
    vector<symbol> syms = symbols(CoeffRing);
    copy(IndetNames.begin(), IndetNames.end(), back_inserter(syms));
    sort(syms.begin(), syms.end());
    const size_t NumSyms = syms.size();
    for (size_t i=0; i < NumSyms-1; ++i)
    {
      if (syms[i] == syms[i+1]) return false;
      if (head(syms[i]) == head(syms[i+1]) &&
          NumIndices(syms[i]) != NumIndices(syms[i+1]))
        return false;
    }
    return true;
  }
  

  //---- Functions for creating/building polynomials

  RingElem monomial(const SparsePolyRing& P, ConstRefRingElem c, ConstRefPPMonoidElem pp)
  {
    if (owner(c) != CoeffRing(P)) CoCoA_ERROR(ERR::MixedCoeffRings, "monomial(P,c,pp)");
    if (owner(pp) != PPM(P)) CoCoA_ERROR(ERR::MixedPPMs, "monomial(P,c,pp)");
    if (IsZero(c)) return zero(P);
    return P->myMonomial(raw(c), raw(pp));
  }

  RingElem monomial(const SparsePolyRing& P, MachineInteger n, ConstRefPPMonoidElem pp)
  {
    return monomial(P, RingElem(CoeffRing(P), n), pp);
  }

  RingElem monomial(const SparsePolyRing& P, const ZZ& N, ConstRefPPMonoidElem pp)
  {
    return monomial(P, RingElem(CoeffRing(P), N), pp);
  }

  RingElem monomial(const SparsePolyRing& P, const QQ& Q, ConstRefPPMonoidElem pp)
  {
    return monomial(P, RingElem(CoeffRing(P), Q), pp);
  }

  RingElem monomial(const SparsePolyRing& P, ConstRefRingElem c, const std::vector<long>& expv)
  {
    return monomial(P, c, PPMonoidElem(PPM(P), expv));
  }

  RingElem monomial(const SparsePolyRing& P, MachineInteger n, const std::vector<long>& expv)
  {
    return monomial(P, RingElem(CoeffRing(P), n), PPMonoidElem(PPM(P), expv));
  }

  RingElem monomial(const SparsePolyRing& P, const ZZ& N, const std::vector<long>& expv)
  {
    return monomial(P, RingElem(CoeffRing(P), N), PPMonoidElem(PPM(P), expv));
  }

  RingElem monomial(const SparsePolyRing& P, const QQ& Q, const std::vector<long>& expv)
  {
    return monomial(P, RingElem(CoeffRing(P), Q), PPMonoidElem(PPM(P), expv));
  }


  namespace // for functions local to this file/compilation unit.
  {
    inline void CheckCompatible(const ConstRefRingElem& x, const ConstRefRingElem& y, const std::string& fn)
    {
      if (owner(x) != owner(y))  CoCoA_ERROR(ERR::MixedRings, fn);
    }

    inline void CheckElemSparsePolyRing(const ConstRefRingElem& f, const std::string& fn)
    {
      if (!IsSparsePolyRing(owner(f))) CoCoA_ERROR(ERR::NotElemSparsePolyRing, fn);
    }

    void CheckCoeffExpv(const SparsePolyRing& P,
                        ConstRefRingElem c, const std::vector<long>& expv,
                        const std::string& s)
    {
      if (CoeffRing(P) != owner(c))    CoCoA_ERROR(ERR::MixedCoeffRings, s);
      if (NumIndets(P) != expv.size()) CoCoA_ERROR(ERR::BadArraySize, s);
    }
    
    void CheckCoeffPP(const SparsePolyRing& P,
                      ConstRefRingElem c, ConstRefPPMonoidElem pp,
                      const std::string& s)
    {
      if (CoeffRing(P) != owner(c))    CoCoA_ERROR(ERR::MixedCoeffRings, s);
      if (PPM(P) != owner(pp))         CoCoA_ERROR(ERR::MixedPPMs, s);
    }
  }
  
  
  RefRingElem PushFront(RefRingElem f, ConstRefRingElem c, const std::vector<long>& expv) /// SHOULD BE vector<ZZ> ????
  {
    CheckElemSparsePolyRing(f, "PushFront(f, c, expv)");
    const SparsePolyRing Rx = AsSparsePolyRing(owner(f));
    CheckCoeffExpv(Rx, c, expv, "PushFront(f, c, expv)");
    if (!IsZero(f) && PPMonoidElem(PPM(Rx), expv) <= LPP(f))
      CoCoA_ERROR(ERR::PPOrder, "PushFront(f, c, expv)");
    Rx->myPushFront(raw(f), raw(c), expv);
    return f;
  }


  RefRingElem PushFront(RefRingElem f, ConstRefRingElem c, ConstRefPPMonoidElem pp)
  {
    CheckElemSparsePolyRing(f, "PushFront(f, c, pp)");
    const SparsePolyRing Rx = AsSparsePolyRing(owner(f));
    CheckCoeffPP(Rx, c, pp, "PushFront(f, c, pp)");
    if (!IsZero(f) && pp <= LPP(f)) CoCoA_ERROR(ERR::PPOrder, "PushFront(f, c, pp)");
    Rx->myPushFront(raw(f), raw(c), raw(pp));
    return f;
  }


  RefRingElem PushBack(RefRingElem f, ConstRefRingElem c, const std::vector<long>& expv) /// SHOULD BE vector<ZZ> ????
  {
    CheckElemSparsePolyRing(f, "PushBack(f, c, expv)");
    const SparsePolyRing Rx = AsSparsePolyRing(owner(f));
    CheckCoeffExpv(Rx, c, expv, "PushBack(f, c, expv)");
    Rx->myPushBack(raw(f), raw(c), expv);
    return f;
  }


  RefRingElem PushBack(RefRingElem f, ConstRefRingElem c, ConstRefPPMonoidElem pp)
  {
    CheckElemSparsePolyRing(f, "PushBack(f, c, pp)");
    const SparsePolyRing Rx = AsSparsePolyRing(owner(f));
    CheckCoeffPP(Rx, c, pp, "PushBack(f, c, pp)");
    Rx->myPushBack(raw(f), raw(c), raw(pp));
    return f;
  }

  /*----------------------------------------------------------------------
    Member functions every concrete SparsePolyRing implementation
    must have in addition to those of PolyRingBase.
    ----------------------------------------------------------------------*/

  bool SparsePolyRingBase::myIsValid(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) return true;
    SparsePolyIter itf = myBeginIter(rawf);
    if (IsZero(coeff(itf))) return false;
    PPMonoidElem PrevPP = PP(itf);
    for (++itf; !IsEnded(itf); ++itf)
    {
      if (IsZero(coeff(itf))) return false;
      if (PrevPP <= PP(itf)) return false;
      PrevPP = PP(itf);
    }
    return true;
  }


  // ANNA: add check if ordering is StdDeg compatible and return StdDeg(LPP)
  long SparsePolyRingBase::myStdDeg(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) CoCoA_ERROR(ERR::ZeroRingElem, "myStdDeg(rawf)");
    long PolyDegree = 0;
    for (SparsePolyIter i=myBeginIter(rawf); !IsEnded(i); ++i)
      PolyDegree = max(PolyDegree, StdDeg(PP(i)));
    return PolyDegree;
  }


  long SparsePolyRingBase::myMaxExponent(ConstRawPtr rawf, size_t index) const
  {
    if (myIsZero(rawf)) CoCoA_ERROR(ERR::ZeroRingElem, "myMaxExponent(rawf, index)");
    long res = 0;
    for (SparsePolyIter i=myBeginIter(rawf); !IsEnded(i); ++i)
      res = max(res, exponent(PP(i), index));
    return res;
  }


  void SparsePolyRingBase::myContent(RawPtr rawcontent, ConstRawPtr rawf) const
  {
    CoCoA_ASSERT(IsGCDDomain(myCoeffRing()));
    const ring& R = myCoeffRing();
    // Compute answer in local var to avoid aliasing problems; also exception clean.
    RingElem ans(R);
    for (SparsePolyIter i=myBeginIter(rawf); !IsEnded(i); ++i)
    {
      R->myGcd(raw(ans), raw(ans), raw(coeff(i))); // ans = GCD(ans, coeff(i));
      if (IsOne(ans)) break;
    }
    // Finally, swap answer into rawcontent.
    R->mySwap(rawcontent, raw(ans));
  }


  void SparsePolyRingBase::myCommonDenom(RawPtr rawcontent, ConstRawPtr rawf) const
  {
    CoCoA_ASSERT(IsFractionField(myCoeffRing()) && IsGCDDomain(BaseRing(AsFractionField(myCoeffRing()))));
    const ring& R = BaseRing(AsFractionField(myCoeffRing()));
    // Compute result in local "ans" to avoid aliasing problems; also exception clean.
    RingElem ans = one(R);
    for (SparsePolyIter it=myBeginIter(rawf); !IsEnded(it); ++it)
    {
      const RingElem D = den(coeff(it));
      ans *= D/gcd(ans, D);
    }
    // Finally, swap answer into rawcontent -- cheaper than assignment.
    R->mySwap(rawcontent, raw(ans));
  }


  void SparsePolyRingBase::myRemoveBigContent(RawPtr rawf) const
  {
    CoCoA_ASSERT(IsGCDDomain(myCoeffRing()));
    CoCoA_ASSERT(!myIsZero(rawf));
    RingElem cont(myCoeffRing());
    myContent(raw(cont), rawf);
    myDivByCoeff(rawf, raw(cont));
  }


  /*----------------------------------------------------------------------
    Member functions inherited from ring with a single implementation
    for all SparsePolyRing implementations
    ----------------------------------------------------------------------*/

  void SparsePolyRingBase::myMul(RawPtr rawlhs, ConstRawPtr rawf, ConstRawPtr rawg) const
  {
    if (myIsZero(rawf) || myIsZero(rawg)) { myAssignZero(rawlhs); return; }
    if (myIsConstant(rawf))
    {
      RingElem ans(ConstRefRingElem(ring(this), rawg));
      myMulByCoeff(raw(ans), raw(myLC(rawf)));  // weak exc guarantee, but not a problem here
      mySwap(rawlhs, raw(ans));
      return;
    }
    if (myIsConstant(rawg) && myCoeffRing()->IamCommutative())  // CoeffRing should be comm
    {
      myMul(rawlhs, rawg, rawf);
      return;
    }
    size_t gLen = myNumTerms(rawg);
    if (IamCommutative() && myNumTerms(rawf) > gLen) { myMul(rawlhs, rawg, rawf); return; }
    SparsePolyRing P(AsSparsePolyRing(ring(this)));
    ConstRefRingElem g(P, rawg);

    if (myIsMonomial(rawf))
    {
      RingElem ans(P);
      myAddMul(raw(ans), rawf, rawg);
      mySwap(raw(ans), rawlhs);
      return;
    }
    geobucket gbk(P);
    for (SparsePolyIter itf=myBeginIter(rawf); !IsEnded(itf); ++itf)
      gbk.myAddMul(monomial(P, coeff(itf), PP(itf)), g, gLen);
    RingElem ans(P);
    AddClear(ans, gbk); // this is for exception safety
    mySwap(raw(ans), rawlhs); // really an assignment
  }


  bool SparsePolyRingBase::myIsDivisible(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    if (myIsZero(rawy)) { return false; }
    if (myIsZero(rawx)) { myAssignZero(rawlhs); return true; }
    if (myIsConstant(rawy))
    {
      RingElem ans(ConstRefRingElem(ring(this), rawx));
      if (!myDivByCoeff(raw(ans), raw(myLC(rawy))))  // weak exc guarantee, but not a problem here
        return false;
      mySwap(rawlhs, raw(ans));
      return true;
    }
    if (!IamCommutative()) { CoCoA_ERROR(ERR::NYI, "SparsePolyRingBase::myDiv non commutative"); }
    SparsePolyRing P = AsSparsePolyRing(ring(this));
    RingElem xCopy(ConstRefRingElem(P, rawx));
    geobucket gbk(P);
    gbk.myAddClear(xCopy, NumTerms(xCopy));
    ConstRefRingElem y(P, rawy);
    size_t yLen = NumTerms(y);
    RingElem ans(P);
    const ring& R = myCoeffRing();
    RingElem coeff(R);
    // if not divisible will throw ERR::BadQuot
    while ( !IsZero(gbk) )
    {
      if (!R->myIsDivisible(raw(coeff), raw(LC(gbk)), raw(myLC(rawy)))) return false;
      if (!IsDivisible(LPP(gbk), LPP(y))) return false;      
      RingElem m(monomial(P, coeff, LPP(gbk)/LPP(y)));
      // xCopy -= m*y;
      gbk.myAddMul(-m, y, yLen);
      // ans += m;
      myAppendClear(raw(ans), raw(m));
    }
    mySwap(raw(ans), rawlhs); // really an assignment
    return true;
  }


  void SparsePolyRingBase::myDeriv(RawPtr rawlhs, ConstRawPtr rawf, ConstRawPtr rawx) const
  {
    if (myIsOne(rawx)) { myAssign(rawlhs, rawf); return; }
    const size_t n = myNumIndets();
    ring R = myCoeffRing();
    SparsePolyRing P = AsSparsePolyRing(ring(this));
    vector<long> expv(n);
    exponents(expv, myLPP(rawx));

    RingElem ans(P);
    for (SparsePolyIter itf=myBeginIter(rawf); !IsEnded(itf); ++itf)
    {
      ZZ scale(1);
      for (size_t indet=0; indet < n; ++indet)
        if (expv[indet] != 0)
        {
          long d = exponent(PP(itf), indet);
          if (d < expv[indet]) { scale = 0; break; }
          scale *= RangeFactorial(d-expv[indet]+1, d);
        }
      if (IsZero(scale)) continue;
      RingElem m(monomial(P, scale*coeff(itf), PP(itf)/myLPP(rawx)));
      if (!IsZero(m)) myAppendClear(raw(ans), raw(m));
    }
    mySwap(raw(ans), rawlhs); // really an assignment
  }


  void SparsePolyRingBase::mySymbols(std::vector<symbol>& SymList) const
  {
    myCoeffRing()->mySymbols(SymList);
    myPPM()->mySymbols(SymList);
  }


  void SparsePolyRingBase::myOutput(std::ostream& out, ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) { out << "0"; return; }

    const ring& R = myCoeffRing();
    const PPMonoid PPM = myPPM();
    const bool IsQuotientOfZ = IsQuotientRing(R) && IsZ(BaseRing(AsQuotientRing(R)));
    //const bool IsQuotientOfZ = IsRingFp(R);
    const bool IsNumberRing = IsZ(R) || IsQuotientOfZ || IsQ(R) || IsRingTwinFloat(R);

//     for (DistrMPoly::summand* it = f.mySummands;
//          it != 0;
//          it = it->myNext, IsFirstCoeff = false)
    bool IsFirstCoeff=true;
    for (SparsePolyIter itf=myBeginIter(rawf); !IsEnded(itf) ; ++itf, IsFirstCoeff=false)
    {
      bool PrintStar = true;
      ConstRefRingElem c = coeff(itf);
      ConstRefPPMonoidElem pp = PP(itf);
      const bool IsOnePP = IsOne(pp);

      //----------------------------------------------------------------------
      // coefficient
      //----------------------------------------------------------------------

      if (!IsNumberRing)
      {
        if (!IsFirstCoeff) out << " +";
        if (R->myIsPrintAtom(raw(c))) out << c; else out << '(' << c << ')';
        goto EndCoeffPrint;
      }
      // assuming Q is printed with positive denominator
      if (!IsFirstCoeff)
      {
        if (R->myIsPrintedWithMinus(raw(c)))
          out << " ";
        else out << " +";
      }
      if (IsOnePP)
      {
        out << c;
        continue;
        goto EndCoeffPrint;
      }
      // Do not print "1 * "...
      if (IsOne(c))
      {
        PrintStar = false;
        goto EndCoeffPrint;
      }
      // Do not print "-1 * "...
      // NB condition below is complicated because in some Z/(n) minus one prints as n-1
      if ( IsMinusOne(c) && R->myIsPrintedWithMinus(raw(c)) )
      {
        PrintStar = false;
        out << "-";
        goto EndCoeffPrint;
      }
      // General case: coeff is neither +1 nor -1
      out << c;

    EndCoeffPrint:

      //----------------------------------------------------------------------
      // PP
      //----------------------------------------------------------------------
      if (!IsOnePP)
      {
        if (PrintStar)  out << "*";
        out << pp;
      }

    }
  }


  bool SparsePolyRingBase::myIsPrintAtom(ConstRawPtr rawx) const
  {
    if (!myIsConstant(rawx)) return false;
    return myCoeffRing()->myIsPrintAtom(raw(myLC(rawx)));
  }


  void SparsePolyRingBase::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("cocoa", "DMPSummands");
    OMOut << myNumTerms(rawx);

    for (SparsePolyIter itf=myBeginIter(rawx); !IsEnded(itf) ; ++itf)
    {
      OMOut << coeff(itf);
      OMOut << PP(itf);
//      R->myOutput(OMOut, it->myCoeff);
//      ordering(PPM)->myOutput(OMOut, it->myOrdv);
    }
    OMOut->mySendApplyEnd();
  }


  bool SparsePolyRingBase::myIsOne(ConstRawPtr rawf) const
  {
    if (!myIsMonomial(rawf)) return false;
    if (!IsOne(myLPP(rawf))) return false;
    return IsOne(myLC(rawf));
  }


  bool SparsePolyRingBase::myIsMinusOne(ConstRawPtr rawf) const
  {
    if (!myIsMonomial(rawf)) return false;
    if (!IsOne(myLPP(rawf))) return false;
    return IsOne(myLC(rawf));
  }


  bool SparsePolyRingBase::myIsConstant(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) return true;
    if (!myIsMonomial(rawf)) return false;
    return (IsOne(myLPP(rawf)));
  }


  bool SparsePolyRingBase::myIsIndet(std::size_t& IndetIndex, ConstRawPtr rawf) const
  {
    if (!myIsMonomial(rawf)) return false;
    if (!IsOne(myLC(rawf))) return false;
    return IsIndet(IndetIndex, myLPP(rawf));
  }


  bool SparsePolyRingBase::myIsHomogPartial(ConstRawPtr rawf, std::size_t n) const
  {
    if (myIsZero(rawf)) { return true; }
    //    SparsePolyRing P(AsSparsePolyRing(ring(this)));
    //    ConstRefRingElem f(P, rawf);
    SparsePolyIter itf=myBeginIter(rawf);
    PPMonoidElem PrevPP=PP(itf);
    for (++itf; !IsEnded(itf); ++itf)
    {
      CoCoA_ASSERT( cmp(PrevPP, PP(itf))>0 ); // assert f is correctly sorted
      if ( CmpWDegPartial(PrevPP, PP(itf), n)!=0 )  return false;
      PrevPP = PP(itf);
    }
    return true;
  }

  bool SparsePolyRingBase::myIsHomog(ConstRawPtr rawf) const
  {
    if (myIsZero(rawf)) { return true; }
    //    SparsePolyRing P(AsSparsePolyRing(ring(this)));
    //    ConstRefRingElem f(P, rawf);
    SparsePolyIter itf=myBeginIter(rawf);
    PPMonoidElem PrevPP=PP(itf);
    for (++itf; !IsEnded(itf); ++itf)
    {
      CoCoA_ASSERT( cmp(PrevPP, PP(itf))>0 ); // assert f is correctly sorted
      if ( CmpWDeg(PrevPP, PP(itf))!=0 )  return false;
      PrevPP = PP(itf);
    }
    return true;
  }


  void SparsePolyRingBase::myHomog(RawPtr rawfHom, ConstRawPtr rawf, ConstRawPtr rawh) const
  {
    SparsePolyRing P(AsSparsePolyRing(ring(this)));
    ConstRefRingElem h(P,rawh);
    CoCoA_ASSERT( myGradingDim()==1 );
    CoCoA_ASSERT( IsIndet(h) );
    CoCoA_ASSERT( IsOne(wdeg(h)[0]) );
    geobucket gbk(P);  // accumulate result into a geobucket for speed
    ConstRefRingElem f(P,rawf);
    ConstRefPPMonoidElem PPh = LPP(h);
    ZZ d = wdeg(f)[0];
    for (SparsePolyIter it=BeginIter(f) ; !IsEnded(it) ; ++it )
      gbk.myAddClear(monomial(P, coeff(it),PP(it)*power(PPh,d-wdeg(PP(it))[0])),1);
    RingElem tmp(P);  // for exception safety
    AddClear(tmp, gbk);
    mySwap(rawfHom, raw(tmp));
  }


  bool SparsePolyRingBase::myIsInvertible(ConstRawPtr rawx) const
  {
    return !myIsZero(rawx) && myIsConstant(rawx) && IsInvertible(myLC(rawx));
  }


  // code for R[x,y]: compute gcd in FractionField(R)[x,y]
  void SparsePolyRingBase::myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    if ( myIsZero(rawx) ) { myAssign(rawlhs, rawy); return; }
    if ( myIsZero(rawy) ) { myAssign(rawlhs, rawx); return; }
    if ( myIsInvertible(rawx) || myIsInvertible(rawy) ) { myAssign(rawlhs, raw(myOne())); return; }
    CoCoA_ASSERT(myNumIndets() != 0);
    SparsePolyRing P = AsSparsePolyRing(ring(this));
    if (myNumIndets()==1)
    {
      ideal I = ideal(ConstRefRingElem(P, rawx), ConstRefRingElem(P, rawy));
      const vector<RingElem>& GB = TidyGens(I);
      if (GB.size() != 1) CoCoA_ERROR("Unable to compute GCD", "SparsePolyRingBase::gcd");
      myAssign(rawlhs, raw(GB[0]));
      return;
    }
    vector<RingElem> L;
    L.push_back(ConstRefRingElem(P,rawx));
    L.push_back(ConstRefRingElem(P,rawy));
    vector<ModuleElem> v;
    ComputeSyz(v, L);
    if ( v.size()!=1 ) 
      CoCoA_ERROR("Unable to compute GCD", "SparsePolyRingBase::gcd");
    RingElem g = L.front()/((v[0])[1]);
    P->mySwap(rawlhs, raw(g));  // exception safe
  }


  bool SparsePolyRingBase::myIsInteger(ZZ& N, ConstRawPtr rawx) const
  {
    if (myIsZero(rawx)) { N = 0; return true; }
    if (myNumTerms(rawx) > 1) return false;
    if (!IsOne(myLPP(rawx))) return false;
    return IsInteger(N, myLC(rawx));
  }


  bool SparsePolyRingBase::myIsRational(QQ& Q, ConstRawPtr rawx) const
  {
    if (myIsZero(rawx)) { Q = 0; return true; }
    if (myNumTerms(rawx) > 1) return false;
    if (!IsOne(myLPP(rawx))) return false;
    return IsRational(Q, myLC(rawx));
  }


  void SparsePolyRingBase::mySquare(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    if (IamCommutative())
    {
      unsigned long NT = myNumTerms(rawx);
      if (NT > 4)
      {
        RingElem f(ring(this));
        RingElem g(ring(this));
        unsigned long count=0;
        for (SparsePolyIter it=myBeginIter(rawx) ; !IsEnded(it) ; ++it )
        {
          if (++count < NT/2) myPushBack(raw(f), raw(coeff(it)), raw(PP(it)));
          else myPushBack(raw(g), raw(coeff(it)), raw(PP(it)));
        }
        RingElem h = power(f, 2) + power(g, 2) + 2*f*g;
        mySwap(rawlhs, raw(h));
        return;
      }
    }
    mySequentialPower(rawlhs, rawx, 2); //??? BUG/SLUG myBinaryPower better if univariate or coeffs are finite field
  }


  void SparsePolyRingBase::myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Assert that we have a genuinely non-trivial case.
    CoCoA_ASSERT(n > 1);
    CoCoA_ASSERT(!myIsZero(rawx) && !myIsOne(rawx) && !myIsMinusOne(rawx));

    // anna 3 apr 2008: under testing
    if (myIsMonomial(rawx) && IamCommutative())
    {
      RingElem m = monomial(AsSparsePolyRing(ring(this)), power(myLC(rawx), n), power(myLPP(rawx), n));
      mySwap(rawlhs, raw(m));
      return;
    }
    if (myIsMonomial(rawx)) { myBinaryPower(rawlhs, rawx, n); return; }
    if (n==2)
    {
      mySquare(rawlhs, rawx);
      return;
    }
    mySequentialPower(rawlhs, rawx, n); //??? BUG/SLUG myBinaryPower better if univariate or coeffs are finite field
  }


  //---- Special functions on RingElem owned by SparsePolyRing

  degree wdeg(ConstRefRingElem f)
  {
    CheckElemSparsePolyRing(f, "wdeg(f)");
    if (IsZero(f)) CoCoA_ERROR(ERR::ZeroRingElem, "wdeg(f)");
    return wdeg(LPP(f)); // yes! see [KR] introduction Sec.4.3
  }


  int CmpWDeg(ConstRefRingElem f1, ConstRefRingElem f2)
  {
    CheckCompatible(f1, f2, "CmpWDeg(f1,f2)");
    CheckElemSparsePolyRing(f1, "CmpWDeg(f1,f2)");
    if (IsZero(f1) || IsZero(f2)) CoCoA_ERROR(ERR::ZeroRingElem, "CmpWDeg(f1,f2)");
    // Now we know that both f1 and f2 are non-zero
    //    return AsSparsePolyRing(owner(f1))->myCmpWDeg(raw(f1), raw(f2));
    return CmpWDeg(LPP(f1), LPP(f2));
  }


  int CmpWDegPartial(ConstRefRingElem f1, ConstRefRingElem f2, std::size_t i)
  {
    CheckCompatible(f1, f2, "CmpWDegPartial(f1,f2,i)");
    CheckElemSparsePolyRing(f1, "CmpWDegPartial(f1,f2,i)");
    if (IsZero(f1) || IsZero(f2)) CoCoA_ERROR(ERR::ZeroRingElem, "CmpWDegPartial(f1,f2,i)");
    // Now we know that both f1 and f2 are non-zero
    //    return AsSparsePolyRing(owner(f1))->myCmpWDeg(raw(f1), raw(f2));
    return CmpWDegPartial(LPP(f1), LPP(f2), i);
  }


  bool IsHomog(ConstRefRingElem f)
  {
    CheckElemSparsePolyRing(f, "IsHomog(f)");
    return AsSparsePolyRing(owner(f))->myIsHomog(raw(f));
  }


  bool IsHomogPartial(ConstRefRingElem f, std::size_t n)
  {
    CheckElemSparsePolyRing(f, "IsHomogPartial(f,n)");
    return AsSparsePolyRing(owner(f))->myIsHomogPartial(raw(f), n);
  }


  RingElem homog(ConstRefRingElem f, ConstRefRingElem h)
  {
    CheckCompatible(f, h, "homog(ConstRefRingElem, ConstRefRingElem)");
    SparsePolyRing P(AsSparsePolyRing(owner(f)));
    if ( GradingDim(P)!=1 )
      CoCoA_ERROR("GrDim must be 1", "homog(ConstRefRingElem, ConstRefRingElem)");
    if ( !IsIndet(h) )
      CoCoA_ERROR("second arg must be an indeterminate", "homog(ConstRefRingElem, ConstRefRingElem)");
    if ( wdeg(h)[0]!=1 )
      CoCoA_ERROR("degree of hom.indet must be 1", "homog(ConstRefRingElem, ConstRefRingElem)");
    RingElem fHom(P);
    P->myHomog(raw(fHom), raw(f), raw(h));
    return fHom;
  }


  //----------------------------------------------------------------------
  //??? the following functions to compute NR will be replaced by GBMill

  int FindReducerIndex(ConstRefPPMonoidElem pp, const vector<RingElem>& v)
  {
    const size_t nelems = v.size();
    for (size_t i=0; i < nelems; ++i)
      if (IsDivisible(pp, LPP(v[i])))
        return i;
    return -1;
  }


  inline int FindReducerIndex(const ReductionCog& F, const vector<RingElem>& v)
  {
    if ( IsActiveZero(F) ) return -1;
    return FindReducerIndex(ActiveLPP(F), v);
  }


  void ReduceActiveLM(ReductionCog& F, const vector<RingElem>& v)
  {
    int i;
    while ( (i = FindReducerIndex(F, v) ) != -1)
      F->myReduce(v[i]);
  }


  void reduce(ReductionCog& F, const vector<RingElem>& v)
  {
    ReduceActiveLM(F, v);
    while ( !IsActiveZero(F) )
    {
      F->myMoveToNextLM();
      ReduceActiveLM(F, v);
    }
  }
  //--------------------------------------------

  RingElem NR(ConstRefRingElem f, const vector<RingElem>& v)
  {
    if ( IsZero(f) ) return f;
    RingElem ans(f);
    ReductionCog F = NewRedCogGeobucketField(AsSparsePolyRing(owner(ans)));
    F->myAssignReset(ans);
    reduce(F, v);
    F->myRelease(ans);
    return ans;
  }


  SparsePolyIter BeginIter(ConstRefRingElem f)
  {
    CheckElemSparsePolyRing(f, "BeginIter(f)");
    return AsSparsePolyRing(owner(f))->myBeginIter(raw(f));
  }


  SparsePolyIter EndIter(ConstRefRingElem f)
  {
    CheckElemSparsePolyRing(f, "EndIter(f)");
    return AsSparsePolyRing(owner(f))->myEndIter(raw(f));
  }

  //-- ideals ----------------------------------------

  bool IsZeroDim(const ideal& I)
  {
    if (!IsSparsePolyRing(AmbientRing(I)))
      CoCoA_ERROR(ERR::NotSparsePolyRing, "IsZeroDim(I)");
    if (IsZero(I) || IsOne(I)) return false;
    // Now we know I is non-trivial.
    SparsePolyRing P = AsSparsePolyRing(AmbientRing(I));
    const vector<RingElem>& GB = TidyGens(I);
    const size_t GBsize = GB.size(); // MUST BE A REDUCED GBASIS !!!
    size_t var; // only used in call to IsIndetPower
    ZZ EXP;     // only used in call to IsIndetPower
    size_t NumIndetPowers = 0;
    for (size_t i=0; i < GBsize; ++i)
      if (IsIndetPosPower(var, EXP, LPP(GB[i])))
        ++NumIndetPowers;
    return (NumIndetPowers == NumIndets(P));
  }


  //-- IdealImpl ----------------------------------------

  ideal SparsePolyRingBase::myIdealCtor(const std::vector<RingElem>& gens) const
  {
    return ideal(new IdealImpl(SparsePolyRing(this), gens)); //??? ugly ???
  }


  SparsePolyRingBase::IdealImpl::IdealImpl(const SparsePolyRing& P, const std::vector<RingElem>& gens):
      myP(P),
      myGensValue(gens),
      myGBasisIsValid(false)
  {
    // IhaveMonomialGensValue = bool3::uncertain; // default for bool3
    if (!IsField(CoeffRing(P)))
      CoCoA_ERROR("NYI ideal of polynomials with coeffs not in a field", "ideal(SparsePolyRing, gens)");//???
    for (size_t i=0; i < gens.size(); ++i)
      if (owner(gens[i]) != myP) CoCoA_ERROR(ERR::MixedRings, "SparsePolyRingBase::IdealImpl ctor");
  }


  IdealBase* SparsePolyRingBase::IdealImpl::clone() const
  {
    return new IdealImpl(*this);
  }


  const SparsePolyRing& SparsePolyRingBase::IdealImpl::myAmbientRing() const
  {
    return myP;
  }


  bool SparsePolyRingBase::IdealImpl::IamZero() const
  {
    return myGBasis().empty();
  }


  // should IsMonomial be visible??? Anna, 2009-09-24
  
  namespace // anonymous
  {
    bool IsMonomial(const std::vector<RingElem>& v)
    {
      // morally:  return find_if(v.begin(), v.end(), not1(IsMonomial)) == v.end();
      return find_if(v.begin(), v.end(),
                     not1(ptr_fun(static_cast<bool(*)(ConstRefRingElem)>(CoCoA::IsMonomial))))
        == v.end(); 
    }
  }


  bool SparsePolyRingBase::IdealImpl::IhaveMonomialGens() const
  {
    if (IsUncertain(IhaveMonomialGensValue))
      IhaveMonomialGensValue = IsMonomial(myGensValue);
    return IsDefinitelyTrue(IhaveMonomialGensValue);
  }


  //------- functions for monomial ideals: where should they go? ------
  void MonomialInterreduce(std::vector<RingElem>& res, const std::vector<RingElem>& l)
  {
    if (l.empty()) { res.clear(); return; }
    SparsePolyRing P = AsSparsePolyRing(owner(l[0]));
    PPVector g(PPM(P), NewDivMaskEvenPowers());
    for (vector<RingElem>::const_iterator it=l.begin(); it!=l.end() ; ++it)
      PushBack(g, LPP(*it));
    interreduce(g);
    convert(res, P, g);
  }


  void MonomialIntersection(std::vector<RingElem>& res, const std::vector<RingElem>& l1,  const std::vector<RingElem>& l2)
  {
    if (l1.empty()) { res.clear(); return; }
    if (l2.empty()) { res.clear(); return; }
    SparsePolyRing P = AsSparsePolyRing(owner(l1[0]));
    PPVector g1(PPM(P), NewDivMaskEvenPowers());
    for (vector<RingElem>::const_iterator it=l1.begin(); it!=l1.end() ; ++it)
      PushBack(g1, LPP(*it));
    PPVector g2(PPM(P), NewDivMaskEvenPowers());
    for (vector<RingElem>::const_iterator it=l2.begin(); it!=l2.end() ; ++it)
      PushBack(g2, LPP(*it));
    PPVector g(PPM(P), NewDivMaskEvenPowers());
    g.myLcms(g1, g2);
    convert(res, P, g);
  }
  //------- functions for monomial ideals: end ------


  void SparsePolyRingBase::IdealImpl::myMaximalTest() const
  {
    CoCoA_ERROR(ERR::NYI, "SparsePolyRingBase::IdealImpl::myMaximalTest()");//???
  }


  void SparsePolyRingBase::IdealImpl::myPrimeTest() const
  {
    CoCoA_ERROR(ERR::NYI, "SparsePolyRingBase::IdealImpl::myPrimeTest()");//???
  }


  void SparsePolyRingBase::IdealImpl::myReduceMod(RingElemRawPtr rawf) const
  {
    //??? very basic default implementation
    RefRingElem(myP, rawf) = NR(ConstRefRingElem(myP, rawf), myGBasis());
  }


  bool SparsePolyRingBase::IdealImpl::contains(RingElemConstRawPtr rawf) const
  {
    RingElem g = ConstRefRingElem(myP, rawf);
    myReduceMod(raw(g));
    return IsZero(g);
  }


  inline const SparsePolyRingBase::IdealImpl* SparsePolyRingBase::IdealImpl::GetPtr(const ideal& I)
  {
    return dynamic_cast<const SparsePolyRingBase::IdealImpl*>(I.myIdealPtr());
  }


  void SparsePolyRingBase::IdealImpl::add(const ideal& Jin)
  {
    const IdealImpl* const J = GetPtr(Jin);
    copy(J->myGensValue.begin(), J->myGensValue.end(), back_inserter(myGensValue));
    if (!IhaveMonomialGens() ||
        !dynamic_cast<const SparsePolyRingBase::IdealImpl*>(Jin.myIdealPtr())->IhaveMonomialGens())
      IhaveMonomialGensValue = bool3::uncertain;
    myGBasisIsValid = false;
    myGBasisValue.clear();
  }


  void SparsePolyRingBase::IdealImpl::mul(const ideal& Jin)
  {
    vector<RingElem> tmpV;
    const IdealImpl* const J = GetPtr(Jin);
    for (vector<RingElem>::const_iterator itI=myGensValue.begin(); itI!=myGensValue.end(); ++itI)
      for (vector<RingElem>::const_iterator itJ=J->myGensValue.begin(); itJ!=J->myGensValue.end(); ++itJ)
        tmpV.push_back((*itI)*(*itJ));
    myGensValue.clear();
    swap(tmpV, myGensValue);  // ANNA does this make copies???  2010-02-03
    if (!IhaveMonomialGens() ||
        !dynamic_cast<const SparsePolyRingBase::IdealImpl*>(Jin.myIdealPtr())->IhaveMonomialGens())
      IhaveMonomialGensValue = bool3::uncertain;
    myGBasisIsValid = false;
    myGBasisValue.clear();
    //    IamPrimeFlag = IamMaximalFlag = bool3::uncertain;
  }


  void SparsePolyRingBase::IdealImpl::intersect(const ideal& J)
  {
    if (IhaveMonomialGens() && dynamic_cast<const SparsePolyRingBase::IdealImpl*>(J.myIdealPtr())->IhaveMonomialGens())
    {
      // IhaveMonomialGensValue = true; // stays constant
      MonomialIntersection(myGensValue, myGensValue, gens(J));
    }
    else
    {
      IhaveMonomialGensValue = bool3::uncertain;
      ComputeIntersection(myGensValue, myGensValue, gens(J));
    }
    myGBasisIsValid = false;
    myGBasisValue.clear();
  }


  void SparsePolyRingBase::IdealImpl::colon(const ideal& J)
  {
    ComputeCColon(myGensValue,myGensValue, gens(J));
    IhaveMonomialGensValue = bool3::uncertain;
    myGBasisIsValid = false;
    myGBasisValue.clear();
 }


  void SparsePolyRingBase::IdealImpl::myDivMod(RingElemRawPtr /*rawlhs*/, RingElemConstRawPtr /*rawnum*/, RingElemConstRawPtr /*rawden*/) const
  {
    CoCoA_ERROR(ERR::NYI, "SparsePolyRingBase::IdealImpl::myDivMod");
  }


  const std::vector<RingElem>& SparsePolyRingBase::IdealImpl::myGens() const
  {
    return myGensValue;
  }


  const std::vector<RingElem>& SparsePolyRingBase::IdealImpl::myTidyGens() const
  {
    return myGBasis();
  }


  // ??? copied from ex-QuotientBasis.C
  bool IsDivisible(ConstRefPPMonoidElem pp, const vector<PPMonoidElem>& ByL)
  {
    for (size_t i=0; i < ByL.size(); ++i)
      if ( IsDivisible(pp, ByL[i])) return true;
    return false;
  }


  const std::vector<RingElem>& SparsePolyRingBase::IdealImpl::myGBasis() const
  {
    if (myGBasisIsValid) return myGBasisValue;
    CoCoA_ASSERT(myGBasisValue.empty());
    //    if ( IsMonomial(myGensValue) )
    if (IhaveMonomialGens())
      MonomialInterreduce(myGBasisValue, myGensValue);
    else
      ComputeGBasis(myGBasisValue, myGensValue);
    myGBasisIsValid = true;
    return myGBasisValue;
  }


  //-- HomImpl ----------------------------------------

  SparsePolyRingBase::HomImpl::HomImpl(const SparsePolyRing& domain, const ring& codomain, const RingHom& CoeffHom, const vector<RingElem>& IndetImages):
      RingHomBase(domain, codomain),
      myCoeffHom(CoeffHom),
      myIndetImages(IndetImages)
  {
    // No need to check anything: checks already made when CoeffHom was built.
  }

namespace
{
  // ??? appropriate use of inheritance here?  this is getting pretty hugly

  // assume image==0
  void ApplySPRCodomain(RefRingElem image, ConstRefRingElem arg, const RingHom CoeffHom, const vector<RingElem>& IndetImages)
  {
    SparsePolyRing S = AsSparsePolyRing(owner(image));
    geobucket gbk(S);

    const size_t NumInd = IndetImages.size();
    for (SparsePolyIter i=BeginIter(arg); !IsEnded(i); ++i)
    {
      RingElem SummandImage = CoeffHom(coeff(i));
      CoCoA_ASSERT(owner(SummandImage) == S);
      if (IsZero(SummandImage)) continue; // efficiency hack????
      ConstRefPPMonoidElem t(PP(i));
      for (size_t ind=0; ind < NumInd; ++ind)
      {
        long d = exponent(t, ind); // ??? should we compute exponents?
        if (d == 0) continue;
        SummandImage *= power(IndetImages[ind], d);
      }
      //        AsSparsePolyRing(S)->myAddClear(raw(ans), raw(SummandImage));
      gbk.myAddClear(SummandImage, NumTerms(SummandImage));
    }
    AddClear(image, gbk);
  }


  void ApplyGeneral(RefRingElem image, ConstRefRingElem arg, const RingHom CoeffHom, const vector<RingElem>& IndetImages)
  {
    ring S = owner(image);
    const size_t NumInd = IndetImages.size();
    for (SparsePolyIter i=BeginIter(arg); !IsEnded(i); ++i)
    {
      RingElem SummandImage = CoeffHom(coeff(i));
      CoCoA_ASSERT(owner(SummandImage) == S);
      if (IsZero(SummandImage)) continue; // efficiency hack????
      ConstRefPPMonoidElem t(PP(i));
      for (size_t ind=0; ind < NumInd; ++ind)
      {
        long d = exponent(t, ind); // ??? should we compute exponents?
        if (d == 0) continue;
        SummandImage *= power(IndetImages[ind], d);
      }
      S->myAdd(raw(image), raw(image), raw(SummandImage));
    }
  }
}  // end of anonymous namespace

  void SparsePolyRingBase::HomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    RingElem ans(myCodomain);  // Putting result into ans is exception safe and avoids aliasing problems.
    if ( IsSparsePolyRing(myCodomain) )
      ApplySPRCodomain(ans, ConstRefRingElem(myDomain, rawarg), myCoeffHom, myIndetImages);
    else
      ApplyGeneral(ans, ConstRefRingElem(myDomain, rawarg), myCoeffHom, myIndetImages);
    myCodomain->mySwap(rawimage, raw(ans));
  }


  void SparsePolyRingBase::HomImpl::myOutputSelfDetails(std::ostream& out) const
  {
    SparsePolyRing P = AsSparsePolyRing(myDomain);
    if (NumIndets(P) == 0) return;
    out << " sending "
        << "(" << indet(P, 0) << " |--> " << myIndetImages[0] << ")";
    for (size_t i=1; i < myIndetImages.size(); ++i)
    {
      out << " & (" << indet(P, i) << " |--> " << myIndetImages[i] << ")";
    }
  }


  RingHom SparsePolyRingBase::myHomCtor(const ring& codomain, const RingHom& CoeffHom, const std::vector<RingElem>& IndetImages) const
  {
    // Args already sanity checked by PolyRingHom (see PolyRing.C)
// DON'T KNOW IF I REALLY WANT TO MAKE THIS CHECK...
//       // Check to see if we're building an identity homomorphism
//       if (ring(this) == codomain && IsIdentity(CoeffHom))
//       {
//         bool IndetsFixed = true;
//         for (size_t i=0; i < myNumIndetsValue; ++i)
//           IndetsFixed &= (myIndetVector[i] == IndetImages[i]);
//         if (IndetsFixed) return IdentityHom(ring(this));
//       }
      // General case
    return RingHom(new HomImpl(SparsePolyRing(this), codomain, CoeffHom, IndetImages));
  }


  RingHom SparsePolyRingBase::myCompose(const RingHom& phi, const RingHom& theta) const
  {
    vector<RingElem> IndetImages;
    for (size_t var=0; var < myNumIndets(); ++var)
      IndetImages.push_back(phi(theta(myIndets()[var])));

    return myHomCtor(codomain(phi), phi(theta(myCoeffEmbeddingHom())), IndetImages);
  }


  //-- CoeffEmbeddingHomImpl ----------------------------------------

  //---------------------------------------------------------------------------
  // Functions for the class SparsePolyRingBase::CoeffEmbeddingHomImpl


  SparsePolyRingBase::CoeffEmbeddingHomImpl::CoeffEmbeddingHomImpl(const SparsePolyRing& P):
    RingHomEmbeddingBase(CoeffRing(P), P)
  {}


  void SparsePolyRingBase::CoeffEmbeddingHomImpl::myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const
  {
    SparsePolyRing P(AsSparsePolyRing(myCodomain));
    RingElem ans(P);  // don't use image here for aliasing
    // ??? ANNA profile this:  (probably better to have myMonomial)
    if (!myDomain->myIsZero(rawarg))
      ans = monomial(P, ConstRefRingElem(myDomain, rawarg), one(PPM(P)));
    P->mySwap(rawimage, raw(ans));
  }


} // end of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SparsePolyRing.C,v 1.47 2010/03/18 13:55:56 abbott Exp $
// $Log: SparsePolyRing.C,v $
// Revision 1.47  2010/03/18 13:55:56  abbott
// Added pseudo-ctors for monomials with QQ coeffs.
//
// Revision 1.46  2010/03/05 18:43:48  abbott
// Added pseudo-ctors allowing polynomial rings to be created specifying
// the ordering using a PPOrderingCtor object.
//
// Revision 1.45  2010/02/04 10:14:38  bigatti
// -- fixed "mul" for ideal
//
// Revision 1.44  2010/02/04 09:57:11  bigatti
// -- added "mul" for ideals.  Implemented only for SparsePolyRing
//
// Revision 1.43  2010/02/03 14:15:58  bigatti
// -- moved IsMonomial(vector<RingElem>) into anomymous namespace
//
// Revision 1.42  2010/01/21 13:17:08  bigatti
// -- commented out unnecessary default definition for bool3
//
// Revision 1.41  2010/01/20 16:49:52  bigatti
// -- fixed IhaveMonomialGensValue in some functions modifying the ideal
//
// Revision 1.40  2009/11/26 17:21:33  bigatti
// -- added PushFront/PushBack(f, c, pp)
// -- added in .C inline functions:
// --   CheckCompatible, CheckElemSparsePolyRing, CheckCoeffExpv, CheckCoeffPP
//
// Revision 1.39  2009/10/29 18:43:23  abbott
// Added necessary include directive for iostream
// (previously was wrongly commented out).
//
// Revision 1.38  2009/10/02 13:27:26  bigatti
// -- unique implementation of myDiv in PolyRing.C
//
// Revision 1.37  2009/09/28 16:19:43  bigatti
// -- unique implementation for myDeriv
//
// Revision 1.36  2009/09/25 13:09:27  bigatti
// -- fixed: reinserted #include QuotientRing
//
// Revision 1.35  2009/09/25 13:02:09  bigatti
// -- myDiv with one implementation in SparsePolyRing
//
// Revision 1.34  2009/09/24 16:24:32  abbott
// Added include directive (after removing it from RingFp.H).
//
// Revision 1.33  2009/09/22 14:07:33  bigatti
// -- added CmpWDegPartial and IsHomogPartial
//
// Revision 1.32  2009/07/30 15:43:56  bigatti
// -- new implementation of IsMonomial(vector<RingElem>) using STL functions
//
// Revision 1.31  2009/07/24 12:26:42  abbott
// Added CommonDenom function for polynomials.
//
// Revision 1.30  2009/07/02 16:32:11  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.29  2009/05/22 10:27:08  bigatti
// -- fixed myOutput and myIsPrintAtom
//
// Revision 1.28  2009/05/21 14:49:33  abbott
// Cleaned up myOutput: now handles -1 in small finite fields correctly
// (if residues are symmetric).
//
// Revision 1.27  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.26  2008/11/18 15:20:09  bigatti
// -- added const to myGBasis return value
// -- added myIdealCtor to RingWeyl for proper inheritance
//
// Revision 1.25  2008/10/09 15:08:42  bigatti
// -- added MonomialIntersection
//
// Revision 1.24  2008/07/04 09:11:04  bigatti
// -- new PPVector class
//
// Revision 1.23  2008/05/29 16:02:32  bigatti
// -- improved myMul for the case [coeff * poly]
//
// Revision 1.22  2008/05/27 16:43:35  bigatti
// -- minor change
//
// Revision 1.21  2008/05/27 16:25:59  bigatti
// -- moved code for monomial ideals into TmpPPVector
// -- small improvement for myMul
//
// Revision 1.20  2008/04/15 15:53:07  bigatti
// -- new for SparsePolyRing: mySquare, first draft
//
// Revision 1.19  2008/04/10 16:36:03  bigatti
// -- prototype for squaring polynomials
//
// Revision 1.18  2008/04/03 13:07:16  bigatti
// -- improved "myPowerSmallExp" for monomial input
//
// Revision 1.17  2008/03/12 16:35:18  bigatti
// -- changed: IsHomogeneous --> IsHomog
// -- changed: ERR:ZeroPoly --> ERR::ZeroRingElem
//
// Revision 1.16  2007/12/07 15:27:01  bigatti
// -- default implementation of "IamOne" in ideal.C
//
// Revision 1.15  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.14  2007/12/04 14:27:06  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.13  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.12  2007/10/05 14:37:42  bigatti
// -- just fixed a comment
//
// Revision 1.11  2007/09/24 14:25:14  abbott
// Changed IsIndetPower to IsIndetPosPower.
// Added parens to shut up gcc-4.3.
//
// Revision 1.10  2007/06/25 11:38:15  bigatti
// -- reverted IsIndetPosPower --> IsIndetPower
//
// Revision 1.9  2007/06/21 21:29:47  abbott
// Changed name of RingFloat into RingTwinFloat.
//
// Revision 1.8  2007/05/31 16:34:37  bigatti
// -- Changed IsValid (now returns true of false and does not throw an error)
// -- using IsValid for sanity check in PushBack
//
// Revision 1.7  2007/05/31 15:43:56  bigatti
// -- added mySymbols and AreGoodIndetNames
//
// Revision 1.5  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.4  2007/05/21 14:46:34  bigatti
// -- PushFront now accepts zero coefficient
//
// Revision 1.3  2007/03/12 16:00:29  bigatti
// -- moved myLog(F, index) into unique implementation in SparsePolyRing
//
// Revision 1.2  2007/03/09 18:56:56  bigatti
// -- added Tmp prefix to Groebner related files
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.32  2007/03/08 18:49:20  bigatti
// -- removed TmpGcd.C: moved code into SparsePolyRing.C
//
// Revision 1.31  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.30  2007/03/08 11:07:12  cocoa
// Made pseudo ctors for polynomial rings more uniform.  This allowed me to
// remove an include of CoCoA/symbol.H  from the RingDistrM*.H files, but then
// I had to put the include in several .C files.
//
// Revision 1.29  2007/03/07 17:04:31  cocoa
// -- several changes by M.Caboara: more operations on ideals,
//    exception cleaner, coding conventions, WSugar, dynamic
//
// Revision 1.28  2007/03/07 14:08:52  bigatti
// -- minor: commented argument names for -Wextra
//
// Revision 1.27  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.26  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.25  2007/02/28 13:51:59  bigatti
// -- added function IsMonomial
//
// Revision 1.24  2007/02/26 15:00:01  bigatti
// -- just a comment fix
//
// Revision 1.23  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.22  2007/01/20 14:07:25  bigatti
// -- moved code for homomorphism into common implementation in SparsePolyRing
//
// Revision 1.21  2006/12/07 17:36:19  cocoa
// -- migrated  myRemoveBigContent myContent myPowerSmallExp  into
//    single implementation in SparsePolyRing
// -- removed  content  from DistrMPoly(..)
//
// Revision 1.20  2006/11/24 17:06:10  cocoa
// -- reorganized includes of header files
//
// Revision 1.19  2006/11/23 17:48:43  cocoa
// -- minor change
//
// Revision 1.18  2006/11/22 17:51:31  cocoa
// -- moved printing functions into unified implementation in SparsePolyRing
//
// Revision 1.17  2006/11/22 15:12:48  cocoa
// -- minor cleaning (indicated by Intel compiler)
//
// Revision 1.16  2006/11/21 18:09:23  cocoa
// -- added myIsMonomial
// -- implemented myIsOne, myIsMinusOne, myIsConstant, myIsIndet in SparsePolyRing
// -- removed the 4 functions from DistrMPoly(..) and RingDistrMPoly(..)
// -- changed all names of RawPtr arguments into "raw(..)"
//
// Revision 1.15  2006/11/17 18:14:01  cocoa
// -- fixed: now Hilbert computes TidyGens instead of cheating
// -- myGBasis is much more efficient on monomial input
//
// Revision 1.14  2006/11/09 17:46:58  cocoa
// -- version 0.9712:
// --   IdealImpl moved to SparsePolyRing from concrete rings
// -- PolyList in GTypes is now vector<RingElem> (was list)
// -- "my" coding convention applied to DistrMPoly
//
// Revision 1.13  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.12  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.11  2006/09/27 14:30:45  cocoa
// -- changed f into rawf for ConstRawPtr and RawPtr variables
//
// Revision 1.10  2006/08/17 09:45:07  cocoa
// -- added: homogenization
//
// Revision 1.9  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.8  2006/07/20 17:06:08  cocoa
// -- moved myStdDeg into SparsePolyRing
//
// Revision 1.7  2006/07/20 16:51:38  cocoa
// -- added common implementation of myStdDeg
//
// Revision 1.6  2006/07/20 14:24:12  cocoa
// -- added special cases for myMul (0) and myGcd (unit)
//
// Revision 1.5  2006/07/17 19:32:55  cocoa
// Added arg check to PushFront.
//
// Revision 1.4  2006/07/17 11:05:53  cocoa
// -- added: myIsValid, myIsHomogeneous, IsHomogeneous
//
// Revision 1.3  2006/06/20 17:25:27  cocoa
// -- added function geobucket::myAddMul(monom, g, gLen);   [without SkipLMFlag]
//
// Revision 1.2  2006/06/08 16:45:27  cocoa
// -- RingDistrMPoly*.H  have been "moved" into RingDistrMPoly*.C
// -- some coding conventions fixed in DistrMPoly*
// -- functions wdeg and CmpWDeg have a common implementation in SparsePolyRing
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.11  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.10  2006/05/12 17:00:34  cocoa
// -- added some function whose implementation in RingDistr*** were identical
// -- added multivariate gcd
//
// Revision 1.9  2006/05/12 13:16:30  cocoa
// Added functions for preprocessing approximate points.
//
// Revision 1.8  2006/04/27 14:08:24  cocoa
// -- sorry: missed parenthesis
//
// Revision 1.7  2006/04/27 14:05:29  cocoa
// -- changed CmpWDeg: I think it should throw an error if polys are 0
//
// Revision 1.6  2006/04/26 16:44:53  cocoa
// -- myMul has now a single implementation in SparsePolyRing
// -- myMul and mul in RingDistrMPoly* and DistrMPoly* have been disabled
//
// Revision 1.5  2006/04/21 14:58:04  cocoa
// Removed myWDeg member function: it is no longer needed.
//
// Revision 1.4  2006/04/11 14:16:29  cocoa
// -- reorganization of fns between reduce,SparsePolyRing,GPoly
// -- added optional "len" argument to myAssignReset in ReductionCog
//
// Revision 1.3  2006/02/13 13:19:01  cocoa
// -- fixed: "const PPMonoidElem&" --> "ConstRefPPMonoidElem"
//
// Revision 1.2  2006/01/19 16:34:42  cocoa
// -- added NF, myReduceMod functions (not yet tested)
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.3  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.2  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.1  2005/07/01 16:08:15  cocoa
// Friday check-in.  Major change to structure under PolyRing:
// now SparsePolyRing and DUPolyRing are separated (in preparation
// for implementing iterators).
//
// A number of other relatively minor changes had to be chased through
// (e.g. IndetPower).
//
