//   Copyright (c)  2006,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// These are from directory TmpFactorDir/
#include "mpz_alias.h"
#include "primes.h"
#include "FF.h"
#include "WGD.h"

// These are from directory TmpFactorDir/multivariate/
#include "DMPZgcd.h"
#include "DMPZfactor.h"
#include "DMPZfactor_modp.h"

/***************************************************************************/
/* CoCoALib includes */

#include "CoCoA/TmpFactor.H"

#include "CoCoA/CanonicalHom.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/NumTheory.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/SparsePolyRing.H"

//#include <iostream>
#include <cstdlib>
using std::malloc;// someone actually uses malloc!!!!
#include <vector>
using std::vector;

/***************************************************************************/

namespace CoCoA
{


// Exception safe???  You're kidding, right?
  DMPZ ConvertToDMPZ(ConstRefRingElem f)
  {
    CoCoA_ASSERT(IsSparsePolyRing(owner(f)));
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    ring R = CoeffRing(P);
    CoCoA_ASSERT(IsFractionField(R));
    CoCoA_ASSERT(IsGCDDomain(BaseRing(AsFractionField(R))));

    const size_t nvars = NumIndets(P);
    DMPZ g = NULL;
    std::vector<long> expv(nvars);
    for (SparsePolyIter it = BeginIter(f); !IsEnded(it); ++it)
    {
      ZZ c;
      if (!IsInteger(c, coeff(it))) CoCoA_ERROR("coeff not in Z","factorize");
      exponents(expv, PP(it));
      int *exps = (int*)malloc(nvars*sizeof(int)); // block absorbed by g
      std::copy(expv.begin(), expv.end(), exps);
      g = DMPZprepend(mpzref(c), exps, g);
    }

    return DMPZreverse(g); /* restore original order of the terms */
  }


  void ConvertDMPZToRingElem(RefRingElem f, DMPZ poly)
  {
    CoCoA_ASSERT(IsSparsePolyRing(owner(f)));
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    PPMonoid M = PPM(P);
    PPMonoidElem t(M);
    f = 0;
    const size_t nvars = NumIndets(P);
    std::vector<long> exps(nvars);
    for (DMPZ iter = poly; iter; iter = iter->next)
    {
      if (mpz_sgn(iter->coeff) == 0) continue;
      ZZ c(CopyFromMPZ, iter->coeff);
      std::copy(iter->exps, iter->exps+nvars, exps.begin());
      f += monomial(P, c, PPMonoidElem(M, exps));
    }
  }

  factorization<RingElem> ConvertToFactorList(DMPZfactors facpows, ring Rx)
  {
    RingElem content = RingElem(Rx, ZZ(CopyFromMPZ, facpows->content));
    RingElem factor(Rx);
    vector<RingElem> facs;
    vector<long> exps;

    DMPZlist iter;
    int i;
    for (i=1, iter=facpows->list; iter; ++i, iter = iter->next)
    {
      ConvertDMPZToRingElem(factor, iter->poly);
      facs.push_back(factor);
      exps.push_back(iter->deg);
    }
    DMPZfactors_dtor(facpows);  // NASTY - we destroy one of our args!!
    return factorization<RingElem>(facs, exps, content);
  }


  factorization<RingElem> factor(ConstRefRingElem f)
  {
    ring R = owner(f);
    if (IsField(R) || !IsGCDDomain(R))
      CoCoA_ERROR("factorization not supported", "factor(RingElem)");
    if (IsZ(R))
    {
      // Factorization is actually over ZZ, so use factor(ZZ) to do the work!
      ZZ N;
      IsInteger(N, f); // guaranteed to succeed
      const factorization<ZZ> Nfacs = factor(N);
      const vector<ZZ>& prime = Nfacs.myFactors;
      vector<RingElem> facs;
      for (size_t i=0; i < prime.size(); ++i)
        facs.push_back(RingElem(R, prime[i]));
      return factorization<RingElem>(facs, Nfacs.myExponents, RingElem(R, Nfacs.myRemainingFactor));
    }
    if (!IsPolyRing(R))
      CoCoA_ERROR(ERR::NotPolyRing, "factor(RingElem)");
/////    if (!IsSparsePolyRing(R))
/////      CoCoA_ERROR(ERR::NYI, "factor(RingElem)  NYI over DUP");
    SparsePolyRing P = AsSparsePolyRing(owner(f));
    if (IsQ(CoeffRing(P)))
    {
      // case P = QQ[x,y,z,...]
      FractionField Q = AsFractionField(CoeffRing(P));
      RingHom coeff = CoeffEmbeddingHom(P)(EmbeddingHom(Q));
      RingElem D = CommonDenom(f);
      NVARS = NumIndets(P); // NVARS is a global!!!
      DMPZ g = ConvertToDMPZ(f*coeff(D));
      factorization<RingElem> ans = ConvertToFactorList(DMPZfactor(g), P); // leaks the factorlist
      DMPZdtor(g);
      if (!IsOne(D)) ans.myRemainingFactor /= CanonicalHom(RingZ(), P)(D);
      return ans;
    }
    if (IsZ(CoeffRing(P)))
    {
      // We do not attempt to factorize the content in ZZ[x,y,z] as it could easily be very costly.
      PolyRing Qx = NewPolyRing(RingQ(), NumIndets(P));
      RingHom phi = PolyRingHom(P, Qx, ZEmbeddingHom(RingQ()), indets(Qx)); // maps ZZ[x,y,z] --> QQ[x,y,z]
      RingElem FinQx = phi(f);
      factorization<RingElem> FactorizationInQx = factor(FinQx);
      const vector<RingElem>& FacsInQx = FactorizationInQx.myFactors;
      vector<RingElem> FacsInP;
      const RingElem ContentInP = CoeffEmbeddingHom(P)(num(FactorizationInQx.myRemainingFactor));
      RingHom psi = PolyRingHom(Qx, P, QEmbeddingHom(RingZ()), indets(P)); // maps QQ[x,y,z] --> ZZ[x,y,z]
      for (size_t i=0; i < FacsInQx.size(); ++i)
      {
        FacsInP.push_back(psi(FacsInQx[i]));
      }
      return factorization<RingElem>(FacsInP, FactorizationInQx.myExponents, ContentInP);
    }
    if (IsQuotientRing(CoeffRing(P)) && IsZ(BaseRing(AsQuotientRing(CoeffRing(P)))))
    {
      // Case ZZ/(p)[x,y,z,...]
      CoCoA_ERROR(ERR::NYI, "Factorization in ZZ/(p)[x,y,...]");
    }
    CoCoA_ERROR(ERR::NYI, "Factorization not in QQ[x,y,...]  or in ZZ/(p)[x,y,...]");
    return factorization<RingElem>(vector<RingElem>(),vector<long>(),RingElem(P)); // NEVER EXECUTED, just to keep compiler quiet
  }


  // temporary: by Anna 2008-01-25
  bool IsIrredPoly(ConstRefRingElem f)
  { 
    if (!IsPolyRing(owner(f)))
      CoCoA_ERROR(ERR::NotPolyRing, "IsIrredPoly");
 // ??? warning: factorize syntax might change: now only for multivariate
    if (IsConstant(f)) return IsIrred(LC(f));
    // ANNA: should do it for Z as well (taking care of constant factors)
    PolyRing P(AsPolyRing(owner(f)));
    if (!IsQ(CoeffRing(P)))
      CoCoA_ERROR(ERR::NYI, "IsIrreducible, !IsQ(CoeffRing)");
    if ( IsSparsePolyRing(P) )
    {
      factorization<RingElem> facs = factor(f);
      return (IsInvertible(facs.myRemainingFactor) && facs.myFactors.size() == 1 && facs.myExponents[0] == 1);
    }
    else
    {
      PolyRing SPR = NewPolyRing(CoeffRing(AsPolyRing(P)), 1);
      RingHom phi(PolyAlgebraHom(P, SPR, indets(SPR)));
      factorization<RingElem> facs = factor(phi(f));
      return (IsInvertible(facs.myRemainingFactor) && facs.myFactors.size() == 1 && facs.myExponents[0] == 1);
    }

  }


} // end of namespace CoCoA

// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/TmpFactor.C,v 1.10 2009/12/03 17:41:20 abbott Exp $
// $Log: TmpFactor.C,v $
// Revision 1.10  2009/12/03 17:41:20  abbott
// Added include of cstdlib for malloc.
//
// Revision 1.9  2009/10/26 15:43:25  bigatti
// -- added CopyFromMPZ in ZZ ctor
//
// Revision 1.8  2009/07/24 14:22:40  abbott
// Changed interface to factorizer: new function name (now "factor") with new signature.
// Did some cleaning too.
//
// Revision 1.7  2009/07/02 16:34:32  abbott
// Minor change to keep the compiler quiet.
//
// Revision 1.6  2009/05/14 09:20:03  abbott
// Changed a comment.
//
// Revision 1.5  2008/10/07 15:46:10  abbott
// Added missing log/history lines at the ends of a few files.
//
//
