//   Copyright (c)  2005  Massimo Caboara

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "CoCoA/FractionField.H"
#include "CoCoA/FreeModule.H"
#include "CoCoA/RingDistrMPolyInlFpPP.H"
#include "CoCoA/RingDistrMPolyInlPP.H"
#include "CoCoA/TmpGOperations.H"
// #include "CoCoA/TmpGReductor.H"  // included in TmpGOperations.H 
#include "CoCoA/io.H" // use io.H just for debugging and statistics
#include "CoCoA/symbol.H"


#include <algorithm>
using std::stable_sort;
#include <list>
using std::list;
#include <functional>
using std::less;
using std::binary_function;
using std::mem_fun_ref; // for calling GPair::complete on GPairList
#include <utility>
using std::make_pair;
#include <iostream>
using std::ostream;
using std::endl;
using std::flush;

namespace CoCoA
{

// GBasis

  void ComputeGBasis(VectorList& theGB, const VectorList& theVL)
  {
    if (theVL.empty())
    {
      theGB.clear();
      return;
    }
    const FreeModule& FM=AsFreeModule(owner(theVL));
    const SparsePolyRing NewP(MakeNewPRingFromModule(FM));
    const SparsePolyRing OldP(AsSparsePolyRing(BaseRing(FM)));
    if (!IsField(CoeffRing(OldP)))
      CoCoA_ERROR("coefficients are not in a field", "ComputeGBasis");
   bool IsSatAlg=false;
   const GRingInfo GRI(NewP,OldP,FM,FM,IsHomog(theVL),
                       IsSatAlg,NewDivMaskEvenPowers());
   GPolyList EmbeddedPolys=EmbedVectorList(theVL,GRI);
   GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
   GBR.myDoGBasis();// homog input standard alg interred
   PolyList OutputPolyList;
   GBR.myGBasis(OutputPolyList);
   theGB.clear();// just to remember to clean this up
   theGB=DeEmbedPolyList(OutputPolyList,GRI);
 }//ComputeGBasis


  void ComputeGBasis(PolyList& theGB, const PolyList& thePL, const int StatLevel)
  {
    if (thePL.empty())
    {
      theGB.clear();
      return;
    }
    SparsePolyRing SPR(AsSparsePolyRing(owner(thePL)));
    // MAX: Adding to ComputeGBasis for modules should suffice.
    // The new ring should be created by MakeNewPRingFromModule
    // and the embeddings:  WithoutDenominators by EmbedVectorList
    //                   :  WithDenominator1Hom by DeEmbedPolyList
    if (!IsField(CoeffRing(SPR)))
      CoCoA_ERROR("coefficients are not in a field", "ComputeGBasis");
    bool IsSatAlg=false;
    if (IsFractionField(CoeffRing(SPR)) && IsGCDDomain(BaseRing(AsFractionField(CoeffRing(SPR)))))
    {
      ring R = BaseRing(AsFractionField(CoeffRing(SPR)));
      SparsePolyRing Rx = NewPolyRing_DMPI(R, symbols(PPM(SPR)), ordering(PPM(SPR)));
      GRingInfo GRI(Rx, IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
      GRI.mySetCoeffRingType(CoeffEncoding::FrFldOfGCDDomain);
      //GReductor GBR(GRI, WithoutDenominators(thePL, Rx), GReductor::ourDefaultStatLevel);
      GReductor GBR(GRI, WithoutDenominators(thePL, Rx), StatLevel);
      GBR.myDoGBasis();// homog input standard alg interred
      PolyList TmpGB;
      GBR.myGBasis(TmpGB);
      theGB.clear();//just to remember to clean this up
      theGB=WithDenominator1Hom(TmpGB, SPR);
    }
    else
    {
      GRingInfo GRI(SPR,IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
      GReductor GBR(GRI, thePL, StatLevel,
                    Reductors::DontUseBorel, GReductor::DontUseDynamicAlg,
                    GReductor::UseCoprimeCrit, GReductor::UseGMCrit,
                    GReductor::UseBackCrit, GReductor::UseDivCrit);
      GBR.myDoGBasis();// homog input standard alg interred
      GBR.myGBasis(theGB);
    }          
  }//ComputeGBasis  
  
  
  void ComputeSATMixGBasis(PolyList& theGB,const PolyList& thePL,const int StatLevel)
  {      
    if (thePL.empty())
    {
      theGB.clear();
      return;
    }
    bool IsSatAlg=false;
    SparsePolyRing SPR(AsSparsePolyRing(owner(thePL)));
    // MAX: Adding to ComputeGBasis for modules should suffice.
    // The new ring should be created by MakeNewPRingFromModule
    // and the embeddings:  WithoutDenominators by EmbedVectorList
    //                   :  WithDenominator1Hom by DeEmbedPolyList
    if (!IsField(CoeffRing(SPR)))
      CoCoA_ERROR("coefficients are not in a field", "ComputeGBasis");
    if (IsFractionField(CoeffRing(SPR)) && IsGCDDomain(BaseRing(AsFractionField(CoeffRing(SPR)))))
    {
      ring R = BaseRing(AsFractionField(CoeffRing(SPR)));
      SparsePolyRing Rx = NewPolyRing_DMPI(R, symbols(PPM(SPR)), ordering(PPM(SPR)));
      GRingInfo GRI(Rx, IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
      GRI.mySetCoeffRingType(CoeffEncoding::FrFldOfGCDDomain);
      //GReductor GBR(GRI, WithoutDenominators(thePL, Rx), GReductor::ourDefaultStatLevel);
      GReductor GBR(GRI, WithoutDenominators(thePL, Rx), StatLevel);
      GBR.myDoGBasis();// homog input standard alg interred
      PolyList TmpGB;
      GBR.myGBasis(TmpGB);
      theGB.clear();//just to remember to clean this up
      theGB=WithDenominator1Hom(TmpGB, SPR);
    }
    else
    {
      GRingInfo GRI(SPR,IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
      GReductor GBR(GRI, thePL, StatLevel,
                    Reductors::DontUseBorel, GReductor::DontUseDynamicAlg,
                    GReductor::UseCoprimeCrit, GReductor::UseGMCrit,
                    GReductor::UseBackCrit, GReductor::UseDivCrit);
      GBR.myDoGBasis();// homog input standard alg interred
      GBR.myGBasis(theGB);
    }
  }//ComputeSATMixGBasis


  void ComputeSATGBasis(PolyList& theGB,const PolyList& thePL,const int StatLevel)
  {
    if (thePL.empty())
    {
      theGB.clear();
      return;
    }
    bool IsSatAlg=true;
    SparsePolyRing SPR(AsSparsePolyRing(owner(thePL)));
    // MAX: Adding to ComputeGBasis for modules should suffice.
    // The new ring should be created by MakeNewPRingFromModule
    // and the embeddings:  WithoutDenominators by EmbedVectorList
    //                   :  WithDenominator1Hom by DeEmbedPolyList
    if (!IsField(CoeffRing(SPR)))
      CoCoA_ERROR("coefficients are not in a field", "ComputeSATGBasis");
    if (IsFractionField(CoeffRing(SPR)) && IsGCDDomain(BaseRing(AsFractionField(CoeffRing(SPR)))))
    {
      ring R = BaseRing(AsFractionField(CoeffRing(SPR)));
      SparsePolyRing Rx = NewPolyRing_DMPI(R, symbols(PPM(SPR)), ordering(PPM(SPR)));
      GRingInfo GRI(Rx, IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
      GRI.mySetCoeffRingType(CoeffEncoding::FrFldOfGCDDomain);
      //GReductor GBR(GRI, WithoutDenominators(thePL, Rx), GReductor::ourDefaultStatLevel);
      GReductor GBR(GRI, WithoutDenominators(thePL, Rx), StatLevel,
                    Reductors::DontUseBorel, GReductor::DontUseDynamicAlg,
                    GReductor::UseCoprimeCrit, GReductor::UseGMCrit,
                    GReductor::UseBackCrit, GReductor::UseDivCrit,
		    GReductor::SaturatingAlg);
      GBR.myDoSATGBasis();// homog input standard alg interred
      PolyList TmpGB;
      GBR.myGBasis(TmpGB);
      theGB.clear();//just to remember to clean this up
      theGB=WithDenominator1Hom(TmpGB, SPR);
    }
    else
    {
      GRingInfo GRI(SPR,IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
      GReductor GBR(GRI, thePL, StatLevel,
                    Reductors::DontUseBorel, GReductor::DontUseDynamicAlg,
                    GReductor::UseCoprimeCrit, GReductor::UseGMCrit,
                    GReductor::UseBackCrit, GReductor::UseDivCrit,
		    GReductor::SaturatingAlg);
      GBR.myDoSATGBasis();// homog input standard alg interred
      GBR.myGBasis(theGB);
    }
  }//ComputeSATGBasis

/*
void ComputeGBasisFrameWork(PolyList& theGB, const PolyList& theInputPolyList)

 {
   theGB.clear();
   if (theInputPolyList.empty())
     return;
   const SparsePolyRing SPR(AsSparsePolyRing(owner(theInputPolyList)));
   GRingInfo GRI(SPR,IsHomog(theInputPolyList),NewDivMaskEvenPowers());
   GReductor GBR(GRI, theInputPolyList,GReductor::ourDefaultStatLevel);


   if (false)
   {
     GBR.myDoGBasis();// homog input standard alg interred
     GBR.myGBasis(theGB);
   }


   if (false)
   {
     GBR.myPrepareGBasis();
     while (GBR.myPairsLen()!=0)
     {
       GBR.myDoGBasis(1);
     }
     GBR.myGBasis(theGB);
   }

  if (false)
  {
   //non null Spoly by non null Spoly
   GBR.myPrepareGBasis();
   //GPoly SP(GRI);
   while (GBR.myPairsLen()!=0)
   {
     GBR.myReduceUntilNonZeroRedSP();
     //SP=GBR.GetSPoly();
     //.......Things.......
     //PolyList PL;
     //GBR.GetCandidateGBasis(PL);
     //GBR.SetSPoly(SP);
     if (GBR.myPairsLen()!=0)
     {
       myTrueReductors.interreduce(mySPoly);
       myUpdateBasisAndPairs();
       if (myCurrentPairDeg!=myOldDeg&&myTrueReductors.IhaveBorelReductors())
         myTrueReductors.myBorelReductorsUpdateInNextDegree(myCurrentPairDeg);
     }
    }
   GBR.myGBasis(theGB);
  }

 }//ComputeGBasisFrameWork
*/



  void ComputeElim(PolyList& theElimResult, const PolyList& thePL, ConstRefPPMonoidElem inds)
  {
    if (thePL.empty())
    {
      theElimResult.clear();
      return;
    }
    PolyList ElimResult;
    PolyList NewPL;
    const SparsePolyRing OldSPR=AsSparsePolyRing(owner(thePL));
    std::vector<long> IndexList=PPMonoidElem2IndexList(inds);
    SparsePolyRing ElimSPR=MakeElimRingFromOld(OldSPR,IndexList,IsHomog(thePL));
    std::vector<RingElem> NewImages=indets(ElimSPR);
    std::vector<RingElem> OldImages=indets(OldSPR);
    RingHom Old2New=PolyRingHom(OldSPR,ElimSPR,CoeffEmbeddingHom(ElimSPR),NewImages);
    RingHom New2Old=PolyRingHom(ElimSPR,OldSPR,CoeffEmbeddingHom(OldSPR),OldImages);
    for (PolyList::const_iterator it=thePL.begin();it!=thePL.end();++it)
      NewPL.push_back(Old2New(*it));
    PPMonoidElem NewInds(LPP(Old2New(monomial(OldSPR,1,inds))));
    PolyList GB;
    ComputeGBasis(GB,NewPL,GReductor::ourDefaultStatLevel);
    for (PolyList::iterator it=GB.begin();it!=GB.end();++it)
      if (IsCoprime(LPP(*it),NewInds))
        ElimResult.push_back(New2Old(*it));
    swap(theElimResult,ElimResult);
    return;
  }//ComputeElim

  void ComputeElim(VectorList& /*theElimResult*/,
                   const VectorList& /*theVL*/,
                   ConstRefPPMonoidElem /*inds*/)
  {
    // Remember to check if ordering is TOPOS ordering here
    // if not, use hom
    CoCoA_ERROR(ERR::NYI, "ComputeElim(const VectorList& theVL,ConstRefPPMonoidElem)");
    return;
  }//ComputeElim


  void ComputeSyz(VectorList& theSyzResult, const VectorList& theVL)
  {
    if (!IsHomog(theVL))
      CoCoA_ERROR("Not Yet Tested for non-homogeneous input", "ComputeSyz(const VectorList&)");
    if (theVL.empty())
    {
      theSyzResult.clear();
      return;
    }
    bool IsSatAlg=false;
    VectorList SyzResult;
    ModOrdTypeForcing MOType;
    if (IsHomog(theVL))
      MOType=WDegPosTO;
    else
      MOType=PosWDegTO;
    const FreeModule FM=AsFreeModule(owner(theVL));
    const SparsePolyRing NewP(MakeNewPRingFromModule(FM,MOType));
    const SparsePolyRing OldP(AsSparsePolyRing(BaseRing(FM)));
    const FreeModule SyzFM=MakeNewFreeModuleForSyz(theVL);
    // Note: the GRI should build itself SyzFM and NewP from the data and deduce FM and OldP.
    //       All the embedding/deembedding functions should be memebers of GRI.
    GRingInfo GRI(NewP,OldP,FM,SyzFM,IsHomog(theVL),IsSatAlg,NewDivMaskEvenPowers());
    GPolyList EmbeddedPolys=SyzEmbedVectorList(theVL,GRI);
    GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
    GBR.myDoGBasis();
    PolyList OutputPolyList;
    GBR.myGBasis(OutputPolyList);
    SyzResult=DeEmbedPolyList(OutputPolyList,GRI,NumCompts(FM));
    swap(theSyzResult,SyzResult);
    return;
  }//ComputeSyz


  void ComputeSyz(VectorList& theSyzResult, const PolyList& thePL)
  {
    //if (!IsHomog(thePL))
    //  CoCoA_ERROR("Not Yet Tested for non-homogeneous input", "ComputeSyz(const VectorList&)");
    if (thePL.empty())
    {
      theSyzResult.clear();
      return;
    }
    bool IsSatAlg=false;
    VectorList SyzResult;
    ModOrdTypeForcing MOType;
    if (IsHomog(thePL))
      MOType=WDegPosTO;
    else
      MOType=PosWDegTO;
    const SparsePolyRing OldP(AsSparsePolyRing(owner(thePL)));
    const SparsePolyRing NewP(MakeNewPRingForSimpleEmbedding(OldP,MOType));
    const FreeModule SyzFM=MakeNewFreeModuleForSyz(thePL);
    GRingInfo GRI(NewP,OldP,SyzFM,IsHomog(thePL),IsSatAlg,NewDivMaskEvenPowers());
    GPolyList EmbeddedPolys=SyzEmbedPolyList(thePL,GRI);
    GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
    GBR.myDoGBasis();
    PolyList OutputPolyList;
    GBR.myGBasis(OutputPolyList);
    SyzResult=DeEmbedPolyList(OutputPolyList,GRI,1);
    swap(theSyzResult,SyzResult);
    return;
  }//ComputeSyz


  void ComputeIntersection(VectorList& theIntersectionResult,
                           const VectorList& theVL1,
                           const VectorList& theVL2)
  {
    bool IsSatAlg=false;
    VectorList IntersectionResult;
    const FreeModule& FM=AsFreeModule(owner(theVL1));
    //const SparsePolyRing NewP(MakeNewPRingFromModule(FM,WDegPosTO));
    const SparsePolyRing NewP(MakeNewPRingFromModulePosFirst(FM,IsHomog(theVL1)&&IsHomog(theVL2)));
    const SparsePolyRing OldP(AsSparsePolyRing(BaseRing(FM)));
    GRingInfo GRI(NewP,OldP,FM,FM,IsHomog(theVL1)&&IsHomog(theVL2),
                  IsSatAlg,NewDivMaskEvenPowers());
    GPolyList EmbeddedPolys=IntEmbedVectorLists(theVL1,theVL2,GRI);
    GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
    GBR.myDoGBasis();// homog input standard alg interred
    PolyList OutputPolyList;
    GBR.myGBasis(OutputPolyList);
    IntersectionResult=DeEmbedPolyList(OutputPolyList,GRI,NumCompts(FM));
    swap(IntersectionResult,theIntersectionResult);
    return;
  }//ComputeIntersection

  void ComputeIntersection(PolyList& theIntersectionResult,
                           const PolyList& thePL1,
                           const PolyList& thePL2)
  {
    bool IsSatAlg=false;
    PolyList IntersectionResult;
    const SparsePolyRing OldP(AsSparsePolyRing(owner(thePL1)));
    const SparsePolyRing NewP(MakeNewPRingForSimpleEmbeddingPosFirst(OldP,IsHomog(thePL1)&&IsHomog(thePL2)));
    GRingInfo GRI(NewP,OldP,IsHomog(thePL1)&&IsHomog(thePL2),IsSatAlg,
                  NewDivMaskEvenPowers());
    GPolyList EmbeddedPolys=IntEmbedPolyLists(thePL1,thePL2,GRI);
    GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
    GBR.myDoGBasis();// homog input standard alg interred
    PolyList OutputPolyList;
    GBR.myGBasis(OutputPolyList);
    IntersectionResult=DeEmbedPolyListToPL(OutputPolyList,GRI,1);
    swap(theIntersectionResult,IntersectionResult);
    return;
  }//ComputeIntersection


// Colon by a single vector
  void ComputeColon(PolyList& theColonResult, const VectorList& theVL1, const VectorList& theVL2)
  {
    bool IsSatAlg=false;
    for (VectorList::const_iterator it=theVL1.begin();it!=theVL1.end();++it)
      if (!IsHomog(*it))
        CoCoA_ERROR(ERR::NYI, "ComputeColon(const VectorList&, const VectorList&) Not Yet Implemented for non-homogeneous input");
    for (VectorList::const_iterator it=theVL2.begin();it!=theVL2.end();++it)
      if (!IsHomog(*it))
        CoCoA_ERROR(ERR::NYI, "ComputeColon(const VectorList&, const VectorList&) Not Yet Implemented for non-homogeneous input");
    PolyList ColonResult;
    const FreeModule& FM=AsFreeModule(owner(theVL1));
    const SparsePolyRing NewP(MakeNewPRingFromModule(FM,WDegPosTO));
    const SparsePolyRing OldP(AsSparsePolyRing(BaseRing(FM)));
    GRingInfo GRI(NewP,OldP,FM,FM,IsHomog(theVL1)&&IsHomog(theVL2),
                  IsSatAlg,NewDivMaskEvenPowers());
    GPolyList EmbeddedPolys=ColonEmbedVectorLists(theVL1,theVL2,GRI);
    GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
    GBR.myDoGBasis();// homog input standard alg interred
    PolyList OutputPolyList;
    GBR.myGBasis(OutputPolyList);
    ColonResult=DeEmbedPolyListToPL(OutputPolyList,GRI,NumCompts(FM));
    swap(ColonResult,theColonResult);
    return;
  }//ComputeColon

// Colon
  void ComputeCColon(PolyList& theCColonResult,
                     const VectorList& theVL1,
                     const VectorList& theVL2)
  {
    if (theVL1.empty() && theVL2.empty())
      CoCoA_ERROR("One argument must be non empty","ComputeCColon");
    PolyList CColonResult;
    if (theVL2.empty())
    {
      CColonResult.push_back(one(BaseRing(owner(theVL1))));
      swap(CColonResult,theCColonResult);
      return;
    }
    if (theVL1.empty())
    {
      swap(CColonResult,theCColonResult);
      return;
    }
    PolyList tmp;
    VectorList::const_iterator it=theVL2.begin();
    ComputeColon(CColonResult,theVL1,MakeVectorList(*it));
    ++it;
    for (;it!=theVL2.end();++it)
    {
      ComputeColon(tmp,theVL1,MakeVectorList(*it));
      ComputeIntersection(CColonResult,CColonResult,tmp);
    }
    swap(CColonResult,theCColonResult);
    return;
  }//ComputeCColon

// Colon by a single poly
  void ComputeColon(PolyList& theColonResult,
                    const PolyList& thePL1,
                    const PolyList& thePL2)
  {
    bool IsSatAlg=false;
    PolyList ColonResult;
    const SparsePolyRing OldP(AsSparsePolyRing(owner(thePL1)));
    const SparsePolyRing NewP(MakeNewPRingForSimpleEmbeddingPosFirst(OldP,IsHomog(thePL1)&&IsHomog(thePL2)));
    GRingInfo GRI(NewP,OldP,IsHomog(thePL1)&&IsHomog(thePL2),
                  IsSatAlg,NewDivMaskEvenPowers());
    GPolyList EmbeddedPolys=ColonEmbedPolyLists(thePL1,thePL2,GRI);
    GReductor GBR(GRI, EmbeddedPolys, GReductor::ourDefaultStatLevel);
    GBR.myDoGBasis();// homog input standard alg interred
    PolyList OutputPolyList;
    GBR.myGBasis(OutputPolyList);
    ColonResult=DeEmbedPolyListToPL(OutputPolyList,GRI,1);
    swap(theColonResult,ColonResult);
    return;
  }//ComputeColon

// Colon
  void ComputeCColon(PolyList& theCColonResult,
                     const PolyList& thePL1,
                     const PolyList& thePL2)
  {
    if (thePL1.empty() && thePL2.empty())
      CoCoA_ERROR("One argument must be non empty","ComputeCColon");
    PolyList CColonResult;
    if (thePL2.empty())
    {
      CColonResult.push_back(one(owner(thePL1)));
      swap(theCColonResult,CColonResult);
      return;
    }
    if (thePL1.empty())
    {
      theCColonResult.clear();
      return;
    }
    PolyList tmp;
    PolyList::const_iterator it=thePL2.begin();
    ComputeColon(CColonResult,thePL1,MakePolyList(*it));
    ++it;
    for (;it!=thePL2.end();++it)
    {
      ComputeColon(tmp,thePL1,MakePolyList(*it));
      ComputeIntersection(CColonResult,CColonResult,tmp);
    }
    swap(theCColonResult,CColonResult);
    return;
  }//ComputeCColon

  void ComputeColon(VectorList& /*theColonResult*/,
                    const VectorList& /*theVL*/,
                    const PolyList& /*thePL*/)
  {
    CoCoA_ERROR(ERR::NYI,"ComputeColon(VectorList, PolyList)");
    return;
  }//ComputeColon


//----------------- These procedures are for operator==
// for PolyList, VectorList used for Saturation

// PolyList::const_iterators are ordered according to
// the LPP or their polys
  bool operator<(const PolyList::const_iterator& it,
                 const PolyList::const_iterator& it1)
  {
    int tmp;
    const SparsePolyRing P(AsSparsePolyRing(owner(*it)));
    tmp=P->myCmpLPP(raw(*it),raw(*it1));
    if (tmp==-1) return true;
    return false;
  }// PolyList::const_iterator operator<

// Two vectors of ptr are equal if the objects pointed at have equal LPP
  bool Equal(const std::vector<PolyList::const_iterator>& theV1,
             const std::vector<PolyList::const_iterator>& theV2)
  {
    if (theV1.size()!=theV2.size())
      return false;
    for (unsigned int i=0;i!=theV1.size();++i)
      if (LPP(*(theV1[i]))!=LPP(*(theV2[i])))
        return false;
    return true;
  }// PolyList::const_iterator operator<

// Two vectors of ptr are equal if the objects pointed at have equal LPP
  bool Equal(const std::vector<VectorList::const_iterator>& theV1,
             const std::vector<VectorList::const_iterator>& theV2)
  {
    if (theV1.size()!=theV2.size())
      return false;
    for (unsigned int i=0;i!=theV1.size();++i)
      //if (*(theV1[i])!=*(theV2[i]))
      // when LT exists if (LT(*(theV1[i]))!=LT(*(theV2[i])))
      return false;
    return true;
  }// VectorList::const_iterator operator<


// Two polys are equal if they have equal LPP
// If you don't like it, write your own order function for PolyList::const_iterator
// This operation is useful when you have I\subset J and you
// want to check I=J
  bool operator==(const PolyList& thePL1, const PolyList& thePL2)
  {
    std::vector<PolyList::const_iterator> V1,V2;
    for (PolyList::const_iterator it=thePL1.begin();it!=thePL1.end();++it)
      V1.push_back(it);
    for (PolyList::const_iterator it=thePL2.begin();it!=thePL2.end();++it)
      V2.push_back(it);
    stable_sort(V1.begin(),V1.end());
    stable_sort(V2.begin(),V2.end());
    return (Equal(V1,V2));
  }// PolyList operator==

// Two polys are equal if they have equal LPP
// If you don't like it, write your own order function for PolyList::const_iterator
// This operation is useful when you have M\subset N and you
// want to check M=N
  bool operator==(const VectorList& theVL1, const VectorList& theVL2)
  {
    std::vector<VectorList::const_iterator> V1,V2;
    for (VectorList::const_iterator it=theVL1.begin();it!=theVL1.end();++it)
      V1.push_back(it);
    for (VectorList::const_iterator it=theVL2.begin();it!=theVL2.end();++it)
      V2.push_back(it);
    stable_sort(V1.begin(),V1.end());
    stable_sort(V2.begin(),V2.end());
    return Equal(V1,V2);
  }//VectorList operator==

//---------------------------------------------------------

  void ComputeSaturation(PolyList& theSaturationResult,
                         const PolyList& thePL1,
                         const PolyList& thePL2)
  {
    PolyList TmpPL1;
    ComputeColon(TmpPL1,thePL1,thePL2);
    PolyList TmpPL2;
    ComputeColon(TmpPL2,TmpPL1,thePL2);
    while (!(TmpPL1==TmpPL2))
    {
      swap(TmpPL1,TmpPL2);
      ComputeColon(TmpPL2,TmpPL1,thePL2);
    }
    swap(theSaturationResult,TmpPL2);
    return;
  }//ComputeSaturation

  void ComputeSSaturation(PolyList& theSSaturationResult,
                          const PolyList& thePL1,
                          const PolyList& thePL2)
  {
    if (thePL1.empty() && thePL2.empty())
      CoCoA_ERROR("One argument must be non empty","ComputeSSaturation");
    if (thePL2.empty())
    {
      theSSaturationResult.clear();// this or sawp? this look better
      theSSaturationResult.push_back(one(owner(thePL1)));
      return;
    }
    if (thePL1.empty())
    {
      theSSaturationResult.clear();
      return;
    }

    PolyList TmpPL1;
    ComputeCColon(TmpPL1,thePL1,thePL2);
    PolyList TmpPL2;
    ComputeCColon(TmpPL2,TmpPL1,thePL2);
    while (!(TmpPL1==TmpPL2))
    {
      swap(TmpPL1,TmpPL2);
      ComputeCColon(TmpPL2,TmpPL1,thePL2);
    }
    swap(theSSaturationResult,TmpPL2);
    return;
  }//ComputeSSaturation


  void ComputeSaturation(VectorList& /*theSaturation*/,
                         const VectorList& /*theVL*/,
                         const PolyList& /*thePL*/)
  {
    CoCoA_ERROR(ERR::NYI,"ComputeSaturation(VectorList, PolyList)");
    return;
  }//ComputeSaturation


  void ComputeSSaturation(VectorList& /*theSSaturation*/,
                          const VectorList& /*theVL*/,
                          const PolyList& /*thePL*/)
  {
    CoCoA_ERROR(ERR::NYI,"ComputeSSaturation(VectorList, PolyList)");
    return;
  }//ComputeSSaturation


  void ComputeHomogenization(VectorList& /*theHomogResult*/,
                             const VectorList& /*theVL*/,
                             const PolyList& /*thePL*/)
  {
    CoCoA_ERROR(ERR::NYI,"ComputeHomogenization(VectorList, PolyList)");
    return;
  }//ComputeHomogenization

  void ComputeHomogenization(PolyList& theHomogResult,
                             const PolyList& thePL1,
                             const PolyList& theIndets)
  {
    if (thePL1.empty())
    {
      theHomogResult.clear();
      return;
    }
    PolyList HomogResult;
    const SparsePolyRing SPR=owner(thePL1);
    PolyList VarProduct;
    VarProduct.push_back(one(SPR));
    std::vector<RingElem> IndetsVector;
    for(PolyList::const_iterator it=theIndets.begin();it!=theIndets.end();++it)
    {
      IndetsVector.push_back(*it);
      VarProduct.front()*=(*it);
    }
    RingElem tmp(SPR);
    for (PolyList::const_iterator it=thePL1.begin();it!=thePL1.end();++it)
    {
      homogenized(tmp,*it,IndetsVector);
      HomogResult.push_back(one(SPR));
      swap(tmp,HomogResult.back());
    }
    ComputeSaturation(HomogResult,HomogResult,VarProduct);
    swap(theHomogResult,HomogResult);
    return;
  }//ComputeHomogenization

// WARN: it supposes ComputeSaturation returns a GB
  bool RadicalMembership(const PolyList& PL,ConstRefRingElem the_f)
  {
    PolyList PL_f;
    PL_f.push_back(the_f);
    PolyList PL1;
    ComputeSaturation(PL1,PL,PL_f);
    if (PL.size()!=1)
      return false;
    monic(PL1);
    if (IsOne(PL1.front()))
      return true;
    else
      return false;
  }//RadicalMembership

  void ComputeLT(VectorList& /*theLTResult*/,const VectorList& /*theVL*/)
  {
    // Waiting for LT of a ModuleElem
    CoCoA_ERROR(ERR::NYI,"ComputeLT(VectorList)");
    return;
  }//ComputeLT


  void ComputeLT(PolyList& theLTResult, const PolyList& thePL)
  {
    PolyList GB;
    ComputeGBasis(GB,thePL,GReductor::ourDefaultStatLevel);
    if (GB.empty())
    {
      swap(theLTResult,GB);
      return;
    }
    PolyList L;
    SparsePolyRing P(AsSparsePolyRing(owner(*(GB.begin()))));
    for (PolyList::const_iterator it=GB.begin();it!=GB.end();++it)
    {
      SparsePolyIter it_p=BeginIter(*it);
      L.push_back(monomial(P, one(CoeffRing(P)), PP(it_p)));
    }
    swap(theLTResult,L);
  }//ComputeLT

}// end namespace cocoa


// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/TmpGOperations.C,v 1.10 2009/04/27 13:18:21 bigatti Exp $
// $Log: TmpGOperations.C,v $
// Revision 1.10  2009/04/27 13:18:21  bigatti
// -- added SaturatingAlg flag
//
// Revision 1.9  2009/01/30 13:41:50  bigatti
// -- enum instead of bool arguments
//
// Revision 1.8  2008/09/19 13:33:42  bigatti
// -- added: Sat algorithm (M.Caboara)
//
// Revision 1.7  2008/09/19 11:34:15  bigatti
// -- new mechanism for passing verbosity level (or StatLevel)
//    [only partially tested]
//
// Revision 1.6  2008/07/09 16:09:47  abbott
// Cosmetic tidying.  Removed bogus comments. Added missing & (for C++ const reference).
//
// Revision 1.5  2008/03/12 16:35:18  bigatti
// -- changed: IsHomogeneous --> IsHomog
// -- changed: ERR:ZeroPoly --> ERR::ZeroRingElem
//
// Revision 1.4  2007/11/09 10:45:52  bigatti
// -- [caboara] preparation for self-saturating algorithm
//
// Revision 1.3  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.2  2007/03/27 13:19:21  bigatti
// -- removed printout
//
// Revision 1.1  2007/03/09 18:56:56  bigatti
// -- added Tmp prefix to Groebner related files
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.17  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.16  2007/03/07 17:04:31  cocoa
// -- several changes by M.Caboara: more operations on ideals,
//    exception cleaner, coding conventions, WSugar, dynamic
//
// Revision 1.15  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.14  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.13  2007/01/25 12:05:20  bigatti
// -- fixed: removed coefficients in LT output
//
// Revision 1.12  2006/12/21 13:48:33  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.11  2006/12/04 13:55:54  cocoa
// -- added: sugar for GradingDim > 0  (called wsugar)
//
// Revision 1.10  2006/11/24 17:01:42  cocoa
// -- reorganized includes of header files
//
// Revision 1.9  2006/11/22 14:43:32  cocoa
// -- minor cleaning (indicated by Intel compiler)
//
// Revision 1.8  2006/11/20 14:57:17  cocoa
// -- added: (standard) sugar for modules
// -- fixed: non-homogeneous sysygies
// -- minor fixes     [M.Caboara]
//
// Revision 1.7  2006/10/06 16:46:17  cocoa
// -- syzygies for non-homogenous polynomials (Max)
// -- wip: evolution of Groebner Framework (Max)
//
// Revision 1.6  2006/10/06 10:48:34  cocoa
// Removed #include references to GradedFreeModule.H
//
// Revision 1.5  2006/08/17 09:35:49  cocoa
// -- changed: checks for homogeneity of input
//
// Revision 1.4  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.3  2006/07/18 10:52:24  cocoa
// -- added check for non-homogeneous input in ComputeElim
//
// Revision 1.2  2006/06/12 13:36:35  cocoa
// -- fixed embarrassing bug in WithDenominator1Hom clearing denominators
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.17  2006/05/16 08:59:16  cocoa
// -- added function for interactive Groebner
//
// Revision 1.16  2006/05/12 17:03:16  cocoa
// -- swapped arguments in homogenized
//
// Revision 1.15  2006/05/11 16:00:22  cocoa
// -- fixed spelling of "homogenize"
//
// Revision 1.14  2006/04/27 14:01:11  cocoa
// -- tidied up include files (using GTypes.H)
//
// Revision 1.13  2006/04/26 10:02:12  cocoa
// -- added GReductor::ourDefaultStatLevel variable to allow CoCoAServer
//    to set statistics level
//
// Revision 1.12  2006/04/21 16:47:06  cocoa
// -- new syntax for ComputeGBasis by Max
//
// Revision 1.11  2006/04/11 16:43:16  cocoa
// -- changed call of LPPNoCopy --> LPP (does not make a copy)
//
// Revision 1.10  2006/04/11 16:22:40  cocoa
// -- added: Elim, LT
//
