//   Copyright (c)  2005-2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for abstract class RingBase, RingElem, RefRingElem.
// Also most operations on ring elements.

#include "CoCoA/ring.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/QQ.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/symbol.H"

#include <cstddef>
using std::size_t;  // used only in RingBase::myBinaryPower
//#include <vector>
using std::vector;
#include <limits>
using std::numeric_limits; // used only in BinaryPower and SequentialPower


namespace CoCoA
{

  long RingBase::NewRingID()
  {
    static long PrevID = 0;
//    if (PrevID == numeric_limits<long>::max()) CoCoA_ERROR("Too many rings", "ring ctor");
    return ++PrevID;
  }


  ZZ characteristic(const ring& R)
  {
    ZZ p;
    R->myCharacteristic(p);
    return p;
  }

  // This makes a wasteful copy.
  vector<symbol> symbols(const ring& R)
  {
    vector<symbol> ans;
    R->mySymbols(ans);
    return ans;
  }


  void swap(RefRingElem f, RefRingElem g)
  {
    if (owner(f) != owner(g))
      CoCoA_ERROR(ERR::MixedRings, "swap(RingElem, RingElem)");
    owner(f)->mySwap(raw(f), raw(g));
  }



  RefRingElem& RefRingElem::operator=(const RefRingElem& x)
  {
    if (myOwner() != owner(x))
      CoCoA_ERROR(ERR::MixedRings, "RefRingElem = RefRingElem (assignment)");
    myOwner()->myAssign(myRawPtr(), raw(x));
    return *this;
  }

  RefRingElem& RefRingElem::operator=(const ConstRefRingElem& x)
  {
    if (myOwner() != owner(x))
      CoCoA_ERROR(ERR::MixedRings, "RefRingElem = ConstRefRingElem (assignment)");
    myOwner()->myAssign(myRawPtr(), raw(x));
    return *this;
  }

  RefRingElem& RefRingElem::operator=(const MachineInteger& n)
  {
    myOwner()->myAssign(myRawPtr(), n);
    return *this;
  }

  RefRingElem& RefRingElem::operator=(const ZZ& N)
  {
    myOwner()->myAssign(myRawPtr(), N);
    return *this;
  }

  RefRingElem& RefRingElem::operator=(const QQ& Q)
  {
    myOwner()->myAssign(myRawPtr(), Q);
    return *this;
  }

  //----------------------------------------------------------------------
  // Default definitions of some virtual functions.

  bool RingBase::IamGCDDomain() const
  {
    return true;
  }


  bool RingBase::IamOrderedDomain() const
  {
    return false;
  }


  RingElemRawPtr RingBase::myNew(const QQ& Q) const
  {
    RingElemRawPtr n = myNew(num(Q));
    RingElemRawPtr d = myNew(den(Q));
    const bool OK = myIsDivisible(n, n, d);
    myDelete(d);
    if (!OK)
    {
      myDelete(n);
      CoCoA_ERROR(ERR::EmbedQQFailed, "RingBase::myNew(QQ) -- generic embedding function");
    }
    return n;
  }

  void RingBase::myAssign(RawPtr rawlhs, const QQ& Q) const
  {
    RingElemRawPtr n = myNew(num(Q));
    RingElemRawPtr d = myNew(den(Q));
    const bool OK = myIsDivisible(n, n, d);
    myDelete(d);
    if (!OK)
    {
      myDelete(n);
      CoCoA_ERROR(ERR::EmbedQQFailed, "RingBase::myAssign(RawPtr,QQ) -- generic embedding function");
    }
    mySwap(rawlhs, n); // really an assignment
    myDelete(n);
  }


  bool RingBase::myIsInvertible(ConstRawPtr rawx) const
  {
    RingElem junk(ring(this));
    return myIsDivisible(raw(junk), raw(myOne()), rawx); // ??? discard quotient???
  }


  bool RingBase::myIsIrred(ConstRawPtr /*rawx*/) const
  {
//     if (myIsZero(rawx)) return CoCoA_ERROR(ERR::ZeroRingElem, "RingBase::myIsIrred(rawx)");
//     if (myIsInvertible(rawx)) return CoCoA_ERROR(ERR::InvertibleRingElem, "RingBase::myIsIrred(rawx)");
//     if (IamGCDDomain()) return CoCoA_ERROR(ERR::NYI, "RingBase::myIsIrred(rawx) for GCDDomain");
//     CoCoA_ERROR(ERR::NotTrueGCDDomain, "RingBase::myIsIrred(rawx)");
    CoCoA_ERROR(ERR::SERIOUS, "RingBase::myIsIrred(rawx)");
    return true; // NEVER REACHED; just to keep compiler quiet.
  }
  

  // This function should never be called, but it must be defined.
  void RingBase::myGcd(RawPtr /*rawlhs*/, ConstRawPtr /*rawx*/, ConstRawPtr /*rawy*/) const
  {
//???    CoCoA_ASSERT(!IamGCDDomain()); // if assert fails, some important code is missing.
    CoCoA_ERROR(ERR::SERIOUS, "RingBase::myGcd(,,)");
  }


  void RingBase::myGcdQuot(RawPtr rawlhs, RawPtr rawxquot, RawPtr rawyquot, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    CoCoA_ASSERT(IamGCDDomain());
    myGcd(rawlhs, rawx, rawy);
    myDiv(rawxquot, rawx, rawlhs);
    myDiv(rawyquot, rawy, rawlhs);
  }


  void RingBase::myExgcd(RawPtr /*rawlhs*/, RawPtr /*rawxcofac*/, RawPtr /*rawycofac*/, ConstRawPtr /*rawx*/, ConstRawPtr /*rawy*/) const
  {
    if (IamGCDDomain()) CoCoA_ERROR(ERR::NYI, "RingBase::myExgcd(,,)");
    CoCoA_ERROR(ERR::NotGCDDomain, "RingBase::myExgcd(,,)");
  }


  // Deal with all trivial cases; pass non-trivial cases to myPowerSmallExp.
  // Assume inputs args are mathematically sensible.
  void RingBase::myPower(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    if (n==0) // zeroth power is trivial
    {
      CoCoA_ASSERT(!myIsZero(rawx));
      myAssign(rawlhs, 1);
      return;
    }
    if (n==1 || myIsOne(rawx) || myIsZero(rawx)) { myAssign(rawlhs, rawx); return; }
    if (myIsMinusOne(rawx))
    {
      if ((n&1) != 0) myAssign(rawlhs, rawx); // -1 to odd power
      else myAssign(rawlhs, 1); // -1 to even power
      return;
    }
    // Non-trivial case: x is not -1, 0, or 1, and n > 1.
    // Handle squaring specially:
    if (n == 2) { mySquare(rawlhs, rawx); return; }
    myPowerSmallExp(rawlhs, rawx, n);
  }

  // Deal with all trivial cases; pass non-trivial cases to myPowerSmallExp or myPowerBigExp
  void RingBase::myPower(RawPtr rawlhs, ConstRawPtr rawx, const ZZ& N) const
  {
    if (IsZero(N))
    {
      CoCoA_ASSERT(!myIsZero(rawx));
      myAssign(rawlhs, 1);
      return;
    }
    if (N==1 || myIsOne(rawx) || myIsZero(rawx)) { myAssign(rawlhs, rawx); return; }
    if (myIsMinusOne(rawx))
    {
      if (IsOdd(N)) myAssign(rawlhs, rawx); // -1 to odd power
      else myAssign(rawlhs, 1);             // -1 to even power
      return;
    }
    // Non-trivial case: x is not -1, 0, or 1, and n > 1.
    // Handle squaring specially:
    if (N == 2) { mySquare(rawlhs, rawx); return; }
    // Call myPowerSmallExp or myPowerBigExp depending on value of exponent.
    unsigned long n;
    if (IsConvertible(n, N))
      myPowerSmallExp(rawlhs, rawx, n);
    else
      myPowerBigExp(rawlhs, rawx, N);
  }


  // Default implementation
  void RingBase::mySquare(RawPtr rawlhs, ConstRawPtr rawx) const
  {
    myMul(rawlhs, rawx, rawx);
  }


  // Default implementation does nothing.
  void RingBase::mySymbols(vector<symbol>& /*SymList*/) const
  {}


  bool RingBase::myIsPrintAtom(ConstRawPtr /*rawx*/) const
  {
    return false;
  }


  bool RingBase::myIsPrintedWithMinus(ConstRawPtr /*rawx*/) const
  {
    //    CoCoA_ERROR(ERR::SERIOUS, "RingBase::myIsPrintedWithMinus");
    return false;
  }


  bool RingBase::myIsMinusOne(ConstRawPtr rawx) const
  {
    RingElem tmp(ring(this));
    myNegate(raw(tmp), rawx);
    return myIsOne(raw(tmp));
  }


  // Default impl of  OBSOLESCENT  function.
  // if N and D are identical then numer overwrites denom.
  bool RingBase::myIsRational(ZZ& N, ZZ& D, ConstRawPtr rawx) const
  {
    QQ Q;
    const bool IsRat = myIsRational(Q, rawx);
    if (IsRat)
    {
      D = den(Q);
      N = num(Q);
    }
    return IsRat;
  }


  bool RingBase::myIsZeroAddMul(RawPtr rawlhs, ConstRawPtr rawfact1, ConstRawPtr rawfact2) const
  {
    RingElem tmp(ring(this));
    myMul(raw(tmp), rawfact1, rawfact2);
    myAdd(rawlhs, rawlhs, raw(tmp));
    return myIsZero(rawlhs);
  }


  // This function should never be called.  It could be called if you forget
  // to define myCmp for an ordered ring.
  int RingBase::myCmp(ConstRawPtr /*rawx*/, ConstRawPtr /*rawy*/) const
  {
    CoCoA_ERROR(ERR::SERIOUS, "RingBase::myCmp should never be called");
    return 0; // NEVER REACHED; just to keep compiler quiet.
  }


  int RingBase::mySign(ConstRawPtr rawx) const
  {
    int tmp = myCmp(rawx, raw(myZero()));
    if (tmp < 0) return -1;
    if (tmp == 0) return 0;
    return 1;
  }


  // This function should never be called.  It could be called if you forget
  // to define myFloor for an ordered ring.
  bool RingBase::myFloor(ZZ& /*N*/, ConstRawPtr /*rawx*/) const
  {
    CoCoA_ERROR(ERR::SERIOUS, "RingBase::myFloor should never be called");
    return false; // NEVER REACHED; just to keep compiler quiet.
  }


  // In myPowerBigExp we are guaranteed that the exponent N is too
  // large to fit into an unsigned long (i.e. it is genuinely large).
  // We also know that x is not -1,0,1; i.e. a non-trivial case.
  // By default we simply complain that the exponent is too big.
  void RingBase::myPowerBigExp(RawPtr /*rawlhs*/, ConstRawPtr /*rawx*/, const ZZ& /*N*/) const
  {
    CoCoA_ERROR(ERR::ExpTooBig, "power(r,N)");
  }

  // Direct iteration for computing powers -- good for multivariate polys.
  void RingBase::mySequentialPower(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Dispose of the trivial cases
    if (n == 0 || myIsOne(rawx)) { myAssign(rawlhs, 1); return; }
    if (myIsZero(rawx)) { myAssignZero(rawlhs); return; }

    RingElem ans(myOne());
    for (unsigned long i = n; i > 0; --i)
      myMul(raw(ans), rawx, raw(ans));
    mySwap(rawlhs, raw(ans));
  }


  // This function is private because of severe restrictions on its args
  // n must be strictly positive, and no aliasing between lhs and x.
  void RingBase::myBinaryPowerLoop(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    if (n == 1) { myAssign(rawlhs, rawx); return; }
    if (n == 2) { mySquare(rawlhs, rawx); return; }
    myBinaryPowerLoop(rawlhs, rawx, n/2);
    mySquare(rawlhs, rawlhs);
    if (n&1) myMul(rawlhs, rawlhs, rawx);
  }

  void RingBase::myBinaryPower(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const
  {
    // Dispose of the trivial cases
    if (n == 0 || myIsOne(rawx)) { myAssign(rawlhs, raw(myOne())); return; }
    if (myIsZero(rawx)) { myAssignZero(rawlhs); return; }
    if (n == 1) { myAssign(rawlhs, rawx); return; }

    RingElem ans(ring(this));
    myBinaryPowerLoop(raw(ans), rawx, n);
    mySwap(rawlhs, raw(ans));
  }


  void RingBase::myBinaryPower(RawPtr rawlhs, ConstRawPtr rawx, const ZZ& N) const
  {
    CoCoA_ASSERT("NEGATIVE EXPONENT" && N >= 0);
    // Dispose of the trivial cases
    if (IsZero(N) || myIsOne(rawx)) { myAssign(rawlhs, 1); return; }
    if (myIsZero(rawx)) { myAssignZero(rawlhs); return; }
    if (N == 1) { myAssign(rawlhs, rawx); return; }

    ConstRefRingElem X(ring(this), rawx);
    RingElem ans(X);
    for (size_t BitPos = mpz_sizeinbase(mpzref(N), 2); BitPos > 0; --BitPos)
    {
      ans *= ans;
      if (mpz_tstbit(mpzref(N), BitPos-1))
        ans *= X;
    }
    mySwap(rawlhs, raw(ans));
  }


  void RingBase::myGcdInField(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const
  {
    CoCoA_ASSERT(IamField());
    // In a field gcd always gives 1, unless both args are 0 when it gives 0.
    if (myIsZero(rawx) && myIsZero(rawy))
      myAssignZero(rawlhs);
    else
      myAssign(rawlhs, raw(myOne()));
  }


  /////////////////////////////////////////////////////////////////////////////

  RingElem::RingElem(const ring& R, const MachineInteger& n):
    RefRingElem(R, R->myNew(n))
  {}


  RingElem::RingElem(const ring& R, const ZZ& N):
    RefRingElem(R, R->myNew(N))
  {}


  RingElem::RingElem(const ring& R, const QQ& Q):
    RefRingElem(R, R->myNew(Q))
  {}


  RingElem::~RingElem()
  {
    myOwner()->myDelete(myRawPtr());
  }


  RingElem& RingElem::operator=(const RingElem& rhs)
  {
//????    if (this == &rhs) return *this;
    if (myOwner() != owner(rhs))
      CoCoA_ERROR(ERR::MixedRings, "RingElem = RingElem (assignment)");
    myOwner()->myAssign(myRawPtr(), raw(rhs));
    return *this;
  }


  RingElem& RingElem::operator=(const ConstRefRingElem& rhs)
  {
//????    if (this == &rhs) return *this;
    if (myOwner() != owner(rhs))
      CoCoA_ERROR(ERR::MixedRings, "RingElem = (Const)RefRingElem (assignment)");
    myOwner()->myAssign(myRawPtr(), raw(rhs));
    return *this;
  }


  RingElem& RingElem::operator=(const MachineInteger& n)
  {
    myOwner()->myAssign(myRawPtr(), n);
    return *this;
  }


  RingElem& RingElem::operator=(const ZZ& N)
  {
    myOwner()->myAssign(myRawPtr(), N);
    return *this;
  }


  RingElem& RingElem::operator=(const QQ& Q)
  {
    myOwner()->myAssign(myRawPtr(), Q);
    return *this;
  }


  bool operator==(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "RingElem == RingElem");
    return owner(x)->myIsEqual(raw(x), raw(y));
  }


  RingElem operator-(ConstRefRingElem x)
  {
    const ring& R = owner(x);
    RingElem ans(R);
    R->myNegate(raw(ans), raw(x));
    return ans;
  }


  RingElem operator+(ConstRefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem + RingElem");

    RingElem ans(Rx);
    Rx->myAdd(raw(ans), raw(x), raw(y));
    return ans;
  }


  RingElem operator-(ConstRefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem - RingElem");

    RingElem ans(Rx);
    Rx->mySub(raw(ans), raw(x), raw(y));
    return ans;
  }


  RingElem operator*(ConstRefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem * RingElem");

    RingElem ans(Rx);
    Rx->myMul(raw(ans), raw(x), raw(y));
    return ans;
  }


  RingElem operator/(ConstRefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem / RingElem");

    if (IsZero(y)) CoCoA_ERROR(ERR::DivByZero, "RingElem / RingElem");
    RingElem ans(Rx);
    if (!Rx->myIsDivisible(raw(ans), raw(x), raw(y)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem / RingElem");
    return ans;
  }


  RingElem gcd(ConstRefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "gcd(RingElem, RingElem)");

    if (!IsGCDDomain(Rx)) CoCoA_ERROR(ERR::NotGCDDomain, "gcd(x,y)");
    RingElem ans(Rx);
    Rx->myGcd(raw(ans), raw(x), raw(y));
    return ans;
  }


  void GcdQuot(RefRingElem gcd, RefRingElem xquot, RefRingElem yquot, ConstRefRingElem x, ConstRefRingElem y)
  {
    const ring& R = owner(gcd);
    if (owner(x) != R || owner(y) != R || owner(xquot) != R || owner(yquot) != R)
      CoCoA_ERROR(ERR::MixedRings, "GcdQuot(g,xq,yq,x,y)");
    if (!IsGCDDomain(R))
      CoCoA_ERROR(ERR::NotGCDDomain, "GcdQuot(g,xq,yq,x,y)");
    R->myGcdQuot(raw(gcd), raw(xquot), raw(yquot), raw(x), raw(y));
  }


  RefRingElem operator+=(RefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem += RingElem");

    Rx->myAdd(raw(x), raw(x), raw(y));
    return x;
  }


  RefRingElem operator-=(RefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem -= RingElem");

    Rx->mySub(raw(x), raw(x), raw(y));
    return x;
  }


  RefRingElem operator*=(RefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem *= RingElem");

    Rx->myMul(raw(x), raw(x), raw(y));
    return x;
  }


  RefRingElem operator/=(RefRingElem x, ConstRefRingElem y)
  {
    const ring& Rx = owner(x);
    const ring& Ry = owner(y);
    if (Rx != Ry)
      CoCoA_ERROR(ERR::MixedRings, "RingElem /= RingElem");

    if (IsZero(y)) CoCoA_ERROR(ERR::DivByZero, "RingElem /= RingElem");
    if (!Rx->myIsDivisible(raw(x), raw(x), raw(y)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem /= RingElem");
    return x;
  }


  std::ostream& operator<<(std::ostream& out, ConstRefRingElem x)
  {
    owner(x)->myOutput(out, raw(x));
    return out;
  }


// Currently replaced by a fully generic template version in tmp.H
//   std::ostream& operator<<(std::ostream& out, const vector<RingElem>& v)
//   {
//     if (v.empty()) return out << "[  ]";
//     vector<RingElem>::const_iterator it=v.begin();
//     out << "[ " << *it;
//     for (++it; it!=v.end() ; ++it) out << ",  " << *it;
//     return out << " ]";
//   }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, ConstRefRingElem x)
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("cocoa", "RingElem");
    OMOut << owner(x);
    owner(x)->myOutput(OMOut, raw(x));
    OMOut->mySendApplyEnd();
    return OMOut;
  }


  bool IsZero(ConstRefRingElem x)
  {
    return owner(x)->myIsZero(raw(x));
  }


  bool IsOne(ConstRefRingElem x)
  {
    return owner(x)->myIsOne(raw(x));
  }


  bool IsMinusOne(ConstRefRingElem x)
  {
    return owner(x)->myIsMinusOne(raw(x));
  }


  bool IsInteger(ZZ& N, ConstRefRingElem x)
  {
    return owner(x)->myIsInteger(N, raw(x));
  }


  // this fn is OBSOLESCENT
  bool IsRational(ZZ& N, ZZ& D, ConstRefRingElem x)
  {
    return owner(x)->myIsRational(N, D, raw(x));
  }


  bool IsRational(QQ& Q, ConstRefRingElem x)
  {
    return owner(x)->myIsRational(Q, raw(x));
  }


  bool IsInvertible(ConstRefRingElem x)
  {
    return owner(x)->myIsInvertible(raw(x));
  }


  bool IsIrred(ConstRefRingElem x)
  {
    return owner(x)->myIsIrred(raw(x));
  }


  bool IsDivisible(ConstRefRingElem num, ConstRefRingElem den)
  {
    const ring& Rnum = owner(num);
    const ring& Rden = owner(den);
    if (Rnum != Rden)
      CoCoA_ERROR(ERR::MixedRings, "IsDivisible(RingElem,RingElem)");

    if (IsZero(den)) return false;
    if (IsZero(num)) return true;
    RingElem quot(Rnum);
    return Rnum->myIsDivisible(raw(quot), raw(num), raw(den));
  }

  bool IsDivisible(const MachineInteger& n, ConstRefRingElem r)
  {
    RingElem tmp(owner(r), n);
    return IsDivisible(tmp, r);
  }

  bool IsDivisible(ConstRefRingElem r, const MachineInteger& n)
  {
    RingElem tmp(owner(r), n);
    return IsDivisible(r, tmp);
  }

  bool IsDivisible(const ZZ& N, ConstRefRingElem r)
  {
    RingElem tmp(owner(r), N);
    return IsDivisible(tmp, r);
  }

  bool IsDivisible(ConstRefRingElem r, const ZZ& N)
  {
    RingElem tmp(owner(r), N);
    return IsDivisible(r, tmp);
  }


  // Check the args are mathematically sensible; use myPower to do the work.
  RingElem power(ConstRefRingElem x, const MachineInteger& n) // deliberately allow negative exponents
  {
    if (IsZero(x) && IsNegative(n))
      CoCoA_ERROR(ERR::BadPwrZero, "power(RingElem, n)");
    if (IsNegative(n) < 0 && !IsInvertible(x))
      CoCoA_ERROR(ERR::NotUnit, "power(RingElem, n) and n < 0");
    if (!IsNegative(n))
    {
      RingElem ans(owner(x));
      owner(x)->myPower(raw(ans), raw(x), AsUnsignedLong(n));
      return ans;
    }
    // Here we know n < 0 and x is invertible
    RingElem invx(owner(x), 1);
    invx /= x;
    RingElem ans(owner(x));
    owner(x)->myPower(raw(ans), raw(invx), -AsSignedLong(n));
    return ans;
  }


  // Check the args are mathematically sensible; use myPower to do the work.
  RingElem power(ConstRefRingElem x, const ZZ& N) // deliberately allow negative exponents
  {
    if (IsZero(x) && N <= 0)
      CoCoA_ERROR(ERR::BadPwrZero, "power(RingElem, N)");
    if (N < 0 && !IsInvertible(x))
      CoCoA_ERROR(ERR::NotUnit, "power(RingElem, N) and N < 0");

    // General case for N a large integer
    if (N >= 0)
    {
      RingElem ans(owner(x));
      owner(x)->myPower(raw(ans), raw(x), N);
      return ans;
    }
    // Here we know N < 0 and x is invertible
    RingElem invx(owner(x), 1);
    invx /= x;
    RingElem ans(owner(x));
    owner(x)->myPower(raw(ans), raw(invx), -N);
    return ans;
  }


  //---------------------------------------------------------------------------
  // More syntactic sugar: arithmetic between RingElems and MachineIntegers

  bool operator==(ConstRefRingElem r, const MachineInteger& n)
  {
    return owner(r)->myIsEqual(raw(r), raw(RingElem(owner(r), n)));
  }


  RingElem operator+(ConstRefRingElem r, const MachineInteger& n)
  {
    RingElem ans(owner(r));
    owner(r)->myAdd(raw(ans), raw(r), raw(RingElem(owner(r), n)));
    return ans;
  }


  RingElem operator-(ConstRefRingElem r, const MachineInteger& n)
  {
    RingElem ans(owner(r));
    owner(r)->mySub(raw(ans), raw(r), raw(RingElem(owner(r), n)));
    return ans;
  }


  RingElem operator*(ConstRefRingElem r, const MachineInteger& n)
  {
    RingElem ans(owner(r));
    owner(r)->myMul(raw(ans), raw(r), raw(RingElem(owner(r), n)));
    return ans;
  }


  RingElem operator/(ConstRefRingElem r, const MachineInteger& n)
  {
    RingElem ans(owner(r), n); // also use ans to convert n to RingElem
    if (IsZero(ans)) CoCoA_ERROR(ERR::DivByZero, "RingElem / MachineInteger");
    if (!owner(r)->myIsDivisible(raw(ans), raw(r), raw(ans)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem / MachineInteger");
    return ans;
  }


  RingElem gcd(ConstRefRingElem r, const MachineInteger& n)
  {
    RingElem ans(owner(r));
    owner(r)->myGcd(raw(ans), raw(r), raw(RingElem(owner(r), n)));
    return ans;
  }


  //---------------------------------------------------------------------------
  // Operations between MachineInteger and RingElem.

  bool operator==(const MachineInteger& n, ConstRefRingElem r)
  {
    return owner(r)->myIsEqual(raw(r), raw(RingElem(owner(r), n)));
  }


  RingElem operator+(const MachineInteger& n, ConstRefRingElem r)
  {
    RingElem ans(owner(r));
    owner(r)->myAdd(raw(ans), raw(RingElem(owner(r), n)), raw(r));
    return ans;
  }


  RingElem operator-(const MachineInteger& n, ConstRefRingElem r)
  {
    RingElem ans(owner(r));
    owner(r)->mySub(raw(ans), raw(RingElem(owner(r), n)), raw(r));
    return ans;
  }


  RingElem operator*(const MachineInteger& n, ConstRefRingElem r)
  {
    RingElem ans(owner(r));
    owner(r)->myMul(raw(ans), raw(RingElem(owner(r), n)), raw(r));
    return ans;
  }


  RingElem operator/(const MachineInteger& n, ConstRefRingElem r)
  {
    if (IsZero(r)) CoCoA_ERROR(ERR::DivByZero, "MachineInteger / RingElem");
    RingElem ans(owner(r), n); // use ans to convert n to a RingElem
    if (!owner(r)->myIsDivisible(raw(ans), raw(ans), raw(r)))
      CoCoA_ERROR(ERR::BadQuot, "MachineInteger / RingElem");
    return ans;
  }


  RingElem gcd(const MachineInteger& n, ConstRefRingElem r)
  {
    RingElem ans(owner(r));
    owner(r)->myGcd(raw(ans), raw(RingElem(owner(r), n)), raw(r));
    return ans;
  }


  RefRingElem operator+=(RefRingElem r, const MachineInteger& n)
  {
    owner(r)->myAdd(raw(r), raw(r), raw(RingElem(owner(r), n)));
    return r;
  }


  RefRingElem operator-=(RefRingElem r, const MachineInteger& n)
  {
    owner(r)->mySub(raw(r), raw(r), raw(RingElem(owner(r), n)));
    return r;
  }


  RefRingElem operator*=(RefRingElem r, const MachineInteger& n)
  {
    owner(r)->myMul(raw(r), raw(r), raw(RingElem(owner(r), n)));
    return r;
  }


  RefRingElem operator/=(RefRingElem r, const MachineInteger& n)
  {
    RingElem den(owner(r), n);
    if (IsZero(den)) CoCoA_ERROR(ERR::DivByZero, "RingElem /= MachineInteger");
    if (!owner(r)->myIsDivisible(raw(r), raw(r), raw(den)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem /= MachineInteger");
    return r;
  }


  /////////////////////////////////////////////////////////////////////////////
  // More syntactic sugar: arithmetic between RingElems and ZZs

  bool operator==(ConstRefRingElem r, const ZZ& N)
  {
    return owner(r)->myIsEqual(raw(r), raw(RingElem(owner(r), N)));
  }


  RingElem operator+(ConstRefRingElem r, const ZZ& N)
  {
    RingElem ans(owner(r));
    owner(r)->myAdd(raw(ans), raw(r), raw(RingElem(owner(r), N)));
    return ans;
  }


  RingElem operator-(ConstRefRingElem r, const ZZ& N)
  {
    RingElem ans(owner(r));
    owner(r)->mySub(raw(ans), raw(r), raw(RingElem(owner(r), N)));
    return ans;
  }


  RingElem operator*(ConstRefRingElem r, const ZZ& N)
  {
    RingElem ans(owner(r));
    owner(r)->myMul(raw(ans), raw(r), raw(RingElem(owner(r), N)));
    return ans;
  }


  RingElem operator/(ConstRefRingElem r, const ZZ& N)
  {
    RingElem ans(owner(r), N); // also use ans to convert N to RingElem
    if (IsZero(ans)) CoCoA_ERROR(ERR::DivByZero, "RingElem / ZZ");
    if (!owner(r)->myIsDivisible(raw(ans), raw(r), raw(ans)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem / ZZ");
    return ans;
  }


  RingElem gcd(ConstRefRingElem r, const ZZ& N)
  {
    RingElem ans(owner(r));
    owner(r)->myGcd(raw(ans), raw(r), raw(RingElem(owner(r), N)));
    return ans;
  }


  bool operator==(const ZZ& N, ConstRefRingElem r)
  {
    return owner(r)->myIsEqual(raw(r), raw(RingElem(owner(r), N)));
  }


  RingElem operator+(const ZZ& N, ConstRefRingElem r)
  {
    RingElem ans(owner(r), N);
    owner(r)->myAdd(raw(ans), raw(ans), raw(r));
    return ans;
  }


  RingElem operator-(const ZZ& N, ConstRefRingElem r)
  {
    RingElem ans(owner(r), N);
    owner(r)->mySub(raw(ans), raw(ans), raw(r));
    return ans;
  }


  RingElem operator*(const ZZ& N, ConstRefRingElem r)
  {
    RingElem ans(owner(r), N);
    owner(r)->myMul(raw(ans), raw(ans), raw(r));
    return ans;
  }


  RingElem operator/(const ZZ& N, ConstRefRingElem r)
  {
    if (IsZero(r)) CoCoA_ERROR(ERR::DivByZero, "ZZ / RingElem");
    RingElem ans(owner(r), N);
    if (!owner(r)->myIsDivisible(raw(ans), raw(ans), raw(r)))
      CoCoA_ERROR(ERR::BadQuot, "ZZ / RingElem");
    return ans;
  }


  RingElem gcd(const ZZ& N, ConstRefRingElem r)
  {
    RingElem ans(owner(r));
    owner(r)->myGcd(raw(ans), raw(RingElem(owner(r), N)), raw(r));
    return ans;
  }


  RefRingElem operator+=(RefRingElem r, const ZZ& N)
  {
    owner(r)->myAdd(raw(r), raw(r), raw(RingElem(owner(r), N)));
    return r;
  }


  RefRingElem operator-=(RefRingElem r, const ZZ& N)
  {
    owner(r)->mySub(raw(r), raw(r), raw(RingElem(owner(r), N)));
    return r;
  }


  RefRingElem operator*=(RefRingElem r, const ZZ& N)
  {
    owner(r)->myMul(raw(r), raw(r), raw(RingElem(owner(r), N)));
    return r;
  }


  RefRingElem operator/=(RefRingElem r, const ZZ& N)
  {
    RingElem den(owner(r), N);
    if (IsZero(den)) CoCoA_ERROR(ERR::DivByZero, "RingElem /= ZZ");
    if (!owner(r)->myIsDivisible(raw(r), raw(r), raw(den)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem /= ZZ");
    return r;
  }


  std::ostream& operator<<(std::ostream& out, const ring& R)
  {
    R->myOutputSelf(out);
    return out;
  }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const ring& R)
  {
    R->myOutputSelf(OMOut);
    return OMOut;
  }


  // Comparisons for arithmetically ordered rings

  int sign(ConstRefRingElem x)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "sign(RingElem)");
    return owner(x)->mySign(raw(x));
  }


  RingElem abs(ConstRefRingElem x)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "abs(RingElem)");
    if (x < 0) return -x;
    return x;
  }


  ZZ floor(ConstRefRingElem x)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "floor(RingElem)");
    ZZ ans;
    owner(x)->myFloor(ans, raw(x));
    return ans;
  }


  ZZ ceil(ConstRefRingElem x)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "ceil(RingElem)");
    ZZ ans;
    if (!owner(x)->myFloor(ans, raw(x)))
      ++ans;
    return ans;
  }


  // This implementation rounds halves TOWARDS +INFINITY.
  ZZ NearestInteger(ConstRefRingElem x)
  {
    const ring& R = owner(x);
    if (!IsOrderedDomain(R))
      CoCoA_ERROR(ERR::NotOrdDom, "NearestInteger(RingElem)");

    // Two special cases:
    if (IsZ(R))
    {
      ZZ ans;
      IsInteger(ans, x);
      return ans;
    }
    RingElem two(R, 2);
    // If ring contains 1/2 then floor(x+1/2) is quickest method.
    if (IsInvertible(two))
      return floor(x+1/two);

    // Following is valid for any ring
    // (but troublesome for RingTwinFloat e.g. with arg x=(2.001,1.999))
    return floor(two*x+1)/2; // NB: integer (floor) division by 2.
  }


  int CmpDouble(ConstRefRingElem x, double z)
  {
    // This impl is simple rather than efficient.
    QQ Q;
    convert(Q, z);
    return cmp(den(Q)*x, num(Q));
  }


  int cmp(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "cmp(RingElem, RingElem)");
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(RingElem, RingElem)");
    return owner(x)->myCmp(raw(x), raw(y));
  }


  bool operator<(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "RingElem < RingElem");
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem < RingElem");
    return owner(x)->myCmp(raw(x), raw(y)) < 0;
  }


  bool operator<=(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "RingElem <= RingElem");
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem <= RingElem");
    return owner(x)->myCmp(raw(x), raw(y)) <= 0;
  }


  bool operator>(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "RingElem > RingElem");
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem > RingElem");
    return owner(x)->myCmp(raw(x), raw(y)) > 0;
  }


  bool operator>=(ConstRefRingElem x, ConstRefRingElem y)
  {
    if (owner(x) != owner(y))
      CoCoA_ERROR(ERR::MixedRings, "RingElem >= RingElem");
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem >= RingElem");
    return owner(x)->myCmp(raw(x), raw(y)) >= 0;
  }


  /////////////////////////////////////////////////////////////////////////////
  // First arg is a MachineInteger, second is a RingElem

  int cmp(const MachineInteger& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(MachineInteger, RingElem)");
    if (IsZero(x)) return -sign(y);
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y));
  }


  bool operator<(const MachineInteger& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "MachineInteger < RingElem");
    if (IsZero(x)) return sign(y) > 0;
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) < 0;
  }


  bool operator<=(const MachineInteger& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "MachineInteger <= RingElem");
    if (IsZero(x)) return sign(y) >= 0;
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) <= 0;
  }


  bool operator>(const MachineInteger& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "MachineInteger > RingElem");
    if (IsZero(x)) return sign(y) < 0;
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) > 0;
  }


  bool operator>=(const MachineInteger& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "MachineInteger >= RingElem");
    if (IsZero(x)) return sign(y) <= 0;
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) >= 0;
  }


  /////////////////////////////////////////////////////////////////////////////
  // First arg is a RingElem, second is a MachineInteger


  int cmp(ConstRefRingElem x, const MachineInteger& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(RingElem, MachineInteger)");
    if (IsZero(y)) return sign(x);
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y)));
  }


  bool operator<(ConstRefRingElem x, const MachineInteger& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem < MachineInteger");
    if (IsZero(y)) return sign(x) < 0;
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) < 0;
  }


  bool operator<=(ConstRefRingElem x, const MachineInteger& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem <= MachineInteger");
    if (IsZero(y)) return sign(x) <= 0;
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) <= 0;
  }


  bool operator>(ConstRefRingElem x, const MachineInteger& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem > MachineInteger");
    if (IsZero(y)) return sign(x) > 0;
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) > 0;
  }


  bool operator>=(ConstRefRingElem x, const MachineInteger& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem >= MachineInteger");
    if (IsZero(y)) return sign(x) >= 0;
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) >= 0;
  }

  ///////////////////////////////////////////////////////////////////////////
  // Comparisons between ZZ and RingElem
  int cmp(const ZZ& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(ZZ, RingElem)");
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y));
  }


  bool operator<(const ZZ& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "ZZ < RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) < 0;
  }


  bool operator<=(const ZZ& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "ZZ <= RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) <= 0;
  }


  bool operator>(const ZZ& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "ZZ > RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) > 0;
  }


  bool operator>=(const ZZ& x, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "ZZ >= RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),x)), raw(y)) >= 0;
  }


  /////////////////////////////////////////////////////////////////////////////
  // First arg is a RingElem, second is a const ZZ&


  int cmp(ConstRefRingElem x, const ZZ& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(RingElem, ZZ)");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y)));
  }


  bool operator<(ConstRefRingElem x, const ZZ& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem < ZZ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) < 0;
  }


  bool operator<=(ConstRefRingElem x, const ZZ& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem <= ZZ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) <= 0;
  }


  bool operator>(ConstRefRingElem x, const ZZ& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem > ZZ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) > 0;
  }


  bool operator>=(ConstRefRingElem x, const ZZ& y)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem >= ZZ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),y))) >= 0;
  }


  /////////////////////////////////////////////////////////////////////////////
  // More syntactic sugar: arithmetic between RingElems and QQs

  bool operator==(ConstRefRingElem r, const QQ& Q)
  {
    return owner(r)->myIsEqual(raw(r), raw(RingElem(owner(r), Q)));
  }


  RingElem operator+(ConstRefRingElem r, const QQ& Q)
  {
    RingElem ans(owner(r));
    owner(r)->myAdd(raw(ans), raw(r), raw(RingElem(owner(r), Q)));
    return ans;
  }


  RingElem operator-(ConstRefRingElem r, const QQ& Q)
  {
    RingElem ans(owner(r));
    owner(r)->mySub(raw(ans), raw(r), raw(RingElem(owner(r), Q)));
    return ans;
  }


  RingElem operator*(ConstRefRingElem r, const QQ& Q)
  {
    RingElem ans(owner(r));
    owner(r)->myMul(raw(ans), raw(r), raw(RingElem(owner(r), Q)));
    return ans;
  }


  RingElem operator/(ConstRefRingElem r, const QQ& Q)
  {
    RingElem ans(owner(r), Q); // also use ans to convert N to RingElem
    if (IsZero(ans)) CoCoA_ERROR(ERR::DivByZero, "RingElem / QQ");
    if (!owner(r)->myIsDivisible(raw(ans), raw(r), raw(ans)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem / QQ");
    return ans;
  }


  bool operator==(const QQ& Q, ConstRefRingElem r)
  {
    return owner(r)->myIsEqual(raw(r), raw(RingElem(owner(r), Q)));
  }


  RingElem operator+(const QQ& Q, ConstRefRingElem r)
  {
    RingElem ans(owner(r), Q);
    owner(r)->myAdd(raw(ans), raw(ans), raw(r));
    return ans;
  }


  RingElem operator-(const QQ& Q, ConstRefRingElem r)
  {
    RingElem ans(owner(r), Q);
    owner(r)->mySub(raw(ans), raw(ans), raw(r));
    return ans;
  }


  RingElem operator*(const QQ& Q, ConstRefRingElem r)
  {
    RingElem ans(owner(r), Q);
    owner(r)->myMul(raw(ans), raw(ans), raw(r));
    return ans;
  }


  RingElem operator/(const QQ& Q, ConstRefRingElem r)
  {
    if (IsZero(r)) CoCoA_ERROR(ERR::DivByZero, "QQ / RingElem");
    RingElem ans(owner(r), Q);
    if (!owner(r)->myIsDivisible(raw(ans), raw(ans), raw(r)))
      CoCoA_ERROR(ERR::BadQuot, "QQ / RingElem");
    return ans;
  }


  RefRingElem operator+=(RefRingElem r, const QQ& Q)
  {
    owner(r)->myAdd(raw(r), raw(r), raw(RingElem(owner(r), Q)));
    return r;
  }


  RefRingElem operator-=(RefRingElem r, const QQ& Q)
  {
    owner(r)->mySub(raw(r), raw(r), raw(RingElem(owner(r), Q)));
    return r;
  }


  RefRingElem operator*=(RefRingElem r, const QQ& Q)
  {
    owner(r)->myMul(raw(r), raw(r), raw(RingElem(owner(r), Q)));
    return r;
  }


  RefRingElem operator/=(RefRingElem r, const QQ& Q)
  {
    RingElem den(owner(r), Q);
    if (IsZero(den)) CoCoA_ERROR(ERR::DivByZero, "RingElem /= QQ");
    if (!owner(r)->myIsDivisible(raw(r), raw(r), raw(den)))
      CoCoA_ERROR(ERR::BadQuot, "RingElem /= QQ");
    return r;
  }


  ///////////////////////////////////////////////////////////////////////////
  // Comparisons between QQ and RingElem
  int cmp(const QQ& Q, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(QQ, RingElem)");
    return owner(y)->myCmp(raw(RingElem(owner(y),Q)), raw(y));
  }


  bool operator<(const QQ& Q, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "QQ < RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),Q)), raw(y)) < 0;
  }


  bool operator<=(const QQ& Q, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "QQ <= RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),Q)), raw(y)) <= 0;
  }


  bool operator>(const QQ& Q, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "QQ > RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),Q)), raw(y)) > 0;
  }


  bool operator>=(const QQ& Q, ConstRefRingElem y)
  {
    if (!IsOrderedDomain(owner(y)))
      CoCoA_ERROR(ERR::NotOrdDom, "QQ >= RingElem");
    return owner(y)->myCmp(raw(RingElem(owner(y),Q)), raw(y)) >= 0;
  }


  /////////////////////////////////////////////////////////////////////////////
  // First arg is a RingElem, second is a QQ


  int cmp(ConstRefRingElem x, const QQ& Q)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "cmp(RingElem, QQ)");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),Q)));
  }


  bool operator<(ConstRefRingElem x, const QQ& Q)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem < QQ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),Q))) < 0;
  }


  bool operator<=(ConstRefRingElem x, const QQ& Q)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem <= QQ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),Q))) <= 0;
  }


  bool operator>(ConstRefRingElem x, const QQ& Q)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem > QQ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),Q))) > 0;
  }


  bool operator>=(ConstRefRingElem x, const QQ& Q)
  {
    if (!IsOrderedDomain(owner(x)))
      CoCoA_ERROR(ERR::NotOrdDom, "RingElem >= QQ");
    return owner(x)->myCmp(raw(x), raw(RingElem(owner(x),Q))) >= 0;
  }




} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ring.C,v 1.14 2009/12/11 11:46:32 abbott Exp $
// $Log: ring.C,v $
// Revision 1.14  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.13  2009/12/03 17:37:22  abbott
// Moved fns characteristic & swap into .C file, so that .H does not
// need to include ZZ.H and error.H.
//
// Revision 1.12  2009/10/29 18:30:09  abbott
// Changed the ring ID value to be a long (previously was size_t).
// Some minor cleaning to include directives.
//
// Revision 1.11  2009/07/02 16:32:10  abbott
// Consequential changes stemming from new class QQ, and modified interface to the member
// function RingBase::myIsRational.  Also some new conversion functions.
//
// Revision 1.10  2008/12/17 12:11:52  abbott
// Changed type from long to MachineInteger in operations which use a machine integer
// in place of a RingElem.  The change is "superficial" but affects many files.
//
// Revision 1.9  2008/11/18 10:24:48  abbott
// Added floor, ceil, NearestInteger and CmpDouble functions.
// Also added myFloor member fn.  Plus some very minor code tidying.
//
// Revision 1.8  2008/04/22 14:44:26  abbott
// Removed top level function square.
// Internal powering routine myBinaryPower now uses mySquare.
//
// Revision 1.7  2008/04/22 13:09:15  abbott
// Removed IsPositive and IsNegative functions for ZZs.
// Comparison between RingElem and 0 is now handled specially (specially fast?).
//
// Revision 1.6  2008/04/15 14:58:06  bigatti
// -- added mySquare, square
//
// Revision 1.5  2008/03/12 16:29:37  bigatti
// -- added: IsIrred
//
// Revision 1.4  2007/10/30 17:14:06  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/05/22 22:45:14  abbott
// Changed fn name IsUnit to IsInvertible.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.8  2006/12/07 17:23:46  cocoa
// -- for compilation with _Wextra: commented out names of unused arguments
//
// Revision 1.7  2006/12/06 17:40:10  cocoa
// -- style: RawPtr args are now called "raw.."
//
// Revision 1.6  2006/11/20 15:55:02  cocoa
// ring is now a class again.  Improved definitions of operator-> in derived classes.
//
// Revision 1.5  2006/11/02 13:25:43  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.4  2006/10/27 19:09:45  cocoa
// Replaced some member functions of CoCoA::symbol by friend functions.
// Removed some include dependency on symbol.H
//
// Revision 1.3  2006/10/16 23:18:58  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.2  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.11  2006/05/29 16:22:37  cocoa
// Third time lucky???
// Added myIsInteger member function to all rings (NYI for RingFloat).
//
// Revision 1.10  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.9  2006/04/27 13:43:35  cocoa
// Added abs function (only for arithmetically ordered rings).
//
// Revision 1.8  2006/04/21 15:01:36  cocoa
// Changed default implementation of RingBase::myGcd -- it now gives a SERIOUS
// error.  All fields must now handle a call to gcd explicitly: they can use
// the new myGcdInField function.  It's now cleaner than it was.
//
// Revision 1.7  2006/04/04 17:20:53  cocoa
// -- fixed: myPower trivial case n==0
//
// Revision 1.6  2006/03/27 12:21:25  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.5  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.4  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.3  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.2  2006/02/14 16:22:20  cocoa
// -- defined "operator<<" for vector<RingElem>&  in ring.H/C
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.4  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.3  2005/07/08 15:09:28  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//
// Revision 1.2  2005/06/30 16:02:18  cocoa
// -- myIsPrintedWithMinus now returns false (used to throw)
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.6  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.5  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.4  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.3  2005/02/11 16:45:24  cocoa
// Removed the useless and misleading functions myInit and myKill
// from the SmallFp*Impl classes; various consequential changes.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.17  2004/11/18 18:33:40  cocoa
// Now every ring know its own "one" element (as well as "zero").
// Several consequential changes.
//
// Revision 1.16  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.15  2004/11/11 14:05:02  cocoa
// -- minor changes for doxygen
//
// Revision 1.14  2004/11/11 11:56:09  cocoa
// (1) Tidied makefiles, and introduced common.mki
// (2) Improved several tests, and cleaned them so that they
//     handle sanely any otherwise unhandled exceptions.
//
// Revision 1.13  2004/11/04 18:47:42  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.12  2004/07/27 16:03:38  cocoa
// Added IsCommutative test and IamCommutative member function
// to all rings.  Tidied geobuckets a little.
//
// Revision 1.11  2004/07/20 15:04:06  cocoa
// The next step in the new "ring element" conversion process:
// handling the case of creating a "const RefRingElem" object
// (since C++ refuses to do this properly itself).
//
// Revision 1.10  2004/07/16 15:45:12  cocoa
// First stage of new RingElem implementation completed.
//
// Revision 1.9  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.8  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.7  2004/04/08 15:33:34  cocoa
// Added function IsInteger, and the related RingBase::myIsInteger
// virtual function, plus all necessary implementations.
//
// Revision 1.6  2004/03/20 17:46:10  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.5  2004/02/03 16:16:20  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.4  2003/11/14 13:06:05  cocoa
// -- New function "myIsPrintAtom" for printing polynomials and fractions
//
// Revision 1.3  2003/10/17 10:51:06  cocoa
// Major cleaning, and new naming convention.
//
// Revision 1.2  2003/10/09 12:13:44  cocoa
// New coding convention for rings.
//
// Revision 1.18  2003/06/23 16:36:46  abbott
// Prior to public release:
//   fixed wrongly placed zero test in operator/(ConstRefRingElem,long)
//
// Revision 1.17  2003/04/16 14:16:46  abbott
// Added code to handle some trivial cases in SequentialPower and
// BinaryPower.  Added definition of new despatch function IsDivisible.
//
// Revision 1.16  2003/01/21 15:38:48  abbott
// Many minor syntactic modifications consequent from the new
// names and structure in ring.H.
//
// Revision 1.15  2002/12/13 15:16:03  abbott
// Added functions IsOne and IsMinusOne (both at global level and
// as member functions of an AbstractRing).  Commented out the
// member function for testing equality between a RingElem and a
// machine integer.  Added member functions for initializing or
// assigning from a big integer (of type ZZ).  Added syntactic
// sugar operators for arithmetic between RingElems and ZZs.
//
// Revision 1.14  2002/12/11 11:49:18  abbott
// Checking in for safety -- this code does at least compile and work.
// Exponents in power functions must now be non-negative (unsigned int).
// Added functions between RingElems and ZZs (big integers).
// Added two member functions for initialising/assigning from big
// integers.
//
// Made op == and != inline because Anna wanted them so; probably revoke
// this after IsOne has been invented.
//
// Revision 1.13  2002/11/14 16:25:32  abbott
// Extensive consequential changes due to renaming in ring.H.
// ring renamed to AbstractRing.
// New typedef RingElem for AbstractRing::elem.
// New typedef ConstRawValue to be used as arg type in ring member functions.
// Old alias subclass has been largely replaced by new ConstRefElem subclass.
// New typedef ConstRefRingElem to be used as arg type for read only RingElems :-(
//
// Revision 1.12  2002/07/05 15:18:17  abbott
// Added definition of member fn SequentialPower (previously simply called power).
// Redefined op/ so that it calls IsDivisible, and gives a useful error mesg
// if the quotient does not exist in the ring.
//
// Revision 1.11  2002/06/27 16:06:34  abbott
// Added BinaryPower and BinaryPowerLoop default definitions.
//
// Revision 1.10  2002/06/21 14:40:16  abbott
// Rewrote the default implementation of IsZeroAddMul; now it is
// cleaner, and possibly more exception safe.
// Added definitions of the "syntactic sugar" arithmetic operations
// (between ring::elems and longs).
//
// Revision 1.9  2002/05/30 13:34:14  abbott
// Moved equals and not-equals operators to ring.H so they can be inline.
//
// Revision 1.8  2002/05/15 14:55:26  abbott
// Added power function default definition.
// Added code to handle alias ring::elem objects.
// General cleaning, and reorganization of indentation.
//
// Revision 1.7  2002/02/15 11:55:11  bigatti
// - added IsGCDDomain, gcd, "=", "==", "!="
//
// Revision 1.6  2002/01/30 15:00:14  abbott
// Tidied the "using" declarations.
// Tidied the "unimplemented: mixed ring arith" error messages.
// Removed the definitions of the two "raw" functions to ring.H.
//
// Revision 1.5  2001/12/07 18:20:40  abbott
// Changed names in accordance with new coding conventions
// (in particular ring::raw_elem became ring::RawValue).
//
// Revision 1.4  2001/11/16 18:45:06  bigatti
// added:   using namespace std;
// for compatibility with gcc-3
//
// Revision 1.3  2001/11/07 20:52:53  abbott
// Added implementations of the functions elem::raw(...).
// These could be inline, but are not yet.  I'll wait until I'm
// sure they're needed before thinking too much about it.
//
// Revision 1.2  2001/10/29 20:34:38  abbott
// Several changes: numerous minor fixes, and the more important correction
// of the standard binary arithmetic operators so that they no longer leak
// heap copies of every value computed (red face time).  The simplistic
// fixes rather imply a strong need for reference counting (or some other
// way of achieving cheap copying) of elem values -- to be discussed!
//
// Revision 1.1  2001/10/05 12:44:31  abbott
// Initial revision
//
