//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Implementation file for the class submodule

#include "CoCoA/submodule.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/FreeModule.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/ring.H"

#include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
#include <vector>
using std::vector;


namespace CoCoA
{

  /*-----------------------------------------------------------------*/
  /** \include submodule.txt  */
  /*-----------------------------------------------------------------*/

  class SubmoduleImpl: public FGModuleBase
  {
    // Two typedefs to save typing.
    typedef ModuleBase::RawPtr RawPtr;
    typedef const ModuleBase::RawPtr& ConstRawPtr;

  public:
    SubmoduleImpl(const module& M, const std::vector<ModuleElem>& gens);
    std::size_t myNumCompts() const;
    const ring& myBaseRing() const;
    const FreeModule& myAmbientFreeModule() const;
    const std::vector<ModuleElem>& myGens() const;
    const std::vector<ModuleElem>& myTidyGens() const;

    const ModuleElem& myZero() const;
    void myNew(RawPtr& rawv) const;
    void myNew(RawPtr& rawv, ConstRawPtr rawt) const;
    void myDelete(RawPtr& rawv) const;                                         // destroys v (incl all resources)
    void mySwap(RawPtr& rawv, RawPtr& raww) const;                           // swap(v, w);
    void myAssign(RawPtr& rawlhs, ConstRawPtr rawv) const;                   // lhs = v;
    ConstRefRingElem myCompt(const RawPtr& rawv, size_t pos) const;          ///< v[pos] (READ ONLY)
    RefRingElem myCompt(RawPtr& rawv, size_t pos) const;                     ///< v[pos]
    void myNegate(RawPtr& rawlhs, ConstRawPtr rawv) const;                   // lhs = -v
    void myAdd(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const;    // lhs = v+w;
    void mySub(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const;    // lhs = v-w;

    void myMul(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const; // lhs = r*v;
    void myDiv(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const; // lhs = (1/r)*v;
    void myOutput(std::ostream& out, ConstRawPtr rawv) const;                // out << v
    void myOutputSelf(std::ostream& out) const;                              // out << M
    void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawv) const;            // OMOut << v
    void myOutputSelf(OpenMathOutput& OMOut) const;                          // OMOut << M
    bool myIsZero(ConstRawPtr rawv) const;                                   // v == 0
//???    bool IsZeroAddMul(RawPtr& rawlhs, RingElemConstRawPtr rawy, ConstRawPtr rawz) const;  // lhs += y*z, result says whether lhs == 0.
    bool myIsEqual(ConstRawPtr, ConstRawPtr) const;

  private: // data members
    const FreeModule myM;
    std::vector<ModuleElem> myGensArray;
    mutable bool myTidyGensIsValid;
    mutable std::vector<ModuleElem> myTidyGensArray;
//???    std::vector<ModuleElem>& ComputeTidyGens() const;
  };



  SubmoduleImpl::SubmoduleImpl(const module& M, const std::vector<ModuleElem>& gens):
      myM(AsFreeModule(M)),
      myGensArray(gens)
  {
    for (size_t i=0; i < gens.size(); ++i)
      if (owner(gens[i]) != M)
        CoCoA_ERROR(ERR::MixedModules, "SubmoduleImpl(M, gens)");
    myRefCountZero();
  }


  std::size_t SubmoduleImpl::myNumCompts() const
  {
    return NumCompts(myM);
  }


  const ring& SubmoduleImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  const FreeModule& SubmoduleImpl::myAmbientFreeModule() const
  {
    return myM;
  }


  const std::vector<ModuleElem>& SubmoduleImpl::myGens() const
  {
    return myGensArray;
  }


  const std::vector<ModuleElem>& SubmoduleImpl::myTidyGens() const
  {
    if (!myTidyGensIsValid)
    {
      CoCoA_ERROR(ERR::NYI, "SubmoduleImpl::myTidyGens");
    }
    return myGensArray;
  }


  const ModuleElem& SubmoduleImpl::myZero() const
  {
    return zero(myM);
  }


  void SubmoduleImpl::myNew(RawPtr& rawv) const
  {
    myM->myNew(rawv);
  }


  void SubmoduleImpl::myNew(RawPtr& rawv, ConstRawPtr rawcopy) const
  {
    myM->myNew(rawv, rawcopy);
  }


  void SubmoduleImpl::myDelete(RawPtr& rawv) const
  {
    myM->myDelete(rawv);
  }


  void SubmoduleImpl::mySwap(RawPtr& rawv, RawPtr& raww) const
  {
    myM->mySwap(rawv, raww);
  }


  void SubmoduleImpl::myAssign(RawPtr& rawlhs, ConstRawPtr rawv) const
  {
    myM->myAssign(rawlhs, rawv);
  }


  ConstRefRingElem SubmoduleImpl::myCompt(const RawPtr& rawv, size_t pos) const
  {
    return myM->myCompt(rawv, pos);
  }

  RefRingElem SubmoduleImpl::myCompt(RawPtr& rawv, size_t pos) const
  {
    return myM->myCompt(rawv, pos);
  }


  void SubmoduleImpl::myNegate(RawPtr& rawlhs, ConstRawPtr rawv) const
  {
    myM->myNegate(rawlhs, rawv);
  }


  void SubmoduleImpl::myAdd(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const
  {
    myM->myAdd(rawlhs, rawv, raww);
  }


  void SubmoduleImpl::mySub(RawPtr& rawlhs, ConstRawPtr rawv, ConstRawPtr raww) const
  {
    myM->mySub(rawlhs, rawv, raww);
  }


  void SubmoduleImpl::myMul(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const
  {
    myM->myMul(rawlhs, rawx, rawv);
  }


  void SubmoduleImpl::myDiv(RawPtr& rawlhs, RingElemConstRawPtr rawx, ConstRawPtr rawv) const
  {
    myM->myDiv(rawlhs, rawx, rawv);
  }


  void SubmoduleImpl::myOutput(std::ostream& out, ConstRawPtr rawv) const
  {
    myM->myOutput(out, rawv);
  }


  void SubmoduleImpl::myOutputSelf(std::ostream& out) const
  {
    out << "submodule(" << myM << ", [";
    if (!myGensArray.empty()) out << myGensArray[0];
    for (size_t i=1; i < myGensArray.size(); ++i)
    {
      out << ", " << myGensArray[i];
    }
    out << "])";
  }


  void SubmoduleImpl::myOutput(OpenMathOutput& OMOut, ConstRawPtr rawv) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("???", "ModuleElement"); // BUG: what should this OMSymbol be???
    OMOut << myM;
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("???", "list"); // BUG: what should this OMSymbol be???
    OMOut << myNumCompts();
    myM->myOutput(OMOut, rawv); // BUG: this should be a "naked" output???
    OMOut->mySendApplyEnd();
    OMOut->mySendApplyEnd();
  }


  void SubmoduleImpl::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("???", "submodule"); // BUG: what should this OMSymbol be???
    OMOut << myM;
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("???", "list"); // BUG: what should this OMSymbol be???
    OMOut << myGensArray.size();
    for (size_t i=0; i < myGensArray.size(); ++i)
      OMOut << myGensArray[i];  // BUG: this should be a "naked" output???
    OMOut->mySendApplyEnd();
    OMOut->mySendApplyEnd();
  }


  bool SubmoduleImpl::myIsZero(ConstRawPtr rawv) const
  {
    return myM->myIsZero(rawv);
  }


//???    bool IsZeroAddMul(RawPtr& rawlhs, RingElemConstRawPtr rawy, ConstRawPtr rawz) const;  // lhs += y*z, result says whether lhs == 0.


  bool SubmoduleImpl::myIsEqual(ConstRawPtr rawv, ConstRawPtr raww) const
  {
    return myM->myIsEqual(rawv, raww);
  }


  FGModule NewSubmodule(const FGModule& M, const std::vector<ModuleElem>& gens)
  {
    return FGModule(new SubmoduleImpl(M, gens));
  }


}  // end of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/submodule.C,v 1.4 2009/12/03 17:26:34 abbott Exp $
// $Log: submodule.C,v $
// Revision 1.4  2009/12/03 17:26:34  abbott
// Renamed EFGModule to FGModule.
// Renamed ModuleBase member fns  myInit -> myNew, myKill -> myDelete.
// Removed some cruft (old code that was not used by anyone).
//
// Revision 1.3  2008/05/30 12:50:48  abbott
// Aligned some comments.
//
// Revision 1.2  2007/10/30 17:14:06  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.4  2007/01/15 14:39:12  bigatti
// -- added prefix "raw" to RawPtr arguments names
//
// Revision 1.3  2006/11/02 13:25:43  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.2  2006/10/06 10:15:52  cocoa
// In response to Susan's bug: a fiasco when compiling with CoCoA_MEMPOOL_DEBUG
// set wrongly.  Moved several implementation classes out of their header files
// into the implementation files.  Several functions had to be uninlined.
// Also corrected position of #include, etc.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.7  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.6  2006/04/21 14:56:33  cocoa
// Changed return type of myCompt member function: now it returns a
// ConstRefRingElem instead of a RingElem (i.e. a copy).
//
// Revision 1.5  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.4  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.3  2005/11/29 13:04:47  cocoa
// -- added "const" to myCompt argument
//
// Revision 1.2  2005/11/24 16:09:38  cocoa
// -- added operator[] for ModuleElem
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.4  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.3  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.6  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.5  2004/11/11 14:06:20  cocoa
// -- moved CVS log to the bottom
// -- minor changes for doxygen
//
// Revision 1.4  2004/06/29 17:10:22  cocoa
// Partially tidied use of "protected" and "private" in various
// base classes.  Checking in at the end of the day -- it works,
// and I wouldn't want it to be lost next time point's disk
// misbehaves.
//
// Revision 1.3  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.2  2004/01/28 15:37:58  cocoa
// Fairly major update: resuscitated "old style" code which didn't compile
// under the current organization.
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
