//   Copyright (c)  2005,2007,2008  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/symbol.H"
#include "CoCoA/utils.H"
#include "CoCoA/error.H"

// #include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
using std::istream;
//#include <string>
using std::string;
//#include <vector>
using std::vector;
#include <algorithm>
using std::sort;
#include <limits>
using std::numeric_limits;

namespace CoCoA
{

  // head must begin with a letter and be comprised of letters or underscore.
  bool symbol::IsValidHead(const std::string& head)
  {
    if (head.empty() || !isalpha(head[0])) return false;
    const size_t n = head.size();
    for (size_t i=1; i < n; ++i)
    {
      if (head[i] != '_' && !isalpha(head[i]))
        return false;
    }
    return true;
  }


  symbol::symbol(const std::string& head):
      myHead(head),
      myIndices(0)
  {
    if (!IsValidHead(myHead))
      CoCoA_ERROR(ERR::BadSymbolHead, "symbol(head)");
  }


  symbol::symbol(const std::string& head, MachineInteger index):
      myHead(head),
      myIndices(1)
  {
    if (!IsValidHead(myHead))
      CoCoA_ERROR(ERR::BadSymbolHead, "symbol(head, index)");
    if (!IsSignedLong(index))
      CoCoA_ERROR(ERR::BadSymbolIndex, "symbol(head,index)");
    myIndices[0] = AsSignedLong(index);
  }



  symbol::symbol(const std::string& head, MachineInteger index1, MachineInteger index2):
      myHead(head),
      myIndices(2)
  {
    if (!IsValidHead(myHead))
      CoCoA_ERROR(ERR::BadSymbolHead, "symbol(head, index)");
    if (!IsSignedLong(index1) || !IsSignedLong(index2))
      CoCoA_ERROR(ERR::BadSymbolIndex, "symbol(head,index)");
    myIndices[0] = AsSignedLong(index1);
    myIndices[1] = AsSignedLong(index2);
  }



  symbol::symbol(const std::string& head, const std::vector<long>& indices):
      myHead(head),
      myIndices(indices)
  {
    if (!IsValidHead(myHead))
      CoCoA_ERROR(ERR::BadSymbolHead, "symbol(head, indices)");
  }


  void symbol::myOutputSelf(std::ostream& out) const
  {
    out << myHead;
    if (myIndices.empty()) return;
    out << '[' << myIndices[0];
    const size_t n = myIndices.size();
    for (size_t i=1; i < n; ++i)
    {
      out << ", " << myIndices[i];
    }
    out << ']';
  }

//    void myOutputSelf(OpenMath::ostream& out) const;????


  // This is semi-exception clean: value of object is modified only
  // if the entire read is successful.
  // This version does not handle white space as I'd like -- see symbol.txt.
  // Normal input fns merely set the failbit of the stream for parse errors.
  void symbol::myInput(istream& in)
  {
    if (!in) return; // quit immediately if stream is not good.
    string head;
    //if (!isalpha(in.peek())) CoCoA_ERROR(ERR::???, "symbol input"); //??? in.setstate(std::ios::failbit);
    while (isalpha(in.peek()))
    {
      char ch;
      in >> ch;
      head.push_back(ch);
    }
    //if (head.empty()) CoCoA_ERROR(ERR::???, "symbol input"); //??? in.setstate(std::ios::failbit);
    if (in.peek() != '[') { swap(myHead, head); myIndices.clear(); return; }
    in.ignore(); // ignore the initial open square bracket
    vector<long> indices;
    while (isdigit(in.peek()))
    {
      long index;
      in >> index;
      indices.push_back(index);
      //if (in.peek() != ',') CoCoA_ERROR(ERR::???, "symbol input");
      in.ignore(); // ignore the comma
    }
    //if (in.peek() != ']') CoCoA_ERROR(ERR::???, "symbol input"); //??? in.setstate(...);
    in.ignore(); // ignore the final close square bracket
    //if (indices.empty()) CoCoA_ERROR(ERR::???, "symbol input"); //??? in.setstate(...);
    swap(myHead, head);
    swap(myIndices, indices);
  }


  int symbol::myCmp(const symbol& sym2) const
  {
    int HeadCmp = LexCmp3(myHead.begin(), myHead.end(),
                          sym2.myHead.begin(), sym2.myHead.end());
    if (HeadCmp != 0) return HeadCmp;
    return LexCmp3(myIndices.begin(), myIndices.end(),
                   sym2.myIndices.begin(), sym2.myIndices.end());
  }


  namespace // anon namespace for file local fn
  {
    void FillRange(vector<symbol>& ans,
                   const string& head,
                   const vector<long>& indices1,
                   const vector<long>& indices2,
                   vector<long>& workspace,
                   size_t i)
    {
      if (i == workspace.size())
      {
        ans.push_back(symbol(head, workspace));
        return;
      }

      for (long j=indices1[i]; /*j <= indices2[i]*/; ++j)
      {
        workspace[i] = j;
        FillRange(ans, head, indices1, indices2, workspace, i+1);
        if (j == indices2[i]) break;
      }
    }
  }


  // This does a sort of cartesian product; not tricky, just long and tedious.
  std::vector<symbol> SymbolRange(const symbol& sym1, const symbol& sym2)
  {
    const size_t n = NumIndices(sym1);
    if (head(sym1) != head(sym2) || NumIndices(sym2) != n)
      CoCoA_ERROR(ERR::BadSymbolRange, "SymbolRange(sym1,sym2)");

    // Boring trivial case; maybe it should be an error?
    if (sym1 == sym2)
      return vector<symbol>(1, sym1);

    vector<symbol> ans;
    const size_t MAX = ans.max_size();
    // Conduct sanity check on index ranges, and compute
    // number of elements in final range; complain if too big.
    size_t RangeSize = 1;
    for (size_t i=0; i < n; ++i)
    {
      if (index(sym1,i) > index(sym2,i))
        CoCoA_ERROR(ERR::BadSymbolRange, "SymbolRange(sym1,sym2)");
      const size_t dim = index(sym2,i) - index(sym1,i);
      // Next "if" checks whether dim+1 would overflow:
      if (dim == numeric_limits<size_t>::max())
        CoCoA_ERROR(ERR::ArgTooBig, "SymbolRange(sym1,sym2)");
      // Next "if" checks whether product Range*(dim+1) would overflow.
      if (MAX/(dim+1) < RangeSize)
        CoCoA_ERROR(ERR::ArgTooBig, "SymbolRange(sym1,sym2)");
      RangeSize *= dim+1; // cannot overflow thanks to checks above.
    }
    ans.reserve(RangeSize);
    vector<long> indices1(n);
    vector<long> indices2(n);
    vector<long> workspace(n);
    for (size_t i=0; i < n; ++i)
    {
      indices1[i] = index(sym1,i);
      indices2[i] = index(sym2,i);
    }
    FillRange(ans, head(sym1), indices1, indices2, workspace, 0);
    return ans;
  }

  std::vector<symbol> SymbolRange(const std::string& head, MachineInteger lo, MachineInteger hi)
  {
    if (!IsSignedLong(lo) || !IsSignedLong(hi))
      CoCoA_ERROR(ERR::BadSymbolIndex, "SymbolRange(hd,ind1,ind2)");
    return SymbolRange(symbol(head,lo), symbol(head,hi));
  }


  long index(const symbol& sym, std::size_t n)
  {
    if (n >= NumIndices(sym)) CoCoA_ERROR(ERR::BadSymbolIndex, "index(symbol,n)");
    return sym.myIndices[n];
  }


  bool AreDistinct(const std::vector<symbol>& syms)
  {
    vector<symbol> s = syms;
    sort(s.begin(), s.end());
    const size_t NumSyms = s.size();
    for (size_t i=0; i < NumSyms-1; ++i)
      if (s[i] == s[i+1]) return false;
    return true;
  }


  bool AreArityConsistent(const std::vector<symbol>& syms)
  {
    vector<symbol> s = syms;
    sort(s.begin(), s.end());
    const size_t NumSyms = s.size();
    for (size_t i=0; i < NumSyms-1; ++i)
      if (head(s[i]) == head(s[i+1]) && NumIndices(s[i]) != NumIndices(s[i+1]))
        return false;
    return true;
  }


  std::ostream& operator<<(std::ostream& out, const symbol& sym)
  {
    sym.myOutputSelf(out);
    return out;
  }


  std::istream& operator>>(std::istream& in, symbol& sym)
  {
    sym.myInput(in);
    return in;
  }


  // The four convenience functions.
  std::vector<symbol> symbols(const std::string& head1)
  {
    vector<symbol> IndetNames;
    IndetNames.push_back(symbol(head1));
    return IndetNames;
  }

  std::vector<symbol> symbols(const std::string& head1, const std::string& head2)
  {
    vector<symbol> IndetNames;
    IndetNames.push_back(symbol(head1));
    IndetNames.push_back(symbol(head2));
    return IndetNames;
  }

  std::vector<symbol> symbols(const std::string& head1, const std::string& head2, const std::string& head3)
  {
    vector<symbol> IndetNames;
    IndetNames.push_back(symbol(head1));
    IndetNames.push_back(symbol(head2));
    IndetNames.push_back(symbol(head3));
    return IndetNames;
  }

  std::vector<symbol> symbols(const std::string& head1, const std::string& head2, const std::string& head3, const std::string& head4)
  {
    vector<symbol> IndetNames;
    IndetNames.push_back(symbol(head1));
    IndetNames.push_back(symbol(head2));
    IndetNames.push_back(symbol(head3));
    IndetNames.push_back(symbol(head4));
    return IndetNames;
  }


} // end of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/symbol.C,v 1.11 2008/12/17 10:49:22 abbott Exp $
// $Log: symbol.C,v $
// Revision 1.11  2008/12/17 10:49:22  abbott
// Change loop control variable from int into long (as it should always have been).
//
// Revision 1.10  2008/12/12 11:29:47  abbott
// Fixed a bug in SymbolRange.  Added example and test for symbols.
//
// Revision 1.9  2008/12/11 20:12:32  abbott
// Now using MachineInteger type in ctors to specify the values of symbol indexes
// (but we check that the values fit into a long).  Corrected two bugs in
// SymbolRange (an undetected overflow and an off-by-1 error).
//
// Revision 1.8  2008/12/11 10:54:01  abbott
// Now indices are [long] values rather than [int].  The extra range will probably
// never be useful, but I prefer to offer it all the same.
//
// Revision 1.7  2008/11/21 21:16:34  abbott
// Added ctor for symbols with 2 indices.
//
// Revision 1.6  2008/04/21 12:32:54  abbott
// Corrected size_t into std::size_t in several header files; in some cases,
// replaced size_t with MachineInteger (with consequent changes to impl files).
//
// Revision 1.5  2007/10/30 17:14:06  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.4  2007/09/25 16:32:30  abbott
// Several minor changes to silence gcc-4.3:
//    more #includes,
//    and fixed a template problemm in RegisterServerOps.C
//
// Revision 1.3  2007/05/31 14:50:09  bigatti
// -- added AreDistinct and AreArityConsistent
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/08 16:55:06  cocoa
// Changed name of "range" function to "SymbolRange".
//
// Revision 1.5  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.4  2006/10/27 19:09:45  cocoa
// Replaced some member functions of CoCoA::symbol by friend functions.
// Removed some include dependency on symbol.H
//
// Revision 1.3  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.2  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.2  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1  2005/07/08 15:09:28  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//
