//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include <cctype>
#include <boost/foreach.hpp>
#include "AST.H"

namespace CoCoA {
namespace AST {

using namespace std;
using namespace boost;
using namespace CoCoA::Lexer;

bool Expression::isLeftValue() const {
	return false;
}

bool IdOrVar::isLeftValue() const {
	return true;
}

bool FieldAccessExpression::isLeftValue() const {
	return targetExp->isLeftValue();
}

bool IndexedAccessExpression::isLeftValue() const {
	return targetExp->isLeftValue();
}

namespace {
	void indent(std::ostream &out, int indentationLevel) {
		while (--indentationLevel>=0)
			out << "  ";
	}
}

void ParsedObject::dumpAsTree(ostream &os) const {
	dumpAsTree(os, 0);
}

ostream &operator<<(ostream &out, const ParsedObject * const parsedObject) {
	return parsedObject->dumpAsString(out);
}

ostream &Identifier::dumpAsString(ostream &out) const {
	return out << lexeme;
}

void Identifier::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << lexeme << '\n';
}

ostream &FullyQualifiedIdentifier::dumpAsString(ostream &out) const {
	return out << lexemePkgname << '.' << lexemeId;
}

void FullyQualifiedIdentifier::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << lexemePkgname << '.' << lexemeId << '\n';
}

ostream &EvalStatement::dumpAsString(ostream &out) const {
	return out << exp << "; ";
}

void EvalStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "; (eval-statement)\n";
	exp->dumpAsTree(out, indentationLevel+1);
}

ostream &ReturnStatement::dumpAsString(ostream &out) const {
	return out << "Return ";
	if (exp)
		out << exp;
	out << "; ";
}

void ReturnStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Return\n";
	if (exp)
		exp->dumpAsTree(out, indentationLevel+1);
}

ostream &SourceStatement::dumpAsString(ostream &out) const {
	return out << (ttype==TT_SOURCE ? "Source ":"<< ") << exp << "; ";
}

void SourceStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << (ttype==TT_SOURCE ? "Source\n":"<<\n");
	exp->dumpAsTree(out, indentationLevel+1);
}

ostream &DescribeStatement::dumpAsString(ostream &out) const {
	return out << "Describe " << exp << "; ";
}

void DescribeStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Describe\n";
	exp->dumpAsTree(out, indentationLevel+1);
}

ostream &UseStatement::dumpAsString(ostream &out) const {
	out << "Use ";
	if (idOrVar)
		out << idOrVar;
	if (ringDefinition) {
		if (idOrVar)
			out << " ::= ";
		out << ringDefinition;
	}
	return out << "; ";
}

void UseStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Use\n";
	if (idOrVar)
		idOrVar->dumpAsTree(out, indentationLevel+1);
	if (ringDefinition) {
		indent(out, indentationLevel);
		out << "::=\n";
		ringDefinition->dumpAsTree(out, indentationLevel+1);
	}
}

ostream &SetStatement::dumpAsString(ostream &out) const {
	out << "Set " << idOrVar;
	if (exp)
		out << " := " << exp;
	return out << "; ";
}

void SetStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Set\n";
	idOrVar->dumpAsTree(out, indentationLevel+1);
	if (exp) {
		indent(out, indentationLevel);
		out << ":=\n";
		exp->dumpAsTree(out, indentationLevel+1);
	}
}

ostream &UnsetStatement::dumpAsString(ostream &out) const {
	return out << "Unset " << idOrVar << "; ";
}

void UnsetStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Unset\n";
	idOrVar->dumpAsTree(out, indentationLevel+1);
}

ostream &BinaryExpression::dumpAsString(ostream &out) const {
	if (isalpha(op[0]))
		return out << leftExp << ' ' << op << ' ' << rightExp;
	return out << leftExp << op << rightExp;
}

void BinaryExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << op << "\n";
	leftExp->dumpAsTree(out, ++indentationLevel);
	rightExp->dumpAsTree(out, indentationLevel);
}

ostream &UnaryExpression::dumpAsString(ostream &out) const {
	if (isalpha(op[0]))
		return out << op << ' ' << exp;
	return out << op << exp;
}

void UnaryExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << op << "\n";
	exp->dumpAsTree(out, indentationLevel+1);
}

ostream &LiteralExpressionTEMPORARY_FIXME::dumpAsString(ostream &out) const {
	return out << beginSourcePosition.stringTo(endSourcePosition);
}

void LiteralExpressionTEMPORARY_FIXME::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << beginSourcePosition.stringTo(endSourcePosition) << '\n';
}

ostream &EmptyListExpression::dumpAsString(ostream &out) const {
	return out << "[]";
}

void EmptyListExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "[]\n";
}

ostream &AssignStatement::dumpAsString(ostream &out) const {
	return out << leftExp << " := " << rightExp << "; ";
}

void AssignStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << ":=\n";
	leftExp->dumpAsTree(out, ++indentationLevel);
	rightExp->dumpAsTree(out, indentationLevel);
}

ostream &RingAssignStatement::dumpAsString(ostream &out) const {
	return out << leftValue << " ::= " << ringDef << "; ";
}

void RingAssignStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "::=\n";
	leftValue->dumpAsTree(out, ++indentationLevel);
	ringDef->dumpAsTree(out, indentationLevel);
}

ostream &ParenthesizedExp::dumpAsString(ostream &out) const {
	return out << "(" << innerExp << ")";
}

void ParenthesizedExp::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "(\n";
	innerExp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << ")\n";
}

ostream &IsDefinedExp::dumpAsString(ostream &out) const {
	return out << "IsDefined(" << exp << ")";
}

void IsDefinedExp::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "IsDefined(\n";
	exp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << ")\n";
}

ostream &TimeStatement::dumpAsString(ostream &out) const {
	return out << "Time " << stmt;
}

void TimeStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Time\n";
	stmt->dumpAsTree(out, indentationLevel+1);
}

ostream &Var::dumpAsString(ostream &out) const {
	return out << "Var(" << exp << ")";
}

void Var::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Var(\n";
	exp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << ")\n";
}

ostream &BackwardCompatibleExp::dumpAsString(ostream &out) const {
	return out << "${" << innerExp << "}$";
}

void BackwardCompatibleExp::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "${\n";
	innerExp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "}$\n";
}

namespace {
	template <typename T>
	void outputExpList(ostream &out, const vector<T> &list, const string &separator) {
		bool first=true;
		BOOST_FOREACH(T e, list) {
			if (first)
				first=false;
			else
				out << separator;
			out << e;
		}
	}

	template <typename T>
	void outputCommaSeparatedExpList(ostream &out, const vector<T> &list) {
		outputExpList(out, list, ", ");
	}
}

ostream &CartesianProductExpression::dumpAsString(ostream &out) const {
	outputExpList(out, operands, "><");
	return out;
}

void CartesianProductExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << "><\n";
	BOOST_FOREACH(intrusive_ptr<const Expression> e, operands)
		e->dumpAsTree(out, indentationLevel);
}

ostream &IdInExpSuchThatExp::dumpAsString(ostream &out) const {
	return out << "[" << identifier << " In " << exp1 << " | " << exp2 << "]";
}

void IdInExpSuchThatExp::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << "[ " << identifier << " In ... | ...]\n";
	exp1->dumpAsTree(out, indentationLevel);
	exp2->dumpAsTree(out, indentationLevel);
}

ostream &ExpSuchThatIdInExpAndExp::dumpAsString(ostream &out) const {
	out << "[" << exp1 << " | " << identifier << " In " << exp2;
	if (optionalExp3)
		out << " And " << optionalExp3;
	return out << "]";
}

void ExpSuchThatIdInExpAndExp::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << "[ ... | " << identifier << " In ... ";
	if (optionalExp3)
		out << "And ... ";
	out << "]\n";
	exp1->dumpAsTree(out, indentationLevel);
	exp2->dumpAsTree(out, indentationLevel);
	if (optionalExp3)
		optionalExp3->dumpAsTree(out, indentationLevel);
}

ostream &NonEmptyListExpression::dumpAsString(ostream &out) const {
	out << "[";
	outputCommaSeparatedExpList(out, exps);
	return out << "]";
}

void NonEmptyListExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << "[..., ...]\n";
	BOOST_FOREACH(intrusive_ptr<const Expression> e, exps)
		e->dumpAsTree(out, indentationLevel);
}

ostream &IndeterminateDeclaration::dumpAsString(ostream &out) const {
	out << idOrVar;
	if (!ranges.empty()) {
		out << "[";
		bool first = true;
		for(vector<pair <intrusive_ptr<const Expression>, intrusive_ptr<const Expression> > >::const_iterator it=ranges.begin(); it!=ranges.end(); ++it) {
			if (first)
				first = false;
			else
				out << ", ";
			out << it->first << ".." << it->second;
		}
		return out << "]";
	}
	return out;
}

void IndeterminateDeclaration::dumpAsTree(ostream &out, int indentationLevel) const {
	idOrVar->dumpAsTree(out, indentationLevel);
	if (ranges.empty())
		return;
	indent(out, ++indentationLevel);
	out << "[]\n";
	++indentationLevel;
	for(vector<pair <intrusive_ptr<const Expression>, intrusive_ptr<const Expression> > >::const_iterator it=ranges.begin(); it!=ranges.end(); ++it) {
		indent(out, indentationLevel);
		out << ".." << "\n";
		it->first->dumpAsTree(out, indentationLevel+1);
		it->second->dumpAsTree(out, indentationLevel+1);
	}
}

ostream &RingDefinition::dumpAsString(ostream &out) const {
	out << identifier;
	if (optionalExp)
		out << "/(" << optionalExp << ")";
	if (indeterminates.size()) {
		out << "[";
		outputCommaSeparatedExpList(out, indeterminates);
		out << "]";
	}
	return out;
}

void RingDefinition::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << identifier << '\n';
	if (optionalExp) {
		indent(out, indentationLevel);
		out << "/()\n";
		optionalExp->dumpAsTree(out, indentationLevel+1);
	}
	if (indeterminates.size()) {
		indent(out, indentationLevel++);
		out << "[]\n";
		BOOST_FOREACH(intrusive_ptr<const IndeterminateDeclaration> id, indeterminates)
			id->dumpAsTree(out, indentationLevel);
	}
}

ostream &FieldAccessExpression::dumpAsString(ostream &out) const {
	return out << targetExp << '.' << name;
}

void FieldAccessExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << ".\n";
	targetExp->dumpAsTree(out, indentationLevel);
	name->dumpAsTree(out, indentationLevel);
}

ostream &IndexedAccessExpression::dumpAsString(ostream &out) const {
	out << targetExp << "[";
	outputCommaSeparatedExpList(out, indexes);
	return out << "]";
}

void IndexedAccessExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << "[]\n";
	targetExp->dumpAsTree(out, indentationLevel);
	BOOST_FOREACH(intrusive_ptr<const Expression> e, indexes)
		e->dumpAsTree(out, indentationLevel);
}

ostream &InvocationExpression::dumpAsString(ostream &out) const {
	out << targetExp << "(";
	if (ellipsis)
		out << "...";
	else
		outputCommaSeparatedExpList(out, args);
	return out << ")";
}

void InvocationExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel++);
	out << "()\n";
	targetExp->dumpAsTree(out, indentationLevel);
	if (ellipsis) {
		indent(out, indentationLevel);
		out << "...";
	} else
		BOOST_FOREACH(intrusive_ptr<const Expression> e, args)
			e->dumpAsTree(out, indentationLevel);
}

ostream &PrintStatement::dumpAsString(ostream &out) const {
	out << (ttype==TT_PRINT ? "Print " : "PrintLn ");
	outputCommaSeparatedExpList(out, exps);
	if (onExp)
		out << " On " << onExp;
	return out << "; ";
}

void PrintStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << (ttype==TT_PRINT ? "Print\n" : "PrintLn\n");
	BOOST_FOREACH(intrusive_ptr<const Expression> e, exps)
		e->dumpAsTree(out, indentationLevel+1);
	if (onExp) {
		indent(out, indentationLevel);
		out << "On\n";
		onExp->dumpAsTree(out, indentationLevel+1);
	}
}

ostream &ClearStatement::dumpAsString(ostream &out) const {
	out << "Clear ";
	outputCommaSeparatedExpList(out, idOrVarList);
	return out << "; ";
}

void ClearStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Clear\n";
	BOOST_FOREACH(intrusive_ptr<const IdOrVar> idOrVar, idOrVarList)
	idOrVar->dumpAsTree(out, indentationLevel+1);
}

ostream &DeleteStatement::dumpAsString(ostream &out) const {
	out << "Delete ";
	outputCommaSeparatedExpList(out, idOrVarList);
	return out << "; ";
}

void DeleteStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Delete\n";
	BOOST_FOREACH(intrusive_ptr<const IdOrVar> idOrVar, idOrVarList)
	idOrVar->dumpAsTree(out, indentationLevel+1);
}

ostream &DestroyStatement::dumpAsString(ostream &out) const {
	out << "Destroy ";
	outputCommaSeparatedExpList(out, idOrVarList);
	return out << "; ";
}

void DestroyStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Destroy\n";
	BOOST_FOREACH(intrusive_ptr<const IdOrVar> idOrVar, idOrVarList)
	idOrVar->dumpAsTree(out, indentationLevel+1);
}

ostream &AliasStatement::dumpAsString(ostream &out) const {
	out << "Alias ";
	bool first=true;
	BOOST_FOREACH(const Binding &b, bindings) {
		if (first)
			first=false;
		else
			out << ", ";
		out << b.idOrVar << " := " << b.packageName;
	}
	if (thereIsIn) {
		out << "In ";
		BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
			out << s;
		return out << "EndAlias ";
	} else
		return out << "; ";
}

void AliasStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Alias\n";
	BOOST_FOREACH(const Binding &b, bindings) {
		indent(out, indentationLevel+1);
		out << b.idOrVar << " := " << b.packageName << "\n";
	}
	if (thereIsIn) {
		indent(out, indentationLevel);
		out << "In\n";
		BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
			s->dumpAsTree(out, indentationLevel+1);
		indent(out, indentationLevel);
		out << "EndAlias\n";
	}
}

ostream &BlockStatement::dumpAsString(ostream &out) const {
	out << "Block ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	return out << "EndBlock ";
}

void BlockStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Block\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndBlock\n";
}

namespace {
	void outParams(bool thereIsEllipsis, const vector<Param> &params, ostream &out) {
		if (thereIsEllipsis) {
			out << "...";
			return;
		}
		bool first=true;
		BOOST_FOREACH(const Param &p, params) {
			if (first)
				first=false;
			else
				out << ", ";
			out << (p.byRef ? "Var ":"") << p.name;
		}
	}
}

ostream &DefineStatement::dumpAsString(ostream &out) const {
	out << "Define " << name << "(";
	outParams(thereIsEllipsis, params, out);
	out << ") ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	return out << "EndDefine ";
}

void DefineStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Define " << name << '(';
	outParams(thereIsEllipsis, params, out);
	out << ")\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndDefine\n";
}

ostream &RecordExpression::dumpAsString(ostream &out) const {
	out << "Record[";
	bool first=true;
	BOOST_FOREACH(const RecordField &f, fields) {
		if (first)
			first=false;
		else
			out << ", ";
		out << f.name << " := " << f.initExp;
	}
	return out << "]";
}

void RecordExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Record[\n";
	BOOST_FOREACH(const RecordField &f, fields) {
		indent(out, indentationLevel);
		out << ":=\n";
		f.name->dumpAsTree(out, indentationLevel+1);
		f.initExp->dumpAsTree(out, indentationLevel+1);
	}
	indent(out, indentationLevel);
	out << "]\n";
}

ostream &PackageStatement::dumpAsString(ostream &out) const {
	out << "Package " << name << ' ';
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	return out << "EndPackage ";
}

void PackageStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Package " << name << '\n';
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndPackage\n";
}

ostream &HelpStatement::dumpAsString(ostream &out) const {
	return out << lexeme << '\n';
}

void HelpStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << lexeme << '\n';
}

ostream &CatchStatement::dumpAsString(ostream &out) const {
	out << "Catch ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	if (leftValue)
		out << "In " << leftValue << ' ';
	return out << "EndCatch ";
}

void CatchStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Catch\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	if (leftValue) {
		indent(out, indentationLevel);
		out << "In\n";
		leftValue->dumpAsTree(out, indentationLevel+1);
	}
	indent(out, indentationLevel);
	out << "EndCatch\n";
}

namespace {
	char operandToChar(TokenType tt) {
		switch (tt) {
		case TT_PLUS:
			return '+';
		case TT_MINUS:
			return '-';
		case TT_STAR:
			return '*';
		case TT_SLASH:
			return '/';
		case TT_MOD:
			return '%';
		case TT_COLON:
			return ':';
		default:
			assert(false);
			return '?';
		}
	}

	ostream &dumpSequenceOfOpsAsString(ostream &out, vector<TokenType>::const_iterator opIt, vector<intrusive_ptr<const Expression> >::const_iterator expIt, vector<intrusive_ptr<const Expression> >::const_iterator expEnd) {
		for(bool isFirst=true; expIt!=expEnd; ++expIt) {
			if (isFirst)
				isFirst = false;
			else
				out << operandToChar(*opIt++);
			out << *expIt;
		}
		return out;
	}

	void dumpSequenceOfOpsAsTree(ostream &out, int indentationLevel, vector<TokenType>::const_iterator opIt, vector<intrusive_ptr<const Expression> >::const_iterator expIt, vector<intrusive_ptr<const Expression> >::const_iterator expEnd) {
		for(bool isFirst=true; expIt!=expEnd; ++expIt) {
			if (isFirst) {
				isFirst = false;
			} else {
				indent(out, indentationLevel);
				out << operandToChar(*opIt++) << '\n';
			}
			(*expIt)->dumpAsTree(out, indentationLevel+1);
		}
	}
}

ostream &SummationExpression::dumpAsString(ostream &out) const {
	return dumpSequenceOfOpsAsString(out, operators.begin(), operands.begin(), operands.end());
}

void SummationExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "summation\n";
	dumpSequenceOfOpsAsTree(out, ++indentationLevel, operators.begin(), operands.begin(), operands.end());
}

ostream &MultiplicationExpression::dumpAsString(ostream &out) const {
	return dumpSequenceOfOpsAsString(out, operators.begin(), operands.begin(), operands.end());
}

void MultiplicationExpression::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "multiplication\n";
	dumpSequenceOfOpsAsTree(out, ++indentationLevel, operators.begin(), operands.begin(), operands.end());
}

ostream &ForStatement::dumpAsString(ostream &out) const {
	out << "For " << ident.lexeme() << ":=" << beginExp << " To " << endExp;
	if (stepExp)
		out << " Step " << stepExp;
	out << " Do ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	out << "EndFor ";
	return out;
}

void ForStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "For " << ident.lexeme() << ":=\n";
	beginExp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "To\n";
	endExp->dumpAsTree(out, indentationLevel+1);
	if (stepExp) {
		indent(out, indentationLevel);
		out << "Step\n";
		stepExp->dumpAsTree(out, indentationLevel+1);
	}
	indent(out, indentationLevel);
	out << "Do\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndFor\n";
}

ostream &UsingStatement::dumpAsString(ostream &out) const {
	out << "Using " << idOrVar << " Do ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	out << "EndUsing ";
	return out;
}

void UsingStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Using " << idOrVar << " Do\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndUsing\n";
}

ostream &ForeachStatement::dumpAsString(ostream &out) const {
	out << "Foreach " << ident.lexeme() << " In " << inExp << " Do ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	out << "EndForeach ";
	return out;
}

void ForeachStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Foreach " << ident.lexeme() << " In\n";
	inExp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "Do\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndForeach\n";
}

ostream &WhileStatement::dumpAsString(ostream &out) const {
	out << "While " << exp << " Do ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	out << "EndWhile ";
	return out;
}

void WhileStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "While\n";
	exp->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "Do\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "EndWhile\n";
}

ostream &RepeatUntilStatement::dumpAsString(ostream &out) const {
	out << "Repeat ";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		out << s;
	out << "Until " << exp << "; ";
	return out;
}

void RepeatUntilStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Repeat\n";
	BOOST_FOREACH(intrusive_ptr<const Statement> s, statements)
		s->dumpAsTree(out, indentationLevel+1);
	indent(out, indentationLevel);
	out << "Until\n";
	exp->dumpAsTree(out, indentationLevel+1);
}

ostream &EmptyStatement::dumpAsString(ostream &out) const {
	return out << "; ";
}

void EmptyStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "; (empty-statement)\n";
}

ostream &IfStatement::dumpAsString(ostream &out) const {
	bool first = true;
	BOOST_FOREACH(IfBranch ib, branches) {
		if (ib.optExp) {
			if (first) {
				first = false;
				out << "If ";
			} else
				out << "Elif ";
			out << ib.optExp << " Then ";
		} else
			out << "Else ";
		BOOST_FOREACH(intrusive_ptr<const Statement> s, ib.statements)
			out << s;
	}
	return out << "EndIf ";
}

void IfStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	bool first = true;
	BOOST_FOREACH(IfBranch ib, branches) {
		indent(out, indentationLevel);
		if (ib.optExp) {
			if (first) {
				first = false;
				out << "If\n";
			} else
				out << "Elif\n";
			ib.optExp->dumpAsTree(out, indentationLevel+1);
			indent(out, indentationLevel);
			out << "Then\n";
		} else
			out << "Else\n";
		BOOST_FOREACH(intrusive_ptr<const Statement> s, ib.statements)
			s->dumpAsTree(out, indentationLevel+1);
	}
	indent(out, indentationLevel);
	out << "EndIf\n";
}

ostream &BreakStatement::dumpAsString(ostream &out) const {
	return out << "Break; ";
}

void BreakStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Break\n";
}

ostream &ContinueStatement::dumpAsString(ostream &out) const {
	return out << "Continue; ";
}

void ContinueStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Continue\n";
}

ostream &SkipStatement::dumpAsString(ostream &out) const {
	return out << "Skip; ";
}

void SkipStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << "Skip\n";
}

ostream &CiaoOrQuitStatement::dumpAsString(ostream &out) const {
	return out << (ttype==TT_CIAO?"Ciao; ":"Quit; ");
}

void CiaoOrQuitStatement::dumpAsTree(ostream &out, int indentationLevel) const {
	indent(out, indentationLevel);
	out << (ttype==TT_CIAO?"Ciao\n":"Quit\n");
}

} // namespace AST
} // namespace CoCoA
