//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#ifndef AST_H_
#define AST_H_

#include <ostream>
#include <string>
#include <vector>
#include <utility>
#include <boost/foreach.hpp>

#include "Lexer.H"

namespace CoCoA {
namespace AST {

class ParsedObject : public CoCoA::Lexer::ReferenceCountedObject {
public:
	const CoCoA::Lexer::CharPointer beginSourcePosition, endSourcePosition;
	ParsedObject(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		beginSourcePosition(_beginSourcePosition),
		endSourcePosition(_endSourcePosition)
	{
		assert(*beginSourcePosition);
		assert(*endSourcePosition);
	}
	virtual std::ostream &dumpAsString(std::ostream &out) const = 0;
	virtual void dumpAsTree(std::ostream &out, int indentationLevel) const = 0;
	void dumpAsTree(std::ostream &os) const;
	std::string toSourceString() const { return beginSourcePosition.stringTo(endSourcePosition); }
	virtual ~ParsedObject() {}
};

std::ostream &operator<<(std::ostream &out, const ParsedObject * const parsedObject);

class Expression : public ParsedObject {
public:
	Expression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		ParsedObject(_beginSourcePosition, _endSourcePosition)
	{}
	virtual bool isLeftValue() const;
};

class BinaryExpression : public Expression {
public:
	const boost::intrusive_ptr<const Expression> leftExp, rightExp;
	const std::string op;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	BinaryExpression(const boost::intrusive_ptr<const Expression> _leftExp, const std::string &_op, const boost::intrusive_ptr<const Expression> _rightExp) :
		Expression(_leftExp->beginSourcePosition, _rightExp->endSourcePosition),
		leftExp(_leftExp),
		rightExp(_rightExp),
		op(_op)
	{
		assert(leftExp);
		assert(rightExp);
		assert(op.length()>0);
	}
};

class UnaryExpression : public Expression {
public:
	const boost::intrusive_ptr<const Expression> exp;
	const std::string op;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	UnaryExpression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const std::string &_op, const boost::intrusive_ptr<const Expression> _exp) :
		Expression(_beginSourcePosition, _exp->endSourcePosition),
		exp(_exp),
		op(_op)
	{
		assert(exp);
		assert(op.length()>0);
	}
};

class LiteralExpressionTEMPORARY_FIXME : public Expression {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	LiteralExpressionTEMPORARY_FIXME(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		Expression(_beginSourcePosition, _endSourcePosition)
	{}
};

class EmptyListExpression : public Expression {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	EmptyListExpression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		Expression(_beginSourcePosition, _endSourcePosition)
	{}
};

class IdInExpSuchThatExp : public Expression {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	const std::string identifier;
	const boost::intrusive_ptr<const Expression> exp1, exp2;
	IdInExpSuchThatExp(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition, const std::string &_identifier, const boost::intrusive_ptr<const Expression> _exp1, const boost::intrusive_ptr<const Expression> _exp2) :
		Expression(_beginSourcePosition, _endSourcePosition),
		identifier(_identifier),
		exp1(_exp1),
		exp2(_exp2)
	{
		assert(identifier.length()>0);
		assert(exp1);
		assert(exp2);
	}
};

class ExpSuchThatIdInExpAndExp : public Expression {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	const std::string identifier;
	const boost::intrusive_ptr<const Expression> exp1, exp2, optionalExp3;
	ExpSuchThatIdInExpAndExp(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition, const boost::intrusive_ptr<const Expression> _exp1, const std::string &_identifier, const boost::intrusive_ptr<const Expression> _exp2, const boost::intrusive_ptr<const Expression> _optionalExp3) :
		Expression(_beginSourcePosition, _endSourcePosition),
		identifier(_identifier),
		exp1(_exp1),
		exp2(_exp2),
		optionalExp3(_optionalExp3)
	{
		assert(identifier.length()>0);
		assert(exp1);
		assert(exp2);
	}
};

class NonEmptyListExpression : public Expression {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	const std::vector<boost::intrusive_ptr<const Expression> > exps;
	NonEmptyListExpression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition, const std::vector<boost::intrusive_ptr<const Expression> > &_exps) :
		Expression(_beginSourcePosition, _endSourcePosition),
		exps(_exps)
	{
		assert(exps.size()>0);
	}
};

class UnaryMinusExpression : public UnaryExpression {
public:
	UnaryMinusExpression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const boost::intrusive_ptr<const Expression> _exp) :
		UnaryExpression(_beginSourcePosition, "-", _exp)
	{
	}
};

class NotExpression : public UnaryExpression {
public:
	NotExpression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const boost::intrusive_ptr<const Expression> _exp) :
		UnaryExpression(_beginSourcePosition, "Not", _exp)
	{
	}
};

class UnaryPlusExpression : public UnaryExpression {
public:
	UnaryPlusExpression(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const boost::intrusive_ptr<const Expression> _exp) :
		UnaryExpression(_beginSourcePosition, "+", _exp)
	{
	}
};

class IndexedAccessExpression : public Expression {
public:
	const boost::intrusive_ptr<const Expression> targetExp;
	const std::vector<boost::intrusive_ptr<const Expression> > indexes;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	bool isLeftValue() const;
	IndexedAccessExpression(const boost::intrusive_ptr<const Expression> _targetExp, const std::vector<boost::intrusive_ptr<const Expression> > &_indexes, const CoCoA::Lexer::CharPointer &closedBracketPosition) :
		Expression(_targetExp->beginSourcePosition, closedBracketPosition),
		targetExp(_targetExp),
		indexes(_indexes)
	{
		assert(targetExp);
		assert(indexes.size()>0);
	}
};

class InvocationExpression : public Expression {
public:
	const boost::intrusive_ptr<const Expression> targetExp;
	const bool ellipsis;
	const std::vector<boost::intrusive_ptr<const Expression> > args;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	InvocationExpression(const boost::intrusive_ptr<const Expression> _targetExp, bool _ellipsis, const std::vector<boost::intrusive_ptr<const Expression> > &_args, const CoCoA::Lexer::CharPointer &closedParenthesisPosition) :
		Expression(_targetExp->beginSourcePosition, closedParenthesisPosition),
		targetExp(_targetExp),
		ellipsis(_ellipsis),
		args(_args)
	{
		assert(targetExp);
	}
};

class CartesianProductExpression : public Expression {
public:
	const std::vector<boost::intrusive_ptr<const Expression> > operands;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	CartesianProductExpression(const std::vector<boost::intrusive_ptr<const Expression> > &_operands) :
		Expression(_operands.front()->beginSourcePosition, _operands.back()->endSourcePosition),
		operands(_operands)
	{
		assert(operands.size()>0);
	}
};

class SummationExpression : public Expression {
public:
	const std::vector<boost::intrusive_ptr<const Expression> > operands;
	const std::vector<CoCoA::Lexer::TokenType> operators;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	SummationExpression(const std::vector<boost::intrusive_ptr<const Expression> > &_operands, const std::vector<CoCoA::Lexer::TokenType> &_operators) :
		Expression(_operands.front()->beginSourcePosition, _operands.back()->endSourcePosition),
		operands(_operands),
		operators(_operators)
	{
		assert(operands.size()>2);
		assert(operators.size()==(operands.size()-1));
		BOOST_FOREACH(CoCoA::Lexer::TokenType tt, operators) {
			assert(tt==CoCoA::Lexer::TT_PLUS || tt==CoCoA::Lexer::TT_MINUS);
		}
	}
};

class MultiplicationExpression : public Expression {
public:
	const std::vector<boost::intrusive_ptr<const Expression> > operands;
	const std::vector<CoCoA::Lexer::TokenType> operators;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	MultiplicationExpression(const std::vector<boost::intrusive_ptr<const Expression> > &_operands, const std::vector<CoCoA::Lexer::TokenType> &_operators) :
		Expression(_operands.front()->beginSourcePosition, _operands.back()->endSourcePosition),
		operands(_operands),
		operators(_operators)
	{
		assert(operands.size()>2);
		assert(operators.size()==(operands.size()-1));
		BOOST_FOREACH(CoCoA::Lexer::TokenType tt, operators) {
			assert(tt==CoCoA::Lexer::TT_STAR || tt==CoCoA::Lexer::TT_SLASH ||
				   tt==CoCoA::Lexer::TT_MOD || tt==CoCoA::Lexer::TT_COLON);
		}
	}
};

class OrExpression : public BinaryExpression {
public:
	OrExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "Or", _rightExp)
	{}
};

class PowerExpression : public BinaryExpression {
public:
	PowerExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "^", _rightExp)
	{}
};

class DotDotExpression : public BinaryExpression {
public:
	DotDotExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "..", _rightExp)
	{}
};

class SumExpression : public BinaryExpression {
public:
	SumExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "+", _rightExp)
	{}
};

class SubtractionExpression : public BinaryExpression {
public:
	SubtractionExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "-", _rightExp)
	{}
};

class ProductExpression : public BinaryExpression {
public:
	ProductExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "*", _rightExp)
	{}
};

class DivisionExpression : public BinaryExpression {
public:
	DivisionExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "/", _rightExp)
	{}
};

class ColonExpression : public BinaryExpression {
public:
	ColonExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, ":", _rightExp)
	{}
};

class ModuloExpression : public BinaryExpression {
public:
	ModuloExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "%", _rightExp)
	{}
};

class AndExpression : public BinaryExpression {
public:
	AndExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "And", _rightExp)
	{}
};

class EqualExpression : public BinaryExpression {
public:
	EqualExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "=", _rightExp)
	{}
};

class NotEqualExpression : public BinaryExpression {
public:
	NotEqualExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "<>", _rightExp)
	{}
};

class IsInExpression : public BinaryExpression {
public:
	IsInExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "IsIn", _rightExp)
	{}
};

class LessThanExpression : public BinaryExpression {
public:
	LessThanExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "<", _rightExp)
	{}
};

class LessOrEqualExpression : public BinaryExpression {
public:
	LessOrEqualExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "<=", _rightExp)
	{}
};

class GreaterThanExpression : public BinaryExpression {
public:
	GreaterThanExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, ">", _rightExp)
	{}
};

class GreaterOrEqualExpression : public BinaryExpression {
public:
	GreaterOrEqualExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, ">=", _rightExp)
	{}
};

class ScopedExpression : public BinaryExpression {
public:
	ScopedExpression(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp) :
		BinaryExpression(_leftExp, "::", _rightExp)
	{}
};

class Statement : public ParsedObject {
public:
	Statement(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		ParsedObject(_beginSourcePosition, _endSourcePosition)
	{}
};

class EvalStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> exp;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	EvalStatement(const boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(_exp->beginSourcePosition, semicolonPosition),
		exp(_exp)
	{
		assert(exp);
	}
};

class ReturnStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> exp;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	ReturnStatement(const CoCoA::Lexer::CharPointer &returnPosition, const boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(returnPosition, semicolonPosition),
		exp(_exp)
	{
	}
};

class SourceStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> exp;
	const CoCoA::Lexer::TokenType ttype;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	SourceStatement(const CoCoA::Lexer::Token &tokSource, const boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(tokSource.getBegin(), semicolonPosition),
		exp(_exp),
		ttype(tokSource.getType())
	{
		assert(exp);
	}
};

class DescribeStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> exp;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	DescribeStatement(const CoCoA::Lexer::CharPointer &describePosition, const boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(describePosition, semicolonPosition),
		exp(_exp)
	{
		assert(exp);
	}
};

class EmptyStatement : public Statement {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	EmptyStatement(const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(semicolonPosition, semicolonPosition)
	{
	}
};

class BreakStatement : public Statement {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	BreakStatement(const CoCoA::Lexer::CharPointer &breakPosition, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(breakPosition, semicolonPosition)
	{
	}
};

class ContinueStatement : public Statement {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	ContinueStatement(const CoCoA::Lexer::CharPointer &continuePosition, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(continuePosition, semicolonPosition)
	{
	}
};

class SkipStatement : public Statement {
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	SkipStatement(const CoCoA::Lexer::CharPointer &skipPosition, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(skipPosition, semicolonPosition)
	{
	}
};

class CiaoOrQuitStatement : public Statement {
	const CoCoA::Lexer::TokenType ttype;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	CiaoOrQuitStatement(const CoCoA::Lexer::Token &tokCiaoOrQuit, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(tokCiaoOrQuit.getBegin(), semicolonPosition),
		ttype(tokCiaoOrQuit.getType())
	{
	}
};

class AssignStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> leftExp, rightExp;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	AssignStatement(const boost::intrusive_ptr<const Expression> _leftExp, const boost::intrusive_ptr<const Expression> _rightExp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(_leftExp->beginSourcePosition, semicolonPosition),
		leftExp(_leftExp),
		rightExp(_rightExp)
	{
		assert(leftExp);
		assert(rightExp);
	}
};

class IdOrVar : public Expression {
public:
	IdOrVar(const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		Expression(_beginSourcePosition, _endSourcePosition)
	{}
	bool isLeftValue() const;
};

class IndeterminateDeclaration : public ParsedObject {
	boost::intrusive_ptr<const IdOrVar> idOrVar;
	const std::vector<std::pair <boost::intrusive_ptr<const Expression>, boost::intrusive_ptr<const Expression> > > ranges;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	IndeterminateDeclaration(const boost::intrusive_ptr<const IdOrVar> _idOrVar, const std::vector<std::pair <boost::intrusive_ptr<const Expression>, boost::intrusive_ptr<const Expression> > > & _ranges, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		ParsedObject(_idOrVar->beginSourcePosition, _endSourcePosition),
		idOrVar(_idOrVar),
		ranges(_ranges)
	{}
};

class RingDefinition : public ParsedObject {
	const std::string identifier;
	const boost::intrusive_ptr<const Expression> optionalExp;
	const std::vector<boost::intrusive_ptr<const IndeterminateDeclaration> > indeterminates;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	RingDefinition(const CoCoA::Lexer::Token &tokId, const boost::intrusive_ptr<const Expression> _optionalExp, const std::vector<boost::intrusive_ptr<const IndeterminateDeclaration> > & _indeterminates, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		ParsedObject(tokId.getBegin(), _endSourcePosition),
		identifier(tokId.lexeme()),
		optionalExp(_optionalExp),
		indeterminates(_indeterminates)
	{
	}
};

class RingAssignStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> leftValue;
	const boost::intrusive_ptr<const RingDefinition> ringDef;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	RingAssignStatement(const boost::intrusive_ptr<const Expression> _leftValue, const boost::intrusive_ptr<const RingDefinition> _ringDef, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(_leftValue->beginSourcePosition, semicolonPosition),
		leftValue(_leftValue),
		ringDef(_ringDef)
	{
		assert(leftValue);
		assert(ringDef);
	}
};

class Identifier : public IdOrVar {
	const std::string lexeme;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	explicit Identifier(const CoCoA::Lexer::Token & tokenId) :
		IdOrVar(tokenId.getBegin(), tokenId.getEnd()),
		lexeme(tokenId.lexeme())
	{
		// note: tokenId may be of type TT_ORD or TT_ELIM (this allows to call those specials functions)
		assert(lexeme.length()>0);
	}
};

class FullyQualifiedIdentifier : public IdOrVar {
	const std::string lexemePkgname, lexemeId;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	explicit FullyQualifiedIdentifier(const CoCoA::Lexer::Token & tokenPkgname, const CoCoA::Lexer::Token & tokenId) :
		IdOrVar(tokenPkgname.getBegin(), tokenId.getEnd()),
		lexemePkgname(tokenPkgname.lexeme()),
		lexemeId(tokenId.lexeme())
	{
		assert(lexemePkgname.length()>0);
		assert(lexemeId.length()>0);
	}
};

class Var : public IdOrVar {
	const boost::intrusive_ptr<const Expression> exp;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	Var(boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		IdOrVar(_beginSourcePosition, _endSourcePosition),
		exp(_exp)
	{
		assert(exp);
	}
};

class FieldAccessExpression : public Expression {
public:
	const boost::intrusive_ptr<const Expression> targetExp;
	const boost::intrusive_ptr<const IdOrVar> name;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	bool isLeftValue() const;
	FieldAccessExpression(const boost::intrusive_ptr<const Expression> _targetExp, boost::intrusive_ptr<const IdOrVar> _name) :
		Expression(_targetExp->beginSourcePosition, _name->endSourcePosition),
		targetExp(_targetExp),
		name(_name)
	{
		assert(targetExp);
		assert(name);
	}
};

struct RecordField {
	boost::intrusive_ptr<const IdOrVar> name;
	boost::intrusive_ptr<const Expression> initExp;
};

class RecordExpression : public Expression {
public:
	const std::vector<RecordField> fields;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	RecordExpression(const CoCoA::Lexer::CharPointer &recordPosition, const std::vector<RecordField> &_fields, const CoCoA::Lexer::CharPointer &closedSquarePosition) :
		Expression(recordPosition, closedSquarePosition),
		fields(_fields)
	{
	}
};

class UseStatement : public Statement {
public:
	const boost::intrusive_ptr<const IdOrVar> idOrVar;
	const boost::intrusive_ptr<const RingDefinition> ringDefinition;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	UseStatement(const CoCoA::Lexer::CharPointer &usePosition, const boost::intrusive_ptr<const IdOrVar> _idOrVar, const boost::intrusive_ptr<const RingDefinition> _ringDefinition, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(usePosition, semicolonPosition),
		idOrVar(_idOrVar),
		ringDefinition(_ringDefinition)
	{
		assert(idOrVar || ringDefinition);
	}
};

class TimeStatement : public Statement {
	const boost::intrusive_ptr<const Statement> stmt;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	TimeStatement(boost::intrusive_ptr<const Statement> _stmt, const CoCoA::Lexer::CharPointer &_beginSourcePosition) :
		Statement(_beginSourcePosition, _stmt->endSourcePosition),
		stmt(_stmt)
	{
		assert(stmt);
	}
};

class SetStatement : public Statement {
public:
	const boost::intrusive_ptr<const IdOrVar> idOrVar;
	const boost::intrusive_ptr<const Expression> exp;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	SetStatement(const CoCoA::Lexer::CharPointer &setPosition, const boost::intrusive_ptr<const IdOrVar> _idOrVar, const boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(setPosition, semicolonPosition),
		idOrVar(_idOrVar),
		exp(_exp)
	{
		assert(idOrVar);
	}
};

class UnsetStatement : public Statement {
public:
	const boost::intrusive_ptr<const IdOrVar> idOrVar;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	UnsetStatement(const CoCoA::Lexer::CharPointer &setPosition, const boost::intrusive_ptr<const IdOrVar> _idOrVar, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(setPosition, semicolonPosition),
		idOrVar(_idOrVar)
	{
		assert(idOrVar);
	}
};

class ForStatement : public Statement {
public:
	const CoCoA::Lexer::Token ident;
	const boost::intrusive_ptr<const Expression> beginExp;
	const boost::intrusive_ptr<const Expression> endExp;
	const boost::intrusive_ptr<const Expression> stepExp;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	ForStatement(const CoCoA::Lexer::CharPointer &forPosition, const CoCoA::Lexer::Token &_ident, boost::intrusive_ptr<const Expression> _beginExp, boost::intrusive_ptr<const Expression> _endExp, boost::intrusive_ptr<const Expression> _stepExp, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endforPosition) :
		Statement(forPosition, endforPosition),
		ident(_ident),
		beginExp(_beginExp),
		endExp(_endExp),
		stepExp(_stepExp),
		statements(_statements)
	{
		assert(beginExp);
		assert(endExp);
	}
};

class UsingStatement : public Statement {
public:
	const boost::intrusive_ptr<const IdOrVar> idOrVar;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	UsingStatement(const CoCoA::Lexer::CharPointer &usingPosition, const boost::intrusive_ptr<const IdOrVar> _idOrVar, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endusingPosition) :
		Statement(usingPosition, endusingPosition),
		idOrVar(_idOrVar),
		statements(_statements)
	{
	}
};

class ForeachStatement : public Statement {
public:
	const CoCoA::Lexer::Token ident;
	const boost::intrusive_ptr<const Expression> inExp;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	ForeachStatement(const CoCoA::Lexer::CharPointer &foreachPosition, const CoCoA::Lexer::Token &_ident, boost::intrusive_ptr<const Expression> _inExp, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endforPosition) :
		Statement(foreachPosition, endforPosition),
		ident(_ident),
		inExp(_inExp),
		statements(_statements)
	{
		assert(inExp);
	}
};

class WhileStatement : public Statement {
public:
	const boost::intrusive_ptr<const Expression> exp;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	WhileStatement(const CoCoA::Lexer::CharPointer &whilePosition, boost::intrusive_ptr<const Expression> _exp, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endwhilePosition) :
                Statement(whilePosition, endwhilePosition),
		exp(_exp),
		statements(_statements)
	{
		assert(exp);
	}
};

class RepeatUntilStatement : public Statement {
public:
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	const boost::intrusive_ptr<const Expression> exp;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	RepeatUntilStatement(const CoCoA::Lexer::CharPointer &repeatPosition, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(repeatPosition, semicolonPosition),
		statements(_statements),
		exp(_exp)

	{
		assert(exp);
	}
};

struct IfBranch {
	boost::intrusive_ptr<const Expression> optExp; // no exp means it's the "else" branch
	std::vector<boost::intrusive_ptr<const Statement> > statements;
};

class IfStatement : public Statement {
public:
	const std::vector<IfBranch> branches;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	IfStatement(const CoCoA::Lexer::CharPointer &ifPosition, const std::vector<IfBranch> _branches, const CoCoA::Lexer::CharPointer &endifPosition) :
		Statement(ifPosition, endifPosition),
		branches(_branches)
	{
		assert(branches.size()>=1);
	}
};

struct Binding {
	boost::intrusive_ptr<const IdOrVar> idOrVar;
	std::string packageName;
};

class AliasStatement : public Statement {
public:
	const std::vector<Binding> bindings;
	const bool thereIsIn;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	AliasStatement(const CoCoA::Lexer::CharPointer &aliasPosition, const std::vector<Binding> &_bindings, const bool _thereIsIn, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endaliasOrSemicolonPosition) :
		Statement(aliasPosition, endaliasOrSemicolonPosition),
		bindings(_bindings),
		thereIsIn(_thereIsIn),
		statements(_statements)
	{
	}
};

class BlockStatement : public Statement {
public:
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	BlockStatement(const CoCoA::Lexer::CharPointer &blockPosition, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endblockPosition) :
		Statement(blockPosition, endblockPosition),
		statements(_statements)
	{
	}
};

struct Param {
	std::string name;
	bool byRef;
	Param() : byRef(false) {}
};

class DefineStatement : public Statement {
public:
	const std::string name;
	const bool thereIsEllipsis;
	const std::vector<Param> params;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	DefineStatement(const CoCoA::Lexer::CharPointer &definePosition, const std::string &_name, const bool _thereIsEllipsis, const std::vector<Param> &_params, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endDefinePosition) :
		Statement(definePosition, endDefinePosition),
		name(_name),
		thereIsEllipsis(_thereIsEllipsis),
		params(_params),
		statements(_statements)
	{
	}
};

class PackageStatement : public Statement {
public:
	const std::string name;
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	PackageStatement(const CoCoA::Lexer::CharPointer &packagePosition, const std::string &_name, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const CoCoA::Lexer::CharPointer &endPackagePosition) :
		Statement(packagePosition, endPackagePosition),
		name(_name),
		statements(_statements)
	{
	}
};

class HelpStatement : public Statement {
public:
	const std::string lexeme;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	HelpStatement(const CoCoA::Lexer::Token &tokHelp) :
		Statement(tokHelp.getBegin(), tokHelp.getEnd()),
		lexeme(tokHelp.lexeme())
	{
	}
};

class CatchStatement : public Statement {
public:
	const std::vector<boost::intrusive_ptr<const Statement> > statements;
	const boost::intrusive_ptr<const Expression> leftValue;
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	CatchStatement(const CoCoA::Lexer::CharPointer &catchPosition, const std::vector<boost::intrusive_ptr<const Statement> > &_statements, const boost::intrusive_ptr<const Expression> _leftValue, const CoCoA::Lexer::CharPointer &endcatchPosition) :
		Statement(catchPosition, endcatchPosition),
		statements(_statements),
		leftValue(_leftValue)
	{
	}
};

class PrintStatement : public Statement {
	const std::vector<boost::intrusive_ptr<const Expression> > exps;
	const boost::intrusive_ptr<const Expression> onExp;
	const CoCoA::Lexer::TokenType ttype;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	PrintStatement(const CoCoA::Lexer::Token tokPrintOrPrintLn, const std::vector<boost::intrusive_ptr<const Expression> > &_exps, const boost::intrusive_ptr<const Expression> _onExp, const CoCoA::Lexer::CharPointer &endifPosition) :
		Statement(tokPrintOrPrintLn.getBegin(), endifPosition),
		exps(_exps),
		onExp(_onExp),
		ttype(tokPrintOrPrintLn.getType())
	{
	}
};

class ClearStatement : public Statement {
	const std::vector<boost::intrusive_ptr<const IdOrVar> > idOrVarList;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	ClearStatement(const CoCoA::Lexer::CharPointer &clearPosition, const std::vector<boost::intrusive_ptr<const IdOrVar> > &_idOrVarList, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(clearPosition, semicolonPosition),
		idOrVarList(_idOrVarList)
	{
		assert(idOrVarList.size()>=1);
	}
};

class DeleteStatement : public Statement {
	const std::vector<boost::intrusive_ptr<const IdOrVar> > idOrVarList;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	DeleteStatement(const CoCoA::Lexer::CharPointer &deletePosition, const std::vector<boost::intrusive_ptr<const IdOrVar> > &_idOrVarList, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(deletePosition, semicolonPosition),
		idOrVarList(_idOrVarList)
	{
		assert(idOrVarList.size()>=1);
	}
};

class DestroyStatement : public Statement {
	const std::vector<boost::intrusive_ptr<const IdOrVar> > idOrVarList;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	DestroyStatement(const CoCoA::Lexer::CharPointer &destroyPosition, const std::vector<boost::intrusive_ptr<const IdOrVar> > &_idOrVarList, const CoCoA::Lexer::CharPointer &semicolonPosition) :
		Statement(destroyPosition, semicolonPosition),
		idOrVarList(_idOrVarList)
	{
		assert(idOrVarList.size()>=1);
	}
};

class ParenthesizedExp : public Expression {
	const boost::intrusive_ptr<const Expression> innerExp;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	ParenthesizedExp(boost::intrusive_ptr<const Expression> _innerExp, const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		Expression(_beginSourcePosition, _endSourcePosition),
		innerExp(_innerExp)
	{
		assert(innerExp);
	}
};

class IsDefinedExp : public Expression {
	const boost::intrusive_ptr<const Expression> exp;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	IsDefinedExp(boost::intrusive_ptr<const Expression> _exp, const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		Expression(_beginSourcePosition, _endSourcePosition),
		exp(_exp)
	{
		assert(exp);
	}
};

class BackwardCompatibleExp : public Expression {
	const boost::intrusive_ptr<const Expression> innerExp;
public:
	std::ostream &dumpAsString(std::ostream &out) const;
	void dumpAsTree(std::ostream &out, int level) const;
	BackwardCompatibleExp(boost::intrusive_ptr<const Expression> _innerExp, const CoCoA::Lexer::CharPointer &_beginSourcePosition, const CoCoA::Lexer::CharPointer & _endSourcePosition) :
		Expression(_beginSourcePosition, _endSourcePosition),
		innerExp(_innerExp)
	{
		assert(innerExp);
	}
};

} // namespace AST
} // namespace CoCoA

#endif /* AST_H_ */
