//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

grammar CoCoA5;

options {
  k=3;
}

// Notes:
// 1) The names "parse<Something>" in the following comments refer to the corresponding methods of 
//    class CoCoA::Parser::Parser
// 2) All keywords are case insensitive so, for instance, 'Use' should be specified as
//    (('U'|'u') ('S'|'s') ('E'|'e')).
//    For the time being, there is an exception: 'RECORD' (all uppercase)
//    is recognized as an Identifier instead of (the uppercase version of) the keyword 'Record',
//    while all the other casing are (e.g., ReCoRD, recoRD, RECORd ...)
//    Needless to say, this is a backward compatibility hack and this issue should be rectified
// 3) Since CoCoA 4 is inconsistent in the treatment of semicolons, the current parser for
//    CoCoA 5 is quite forgiving when semicolons are omitted (the extra-pedantic warning level
//    has to be turned on to get a feedback about this issue). Future versions may become strictier.

/* parseStatement */ statement
    : /* parseUseStatement         */ 'Use' (ringDefinition | idOrVar ('::=' ringDefinition )? )';'
    | /* parseSetStatement         */ 'Set' idOrVar ( ':=' expression )? ';'
    | /* parseUnsetStatement       */ 'Unset' idOrVar ';'
    | /* parseForStatement         */ 'For' Identifier ':=' expression 'To' expression ('Step' expression)? 'Do' statement* ('EndFor'|'End')
    | /* parseForeachStatement     */ 'Foreach' Identifier 'In' expression 'Do' statement* ('EndForeach'|'End')
    | /* parseIfStatment           */ 'If' expression 'Then' statement*
                                      ('Elif' expression 'Then' statement*)* ('Else' statement*) ('Endif'|'End')
    | /* parseReturnStatement      */ 'Return' expression? ';'
    | /* parseUsing                */ 'Using' Identifier 'Do' statement* ('EndUsing'|'End')
    | /* parsePrint                */ ('Print'|'PrintLn') (expressionOrNewline (',' expressionOrNewline)*)? ('On' expression)? ';'
    | /* parseBreakStatement       */ 'Break' ';'
    | /* parseRepeatUntilStatement */ 'Repeat' statement* 'Until' expression ';' 
    | /* parseWhileStatement       */ 'While' expression 'Do' statement* ('EndWhile'|'End')
    | /* parseDescribeStatement    */ 'Describe' expression ';'
    | /* parseClearStatement       */ 'Clear' idOrVarList ';'
    | /* parseDeleteStatement      */ 'Delete' idOrVarList ';'
    | /* parseDestroyStatement     */ 'Destroy' idOrVarList ';'
    | /* parseAliasStatement       */ 'Alias' bindingList (('In' statement* ('EndAlias'|'End')) |';')
    | /* parseCiaoOrQuitStatement  */ ('Ciao'|'Quit') ';'
    | /* parseSkipStatement        */ 'Skip' ';'
    | /* parseCatchStatement       */ 'Catch' statement* ('In' idOrVar)? ('EndCatch'|'End')
    | /* parseBlockStatement       */ 'Block' statement* ('EndBlock'|'End')
    | /* parsePackageStatement     */ 'Package' PackageName statement* ('EndPackage'|'End')
    | /* (inside the method)       */ HelpStatement
    | /* parseSourceStatement      */ ('<<'|'Source') expression ';'
    | /* parseDefineStatement      */ 'Define' Identifier '(' (paramList|'...') ')' statement* ('EndDefine'|'End')
    | /* parseTryStatement         */ 'Try' statement* 'UponError' Identifier 'Do' statement* ('EndTry'|'End')
    | /* (end of the method)       */ expression (':=' expression | '::=' ringDefinition)? ';'
    | 'Time' statement // but not an expression-statement
    ;
    
/* parseBindingList */
bindingList
    : idOrVar ':=' PackageName (',' idOrVar ':=' PackageName)*
    ;
        
/* parseExpOrNewLine */
expressionOrNewline
    : expression
    | 'NewLine'
    ;
    
/* parseIdOrVarList */
idOrVarList
    : (idOrVar (',' idOrVar)*)?
    ;
    
/* parseParamList */
paramList
    : ( 'Var'? Identifier (',' 'Var'? Identifier)*)?
    ;
        
/* parseRingDefinition */
// FIXME: this definition is more general than it should be (that is, it currently accepts too much stuff)
ringDefinition
    : ( Identifier ('/' '(' expression ')')?) '[' indeterminateDecl ( ',' indeterminateDecl )* ']'
        ( ',' ('ToPos' | 'PosTo' | ('Weights' '(' expression (',' expression)* ')')|'Lex'|'Xel'|'DegRevLex'|'DegLex'|('Ord' '(' expression ')')|
        ('Elim' '(' elimIndeterminateDecl ('..' elimIndeterminateDecl)? ')')))*
    ;

/* parseIndeterminateDecl(isInsideElim=false) */
indeterminateDecl    
    : idOrVar ('[' additiveExpression'..'additiveExpression (',' additiveExpression'..'additiveExpression)* ']')?
        // here we use additiveExpression because '..' creates ambiguities
    ;

/* parseIndeterminateDecl(isInsideElim=true) */
elimIndeterminateDecl    
    : idOrVar ('[' additiveExpression  (',' additiveExpression)* ']')?
        // here we use additiveExpression because '..' creates ambiguities
    ;
    
    
/* parseExpressionList */
expressionList
    : expression (',' expression)*
    ;
  
/* parseExpression */
expression
    : conditionalOrExpression
    | '${' expression '}$'
    ;

/* parseConditionOrExpression */
conditionalOrExpression
    : conditionalAndExpression ( 'Or' conditionalAndExpression )*
    ;

/* parseConditionalAndExpression */
conditionalAndExpression
    : equalityExpression ( 'And' equalityExpression )*
    ;

/* parseEqualityExpression */
equalityExpression
    : relationalExpression ( ('=' | '<>' | 'IsIn') relationalExpression )*
    ;

/* parseRelationalExpression */
relationalExpression
    : cartesianProductExpression ( ('<='|'<'|'>'|'>=') cartesianProductExpression )*
    ;
    
/* parseCartesianProductExpression */
cartesianProductExpression
    : listExpression ( '><' listExpression )* // note: this is an n-ary operator
        ;
  
/* parseListExpression */
listExpression
    : additiveExpression ('..' additiveExpression)?
    ;

/* parseAdditiveExpression */
additiveExpression
    : multiplicativeExpression ( ('+' | '-') multiplicativeExpression )*
    ;

/* parseMultiplicativeExpression */
multiplicativeExpression
    : powerExpression ( ( '*' | '/' | '%' ) powerExpression
    | (Identifier|'Var') => limitedPowerExpression)*
    ;
    
/* parsePowerExpression */
powerExpression
    : unaryExpression ('^' unaryExpression)*
    ;
  
/* parseLimitedPowerExpression */
limitedPowerExpression
    : idOrVar selector* ('^' unaryExpression)*
    ;
     
/* parseUnaryExpression */
unaryExpression
    : '+' unaryExpression
    | '-' unaryExpression
    | 'Not' unaryExpression
    | primary selector* 
    ;
 
/* parsePrimary */
primary
    : '(' expression ')'
    | 'IsDefined' '(' expression ')'
    | 'True'
    | 'False'
    | (PackageName '.')? Identifier ('::' primary)?
    | 'Var' '(' expression ')' ('::' primary)?
    | FLOAT_LITERAL
    | INT_LITERAL
    | STRING_LITERAL STRING_LITERAL*
    | /* parseListPrimary */ '[' ']'
    | /* parseListPrimary */ '[' Identifier 'In' expression '|' expression ']' 
    | /* parseListPrimary */ '[' expression ('|' Identifier 'In' equalityExpression ('And' expression)? | (',' expression)*) ']'
                                // here we use equalityExpression because 'And' creates ambiguities
    | /* parseRecord       */ 'Record' '[' ((idOrVar ('='|':=') expression) (',' idOrVar ('='|':=') expression)*)? ']'
    | 'Ord' '(' expressionList? ')'
    | 'Elim' '(' expressionList? ')'
    ;
    
/* parseIdOrVar */
idOrVar
    :    Identifier | ('Var' '(' expression ')') 
    ;
   
/* parseSelectors */
selector
    :   '.' idOrVar
    |   '[' expressionList ']'
    |   '(' expressionList? ')'
    |	'(' '...' ')'
    ;

// Lexer:
 
Identifier 
    : ('a'..'z'|'A'..'Z'|'_')('a'..'z'|'A'..'Z'|'_'|'0'..'9')*
    ;

PackageName
    : '$' ('/' | 'a'..'z'|'A'..'Z'|'_'|'0'..'9')*
    ;

WS
    : (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=HIDDEN;}
    ;

COMMENT
    : '/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    ;

HelpStatement
    : '?' ~('\n'|'\r')* '\r'? '\n'
    ;
    
LINE_COMMENT
    : ('//'|'--') ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;}
    ;
    
INT_LITERAL
    : ('0'..'9')+
    ;
    
FLOAT_LITERAL
    : ('0'..'9')+ '.' ('0'..'9'+)
    ;
    
STRING_LITERAL // note: newlines are not allowed inside string literals (users must use the escape sequences)
    : '"'  ( ESCAPE_SEQUENCE | ~('\\'|'"' ) )* '"'
    | '\'' ( ESCAPE_SEQUENCE | ~('\\'|'\'') )* '\''
    ;

fragment
ESCAPE_SEQUENCE
    : '\\' ('a'|'t'|'n'|'r'|'\"'|'\''|'\\')
    ;

