//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
#include <boost/tr1/memory.hpp>
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/foreach.hpp>
#include <boost/scope_exit.hpp>
#include <readline/readline.h>
#include <readline/history.h>
#include <cstring>
#include <cctype>
#include <sstream>

#include "Lexer.H"

namespace CoCoA {
namespace Lexer {

using namespace std;
using namespace boost;

void Lexer::reportWarningMayBeRemoved(const string &prefixMsg, const CharPointer &from, const CharPointer &to)
{
	if (warnAboutFeaturesThatMayBeRemoved)
		reportWarning(prefixMsg+" will be probably removed in the final version of CoCoA 5", from, to, WS_PEDANTIC);
}

string Token::lexeme() const {
	return type==TT_EOF ? begin.toPrettyString() : begin.stringTo(end);
}

void Lexer::resetCompatibilityMode() {
	status.inCoCoA4CompatibilityMode=false;
}

int Lexer::getErrorCount() const {
	return nErrors;
}

int Lexer::getWarningCount() const {
	return nWarnings;
}

void Lexer::resetErrorCounts() {
	nErrors = nWarnings = 0;
}

inline void Lexer::ungetCP(CharPointer cp) {
	ungetCPs.push(cp);
}

void Lexer::consumeSingleLineComment(const CoCoA::Parser::ParserStatus &ps) {
	for(;;) {
		char c = *getCP(ps);
		if (c=='\n' || !c)
			return;
	}
}

Token Lexer::getHelpStatement(const CharPointer &begin, const CoCoA::Parser::ParserStatus &ps) {
	CharPointer end=begin;
	for(;;) {
		CharPointer cp = getCP(ps);
		char c = *cp;
		if (c=='\n' || !c)
			return Token(begin, end, TT_HELP);
		end = cp;
	}
}

CharPointer Lexer::getCP(const CoCoA::Parser::ParserStatus &ps) {
	if (!ungetCPs.empty()) {
		CharPointer cp = ungetCPs.top();
		ungetCPs.pop();
		return cp;
	}
	while (!currentLine || currentPositionInLine>=currentLine->chars.length()) {
		string chars;
		if (!lineProvider->readNextLine(this->status, ps, chars))
			return CharPointer::Null;
		chars.append("\n");
		currentLine = intrusive_ptr<Line>(new Line(lineProvider->currentLineNumber(), chars, lineProvider));
		if (previousLine)
			previousLine->nextLine = currentLine;
		previousLine = currentLine;
		currentPositionInLine = 0;
	}
	return CharPointer(currentLine, currentPositionInLine++);
}

inline CharPointer Lexer::peekCP(const CoCoA::Parser::ParserStatus &ps) {
	CharPointer cp = getCP(ps);
	ungetCP(cp);
	return cp;
}

Token Lexer::getToken(const CoCoA::Parser::ParserStatus &ps) {
	if (!ungetTokens.empty()) {
		Token t = ungetTokens.top();
		ungetTokens.pop();
		return t;
	}
	for(;;) {
		CharPointer begin = getCP(ps), end=begin;
		char c = *begin;
		switch (c) {
		case ' ':
		case '\n':
		case '\r':
		case '\t':
		case '\f':
			continue;
		case '\0':
			return Token::EndOfFile;
		case ';':
			return Token(begin, begin, TT_SEMICOLON);
		case '|':
			return Token(begin, begin, TT_PIPE);
		case '^':
			return Token(begin, begin, TT_POWER);
		case ',':
			return Token(begin, begin, TT_COMMA);
		case '%':
			return Token(begin, begin, TT_MOD);
		case '+':
			return Token(begin, begin, TT_PLUS);
		case '(':
			return Token(begin, begin, TT_OPEN_ROUND);
		case ')':
			return Token(begin, begin, TT_CLOSED_ROUND);
		case '[':
			return Token(begin, begin, TT_OPEN_SQUARE);
		case ']':
			return Token(begin, begin, TT_CLOSED_SQUARE);
		case '*':
			return Token(begin, begin, TT_STAR);
		case '=':
			return Token(begin, begin, TT_EQUAL);
		case '<':
			end=getCP(ps);
			switch (*end) {
			case '=':
				return Token(begin, end, TT_LE);
			case '>':
				return Token(begin, end, TT_NOTEQUAL);
			case '<':
				return Token(begin, end, TT_SOURCE_AS_LSHIFT);
			}
			ungetCP(end);
			return Token(begin, begin, TT_LT);
		case '>':
			end=getCP(ps);
			switch (*end) {
			case '=':
				return Token(begin, end, TT_GE);
			case '<':
				return Token(begin, end, TT_CART_PROD);
			}
			ungetCP(end);
			return Token(begin, begin, TT_GT);
		case '$':
			end=getCP(ps);
			if (*end=='{') {
				if (status.isInCoCoA4CompatibilityMode())
					throw LexerException("Compatibility parentheses cannot be nested", begin, end, false);
				status.inCoCoA4CompatibilityMode = true;
				return Token(begin, end, TT_OPEN_COMPAT);
			}
			ungetCP(end);
			return getPackageName(begin, ps);
		case '}':
			end=getCP(ps);
			if (*end=='$') {
				if (!status.isInCoCoA4CompatibilityMode())
					throw LexerException("Trying to close an unopened compatibility parenthesis", begin, end, false);
				status.inCoCoA4CompatibilityMode = false;
				return Token(begin, end, TT_CLOSE_COMPAT);
			}
			ungetCP(end);
			break;
		case '.':
			end=getCP(ps);
			if (*end=='.') {
				CharPointer cp = getCP(ps);
				if (*cp=='.')
					return Token(begin, end, TT_ELLIPSIS);
				ungetCP(cp);
				return Token(begin, end, TT_DOTDOT);
			}
			ungetCP(end);
			return Token(begin, begin, TT_DOT);
		case '-':
			if (*peekCP(ps)=='-') {
				consumeSingleLineComment(ps);
				continue;
			}
			return Token(begin, begin, TT_MINUS);
		case '?':
			return getHelpStatement(begin, ps);
		case '/':
			switch(*peekCP(ps)) {
			case '/':
				consumeSingleLineComment(ps);
				continue;
			case '*':
				assert(!status.inMultiLineComment);
				status.inMultiLineComment=true;
				{
					BOOST_SCOPE_EXIT( (&status) ) {
						status.inMultiLineComment=false;
					} BOOST_SCOPE_EXIT_END
					getCP(ps); // consumes the '*'
					bool lastWasStar = false;
					CharPointer lastReadCP = begin;
					for(;;) {
						CharPointer cp = getCP(ps);
						char c = *cp;
						if (!c)
							throw LexerException("Unclosed multi-line comment", begin, lastReadCP, false);
						lastReadCP = cp;
						if (c=='/' && lastWasStar)
							break;
						lastWasStar = c=='*';
					}
				}
				continue;
			}
			return Token(begin, begin, TT_SLASH);
		case '\'':
			reportWarning("the use of single-quotes for defining string constants is discouraged; use the double quotes (\") to avoid this warning", WS_NORMAL);
		case '\"':
			return getStringLiteral(begin, ps);
		case ':':
			end=getCP(ps);
			switch(*end) {
			case '=':
				return Token(begin, end, TT_ASSIGN);
			case ':':
				if (*peekCP(ps)=='=')
					return Token(begin, getCP(ps), TT_RING_ASSIGN);
				return Token(begin, end, TT_COLONCOLON);
			}
			ungetCP(end);
			return Token(begin, begin, TT_COLON);
		}
		if (isalpha(c) || c=='_')
			return getIdentifierOrKeyword(begin, ps);
		if (isdigit(c))
			return getNumberLiteral(begin, ps);
		if (c=='{' || c=='}')
			throw LexerException("Unknown symbol "+begin.toPrettyString()+", did you forget a $? Backward compatibility parentheses are ${ and }$", begin, begin, true);
		throw LexerException("Unknown symbol "+begin.toPrettyString(), begin, begin, true);
	}
}

namespace {
	struct Keyword {
		const string realId;
		const string lowerId;
		const TokenType tokenType;
		Keyword(const string & _realId, TokenType _tokenType) :
			realId(_realId),
			lowerId(to_lower_copy(_realId)),
			tokenType(_tokenType)
		{}
	} keywordTable [] = {
			Keyword("Alias", TT_ALIAS),
			Keyword("And", TT_AND),
			Keyword("Block", TT_BLOCK),
			Keyword("Break", TT_BREAK),
			Keyword("Catch", TT_CATCH),
			Keyword("Ciao", TT_CIAO),
			Keyword("Clear", TT_CLEAR),
			Keyword("Continue", TT_CONTINUE),
			Keyword("Define", TT_DEFINE),
			Keyword("DegLex", TT_DEGLEX),
			Keyword("DegRevLex", TT_DEGREVLEX),
			Keyword("Delete", TT_DELETE),
			Keyword("Describe", TT_DESCRIBE),
			Keyword("Destroy", TT_DESTROY),
			Keyword("Do", TT_DO),
			Keyword("Elif", TT_ELIF),
			Keyword("Elim", TT_ELIM),
			Keyword("Else", TT_ELSE),
			Keyword("EndAlias", TT_ENDALIAS),
			Keyword("EndBlock", TT_ENDBLOCK),
			Keyword("EndCatch", TT_ENDCATCH),
			Keyword("EndDefine", TT_ENDDEFINE),
			Keyword("EndForeach", TT_ENDFOREACH),
			Keyword("EndFor", TT_ENDFOR),
			Keyword("EndIf", TT_ENDIF),
			Keyword("EndPackage", TT_ENDPACKAGE),
			Keyword("End", TT_END),
			Keyword("EndUsing", TT_ENDUSING),
			Keyword("EndWhile", TT_ENDWHILE),
			Keyword("False", TT_FALSE),
			Keyword("Foreach", TT_FOREACH),
			Keyword("For", TT_FOR),
			Keyword("If", TT_IF),
			Keyword("In", TT_IN),
			Keyword("IsDefined", TT_ISDEFINED),
			Keyword("IsIn", TT_ISIN),
			Keyword("Lex", TT_LEX),
			Keyword("Not", TT_NOT),
			Keyword("On", TT_ON),
			Keyword("Ord", TT_ORD),
			Keyword("Or", TT_OR),
			Keyword("Package", TT_PACKAGE),
			Keyword("PosTo", TT_POSTO),
			Keyword("PrintLn", TT_PRINTLN),
			Keyword("Print", TT_PRINT),
			Keyword("Protect", TT_PROTECT),
			Keyword("Quit", TT_QUIT),
			Keyword("Record", TT_RECORD),
			Keyword("Repeat", TT_REPEAT),
			Keyword("Return", TT_RETURN),
			Keyword("Set", TT_SET),
			Keyword("Skip", TT_SKIP),
			Keyword("Source", TT_SOURCE),
			Keyword("Step", TT_STEP),
			Keyword("Then", TT_THEN),
			Keyword("Time", TT_TIME),
			Keyword("To", TT_TO),
			Keyword("ToPos", TT_TOPOS),
			Keyword("True", TT_TRUE),
			Keyword("Unset", TT_UNSET),
			Keyword("Until", TT_UNTIL),
			Keyword("Unprotect", TT_UNPROTECT),
			Keyword("Use", TT_USE),
			Keyword("Using", TT_USING),
			Keyword("Var", TT_VAR),
			Keyword("Weights", TT_WEIGHTS),
			Keyword("While", TT_WHILE),
			Keyword("Xel", TT_XEL)
	};
}

Token Lexer::getNumberLiteral(const CharPointer &begin, const CoCoA::Parser::ParserStatus &ps) {
	CharPointer end = begin;
	for(;;) {
		CharPointer cp = getCP(ps);
		char c = *cp;
		if (c=='.') {
			if (isdigit(*peekCP(ps))) {
				for(;;) {
					cp = getCP(ps);
					if (!isdigit(*cp)) {
						ungetCP(cp);
						break;
					}
					end = cp;
				}
				return Token(begin, end, TT_FLOAT_LITERAL);
			}
		}
		if (!isdigit(c)) {
			ungetCP(cp);
			break;
		}
		end = cp;
	}
	return Token(begin, end, TT_INT_LITERAL);
}

Token Lexer::getStringLiteral(const CharPointer &begin, const CoCoA::Parser::ParserStatus &ps) {
	const char matchingQuote = *begin;
	status.inStringLiteral = true;
	BOOST_SCOPE_EXIT( (&status) ) {
		status.inStringLiteral=false;
	} BOOST_SCOPE_EXIT_END
	CharPointer end = begin;
	CharPointer previousCharP = begin;
	bool lastWasEscape=false;
	for(;;) {
		CharPointer cp = getCP(ps);
		char c = *cp;
		if (!c)
			throw LexerException("Unclosed string literal", begin, end, false);
		previousCharP = end;
		end = cp;
		if (c==matchingQuote && !lastWasEscape) {
			CharPointer nextCharP = getCP(ps);
			if (*nextCharP==matchingQuote)
				continue;
			ungetCP(nextCharP);
			break;
		}
		if (lastWasEscape) {
			switch (c) {
			case 'n':
			case 'r':
			case 't':
			case 'a':
			case '\'':
			case '\"':
			case '\\':
				break;
			default:
				reportError("Unrecognized escape sequence", previousCharP, end);
			}
		}
		lastWasEscape = !lastWasEscape && c=='\\';
		if (c=='\n')
			reportError("Newline characters are not allowed inside string literals (you can insert them using escape sequences, for instance, \\n)", end, end);
	}
	return Token(begin, end, TT_STRING_LITERAL);
}

Token Lexer::getPackageName(const CharPointer &begin, const CoCoA::Parser::ParserStatus &ps) {
	CharPointer end = begin;
	CharPointer cp = begin;
	char c;
	for(;;) {
		cp = getCP(ps);
		c = *cp;
		if (!(isalnum(c) || c=='_' || c=='/')) {
			ungetCP(cp);
			break;
		}
		end = cp;
	}
	return Token(begin, end, TT_PACKAGENAME);
}

Token Lexer::getIdentifierOrKeyword(const CharPointer &begin, const CoCoA::Parser::ParserStatus &ps) {
	CharPointer end = begin;
	if (status.isInCoCoA4CompatibilityMode() && islower(*begin)) {
		CharPointer cp = peekCP(ps);
		if (isupper(*cp))
			throw LexerException("This form of implicit multiplication is not allowed (anymore)", begin, cp, false);
		return Token(begin, begin, TT_IDENTIFIER);
	}
	for(;;) {
		CharPointer cp = getCP(ps);
		char c = *cp;
		if (!(isalnum(c) || c=='_')) {
			ungetCP(cp);
			break;
		}
		end = cp;
	}
	const size_t beginPos = begin.getPositionInLine();
	const size_t endPos = end.getPositionInLine();
	assert(begin.getLine()==end.getLine());
	const string chars = begin.getLine()->chars.substr(beginPos, endPos-beginPos+1);
	if (chars=="RECORD")
		return Token(begin, end, TT_IDENTIFIER);
	if (warningLevel!=WS_EXTRA_PEDANTIC) {
		// upper cases "TRUE" and "FALSE" are for backward compatibility (they should be "True" and "False")
		if (chars=="TRUE")
			return Token(begin, end, TT_TRUE);
		if (chars=="FALSE")
			return Token(begin, end, TT_FALSE);
	}
	const string lowerChars = to_lower_copy(chars);
	TokenType type = TT_IDENTIFIER;
	BOOST_FOREACH(Keyword k, keywordTable) {
		if (chars==k.realId || chars==k.lowerId) {
			type = k.tokenType;
			break;
		}
		if (lowerChars==k.lowerId) {
			string msg(chars+" recognized as the keyword "+k.realId+"; use \""+k.realId+"\" ");
			msg += status.isInCoCoA4CompatibilityMode() ?
					"to avoid this warning (note: outside CoCoA4 compatibility mode, \""+k.lowerId+"\" is allowed too)" :
					"or \""+k.lowerId+"\" to avoid this warning (note: in CoCoA4 compatibility mode, only the former casing is allowed)";
			reportWarning(msg, begin, end, WS_NORMAL);
			type = k.tokenType;
			break;
		}
	}
	return Token(begin, end, type);
}

namespace {
	const char HICHAR = '^';
	const size_t MAX_PREFIX = 30;
	const size_t MAX_SUFFIX = MAX_PREFIX;

	inline void outputChar(ostream &out, char c, size_t n) {
		for(size_t i=0; i<n; ++i)
			out << c;
	}

	void outputUnderlinedLine(ostream &out, intrusive_ptr<const Line> line, size_t from, size_t to) {
		string prefix, suffix;
		const string &chars = line->chars;
                size_t printingStart=0, printingEnd=chars.length()==0 ? 0 : chars.length()-1;
		if (from>=MAX_PREFIX) {
			printingStart = from-MAX_PREFIX;
			prefix = "... ";
		}
		if ( (printingEnd-to)>=MAX_SUFFIX ) {
			printingEnd = to+MAX_SUFFIX;
			suffix = " ...";
		}
		assert(printingStart<=printingEnd);
		assert(printingEnd<chars.length());
		out << prefix;
		for(size_t i=printingStart; i<=printingEnd;) {
			const char c = chars[i++];
			if (c=='\n')
				break;
			out << (c=='\t' ? ' ':c);
		}
		out << suffix << '\n';
		outputChar(out, ' ', prefix.length()+from-printingStart);
		outputChar(out, HICHAR, to-from+1);
		out << '\n';
	}

	void outputUnderlinedChars(ostream &out, const CharPointer &from, const CharPointer &to) {
		if (!*from) {
			out << "<End of file>" << endl;
			return;
		}
		intrusive_ptr<const Line> fromLine = from.getLine();
		intrusive_ptr<const Line> toLine = to.getLine();
		if (fromLine==toLine)
			return outputUnderlinedLine(out, fromLine, from.getPositionInLine(), to.getPositionInLine());
		if (!*to) {
			out << "... <End of file>" << endl;
			return;
		}
		outputUnderlinedLine(out, fromLine, from.getPositionInLine(), fromLine->chars.length()-2); // the last char of every line is \n and we don't want to underline it
		int skippedLines=0;
		intrusive_ptr<const Line> currLine = fromLine;
		while ( (currLine = currLine->getNextLineInBuffer())!= toLine )
			++skippedLines;
		if (skippedLines) {
			out << "... (";
			if (skippedLines==1)
				out << "another source line";
			else
				out << "other " << skippedLines << " source lines";
			out << ") ...\n";
		}
		outputUnderlinedLine(out, toLine, 0, to.getPositionInLine());
	}
}

intrusive_ptr<const Line> CharPointer::emptyLine(new Line());

void reportLineNumberWhenMeaningful(const CharPointer &fromPos, const CharPointer &toPos) {
	intrusive_ptr<const Line> l = fromPos.getLine();
	if (intrusive_ptr<const FileLineProvider> p = dynamic_pointer_cast<const FileLineProvider>(l->provider)) {
		int l1 = l->number;
		int l2 = toPos.getLine()->number;
		if (l1==l2)
			cerr << " at line " << l1 << " (column " << (fromPos.getPositionInLine()+1) << ")";
		else
			cerr << " at lines from " << l1 << " (column " << (fromPos.getPositionInLine()+1) << ") to " << l2 << " (column " << (toPos.getPositionInLine()+1) << ")";
		cerr << " of file " << p->filename;
	}
}

namespace {
	const string ERROR("Error");
	const string WARNING("Warning");
}

void Lexer::incrementErrorCounter() {
	if (++nErrors >= TooManyErrorsException::MAX_ERRORS)
		throw TooManyErrorsException();
}

void Lexer::reportError(const string &msg) {
	cerr << ERROR << ": " << msg << '\n';
	incrementErrorCounter();
}

void Lexer::reportError(const string &msg, const CharPointer &from, const CharPointer &to) {
	if (from==lastReportedErrorPosition) // may happen because of parsing error recovery
		return;
	lastReportedErrorPosition = from;
	cerr << ERROR;
	reportLineNumberWhenMeaningful(from, to);
	cerr << ": " << msg << '\n';
	outputUnderlinedChars(cerr, from, to);
	incrementErrorCounter();
}

void Lexer::reportWarning(const string &msg, WarningSeverity severity) {
	if (warningLevel>=severity) {
		++nWarnings;
		cerr << WARNING << ": " << msg << '\n';
	}
}

void Lexer::reportWarning(const string &msg, const CharPointer &from, const CharPointer &to, WarningSeverity severity) {
	if (warningLevel>=severity) {
		++nWarnings;
		cerr << WARNING;
		reportLineNumberWhenMeaningful(from, to);
		cerr << ": " << msg << '\n';
		outputUnderlinedChars(cerr, from, to);
	}
}

string CharPointer::toPrettyString() const {
	char c = **this;
	if (c>' ')
		return '\"'+std::string(1, c)+'\"';
	switch(c) {
	case '\0': return "<end of file>";
	case '\t': return "<tab>";
	case '\n': return "<newline>";
	case '\r': return "<carriage return>";
	case ' ' : return "<space>";
	default  : std::ostringstream os;
			   os << "<unprintable char (code=" << static_cast<int>(c) << ")>";
			   return os.str();
	}
}

string CharPointer::stringTo(const CharPointer &end) const {
	assert(line);
	assert(end.line);
	std::string result;
	boost::intrusive_ptr<const Line> l = line;
	size_t pos = positionInLine;
	for(; l!=end.line; l=l->getNextLineInBuffer()) {
		assert(l);
		assert(pos<l->chars.length());
		result.append(l->chars.substr(pos));
		pos=0;
	}
	assert(pos<=end.positionInLine);
	result.append(l->chars.substr(pos, end.positionInLine-pos+1));
	return result;
}

const CharPointer CharPointer::Null;

Token Token::EndOfFile;

} // namespace Lexer
} // namespace CoCoA
