//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LEXER_H_
#define LEXER_H_

#include <boost/utility.hpp>
#include <string>
#include <cassert>
#include <stack>

#include "LineProviders.H"

namespace CoCoA {
namespace Lexer {

class CharPointer;

class Line : public ReferenceCountedObject {
private:
	friend class Lexer;
	friend class CharPointer;
	boost::intrusive_ptr<const Line> nextLine;
	Line() :
		number(0),
		chars(1, '\0')
	{}
public:
	const int number;
	const std::string chars;
	const boost::intrusive_ptr<const LineProvider> provider;
	Line(const int _number, const std::string &_chars, const boost::intrusive_ptr<const LineProvider> _provider) :
		number(_number),
		chars(_chars),
		provider(_provider)
	{
		assert(number>=0);
		assert(provider);
	}
	const boost::intrusive_ptr<const Line> getNextLineInBuffer() const { return nextLine; }
};

class CharPointer {
	boost::intrusive_ptr<const Line> line;
	size_t positionInLine;
	static boost::intrusive_ptr<const Line> emptyLine;
	CharPointer() :
		line(emptyLine),
		positionInLine(0)
	{}
public:
	static const CharPointer Null;
	CharPointer(boost::intrusive_ptr<const Line> _line, const size_t _positionInLine) :
		line(_line),
		positionInLine(_positionInLine)
	{
		assert(line);
		assert(positionInLine<line->chars.length());
	}
	inline boost::intrusive_ptr<const Line> getLine() const { return line; }
	inline size_t getPositionInLine() const { return positionInLine; }
	inline char operator *() const {
		return line->chars[positionInLine];
	}
	std::string toPrettyString() const;
	std::string stringTo(const CharPointer &end) const;
	bool operator==(const CharPointer &other) const {
		return positionInLine==other.positionInLine && line==other.line;
	}
	bool operator!=(const CharPointer &other) const {
		return positionInLine!=other.positionInLine || line!=other.line;
	}
};

enum TokenType {
	TT_ALIAS,
	TT_AND,
	TT_ASSIGN, // :=
	TT_BLOCK,
	TT_BREAK,
	TT_CART_PROD, // ><
	TT_CATCH,
	TT_CIAO,
	TT_CLEAR,
	TT_CLOSE_COMPAT, // }$
	TT_CLOSED_ROUND, // )
	TT_CLOSED_SQUARE, // ]
	TT_COLON,
	TT_COLONCOLON, // ::
	TT_COMMA,
	TT_CONTINUE,
	TT_DEFINE,
	TT_DEGLEX,
	TT_DEGREVLEX,
	TT_DELETE,
	TT_DESCRIBE,
	TT_DESTROY,
	TT_DO,
	TT_DOT,
	TT_DOTDOT,
	TT_ELLIPSIS,
	TT_ELIF,
	TT_ELIM,
	TT_ELSE,
	TT_END,
	TT_ENDALIAS,
	TT_ENDBLOCK,
	TT_ENDCATCH,
	TT_ENDDEFINE,
	TT_ENDFOR,
	TT_ENDFOREACH,
	TT_ENDIF,
	TT_ENDPACKAGE,
	TT_ENDUSING,
	TT_ENDWHILE,
	TT_EOF,
	TT_EQUAL,
	TT_FALSE,
	TT_FLOAT_LITERAL,
	TT_FOR,
	TT_FOREACH,
	TT_GE, // >=
	TT_GT, // >
	TT_HELP, // ? .... <newline>
	TT_IDENTIFIER,
	TT_IF,
	TT_IN, // In
	TT_INT_LITERAL,
	TT_ISDEFINED,
	TT_ISIN, // IsIn
	TT_LE, // <=
	TT_LEX,
	TT_LT, // <
	TT_MINUS,
	TT_MOD, // %
	TT_NOT,
	TT_NOTEQUAL, // <>
	TT_ON,
	TT_OPEN_COMPAT, // ${
	TT_OPEN_ROUND, // (
	TT_OPEN_SQUARE, // [
	TT_OR,
	TT_ORD,
	TT_PACKAGE,
	TT_PACKAGENAME,
	TT_PIPE, // |
	TT_PLUS,
	TT_POSTO,
	TT_POWER, // ^
	TT_PRINT,
	TT_PRINTLN,
	TT_PROTECT,
	TT_QUIT,
	TT_RECORD,
	TT_REPEAT,
	TT_RETURN,
	TT_RING_ASSIGN, // ::=
	TT_SEMICOLON,
	TT_SET,
	TT_SKIP,
	TT_SLASH,
	TT_SOURCE,
	TT_SOURCE_AS_LSHIFT, // <<
	TT_STAR, // *
	TT_STEP,
	TT_STRING_LITERAL,
	TT_THEN,
	TT_TIME,
	TT_TO,
	TT_TOPOS,
	TT_TRUE,
	TT_UNPROTECT,
	TT_UNSET,
	TT_UNTIL,
	TT_USE,
	TT_USING,
	TT_VAR,
	TT_WEIGHTS,
	TT_WHILE,
	TT_XEL
};

class Token {
	TokenType type;
	CharPointer begin, end;
	Token() :
		type(TT_EOF),
		begin(CharPointer::Null),
		end(CharPointer::Null)
	{}
public:
	static Token EndOfFile;
	Token(const CharPointer &_begin, const CharPointer &_end, TokenType _type) :
		type(_type),
		begin(_begin),
		end(_end)
	{
		assert(type!=TT_EOF);
		assert(*begin);
		assert(*end);
	}
	inline TokenType getType() const { return type; }
	inline const CharPointer &getBegin() const { return begin; }
	inline const CharPointer &getEnd() const { return end; }
	std::string lexeme() const;
};

enum WarningSeverity {
	WS_NONE=0,
	WS_NORMAL=1,
	WS_PEDANTIC=2,
	WS_EXTRA_PEDANTIC=3
};

class Lexer : public ReferenceCountedObject {
	boost::intrusive_ptr<LineProvider> lineProvider;
	boost::intrusive_ptr<Line> previousLine, currentLine;
	size_t currentPositionInLine;
	std::stack<CharPointer> ungetCPs;
	std::stack<Token> ungetTokens;
	LexerStatus status;
	void ungetCP(CharPointer cp);
	CharPointer getCP(const CoCoA::Parser::ParserStatus &ps);
	CharPointer peekCP(const CoCoA::Parser::ParserStatus &ps);
	Token getIdentifierOrKeyword(const CharPointer &beginCP, const CoCoA::Parser::ParserStatus &ps);
	Token getPackageName(const CharPointer &beginCP, const CoCoA::Parser::ParserStatus &ps);
	Token getNumberLiteral(const CharPointer &beginCP, const CoCoA::Parser::ParserStatus &ps);
	Token getStringLiteral(const CharPointer &beginCP, const CoCoA::Parser::ParserStatus &ps);
	Token getHelpStatement(const CharPointer &beginCP, const CoCoA::Parser::ParserStatus &ps);
	void consumeSingleLineComment(const CoCoA::Parser::ParserStatus &ps);
	void incrementErrorCounter();
	CharPointer lastReportedErrorPosition;
	WarningSeverity warningLevel;
	int nErrors, nWarnings;
	bool warnAboutFeaturesThatMayBeRemoved;
	bool interactiveInput;
public:
	explicit Lexer(boost::intrusive_ptr<LineProvider> _lineProvider, WarningSeverity _warningLevel, bool _warnAboutFeaturesThatMayBeRemoved, bool _interactiveInput) :
		lineProvider(_lineProvider),
		currentPositionInLine(0),
		lastReportedErrorPosition(CharPointer::Null),
		warningLevel(_warningLevel),
		nErrors(0),
		nWarnings(0),
		warnAboutFeaturesThatMayBeRemoved(_warnAboutFeaturesThatMayBeRemoved),
		interactiveInput(_interactiveInput)
	{
		assert(lineProvider);
	}
	bool isInteractive() { return interactiveInput; }
	inline const LexerStatus & getStatus() const { return status; }
	inline void ungetToken(const Token &t) {
		ungetTokens.push(t);
	}
	Token getToken(const CoCoA::Parser::ParserStatus &ps);
	void resetCompatibilityMode();
	int getErrorCount() const;
	int getWarningCount() const;
	void resetErrorCounts();
	void reportError(const std::string &msg);
	void reportError(const std::string &msg, const CharPointer &from, const CharPointer &to);
	void reportWarning(const std::string &msg, WarningSeverity severity);
	void reportWarning(const std::string &msg, const CharPointer &from, const CharPointer &to, WarningSeverity severity);
	void reportWarningMayBeRemoved(const std::string &prefixMsg, const CharPointer &from, const CharPointer &to);
};

class ExceptionWithSourcePosition : public BaseException {
public:
	const CharPointer from, to;
	const bool needsRecovery;
	ExceptionWithSourcePosition(const std::string &_reason, const CharPointer &_from, const CharPointer &_to, bool _needsRecovery) :
		BaseException(_reason),
		from(_from),
		to(_to),
		needsRecovery(_needsRecovery)
	{}
	ExceptionWithSourcePosition(const std::string &_reason, const Token &token, bool _needsRecovery) :
		BaseException(_reason),
		from(token.getBegin()),
		to(token.getEnd()),
		needsRecovery(_needsRecovery)
	{}
	~ExceptionWithSourcePosition() throw () {}
};

class LexerException : public ExceptionWithSourcePosition {
public:
	LexerException(const std::string &_reason, const CharPointer &_from, const CharPointer &_to, bool _needsRecovery) :
		ExceptionWithSourcePosition(_reason, _from, _to, _needsRecovery)
	{}
	LexerException(const std::string &_reason, const Token &token, bool _needsRecovery) :
		ExceptionWithSourcePosition(_reason, token.getBegin(), token.getEnd(), _needsRecovery)
	{}
	~LexerException() throw () {}
};

class TooManyErrorsException : public BaseException {
	static std::string humanReadable(CoCoA::Lexer::TokenType tt);
public:
	static const int MAX_ERRORS = 15;
	TooManyErrorsException() :
		BaseException("Too many errors, bailing out")
	{}
	~TooManyErrorsException() throw () {}
};

} // namespace Lexer
} // namespace CoCoA
#endif /* LEXER_H_ */
