//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LINEPROVIDERS_H_
#define LINEPROVIDERS_H_

#include <fstream>
#include <boost/intrusive_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/iostreams/device/file.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <exception>
#include <vector>
#include <string>

namespace CoCoA {

namespace Parser {

class ParserStatus : boost::noncopyable {
	friend class Parser;
	bool inTheMiddleOfAStatement;
	bool inRecoveryMode;
	std::vector<std::string> openBlocks;
public:
	ParserStatus() :
		inTheMiddleOfAStatement(false),
		inRecoveryMode(false)
	{}
	bool isInTheMiddleOfAStatement() const { return inTheMiddleOfAStatement; }
	bool isInRecoveryMode() const { return inRecoveryMode; }
	std::string promptHeader() const { return openBlocks.empty() ? "" : openBlocks.back(); }
};

struct AskingForNewInteractiveInputDuringRecoveryException : std::exception {};

} // namespace Parser

namespace Lexer {

class LexerStatus : boost::noncopyable {
	friend class Lexer;
	bool inStringLiteral;
	bool inMultiLineComment;
	bool inCoCoA4CompatibilityMode;
public:
	LexerStatus() :
		inStringLiteral(false),
		inMultiLineComment(false),
		inCoCoA4CompatibilityMode(false)
	{}
	bool isInStringLiteral() const { return inStringLiteral; }
	bool isInMultiLineComment() const { return inMultiLineComment; }
	bool isInCoCoA4CompatibilityMode() const { return inCoCoA4CompatibilityMode; }
};

class ReferenceCountedObject : boost::noncopyable {
	mutable int refCount;
	friend void intrusive_ptr_add_ref(const ReferenceCountedObject *);
	friend void intrusive_ptr_release(const ReferenceCountedObject *);
public:
	ReferenceCountedObject() : refCount(0) {}
	virtual ~ReferenceCountedObject() = 0;
};

inline void intrusive_ptr_add_ref(const ReferenceCountedObject *o) {
	++o->refCount;
}

inline void intrusive_ptr_release(const ReferenceCountedObject *o) {
	if (--o->refCount==0)
		delete o;
}

class LineProvider : public ReferenceCountedObject {
	int lineNumber;
protected:
	virtual bool doReadNextLine(const LexerStatus &, const CoCoA::Parser::ParserStatus &, std::string &chars) = 0;
public:
	LineProvider() : lineNumber(0) {}
	inline int currentLineNumber() const { return lineNumber; }
	// Knowing the current state of the Lexer and the Parser, the ReadlineProvider (that is, the one
	// that fetches lines from the console) can provide meaningful prompts (indicating, for instance,
	// the nesting level of the current statement) and warning (for instance, newline inside a string literal).
	inline bool readNextLine(const LexerStatus &ls, const CoCoA::Parser::ParserStatus &ps, std::string &chars) {
		bool b = doReadNextLine(ls, ps, chars);
		if (b)
			++lineNumber;
		return b;
	}
	virtual ~LineProvider() {}
};

class InteractiveLineProvider : public LineProvider {
protected:
	static const char *prompt(const LexerStatus &, const CoCoA::Parser::ParserStatus &);
};

class ReadlineLineProvider : public InteractiveLineProvider {
protected:
	bool doReadNextLine(const LexerStatus &, const CoCoA::Parser::ParserStatus &, std::string &chars);
};

class GetlineLineProvider : public InteractiveLineProvider {
protected:
	bool doReadNextLine(const LexerStatus &, const CoCoA::Parser::ParserStatus &, std::string &chars);
};

class FileLineProvider : public LineProvider {
	boost::iostreams::filtering_istream input;
	bool echoToCout;
protected:
	bool doReadNextLine(const LexerStatus &, const CoCoA::Parser::ParserStatus &, std::string &chars);
public:
	const std::string filename;
	FileLineProvider(const std::string &_filename, bool _echoToCout);
};

class BaseException : public std::exception {
public:
	const std::string reason;
	explicit BaseException(const std::string &_reason) : reason(_reason) {
		assert(reason.length()>0);
	}
	const char *what() const throw () { return reason.c_str(); }
	~BaseException() throw () {}
};

struct IOException : public BaseException {
	explicit IOException(const std::string &_reason) :
		BaseException(_reason)
	{}
};

struct FailedAssertionException : public BaseException {
	explicit FailedAssertionException(const std::string &_reason) :
		BaseException(_reason)
	{}
};

} // namespace Lexer
} // namespace CoCoA

#endif /* LINEPROVIDERS_H_ */
