//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
#include <boost/tr1/memory.hpp>
#include <cstdlib>
#include "Parser.H"

using namespace std;
using namespace boost;
using namespace CoCoA::Lexer;
using namespace CoCoA::Parser;
using namespace CoCoA::AST;

namespace {

void printErrors(int nErrors, int nWarnings) {
	cout << "Got " << nErrors << " error(s) and " << nWarnings << " warning(s).\n";
}

}

int process(intrusive_ptr<LineProvider> lineProvider, WarningSeverity warningLevel, bool warnAboutCocoa5, bool interactiveInput)
{
	intrusive_ptr<Lexer> lexer;
	try {
		lexer = intrusive_ptr<Lexer>(new Lexer(lineProvider, warningLevel, warnAboutCocoa5, interactiveInput));
		intrusive_ptr<Parser> parser(new Parser(lexer));
		for(;;) {
			try	{
				if (interactiveInput)
					lexer->resetErrorCounts();
				Token t = lexer->getToken(parser->getStatus());
				if (t.getType()==TT_EOF)
					break;
				lexer->ungetToken(t);
				intrusive_ptr<const Statement> s = parser->parseStatement();
				if (interactiveInput) {
					int nErrors = lexer->getErrorCount();
					int nWarnings = lexer->getWarningCount();
					if (nErrors || nWarnings)
						printErrors(nErrors, nWarnings);
					else
						cout << "Ok.\n";
//					if (s) {
//						cout << "Parsed statement = " << s << '\n';
//						cout << "Parsed tree =" << '\n';
//						s->dumpAsTree(cout);
//					}
//					cout << endl;
				}
			} catch (const UnexpectedTokenException &ete) {
				lexer->reportError(ete.reason, ete.found.getBegin(), ete.found.getEnd());
				if (ete.found.getType()==TT_EOF)
					return EXIT_FAILURE;
				if (ete.needsRecovery)
					parser->tryToRecover(ete.from);
			} catch (const ExceptionWithSourcePosition &ewsp) {
				lexer->reportError(ewsp.reason, ewsp.from, ewsp.to);
				if (ewsp.needsRecovery)
					parser->tryToRecover(ewsp.from);
			}
		}
		if (!interactiveInput)
			printErrors(lexer->getErrorCount(), lexer->getWarningCount());
		return EXIT_SUCCESS;
	} catch (const BaseException &e) {
		cerr << e.reason << '\n';
		if (lexer)
			printErrors(lexer->getErrorCount(), lexer->getWarningCount());
	} catch (const std::exception &e) {
		cerr << "Fatal error: " << e.what() << '\n';
		if (lexer)
			printErrors(lexer->getErrorCount(), lexer->getWarningCount());
	}
	return EXIT_FAILURE;
}

void printUsage(char *progname) {
	cout << "Usage:\n" << progname << " [<option> ... <option>] [<filename> ... <filename>]\n"
		"Options:\n"
		"-getline             Uses ANSI getline instead of GNU readline\n"
		"-w0, -w1, -w2, -w3   Sets the warning level: -w0 none, -w1 normal, -w2 pedantic (the default), -w3 extra-pedantic\n"
		"-wc5                 Warns about features that may be removed in CoCoA 5\n"
		"-help                Prints this help\n"
		"\nIf no input filenames are given, then an interactive session is started\n";
}

int main(int argc, char **argv)
{
	cout << "CoCoA-5 alpha, for testing purposes only, compiled on " << __DATE__ << " " << __TIME__ << ".\nPlease, DO NOT DISTRIBUTE this file.\n\n";
	bool useGetline=false;
	bool warnAboutCocoa5=false;
	WarningSeverity warningLevel = WS_PEDANTIC;
	int paramsIndex=1;
	while (paramsIndex<argc) {
		string arg(argv[paramsIndex]);
		if (arg[0]!='-')
			break;
		++paramsIndex;
		if (arg=="-getline") {
			useGetline = true;
			continue;
		}
		if (arg=="-w0") {
			warningLevel = WS_NONE;
			continue;
		}
		if (arg=="-w1") {
			warningLevel = WS_NORMAL;
			continue;
		}
		if (arg=="-w2") {
			warningLevel = WS_PEDANTIC;
			continue;
		}
		if (arg=="-w3") {
			warningLevel = WS_EXTRA_PEDANTIC;
			continue;
		}
		if (arg=="-wc5") {
			warnAboutCocoa5 = true;
			continue;
		}
		printUsage(*argv);
		if (arg=="-help" || arg=="--help" || arg=="-?" || arg=="--?")
			return EXIT_SUCCESS;
		cerr << "\nUnknown option: " << arg << endl;
		return EXIT_FAILURE;
	}
	if (paramsIndex==argc) {
		cout << "No filenames given, starting interactive session (run with --help to see all the options)\n";
		return process(intrusive_ptr<LineProvider>(useGetline ?
			static_cast<LineProvider *>(new GetlineLineProvider()) :
			static_cast<LineProvider *>(new ReadlineLineProvider())), warningLevel, warnAboutCocoa5, true);
	}
	while (paramsIndex<argc) {
		cout <<
		    "\n\n"
			"##########################################################################\n"
			<< "Parsing file " << argv[paramsIndex] << '\n' <<
			"##########################################################################\n"
			<< endl;
		try {
			process(intrusive_ptr<LineProvider>(
				static_cast<LineProvider *>(new FileLineProvider(argv[paramsIndex++], false))), warningLevel, warnAboutCocoa5, false);
		} catch (const BaseException &e) {
			cerr << e.reason << '\n';
		} catch (const std::exception &e) {
			cerr << "Fatal error: " << e.what() << '\n';
		}
	}
	return EXIT_SUCCESS;
}
