//   Copyright (c) 2009 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
#include <exception>
#include <boost/scope_exit.hpp>

#include "Parser.H"

namespace CoCoA {
namespace Parser {

using namespace std;
using namespace boost;
using namespace CoCoA::AST;
using namespace CoCoA::Lexer;

namespace {
	WarningSeverity WSEndInsteadOfErrBlabla = WS_PEDANTIC;
}

Token Parser::expecting(TokenType expectedTT) {
	const Token token = lexer->getToken(status);
	const TokenType type = token.getType();
	if (type!=expectedTT) {
		lexer->ungetToken(token);
		throw UnexpectedTokenException(expectedTT, token);
	}
	return token;
}

std::string UnexpectedTokenException::humanReadable(TokenType type) {
	switch (type) {
	case TT_ALIAS: return "Alias";
	case TT_AND: return "And";
	case TT_ASSIGN: return "an assignment operator (:=)";
	case TT_BLOCK: return "Block";
	case TT_BREAK: return "Break";
	case TT_CART_PROD: return "the >< operator";
	case TT_CATCH: return "Catch";
	case TT_CIAO: return "Ciao";
	case TT_CLEAR: return "Clear";
	case TT_CLOSE_COMPAT: return "the backward-compatibility close marker";
	case TT_CLOSED_ROUND: return "a closed parenthesis";
	case TT_CLOSED_SQUARE: return "a closed square bracket";
	case TT_COLONCOLON: return "scope operator (::)";
	case TT_COLON: return "colon";
	case TT_COMMA: return "a comma";
	case TT_CONTINUE: return "Continue";
	case TT_DEFINE: return "Define";
	case TT_DEGLEX: return "DegLex";
	case TT_DEGREVLEX: return "DegRevLex";
	case TT_DELETE: return "Delete";
	case TT_DESCRIBE: return "Describe";
	case TT_DESTROY: return "Destroy";
	case TT_DO: return "Do";
	case TT_DOTDOT: return "the .. operator";
	case TT_DOT: return "a dot";
	case TT_ELLIPSIS: return "an ellipsis (...)";
	case TT_ELIF: return "Elif";
	case TT_ELIM: return "Elim";
	case TT_ELSE: return "Else";
	case TT_ENDALIAS: return "EndAlias";
	case TT_ENDBLOCK: return "EndBlock";
	case TT_ENDCATCH: return "EndCatch";
	case TT_ENDDEFINE: return "EndDefine";
	case TT_ENDFOREACH: return "EndForeach";
	case TT_ENDFOR: return "EndFor";
	case TT_ENDIF: return "Endif";
	case TT_ENDPACKAGE: return "EndPackage";
	case TT_END: return "End";
	case TT_ENDUSING: return "EndUsing";
	case TT_ENDWHILE: return "EndWhile";
	case TT_EOF: return "<EOF>";
	case TT_EQUAL: return "the = operator";
	case TT_FALSE: return "False";
	case TT_FLOAT_LITERAL: return "a constant floating-point";
	case TT_FOREACH: return "Foreach";
	case TT_FOR: return "For";
	case TT_GE: return "the >= operator";
	case TT_GT: return "the > operator";
	case TT_HELP: return "the help statement";
	case TT_IDENTIFIER: return "an identifier";
	case TT_ISDEFINED: return "IsDefined";
	case TT_IF: return "If";
	case TT_IN: return "In";
	case TT_INT_LITERAL: return "a constant integer";
	case TT_ISIN: return "IsIn";
	case TT_LE: return "the <= operator";
	case TT_LEX: return "Lex";
	case TT_LT: return "the < operator";
	case TT_MINUS: return "the - operator";
	case TT_MOD: return "the % operator";
	case TT_NOTEQUAL: return "the <> operator";
	case TT_NOT: return "Not";
	case TT_ON: return "On";
	case TT_OPEN_COMPAT: return "the backward-compatibility open marker";
	case TT_OPEN_ROUND: return "an open parenthesis";
	case TT_OPEN_SQUARE: return "an open square bracket";
	case TT_ORD: return "Ord";
	case TT_OR: return "Or";
	case TT_PACKAGE: return "Package";
	case TT_PACKAGENAME: return "a package name";
	case TT_PIPE: return "a pipe (|)";
	case TT_PLUS: return "the + operator";
	case TT_POSTO: return "PosTo";
	case TT_POWER: return "the ^ operator";
	case TT_PRINTLN: return "PrintLn";
	case TT_PRINT: return "Print";
	case TT_PROTECT: return "Protect";
	case TT_QUIT: return "Quit";
	case TT_RECORD: return "Record";
	case TT_REPEAT: return "Repeat";
	case TT_RETURN: return "Return";
	case TT_RING_ASSIGN: return "a ring-assignment operator (::=)";
	case TT_SEMICOLON: return "a semicolon";
	case TT_SET: return "Set";
	case TT_SKIP: return "Skip";
	case TT_SLASH: return "the / operator";
	case TT_SOURCE: return "Source";
	case TT_SOURCE_AS_LSHIFT: return "the source operator (<<)";
	case TT_STEP: return "Step";
	case TT_STAR: return "the * operator";
	case TT_STRING_LITERAL: return "a constant string";
	case TT_THEN: return "Then";
	case TT_TIME: return "Time";
	case TT_TO: return "To";
	case TT_TOPOS: return "ToPos";
	case TT_TRUE: return "True";
	case TT_UNPROTECT: return "Unprotect";
	case TT_UNSET: return "Unset";
	case TT_UNTIL: return "Until";
	case TT_USE: return "Use";
	case TT_USING: return "Using";
	case TT_VAR: return "Var";
	case TT_WEIGHTS: return "Weights";
	case TT_WHILE: return "While";
	case TT_XEL: return "Xel";
	}
	assert(false);
	throw FailedAssertionException("Unknown token type");
}

void Parser::reportSkippedInput(const CharPointer &from, const CharPointer &to) {
	if (from!=to)
		lexer->reportWarning("Skipping some input symbols while trying to continue the parsing after the error", from, to, WS_EXTRA_PEDANTIC);
	lastRecoveredPosition = to;
}

namespace {
	// the semicolon is rather special: it might start a statement, but only the empty one!
	// So, having the semicolon here would do more harm than good
	bool mayStartOrFollowAStatement(const TokenType tt) {
		switch (tt) {
		case TT_ALIAS:
		case TT_BLOCK:
		case TT_BREAK:
		case TT_CATCH:
		case TT_CIAO:
		case TT_CLEAR:
		case TT_DEFINE:
		case TT_DELETE:
		case TT_DESCRIBE:
		case TT_DESTROY:
		case TT_ELIF:
		case TT_ELSE:
		case TT_END:
		case TT_ENDALIAS:
		case TT_ENDBLOCK:
		case TT_ENDCATCH:
		case TT_ENDDEFINE:
		case TT_ENDFOR:
		case TT_ENDFOREACH:
		case TT_ENDIF:
		case TT_ENDPACKAGE:
		case TT_ENDUSING:
		case TT_ENDWHILE:
		case TT_FOR:
		case TT_FOREACH:
		case TT_HELP:
		case TT_IF:
		case TT_PACKAGE:
		case TT_PRINT:
		case TT_PRINTLN:
		case TT_QUIT:
		case TT_REPEAT:
		case TT_RETURN:
		case TT_SET:
		case TT_SKIP:
		case TT_SOURCE:
		case TT_SOURCE_AS_LSHIFT:
		case TT_UNSET:
		case TT_USE:
		case TT_USING:
		case TT_WHILE:
			return true;
		default:
			return false;
		}
	}
}

namespace {
	string aSemicolonShouldBeInsertedHere("A semicolon should be inserted here");
}

Token Parser::getTokenWithVirtualSemicolon(const CharPointer &position) {
	const Token token = lexer->getToken(status);
	switch (token.getType()) {
	case TT_ELIF:
	case TT_ELSE:
	case TT_END:
	case TT_ENDALIAS:
	case TT_ENDBLOCK:
	case TT_ENDCATCH:
	case TT_ENDDEFINE:
	case TT_ENDFOR:
	case TT_ENDFOREACH:
	case TT_ENDIF:
	case TT_ENDPACKAGE:
	case TT_ENDUSING:
	case TT_ENDWHILE:
	case TT_IN:
		lexer->ungetToken(token);
		lexer->reportWarning(aSemicolonShouldBeInsertedHere, position, token.getBegin(), WS_EXTRA_PEDANTIC);
		return Token(position, position, TT_SEMICOLON);
	default:
		return token;
	}
}

Token Parser::expectingSemicolon(const CharPointer &position, const char * const errorMsg) {
	const Token token = lexer->getToken(status);
	switch (token.getType()) {
	case TT_SEMICOLON:
		return token;
	case TT_ELIF:
	case TT_ELSE:
	case TT_END:
	case TT_ENDALIAS:
	case TT_ENDBLOCK:
	case TT_ENDCATCH:
	case TT_ENDDEFINE:
	case TT_ENDFOR:
	case TT_ENDFOREACH:
	case TT_ENDIF:
	case TT_ENDPACKAGE:
	case TT_ENDUSING:
	case TT_ENDWHILE:
	case TT_IN:
		lexer->ungetToken(token);
		lexer->reportWarning(aSemicolonShouldBeInsertedHere, position, token.getBegin(), WS_EXTRA_PEDANTIC);
		return Token(position, position, TT_SEMICOLON);
	default:
		lexer->ungetToken(token);
		if (errorMsg)
			throw UnexpectedTokenException(errorMsg, token);
		throw UnexpectedTokenException(TT_SEMICOLON, token);
	}
}

void Parser::tryToRecover(const CharPointer &beginPosition) {
	lexer->resetCompatibilityMode();
	status.inRecoveryMode = true;
	BOOST_SCOPE_EXIT( (&status) ) {
		status.inRecoveryMode = false;
	} BOOST_SCOPE_EXIT_END
	CharPointer endPosition = beginPosition;
	for(;;) {
		try	{
			const Token token = lexer->getToken(status);
			switch (TokenType tt = token.getType()) {
			case TT_SEMICOLON:
				endPosition = token.getEnd();
				// no break here
			case TT_EOF:
				reportSkippedInput(beginPosition, endPosition);
				return;
			default:
				if (mayStartOrFollowAStatement(tt) && endPosition!=lastRecoveredPosition) {
					lexer->ungetToken(token);
					reportSkippedInput(beginPosition, endPosition);
					return;
				}
			}
			endPosition = token.getEnd();
		} catch (const AskingForNewInteractiveInputDuringRecoveryException &) {
			// we've consumed all (interactive) input after an error, we're good to go
			reportSkippedInput(beginPosition, endPosition);
			return;
		} catch (const LexerException &pe) {
			// we ignore all lexer errors (for instance, unbalanced compatibility-parentheses) until we're synchronized
		}
	}
}

intrusive_ptr<const IdOrVar> Parser::parseIdOrVar() {
	return parseIdOrVar(lexer->getToken(status));
}

intrusive_ptr<const IdOrVar> Parser::parseIdOrVar(const Token &tokIdOrVar) {
	TokenType tt = tokIdOrVar.getType();
	if (tt==TT_IDENTIFIER)
		return intrusive_ptr<const IdOrVar>(new Identifier(tokIdOrVar));
	if (tt==TT_VAR) {
		expecting(TT_OPEN_ROUND);
		intrusive_ptr<const Expression> exp = parseExpression();
		Token closedRound = expecting(TT_CLOSED_ROUND);
		lexer->reportWarningMayBeRemoved("The keyword Var", tokIdOrVar.getBegin(), closedRound.getEnd());
		return intrusive_ptr<const IdOrVar>(new Var(exp, tokIdOrVar.getBegin(), closedRound.getEnd()));
	}
	throw UnexpectedTokenException("Expecting an identifier or Var(...)", tokIdOrVar);
}

intrusive_ptr<const UseStatement> Parser::parseUseStatement(const Token &tokUse) {
	intrusive_ptr<const IdOrVar> idOrVar;
	intrusive_ptr<const RingDefinition> ringDefinition;
	Token t = lexer->getToken(status);
	const string lexeme(t.lexeme());
	TokenType tt = t.getType();
	if (tt==TT_IDENTIFIER) {
		Token lookahead = lexer->getToken(status);
		TokenType laTT = lookahead.getType();
		if (laTT==TT_SLASH || laTT==TT_OPEN_SQUARE) {
			lexer->ungetToken(lookahead);
			lexer->ungetToken(t);
			ringDefinition = parseRingDefinition();
		} else
			lexer->ungetToken(lookahead);
	}
	if (!ringDefinition) {
		if (tt==TT_IDENTIFIER || tt==TT_VAR) {
			idOrVar = parseIdOrVar(t);
			t = lexer->getToken(status);
			if (t.getType()!=TT_RING_ASSIGN)
				lexer->ungetToken(t);
			else
				ringDefinition = parseRingDefinition();
		} else {
			lexer->ungetToken(t);
			throw UnexpectedTokenException("Expecting an identifier or Var(...)", t);
		}
	}
	assert(idOrVar || ringDefinition);
	Token tokSemicolon = expectingSemicolon(ringDefinition ? ringDefinition->endSourcePosition : idOrVar->endSourcePosition);
	return intrusive_ptr<const UseStatement>(new UseStatement(tokUse.getBegin(), idOrVar, ringDefinition, tokSemicolon.getEnd()));
}

intrusive_ptr<const SetStatement> Parser::parseSetStatement(const Token &tokSet) {
	intrusive_ptr<const IdOrVar> idOrVar = parseIdOrVar();
	intrusive_ptr<const Expression> exp;
	Token t = lexer->getToken(status);
	if (t.getType()==TT_ASSIGN)
		exp = parseExpression();
	else
		lexer->ungetToken(t);
	Token tokSemicolon = expectingSemicolon(exp ? exp->endSourcePosition : idOrVar->endSourcePosition);
	lexer->reportWarningMayBeRemoved("The keywords Set/Unset", tokSet.getBegin(), tokSemicolon.getEnd());
	return intrusive_ptr<const SetStatement>(new SetStatement(tokSet.getBegin(), idOrVar, exp, tokSemicolon.getEnd()));
}

intrusive_ptr<const UnsetStatement> Parser::parseUnsetStatement(const Token &tokUnset) {
	intrusive_ptr<const IdOrVar> idOrVar = parseIdOrVar();
	Token tokSemicolon = expectingSemicolon(idOrVar->endSourcePosition);
	lexer->reportWarningMayBeRemoved("The keywords Set/Unset", tokUnset.getBegin(), tokSemicolon.getEnd());
	return intrusive_ptr<const UnsetStatement>(new UnsetStatement(tokUnset.getBegin(), idOrVar, tokSemicolon.getEnd()));
}

intrusive_ptr<const PrintStatement> Parser::parsePrintStatement(const Token &tokPrintOrPrintln) {
	vector<intrusive_ptr<const Expression> > exps;
	CharPointer position = tokPrintOrPrintln.getEnd();
	Token t = getTokenWithVirtualSemicolon(position);
	TokenType tt = t.getType();
	if (tt!=TT_ON && tt!=TT_SEMICOLON) {
		lexer->ungetToken(t);
		for(;;) {
			intrusive_ptr<const Expression> e = parseExpression();
			position = e->endSourcePosition;
			exps.push_back(e);
			t = getTokenWithVirtualSemicolon(position);
			tt = t.getType();
			if (tt!=TT_COMMA)
				break;
		}
	}
	if (tt==TT_ON) {
		intrusive_ptr<const Expression> onExp = parseExpression();
		Token tokSemicolon = expectingSemicolon(onExp->endSourcePosition);
		return intrusive_ptr<const PrintStatement>(new PrintStatement(tokPrintOrPrintln, exps, onExp, tokSemicolon.getEnd()));
	}
	if (tt!=TT_SEMICOLON)
		throw UnexpectedTokenException("Expecting an \"On\" or a semicolon to end the Print(Ln) statement", t);
	return intrusive_ptr<const PrintStatement>(new PrintStatement(tokPrintOrPrintln, exps, intrusive_ptr<const Expression>(), t.getEnd()));
}

intrusive_ptr<const ForeachStatement> Parser::parseForeachStatement(const Token &tokForeach) {
	Token id = expecting(TT_IDENTIFIER);
	expecting(TT_IN);
	intrusive_ptr<const Expression> inExp = parseExpression();
	expecting(TT_DO);
	enteringBlock("Foreach");
	++nNestingLoops;
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		--This.nNestingLoops;
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokForeach; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDFOREACH)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndForeach", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const ForeachStatement>(new ForeachStatement(tokForeach.getBegin(), id, inExp, statements, t.getEnd()));
}

intrusive_ptr<const BlockStatement> Parser::parseBlockStatement(const Token &tokBlock) {
	enteringBlock("Block");
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokBlock; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDBLOCK)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndBlock", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const BlockStatement>(new BlockStatement(tokBlock.getBegin(), statements, t.getEnd()));
}

intrusive_ptr<const PackageStatement> Parser::parsePackageStatement(const Token &tokPackage) {
	Token tokPackageName = expecting(TT_PACKAGENAME);
	if (tokPackageName.lexeme()=="$")
		throw ExceptionWithSourcePosition("Invalid package name", tokPackageName.getBegin(), tokPackageName.getEnd(), false);
	if (insideDefine)
		throw ExceptionWithSourcePosition("Packages cannot be declared inside function declarations", tokPackage.getBegin(), tokPackageName.getEnd(), false);
	if (!status.openBlocks.empty())
		throw ExceptionWithSourcePosition("Packages must be declared at top-level", tokPackage.getBegin(), tokPackageName.getEnd(), false);
	insidePackage = true;
	enteringBlock("Package");
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		This.insidePackage = false;
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokPackage; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDPACKAGE)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndPackage", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const PackageStatement>(new PackageStatement(tokPackage.getBegin(), tokPackageName.lexeme(), statements, t.getEnd()));
}

intrusive_ptr<const CatchStatement> Parser::parseCatchStatement(const Token &tokCatch) {
	enteringBlock("Catch");
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	intrusive_ptr<const Expression> leftValue;
	Token t = tokCatch; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_IN) {
			if (leftValue)
				throw UnexpectedTokenException("Expecting \"EndCatch\" (you've already specified an In-clause)", t);
			leftValue = parseExpression();
			if (!leftValue->isLeftValue())
				lexer->reportError("A left-value (an identifier, a field access or an indexed-access) is required here", leftValue->beginSourcePosition, leftValue->endSourcePosition);
			continue;
		}
		if (tt==TT_ENDCATCH)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndCatch", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const CatchStatement>(new CatchStatement(tokCatch.getBegin(), statements, leftValue, t.getEnd()));
}

vector<Param> Parser::parseParamList() {
	vector<Param> params;
	for(;;) {
		Param p;
		Token t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_CLOSED_ROUND) {
			lexer->ungetToken(t);
			return params;
		}
		if (tt==TT_VAR) {
			p.byRef = true;
			p.name = expecting(TT_IDENTIFIER).lexeme();
		} else if (tt==TT_IDENTIFIER) {
			p.name = t.lexeme();
		} else
			throw UnexpectedTokenException("Expecting an identifier or a closed round parenthesis", t);
		params.push_back(p);
		t = lexer->getToken(status);
		if (t.getType()!=TT_COMMA) {
			lexer->ungetToken(t);
			return params;
		}
	}
}

intrusive_ptr<const DefineStatement> Parser::parseDefineStatement(const Token &tokDefine) {
	Token tokIdentifier = expecting(TT_IDENTIFIER);
	string identifier = tokIdentifier.lexeme();
	if (insideDefine)
		throw ExceptionWithSourcePosition("Function declarations cannot be nested", tokDefine.getBegin(), tokIdentifier.getEnd(), false);
	if (!(status.openBlocks.empty() || (status.openBlocks.size()==1 && insidePackage)))
		throw ExceptionWithSourcePosition("Function declarations must be declared at top-level or inside a package", tokDefine.getBegin(), tokIdentifier.getEnd(), false);
	enteringBlock("Define");
	insideDefine = true;
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		This.insideDefine = false;
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	expecting(TT_OPEN_ROUND);
	vector<Param> params;
	bool thereIsEllipsis = false;
	Token t = lexer->getToken(status);
	if (t.getType()==TT_ELLIPSIS) {
		thereIsEllipsis = true;
	} else {
		lexer->ungetToken(t);
		params = parseParamList();
	}
	expecting(TT_CLOSED_ROUND);
	vector<intrusive_ptr<const Statement> > statements;
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDDEFINE)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndDefine", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const DefineStatement>(new DefineStatement(tokDefine.getBegin(), identifier, thereIsEllipsis, params, statements, t.getEnd()));
}

intrusive_ptr<const AliasStatement> Parser::parseAliasStatement(const Token &tokAlias) {
	CharPointer position = tokAlias.getEnd();
	vector<Binding> bindingList = parseBindingList(position);
	bool thereIsIn = false;
	vector<intrusive_ptr<const Statement> > statements;
	Token t = getTokenWithVirtualSemicolon(position);
	TokenType tt = t.getType();
	if (tt!=TT_SEMICOLON) {
		if (tt!=TT_IN)
			throw UnexpectedTokenException("Binding lists must be followed by \"In\" or a semicolon", t);
		thereIsIn = true;
		enteringBlock("Alias");
		Parser &This = *this;
		BOOST_SCOPE_EXIT( (&This) ) {
			This.exitingBlock();
		} BOOST_SCOPE_EXIT_END
		for(;;) {
			status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
			t = lexer->getToken(status);
			tt = t.getType();
			if (tt==TT_ENDALIAS)
				break;
			if (tt==TT_END) {
				lexer->reportWarning("Interpreting End as EndAlias", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
				break;
			}
			lexer->ungetToken(t);
			intrusive_ptr<const Statement> stmt = tryParseStatement();
			if (stmt)
				statements.push_back(stmt);
		}
	}
	return intrusive_ptr<const AliasStatement>(new AliasStatement(tokAlias.getBegin(), bindingList, thereIsIn, statements, t.getEnd()));
}

intrusive_ptr<const WhileStatement> Parser::parseWhileStatement(const Token &tokWhile) {
	intrusive_ptr<const Expression> exp = parseExpression();
	expecting(TT_DO);
	enteringBlock("While");
	++nNestingLoops;
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		--This.nNestingLoops;
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokWhile; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDWHILE)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndWhile", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const WhileStatement>(new WhileStatement(tokWhile.getBegin(), exp, statements, t.getEnd()));
}

intrusive_ptr<const RepeatUntilStatement> Parser::parseRepeatUntilStatement(const Token &tokRepeat) {
	enteringBlock("Repeat");
	++nNestingLoops;
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		--This.nNestingLoops;
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokRepeat; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_UNTIL)
			break;
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	intrusive_ptr<const Expression> exp = parseExpression();
	t = expectingSemicolon(exp->endSourcePosition);
	return intrusive_ptr<const RepeatUntilStatement>(new RepeatUntilStatement(tokRepeat.getBegin(), statements, exp, t.getEnd()));
}

void Parser::enteringBlock(std::string blockName) {
	vector<string> &blocks = status.openBlocks;
	if (blocks.empty())
		blocks.push_back(blockName + " ");
	else {
		string current = blocks.back();
		blocks.push_back(current + ">> " + blockName + " ");
	}
}

void Parser::exitingBlock() {
	status.openBlocks.pop_back();
}

intrusive_ptr<const ForStatement> Parser::parseForStatement(const Token &tokFor) {
	Token id = expecting(TT_IDENTIFIER);
	expecting(TT_ASSIGN);
	intrusive_ptr<const Expression> beginExp = parseExpression();
	expecting(TT_TO);
	intrusive_ptr<const Expression> endExp = parseExpression();
	intrusive_ptr<const Expression> stepExp;
	Token maybeTokStep = lexer->getToken(status);
	if (maybeTokStep.getType()==TT_STEP)
		stepExp = parseExpression();
	else
		lexer->ungetToken(maybeTokStep);
	expecting(TT_DO);
	enteringBlock("For");
	++nNestingLoops;
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		--This.nNestingLoops;
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokFor; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDFOR)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndFor", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const ForStatement>(new ForStatement(tokFor.getBegin(), id, beginExp, endExp, stepExp, statements, t.getEnd()));
}

intrusive_ptr<const UsingStatement> Parser::parseUsingStatement(const Token &tokUsing) {
	const intrusive_ptr<const IdOrVar> idOrVar = parseIdOrVar();
	expecting(TT_DO);
	enteringBlock("Using");
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This) ) {
		This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	vector<intrusive_ptr<const Statement> > statements;
	Token t = tokUsing; // because we cannot default-construct it
	for(;;) {
		status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
		t = lexer->getToken(status);
		TokenType tt = t.getType();
		if (tt==TT_ENDUSING)
			break;
		if (tt==TT_END) {
			lexer->reportWarning("Interpreting End as EndUsing", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
			break;
		}
		lexer->ungetToken(t);
		intrusive_ptr<const Statement> stmt = tryParseStatement();
		if (stmt)
			statements.push_back(stmt);
	}
	return intrusive_ptr<const UsingStatement>(new UsingStatement(tokUsing.getBegin(), idOrVar, statements, t.getEnd()));
}

intrusive_ptr<const ReturnStatement> Parser::parseReturnStatement(const Token &tokReturn) {
	Token t = getTokenWithVirtualSemicolon(tokReturn.getEnd());
	intrusive_ptr<const Expression> exp;
	if (t.getType()!=TT_SEMICOLON) {
		lexer->ungetToken(t);
		exp = parseExpression();
	}
	Token tokSemicolon = expectingSemicolon(exp ? exp->endSourcePosition:tokReturn.getEnd());
	if (!insideDefine)
		throw ExceptionWithSourcePosition("Return statements can be only used inside functions", tokReturn.getBegin(), tokSemicolon.getEnd(), false);
	return intrusive_ptr<const ReturnStatement>(new ReturnStatement(tokReturn.getBegin(), exp, tokSemicolon.getEnd()));
}

intrusive_ptr<const SourceStatement> Parser::parseSourceStatement(const Token &tokSource) {
	intrusive_ptr<const Expression> exp = parseExpression();
	Token tokSemicolon = expectingSemicolon(exp->endSourcePosition);
	return intrusive_ptr<const SourceStatement>(new SourceStatement(tokSource, exp, tokSemicolon.getEnd()));
}

intrusive_ptr<const ClearStatement> Parser::parseClearStatement(const Token &tokClear) {
	CharPointer position = tokClear.getEnd();
	const vector<intrusive_ptr<const IdOrVar> > list = parseIdOrVarList(position);
	Token tokSemicolon = expectingSemicolon(position);
	return intrusive_ptr<const ClearStatement>(new ClearStatement(tokClear.getBegin(), list, tokSemicolon.getEnd()));
}

intrusive_ptr<const DeleteStatement> Parser::parseDeleteStatement(const Token &tokDelete) {
	CharPointer position = tokDelete.getEnd();
	const vector<intrusive_ptr<const IdOrVar> > list = parseIdOrVarList(position);
	Token tokSemicolon = expectingSemicolon(position);
	return intrusive_ptr<const DeleteStatement>(new DeleteStatement(tokDelete.getBegin(), list, tokSemicolon.getEnd()));
}

intrusive_ptr<const DestroyStatement> Parser::parseDestroyStatement(const Token &tokDestroy) {
	CharPointer position = tokDestroy.getEnd();
	const vector<intrusive_ptr<const IdOrVar> > list = parseIdOrVarList(position);
	Token tokSemicolon = expectingSemicolon(position);
	return intrusive_ptr<const DestroyStatement>(new DestroyStatement(tokDestroy.getBegin(), list, tokSemicolon.getEnd()));
}

intrusive_ptr<const DescribeStatement> Parser::parseDescribeStatement(const Token &tokDescribe) {
	intrusive_ptr<const Expression> exp = parseExpression();
	Token tokSemicolon = expectingSemicolon(exp->endSourcePosition);
	return intrusive_ptr<const DescribeStatement>(new DescribeStatement(tokDescribe.getBegin(), exp, tokSemicolon.getEnd()));
}

intrusive_ptr<const BreakStatement> Parser::parseBreakStatement(const Token &tokBreak) {
	Token tokSemicolon = expectingSemicolon(tokBreak.getEnd());
	if (!nNestingLoops)
		throw ExceptionWithSourcePosition("Break (and Continue) statements can be only used inside loops", tokBreak.getBegin(), tokSemicolon.getEnd(), false);
	return intrusive_ptr<const BreakStatement>(new BreakStatement(tokBreak.getBegin(), tokSemicolon.getEnd()));
}

intrusive_ptr<const ContinueStatement> Parser::parseContinueStatement(const Token &tokContinue) {
	Token tokSemicolon = expectingSemicolon(tokContinue.getEnd());
	if (!nNestingLoops)
		throw ExceptionWithSourcePosition("Continue (and Break) statements can be only used inside loops", tokContinue.getBegin(), tokSemicolon.getEnd(), false);
	return intrusive_ptr<const ContinueStatement>(new ContinueStatement(tokContinue.getBegin(), tokSemicolon.getEnd()));
}

intrusive_ptr<const SkipStatement> Parser::parseSkipStatement(const Token &tokSkip) {
	Token tokSemicolon = expectingSemicolon(tokSkip.getEnd());
	return intrusive_ptr<const SkipStatement>(new SkipStatement(tokSkip.getBegin(), tokSemicolon.getEnd()));
}

intrusive_ptr<const CiaoOrQuitStatement> Parser::parseCiaoOrQuitStatement(const Token &tokCiaoOrQuit) {
	Token tokSemicolon = expectingSemicolon(tokCiaoOrQuit.getEnd());
	return intrusive_ptr<const CiaoOrQuitStatement>(new CiaoOrQuitStatement(tokCiaoOrQuit, tokSemicolon.getEnd()));
}

intrusive_ptr<const IfStatement> Parser::parseIfStatement(const Token &tokIf) {
	vector<IfBranch> branches;
	bool foundElse = false;
	bool firstTime = true;
	Parser &This = *this;
	BOOST_SCOPE_EXIT( (&This)(&firstTime) ) {
		if (!firstTime)
			This.exitingBlock();
	} BOOST_SCOPE_EXIT_END
	Token t = tokIf; // because we cannot default-construct it
	TokenType tt = TT_IF; // anything != TT_END, TT_ENDIF (to enter the loop) and TT_ELSE (to read the expression) would work
	while (tt!=TT_END && tt!=TT_ENDIF) {
		IfBranch branch;
		if (tt!=TT_ELSE) {
			branch.optExp = parseExpression();
			expecting(TT_THEN);
			if (firstTime) {
				firstTime = false;
				enteringBlock("If");
			}
		}
		for(;;) {
			status.inTheMiddleOfAStatement = false; // plainly false, but it gives a better feedback
			t = lexer->getToken(status);
			tt = t.getType();
			if (tt==TT_ENDIF || tt==TT_ELIF)
				break;
			if (tt==TT_END) {
				lexer->reportWarning("Interpreting End as EndIf", t.getBegin(), t.getEnd(), WSEndInsteadOfErrBlabla);
				break;
			}
			if (tt==TT_ELSE) {
				 if (foundElse)
					 throw UnexpectedTokenException("An If statement cannot have more than one Else branch, I was expecting Endif", t);
				 foundElse = true;
				 break;
			}
			lexer->ungetToken(t);
			intrusive_ptr<const Statement> stmt = tryParseStatement();
			if (stmt)
				branch.statements.push_back(stmt);
		}
		branches.push_back(branch);
	}
	return intrusive_ptr<const IfStatement>(new IfStatement(tokIf.getBegin(), branches, t.getEnd()));
}

intrusive_ptr<const Statement> Parser::tryParseStatement() {
	if (lexer->isInteractive())
		return parseStatement();
	try {
		return parseStatement();
	} catch (const UnexpectedTokenException &ete) {
		const Token &foundToken = ete.found;
		if (foundToken.getType()==TT_EOF)
			throw;
		lexer->reportError(ete.reason, foundToken.getBegin(), foundToken.getEnd());
		tryToRecover(foundToken.getBegin());
		return intrusive_ptr<const Statement>();
	} catch (const ExceptionWithSourcePosition &ewsp) {
		lexer->reportError(ewsp.reason, ewsp.from, ewsp.to);
		if (ewsp.needsRecovery)
			tryToRecover(ewsp.from);
		return intrusive_ptr<const Statement>();
	}
}

intrusive_ptr<const Statement> Parser::parseStatement() {
	status.inTheMiddleOfAStatement = false; // not strictly true (it could be nested), but it gives a better feedback
	Token t = lexer->getToken(status);
	status.inTheMiddleOfAStatement = true;
	BOOST_SCOPE_EXIT( (&status) ) {
		status.inTheMiddleOfAStatement = false;
	} BOOST_SCOPE_EXIT_END
	switch (t.getType()) {
	case TT_SEMICOLON:
		//lexer->reportWarning("Empty statement", t.getBegin(), t.getEnd(), WS_PEDANTIC);
		return intrusive_ptr<const Statement>(new EmptyStatement(t.getBegin()));
	case TT_USE:
		return parseUseStatement(t);
	case TT_SET:
		return parseSetStatement(t);
	case TT_UNSET:
		return parseUnsetStatement(t);
	case TT_FOR:
		return parseForStatement(t);
	case TT_FOREACH:
		return parseForeachStatement(t);
	case TT_IF:
		return parseIfStatement(t);
	case TT_RETURN:
		return parseReturnStatement(t);
	case TT_USING:
		return parseUsingStatement(t);
	case TT_PRINT:
	case TT_PRINTLN:
		return parsePrintStatement(t);
	case TT_BREAK:
		return parseBreakStatement(t);
	case TT_CONTINUE:
		return parseContinueStatement(t);
	case TT_SKIP:
		return parseSkipStatement(t);
	case TT_QUIT:
	case TT_CIAO:
		return parseCiaoOrQuitStatement(t);
	case TT_DESCRIBE:
		return parseDescribeStatement(t);
	case TT_REPEAT:
		return parseRepeatUntilStatement(t);
	case TT_WHILE:
		return parseWhileStatement(t);
	case TT_CLEAR:
		return parseClearStatement(t);
	case TT_DELETE:
		return parseDeleteStatement(t);
	case TT_DESTROY:
		return parseDestroyStatement(t);
	case TT_ALIAS:
		return parseAliasStatement(t);
	case TT_BLOCK:
		return parseBlockStatement(t);
	case TT_CATCH:
		return parseCatchStatement(t);
	case TT_PACKAGE:
		return parsePackageStatement(t);
	case TT_DEFINE:
		return parseDefineStatement(t);
	case TT_SOURCE:
	case TT_SOURCE_AS_LSHIFT:
		return parseSourceStatement(t);
	case TT_HELP:
		return intrusive_ptr<const Statement>(new HelpStatement(t));
	case TT_TIME:
		{
			intrusive_ptr<const Statement> stmt = parseStatement();
			if (intrusive_ptr<const EvalStatement> evalStmt = dynamic_pointer_cast<const EvalStatement>(stmt))
				lexer->reportWarning("This use of Time statement is discouraged; please use an assignment to avoid this warning", evalStmt->beginSourcePosition, evalStmt->endSourcePosition, WS_NORMAL);
			return intrusive_ptr<const Statement>(new TimeStatement(stmt, t.getBegin()));
		}
	default:
		break; // apparently useless, but it shuts the compiler (blabla not handled in switch)
	}
	lexer->ungetToken(t);
	intrusive_ptr<const Expression> exp = parseExpression();
	const Token t2 = getTokenWithVirtualSemicolon(exp->endSourcePosition);
	const TokenType tt2 = t2.getType();
	if (tt2==TT_RING_ASSIGN) {
		intrusive_ptr<const RingDefinition> rdef = parseRingDefinition();
		Token tokSemicolon = expectingSemicolon(rdef->endSourcePosition, "Expecting a semicolon after a ring assignment");
		if (!exp->isLeftValue())
			lexer->reportError("Invalid left-value for ring-assignment", exp->beginSourcePosition, exp->endSourcePosition);
		return intrusive_ptr<const Statement>(new RingAssignStatement(exp, rdef, tokSemicolon.getEnd()));
	}
	if (tt2==TT_ASSIGN) {
		intrusive_ptr<const Expression> rightExp = parseExpression();
		Token tokSemicolon = expectingSemicolon(rightExp->endSourcePosition, "Expecting a semicolon after an assignment");
		if (!exp->isLeftValue())
			lexer->reportError("Invalid left-value for assignment", exp->beginSourcePosition, exp->endSourcePosition);
		return intrusive_ptr<const Statement>(new AssignStatement(exp, rightExp, tokSemicolon.getEnd()));
	}
	if (tt2!=TT_SEMICOLON)
		throw UnexpectedTokenException("Expecting a semicolon, to end the statement, or an assignment operator", t2);
	return intrusive_ptr<const Statement>(new EvalStatement(exp, t2.getEnd()));
}

intrusive_ptr<const RingDefinition> Parser::parseRingDefinition() {
	const Token tokRingIdentifier = expecting(TT_IDENTIFIER);
	Token lastToken = tokRingIdentifier;
	intrusive_ptr<const Expression> optionalExp;
	const Token maybeSlash = lexer->getToken(status);
	if (maybeSlash.getType()!=TT_SLASH)
		lexer->ungetToken(maybeSlash);
	else {
		expecting(TT_OPEN_ROUND);
		optionalExp = parseExpression();
		lastToken = expecting(TT_CLOSED_ROUND);
	}
	std::vector<boost::intrusive_ptr<const IndeterminateDeclaration> > indeterminates;
	const Token maybeOpenSquare = lexer->getToken(status);
	if (maybeOpenSquare.getType()!=TT_OPEN_SQUARE)
		lexer->ungetToken(maybeOpenSquare);
	else {
		for(;;) {
			indeterminates.push_back(parseIndeterminateDecl(false));
			const Token t = lexer->getToken(status);
			if (t.getType()!=TT_COMMA) {
				lexer->ungetToken(t);
				break;
			}
		}
		lastToken = expecting(TT_CLOSED_SQUARE);
	}
	for(;;) {
		Token maybeComma = lexer->getToken(status);
		if (maybeComma.getType()!=TT_COMMA) {
			lexer->ungetToken(maybeComma);
			break;
		} else {
			lastToken = lexer->getToken(status);
			// FIXME: to finish the handling (the AST knows nothing about this)
			switch (lastToken.getType()) {
			case TT_LEX:
			case TT_XEL:
			case TT_DEGREVLEX:
			case TT_DEGLEX:
			case TT_POSTO:
			case TT_TOPOS:
				break;
			case TT_ORD:
				expecting(TT_OPEN_ROUND);
				parseExpression();
				lastToken = expecting(TT_CLOSED_ROUND);
				break;
			case TT_ELIM: {
				expecting(TT_OPEN_ROUND);
				parseIndeterminateDecl(true);
				Token maybeDotdot = lexer->getToken(status);
				if (maybeDotdot.getType()==TT_DOTDOT)
					parseIndeterminateDecl(true);
				else
					lexer->ungetToken(maybeDotdot);
				lastToken = expecting(TT_CLOSED_ROUND);
				break;
			}
			case TT_WEIGHTS:
				expecting(TT_OPEN_ROUND);
				parseExpressionList();
				lastToken = expecting(TT_CLOSED_ROUND);
				break;
			default:
				throw UnexpectedTokenException("Expecting Lex, Xel, DegLex, DegRevLex or Ord", lastToken);
			}
		}
	}
	return intrusive_ptr<const RingDefinition>(new RingDefinition(tokRingIdentifier, optionalExp, indeterminates, lastToken.getEnd()));
}

intrusive_ptr<const IndeterminateDeclaration> Parser::parseIndeterminateDecl(bool isInsideElim) {
	const intrusive_ptr<const IdOrVar> idOrVar = parseIdOrVar();
	CharPointer endPosition = idOrVar->endSourcePosition;
	vector<pair <intrusive_ptr<const Expression>, intrusive_ptr<const Expression> > > ranges;
	Token t = lexer->getToken(status);
	if (t.getType()!=TT_OPEN_SQUARE)
		lexer->ungetToken(t);
	else {
		for(;;) {
			intrusive_ptr<const Expression> e1 = parseAdditiveExpression();
			intrusive_ptr<const Expression> e2;
			if (!isInsideElim) {
				expecting(TT_DOTDOT);
				e2 = parseAdditiveExpression();
			}
			ranges.push_back(make_pair(e1, e2));
			t = lexer->getToken(status);
			if (t.getType()!=TT_COMMA) {
				lexer->ungetToken(t);
				break;
			}
		}
		Token closedSquare = expecting(TT_CLOSED_SQUARE);
		endPosition = closedSquare.getEnd();
	}
	return intrusive_ptr<const IndeterminateDeclaration>(new IndeterminateDeclaration(idOrVar, ranges, endPosition));
}

vector<intrusive_ptr<const Expression> > Parser::parseExpressionList() {
	vector<intrusive_ptr<const Expression> > list;
	for(;;) {
		const intrusive_ptr<const Expression> exp = parseExpression();
		list.push_back(exp);
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_COMMA) {
			lexer->ungetToken(t);
			break;
		}
	}
	return list;
}

vector<intrusive_ptr<const IdOrVar> > Parser::parseIdOrVarList(CharPointer &position) {
	vector<intrusive_ptr<const IdOrVar> > list;
	for(;;) {
		const intrusive_ptr<const IdOrVar> idOrVar = parseIdOrVar();
		position = idOrVar->endSourcePosition;
		list.push_back(idOrVar);
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_COMMA) {
			lexer->ungetToken(t);
			break;
		}
	}
	return list;
}

vector<Binding> Parser::parseBindingList(CharPointer &position) {
	vector<Binding> list;
	for(;;) {
		Binding b;
		b.idOrVar = parseIdOrVar();
		expecting(TT_ASSIGN);
		Token pn = expecting(TT_PACKAGENAME);
		position = pn.getEnd();
		b.packageName = pn.lexeme();
		list.push_back(b);
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_COMMA) {
			lexer->ungetToken(t);
			break;
		}
	}
	return list;
}

intrusive_ptr<const Expression> Parser::parseExpression() {
	const Token t = lexer->getToken(status);
	if (t.getType()==TT_OPEN_COMPAT) {
		const intrusive_ptr<const Expression> exp = parseExpression();
		const Token t2=expecting(TT_CLOSE_COMPAT);
		return intrusive_ptr<const Expression>(new BackwardCompatibleExp(exp, t.getBegin(), t2.getEnd()));
	}
	lexer->ungetToken(t);
	return parseConditionalOrExpression();
}

intrusive_ptr<const Expression> Parser::parseConditionalOrExpression() {
	intrusive_ptr<const Expression> returnValue = parseConditionalAndExpression();
	for(;;) {
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_OR) {
			lexer->ungetToken(t);
			break;
		}
		intrusive_ptr<const Expression> exp = parseExpression();
		returnValue = intrusive_ptr<const Expression>(new OrExpression(returnValue, exp));
	}
	return returnValue;
}

intrusive_ptr<const Expression> Parser::parseConditionalAndExpression() {
	intrusive_ptr<const Expression> returnValue = parseEqualityExpression();
	for(;;) {
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_AND) {
			lexer->ungetToken(t);
			break;
		}
		const intrusive_ptr<const Expression> exp = parseEqualityExpression();
		returnValue = intrusive_ptr<const Expression>(new AndExpression(returnValue, exp));
	}
	return returnValue;
}

intrusive_ptr<const Expression> Parser::parseEqualityExpression() {
	intrusive_ptr<const Expression> returnValue = parseRelationalExpression();
	for(;;) {
		const Token t = lexer->getToken(status);
		const TokenType type = t.getType();
		if (type!=TT_EQUAL && type!=TT_NOTEQUAL && type!=TT_ISIN) {
			lexer->ungetToken(t);
			break;
		}
		const intrusive_ptr<const Expression> exp = parseRelationalExpression();
		switch (type) {
		case TT_EQUAL:
			returnValue = intrusive_ptr<const Expression>(new EqualExpression(returnValue, exp));
			break;
		case TT_NOTEQUAL:
			returnValue = intrusive_ptr<const Expression>(new NotEqualExpression(returnValue, exp));
			break;
		case TT_ISIN:
			returnValue = intrusive_ptr<const Expression>(new IsInExpression(returnValue, exp));
			break;
		default:
			assert(false);
		}
	}
	return returnValue;
}

intrusive_ptr<const Expression> Parser::parseRelationalExpression() {
	intrusive_ptr<const Expression> returnValue = parseCartesianProductExpression();
	for(;;) {
		const Token t = lexer->getToken(status);
		const TokenType type = t.getType();
		if (type!=TT_LE && type!=TT_LT && type!=TT_GE && type!=TT_GT) {
			lexer->ungetToken(t);
			break;
		}
		const intrusive_ptr<const Expression> exp = parseCartesianProductExpression();
		switch (type) {
		case TT_LE:
			returnValue = intrusive_ptr<const Expression>(new LessOrEqualExpression(returnValue, exp));
			break;
		case TT_LT:
			returnValue = intrusive_ptr<const Expression>(new LessThanExpression(returnValue, exp));
			break;
		case TT_GE:
			returnValue = intrusive_ptr<const Expression>(new GreaterOrEqualExpression(returnValue, exp));
			break;
		case TT_GT:
			returnValue = intrusive_ptr<const Expression>(new GreaterThanExpression(returnValue, exp));
			break;
		default:
			assert(false);
		}
	}
	return returnValue;
}

intrusive_ptr<const Expression> Parser::parseCartesianProductExpression() {
	intrusive_ptr<const Expression> exp = parseListExpression();
	const Token t = lexer->getToken(status);
	if (t.getType()!=TT_CART_PROD) {
		lexer->ungetToken(t);
		return exp;
	}
	vector<intrusive_ptr<const Expression> > operands;
	operands.push_back(exp);
	for(;;) {
		exp = parseListExpression();
		operands.push_back(exp);
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_CART_PROD) {
			lexer->ungetToken(t);
			break;
		}
	}
	return intrusive_ptr<const Expression>(new CartesianProductExpression(operands));
}

intrusive_ptr<const Expression> Parser::parseListExpression() {
	const intrusive_ptr<const Expression> exp = parseAdditiveExpression();
	const Token t = lexer->getToken(status);
	if (t.getType()!=TT_DOTDOT) {
		lexer->ungetToken(t);
		return exp;
	}
	const intrusive_ptr<const Expression> exp2 = parseAdditiveExpression();
	return intrusive_ptr<const Expression>(new DotDotExpression(exp, exp2));
}

intrusive_ptr<const Expression> Parser::parseListPrimary(const Token &tokOpenSquare) {
	Token t = lexer->getToken(status);
	if (t.getType()==TT_CLOSED_SQUARE)
		return intrusive_ptr<const Expression>(new EmptyListExpression(tokOpenSquare.getBegin(), t.getEnd()));
	if (t.getType()==TT_IDENTIFIER) {
		const Token tokId = t;
		t = lexer->getToken(status);
		if (t.getType()==TT_IN) {
			const intrusive_ptr<const Expression> exp1 = parseExpression();
			expecting(TT_PIPE);
			const intrusive_ptr<const Expression> exp2 = parseExpression();
			const Token tokClosedSquare = expecting(TT_CLOSED_SQUARE);
			return intrusive_ptr<const Expression>(new IdInExpSuchThatExp(tokOpenSquare.getBegin(), tokClosedSquare.getEnd(), tokId.lexeme(), exp1, exp2));
		}
		lexer->ungetToken(t);
		lexer->ungetToken(tokId);
	} else
		lexer->ungetToken(t);
	const intrusive_ptr<const Expression> exp1 = parseExpression();
	t = lexer->getToken(status);
	if (t.getType()==TT_COMMA || t.getType()==TT_CLOSED_SQUARE) {
		vector<intrusive_ptr<const Expression> > exps;
		exps.push_back(exp1);
		while (t.getType()==TT_COMMA) {
			exps.push_back(parseExpression());
			t = lexer->getToken(status);
		}
		lexer->ungetToken(t);
		const Token tokClosedSquare = expecting(TT_CLOSED_SQUARE);
		return intrusive_ptr<const Expression>(new NonEmptyListExpression(tokOpenSquare.getBegin(), tokClosedSquare.getEnd(), exps));
	} else
		lexer->ungetToken(t);
	expecting(TT_PIPE);
	const Token tokId = expecting(TT_IDENTIFIER);
	expecting(TT_IN);
	const intrusive_ptr<const Expression> exp2 = parseEqualityExpression();
	intrusive_ptr<const Expression> exp3;
	t = lexer->getToken(status);
	if (t.getType()==TT_AND) {
		exp3 = parseExpression();
	} else
		lexer->ungetToken(t);
	const Token tokClosedSquare = expecting(TT_CLOSED_SQUARE);
	return intrusive_ptr<const Expression>(new ExpSuchThatIdInExpAndExp(tokOpenSquare.getBegin(), tokClosedSquare.getEnd(), exp1, tokId.lexeme(), exp2, exp3));
}

intrusive_ptr<const Expression> Parser::parseAdditiveExpression() {
	const intrusive_ptr<const Expression> firstExp = parseMultiplicativeExpression();
	vector<intrusive_ptr<const Expression> > exps;
	vector<TokenType> ops;
	exps.push_back(firstExp);
	for(;;) {
		const Token t = lexer->getToken(status);
		const TokenType type = t.getType();
		if (type!=TT_PLUS && type!=TT_MINUS) {
			lexer->ungetToken(t);
			break;
		}
		ops.push_back(type);
		exps.push_back(parseMultiplicativeExpression());
	}
	const size_t size = exps.size();
	if (size==1)
		return firstExp;
	if (size==2) {
		switch (ops.front()) {
		case TT_PLUS:
			return intrusive_ptr<const Expression>(new SumExpression(exps[0], exps[1]));
		case TT_MINUS:
			return intrusive_ptr<const Expression>(new SubtractionExpression(exps[0], exps[1]));
		default:
			assert(false);
		}
	}
	return intrusive_ptr<const Expression>(new SummationExpression(exps, ops));
}

intrusive_ptr<const Expression> Parser::parseMultiplicativeExpression() {
	const intrusive_ptr<const Expression> firstExp = parsePowerExpression();
	bool usedStarOrModAfterSlashOrColon=false;
	int nSlashesOrColons=0;
	vector<intrusive_ptr<const Expression> > exps;
	vector<TokenType> ops;
	exps.push_back(firstExp);
	for(;;) {
		const Token t = lexer->getToken(status);
		const TokenType type = t.getType();
		if (type==TT_IDENTIFIER) {
			if (!lexer->getStatus().isInCoCoA4CompatibilityMode())
				throw UnexpectedTokenException("Unexpected identifier, are you forgetting a \"*\" or a \";\" ?", t);
			lexer->ungetToken(t);
			ops.push_back(TT_STAR);
			exps.push_back(parseLimitedPowerExpression());
			usedStarOrModAfterSlashOrColon = nSlashesOrColons>0;
			continue;
		}
		switch(type) {
		case TT_STAR:
			usedStarOrModAfterSlashOrColon = nSlashesOrColons>0;
			break;
		case TT_SLASH:
			++nSlashesOrColons;
			break;
		case TT_MOD:
			usedStarOrModAfterSlashOrColon = nSlashesOrColons>0;
			break;
		case TT_COLON:
			++nSlashesOrColons;
			break;
		default:
			lexer->ungetToken(t);
			goto longBreak; // I'm neither ashamed nor afraid of using an "evil" goto in situations like this one
		}
		ops.push_back(type);
		exps.push_back(parsePowerExpression());
	}
longBreak:
	const size_t size = exps.size();
	if (size==1)
		return firstExp;
	if (nSlashesOrColons>1)
		lexer->reportWarning("Using two or more operators \"/\" or \":\" at the same level is potentially ambiguous; please parenthesize", firstExp->beginSourcePosition, exps.back()->endSourcePosition, WS_NORMAL);
	if ( usedStarOrModAfterSlashOrColon )
		lexer->reportWarning("Using operators \"*\" or \"%\" after \"/\" or \":\" is potentially ambiguous; please parenthesize", firstExp->beginSourcePosition, exps.back()->endSourcePosition, WS_NORMAL);
	if (size==2) {
		switch (ops.front()) {
		case TT_STAR:
			return intrusive_ptr<const Expression>(new ProductExpression(exps[0], exps[1]));
		case TT_SLASH:
			return intrusive_ptr<const Expression>(new DivisionExpression(exps[0], exps[1]));
		case TT_MOD:
			return intrusive_ptr<const Expression>(new ModuloExpression(exps[0], exps[1]));
		case TT_COLON:
			return intrusive_ptr<const Expression>(new ColonExpression(exps[0], exps[1]));
		default:
			assert(false);
		}
	}
	return intrusive_ptr<const Expression>(new MultiplicationExpression(exps, ops));
}

intrusive_ptr<const Expression> Parser::parsePowerExpression() {
	intrusive_ptr<const Expression> returnValue = parseUnaryExpression();
	for(;;) {
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_POWER) {
			lexer->ungetToken(t);
			break;
		}
		const intrusive_ptr<const Expression> exp = parseUnaryExpression();
		returnValue = intrusive_ptr<const Expression>(new PowerExpression(returnValue, exp));
	}
	return returnValue;
}

intrusive_ptr<const Expression> Parser::parseSelectors(intrusive_ptr<const Expression> targetExp) {
	intrusive_ptr<const Expression> returnValue = targetExp;
	for(;;) {
		const Token t = lexer->getToken(status);
		const TokenType type=t.getType();
		if (type!=TT_DOT && type!=TT_OPEN_SQUARE && type!=TT_OPEN_ROUND) {
			lexer->ungetToken(t);
			return returnValue;
		}
		switch(type) {
		case TT_DOT:
			{
			returnValue = intrusive_ptr<const Expression>(new FieldAccessExpression(returnValue, parseIdOrVar()));
			}
			break;
		case TT_OPEN_SQUARE:
			{
			const vector<intrusive_ptr<const Expression> > eList=parseExpressionList();
			const Token tokClosedBracket = expecting(TT_CLOSED_SQUARE);
			returnValue = intrusive_ptr<const Expression>(new IndexedAccessExpression(returnValue, eList, tokClosedBracket.getEnd()));
			}
			break;
		case TT_OPEN_ROUND:
			{
				vector<intrusive_ptr<const Expression> > args;
				const Token t = lexer->getToken(status);
				if (t.getType()==TT_ELLIPSIS) {
					const Token tokClosedRound = expecting(TT_CLOSED_ROUND);
					returnValue = intrusive_ptr<const Expression>(new InvocationExpression(returnValue, true, args, tokClosedRound.getEnd()));
				} else {
					lexer->ungetToken(t);
					if (t.getType()!=TT_CLOSED_ROUND)
						args = parseExpressionList();
					const Token tokClosedRound = expecting(TT_CLOSED_ROUND);
					returnValue = intrusive_ptr<const Expression>(new InvocationExpression(returnValue, false, args, tokClosedRound.getEnd()));
				}
			}
			break;
		default:
			assert(false);
			throw FailedAssertionException("Parser::parseSelectors unknown token type");
		}
	}
}

intrusive_ptr<const Expression> Parser::parseLimitedPowerExpression() {
	const intrusive_ptr<const Expression> idOrVar = parseIdOrVar();
	intrusive_ptr<const Expression> returnValue = parseSelectors(idOrVar);
	for(;;) {
		const Token t = lexer->getToken(status);
		if (t.getType()!=TT_POWER) {
			lexer->ungetToken(t);
			return returnValue;
		}
		const intrusive_ptr<const Expression> exp = parseUnaryExpression();
		returnValue = intrusive_ptr<const Expression>(new PowerExpression(returnValue, exp));
	}
}

intrusive_ptr<const RecordExpression> Parser::parseRecord(const Token &tokRecord) {
	vector<RecordField> fields;
	Token t = expecting(TT_OPEN_SQUARE);
	t = lexer->getToken(status);
	if (t.getType()!=TT_CLOSED_SQUARE) {
		lexer->ungetToken(t);
		for(;;) {
			RecordField f;
			f.name = parseIdOrVar();
			t = lexer->getToken(status);
			TokenType tt = t.getType();
			if (tt==TT_EQUAL)
				lexer->reportWarning("You should use \":=\" instead of \"=\" inside records", t.getBegin(), t.getEnd(), WS_NORMAL);
			else if (tt!=TT_ASSIGN)
				throw UnexpectedTokenException("Expected \":=\" before the field initialization expression", t);
			f.initExp = parseExpression();
			fields.push_back(f);
			t = lexer->getToken(status);
			tt = t.getType();
			if (tt==TT_CLOSED_SQUARE)
				break;
			if (tt!=TT_COMMA)
				throw UnexpectedTokenException("Expected a comma or a closed square bracket to end the record definition", t);
		}
	}
	return intrusive_ptr<const RecordExpression>(new RecordExpression(tokRecord.getBegin(), fields, t.getEnd()));
}

intrusive_ptr<const Expression> Parser::parsePrimary(Token t) {
	Token tokPackagename=Token::EndOfFile;
	bool thereIsPackagename=false;
	switch (t.getType()) {
	case TT_RECORD:
		return parseRecord(t);
	case TT_OPEN_SQUARE:
		return parseListPrimary(t);
	case TT_OPEN_ROUND:
		{
			const intrusive_ptr<const Expression> exp = parseExpression();
			const Token tokClosedRound = expecting(TT_CLOSED_ROUND);
			return intrusive_ptr<const Expression>(new ParenthesizedExp(exp, t.getBegin(), tokClosedRound.getEnd()));
		}
	case TT_VAR:
		{
			expecting(TT_OPEN_ROUND);
			const intrusive_ptr<const Expression> exp = parseExpression();
			const Token tokClosedRound = expecting(TT_CLOSED_ROUND);
			lexer->reportWarningMayBeRemoved("The keyword Var", t.getBegin(), tokClosedRound.getEnd());
			intrusive_ptr<const Expression> var(new Var(exp, t.getBegin(), tokClosedRound.getEnd()));
			Token maybeColonColon = lexer->getToken(status);
			if (maybeColonColon.getType()!=TT_COLONCOLON) {
				lexer->ungetToken(maybeColonColon);
				return var;
			}
			Token nextToken = lexer->getToken(status);
			return intrusive_ptr<const Expression>(new ScopedExpression(var, parsePrimary(nextToken)));
		}
	case TT_ISDEFINED:
		{
			Token tokOpenRound = expecting(TT_OPEN_ROUND);
			intrusive_ptr<const Expression> exp = parseExpression();
			Token tokClosedRound = expecting(TT_CLOSED_ROUND);
			return intrusive_ptr<const Expression>(new IsDefinedExp(exp, tokOpenRound.getBegin(), tokClosedRound.getEnd()));
		}
	case TT_PACKAGENAME:
		tokPackagename = t;
		thereIsPackagename = true;
		expecting(TT_DOT);
		t = expecting(TT_IDENTIFIER);
		// no break here, we're falling through
	case TT_ORD:
	case TT_ELIM: {
		Token openRound = expecting(TT_OPEN_ROUND);
		lexer->ungetToken(openRound);
		return intrusive_ptr<const Expression>(new Identifier(t)); // a blatant lie that allows to invoke these two special functions
	}
	case TT_IDENTIFIER:
		{
			intrusive_ptr<const Expression> identifier(thereIsPackagename ?
					static_cast<Expression *>(new FullyQualifiedIdentifier(tokPackagename, t)) :
					static_cast<Expression *>(new Identifier(t)));
			Token maybeColonColon = lexer->getToken(status);
			if (maybeColonColon.getType()!=TT_COLONCOLON) {
				lexer->ungetToken(maybeColonColon);
				return identifier;
			}
			Token nextToken = lexer->getToken(status);
			return intrusive_ptr<const Expression>(new ScopedExpression(identifier, parsePrimary(nextToken)));
		}
	case TT_INT_LITERAL:
	case TT_FLOAT_LITERAL:
	case TT_TRUE:
	case TT_FALSE:
		return intrusive_ptr<const Expression>(new LiteralExpressionTEMPORARY_FIXME(t.getBegin(), t.getEnd()));
	case TT_STRING_LITERAL: {
		CharPointer begin = t.getBegin();
		CharPointer end = t.getEnd();
		for(;;) {
			t = lexer->getToken(status);
			if (t.getType()!=TT_STRING_LITERAL) {
				lexer->ungetToken(t);
				break;
			}
			end = t.getEnd();
		}
		return intrusive_ptr<const Expression>(new LiteralExpressionTEMPORARY_FIXME(begin, end));
	}
	case TT_EOF:
		throw UnexpectedTokenException("Found End-of-File while looking for the start of an expression", t);
	default:
		lexer->ungetToken(t);
		throw UnexpectedTokenException("Invalid start of expression", t);
	}
}

void Parser::checkThereIsntAnotherUnaryPlusOrMinus(const CoCoA::Lexer::Token &tokFirstOp) {
	const Token t = lexer->getToken(status);
	lexer->ungetToken(t);
	const TokenType tt = t.getType();
	if (tt==TT_MINUS && tokFirstOp.getType()==TT_MINUS)
		lexer->reportError("Using two unary minus operators in a row is not allowed because it looks suspiciously like as a mistyped single-line comment; please parenthesize the inner expression or remove the blank(s) to make your intentions clear", tokFirstOp.getBegin(), t.getEnd());
	else if (tt==TT_PLUS || tt==TT_MINUS)
		lexer->reportError("Using two unary operators in a row is not allowed; please check your expression and parenthesize its inner sub-expression if this is really what you want", tokFirstOp.getBegin(), t.getEnd());
}

intrusive_ptr<const Expression> Parser::parseUnaryExpression() {
	const Token t = lexer->getToken(status);
	switch (t.getType()) {
	case TT_PLUS:
		checkThereIsntAnotherUnaryPlusOrMinus(t);
		return intrusive_ptr<const Expression>(new UnaryPlusExpression(t.getBegin(), parseUnaryExpression()));
	case TT_MINUS:
		checkThereIsntAnotherUnaryPlusOrMinus(t);
		return intrusive_ptr<const Expression>(new UnaryMinusExpression(t.getBegin(), parseUnaryExpression()));
	case TT_NOT:
		return intrusive_ptr<const Expression>(new NotExpression(t.getBegin(), parseUnaryExpression()));
	default:
		return parseSelectors(parsePrimary(t));
	}
}

} // namespace Parser
} // namespace CoCoA
