//   Copyright (c)  2009  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/PPWithMask.H"
#include "CoCoA/PPMonoidEv.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingDistrMPoly.H"
#include "CoCoA/RingDistrMPolyInlFpPP.H"
#include "CoCoA/RingDistrMPolyInlPP.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/TmpPPVector.H"
#include "CoCoA/assert.H"  // for using TEST_ASSERT
#include "CoCoA/error.H"
#include "CoCoA/io.H"
#include "CoCoA/symbol.H"

using namespace CoCoA;

#include <algorithm>
using std::min;
#include <iostream>
using std::endl;
#include <vector>
using std::vector;

//----------------------------------------------------------------------
// Test for PPVector operations
// see also test-MonomialIdeal1
// functions: 
// environments: 
//----------------------------------------------------------------------

// Handy macro for making assertions.
#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)

ConstRefPPMonoidElem CsbSquareIndet(PPMonoid PPM, size_t l, size_t sq1, size_t sq2)
{
  CoCoA_ASSERT( l*l <= NumIndets(PPM) );
  CoCoA_ASSERT( sq1 <= l && sq2 <= l );
  return indet(PPM, (sq1-1)*l + (sq2-1));
}


void AppendQueenMovesFrom(PPVector& PPV, size_t Csb, size_t sq1, size_t sq2)
{
  ConstRefPPMonoidElem x = CsbSquareIndet(PPM(PPV), Csb, sq1, sq2);
  for ( size_t i=sq2+1 ; i<=Csb ; ++i )
    PushBack(PPV, x * CsbSquareIndet(PPM(PPV), Csb, sq1, i));
  for ( size_t i=sq1+1 ; i<=Csb ; ++i )
    PushBack(PPV, x * CsbSquareIndet(PPM(PPV), Csb, i, sq2));
  for ( size_t i=min(Csb-sq1,Csb-sq2) ; i>0 ; --i )
    PushBack(PPV, x * CsbSquareIndet(PPM(PPV), Csb, sq1+i, sq2+i));
  for ( size_t i=min(Csb-sq1, sq2-1) ; i>0 ; --i )
    PushBack(PPV, x * CsbSquareIndet(PPM(PPV), Csb, sq1+i, sq2-i));
}


void QueenPPVector(PPVector& PPV, size_t Csb)
{
  for ( size_t sq1=1 ; sq1<=Csb ; ++sq1 )
    for ( size_t sq2=1 ; sq2<=Csb ; ++sq2 )
      AppendQueenMovesFrom(PPV, Csb, sq1, sq2);
}


// ----------------------------------------------------------------------

void program()
{
  GlobalManager CoCoAFoundations;

  int N = 7;
  PPOrdering ord = NewLexOrdering(N*N);
  PPMonoid PPM1 = NewPPMonoidEv(SymbolRange("x", 0, NumIndets(ord)-1), ord);
  DivMaskRule DMR1 = NewDivMaskEvenPowers();

  PPVector PPV(PPM1, DMR1);
  QueenPPVector(PPV, N);
  interreduce(PPV);


  const std::vector<PPMonoidElem> x = indets(PPM1);
  PPVector f(PPM1, DMR1);
  PPVector g(PPM1, DMR1);
  PPVector h(PPM1, DMR1);
  PushBack(f, x[1]*x[2]);  PushBack(f, x[3]*x[2]);
  PushBack(g, x[1]*x[5]);  PushBack(g, x[6]*x[7]);
  
  lcms(h, f, g);

  ring P = NewPolyRing(RingQ(), NumIndets(PPM1));
  std::vector<RingElem> v;
  convert(v, P, h);
  PPVector l(PPM1, DMR1);
  convert(l, v);

  TEST_ASSERT(l.mySize() == h.mySize());
  TEST_ASSERT(PPM(l) == PPM(h));
  TEST_ASSERT(owner(PP(l[0])) == owner(PP(h[0])));
  for (size_t i=0; i<l.mySize() ; ++i)    TEST_ASSERT(l[i] == h[i]);
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
