//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingFpDouble.H"
#include "CoCoA/RingFpLog.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/ZZ.H"
#include "CoCoA/convert.H"
#include "CoCoA/io.H"
#include "CoCoA/ring.H"

using namespace CoCoA;

#include <exception>
using std::exception;
#include <iostream>
using std::endl;
#include<vector>
using std::vector;

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


void test(const QuotientRing& Fp)
{
  unsigned long p;
  TEST_ASSERT(IsConvertible(p, characteristic(Fp)));
  RingElem x(Fp);
  TEST_ASSERT(x == p);

  // We shall do some arithmetic spot checks; use Fibonacci numbers as "random" numbers.
  const size_t NumTrials = 100;
  vector<ZZ> fibonacci(NumTrials);
  fibonacci[0] = 0;
  fibonacci[1] = 1;
  for (size_t n=2; n < NumTrials; ++n)
    fibonacci[n] = fibonacci[n-1] + fibonacci[n-2];

  // This loop checks Fermat's little theorem: x^(p-1) = 1 mod p.
  // For speed check only for x values in the array fibonacci.
  for (size_t i=1; i < NumTrials; ++i)
  {
    RingElem x(Fp, fibonacci[i]);
    if (IsZero(x)) continue;
    TEST_ASSERT(power(x, p-1) == 1);
    if (p==2) continue;
    RingElem powx = power(x, (p-1)/2);
    TEST_ASSERT(powx == 1 || powx == -1);
  }

  // This variable will be handy in the loops below.
  ZZ preimage;
  // Do some spot checks on multiplication.
  for (size_t i=0; i < NumTrials; ++i)
  {
    for (size_t j=0; j < NumTrials; ++j)
    {
      ZZ prod = fibonacci[i]*fibonacci[j];
      RingElem ModProd = RingElem(Fp,fibonacci[i]) * RingElem(Fp,fibonacci[j]);
      TEST_ASSERT(IsInteger(preimage, ModProd));
      TEST_ASSERT(IsZero((prod-preimage)%p));
    }
  }

  // Do some spot checks on division.
  for (size_t i=0; i < NumTrials; ++i)
  {
    for (size_t j=0; j < NumTrials; ++j)
    {
      RingElem denom(Fp, fibonacci[j]);
      if (IsZero(denom)) continue;
      RingElem ModQuot = RingElem(Fp,fibonacci[i])/denom;
      TEST_ASSERT(IsInteger(preimage, ModQuot));
      TEST_ASSERT(IsZero((fibonacci[i] - fibonacci[j]*preimage)%p));
    }
  }
}


void program()
{
  GlobalManager CoCoAFoundations;

  const ring Z = RingZ();

  vector<unsigned long> p;
  p.push_back(2);
  p.push_back(3);
  p.push_back(101);
  p.push_back(257);
  p.push_back(1009);
  p.push_back(32749);        // Largest prime below 2^15.
  p.push_back(46337);        // Largest prime whose square is below 2^31.
  p.push_back(65521);        // Largest prime below 2^16.
  p.push_back(65537);        // First prime greater than 2^16.
  p.push_back(67108859);     // Upper limit for RingFpDouble (with IEEE doubles).
  p.push_back(67108879);     // First prime too big for RingFpDouble (with IEEE doubles).
  p.push_back(3037000493ul); // Largest prime whose square is below 2^63.
  p.push_back(4294967291ul); // Largest prime below 2^32.

  for (size_t i=0; i < p.size(); ++i)
  {
    ideal I = ideal(Z, p[i]);
    // NewRingFp ctor will fail if char is too big.
    try { test(NewRingFp(I)); }
    catch (const CoCoA::ErrorInfo& err) { if (err != ERR::BadSmallFpChar) throw; }

    // NewRingFpLog ctor will fail if char is too big.
    try { test(NewRingFpLog(I)); }
    catch (const CoCoA::ErrorInfo& err) { if (err != ERR::BadSmallFpChar) throw; }

    // NewRingFpDouble ctor will fail if char is too big.
    try { test(NewRingFpDouble(I)); }
    catch (const CoCoA::ErrorInfo& err) { if (err != ERR::BadSmallFpChar) throw; }

    test(NewQuotientRing(Z, I));
  }
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
