//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/CanonicalHom.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingDistrMPoly.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/io.H"
#include "CoCoA/ring.H"

using namespace CoCoA;

#include <iostream>
using std::endl;
#include <vector>
using std::vector;

//----------------------------------------------------------------------
// Test for RingHom
// Z->Q, Q->Q[x,y,z,t], Z->Fp, Q->Fpx, Q[x,y,z,t]->Fp[x], Fp[x]->Fp
// CompositeHom
//----------------------------------------------------------------------

void program()
{
  GlobalManager CoCoAFoundations(UseNonNegResidues);

  // Create the rings we will be using.
  const FractionField Q = RingQ(); // used to fail for non-uniqueness of Z
  const QuotientRing Fp = NewZmod(7);

  const PolyRing Qxyzt = NewPolyRing(Q, 4);
  const PolyRing Fpx = NewPolyRing(Fp, 1);

  // Some simple homomorphisms...
//   RingHom Z2Q = EmbeddingHom(Q);
//   RingHom Q2Qxyzt = CoeffEmbeddingHom(Qxyzt);
//   RingHom Z2Fp = QuotientingHom(Fp);
  RingHom Z2Q = CanonicalHom(RingZ(), Q);
  RingHom Q2Qxyzt = CanonicalHom(Q, Qxyzt);
  RingHom Z2Fp = CanonicalHom(RingZ(), Fp);
  RingHom Q2Fpx = InducedHom(Q, ZEmbeddingHom(Fpx)); // only partial!

  // A homomorphism from the polynomial ring Qxyzt to Fpx.
  vector<RingElem> v1;
  v1.push_back(zero(Fpx));       // x |-> 0
  v1.push_back(indet(Fpx,0));    // y |-> x in Fpx
  v1.push_back(RingElem(Fpx,1)); // z |-> 1 in Fpx
  v1.push_back(zero(Fpx));       // t |-> 0
  RingHom Qxyzt2Fpx = PolyRingHom(Qxyzt, Fpx, Q2Fpx, v1);

  // A homomorphism from Fpx to Fp.
  vector<RingElem> v2;
  v2.push_back(RingElem(Fp,-1)); // x |-> -1
  RingHom Fpx2Fp = PolyRingHom(Fpx, Fp, IdentityHom(Fp), v2);

  // Compose four homomorphisms
  RingHom CompositeHom = Fpx2Fp(Qxyzt2Fpx(Q2Qxyzt(Z2Q)));
  GlobalOutput() << "Composite homomorphism is " << CompositeHom << endl << endl;

  for (int i=1; i <= 10; ++i)
  {
    RingElem image = CompositeHom(i); // i is a machine int, and is mapped into Z automatically.
    GlobalOutput() << "Using composite hom: " << i << " maps to "
                   << image << " in " << owner(image) << endl;
  }

  GlobalOutput() << "Image of  3  under Z2Q       is " << Z2Q(3) << endl;
  GlobalOutput() << "Image of 3/5 under Q2Qxyzt   is " << Q2Qxyzt(RingElem(Q,3)/5) << endl;
  GlobalOutput() << "Image of 3/5 under Qxyzt2Fpx is " << Qxyzt2Fpx(RingElem(Qxyzt,3)/5) << endl;
  GlobalOutput() << "Image of 3/5 under Q2Fpx     is " << Q2Fpx(3/RingElem(Q,5)) << endl; // works fine
  try
  {
    GlobalOutput() << "Image of 1/7 under Q2Fpx is " << std::flush;
    GlobalOutput() << Q2Fpx(1/RingElem(Q,7)) << endl; // generates an error
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    if (err != ERR::BadRingHomArg2) throw; // rethrow if err is not the expected error.
    GlobalOutput() << "***" << endl
                   << "Program correctly detected improper use of a partial ring hom." << endl;
  }
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
