//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/ZZ.H"
#include "CoCoA/error.H"
#include "CoCoA/io.H"
#include "CoCoA/ring.H"

using namespace CoCoA;

#include <iostream>
using std::cout;
using std::endl;

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


void tautology(const RingElem& n)
{
  TEST_ASSERT(IsZero(n) ^ IsInvertible(n));
  ZZ N;
  TEST_ASSERT(IsInteger(N, n));
  TEST_ASSERT(n == N);
  TEST_ASSERT(N == n);
  RingElem n2(owner(n), N);
  TEST_ASSERT(n == n2);
  TEST_ASSERT(n2 == n);
}


void greater(const RingElem& n, int m)
{
  TEST_ASSERT(!(n == m));
  TEST_ASSERT(n != m);
  TEST_ASSERT(n >= m);
  TEST_ASSERT(n > m);
  TEST_ASSERT(!(n <= m));
  TEST_ASSERT(!(n < m));
  TEST_ASSERT(!(m == n));
  TEST_ASSERT(m != n);
  TEST_ASSERT(!(m >= n));
  TEST_ASSERT(!(m > n));
  TEST_ASSERT(m <= n);
  TEST_ASSERT(m < n);
}

void equal(const RingElem& n, int m)
{
  TEST_ASSERT(n == m);
  TEST_ASSERT(!(n != m));
  TEST_ASSERT(n >= m);
  TEST_ASSERT(!(n > m));
  TEST_ASSERT(n <= m);
  TEST_ASSERT(!(n < m));
  TEST_ASSERT(m == n);
  TEST_ASSERT(!(m != n));
  TEST_ASSERT(m >= n);
  TEST_ASSERT(!(m > n));
  TEST_ASSERT(m <= n);
  TEST_ASSERT(!(m < n));
}

void less(const RingElem& n, int m)
{
  TEST_ASSERT(!(n == m));
  TEST_ASSERT(n != m);
  TEST_ASSERT(!(n >= m));
  TEST_ASSERT(!(n > m));
  TEST_ASSERT(n <= m);
  TEST_ASSERT(n < m);
  TEST_ASSERT(!(m == n));
  TEST_ASSERT(m != n);
  TEST_ASSERT(m >= n);
  TEST_ASSERT(m > n);
  TEST_ASSERT(!(m <= n));
  TEST_ASSERT(!(m < n));
}


void greater(const RingElem& n, const ZZ& m)
{
  TEST_ASSERT(!(n == m));
  TEST_ASSERT(n != m);
  TEST_ASSERT(n >= m);
  TEST_ASSERT(n > m);
  TEST_ASSERT(!(n <= m));
  TEST_ASSERT(!(n < m));
  TEST_ASSERT(!(m == n));
  TEST_ASSERT(m != n);
  TEST_ASSERT(!(m >= n));
  TEST_ASSERT(!(m > n));
  TEST_ASSERT(m <= n);
  TEST_ASSERT(m < n);
}

void equal(const RingElem& n, const ZZ& m)
{
  TEST_ASSERT(n == m);
  TEST_ASSERT(!(n != m));
  TEST_ASSERT(n >= m);
  TEST_ASSERT(!(n > m));
  TEST_ASSERT(n <= m);
  TEST_ASSERT(!(n < m));
  TEST_ASSERT(m == n);
  TEST_ASSERT(!(m != n));
  TEST_ASSERT(m >= n);
  TEST_ASSERT(!(m > n));
  TEST_ASSERT(m <= n);
  TEST_ASSERT(!(m < n));
}

void less(const RingElem& n, const ZZ& m)
{
  TEST_ASSERT(!(n == m));
  TEST_ASSERT(n != m);
  TEST_ASSERT(!(n >= m));
  TEST_ASSERT(!(n > m));
  TEST_ASSERT(n <= m);
  TEST_ASSERT(n < m);
  TEST_ASSERT(!(m == n));
  TEST_ASSERT(m != n);
  TEST_ASSERT(m >= n);
  TEST_ASSERT(m > n);
  TEST_ASSERT(!(m <= n));
  TEST_ASSERT(!(m < n));
}


void arithmetic(RingElem n, const RingElem& m)
{
  ZZ N;
  TEST_ASSERT(IsInteger(N, n));
  ZZ M;
  TEST_ASSERT(IsInteger(M, m));
  TEST_ASSERT(n == N);
  TEST_ASSERT(m == M);

  // Some simple arithmetic
  n = n+m; N=N+M; TEST_ASSERT(n == N);
  n = n-m; N=N-M; TEST_ASSERT(n == N);
  n = n*m; N=N*M; TEST_ASSERT(n == N);
  n = n/m; N=N/M; TEST_ASSERT(n == N);

  n = n+M; N=N+M; TEST_ASSERT(n == N);
  n = n-M; N=N-M; TEST_ASSERT(n == N);
  n = n*M; N=N*M; TEST_ASSERT(n == N);
  n = n/M; N=N/M; TEST_ASSERT(n == N);

  n = N+m; N=N+M; TEST_ASSERT(n == N);
  n = N-m; N=N-M; TEST_ASSERT(n == N);
  n = N*m; N=N*M; TEST_ASSERT(n == N);
  n = N/m; N=N/M; TEST_ASSERT(n == N);

  // Checks for aliasing problems
  n = n+n; N=N+N; TEST_ASSERT(n == N);
  n = n*n; N=N*N; TEST_ASSERT(n == N);
  n = n/n; N=N/N; TEST_ASSERT(n == N);
  n = n-n; N=N-N; TEST_ASSERT(n == N);

  // assignment arithmetic
  N = 123456789;  n = N; TEST_ASSERT(n == N);
  n += m; N+=M; TEST_ASSERT(n == N);
  n -= m; N-=M; TEST_ASSERT(n == N);
  n *= m; N*=M; TEST_ASSERT(n == N);
  n /= m; N/=M; TEST_ASSERT(n == N);

  // Checks for aliasing problems
  n = m; n += n; N=M; N+=M; TEST_ASSERT(n == N);
  n = m; n -= n; N=M; N-=M; TEST_ASSERT(n == N);
  n = m; n *= n; N=M; N*=M; TEST_ASSERT(n == N);
  n = m; n /= n; N=M; N/=M; TEST_ASSERT(n == N);
}

void program()
{
  GlobalManager CoCoAFoundations;

  const ring Q = RingQ();

  TEST_ASSERT(IsFractionField(Q));
  FractionField FF = AsFractionField(Q);
  TEST_ASSERT(IsZ(BaseRing(FF)));
  TEST_ASSERT(BaseRing(FF) == RingZ());

  const ZZ Zero;
  RingElem n(Q);
  // Check an implementation detail
  RingElem m = n; TEST_ASSERT(m == n); TEST_ASSERT(raw(m).myRawPtr() != raw(n).myRawPtr());

  TEST_ASSERT(IsZero(n));
  TEST_ASSERT(!IsInvertible(n));
  TEST_ASSERT(!IsOne(n));
  TEST_ASSERT(!IsMinusOne(n));
  equal(n, 0);
  equal(n, Zero);
  greater(n, -1);
  greater(n, ZZ(-1));
  less(n, 1);
  less(n, ZZ(1));

  n = 1;
  TEST_ASSERT(!IsZero(n));
  TEST_ASSERT(IsInvertible(n));
  TEST_ASSERT(IsOne(n));
  TEST_ASSERT(!IsMinusOne(n));
  equal(n, 1);
  equal(n, ZZ(1));
  greater(n, -1);
  greater(n, ZZ(-1));
  less(n, 2);
  less(n, ZZ(2));

  n = -1;
  TEST_ASSERT(!IsZero(n));
  TEST_ASSERT(IsInvertible(n));
  TEST_ASSERT(!IsOne(n));
  TEST_ASSERT(IsMinusOne(n));
  equal(n, -1);
  equal(n, ZZ(-1));
  greater(n, -2);
  greater(n, ZZ(-2));
  less(n, 1);
  less(n, ZZ(1));

  n = power(RingElem(Q,10), 1000);
  TEST_ASSERT(IsInvertible(n));
  TEST_ASSERT(n == power(ZZ(10), 1000));

  tautology(RingElem(Q, 0));
  tautology(RingElem(Q, 1));
  tautology(RingElem(Q, -1));
  tautology(RingElem(Q, 32003));
  tautology(RingElem(Q, -32003));
  tautology(RingElem(Q, power(ZZ(17),99)));
  tautology(RingElem(Q, power(ZZ(-17),99)));

  // Check correct arithmetic.  Cases are small/large, positive/negative.
  arithmetic(RingElem(Q, 23), RingElem(Q, 37));
  arithmetic(RingElem(Q, 100), RingElem(Q, power(ZZ(10), 100)));
  arithmetic(RingElem(Q, power(ZZ(10), 100)), RingElem(Q, 100));
  arithmetic(RingElem(Q, power(ZZ(10), 100)), RingElem(Q, power(ZZ(2), 301)));

  arithmetic(RingElem(Q, 23), -RingElem(Q, 37));
  arithmetic(RingElem(Q, 100), -RingElem(Q, power(ZZ(10), 100)));
  arithmetic(RingElem(Q, power(ZZ(10), 100)), -RingElem(Q, 100));
  arithmetic(RingElem(Q, power(ZZ(10), 100)), -RingElem(Q, power(ZZ(2), 301)));

  arithmetic(-RingElem(Q, 23), RingElem(Q, 37));
  arithmetic(-RingElem(Q, 100), RingElem(Q, power(ZZ(10), 100)));
  arithmetic(-RingElem(Q, power(ZZ(10), 100)), RingElem(Q, 100));
  arithmetic(-RingElem(Q, power(ZZ(10), 100)), RingElem(Q, power(ZZ(2), 301)));

  arithmetic(-RingElem(Q, 23), -RingElem(Q, 37));
  arithmetic(-RingElem(Q, 100), -RingElem(Q, power(ZZ(10), 100)));
  arithmetic(-RingElem(Q, power(ZZ(10), 100)), -RingElem(Q, 100));
  arithmetic(-RingElem(Q, power(ZZ(10), 100)), -RingElem(Q, power(ZZ(2), 301)));

  // assignment arithmetic
  n = 355; n /= 113; // Aside: 355/113 is a famous approximation for pi.
  m = n;

  n += m;
  n -= m;
  n *= m;
  n /= m;

  n = m; n += n; TEST_ASSERT(n == 2*m);
  n = m; n -= n; TEST_ASSERT(IsZero(n));
  n = m; n *= n; TEST_ASSERT(n == m*m);
  n = m; n /= n; TEST_ASSERT(IsOne(n));

  // compute a large number
  n = power(m, 100);
  TEST_ASSERT(IsInvertible(n));
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA::Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
