//   Copyright (c)  2007,2008  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/MatrixViews.H"
#include "CoCoA/error.H"
#include "CoCoA/io.H"
#include "CoCoA/matrix.H"

using namespace CoCoA;
#include <algorithm>
using std::min;
#include <cstddef>
using std::size_t;
#include <iostream>
using std::endl;
#include <vector>
using std::vector;

//----------------------------------------------------------------------
// Test for matrix
// functions: IdentityMat, submat, BlockMat, ConcatVer, ConcatHor
//----------------------------------------------------------------------

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


void TestZeroMat(ring R)
{
  // A few quick checks on a (large) zero matrix.
  ConstMatrixView ZM = ZeroMat(R, 12345,67890);
  for (size_t i=0; i < NumRows(ZM); ++i)
    TEST_ASSERT(IsZeroRow(ZM, i));
  for (size_t j=0; j < NumCols(ZM); ++j)
    TEST_ASSERT(IsZeroCol(ZM, j));
  for (size_t i=0; i < min(NumRows(ZM), NumCols(ZM)); ++i)
    TEST_ASSERT(IsZero(ZM(i, i)));
}


void TestIdentityMat(ring R)
{
  // An IdentityMatrix is constant; you cannot change it, but you can copy it.
  ConstMatrixView Id = IdentityMat(R, 10);
  TEST_ASSERT(NumCols(Id) == 10);
  TEST_ASSERT(NumCols(Id) == NumRows(Id));
  // SetEntry(Id,1,1,1); <-- as expected, this does not compile because I10 is constant
  for (size_t i=0; i < NumRows(Id); ++i)
    for (size_t j=0; j < NumCols(Id); ++j)
    {
      if (i == j) TEST_ASSERT(IsOne(Id(i, j)));
      else TEST_ASSERT(IsZero(Id(i, j)));
    }

  // Create a submatrix view of I10; since the row and column
  // indices are the same lists, the result will be another
  // identity matrix.
  vector<size_t> RowIndices;
  RowIndices.push_back(2);
  RowIndices.push_back(7);
  RowIndices.push_back(1);
  RowIndices.push_back(8);
  vector<size_t> ColIndices = RowIndices;
  ConstMatrixView SubM = submat(Id, RowIndices, ColIndices);
  for (size_t i=0; i < NumRows(SubM); ++i)
    for (size_t j=0; j < NumCols(SubM); ++j)
    {
      if (i == j) TEST_ASSERT(IsOne(SubM(i, j)));
      else TEST_ASSERT(IsZero(SubM(i, j)));
    }
}


void TestBlockMat(ring R)
{
  // Create a 6x6 matrix M containing the numbers 0 to 35
  matrix M(NewDenseMat(R, 6, 6));
  for (size_t i=0; i < 6; ++i)
    for (size_t j=0; j < 6; ++j)
      SetEntry(M, i, j, 6*i+j);

  vector<size_t> R1, R2, R3;
  R1.push_back(0);
  R2.push_back(1); R2.push_back(2);
  R3.push_back(3); R3.push_back(4); R3.push_back(5);
  vector<size_t> C1 = R1;
  vector<size_t> C2 = R2;
  vector<size_t> C3 = R3;

  // Cut M up into 9 submatrices
  ConstMatrixView M11(submat(M, R1, C1));
  ConstMatrixView M12(submat(M, R1, C2));
  ConstMatrixView M13(submat(M, R1, C3));

  ConstMatrixView M21(submat(M, R2, C1));
  ConstMatrixView M22(submat(M, R2, C2));
  ConstMatrixView M23(submat(M, R2, C3));

  ConstMatrixView M31(submat(M, R3, C1));
  ConstMatrixView M32(submat(M, R3, C2));
  ConstMatrixView M33(submat(M, R3, C3));

  // Reassemble the submatrices into a single matrix (method 1)
  ConstMatrixView Mcopy1 = BlockMat(BlockMat(M11, M12, M21, M22),
                                    ConcatVer(M13, M23),
                                    ConcatHor(M31, M32),
                                    M33);

  // Reassemble the submatrices into a single matrix (method 2)
  ConstMatrixView Mcopy2 = BlockMat(M11,
                                    ConcatHor(M12, M13),
                                    ConcatVer(M21, M31),
                                    BlockMat(M22, M23, M32, M33));

  // Check that the reassembled matrices are equal to M.
  for (size_t i=0; i < NumRows(M); ++i)
    for (size_t j=0; j < NumCols(M); ++j)
    {
      TEST_ASSERT(M(i, j) == Mcopy1(i, j));
      TEST_ASSERT(M(i, j) == Mcopy2(i, j));
    }
}


void program()
{
  GlobalManager CoCoAFoundations;

  const ring Z = RingZ();

  TestZeroMat(Z);
  TestZeroMat(RingQ());
  TestZeroMat(NewZmod(2));

  TestIdentityMat(Z);
  TestIdentityMat(RingQ());
  TestIdentityMat(NewZmod(7));

  TestBlockMat(Z);

  matrix M(NewDenseMat(Z, 2, 3));
  SetEntry(M,1,0,3);
  SetEntry(M,1,1,1);
  SetEntry(M,1,2,4);

  TEST_ASSERT(IsZero(M(0,0)));

  TEST_ASSERT(M->myIsZeroRow(0));
  TEST_ASSERT(!M->myIsZeroRow(1));

  M->myAddRowMul(0, 1, RingElem(Z, 2));
  TEST_ASSERT(!M->myIsZeroRow(0));
  TEST_ASSERT(M(0,1) == 2);
  M->myAddRowMul(0, 1, M(0,0)); // alias test
  TEST_ASSERT(M(0,0) == 8*M(1,0));
  TEST_ASSERT(M(0,1) == 8*M(1,1));
  TEST_ASSERT(M(0,2) == 8*M(1,2));

  // Some simple tests on the transposed view of a matrix.
  MatrixView TrM = transpose(M);
  TEST_ASSERT(NumRows(TrM) == NumCols(M));
  TEST_ASSERT(NumRows(M) == NumCols(TrM));
  MatrixView TrTrM = transpose(TrM);
  for (size_t i = 0; i < NumRows(M); ++i)
    for (size_t j = 0 ; j < NumCols(M); ++j)
      TEST_ASSERT(TrTrM(i,j) == M(i,j));
  for (size_t i = 0; i < NumRows(M); ++i)
    for (size_t j = 0 ; j < NumCols(M); ++j)
      SetEntry(TrTrM,i,j, 2*M(i,j));
  for (size_t i = 0; i < NumRows(M); ++i)
    for (size_t j = 0 ; j < NumCols(M); ++j)
      TEST_ASSERT(TrTrM(i,j) == M(i,j));
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
