//   Copyright (c)  2005,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/SmallFpImpl.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/SmallPrime.H"
#include "CoCoA/error.H"
#include "CoCoA/convert.H"
#include "CoCoA/OpenMath.H"

#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits;

namespace CoCoA
{

  namespace // anonymous namespace
  {
    // JAA believes that this routine always returns a reduced value,
    // but does not have a proof of this currently.
    SmallFpImpl::value_t exgcd(SmallFpImpl::value_t r, SmallFpImpl::value_t m)
    {
      CoCoA_ASSERT(r != 0);  CoCoA_ASSERT(m != 0);
      SmallFpImpl::value_t p = m;
      SmallFpImpl::value_t cr = 1;
      SmallFpImpl::value_t cm = 0; // this is minus what you think it is!
      while (r != 0)
      {
	SmallFpImpl::value_t q = m/r;
	m -= q*r;
	cm += q*cr;
	if (m == 0) break;
	q = r/m;
	r -= q*m;
	cr += q*cm;
      }
      //???  if (r+m != 1) error("cannot invert zero-divisor");
      if (r == 0) return p-cm;
      return cr;
    }


    // This function computes least e such that 2^e > n.
    // I have preferred simplicity over speed here.
    unsigned short NumBits(unsigned long n)
    {
      size_t ans = 0;
      while (n > 0)
      {
        ++ans;
        n >>= 1;
      }
      return ans;
    }


  } // end of anonymous namespace


  // These three inline fns are used only in the ctors.
  inline SmallFpImpl::value_t SmallFpImpl::CalcRecip(value_t p)
  {
    const value_t MAX = numeric_limits<value_t>::max();
    return MAX/p; // A sort of reciprocal of p.
  }

  inline SmallFpImpl::value_t SmallFpImpl::CalcDrop(value_t p)
  {
    const value_t MAX = numeric_limits<value_t>::max();
    return p*(MAX/p/2); // Largest multiple of p <= MAX/2; exploits integer division.
  }

  inline SmallFpImpl::value_t SmallFpImpl::CalcIterLimit(value_t p)
  {
    const value_t MAX = numeric_limits<value_t>::max();
    return MAX/(p-1)/(p-1)/2; // Max no. of unreduced products you can sum without exceeding MAX/2.
  }


  SmallFpImpl::SmallFpImpl(unsigned long p):
      myModulus(CheckCtorArg(p)),
      myResiduesAreSymm(DefaultResiduesAreSymm()),
      myRecip(CalcRecip(myModulus)),
      myMulShift1(NumBits(myModulus)),
      myMulShift2(8*sizeof(value_t)-myMulShift1),
      myDrop(CalcDrop(myModulus)),
      myIterLimit(CalcIterLimit(myModulus))
  {}


  SmallFpImpl::SmallFpImpl(const ZZ& P):
      myModulus(CheckCtorArg(P)),
      myResiduesAreSymm(DefaultResiduesAreSymm()),
      myRecip(CalcRecip(myModulus)),
      myMulShift1(NumBits(myModulus)),
      myMulShift2(8*sizeof(value_t)-myMulShift1),
      myDrop(CalcDrop(myModulus)),
      myIterLimit(CalcIterLimit(myModulus))
  {}


  void SmallFpImpl::myDiv(value_t& lhs, value_t x, value_t y) const
  {
    CoCoA_ASSERT(x == x%myModulus);
    CoCoA_ASSERT(y == y%myModulus);
    CoCoA_ASSERT(!myIsZero(y));
    value_t recip = exgcd(y, myModulus);
    myMul(lhs, x, recip);
  }


  void SmallFpImpl::myPower(value_t& lhs, value_t x, unsigned long n) const
  {
    CoCoA_ASSERT(!myIsZero(x) || n != 0); // Cannot compute 0^0
    if (myIsZero(x)) { lhs = 0; return; }
    if (myIsOne(x)) { lhs = 1; return; }
    n %= myModulus - 1; // OK by fermat's little theorem.
    if (n == 0) { lhs = 1; return; }
    if (n == 1) { lhs = x; return; }
    // Here we know that n >= 2 and x is not 0 or 1.
    // Below is an iterative version of repeated squaring.
    unsigned long mask = 1;
    unsigned long quartern = n/4;
    while (mask <= quartern) mask <<= 1;
    value_t X = x; // copy x in case lhs aliases x
    lhs = x;
    for (; mask; mask >>= 1)
    {
      myMul(lhs, lhs, lhs);
      if (n & mask) myMul(lhs, lhs, X);
    }
  }


  void SmallFpImpl::myOutput(ostream& out, value_t x) const
  {
    CoCoA_ASSERT(x == x%myModulus);
    if (!myIsPrintedWithMinus(x))
      out << x;
    else
      out << long(x)-long(myModulus);
  }


  void SmallFpImpl::myOutput(OpenMathOutput& OMOut, value_t x) const
  {
    CoCoA_ASSERT(x == x%myModulus);
    OMOut << x;
  }


  // If p is a small prime, return p as a value_t.
  // Otherwise throw an exception.
  SmallFpImpl::value_t SmallFpImpl::CheckCtorArg(unsigned long p)
  {
    if (CalcRecip(p) < p || !IsSmallPrime(p) || CalcIterLimit(p) == 0)
      CoCoA_ERROR(ERR::BadSmallFpChar, "SmallFpImpl ctor");
    return p;
  }

  // Check that P is a small (positive) prime; otherwise throw.
  SmallFpImpl::value_t SmallFpImpl::CheckCtorArg(const ZZ& P)
  {
    // First convert to ulong, if successful then check value is small enough.
    unsigned long p;
    if (P <= 0 || !IsConvertible(p, P))
      CoCoA_ERROR(ERR::BadSmallFpChar, "SmallFpImpl ctor");
    return CheckCtorArg(p);
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SmallFpImpl.C,v 1.5 2009/12/11 11:46:32 abbott Exp $
// $Log: SmallFpImpl.C,v $
// Revision 1.5  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.4  2009/05/14 09:39:29  abbott
// Added possibility to specify "symmetric" or "non-negative" residues
// in quotients of ZZ.  Affects printing of elements in quotients of ZZ
// (also changed printing of elements in general quotient rings).
// Consequent changes in several tests.
//
// Revision 1.3  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.4  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.3  2006/11/27 13:06:22  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.2  2006/11/02 13:25:43  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.4  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.3  2005/10/12 15:52:09  cocoa
// Completed test-RingFp1 and corrected/cleaned the SmallFp*
// and RingFp* files.
//
// Some minor tidying elsewhere.
//
// Revision 1.2  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.3  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.2  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.4  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.3  2004/07/20 15:04:06  cocoa
// The next step in the new "ring element" conversion process:
// handling the case of creating a "const RefRingElem" object
// (since C++ refuses to do this properly itself).
//
// Revision 1.2  2004/07/14 16:40:42  cocoa
// Separated RingFpLog from its implementation which now resides in
// a new class: SmallFpLogImpl.  This is analogous to the change made
// to RingFp yesterday.
//
// Some tidying and other sundry minor changes.
//
// Revision 1.1  2004/07/13 16:32:26  cocoa
// First stage of major revamp of ring elements.
// Implementation of RingFp has been split into "ring interface"
// and "algorithms plus data structures".
//
//
