//   Copyright (c)  2005,2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/SmallFpLogImpl.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/SmallPrime.H"
#include "CoCoA/error.H"
#include "CoCoA/convert.H"
#include "CoCoA/OpenMath.H"

#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits; // only in SmallFpLogImpl ctor

namespace CoCoA
{

  // These two inline fns are used only in the ctors.
  inline SmallFpLogImpl::value_t SmallFpLogImpl::CalcDrop(value_t p)
  {
    const value_t MAX = numeric_limits<value_t>::max();
    return p*(MAX/p/2); // Largest multiple of p <= MAX/2; exploits integer division.
  }

  inline SmallFpLogImpl::value_t SmallFpLogImpl::CalcIterLimit(value_t p)
  {
    const value_t MAX = numeric_limits<value_t>::max();
    return MAX/(p-1)/(p-1)/2; // Max no. of unreduced products you can sum without exceeding MAX/2.
  }


  SmallFpLogImpl::SmallFpLogImpl(unsigned long p):
      myModulus(CheckCtorArg(p)),
      myResiduesAreSymm(DefaultResiduesAreSymm()),
      myDrop(CalcDrop(myModulus)),
      myIterLimit(CalcIterLimit(myModulus)),
      myRoot(FindPrimRoot(myModulus)),
      myLog(myModulus),
      myExp(2*myModulus-1)
  {
    myCtorBody();
  }

  SmallFpLogImpl::SmallFpLogImpl(const ZZ& P):
      myModulus(CheckCtorArg(P)),
      myResiduesAreSymm(DefaultResiduesAreSymm()),
      myDrop(CalcDrop(myModulus)),
      myIterLimit(CalcIterLimit(myModulus)),
      myRoot(FindPrimRoot(myModulus)),
      myLog(myModulus),
      myExp(2*myModulus-1)
  {
    myCtorBody();
  }


  void SmallFpLogImpl::myPower(value_t& lhs, value_t x, unsigned long n) const
  {
    CoCoA_ASSERT(!myIsZero(x) || n != 0); // Cannot compute 0^0
    if (n == 0 || myIsOne(x)) { lhs = 1; return; }
    if (n == 1 || myIsZero(x)) { lhs = x; return; }
    n %= myModulus-1;
    // At this point 0 <= n < myModulus-1 and import(x) != 0
    // (n*myLog[x]) cannot overflow, since square of myModulus-1 fits in a value_t (see ctor)
    lhs = myExp[(n*myLog[x])%(myModulus-1)];  // no risk of overflow in the product
  }


  void SmallFpLogImpl::myOutput(ostream& out, value_t x) const
  {
    CoCoA_ASSERT(x == x%myModulus);
    if (!myIsPrintedWithMinus(x))
      out << x;
    else
      out << long(x)-long(myModulus);
  }


  void SmallFpLogImpl::myOutput(OpenMathOutput& OMOut, value_t x) const
  {
    CoCoA_ASSERT(x == x%myModulus);
    OMOut << x;
  }


  // If p is a small prime, return p as a value_t.
  // Otherwise throw an exception.
  SmallFpLogImpl::value_t SmallFpLogImpl::CheckCtorArg(unsigned long p)
  {
    if (p > 65535 ||
        p > numeric_limits<FpTableElem>::max() ||
        !IsSmallPrime(p) ||
        CalcIterLimit(p) == 0)
      CoCoA_ERROR(ERR::BadSmallFpChar, "SmallFpLogImpl ctor");
    return p;
  }

  // Check that P is a small (positive) prime; otherwise throw.
  SmallFpLogImpl::value_t SmallFpLogImpl::CheckCtorArg(const ZZ& P)
  {
    // First convert to ulong, if successful then check value is small enough.
    unsigned long p;
    if (P <= 0 || !IsConvertible(p, P))
      CoCoA_ERROR(ERR::BadSmallFpChar, "SmallFpLogImpl ctor");
    return CheckCtorArg(p);
  }


  // NOTE: When this is called the Log/Exp tables are still empty.
  void SmallFpLogImpl::myCtorBody()
  {
    const value_t p = myModulus; // Convenient short-hand.

    // Build the log/exp tables.
    if (p == 2) // simplest to handle this special case separately
    {
      myLog[1] = 0;
      myExp[0] = myExp[1] = 1;
      return;
    }

    // From here on we know p is odd.
    const value_t p1 = p-1;
    const value_t half_p1 = (p-1)/2;
    const value_t ThreeHalves = p1 + half_p1;

    value_t s = 1;
    for (value_t i = 0; i < half_p1; ++i)
    {
      // All assignments to myLog and myExp truncate safely.
      myLog[s] = i;
      myLog[p-s] = i+half_p1;
      myExp[i] = s;
      myExp[i+half_p1] = p-s;
      myExp[i+p1] = s;
      myExp[i+ThreeHalves] = p-s;
      s = (s*myRoot)%p;
    }
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SmallFpLogImpl.C,v 1.5 2009/12/11 11:46:32 abbott Exp $
// $Log: SmallFpLogImpl.C,v $
// Revision 1.5  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.4  2009/05/14 09:39:29  abbott
// Added possibility to specify "symmetric" or "non-negative" residues
// in quotients of ZZ.  Affects printing of elements in quotients of ZZ
// (also changed printing of elements in general quotient rings).
// Consequent changes in several tests.
//
// Revision 1.3  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.2  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.4  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.3  2006/11/27 13:06:22  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.2  2006/11/02 13:25:43  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.5  2005/10/14 15:25:07  cocoa
// Major tidying and cleaning to small prime finite fields.
// Several consequential changes.  Improved their documentation.
//
// Added Makefile and script to include/CoCoA/ directory to
// keep library.H up to date.
//
// Revision 1.4  2005/10/12 15:52:09  cocoa
// Completed test-RingFp1 and corrected/cleaned the SmallFp*
// and RingFp* files.
//
// Some minor tidying elsewhere.
//
// Revision 1.3  2005/10/11 16:37:30  cocoa
// Added new small prime finite field class (see RingFpDouble).
//
// Cleaned makefiles and configuration script.
//
// Tidied PPMonoid code (to eliminate compiler warnings).
//
// Fixed bug in RingFloat::myIsInteger.
//
// Revision 1.2  2005/09/22 18:04:17  cocoa
// It compiles; the tests run OK.  The examples compile.
// No documentation -- the mindless eurocrats have rendered
// me mindless too.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.3  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.2  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.2  2004/07/16 15:45:12  cocoa
// First stage of new RingElem implementation completed.
//
// Revision 1.1  2004/07/14 16:40:42  cocoa
// Separated RingFpLog from its implementation which now resides in
// a new class: SmallFpLogImpl.  This is analogous to the change made
// to RingFp yesterday.
//
// Some tidying and other sundry minor changes.
//
//
//
